/* System headers */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>

#define MAIN_SOCKET_PATH "/tmp/.OroborOSX_main_socket"

int init_socket_connect(char *pathstr, char *myname)
{
    int s, len, dummy, sret;
    struct sockaddr_un remote;
    
    if ((s = socket(AF_UNIX,SOCK_STREAM,0))==-1) {
        perror("socket");
        return -1;
    }

    printf("keepalive info: Trying to connect to OroborOSX socket...\n");

    remote.sun_family=AF_UNIX;
    strcpy(remote.sun_path,pathstr);
    len = strlen(remote.sun_path) + sizeof(remote.sun_family) + 1;
    if(connect(s,(struct sockaddr *)&remote,len)==-1) {
        perror("connect");
        return -2;
    }

    printf("keepalive info: Connected. (%d)\n",s);
    
    len = strlen(myname);
    if(send(s,&len,4,0)==-1) {
        perror("keepalive: len send in init_socket_connect");
        return -3;
    }
    if(send(s,myname,len,0)==-1) {
        perror("keepalive: name send in init_socket_connect");
        return -4;
    }

    sret = fcntl(s,F_GETFL,dummy);
    sret |= O_NONBLOCK;
    dummy = fcntl(s,F_SETFL,sret);

    return s;
}

int send_bytes_to_socket(int s, int msglen, char *str)
{
    if(send(s,&msglen,4,0)==-1) {
        perror("keepalive: len send in send_bytes_to_socket");
        return -3;
    }
    
    if(msglen>0)
        if(send(s,str,msglen,0)==-1) {
            perror("keepalive: name send in send_bytes_to_socket");
            return -4;
        }
        
    return 0;
}

int check_listen_recv(int s)
{
    int n=0, rcvlen;
    
    n = recv(s, &rcvlen, 4, 0);
    if(n<0) {
        if(errno==EAGAIN)
            return -1;
        perror("keepalive: recv");	/* got some error */
        return n-2;
    } else if(n==0)
        return -2;	/* socket disappeared, so signal end */
    if(rcvlen==0)
        return -2;	/* client has indicated it wants to end */
    if(rcvlen<0)
        return rcvlen-2;	/* something weird going on here... */
        
    return rcvlen;
}

int get_bytes_from_socket(int s, int rcvlen, char *str)
{
    int n;
    
    while(n>0) {
        n = recv(s, str, rcvlen, 0);
        if( n < 0 ) {
            if( errno==EAGAIN) {
                printf("keepalive info: waiting for more... %d\n",n);
                n=1;
            } else {
                perror("keepalive: recv");
                return -1;
            }
        } else if(n==0) {
            fprintf(stderr,"keepalive: Lost connection...\n");
            perror("keepalive: recv");
            return -1;
        } else {
            if( n<rcvlen )
                printf("keepalive warning: Did not get all? %d %d\n",n,rcvlen);
            return n;
        }
    }
}

int check_listen_get_bytes(int s, char *str)
{
    int rcvlen, gotlen;
    
    rcvlen = check_listen_recv(s);
    if(rcvlen==0) {
        printf("Got zero...\n");
        return 0;
    } else if(rcvlen==-1) {
        return -1;
    } else if(rcvlen==-2) {
        return -2;
    } else if(rcvlen<0) {
        fprintf(stderr,"keepalive: check_listen_get_bytes got error from check_listen_recv: %d\n",rcvlen);
        return rcvlen-3;
    } else {
        //printf("Expecting %d bytes\n",rcvlen);
        gotlen = get_bytes_from_socket(s,rcvlen,str);
        if(gotlen!=rcvlen) {
            fprintf(stderr,"keepalive: Something went wrong: %d %d\n",rcvlen,gotlen);
            return -3;
        } else {
            str[rcvlen]=0;
        }
        return rcvlen;
    }
}


int main(int argc, char *argv[])
{
	char 	*display, *randomID, *socket_path;
	char	buf[100];
	int	oroborosx_socket=0, oroborosx_status=0;
	int	i, rcvlen, send_err;

	printf("keepalive info: started\n");

	randomID = getenv( "OROBOROSX_RANDOMID" );
        printf("keepalive info: Got OROBOROSX_RANDOMID = %s\n",randomID);

        socket_path = malloc(strlen(MAIN_SOCKET_PATH)+strlen(randomID)+5);
        sprintf(socket_path,"%s.%s",MAIN_SOCKET_PATH,randomID);

	oroborosx_socket = init_socket_connect(socket_path,"keepalive");
        free(socket_path);

	if( oroborosx_socket < 0 ) {
		fprintf(stderr,"keepalive: failed to connect to OroborOSX socket: %d\n",oroborosx_socket);
		return 1;
	}

	printf("keepalive info: connected to OroborOSX socket: %d\n",oroborosx_socket);

	display = getenv( "DISPLAY" );

	if(!display) {
		fprintf(stderr,"keepalive: failed to get DISPLAY\n");
		send_err = close(oroborosx_socket);
		return 2;
	}

        printf("keepalive info: got DISPLAY ('%s')\n",display);

	send_err = send_bytes_to_socket(oroborosx_socket,strlen(display),display);

	if( send_err < 0 ) {
		printf("keepalive: error sending DISPLAY: %d\n",send_err);
		send_err = close(oroborosx_socket);
		return 3;
	}

	printf("keepalive info: sent DISPLAY (%d)\n",send_err);

	i = 0;

	while( 1 ) {

		rcvlen = check_listen_get_bytes(oroborosx_socket,buf);
		if(rcvlen==0) {
			fprintf(stderr,"keepalive: Got zero from OroborOSX.\n");
			send_err = close(oroborosx_socket);
			return 4;
		} else if(rcvlen==-2) {
			printf("keepalive info: OroborOSX ended session. Finishing...\n");
			return 0;
		} else if(rcvlen==-1) { /* nothing in receive buffer */
			i++;
			if( i > 18 && oroborosx_status==0 ) {
				printf("keepalive warning: getting debug report from OroborOSX and XDarwin...\n");
				/* get debug report... */
				system("./gdbthem");
				i = 0;
			} else if( i > 14 ) {
				printf("keepalive warning: no contact with OroborOSX for %d seconds\n",i);
			} //else
				//printf("keepalive info: no contact with OroborOSX for %d seconds\n",i);
			rcvlen = sleep(1);
		} else if(rcvlen<0) {
			fprintf(stderr,"keepalive: Got an error from check_listen_get_bytes: %d\n",rcvlen);
			send_err = close(oroborosx_socket);
			return 5;
		} else if(rcvlen!=4) {
			buf[rcvlen] = 0;
			printf("keepalive warning: got something weird from OroborOSX (%d,%d)\n",rcvlen,buf[0]);
			i = 0;
		} else { /* got something... */
			buf[rcvlen] = 0;
			//printf("keepalive info: got something from OroborOSX (%d,%d)\n",rcvlen,buf[0]);
			i = 0;
			if( buf[0] == 111 ) {
				printf("keepalive info: OroborOSX is finishing...\n");
				send_err = close(oroborosx_socket);
				return 0;
			} else if( buf[0] == 100 ) {
				oroborosx_status = 1;
				//printf("keepalive info: OroborOSX sent inhibitor\n");
			} else {
				if( oroborosx_status == 1 ) {
					//printf("keepalive info: OroborOSX unhibited\n");
					oroborosx_status = 0;
				}
				if( buf[0] != 10 )
					printf("keepalive warning: got strange value from OroborOSX (%d)\n",buf[0]);
			}
		}

	}

	send_err = close(oroborosx_socket);

	return 0;
}

