/******************************************/ 
/* Carbon-compatible section of OroborOSX */
/* (c) Adrian Umpleby Jan 2002            */
/*                                        */
/* Original oroborus window manager code  */
/* (c) Ken Lynch Jan 2001                 */
/*                                        */
/* This file started life as main.c from  */
/* Apple's prototype Carbon application   */
/*                                        */
/* Distributed under the GPL              */ 
/* See LICENSE for more details           */ 
/******************************************/ 
 
/* macevents.c */

#include <string.h>
#include "main.h"
#include "hackcmds.h"
#include "communication.h"

int	idleCount=0, noIdleMax=3, wasWholeScreenMode=0, justSetWholeScreenMode=0, gotCommandDown=0, doingQuit=0;
long	iconbar_WaitTicks=100, gGotMenuClick=0, lastMovedTen=0, time_to_leave=0, count_time_start=0;
long	lastsentmods=0;
Cursor	cmdKeyCrsr;
OSErr OpenGeneralFile(FSSpecPtr spec, char *thePath, char *theExt);


static OSStatus LowRunAppleScript(const void* text, long textLength,
                                    AEDesc *resultData) {
    ComponentInstance theComponent;
    AEDesc scriptTextDesc;
    OSStatus err;
    OSAID scriptID, resultID;

        /* set up locals to a known state */
    theComponent = NULL;
    AECreateDesc(typeNull, NULL, 0, &scriptTextDesc);
    scriptID = kOSANullScript;
    resultID = kOSANullScript;

        /* open the scripting component */
    theComponent = OpenDefaultComponent(kOSAComponentType,
                    typeAppleScript);
    if (theComponent == NULL) { err = paramErr; goto bail; }

        /* put the script text into an aedesc */
    err = AECreateDesc(typeChar, text, textLength, &scriptTextDesc);
    if (err != noErr) goto bail;

        /* compile the script */
    err = OSACompile(theComponent, &scriptTextDesc,
                    kOSAModeNull, &scriptID);
    if (err != noErr) goto bail;

        /* run the script */
    err = OSAExecute(theComponent, scriptID, kOSANullScript,
                    kOSAModeNull, &resultID);

        /* collect the results - if any */
    if (resultData != NULL) {
        AECreateDesc(typeNull, NULL, 0, resultData);
        if (err == errOSAScriptError) {
            OSAScriptError(theComponent, kOSAErrorMessage,
                        typeChar, resultData);
        } else if (err == noErr && resultID != kOSANullScript) {
            OSADisplay(theComponent, resultID, typeChar,
                        kOSAModeNull, resultData);
        }
    }
bail:
    AEDisposeDesc(&scriptTextDesc);
    if (scriptID != kOSANullScript) OSADispose(theComponent, scriptID);
    if (resultID != kOSANullScript) OSADispose(theComponent, resultID);
    if (theComponent != NULL) CloseComponent(theComponent);
    printf("LowRunAppleScript returned: %d\n",err);
    return err;
}

    /* SimpleRunAppleScript compiles and runs the AppleScript in
    the c-style string provided as a parameter.  The result returned
    indicates the success of the operation. */
static OSStatus SimpleRunAppleScript(const char* theScript) {
    return LowRunAppleScript(theScript, strlen(theScript), NULL);
}

void OpenXDarwinPrefs()
{
    SimpleRunAppleScript(
        "tell application \"XDarwin\"\n"
	"   set visible of window named \"Preferences\" to true\n"
	"   set visible of window named \"Voorkeuren\" to true\n"
	"   set visible of window named \"Preferencias\" to true\n"
	"   set visible of window named \"Preferenze\" to true\n"
	"   set visible of window named \"Instllningar\" to true\n"
	"   set visible of window named \"Prfrences\" to true\n"
	"   set visible of window named \"Einstellungen\" to true\n"
//	"   set visible of window named \"tX-[\" to true\n"
//	"   set visible of window named \"X$\" to true\n"
	"   activate\n"
        "end tell\n");
    SendSimpleHackCommand(HACKCMD_SHOWINGPREFS,1);
}

void HideFolderExtensions()
{
    OpenGeneralFile(NULL,NULL,NULL);
    SimpleRunAppleScript(
        "tell application \"Finder\"\n"
	"set extension hidden of folder \"Themes.localized\" of folder \"OroborOSX\" of folder \"Preferences\" of folder \"Library\" of home to true\n"
	"set extension hidden of folder \"Launch Menu Items.localized\" of folder \"OroborOSX\" of folder \"Preferences\" of folder \"Library\" of home to true\n"
	"set extension hidden of folder \"Helper Scripts.localized\" of folder \"OroborOSX\" of folder \"Preferences\" of folder \"Library\" of home to true\n"
	"set extension hidden of folder \"Keymaps.localized\" of folder \"OroborOSX\" of folder \"Preferences\" of folder \"Library\" of home to true\n"
        "end tell\n");
}

static OSErr QuitAppleEventHandler( const AppleEvent *appleEvt, AppleEvent* reply, UInt32 refcon )
{
    int doquit=0;
    
    sprintf(lastsub,"TRACE1: QuitAppleEventHandler\n");
    subtrace();
//#ifdef SHOW_TRACE1
    printf("TRACE1: QuitAppleEventHandler\n");
//#endif

    if(alertWin==NULL) 
        doingQuit=0;
    else if(doingQuit==0) {
        // don't want to quit while an alert window is showing, unless it's the quit alert
        printf("OroborOSX info: got quit event while showing alert\n");
        SysBeep(10);
        return noErr;
    }
    
    if(doingQuit==1) {
        printf("OroborOSX info: got into quit event again...\n");
        gQuitFlag =  GOTQUITAE_ERRNO;
        return noErr;
    }
    
    if(launched_xdarwin==0) {
        doingQuit=1;
        Bring_OOSX_To_Front();	// should put a bouncing dock icon type notification here...
        // Then again, maybe OK, since if we're not front, user probably told us to quit via dock menu...?
        // But if the user says not to quit then the original front process should really come in front again...
        doquit=ShowAlert(QUITWM_NAME,1,1);	// confirmation dialog if we did not start XDarwin
    } else if(client_count==0)	// don't bother with confirmation if there's no clients
        doquit=1;
    else {
        doingQuit=1;
        Bring_OOSX_To_Front();	// should put a bouncing dock icon type notification here...
        doquit=ShowAlert(QUITX11_NAME,1,1);	// confirmation dialog if we started XDarwin
    }
    
    doingQuit=0;

    if(doquit) {
        printf("OroborOSX info: quitting...\n");
        Allow_HUPs(0);
        quit();
        gQuitFlag =  GOTQUITAE_ERRNO;
    }
    
    return noErr;
}

static OSErr RAPPAppleEventHandler( const AppleEvent *appleEvt, AppleEvent* reply, UInt32 refcon )
{
    sprintf(lastsub,"TRACE1: RAPPAppleEventHandler\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: RAPPAppleEventHandler\n");
#endif

#ifdef DEBUG
    printf("Got a 'rapp' AppleEvent\n");
#endif
    if(gInvisible!=0) {
        gInvisible=0;
        map_all_clients();
    }
    raise_all_clients();
    
    DestroyBadWindows((WindowRef) 0);
    
    return noErr;
}

char *GetExtensionFromPath(char *thePath)
{
    int ipos=0;
    char *theExt=NULL, *theName=NULL;
    
    // look for final slash
    theName = rindex(thePath,47);
    if(theName==NULL) theName=thePath;
    // look for final dot
    theExt = rindex(theName,46);
    // check for just a single dot
    if(theExt && strlen(theExt)==1 && !strncmp(theExt,".",1)) theExt = NULL;
    
    if(theExt) {
        char *lowExt = malloc(strlen(theExt)+4);
        // make it all lower case...
        for(ipos=0;ipos<strlen(theExt);ipos++)
            lowExt[ipos] = tolower(theExt[ipos]);
        lowExt[strlen(theExt)]=0;
        theExt = lowExt;
    }
    return theExt;
}

char *GetFullPathFromFSSpec(FSSpecPtr spec)
{
/*    CInfoPBRec	myPB;
    char 	*fullPath, dirName[200], *oneName=&spec.name;
    OSErr	myErr;
    int		i;*/
    char	*fullPath, *retval, *tmpstr;
    int		fullPathLen=512, retlen, i, j;
    CFURLRef	myURL;
    FSRef	fsref;
    OSErr	myErr;
    
    fullPath=malloc(520);
    myErr = FSpMakeFSRef(spec,&fsref);
    myURL = CFURLCreateFromFSRef(kCFAllocatorDefault,&fsref);
    CFURLGetFileSystemRepresentation(myURL,TRUE,fullPath,fullPathLen);
    if(fullPath==NULL) return NULL;
    retval = strdup(fullPath);
    retlen = strlen(fullPath);
    j = 0;
    for(i=0;i<=retlen;i++) {
        fullPath[j]=retval[i];
        j++;
        if(retval[i]==39) {
            if(j>515) {
                fprintf(stderr,"%s: Path name too long! %s\n",PROGNAME,retval);
                return NULL; }
            fullPath[j]=92;
            j++;
            fullPath[j]=39;
            j++;
            fullPath[j]=39;
            j++;
        }
    }
    free(retval);
    retval = strdup(fullPath);
    free(fullPath);
    return retval;

/*    fullPath = malloc(oneName[0]+2);
    fullPath[0] = '/';
    for(i=0;i<oneName[0];i++)
        fullPath[i+1] = oneName[i+1];
    myPB.ioNamePtr = dirName;
    myPB.ioVRefnum = spec.vRefnum;
    myPB.ioDrParID = mySpec.parID;
    myPB.ioFDirIndex = -1;
    do {
        int dirLen;
        char *tmpstr;
        myPB.ioDrDirID = myPB.ioDrParID;
        myErr = PBGetCatInfo(&myPB,False);
        dirLen = dirName[0];
        tmpstr = malloc(strlen(fullPath)+dirLen+2);
        tmpstr[0] = '/';
        for(i=1;i<=dirLen;i++)
            tmpstr[i] = dirName[i];
        for(i=0;i<strlen(fullPath);i++)
            tmpstr[i+dirLen+1] = fullPath[i];
        free(fullPath);
        fullPath = tmpstr;
    } while(myPB.ioDrDirID!=fsRtDirID);
    GetFullPath = fullPath;*/
}

void GetTypeCreatorFromFSSpec(FSSpecPtr spec,char *theType, char *theCr)
{
    OSErr	myErr;
    FInfo	fndrinfo;
    
    myErr = FSpGetFInfo(spec,&fndrinfo);
    *(OSType *) theType = fndrinfo.fdType;
    *(OSType *) theCr = fndrinfo.fdCreator;
    theType[4] = 0;
    theCr[4] = 0;
}

char *check_compressed(char *thePath, char *theExt)
{
    char *script=NULL, *filename=NULL, *clipPath=NULL;
    int err=0;
    
    if(theExt==NULL) return NULL;
    
    if( strlen(theExt)<2 ) return theExt;
    
    filename=rindex(thePath,47);
    if(filename==NULL)
        filename=thePath;

    clipPath=strdup(thePath);
    clipPath[strlen(thePath)-strlen(theExt)]=0;
    script=malloc(60+strlen(thePath)+strlen(theExt)+strlen(clipPath));
    
    // Ask expander script to check before overwriting
    sprintf(script,"./Contents/Scripts/FindExpander '%s' '%s' '%s' check",thePath,clipPath,theExt);
    
    printf("running: %s\n",script);
    err = system(script);
    
    if(err>=2560000) {
        fprintf(stderr,"%s: Got error %d from FindExpander script\n",PROGNAME,err);
        err=256;
    }
    
    if(err==5120) {
        int allowit=0;
        // The expander script noticed that the file already exists
        printf("%s info: FindExpander needs to know if file can be overwritten\n",PROGNAME);
        // Check with user
        allowit = ShowAlert("ExpandExists",1,1);
        if(allowit==1) {
            // run expander script again, but tell it to overwrite this time
            sprintf(script,"./Contents/Scripts/FindExpander '%s' '%s' '%s' force",thePath,clipPath,theExt);
            err = system(script);
        } else {
            free(clipPath);
            return (char *) -1; // YUK!
        }
    }
    
    free(script);
    free(clipPath);
    
    if(err) {
        if(err!=256) {
            fprintf(stderr,"%s: Got error %d from FindExpander script\n",PROGNAME,err);
            ShowAlert("ExpandFailed",1,0);
            return (char *) -1; // YUK!
        }
        printf("%s info: did not find an expander for this file - assuming none needed...\n",PROGNAME);
        thePath=NULL;
    } else {
        // strip off extension - this should now be the name of the expanded file
        thePath[strlen(thePath)-strlen(theExt)]=0;
    }
/*
    // Hardwired formats for the moment...
    if( strlen(theExt)==2 && !strncmp(theExt,".z",2) ) {
        char cmd[15];
        script=malloc(20+strlen(thePath));
        if( thePath[strlen(thePath)-1]==122 ) {	// lower case z - use unpack
            sprintf(cmd,"unpack");
        else {	// upper case Z - use uncompress
            sprintf(cmd,"uncompress");
        }
        sprintf(script,"%s '%s'",cmd,thePath);
        err = system(script);
        if(err) {
            fprintf(stderr,"%s: got error %d when trying to %s file \"%s\"\n",PROGNAME,err,cmd,filename);
            thePath=NULL;	// this only has local effect, within check_compressed
        } else
            thePath[strlen(thePath)-2]=0;
    } else if ( strlen(theExt)==3 && !strncmp(theExt,".gz",3) ) {
        script=malloc(20+strlen(thePath));
        sprintf(script,"gunzip '%s'",thePath);
        err = system(script);
        if(err) {
            fprintf(stderr,"%s: got error %d when trying to gunzip file \"%s\"\n",PROGNAME,err,filename);
            thePath=NULL;	// this only has local effect, within check_compressed
        } else
            thePath[strlen(thePath)-3]=0;
    } else if ( strlen(theExt)==4 && !strncmp(theExt,".uue",4) ) {
        script=malloc(20+strlen(thePath));
        sprintf(script,"uudecode '%s'",thePath);
        err = system(script);
        if(err) {
            fprintf(stderr,"%s: got error %d when trying to uudecode file \"%s\"\n",PROGNAME,err,filename);
            thePath=NULL;	// this only has local effect, within check_compressed
        } else
            thePath[strlen(thePath)-4]=0;
    } else
        return theExt;
*/
    if(thePath) return GetExtensionFromPath(thePath);
    // thePath was set to NULL if something went wrong, so we just return the standard extension
    return theExt;
}

OSErr OpenGeneralFile(FSSpecPtr spec, char *thePath, char *theExt)
{
    char *script=NULL, *theExt2=NULL, theType[6], theCr[6];
    int err, did_uncompress=0;
    
    if(thePath==NULL) {
        script = malloc(60);
        sprintf(script,"./Contents/Scripts/FindHelper check check check check");
        err = system(script);
        if(err) {
            fprintf(stderr,"%s: FindHelper failed with error %d\n",PROGNAME,err);
            ShowAlert(BADFINDHELPER_NAME,1,0);
        }
        free(script);
        return 0;
    }
    
    theExt2=check_compressed(thePath,theExt);
    
    if(theExt2==(char *)-1) {
        // special (YUKKY!) code for when the expansion did not happen (even though it found an expander)
        printf("%s info: expander aborted\n",PROGNAME);
        return -9999;
    } else if(theExt2!=theExt) {
        theExt = theExt2; // the original "theExt" will be freed in ODOCAppleEventHandler
        did_uncompress=1;
        // can't really deal with getting Type/Creator from compressed file
        theType[0] = 0;
        theCr[0] = 0;
    } else {
        GetTypeCreatorFromFSSpec(spec,theType,theCr);
    }

    if(gVerboseConsole)
        printf("%s info: got ext,type,creator: %s '%s' '%s'\n",PROGNAME,theExt,theType,theCr);
    {
        char *ttt=malloc(strlen(thePath)+15);
        sprintf(ttt,"ls -l '%s'",thePath);
        system(ttt);
        free(ttt);
    }
    if(theExt) {
        script = malloc(65+strlen(thePath)+strlen(theExt));
        sprintf(script,"./Contents/Scripts/FindHelper '%s' '%s' '%s' '%s' '%s'",
                                            thePath,theExt,theType,theCr,Display_Name);
    } else {
        script = malloc(strlen(thePath)+65);
        sprintf(script,"./Contents/Scripts/FindHelper '%s' '' '%s' '%s' '%s'",
                                                    thePath,theType,theCr,Display_Name);
    }
    err = system(script);
/*        if(err==NOTEXIST_STAT)
            ShowAlert(NOTEXISTLAUNCH_NAME,1,0);
        else {
            printf("%s warning: Script returned with error: %d\n",PROGNAME,err);
            ShowAlert(WARNLAUNCH_NAME,1,0); }*/
    if(err==NOHELPERSMATCH_STAT) {
        ShowAlert(NOHELPERSMATCH_NAME,1,0);
    } else if(err==GOTX11EXEC_STAT) {
        printf("%s info: Running an X11 executable\n",PROGNAME);
    } else if(err==NOFILEINFO_STAT) {
        ShowAlert(NOFILEINFO_NAME,1,0);
    } else if(err) {
        fprintf(stderr,"%s: FindHelper failed with error %d\n",PROGNAME,err);
	ShowAlert(BADFINDHELPER_NAME,1,0);
    } else {
        free(script);
        if(gVerboseConsole)
            printf("%s info: FindHelper worked...\n",PROGNAME);
        script = malloc(90+strlen(thePath)+strlen(Display_Name));
        sprintf(script,"./Contents/Scripts/RunHelpers '%s' '%s' < /tmp/oroborosx-%s.dir/helperlist",
                                                                        Display_Name,thePath,randomID);
        err = system(script);
        if(err==ALLHELPERSFAILED_STAT) {
            ShowAlert(ALLHELPERSFAILED_NAME,1,0);
        } else if(err) {
            fprintf(stderr,"%s: helper script returned error %d\n",PROGNAME,err);
            ShowAlert(BADRUNHELPERS_NAME,1,0);
        }
    }
    free(script);
    if(did_uncompress) free(theExt2);
    return 0;
}

static OSErr ODOCAppleEventHandler( const AppleEvent *appleEvt, AppleEvent* reply, UInt32 refcon )
{
    OSType	HLEclass, HLEmsgID;
    OSErr	myErr;
    long	i, ni;
    int		namelen;
    DescType	typeCode;
    Size	dataSize;
    AEDescList	myDesc, newDesc;
    AEKeyword	theAEKeyword;
    Boolean	changed, dummy;
    FSSpec	mySpec;
    AliasHandle	myAlias;
    pstr_t	*pStr;
    char	*cStr=NULL, *thePath=NULL;
    
    sprintf(lastsub,"TRACE1: ODOCAppleEventHandler\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: ODOCAppleEventHandler\n");
#endif

    HLEmsgID = appleEvt->descriptorType;
    HLEclass = reply->descriptorType;
    
    myErr=AECountItems(appleEvt,&ni);
    
    if(gVerboseConsole)
        printf("  Number of initial items: %d\n",ni);
    
    if(ni>0) {
        myErr=AESizeOfNthItem(appleEvt, 1, &typeCode, &dataSize);
        if(gVerboseConsole)
            printf("  typeCode,dataSize: %d,%d\n",typeCode,dataSize);
        myErr=AEGetNthDesc(appleEvt,1,typeCode,&theAEKeyword,&myDesc);
        HLEclass=(OSType) theAEKeyword;
        HLEmsgID=myDesc.descriptorType;
        if(gVerboseConsole)
            printf("  HLEclass,HLEmsgID: %d,%d\n",HLEclass,HLEmsgID);
        myErr=AECountItems(&myDesc,&ni);
        if(gVerboseConsole)
            printf("  Number of actual items: %d\n",ni);
        
        for(i=1;i<=ni;i++) {
            myErr=AESizeOfNthItem(&myDesc,i,&typeCode,&dataSize);
            if(gVerboseConsole)
                printf("  item %d   typeCode,dataSize: %#lx,%#lx\n",i,typeCode,dataSize);
            myErr=AEGetNthDesc(&myDesc,i,typeCode,&theAEKeyword,&newDesc);
            myAlias=(AliasHandle) newDesc.dataHandle;
            if(gVerboseConsole)
                printf("  item %d   myAlias: %#lx\n",i,myAlias);
            myErr=ResolveAlias(NULL,myAlias,&mySpec,&changed);
            if(myErr==noErr) {
                char *theExt=NULL;
// Use the horrible hack to get C string out of a Pascal string
// -there's probably a macro somewhere to do this 'legally'...
                pStr=&(mySpec.name);
                namelen=pStr->len;
                cStr=&(pStr->thestr);
                cStr[namelen]=0;
                thePath = GetFullPathFromFSSpec(&mySpec); // *Allocated* -must be freed
                theExt = GetExtensionFromPath(thePath); // *Allocated!* -since the case may have been changed from end of thePath
                if(gVerboseConsole)
                    printf("  item %d: %s\n",i,thePath);
                if(theExt && strlen(theExt)==7 && !strncmp(theExt,".x11app",7)) {
                    // Now look for that item in the Launch menu, and run the associated script
                    myErr=RunX11Script(0,0,thePath);
                    //myErr=LaunchMenuItem(cStr,1,0,0);
                    if(myErr!=0)
                        printf("%s warning: could not launch item %s -got error %d\n",PROGNAME,cStr,myErr);
                } else if(theExt && strlen(theExt)==10 && !strncmp(theExt,".x11helper",10)) {
                    cStr = malloc(20+strlen(thePath));
                    sprintf(cStr,"open -a TextEdit '%s'",thePath);
                    myErr = system(cStr);
                    if( myErr ) 
                        fprintf(stderr,"%s: got error %d when trying to open x11helper with TextEdit!\n",PROGNAME,myErr);
                    else {
                        printf("%s info: opening .x11helper script with TextEdit\n",PROGNAME);
                        sprintf(cStr,"sleep 1 ; open -a TextEdit &");
                        system(cStr);
                    }
                    free(cStr);
                } else {
                    myErr = OpenGeneralFile(&mySpec,thePath,theExt);
                    //printf("%s warning: need to get extension, type, full path and send to helper script...\n",PROGNAME);
                }
                free(thePath);
                if(theExt) free(theExt);
            } else
                printf("  item %d  got an error: %d\n",i,myErr);
        }
    }
    
    return noErr;
}

void InitAEHandlers()
{
    int	err;
    
// Install the quit AppleEvent handler
    err = AEInstallEventHandler( kCoreEventClass, kAEQuitApplication, NewAEEventHandlerUPP((AEEventHandlerProcPtr)QuitAppleEventHandler), 0, false );
    if (err != noErr) {
        ClearUpSockets();
        ExitToShell();
    }
    
// Install the rapp AppleEvent handler
    err = AEInstallEventHandler( kCoreEventClass, kAEReopenApplication, NewAEEventHandlerUPP((AEEventHandlerProcPtr)RAPPAppleEventHandler), 0, false );
    if (err != noErr) {
        ClearUpSockets();
        ExitToShell();
    }

// Install the odoc AppleEvent handler
    err = AEInstallEventHandler( kCoreEventClass, kAEOpenDocuments, NewAEEventHandlerUPP((AEEventHandlerProcPtr)ODOCAppleEventHandler), 0, false );
    if (err != noErr) {
        ClearUpSockets();
        ExitToShell();
    }
        
}

/*
 * Glue code to set command-key-shortcut cursor from X11
 */
void SetCmdKeyCursor(int on)
{
    if(on) {
        SetCursor(&cmdKeyCrsr);
    } else
        InitCursor();
}

void EventLoop()
{
    int		moving;
    Point	mspt;
    int		dummy;
    Boolean	gotEvent;
    EventRecord	event;
    char	script[100];
        
    sprintf(lastsub,"TRACE1: EventLoop\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: EventLoop\n");
#endif

    gQuitFlag = 0;

// Make sure we're all in front once started...
    Restart_XDarwin();
    Bring_To_Front(0);
    Stop_XDarwin();

    idleCount=0;
    noIdleMax=2;
// Leave longer gaps between background mouse tracking if tracers are on
#ifdef SHOW_TRACE1
    noIdleMax=10;
#endif
#ifdef SHOW_TRACE2
    noIdleMax=20;
#endif
#ifdef SHOW_TRACE3
    noIdleMax=50;
#endif

// all ready, so make sure the idle loop is called, and that HUPs are allowed...
    gInited=1;
    //Allow_HUPs(0);
/*    
    dummy = check_for_keepalive();
    if(dummy<0) {
        fprintf(stderr,"%s: error waiting for keepalive (%d)",PROGNAME,dummy);
        exit(1);
    } else if(dummy==0)*/
        Allow_HUPs(0);

    SendSimpleHackCommand(HACKCMD_ALLOWFRONT,0);

// xinerama should have been checked by now, so add screens to Options menu
    printf("%s info: heads: %d  screen: %d\n",PROGNAME,count_xinerama_heads(),iconbar_screen);
    if( iconbar_screen>count_xinerama_heads() )
        iconbar_screen=count_xinerama_heads();
    else if( iconbar_screen < 0 )
        iconbar_screen=0;
    Setup_Options_Menu(itemScreenListStart + iconbar_screen);

    Allow_HUPs(1);	// make sure the "HUP-stopper" has also gone for now
    Create_Keymaps_Menu(1);
    Allow_HUPs(0);

// make sure the folders in the user's prefs folder have no '.localized' extension
    HideFolderExtensions();

// and launch the items in the StartupItems file
    Launch_Startup_Items();
// make sure the personal Helper Scripts folder exists

// Set up cursor for command-key-shortcut mode
    cmdKeyCrsr.data[1]  = 0x0000;
    cmdKeyCrsr.data[2]  = 0x3838;
    cmdKeyCrsr.data[3]  = 0x4444;
    cmdKeyCrsr.data[4]  = 0x4444;
    cmdKeyCrsr.data[5]  = 0x4444;
    cmdKeyCrsr.data[6]  = 0x3FF8;
    cmdKeyCrsr.data[7]  = 0x0440;
    cmdKeyCrsr.data[8]  = 0x0440;
    cmdKeyCrsr.data[9]  = 0x0440;
    cmdKeyCrsr.data[10] = 0x3FF8;
    cmdKeyCrsr.data[11] = 0x4444;
    cmdKeyCrsr.data[12] = 0x4444;
    cmdKeyCrsr.data[13] = 0x4444;
    cmdKeyCrsr.data[14] = 0x3838;
    cmdKeyCrsr.data[15] = 0x0000;
    cmdKeyCrsr.data[16] = 0x0000;
    cmdKeyCrsr.mask[1]  = 0x3838;
    cmdKeyCrsr.mask[2]  = 0x7c7c;
    cmdKeyCrsr.mask[3]  = 0xfefe;
    cmdKeyCrsr.mask[4]  = 0xeeee;
    cmdKeyCrsr.mask[5]  = 0xfffe;
    cmdKeyCrsr.mask[6]  = 0x7ffc;
    cmdKeyCrsr.mask[7]  = 0x3ff8;
    cmdKeyCrsr.mask[8]  = 0x0ec0;
    cmdKeyCrsr.mask[9]  = 0x3ff8;
    cmdKeyCrsr.mask[10] = 0x7ffc;
    cmdKeyCrsr.mask[11] = 0xfffe;
    cmdKeyCrsr.mask[12] = 0xeeee;
    cmdKeyCrsr.mask[13] = 0xfefe;
    cmdKeyCrsr.mask[14] = 0x7c7c;
    cmdKeyCrsr.mask[15] = 0x3838;
    cmdKeyCrsr.mask[16] = 0x0000;
    cmdKeyCrsr.hotSpot.h = 8;
    cmdKeyCrsr.hotSpot.v = 8;
    
    do
    {
        /*if(keepalive_socket==0 && can_send_hackcmds!=0 && main_socket!=0) {
            dummy=check_for_keepalive();
            if(dummy<0) {
                fprintf(stderr,"%s: error waiting for keepalive (%d)",PROGNAME,dummy);
                exit(1);
            } else if(dummy==0)
                Allow_HUPs(0);
        }*/
        
        if(TickCount()>lastOKTicks+600) {
            lastOKTicks=TickCount();
// Write to console for the moment, to try and keep track of when it hangs
//            sprintf(script,"echo -n 'reRead: %d    ';date",reReadOnNextFront);
#ifdef SHOW_10S_LOG
//            sprintf(script,"echo -n '%s still OK on  ';date;/usr/bin/touch /tmp/oroborosx-%s.dir/check",PROGNAME,randomID);
            sprintf(script,"echo -n '%s still OK on  ';date",PROGNAME);
            system(script);
//#else
//            sprintf(script,"/usr/bin/touch /tmp/oroborosx-%s.dir/check",randomID);
#endif
//            system(script);
            Allow_HUPs(1);	// make sure the "HUP-stopper" has also gone
// Do we still need to recheck the Launch Menu Items folder...?
/*            if(gX11InFront==1) {
                if(reReadOnNextFront<0) {
                    Create_Launch_Menu();
                    reReadOnNextFront=-reReadOnNextFront;
                }
            } else if(reReadOnNextFront>0) {
                Create_Launch_Menu();
                reReadOnNextFront--;
            }
*/
            check_osx_windows_bounds();
        }

        
//        sprintf(lastsub,"TRACE1: Starting GetNextEvent/usleep in Eventloop\n");
//        subtrace();
//#ifdef SHOW_TRACE1
//        printf("TRACE1: Starting GetNextEvent/usleep in Eventloop\n");
//#endif
        Stop_XDarwin();
        dummy=0;
        gotEvent = GetNextEvent(everyEvent,&event);
        
        GetGlobalMouse(&mspt);

        real_x=mspt.h-diffscrn_x;
        real_y=mspt.v-diffscrn_y;
    
        moving=0;
        if(real_x>lastrl_x+3)
            moving=2;
        else if(real_y>lastrl_y+3)
            moving=2;
        else if(real_x<lastrl_x-3)
            moving=2;
        else if(real_y<lastrl_y-3)
            moving=2;
        else if(real_x!=lastrl_x)
            moving=1;
        else if(real_y!=lastrl_y)
            moving=1;
                
        while(gotEvent==FALSE && moving==0 && dummy<5 && mouse_mode==MOUSE_NORMAL && delayed_enter_ticks==0) {
            usleep(5000);
            gotEvent = GetNextEvent(everyEvent,&event);
            dummy++;

            GetGlobalMouse(&mspt);

            real_x=mspt.h-diffscrn_x;
            real_y=mspt.v-diffscrn_y;
    
            moving=0;
            if(real_x>lastrl_x+3)
                moving=2;
            else if(real_y>lastrl_y+3)
                moving=2;
            else if(real_x<lastrl_x-3)
                moving=2;
            else if(real_y<lastrl_y-3)
                moving=2;
            else if(real_x!=lastrl_x)
                moving=1;
            else if(real_y!=lastrl_y)
                moving=1;
        }
//        if(dummy<5) printf("Out of GetNextEvent early: %d, %d, %d\n",dummy,event.what, moving);
        
        Restart_XDarwin();
        if(gotEvent==FALSE && event.modifiers>>8!=lastsentmods) {
            SendModsToXDarwin(event.modifiers>>8,lastsentmods);
            if( (event.modifiers & 0x100 ) && (lastsentmods & 0x1)==0 && keepInFrontUntil==-1 ) {
                gotCommandDown=1;
                if(gVerboseConsole)
                    printf("%s info: Noticed command key go down...\n",PROGNAME);
            } else if( gotCommandDown ) {
                if( (event.modifiers & 0x100 )==0 && (lastsentmods & 0x1) ) {
                    keepInFrontUntil = 0;
                    gotCommandDown = 0;
                	if(gVerboseConsole)
                            printf("%s info: Switching out of command-key-shortcut mode\n",PROGNAME);
                        InitCursor();
                } else {
                    gotCommandDown = 0;
                    if(gVerboseConsole)
                        printf("%s info: Got another modifier - cancelling switch from cmd-key mode\n",PROGNAME);
                }
            }
            lastsentmods = event.modifiers>>8;
            if(gVerboseConsole)
                printf("%s info: Sent mods change to XDarwin: %#lx\n",PROGNAME,lastsentmods);
        }
//        gotEvent = WaitNextEvent(everyEvent,&event,2,nil);

//    sprintf(lastsub,"TRACE1: Back from GetNextEvent/usleep in Eventloop\n");
//    subtrace();
//#ifdef SHOW_TRACE1
//    printf("TRACE1: Back from GetNextEvent/usleep in Eventloop\n");
//#endif

        
/*        if(gX11InFront!=0 || mouse_mode!=MOUSE_NORMAL) {
            if( moving==2 )
                gotEvent = WaitNextEvent(everyEvent,&event,2,nil);
            else
                gotEvent = WaitNextEvent(everyEvent,&event,6,nil);
        } else if( moving==2 )
            gotEvent = WaitNextEvent(everyEvent,&event,3,nil);
        else
            gotEvent = WaitNextEvent(everyEvent,&event,8,nil);
*/        
        if(gQuitFlag==0) {
            if (event.what!=0)
                DoEvent(&event);
            if(gQuitFlag==0) {
                if(idleCount==noIdleMax || mouse_mode!=MOUSE_NORMAL) {
                    DoIdle(mspt,1);
                    idleCount=0;
                } else {
                    DoIdle(mspt,moving);
                    idleCount++;
                }
            }
        }
        
        lastrl_x=real_x;
        lastrl_y=real_y;
        
    } while (gQuitFlag==0);
    
    SavePrefs();
    
    dummy=sleep(1);
    //system("/bin/sleep 1");
    sprintf(script,"/bin/rm -rf /tmp/oroborosx-%s.dir",randomID);
    system(script);
    ClearUpSockets();
    ExitToShell();					
}

void Allow_HUPs(int allow)
{
    int		send_err;
    char	script[55];
    
    if(can_send_hackcmds==0) return;
    if(keepalive_socket==0) return;
    
    if(alertWin!=NULL) allow=0;

    if(allow==0)
        script[0] = 100; // This value tells keepalive that OroborOSX will be busy
    else if(allow==-1)
        script[0] = 111; // This value tells keepalive that OroborOSX is finishing
    else
        script[0] = 10;	 // This value tells keepalive that OroborOSX is running as normal
    
    send_err = send_bytes_to_socket(keepalive_socket, 4, script);
    
    if(send_err!=0)
        printf("%s warning: Got error %d sending %d to keepalive\n",PROGNAME,send_err,script[0]);
    else if(gVerboseConsole)
        printf("%s info: Sent %d to keepalive\n",PROGNAME,script[0]);
}

void Start_Time_Check(long addon)
{
    count_time_start=TickCount();
    time_to_leave=count_time_start+addon;
}

int Is_Time_Up()
{
    if(TickCount()>time_to_leave) return 1;
    if(TickCount()<count_time_start) return 1;	// just to cover case of wrap-around
    return 0;
}

void DoIdle(Point mspt, int moving)
{
    static Boolean	wasInMenubar=FALSE, wasSame=FALSE;
    short		myErr;
    ProcessSerialNumber	frontPSN;
    Boolean		x11InFront;
    Boolean		same, cursorInMenubar=FALSE;
    WindowRef		currentFrontWin;

#ifdef DEBUG
    ProcessInfoRec	procInfo;
    pstr_t		myName;
    pstr_t		procName;
#endif

//    sprintf(lastsub,"TRACE1: DoIdle (gX11InFront,moving: %d,%d)\n",gX11InFront,moving);
//    subtrace();
//#ifdef SHOW_TRACE1
//    printf("TRACE1: DoIdle (gX11InFront,moving: %d,%d)\n",gX11InFront,moving);
//#endif

    myErr=GetFrontProcess(&frontPSN);
    if(myErr!=noErr)
        return;
        
    myErr = GetCurrentProcess(&myPSN);
    if(myErr!=noErr)
        return;
        
#ifdef DEBUG
    procInfo.processAppSpec=NULL;
    procInfo.processName=(char *) &myName;
    myErr=GetProcessInformation(&myPSN,&procInfo);
    if(myErr!=noErr)
        return;
    myName.thestr[(int) myName.len]=0;

    procInfo.processAppSpec=NULL;
    procInfo.processName=(char *) &procName;
    myErr=GetProcessInformation(&frontPSN,&procInfo);
    if(myErr!=noErr)
        return;
    procName.thestr[(int) procName.len]=0;
#endif

    myErr=SameProcess(&xPSN,&frontPSN,&x11InFront);
    if(myErr!=noErr)
        return;

    if(x11InFront) {
        xPSN=frontPSN;
        if( gX11InFront==0 && Is_Whole_Screen_Mode()==1 )
            Wait_OOSX_To_Front();
    }
        
    myErr=SameProcess(&myPSN,&frontPSN,&same);
    if(myErr!=noErr)
        return;
        
    if(!same) gotCommandDown = 0;
    
    cursorInMenubar=FALSE;
    
    if(x11InFront || same)
        //Check_End_Whole_Screen_Mode();
        if(mspt.h>=gMainRect.left && mspt.v>=gMainRect.top
                        && mspt.v<gMainRect.top+22 && mspt.h<=gMainRect.right) {
            cursorInMenubar=TRUE;
            left_client();
            if( Is_Whole_Screen_Mode() && justSetWholeScreenMode==0 && same ) {
                if( gVerboseConsole ) printf("Back in menubar - showing it...\n");
                wasWholeScreenMode = 1;
                End_Whole_Screen_Mode();
            }
        } else if( justSetWholeScreenMode ) {
            if( gVerboseConsole ) printf("Left menubar - unsetting justSetWholeScreenMode...\n");
            justSetWholeScreenMode = 0;
        }/* else if( wasWholeScreenMode )
            printf("Detected left menubar - should retrun to fullscreen mode...\n");*/

// Fix for extra items on far right of menubar in OSX10.1 -only works for holding down button,
// not for click-release in menubar item...
    if(cursorInMenubar) {
        if( Button() )
// gives an extra 30 ticks after button release before allowing switch to XDarwin
            gGotMenuClick=TickCount()+30;
        else
            gGotMenuClick=0;
    } else if (gGotMenuClick!=0) {
        if( Button() ) {
// gives an extra 30 ticks after button release before allowing switch to XDarwin
            gGotMenuClick=TickCount()+30;
            cursorInMenubar=TRUE;
        } else if(gGotMenuClick>TickCount()) {
            cursorInMenubar=TRUE;
        } else
            gGotMenuClick=0;
    }
    
    if( cursorInMenubar==FALSE && wasWholeScreenMode && same ) {
        if( gVerboseConsole ) printf("Cursor left menubar - back to whole screen mode...\n");
        wasWholeScreenMode = 0;
        Start_Whole_Screen_Mode();
        if(keepInFrontUntil==-1) SetCmdKeyCursor(1);
    }
    
    if( keepInFrontUntil==-1 && same ) {
        // Deal with the cursor change
        if( cursorInMenubar && !wasInMenubar )
            SetCmdKeyCursor(0); // back to arrow when in menubar
        else if( !cursorInMenubar && wasInMenubar )
            SetCmdKeyCursor(1); // back to clover when out of menubar
        else if( !wasSame )
            SetCmdKeyCursor(1); // ensure it becomes clover when we come back in front
    }
    
    wasInMenubar = cursorInMenubar;
    wasSame = same;

    currentFrontWin=FrontWindow();

    if( currentFrontWin==NULL || !IsWindowVisible(currentFrontWin) ) {
        if(gInvisible==0) {
            if(gVerboseConsole)
                printf("%s info: Front window is invisible\n",PROGNAME);
            gInvisible=1;
            unmap_all_clients();
        }
    } else {
        if(gInvisible!=0) {
            gInvisible=0;
            if(DestroyBadWindows(currentFrontWin)) {
                printf("%s info: Front window visible, but removed\n",PROGNAME);
                currentFrontWin=FrontWindow();
            } else if(gVerboseConsole)
                printf("%s info: Front window is visible again\n",PROGNAME);
            map_all_clients();
        }
    }
            
    if(x11InFront && gInvisible==0) {
#ifdef SHOW_FRONT
        printf("FRONT: XDarwin\n");
#endif
// Make sure OroborOSX application is in front if the about window is frontmost...
        if( keepInFrontUntil != 0 ) {
            if( keepInFrontUntil==-1 ) {
                if( mouse_mode==0 ) forceOOSXInFront = 1;
            } else {
                keepInFrontUntil = TickCount()+10;
                if( gVerboseConsole ) printf("Forcing in front so Dock goes away...\n");
                if( keepInFrontUntil==-1 || keepInFrontUntil==0 ) keepInFrontUntil = 1;
                if( mouse_mode==0 ) forceOOSXInFront = 1;
            }
        }
        if(currentFrontWin==aboutWin || currentFrontWin==alertWin || (oosxWasInFront==0 && can_send_hackcmds==0)) {
            if(gVerboseConsole)
                printf("Bringing to front 1... %d,%d,%d\n",
                    oosxWasInFront,can_send_hackcmds,forceOOSXInFront);
            Bring_OOSX_To_Front();
            oosxWasInFront=1;
            same=TRUE;
            x11InFront=FALSE;
            forceOOSXInFront=FALSE;
// Also needs to be frontmost if the cursor is in the menubar - but we still want to
// act as if XDarwin is in front (i.e. keep the current X11 focus window)
        } else if(cursorInMenubar) {
            if(gVerboseConsole)
                printf("Bringing to front 2... %d,%d,%d\n",
                    oosxWasInFront,can_send_hackcmds,forceOOSXInFront);
            Bring_OOSX_To_Front();
            oosxWasInFront=1;
            forceOOSXInFront=FALSE;
        } else if(forceOOSXInFront || gX11InFront==0) {
            /*if(lastsentmods&0x8) {
                printf("%s info: option click causes trouble...",PROGNAME);
                come_to_front();
            }*/
            if(gVerboseConsole)
                printf("Bringing to front 3... %d,%d,%d\n",
                    oosxWasInFront,can_send_hackcmds,forceOOSXInFront);
            Bring_OOSX_To_Front();
            if(forceOOSXInFront) gX11InFront=1;
            oosxWasInFront=1;
            forceOOSXInFront=FALSE;
        }
    } else if(same) {
        oosxWasInFront=1;
#ifdef SHOW_FRONT
        printf("FRONT: %s\n",PROGNAME);
#endif
// But if the OroborOSX application is in front, and the about window is not frontmost...
        if(currentFrontWin!=aboutWin && currentFrontWin!=alertWin) {
// ...then we must make sure the right window has the focus (before XDarwin comes in front)
            if(currentFrontWin!=lastFrontWin) {
                MoveWindow(currentFrontWin,-5000,-5000,false);
#ifdef SHOW_EVENTS
                printf("New window in front: %#lx %#lx\n",currentFrontWin,lastFrontWin);
#else
                if(gVerboseConsole)
                    printf("New window in front: %#lx %#lx\n",currentFrontWin,lastFrontWin);
#endif
// tell the window manager that a new X11 window has come in front
		Restart_XDarwin();
                change_focus((long) currentFrontWin);
		Stop_XDarwin();
                lastFrontWin=currentFrontWin;
            }
// ...and then we want XDarwin in front
// (unless the cursor is in the menubar, in which case we pretend it is)
            if(keepInFrontUntil!=0) {
                if( keepInFrontUntil==-1 ) {
                    if(!cursorInMenubar && mouse_mode!=0) {
                        Restart_XDarwin();
                        Bring_X_To_Front(0);
                        Stop_XDarwin();
                    }
                } else if( TickCount() >= keepInFrontUntil ) {
                    if( gVerboseConsole ) printf("Enough of keeping in front...\n");
                    keepInFrontUntil = 0;
                }
            } else if(!cursorInMenubar) {
		Restart_XDarwin();
                Bring_X_To_Front(0);
            	Stop_XDarwin();
            } else if(forceOOSXInFront)
                forceOOSXInFront=FALSE;
            same=FALSE;
            x11InFront=TRUE;
        }
    } else {
        oosxWasInFront=0;
#ifdef SHOW_FRONT
        printf("FRONT: unknown\n");
#endif
    }

// Check to see if the frontmost window has changed
    if(currentFrontWin!=lastFrontWin) {
        if(currentFrontWin==aboutWin || currentFrontWin==alertWin) {
// Disable the window menu if the about window is frontmost
            Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
        } else {
// And now make sure the frontmost window is off the screen
            MoveWindow(currentFrontWin,-5000,-5000,false);
#ifdef SHOW_EVENTS
            printf("New window in front: %#lx %#lx\n",currentFrontWin,lastFrontWin);
#else
            if(gVerboseConsole)
                printf("New window in front: %#lx %#lx\n",currentFrontWin,lastFrontWin);
#endif
            if(x11InFront) {
// tell the window manager that a new X11 window has come in front
                Restart_XDarwin();
                change_focus((long) currentFrontWin);
                Stop_XDarwin();
            }
        }
        lastFrontWin=currentFrontWin;
    }

    if(moving!=0) {
#ifdef SHOW_MOUSE
        printf("Mouse moved: %d,%d  (from %d,%d)\n",real_x,real_y,lastrl_x,lastrl_y);
#endif
        lastMovedTicks=TickCount();
        lastMovedTen=lastMovedTicks;
    } else if(TickCount()>lastMovedTen+10) {
        lastMovedTen=TickCount();
        moving=10;
    }

    if(x11InFront) {
    
//        printf("Calling do_event_loop with 2000 or 1000...\n");
        if( moving==0 && delayed_enter_ticks==0 )
            do_event_loop(2000,0);
        else
            do_event_loop(1000,moving);

	if(gX11InFront==0) {
#ifdef DEBUG
            printf("%s has come to front\n",procName.thestr);
#endif
            gX11InFront=1;
            Restart_XDarwin();
            come_to_front();
            Stop_XDarwin();
        }

    } else {
    
        if(gX11InFront!=0) {
#ifdef DEBUG
            if(same)
                printf("%s: I am in front!\n",procName.thestr);
            else
                printf("%s has come to front\n",procName.thestr);
#endif
            gX11InFront=0;
            Restart_XDarwin();
            gone_to_back();
            Stop_XDarwin();
        }
        
//        printf("Calling do_event_loop with 4000 or 2000...\n");
        if( moving==0 )
            do_event_loop(5000,0);
        else
            do_event_loop(3000,moving);
    }
    
    if(gQuitFlag==0)
        if(lastIconbarOutTicks!=0)
            if(TickCount()>lastIconbarOutTicks+iconbar_WaitTicks) {
                Restart_XDarwin();
                hide_all_iconbars();
            	Stop_XDarwin();
            }
}

void DoEvent(EventRecord *event)
{
    short		part;
    Boolean		hit;
    unsigned char	key;
    Rect		tempRect;
    WindowRef		whichWindow;
    long		menusel;
    Point		mouse;
    ControlHandle	control;
    int			wasNotWholeScreenMode=0, keycode=0;
        
    sprintf(lastsub,"TRACE1: DoEvent (what,where,msg,mods: %d %#lx %#lx %#lx)\n",
                    event->what,event->where,event->message,event->modifiers);
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: DoEvent (what,where,msg,mods: %d %#lx %#lx %#lx)\n",
                    event->what,event->where,event->message,event->modifiers);

#else
    if(gVerboseConsole)
        printf("%s info: DoEvent (what,where,msg,mods: %d %#lx %#lx %#lx)\n",
                    PROGNAME,event->what,event->where,event->message,event->modifiers);
#endif

    if( Is_Whole_Screen_Mode()==0 ) wasNotWholeScreenMode = 1;

    switch (event->what) 
    {
        case mouseDown:
#ifdef SHOW_EVENTS
            printf("%s event: Got mouse down...",PROGNAME);
#endif
            gotCommandDown = 0;
            part = FindWindow(event->where, &whichWindow);
            switch (part)
            {
                case inMenuBar:  /* process a moused menu command */
                    SendSimpleHackCommand(HACKCMD_ALLOWFRONT,1); // X11 window click must activate XDarwin
                    Allow_HUPs(0);
                    menusel=MenuSelect(event->where);
                    SendSimpleHackCommand(HACKCMD_ALLOWFRONT,0);
                    if(alertWin==NULL) Allow_HUPs(1);
#ifdef SHOW_EVENTS
                    printf(" in menubar (%#lx)...\n",menusel);
#endif
                    if(menusel!=0)
                        DoMenuCommand(menusel,0,0);
                    break;
                    
                case inSysWindow:
                    break;
                
                case inContent:
#ifdef SHOW_EVENTS
                    printf(" in window (%#lx)...\n",whichWindow);
#endif
                    if(whichWindow != FrontWindow()) 
                        SelectWindow(whichWindow);
                    if(whichWindow==alertWin) {
                        SetPort(GetWindowPort(whichWindow));
                        mouse = event->where;    //get the mouse location
                        GlobalToLocal(&mouse);   //convert to local coordinates
                        part = FindControl(mouse, whichWindow, &control);
                        printf("   ...got control part: %d\n",part);
                    }
                    break;
                
                case inDrag:	/* pass screenBits.bounds */
#ifdef SHOW_EVENTS
                    printf(" in window title (%#lx)...\n",whichWindow);
#endif
                    Restart_XDarwin();
                    if(showingAboutWin)
                        if(whichWindow==aboutWin)
                            if(gX11InFront==1) {
                                gX11InFront=0;
                                gone_to_back();
                            }
                    GetRegionBounds(GetGrayRgn(), &tempRect);
                    DragWindow(whichWindow, event->where, &tempRect);
                    Stop_XDarwin();
                    break;
                    
                case inGrow:
                    printf(" in window resize widget (%#lx)...\n",whichWindow);
                    break;
                    
                case inGoAway:
#ifdef SHOW_EVENTS
                    printf(" in window close widget (%#lx)...\n",whichWindow);
#endif
                    Restart_XDarwin();
                    hit = TrackBox(whichWindow, event->where, part);
                    Stop_XDarwin();
                    if (hit) 
                    {
                        if(whichWindow==aboutWin) {
#ifdef DEBUG
                            printf("Hiding about window\n");
#endif
                            showingAboutWin=0;
                            DisposeWindow(aboutWin);
                            aboutWin=NULL;
                            if(showingNoFocusWin!=0) {
                                ShowWindow(noFocusWin);
                                SelectWindow(noFocusWin);
                            }
                        } else {
#ifdef DEBUG
                            printf("Closing window: %#lx (%#lx)\n",whichWindow,aboutWin);
#endif
                            DisposeWindow(whichWindow); }
/*                        quit();
                        system("/bin/sleep 1");
                        ExitToShell();*/
                    }
                    break;
                    
                case inZoomIn:
                case inZoomOut:
                    printf(" in window zoom widget (%#lx)...\n",whichWindow);
                    Restart_XDarwin();
                    hit = TrackBox(whichWindow, event->where, part);
                    Stop_XDarwin();
                    if (hit) 
                    {
                        SetPort(GetWindowPort(whichWindow));   // window must be current port
                        EraseRect(GetWindowPortBounds(whichWindow, &tempRect));   // inval/erase because of ZoomWindow bug
                        ZoomWindow(whichWindow, part, true);
                        InvalWindowRect(whichWindow, GetWindowPortBounds(whichWindow, &tempRect));	
                    }
                    break;
            }
            break;
            
            case keyUp:
                gotCommandDown = 0;
                key = event->message & charCodeMask;
                keycode = (event->message & keyCodeMask)>>8;
#ifndef SHOW_EVENTS
                if(gVerboseConsole)
#endif
                    printf("%s event: Got key up: %#lx %#lx\n",PROGNAME,key,keycode);
                if((event->modifiers & cmdKey)==0)
                    SendKeyToXDarwin(keycode,0);
                break;
            
            case keyDown:
            case autoKey:
                gotCommandDown = 0;
                key = event->message & charCodeMask;
                keycode = (event->message & keyCodeMask)>>8;
#ifndef SHOW_EVENTS
                if(gVerboseConsole) {
#endif
                    if(event->what == keyDown)
                        printf("%s event: Got key down: %#lx %#lx  (mods:",PROGNAME,key,keycode);
                    else
                        printf("%s event: Got autokey: %#lx %#lx  (mods:",PROGNAME,key,keycode);
                    if(event->modifiers & cmdKey) printf(" cmd");
                    if(event->modifiers & controlKey) printf(" ctrl");
                    if(event->modifiers & optionKey) printf(" opt");
                    if(event->modifiers & shiftKey) printf(" shift");
                    if((event->modifiers & (cmdKey|controlKey|optionKey|shiftKey))==0) printf(" none");
                    printf(")\n");
#ifndef SHOW_EVENTS
                }
#endif
                if((event->modifiers & cmdKey)==0) {
//                    if(passwordWin!=NULL && key==65 && passwordWin==alertWin) {
//                        printf("Got rid of password window\n");
//                        DisposeWindow(passwordWin);
//                        passwordWin=NULL;
//                        alertWin=NULL
//                    }
                    SendKeyToXDarwin(keycode,1);
                } else if (event->what == keyDown) {
                    if((event->modifiers & optionKey)!=0) {
                        // deal with the option-key menu items
                        // ANNOYING!!! - why won't this work automatically?!?
                        if((event->modifiers & controlKey)!=0) {
                            if(key==26 || key==122)
                                DoMenuCommand((long) 0,menuOptions,itemWholeScreenMode);
                        } else if(groupCycleOff==0) {
                            if((event->modifiers & shiftKey)!=0) {
                                if(key==96 || key==126 || key==217)
                                    DoMenuCommand((long) 0,menuWindow,itemNextGroup);
                            } else {
                                if(key==96 || key==126 || key==217)
                                    DoMenuCommand((long) 0,menuWindow,itemLowerGroup);
                            }
                        }
                    } else if((event->modifiers & shiftKey)!=0) {
                        // deal with the shift-key menu items
                        // ANNOYING!!! - why won't this work automatically?!?
                        if(key==47)		// "OroborOSX Help" in Help menu
                            DoMenuCommand((long) 0,menuHelp,itemHelp);
                        else if(key==102)	// "Auto-focus" in Window menu
                            DoMenuCommand((long) 0,menuProperties,itemAutoFocus);
                        else if(key==114)	// "Auto-raise" in Window menu
                            DoMenuCommand((long) 0,menuProperties,itemAutoRaise);
                        else if(key==116)	// "Auto-lower" in Window menu
                            DoMenuCommand((long) 0,menuProperties,itemAutoLower);
                        else {
                            if(cycleOff==0) {
                                // "Next" in Window menu -as shift backquote or tilde or shift right arrow
                                if(key==96 || key==126 || key==217 || key==29)
                                    DoMenuCommand((long) 0,menuWindow,itemNext);
                                else if(key==28)	// "Send to back" -as shift left arrow
                                    DoMenuCommand((long) 0,menuWindow,itemLower);
                            }
                            if(groupCycleOff==0) {
                                if(key==30)	// "Next Group" -as shift up arrow
                                    DoMenuCommand((long) 0,menuWindow,itemNextGroup);
                                else if(key==31)	// "Send group to back" -as shift down arrow
                                    DoMenuCommand((long) 0,menuWindow,itemLowerGroup);
                            }
                        }
                    } else {
                        if(key==96 || key==126 || key==217)	// "Send To Back" in Window menu -as tilde or backquote
                            DoMenuCommand((long) 0,menuWindow,itemLower);
                        else
                            DoMenuCommand(MenuKey(key),0,0);
                    }
                }
                break;
                
            case activateEvt:	       /* if you needed to do something special */
                break;
                    
            case updateEvt:
                    DrawWindow((WindowRef) event->message);
                    break;
                        
            case kHighLevelEvent:
#ifndef SHOW_EVENTS
                    if(gVerboseConsole)
#endif
                        printf("%s event: Got an AppleEvent...\n",PROGNAME);
                    AEProcessAppleEvent( event );
                    break;
		
            case diskEvt:
                    break;
                
            default:
#ifndef SHOW_EVENTS
                    if(gVerboseConsole)
#endif
                        printf("%s event: Got unhandled event number %d\n",PROGNAME,event->what);
                    break;
    }
        
    if( Is_Whole_Screen_Mode()==1 && wasNotWholeScreenMode ) {
    	if( gVerboseConsole ) printf("Went into whole screen mode - setting justSetWholeScreenMode...\n");
	justSetWholeScreenMode = 1;
    }

    sprintf(lastsub,"TRACE1: leaving DoEvent\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: leaving DoEvent\n");
#endif
}

