/******************************************/ 
/* Carbon-compatible section of OroborOSX */
/* (c) Adrian Umpleby Jan 2002            */
/*                                        */
/* Original oroborus window manager code  */
/* (c) Ken Lynch Jan 2001                 */
/*                                        */
/* Distributed under the GPL              */ 
/* See LICENSE for more details           */ 
/******************************************/ 
 
/* menus.c */

#include "hackcmds.h"
#include "main.h"

#define DISABLE_CHANGEGROUP_MENU

char		*nextItem=NULL, *nextGroupItem=NULL, *aboutThemeItem=NULL, *orgGroupItem=NULL;
int		keymapItemsCount=7, themeItemsCount=8, launchItemsCount=17, gDoneSetupMenus=0;
int		itemPersonalKeymaps=initPersonalKeymaps; // this wil increase as global keymaps are added
// guess I should really make sure the Launch menu is not too long...
int		launch_counter[250];	// should be enough items, though   :-)

/* Lookup table iconbar_steps -> menu item. */
int steps_menu_items[][2] = {{ICONSTEPS_FAST,   itemFastIconbars  },
                            {  ICONSTEPS_MED,   itemMedIconbars   },
                            {  ICONSTEPS_SLOW,  itemSlowIconbars  }};
extern int iconbar_steps;


void Edit_Startup_Items()
{
    char	*script=NULL;
    int		err;

    sprintf(lastsub,"TRACE1: Edit_Startup_Items\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: Edit_Startup_Items\n");
#endif

#ifdef DEBUG
    printf("About to malloc script in Edit_Startup_Items: %s\n",Display_Name);
#endif
    script=malloc(45+strlen(Display_Name));
#ifdef DEBUG
    printf("script: %#lx\n",(long) script);
#endif
    if(script==NULL) {
        fprintf(stderr, "%s: Memory allocation failed in Edit_Startup_Items.\n", PROGNAME);
        if(!on_cmd_line) ShowAlert(WARNMEM_NAME,1,0);
        return;
    }

// crude editing method - just open up the file in xedit for the user to fiddle with...
    sprintf(script,"Contents/Scripts/EditStartups '%s'",Display_Name);
    printf("%s info: Running: %s\n",PROGNAME,script);
    err=system(script);
    if(err==NOLMIFLDR_STAT)
        ShowAlert(NOLMIFLDR_NAME,1,0);
    else if(err==NOTEXIST_STAT)
        ShowAlert(NOTEXISTLAUNCH_NAME,1,0);
    else if(err!=0) {
        printf("Script returned with error: %d\n",err);
        ShowAlert(WARNLAUNCH_NAME,1,0);
    }

#ifdef DEBUG
    printf("About to free script in Edit_Startup_Items...\n");
#endif
    free(script);
#ifdef DEBUG
    printf("script freed OK\n");
#endif
}

int RunX11App(char *name, int namelen, char *args, int arglen, int inhibit)
{
    char	*script=NULL;
    int		err;

    sprintf(lastsub,"TRACE1: RunX11App (name,args: %s %s)\n",name,args);
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: RunX11App (name,args: %s %s)\n",name,args);
#endif

#ifdef DEBUG
    printf("About to malloc script in RunX11App: %d\n",namelen);
#endif
    script=malloc(namelen+arglen+strlen(Display_Name)+50);
#ifdef DEBUG
    printf("script: %#lx\n",(long) script);
#endif

    if(script==NULL)
    {
	if(on_cmd_line || inhibit) {
            fprintf(stderr, "%s: Memory allocation failed in RunX11App.\n", PROGNAME);
            return -1;
        } else {
            fprintf(stderr, "%s: Memory allocation failed in RunX11App.\n", PROGNAME);
            ShowAlert(WARNMEM_NAME,1,0);
            return -1;
        }
    }
    
    if (arglen==0)
        sprintf(script,"Contents/LaunchScripts/generic '%s' '%s'", Display_Name, name);
    else
        sprintf(script,"Contents/LaunchScripts/generic '%s' '%s' %s", Display_Name, name, args);

//#ifdef DEBUG
    printf("%s info: Running: %s\n",PROGNAME,script);
//#endif

// run script to launch X11 application...
    err=system(script);
    if(inhibit==0) {
        if(err==NOLMIFLDR_STAT)
            ShowAlert(NOLMIFLDR_NAME,1,0);
        else if(err==NOTEXIST_STAT)
            ShowAlert(NOTEXISTLAUNCH_NAME,1,0);
        else if(err!=0) {
            printf("Script returned with error: %d\n",err);
            ShowAlert(WARNLAUNCH_NAME,1,0); }
    }
    
#ifdef DEBUG
    printf("About to free script in RunX11App...\n");
#endif
    free(script);
#ifdef DEBUG
    printf("script freed OK\n");
#endif
    return err;
}

int RunX11Script(int item, int inhibit, char *thePath)
{
    MenuRef	menu;
    Str255	itemStr;
    int		namelen, cntr, err=0;
    pstr_t	*pStr;
    char	*cStr;
    char	*script=NULL;

    sprintf(lastsub,"TRACE1: RunX11Script (item: %d)\n",item);
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: RunX11Script (item: %d)\n",item);
#endif

    if(thePath==NULL) {
        menu = GetMenuHandle( menuLaunch );
        GetMenuItemText(menu,item,itemStr);
    
// hack to get the string out of the pascal string
// - there's probably a macro to do this 'legally' somewhere!
        pStr=&itemStr;
        namelen=pStr->len;
        cStr=&(pStr->thestr);
        cStr[namelen]=0;
    } else
        cStr = thePath;
    
    // Make sure cStr has any single quotes escaped
    if(thePath==NULL) {
        script=Escape_Quotes(cStr,250);
        if(script) { // set 'err' as a flag so we know to free cStr later on
            cStr=script;
            err=1; }
    }
    namelen=strlen(cStr);

#ifdef DEBUG
    printf("Got item: (%d) %s\n",namelen,cStr);    
    printf("About to malloc script in RunX11Script: %d\n",namelen);
#endif
    script=malloc(namelen+strlen(Display_Name)+65);
#ifdef DEBUG
    printf("script: %#lx\n",(long) script);
#endif
    if(script==NULL)
    {
	if(on_cmd_line || inhibit) {
            fprintf(stderr, "%s: Memory allocation failed in RunX11Script.\n", PROGNAME);
            if(err) free(cStr);
            return -1;
        } else {
            fprintf(stderr, "%s: Memory allocation failed in RunX11Script.\n", PROGNAME);
            ShowAlert(WARNMEM_NAME,1,0);
            if(err) free(cStr);
            return -1;
        }
    }
    
    cntr=0;
    if(thePath==NULL) {
        // counters start at zero... (oops! never mind -just add one before launching...)
        cntr=launch_counter[item-itemUserItems-1]+1;
        // Tell the LaunchList script to launch the Launch menu item
        if(cntr==1)
            sprintf(script,"Contents/Scripts/LaunchList launch './%s.x11app' '%s'",cStr,Display_Name);
        else
            sprintf(script,"Contents/Scripts/LaunchList launch './%s.x11app' '%s' '%d'",cStr,Display_Name,cntr);
    } else // Since this is not an item from the Launch menu, launch it without a counter
        sprintf(script,"Contents/Scripts/LaunchList launch '%s' '%s'",cStr,Display_Name);
    
    // if err is set then cStr was malloc'd in Escape_Quotes, so free it here
    if(err) free(cStr);

    if(gVerboseConsole)
        printf("%s info: Running: %s\n",PROGNAME,script);

    err=system(script);
    if(err==0) {
        if(thePath!=NULL) launch_counter[item-itemUserItems-1]=cntr;
    } else if(inhibit==0) {
        if(err==NOTEXIST_STAT)
            ShowAlert(NOTEXISTLAUNCH_NAME,1,0);
        else {
            printf("%s warning: Script returned with error: %d\n",PROGNAME,err);
            ShowAlert(WARNLAUNCH_NAME,1,0); }
    }
        
#ifdef DEBUG
    printf("About to free script in RunX11Script...\n");
#endif
    free(script);
#ifdef DEBUG
    printf("script freed OK\n");
#endif
    return err;
}

int Select_Keymaps_Menu(int which, int changed)
{
    MenuRef	menu;
    int		cnt,i,namelen,err,buflen,trycnt=0;
    Str255	itemStr;
    pstr_t	*pStr;
    char	*cStr, *script=NULL, *buf=NULL, folder[30];
    OSStatus	stat;
    FILE	*file;
    
    sprintf(lastsub,"TRACE3: Select_Keymaps_Menu (which: %d)\n",which);
    subtrace();
//#ifdef SHOW_TRACE3
    printf("TRACE3: Select_Keymaps_Menu (which: %d)\n",which);
//#endif

//    menu=GetMenuHandle( menuThemes );
    stat = GetMenuItemHierarchicalMenu( GetMenuHandle(menuOptions), itemKeymapsSubmenu, &menu );
    cnt=CountMenuItems(menu);
    
    pStr=&itemStr;
    cStr=&(pStr->thestr);
    sprintf(cStr,"open-keymap-folder");

    if( changed<0 ) changed=0;

tryagain:
    if(which==itemMergeXmodmap) {
        sprintf(folder,"personal");
        sprintf(cStr,"merge-personal-keymap");
        changed=-2;
    } else if(which==itemSaveKeymap) {
        char *keymapName=NULL;
        textInputWin=MakeWindow(keymapNameWintype);
        if(textInputWin==NULL) {
            printf("%s warning: Failed to show keymap name window!\n",PROGNAME);
            return 0;
        }
        keymapName=DoTextInputWindow(0);
        if( keymapName==NULL ) {
            printf("%s info: keymap name not received\n",PROGNAME);
            return 0;
        }
        sprintf(folder,"save-current-keymap");
        sprintf(cStr,keymapName);
        free(keymapName);
        changed=-3;	// want to update menu
    } else if(which==itemPersonalKeymaps) {
        changed=-1;
        sprintf(folder,"personal");
    } else if(which==itemGlobalKeymaps) {
        changed=-1;
        sprintf(folder,"global");
    } else if(which==itemPersonalKeymaps-1 || which==itemGlobalKeymaps-1)
        changed=0;	// there are blank items before itemPersonal/GlobalKeymaps, so shouldn't get this!
    else if(which>itemGlobalKeymaps && which<=cnt) {
        for(i=itemGlobalKeymaps+1;i<=cnt;i++)
            CheckMenuItem( menu, i, false );
        SetItemMark( menu, which, diamondMark );
        //printf("Put diamond mark on item: %d (out of %d)\n",which,cnt);
        if(changed) {
            // now get the name of the item and put it into gKeymapName
            GetMenuItemText(menu,which,itemStr);
            pStr=&itemStr;
            namelen=pStr->len;
            cStr=&(pStr->thestr);
            cStr[namelen]=0;
            if(gKeymapName!=NULL) free(gKeymapName);
            cStr+=2;
            gKeymapName=strdup(cStr);
            cStr-=2;
            printf("%s info: Set new keymap file from menu item: %s\n",PROGNAME,gKeymapName);
            if(which>itemPersonalKeymaps) {
                gKeymapGlobal = 0;
                sprintf(folder,"personal");
            } else {
                gKeymapGlobal = 1;
                sprintf(folder,"global");
            }
            // append extension for filename
            sprintf(cStr,"%s.xmodmap",gKeymapName);
            SavePrefs();
        }
    }

    if(changed==0) return 0;
    
    script=malloc(85+strlen(cStr)+strlen(Display_Name));
    buf=malloc(256); // don't need to do this here, but means we don't have to check mem again
    if(script==NULL || buf==NULL)
    {
        if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in Select_Keymaps_Menu.\n", PROGNAME);
            ClearUpSockets();
            exit(1);
        } else {
            fprintf(stderr, "%s: Memory allocation failed in Select_Keymaps_Menu.\n", PROGNAME);
            ShowAlert(FATALMEM_NAME,0,0);
            gQuitFlag=FATALMEM_ERRNO;
            return 0;
        }
    }
    
    err = MyXServerGrabLevel();
    printf("---- MyXServerGrabLevel is %d\n",err);
    err = 0;
// This is the file that the script will create
    sprintf(script,"/bin/rm -f /tmp/keymaplist.%s",randomID);
    printf("%s info: Removing file: %s\n",PROGNAME,script);
    system(script);
runscript:
    gWaitingForSIGCHLD=1;
    sprintf(script,"Contents/Scripts/KeymapList '%s' '%s' '%s' '%s'",randomID,folder,cStr,Display_Name);
    printf("Running command: %s\n",script);
    err=system(script);
    if(err!=noErr) {
        fprintf(stderr,"%s warning: got error %d from KeymapList script\n",PROGNAME,err);
        if(err==NOKEYMAPSFLDR1_STAT)
            ShowAlert(NOKEYMAPSFLDR1_NAME,1,0);
        else if(err==NOKEYMAPSFLDR2_STAT)
            ShowAlert(NOKEYMAPSFLDR2_NAME,1,0);
        else if(err==NOXMODMAP_STAT)
            ShowAlert(NOXMODMAP_NAME,1,0);
        else if(err==KEYMAPWRITEFAIL_STAT)
            ShowAlert(KEYMAPWRITEFAIL_NAME,1,0);
        else if(err==KEYMAPEXISTS_STAT) {
            sprintf(folder,"save-current-keymap-force");
            err=ShowAlert(KEYMAPEXISTS_NAME,1,1);
            printf("Got err=%d from ShowAlert...\n",err);
            if(err!=0)
                goto runscript;
            else {
                free(script);
                free(buf);
                goto tryagain;
            }
        } else
            ShowAlert(PASSWINFAIL_NAME,1,0);
        changed = 0;
    }
    
    if(changed<=0) {
        free(script);
        free(buf);
        if(changed==-3) {
            if(gKeymapName) free(gKeymapName);
            gKeymapName=strdup(cStr);
        }
        return changed;
    }
    
// This is the file that the listing script should have just created
    sprintf(script,"/tmp/keymaplist.%s",randomID);
//#ifdef DEBUG
    printf("%s info: Keymaplist file: '%s'\n",PROGNAME,script);
//#endif

    while(trycnt<100) {
// Now check to see if the file exists
    if(gWaitingForSIGCHLD && trycnt<50) {
        printf("%s info: waiting for SIGCHLD signal after KeymapList script...\n",PROGNAME);
        trycnt++;
        do_event_loop(100000,-999);
    } else if((file=fopen(script, "r"))) 
    {
        int modlist[6][2], key, side, keycode;
        for(i=0;i<6;i++) {
            modlist[i][0] = 0;
            modlist[i][1] = 0;
        }
// now extract the lines from file (max 254 characters)
        while(fgets(buf, 254, file)) 
        {
// only use a line if it does not start with a hash and it is not blank
// (remember that a line includes the newline at the end of it)
            buflen=strlen(buf);
            if(buflen<2) {
                if(gVerboseConsole) printf("Short line: (%d) %s",buflen,buf);
            } else if(strncmp(buf,"#",1)) {
                if(gVerboseConsole) printf("%s info: Got an item: %s  ",PROGNAME,buf);
                buf[1] = 0;
                key = atoi(buf);
                buf+=2;
                buf[1] = 0;
                side = atoi(buf);
                buf+=2;
                keycode = atoi(buf);
                buf-=4;
                if(key>0 && key<6 && side>0 && side<3 && keycode>7 && keycode<256) {
                    modlist[key-1][side-1] = keycode;
                } else
                    printf("Failed to read: %d %d %d\n",key,side,keycode);
            } else if(gVerboseConsole)
                printf("Comment: %s",buf); // shouldn't be any of these anyway...?
        } 
// close the file and get rid of it
        fclose(file);
        sprintf(script,"/bin/rm -f /tmp/keymaplist.%s",randomID);
        printf("%s info: Removing file: %s\n",PROGNAME,script);
        system(script);
        SendKeycodeModsToXDarwin(&modlist[0][0]);
        trycnt=100;
    } else {
        printf("%s warning: keymaplist file not found: '%s'\n",PROGNAME,script);
        trycnt++;
        do_event_loop(100000,-999);
    }
    }
    
    free(buf);
    free(script);
    return 0;
}

/*
 * Rather 'non-Macish' way to get a list of files in the Keymaps folders,
 * and then use them to create the items for the "X11 Keymaps" submenu
 */
void Create_Keymaps_Menu(int setKeymap)
{
    MenuRef	menu;
    int		i, buflen, err, jjj, changed=0;
    char	*script=NULL, *buf=NULL;
    FILE	*file;
    OSStatus	stat;
    pstr_t	*pStr;
    char	*cStr;
    Str255	itemStr;

//    menu = GetMenuHandle( menuThemes );
    stat = GetMenuItemHierarchicalMenu( GetMenuHandle(menuOptions), itemKeymapsSubmenu, &menu );
    
    sprintf(lastsub,"TRACE1: Create_Keymaps_Menu (set, menu: %d, %#lx)\n",setKeymap,menu);
    subtrace();
#ifdef SHOW_TRACE1    
    printf("TRACE1: Create_Keymaps_Menu (menu: %#lx)\n",menu);
#endif

#ifdef DEBUG    
    printf("About to malloc script & buf: 55 & 256\n");
#endif
    script=malloc(60+strlen(Display_Name));
    buf=malloc(256);
#ifdef DEBUG    
    printf("script,buf: %#lx,%#lx\n",(long) script, (long) buf);
#endif

    if(script==NULL || buf==NULL)
    {
        if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in Create_Keymaps_Menu.\n", PROGNAME);
            ClearUpSockets();
            exit(1);
        } else {
            fprintf(stderr, "%s: Memory allocation failed in Create_Keymaps_Menu.\n", PROGNAME);
            ShowAlert(FATALMEM_NAME,0,0);
            gQuitFlag=FATALMEM_ERRNO;
            return;
        }
    }
    
    sprintf(script,"Contents/Scripts/KeymapList '%s' list list '%s'",randomID,Display_Name);
    if(gVerboseConsole)
        printf("%s info: Running: %s\n",PROGNAME,script);
// run script to get list of items...
    err=system(script);
    if(err!=noErr) {
        fprintf(stderr,"%s warning: got error %d from KeymapList script\n",PROGNAME,err);
        if(err==NOKEYMAPSFLDR1_STAT)
            ShowAlert(NOKEYMAPSFLDR1_NAME,1,0);
        else if(err==NOKEYMAPSFLDR2_STAT)
            ShowAlert(NOKEYMAPSFLDR2_NAME,1,0);
    }
    
// This is the file that the script should have just created
    sprintf(script,"/tmp/keymaplist.%s",randomID);
#ifdef DEBUG
    printf("%s info: Keymaplist file: '%s'\n",PROGNAME,script);
#endif

// Now check to see if the file exists
    if((file=fopen(script, "r"))) 
    {
        int insertPos = itemGlobalKeymaps; // start off appending to global entries
// Now that we know the file exists, we start fiddling the menu entries
        keymapItemsCount=CountMenuItems(menu);
        for(i=itemPersonalKeymaps;i<keymapItemsCount;i++)
            DeleteMenuItem(menu,itemPersonalKeymaps+1);
        keymapItemsCount=CountMenuItems(menu);
        if(keymapItemsCount!=itemPersonalKeymaps)
            printf("%s warning: inconsistency in Keymaps menu (%d vs %d) - could be bad!",
                                            PROGNAME,keymapItemsCount,itemPersonalKeymaps);
        itemPersonalKeymaps = keymapItemsCount;
        for(i=itemGlobalKeymaps;i<itemPersonalKeymaps-2;i++)
            DeleteMenuItem(menu,itemGlobalKeymaps+1);
        keymapItemsCount=CountMenuItems(menu);
        if(keymapItemsCount!=initPersonalKeymaps)
            printf("%s warning: inconsistency in Keymaps menu (%d vs %d) - could be bad!",
                                            PROGNAME,keymapItemsCount,initPersonalKeymaps);
        itemPersonalKeymaps = initPersonalKeymaps;
        
// hack to easily convert the string to a pascal string
        buf+=3;
        
// now extract the lines from file (max 254 characters)
        while(fgets(buf, 254, file)) 
        {
// only use a line if it does not start with a hash and it is not blank
// (remember that a line includes the newline at the end of it)
            buflen=strlen(buf);
            if(buflen<2) {
                if(gVerboseConsole) printf("Short line: (%d) %s",buflen,buf);
            } else if(!strncmp(buf,"SWITCH TO PERSONAL...",19)) {
                insertPos = 0; // got special switch string - now on personal keymap list
            } else if(strncmp(buf,"#",1)) {
                if(gVerboseConsole) printf("%s info: Got an item: %s  ",PROGNAME,buf);
                // now hack it to a pascal string
                buf-=3;
                buf[0]=buflen+2;
                buf[1]=32;
                buf[2]=32;
                if(insertPos==0)
                    AppendMenu(menu,buf);
                else {
                    InsertMenuItem(menu,buf,insertPos);
                    insertPos++;
                    itemPersonalKeymaps++;
                }
                buf+=3;
                keymapItemsCount++;
            } else if(gVerboseConsole)
                printf("Comment: %s",buf); // shouldn't be any of these anyway...?
        } 
// close the file and get rid of it
        fclose(file);
        if(gVerboseConsole) {
            sprintf(script,"/bin/cat /tmp/keymaplist.%s",randomID);
            system(script);
        }
        sprintf(script,"/bin/rm -f /tmp/keymaplist.%s",randomID);
        printf("%s info: Removing file: %s\n",PROGNAME,script);
        system(script);
        
// make sure that we undo the hack before freeing this...
        buf-=3;

    } else
        printf("%s warning: keymaplist file not found: '%s'\n",PROGNAME,script);
    
// and make sure we have the right number of menu items
    keymapItemsCount=CountMenuItems(menu);

    if(keymapItemsCount<=initPersonalKeymaps) {
        printf("%s: No Keymaps found!! Could end up fatal...\n",PROGNAME);
        ShowAlert(NOKEYMAPS_NAME,1,0);
        return;
    }
    
#ifdef DEBUG
    printf("About to free buf and script in Create_Keymaps_Menu...\n");
#endif
    free(buf);
    free(script);
#ifdef DEBUG
    printf("buf and script freed OK\n");
#endif

// search for the current theme, and set it
    if(setKeymap!=0) {
        printf("%s info: searching for item %s...\n",PROGNAME,gKeymapName);
        setKeymap=0;
        if(gKeymapName==NULL) {
            printf("%s info: Create_Keymaps_Menu with NULL keymap name\n",PROGNAME);
        } else {
            int startPos = itemGlobalKeymaps;
            if( gKeymapGlobal==0 ) startPos = itemPersonalKeymaps;
tryagain:
            for(i=startPos;i<keymapItemsCount;i++) {
                GetMenuItemText(menu,i+1,itemStr);
                pStr=&itemStr;
                buflen=pStr->len;
                cStr=&(pStr->thestr);
                cStr[buflen]=0;
                cStr+=2;
                if(gVerboseConsole) printf("%s info: Got item %d: %s\n",PROGNAME,i+1,cStr);
                if(buflen==strlen(gKeymapName)+2)
                    if(!strncasecmp(cStr,gKeymapName,buflen-2))
                        setKeymap=i+1;
            }
            if(setKeymap==0 && gKeymapGlobal==0 ) {
                setKeymap = -1;
                startPos = itemGlobalKeymaps;
                goto tryagain;
            }
            if(setKeymap<0) setKeymap=0;
        }
        if(setKeymap==0) {
            if(gKeymapName!=NULL) {
                printf("%s warning: Could not find current keymap in menu: %s\n",PROGNAME,gKeymapName);
                ShowAlert(NOCURKEYMAP_NAME,1,0);
            } else {
                printf("%s warning: No keymap set; using first in menu...\n",PROGNAME);
                setKeymap=itemGlobalKeymaps+1;
                changed = 1;
            }
        } else
            changed = 1;
        if(setKeymap!=0)
            Select_Keymaps_Menu(setKeymap,changed);
    }
}

/*
 * Rather 'non-Macish' way to get a list of files in the Themes folder,
 * and then use it to create the user items for the "Themes" submenu
 */
void Create_Themes_Menu(int setThemeDir)
{
    MenuRef	menu;
    int		i, buflen, err, jjj;
    char	*script=NULL, *buf=NULL;
    FILE	*file;
    OSStatus	stat;
    pstr_t	*pStr;
    char	*cStr;
    Str255	itemStr;

//    menu = GetMenuHandle( menuThemes );
    stat = GetMenuItemHierarchicalMenu( GetMenuHandle(menuOptions), itemThemeSubmenu, &menu );
    
    sprintf(lastsub,"TRACE1: Create_Themes_Menu (set, menu: %d, %#lx)\n",setThemeDir,menu);
    subtrace();
#ifdef SHOW_TRACE1    
    printf("TRACE1: Create_Themes_Menu (menu: %#lx)\n",menu);
#endif

#ifdef DEBUG    
    printf("About to malloc script & buf: 55 & 256\n");
#endif
    script=malloc(55);
    buf=malloc(256);
#ifdef DEBUG    
    printf("script,buf: %#lx,%#lx\n",(long) script, (long) buf);
#endif

    if(script==NULL || buf==NULL)
    {
        if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in Create_Themes_Menu.\n", PROGNAME);
            ClearUpSockets();
            exit(1);
        } else {
            fprintf(stderr, "%s: Memory allocation failed in Create_Themes_Menu.\n", PROGNAME);
            ShowAlert(FATALMEM_NAME,0,0);
            gQuitFlag=FATALMEM_ERRNO;
            return;
        }
    }
    
    sprintf(script,"Contents/Scripts/ThemeList '%s'",randomID);
    if(gVerboseConsole)
        printf("%s info: Running: %s\n",PROGNAME,script);
// run script to get list of items...
    err=system(script);
    if(err==NOTHEMEFLDR_STAT)
        ShowAlert(NOTHEMEFLDR_NAME,1,0);

// This is the file that the script should have just created
    sprintf(script,"/tmp/themelist.%s",randomID);
#ifdef DEBUG
    printf("%s info: Themelist file: '%s'\n",PROGNAME,script);
#endif

// Now check to see if the file exists
    if((file=fopen(script, "r"))) 
    {
// Now that we know the file exists, we start fiddling the menu entries
        themeItemsCount=CountMenuItems(menu);
        for(i=itemUserThemes;i<themeItemsCount;i++)
            DeleteMenuItem(menu,itemUserThemes+1);
        themeItemsCount=itemUserThemes;
        
// hack to easily convert the string to a pascal string
        buf++;
        
// now extract the lines from file (max 254 characters)
        while(fgets(buf, 254, file)) 
        {
// only use a line if it does not start with a hash and it is not blank
// (remember that a line includes the newline at the end of it)
            buflen=strlen(buf);
            if(buflen<2) {
                if(gVerboseConsole) printf("Short line: (%d) %s",buflen,buf);
            } else if(strncmp(buf,"#",1)) {
                if(gVerboseConsole) printf("%s info: Got an item: %s  ",PROGNAME,buf);
                // now hack it to a pascal string
                buf--;
                buf[0]=buflen;
                AppendMenu(menu,buf);
                buf++;
                themeItemsCount++;
            } else if(gVerboseConsole)
                printf("Comment: %s",buf); // shouldn't be any of these anyway...?
        } 
// close the file and get rid of it
        fclose(file);
        if(gVerboseConsole) {
            sprintf(script,"/bin/cat /tmp/themelist.%s",randomID);
            system(script);
        }
        sprintf(script,"/bin/rm -f /tmp/themelist.%s",randomID);
        printf("%s info: Removing file: %s\n",PROGNAME,script);
        system(script);
        
// make sure that we undo the hack before freeing this...
        buf--;

    } else
        printf("%s warning: themelist file not found: '%s'\n",PROGNAME,script);
    
// and make sure we have the right number of menu items
    themeItemsCount=CountMenuItems(menu);

    if(themeItemsCount<=itemUserThemes) {
        printf("%s: No Themes found!! Could end up fatal...\n",PROGNAME);
        ShowAlert(NOTHEMES_NAME,1,0);
        return;
    }
    
#ifdef DEBUG
    printf("About to free buf and script in Create_Themes_Menu...\n");
#endif
    free(buf);
    free(script);
#ifdef DEBUG
    printf("buf and script freed OK\n");
#endif

// search for the current theme, and set it
    if(setThemeDir!=0) {
        printf("%s info: searching for item %s...\n",PROGNAME,gThemeDir);
        setThemeDir=0;
        if(gThemeDir==NULL) {
            printf("%s info: Create_Themes_Menu with NULL theme dir set\n",PROGNAME);
        } else {
            for(i=itemUserThemes;i<themeItemsCount;i++) {
                GetMenuItemText(menu,i+1,itemStr);
                pStr=&itemStr;
                buflen=pStr->len;
                cStr=&(pStr->thestr);
                cStr[buflen]=0;
                if(gVerboseConsole) printf("%s info: Got item %d: %s\n",PROGNAME,i+1,cStr);
                if(buflen==strlen(gThemeDir))
                    if(!strncasecmp(cStr,gThemeDir,buflen))
                        setThemeDir=i+1;
            }
        }
        if(setThemeDir==0) {
            if(gThemeDir!=NULL) {
                printf("%s warning: Could not find current theme in menu: %s\n",PROGNAME,gThemeDir);
                ShowAlert(NOCURTHEME_NAME,1,0);
                gThemeDir=NULL;
            } else {
                printf("%s warning: No theme set; using first in menu...\n",PROGNAME);
                setThemeDir=itemUserThemes+1;
            }
        }
        if(setThemeDir!=0)
            Setup_Themes_Menu(setThemeDir);
    }
}

/*
 * Rather 'non-Macish' way to get a list of files in the Launch Menu Items folder,
 * and then use it to create the user items for the launch menu
 */
void Create_Launch_Menu()
{
    MenuRef	menu;
    int		i, buflen, err;
    char	*script=NULL, *buf=NULL;
    FILE	*file;
    
    menu = GetMenuHandle( menuLaunch );

    sprintf(lastsub,"TRACE1: Create_Launch_Menu\n");
    subtrace();
#ifdef SHOW_TRACE1    
    printf("TRACE1: Create_Launch_Menu\n");
#endif

#ifdef DEBUG    
    printf("About to malloc script & buf: 55 & 256\n");
#endif
    script=malloc(55);
    buf=malloc(256);
#ifdef DEBUG    
    printf("script,buf: %#lx,%#lx\n",(long) script, (long) buf);
#endif

    if(script==NULL || buf==NULL)
    {
        if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in Create_Launch_Menu.\n", PROGNAME);
            ClearUpSockets();
            exit(1);
        } else {
// need a nice alert here...
            fprintf(stderr, "%s: Memory allocation failed in Create_Launch_Menu.\n", PROGNAME);
            ShowAlert(FATALMEM_NAME,0,0);
            gQuitFlag=FATALMEM_ERRNO;
            return;
        }
    }
    
    sprintf(script,"Contents/Scripts/LaunchList '%s'",randomID);
    printf("%s info: Running: %s\n",PROGNAME,script);
// run script to get list of items...
    err=system(script);
    if(err==NOLMIFLDR_STAT)
        ShowAlert(NOLMIFLDR_NAME,1,0);

// This is the file that the script should have just created
    sprintf(script,"/tmp/launchlist.%s",randomID);
//#ifdef DEBUG
    printf("%s info: Launchlist file: '%s'\n",PROGNAME,script);
//#endif

// Now check to see if the file exists
    if((file=fopen(script, "r"))) 
    {
// Now that we know the file exists, we start fiddling the menu entries
        launchItemsCount=CountMenuItems(menu);
        for(i=itemUserItems;i<launchItemsCount;i++) {
            DeleteMenuItem(menu,itemUserItems+1); }
        launchItemsCount=itemUserItems;
        
// hack to easily convert the string to a pascal string
        buf++;
        
// now extract the lines from file (max 254 characters)
        while(fgets(buf, 254, file)) 
        {
// only use a line if it does not start with a hash and it is not blank
// (remember that a line includes the newline at the end of it)
            buflen=strlen(buf);
            if(buflen<2) {
                if(gVerboseConsole) printf("Short line: (%d) %s",buflen,buf);
            } else if(strncmp(buf,"#",1)) {
                if(gVerboseConsole) printf("%s info: Got an item: %s  ",PROGNAME,buf);
                // now hack it to a pascal string
                buf--;
                buf[0]=buflen;
                AppendMenu(menu,buf);
                buf++;
                // just have to reset these counters to zero, I think...
                // Without storing all the item names and checking to see how they have changed
                // it's difficult to keep track of which counters go where if the item list changes...
                launch_counter[launchItemsCount-itemUserItems]=0;
                launchItemsCount++;
            } else if(gVerboseConsole)
                printf("Comment: %s",buf); // shouldn't be any of these anyway...?
        } 
// close the file and get rid of it
        fclose(file);
        sprintf(script,"/bin/rm -f /tmp/launchlist.%s",randomID);
        printf("%s info: Removing file: %s\n",PROGNAME,script);
        system(script);
        
// make sure that we undo the hack before freeing this...
        buf--;

    } else
        printf("%s warning: launchlist file not found: '%s'\n",PROGNAME,script);
    
#ifdef DEBUG
    printf("About to free buf and script in Create_Launch_Menu...\n");
#endif
    free(buf);
    free(script);
#ifdef DEBUG
    printf("buf and script freed OK\n");
#endif
    
// and make sure we have the right number of menu items
    launchItemsCount=CountMenuItems(menu);
}

void MakeMenu()		/* Put up a menu */
{
    Handle	menuBar;
    MenuRef	menu;
    OSStatus	myStat;
    IBNibRef	theNIBRef;
    
    sprintf(lastsub,"TRACE1: MakeMenu\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: MakeMenu\n");
#endif

    myStat = CreateNibReference(CFSTR("menubar"),&theNIBRef);
    if(myStat!=0) {
        printf("%s: Failed to get menubar from application bundle! (err: %d)\n",PROGNAME,myStat);
        ShowAlert(BADMENUNIB_NAME,0,0);
        gQuitFlag=-1000;
        return;
    }
    
    myStat = SetMenuBarFromNib(theNIBRef,CFSTR("MainMenu"));
    DisposeNibReference(theNIBRef);
    if(myStat!=0) {
        printf("%s: Failed to get menu from menubar.nib! (err: %d)\n",PROGNAME,myStat);
        ShowAlert(BADMENUNIB_NAME,0,0);
        gQuitFlag=-1001;
        return;
    }
    
    gDoneSetupMenus=0;

}

void SetupMenus()
{
    Handle	menuBar;
    MenuRef	menu;
    int		(*i)[2];
    
    if(gDoneSetupMenus) return;
    
//    gHiding_Iconbars=1;
//    shrinking_iconbars=1;
//    gInterleaving_Windows=0;
    
//    iconbarpos=ICONBARPOS_TOPLEFT;
//    Setup_Options_Menu(itemTopLeft);
    Setup_Options_Menu(itemTopLeft + iconbarpos);

//    Setup_Options_Menu(itemMedIconbars);
    /* Set up the right iconbar_steps item. If none apply, don't set any. */
    for (i = steps_menu_items; i < steps_menu_items+sizeof(steps_menu_items); ++i) {
        if ((*i)[0] == iconbar_steps) {
            Setup_Options_Menu((*i)[1]);
            break;
        }
    }
        
// can't do Options menu items for the screens yet -xinerama is not ready. Done later, in EventLoop
//    Setup_Options_Menu(itemScreenListStart + iconbar_screen);

    Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    Create_Launch_Menu();
//    Create_Themes_Menu(0);
//    Setup_Themes_Menu(itemUserThemes+1);
    if (NULL == gThemeDir) {
        Create_Themes_Menu(0);        // Pick a default theme.
        Setup_Themes_Menu(itemUserThemes+1);
    } else {
        Create_Themes_Menu(1);        // Use the stored theme.
    }
    
    gDoneSetupMenus=1;
}

void Setup_Themes_Menu(int which)
{
    MenuRef	menu;
    int		cnt,i,namelen,kjkasjdhk;
    Str255	itemStr;
    pstr_t	*pStr;
    char	*cStr;
    OSStatus	stat;
    
    sprintf(lastsub,"TRACE3: Setup_Themes_Menu (which: %d)\n",which);
    subtrace();
//#ifdef SHOW_TRACE3
    printf("TRACE3: Setup_Themes_Menu (which: %d)\n",which);
//#endif

//    menu=GetMenuHandle( menuThemes );
    stat = GetMenuItemHierarchicalMenu( GetMenuHandle(menuOptions), itemThemeSubmenu, &menu );
    cnt=CountMenuItems(menu);
    
    if(gUseShapedTheme==0) {
//        printf("Removed mark from item %d (out of %d) on menu %#lx\n",itemToggleShaped,cnt,menu);
        CheckMenuItem(menu, itemToggleShaped, false );
    } else {
//        printf("Put checkmark on item %d (out of %d) on menu %#lx\n",itemToggleShaped,cnt,menu);
        CheckMenuItem(menu, itemToggleShaped, true );
    }
    
    if(which>itemUserThemes) {
        for(i=itemUserThemes+1;i<=cnt;i++)
            CheckMenuItem( menu, i, false );
        SetItemMark( menu, which, diamondMark );
        //printf("Put diamond mark on item: %d (out of %d)\n",which,cnt);
        // now get the name of the item and put it into gThemeDir
        GetMenuItemText(menu,which,itemStr);
        pStr=&itemStr;
        namelen=pStr->len;
        cStr=&(pStr->thestr);
        cStr[namelen]=0;
        if(gThemeDir!=NULL) free(gThemeDir);
        gThemeDir=strdup(cStr);
        printf("%s info: Set new theme directory from menu item: %s\n",PROGNAME,gThemeDir);
        GetMenuItemText(menu,itemAboutThisTheme,itemStr);
        pStr=&itemStr;
        namelen=pStr->len;
        cStr=&(pStr->thestr);
        cStr[namelen]=0;
        if(aboutThemeItem==NULL)
            aboutThemeItem=strdup(cStr);
        sprintf(cStr,"%s %s",aboutThemeItem,gThemeDir);
        namelen=strlen(cStr);
        pStr->len=namelen;
        SetMenuItemText(menu,itemAboutThisTheme,itemStr);
    }
}

void Setup_Help_Menu()
{
    MenuRef menu = GetMenuHandle( menuHelp );
    
    if(gVerboseConsole==0)
        CheckMenuItem( menu, itemVerbose, false );
    else
        CheckMenuItem( menu, itemVerbose, true );
}

void Setup_FullScreen_MenuOption(int ticked)
{
    MenuRef menu = GetMenuHandle( menuOptions );

    if(ticked)
        CheckMenuItem( menu, itemWholeScreenMode, true );
    else
        CheckMenuItem( menu, itemWholeScreenMode, false );
}

void Setup_Options_Menu(int which)
{
    MenuRef	menu;
    int		cnt,i,namelen;
    Str255	itemStr;
    pstr_t	*pStr;
    char	*cStr;

    printf("In Setup_Options_Menu: %d %d %d\n",which,gGotCVDHidden,itemAllowOffScreen);
    sprintf(lastsub,"TRACE3: Setup_Options_Menu (which: %d %d)\n",which,can_send_hackcmds);
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: Setup_Options_Menu (which: %d %d)\n",which,can_send_hackcmds);
#endif

    menu = GetMenuHandle( menuOptions );
    
    if(gQuietStartup==0)
        CheckMenuItem( menu, itemQuietStartup, false );
    else
        CheckMenuItem( menu, itemQuietStartup, true );
    
    if(gHiding_Iconbars==0 && can_send_hackcmds)
        CheckMenuItem( menu, itemHideIconbars, false );
    else {
        CheckMenuItem( menu, itemHideIconbars, true );
    }
    
    if(gMinMethod==minmethod_dock && can_send_hackcmds)
        CheckMenuItem( menu, itemMinToDock, true );
    else {
        CheckMenuItem( menu, itemMinToDock, false );
    }
    
    if(gInterleaving_Windows!=0 && can_send_hackcmds!=0)
        CheckMenuItem( menu, itemInterleave, true );
    else
        CheckMenuItem( menu, itemInterleave, false );
    
    if(can_send_hackcmds==0 || gGotCVDHidden==0)
        DisableMenuItem( menu, itemAllowOffScreen );
    else {
        EnableMenuItem( menu, itemAllowOffScreen );
        if(gIncludeOffScreen==0)
            CheckMenuItem( menu, itemAllowOffScreen, false );
        else
            CheckMenuItem( menu, itemAllowOffScreen, true );
    }
    
    if(can_send_hackcmds==0) {
        DisableMenuItem( menu, itemHideIconbars );
        DisableMenuItem( menu, itemMinToDock );
        DisableMenuItem( menu, itemInterleave );
        DisableMenuItem( menu, itemBGWinDimmed );
        DisableMenuItem( menu, itemBGWinTranslucent );
    } else {
        EnableMenuItem( menu, itemHideIconbars );
        EnableMenuItem( menu, itemMinToDock );
        EnableMenuItem( menu, itemInterleave );
        DisableMenuItem( menu, itemBGWinDimmed );	// doesn't work yet...
        EnableMenuItem( menu, itemBGWinTranslucent );
        
        CheckMenuItem( menu, itemBGWinDimmed, false );
        CheckMenuItem( menu, itemBGWinTranslucent, false );

        if((gBGWinSetting & BGWIN_DIMMED)==BGWIN_DIMMED)
            CheckMenuItem( menu, itemBGWinDimmed, true );
        if((gBGWinSetting & BGWIN_TRANSLUCENT)==BGWIN_TRANSLUCENT)
            CheckMenuItem( menu, itemBGWinTranslucent, true );
    }
    
//    if(gMenubarSpace==0)
//        CheckMenuItem( menu, itemLeaveMenubar, false );
//    else
//        CheckMenuItem( menu, itemLeaveMenubar, true );

    if(shrinking_iconbars==0)
        CheckMenuItem( menu, itemFancyIconbars, false );
    else
        CheckMenuItem( menu, itemFancyIconbars, true );
    
    if(which>=itemScreenListStart) {
        cnt=CountMenuItems(menu);
        if(cnt>itemScreenListStart) {
            for(i=itemScreenListStart+1;i<=cnt;i++)
                DeleteMenuItem( menu, itemScreenListStart+1 ); }
        cnt=count_xinerama_heads();
        printf("%s info: Screens for Options menu: %d (%d)\n",PROGNAME,cnt,itemScreenListStart);
        if(cnt>1) {
            GetMenuItemText(menu,itemScreenListStart,itemStr);
// hack (again) to get the string out of the pascal string
// - there's probably a macro to do this 'legally' somewhere!
            pStr=&itemStr;
            namelen=pStr->len;
            cStr=&(pStr->thestr);
            for(i=1;i<cnt;i++) {
// strip off the screen number from the string
                cStr[namelen-1]=0;
                printf("  adding %d\n",i);
                sprintf(cStr,"%s%d",cStr,i);
                pStr->len=strlen(cStr);
                AppendMenu(menu,itemStr);
            }
        }
        cnt+=itemScreenListStart;
        if(which<=cnt) {
            for(i=itemScreenListStart;i<cnt;i++)
                CheckMenuItem( menu, i, false );
            SetItemMark( menu, which, diamondMark );
        }
    } else if(which>=itemTopLeft) {
        CheckMenuItem( menu, itemTopLeft, false );
        CheckMenuItem( menu, itemTopRight, false );
        CheckMenuItem( menu, itemBotLeft, false );
        CheckMenuItem( menu, itemBotRight, false );
        SetItemMark( menu, which, diamondMark );
    } else if(which>=itemSlowIconbars) {
        CheckMenuItem( menu, itemSlowIconbars, false );
        CheckMenuItem( menu, itemMedIconbars, false );
        CheckMenuItem( menu, itemFastIconbars, false );
        SetItemMark( menu, which, diamondMark );
    } 
}

void Setup_Properties_Menu_Translucency(OSXWinInfo *winfo)
{
    MenuRef	menu, propmenu;
    OSStatus	stat;

    
    menu = GetMenuHandle( menuWindow );
    stat = GetMenuItemHierarchicalMenu( menu, itemProperties, &propmenu );
    
    CheckMenuItem(propmenu,itemTranslucent0,false);
    CheckMenuItem(propmenu,itemTranslucent1,false);
    CheckMenuItem(propmenu,itemTranslucent2,false);
    
    if(winfo==NULL || can_send_hackcmds==0) {
        DisableMenuItem(propmenu,itemTranslucent0);
        DisableMenuItem(propmenu,itemTranslucent1);
        DisableMenuItem(propmenu,itemTranslucent2);
    } else {
        EnableMenuItem(propmenu,itemTranslucent0);
        EnableMenuItem(propmenu,itemTranslucent1);
        EnableMenuItem(propmenu,itemTranslucent2);
        if(winfo->translucency==ACTIVE_TRANSLUCENT0)
            SetItemMark(propmenu,itemTranslucent0,diamondMark);
        else if(winfo->translucency==ACTIVE_TRANSLUCENT1)
            SetItemMark(propmenu,itemTranslucent1,diamondMark);
        else if(winfo->translucency==ACTIVE_TRANSLUCENT2)
            SetItemMark(propmenu,itemTranslucent2,diamondMark);
    }
}

void Setup_Group_Menu(int enabled)
{
    int		cnt, namelen, i;
    char	*name=NULL, *cStr;
    MenuRef	menu, groupmenu;
    OSStatus	stat;
    Str255	itemStr;
    pstr_t	*pStr;
    
    menu = GetMenuHandle( menuWindow );
    stat = GetMenuItemHierarchicalMenu( menu, itemGroupSubmenu, &groupmenu );

        GetMenuItemText(groupmenu,itemOriginalGroup,itemStr);
        pStr=&itemStr;
        namelen=pStr->len;
        cStr=&(pStr->thestr);
        cStr[namelen]=0;

        if(orgGroupItem==NULL)
            printf("%s info: Starting out orgGroupItem as %s\n",PROGNAME,cStr);
        if(orgGroupItem==NULL)
            orgGroupItem=strdup(cStr);

        name=NULL;
        if(enabled) {
#ifndef DISABLE_CHANGEGROUP_MENU
            EnableMenuItem(menu,itemGroupSubmenu);
            EnableMenuItem(groupmenu,0);
#endif
            name=get_focus_client_grouporg_name();
            if(name==NULL) {
                name=get_focus_client_group_name();
                DisableMenuItem(groupmenu,itemOriginalGroup);
            } else
                EnableMenuItem(groupmenu,itemOriginalGroup);
#ifdef DISABLE_CHANGEGROUP_MENU
            DisableMenuItem(menu,itemGroupSubmenu);
            DisableMenuItem(groupmenu,0);
#endif
        } else {
            DisableMenuItem(menu,itemGroupSubmenu);
            DisableMenuItem(groupmenu,0);
        }
        if(name==NULL) {
            sprintf(cStr,"%s",orgGroupItem);
            DisableMenuItem(groupmenu,itemOriginalGroup);
        } else {
            sprintf(pStr->thestr,"%s: \"%s\"",orgGroupItem,name);
            cStr=&(pStr->thestr);
        }
        
    if(gVerboseConsole)
        printf("%s info: Setting orgGroupItem to %s\n",PROGNAME,cStr);
    namelen=strlen(cStr);
    pStr->len=namelen;
    SetMenuItemText(groupmenu,itemOriginalGroup,itemStr);
    
    namelen = 0;
    if(name!=NULL) namelen = strlen(name);
    cnt=CountMenuItems(groupmenu);
    
    for(i=itemStartGroups;i<=cnt;i++) {
        CheckMenuItem(groupmenu,i,false);
        if(name!=NULL) {
            GetMenuItemText(groupmenu,i,itemStr);
            pStr=&itemStr;
            if(namelen==pStr->len)
                if(!strncmp(pStr->thestr,name,namelen))
                    SetItemMark(groupmenu,i,diamondMark);
        }
    }
}

void Add_Group_to_Menu(char *name)
{
    MenuRef	menu, groupmenu;
    OSStatus	stat;
    Str255	itemStr;
    pstr_t	*pStr;
    int 	namelen, cnt;

    if(name==NULL) return;
    
    menu = GetMenuHandle( menuWindow );
    stat = GetMenuItemHierarchicalMenu( menu, itemGroupSubmenu, &groupmenu );

    cnt=CountMenuItems(groupmenu);
    printf("Adding \"%s\" to group menu (%d)\n",name,cnt);
        
    pStr=&itemStr;
    sprintf(pStr->thestr,"%s",name);
    namelen=strlen(name);
    pStr->len=namelen;
    AppendMenu(groupmenu,itemStr);
}

void Remove_Group_from_Menu(char *name)
{
    MenuRef	menu, groupmenu;
    OSStatus	stat;
    Str255	itemStr;
    pstr_t	*pStr;
    int 	namelen, cnt, i;

    if(name==NULL) return;

    printf("Removing \"%s\" from group menu\n",name);
    namelen = strlen(name);
    menu = GetMenuHandle( menuWindow );
    stat = GetMenuItemHierarchicalMenu( menu, itemGroupSubmenu, &groupmenu );

    cnt=CountMenuItems(groupmenu);
    
    for(i=0;i<cnt;i++) {
        GetMenuItemText(groupmenu,itemStartGroups+i,itemStr);
        pStr=&itemStr;
        if(pStr->len==namelen)
            pStr->thestr[namelen]=0;
            if(!strncmp(pStr->thestr,name,namelen)) {
                DeleteMenuItem(groupmenu,itemStartGroups+i);
                return;
            }
    }
    
    printf("%s warning: failed to find group %s to remove from menu!\n",PROGNAME,name);
}

void Setup_Window_Menu(int all_off, int has_frame, int has_close, int has_min, int has_zoom, int is_float, int can_af, int can_ar, int can_al, int first_click, int click_raise, int live_resize, int no_grab_focus, int retain_focus, int client_modality)
{
    MenuRef	menu, propmenu, groupmenu;
    OSStatus	stat;
    long	nextwin;
    Str255	itemStr;
    pstr_t	*pStr;
    char	*cStr, *full_title=NULL;
    int		namelen;
    
    sprintf(lastsub,"TRACE3: Setup_Window_Menu (has_frame,can_af,can_ar,can_al: %d,%d,%d,%d)\n",
            has_frame,can_af,can_ar,can_al);
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: Setup_Window_Menu (has_frame,can_af,can_ar,can_al: %d,%d,%d,%d)\n",
            has_frame,can_af,can_ar,can_al);
#endif

    menu = GetMenuHandle( menuWindow );
    stat = GetMenuItemHierarchicalMenu( menu, itemProperties, &propmenu );
    stat = GetMenuItemHierarchicalMenu( menu, itemGroupSubmenu, &groupmenu );

    if(all_off==0) {
        EnableMenuItem(menu,0);
        DisableMenuItem(menu,itemWindowshade);
        DisableMenuItem(menu,itemMinimise);
        DisableMenuItem(menu,itemZoom);
        DisableMenuItem(menu,itemClose);
        DisableMenuItem(menu,itemLower);
        DisableMenuItem(menu,itemNext);
        DisableMenuItem(menu,itemLowerGroup);
        DisableMenuItem(menu,itemNextGroup);
        DisableMenuItem(menu,itemHideGroup);
        cycleOff=1;
        groupCycleOff=1;
        DisableMenuItem(menu,itemProperties);
        DisableMenuItem(propmenu,0);
        Setup_Group_Menu(0);
        CheckMenuItem(propmenu,itemFirstClick,false);
        CheckMenuItem(propmenu,itemFirstClickChild,false);
        CheckMenuItem(propmenu,itemFirstClickDrag,false);
        CheckMenuItem(propmenu,itemIgnoreKeys,false);
        CheckMenuItem(propmenu,itemIgnoreClicks,false);
        CheckMenuItem(propmenu,itemAutoFocus,false);
        CheckMenuItem(propmenu,itemAutoRaise,false);
        CheckMenuItem(propmenu,itemAutoLower,false);
        CheckMenuItem(propmenu,itemClickRaise,false);
        CheckMenuItem(propmenu,itemPopup,false);
        CheckMenuItem(propmenu,itemNoframes,false);
        CheckMenuItem(propmenu,itemModal,false);
        CheckMenuItem(propmenu,itemLiveResize,false);
        CheckMenuItem(propmenu,itemNoGrabFocus,false);
        CheckMenuItem(propmenu,itemRetainFocus,false);
//        DisableMenuItem(menu,0);
        if( Is_Whole_Screen_Mode()==0 ) DrawMenuBar();
        return;
    }
    
    EnableMenuItem(menu,0);

        // the "Change Group" submenu is disabled at the moment
    Setup_Group_Menu(1);

    EnableMenuItem(propmenu,itemNoframes);
    if(has_frame==0) {
        CheckMenuItem(propmenu,itemNoframes,true);
        DisableMenuItem(menu,itemWindowshade);
    } else {
        CheckMenuItem(propmenu,itemNoframes,false);
        EnableMenuItem(menu,itemWindowshade);
    }
    EnableMenuItem(menu,itemMinimise);
    EnableMenuItem(menu,itemZoom);
    EnableMenuItem(menu,itemClose);
    EnableMenuItem(menu,itemLower);
    
        GetMenuItemText(menu,itemNext,itemStr);
        pStr=&itemStr;
        namelen=pStr->len;
        cStr=&(pStr->thestr);
        cStr[namelen]=0;
        if(nextItem==NULL)
            printf("%s info: Starting out nextItem as %s\n",PROGNAME,cStr);
        if(nextItem==NULL)
            nextItem=strdup(cStr);
        full_title=NULL;
//        nextwin=get_next_osxwinref(0);
        full_title=get_next_title_in_group(0);
        if(full_title==NULL) {
            sprintf(cStr,"%s",nextItem);
            DisableMenuItem(menu,itemNext);
            DisableMenuItem(menu,itemLower);
            cycleOff=1;
        } else {
            cycleOff=0;
            sprintf(pStr->thestr,"%s: \"%s\"",nextItem,full_title);
            free(full_title);
            cStr=&(pStr->thestr);
            EnableMenuItem(menu,itemNext);
        }
/*        } else {
            cycleOff=0;
            GetWTitle((WindowRef) nextwin,itemStr);
            pStr=&itemStr;
            namelen=pStr->len;
            cStr=&(pStr->thestr);
            cStr[namelen]=0;
            cStr=strdup(pStr->thestr);
            sprintf(pStr->thestr,"%s:  \"%s\"",nextItem,cStr);
            free(cStr);
            cStr=&(pStr->thestr);
            EnableMenuItem(menu,itemNext);
        }*/
    if(gVerboseConsole)
        printf("%s info: Setting next item to %s\n",PROGNAME,cStr);

    namelen=strlen(cStr);
    pStr->len=namelen;
    SetMenuItemText(menu,itemNext,itemStr);
        
        GetMenuItemText(menu,itemNextGroup,itemStr);
        pStr=&itemStr;
        namelen=pStr->len;
        cStr=&(pStr->thestr);
        cStr[namelen]=0;
        if(nextGroupItem==NULL)
            printf("%s info: Starting out nextGroupItem as %s\n",PROGNAME,cStr);
        if(nextGroupItem==NULL)
            nextGroupItem=strdup(cStr);
        full_title=NULL;
        full_title=get_next_group_name();
        if(full_title==NULL) {
            sprintf(cStr,"%s",nextGroupItem);
            DisableMenuItem(menu,itemHideGroup);
            DisableMenuItem(menu,itemNextGroup);
            DisableMenuItem(menu,itemLowerGroup);
            groupCycleOff=1;
        } else {
            groupCycleOff=0;
            sprintf(pStr->thestr,"%s: \"%s\"",nextGroupItem,full_title);
            free(full_title);
            cStr=&(pStr->thestr);
            EnableMenuItem(menu,itemHideGroup);
            EnableMenuItem(menu,itemNextGroup);
            EnableMenuItem(menu,itemLowerGroup);
        }

    if(gVerboseConsole)
        printf("%s info: Setting next group item to %s\n",PROGNAME,cStr);

    namelen=strlen(cStr);
    pStr->len=namelen;
    SetMenuItemText(menu,itemNextGroup,itemStr);
        
    EnableMenuItem(menu,itemProperties);

    if(has_close==0)
        DisableMenuItem(menu,itemClose);
    
    if(has_min==0)
        DisableMenuItem(menu,itemMinimise);

    if(has_zoom==0)
        DisableMenuItem(menu,itemZoom);

    EnableMenuItem(propmenu,0);

    EnableMenuItem(propmenu,itemFloat);
    EnableMenuItem(propmenu,itemAutoFocus);
    EnableMenuItem(propmenu,itemAutoRaise);
    EnableMenuItem(propmenu,itemModal);

    if(no_grab_focus==1)
        CheckMenuItem(propmenu,itemNoGrabFocus,true);
    else
        CheckMenuItem(propmenu,itemNoGrabFocus,false);
        
    if(retain_focus==1)
        CheckMenuItem(propmenu,itemRetainFocus,true);
    else
        CheckMenuItem(propmenu,itemRetainFocus,false);
        
    if(is_float==0)
        DisableMenuItem(propmenu,itemFloat);
    else if(is_float==1)
        CheckMenuItem(propmenu,itemFloat,true);
    else if(is_float==-1)
        CheckMenuItem(propmenu,itemFloat,false);
        
    if(can_al==0) {
        CheckMenuItem(propmenu,itemAutoLower,false);
        DisableMenuItem(propmenu,itemAutoLower);
    } else if(can_al==1)
        CheckMenuItem(propmenu,itemAutoLower,true);
    else if(can_al==-1)
        CheckMenuItem(propmenu,itemAutoLower,false);

    if(client_modality==0) {
        CheckMenuItem(propmenu,itemModal,false);
        DisableMenuItem(propmenu,itemModal);
    } else if(client_modality==1)
        CheckMenuItem(propmenu,itemModal,true);
    else if(client_modality==-1)
        CheckMenuItem(propmenu,itemModal,false);

    if(can_ar==0) {
        CheckMenuItem(propmenu,itemAutoRaise,false);
        DisableMenuItem(propmenu,itemAutoRaise);
        DisableMenuItem(propmenu,itemAutoLower);
        DisableMenuItem(propmenu,itemClickRaise);
        DisableMenuItem(propmenu,itemFirstClick);
        DisableMenuItem(propmenu,itemFirstClickChild);
    } else if(can_ar==1) {
        CheckMenuItem(propmenu,itemAutoRaise,true);
        EnableMenuItem(propmenu,itemAutoLower);
        DisableMenuItem(propmenu,itemClickRaise);
        DisableMenuItem(propmenu,itemFirstClick);
        DisableMenuItem(propmenu,itemFirstClickChild);
    } else if(can_ar==-1) {
        CheckMenuItem(propmenu,itemAutoRaise,false);
        DisableMenuItem(propmenu,itemAutoLower);
        EnableMenuItem(propmenu,itemClickRaise);
        EnableMenuItem(propmenu,itemFirstClick);
        EnableMenuItem(propmenu,itemFirstClickChild);
    }
    
    if(click_raise==0)
        CheckMenuItem(propmenu,itemClickRaise,false);
    else
        CheckMenuItem(propmenu,itemClickRaise,true);

    if(live_resize==0)
        CheckMenuItem(propmenu,itemLiveResize,false);
    else
        CheckMenuItem(propmenu,itemLiveResize,true);

    if(can_af==0) {
        CheckMenuItem(propmenu,itemAutoFocus,false);
        DisableMenuItem(propmenu,itemAutoFocus);
        DisableMenuItem(propmenu,itemAutoRaise);
        DisableMenuItem(propmenu,itemAutoLower);
        DisableMenuItem(propmenu,itemClickRaise);
        DisableMenuItem(propmenu,itemFirstClick);
        DisableMenuItem(propmenu,itemFirstClickChild);
        DisableMenuItem(propmenu,itemFirstClickDrag);
        DisableMenuItem(propmenu,itemRetainFocus);
        DisableMenuItem(propmenu,itemNoGrabFocus);
        DisableMenuItem(propmenu,itemIgnoreClicks);
        DisableMenuItem(propmenu,itemIgnoreKeys);
    } else if(can_af==1) {
        CheckMenuItem(propmenu,itemAutoFocus,true);
        DisableMenuItem(propmenu,itemFirstClick);
        DisableMenuItem(propmenu,itemFirstClickChild);
        DisableMenuItem(propmenu,itemFirstClickDrag);
        DisableMenuItem(propmenu,itemRetainFocus);
        EnableMenuItem(propmenu,itemNoGrabFocus);
        EnableMenuItem(propmenu,itemIgnoreClicks);
        DisableMenuItem(propmenu,itemIgnoreKeys);
    } else if(can_af==-1) {
        CheckMenuItem(propmenu,itemAutoFocus,false);
        DisableMenuItem(propmenu,itemAutoRaise);
        DisableMenuItem(propmenu,itemAutoLower);
        DisableMenuItem(propmenu,itemClickRaise);
        EnableMenuItem(propmenu,itemFirstClick);
        EnableMenuItem(propmenu,itemFirstClickChild);
        EnableMenuItem(propmenu,itemFirstClickDrag);
        EnableMenuItem(propmenu,itemRetainFocus);
        EnableMenuItem(propmenu,itemIgnoreClicks);
        DisableMenuItem(propmenu,itemIgnoreKeys);   // Will be Enabled at some point...
    }
    
    if((first_click & 1)==1)
        CheckMenuItem(propmenu,itemFirstClick,true);
    else
        CheckMenuItem(propmenu,itemFirstClick,false);
    
    if((first_click & 2)==2)
        CheckMenuItem(propmenu,itemFirstClickChild,true);
    else
        CheckMenuItem(propmenu,itemFirstClickChild,false);
    
    if((first_click & 4)==4) {
        CheckMenuItem(propmenu,itemFirstClickDrag,true);
        DisableMenuItem(propmenu,itemFirstClick);
        DisableMenuItem(propmenu,itemFirstClickChild);
    } else
        CheckMenuItem(propmenu,itemFirstClickDrag,false);
    
    if((first_click & 8)==8) {
        CheckMenuItem(propmenu,itemIgnoreClicks,true);
        DisableMenuItem(propmenu,itemFirstClick);
        DisableMenuItem(propmenu,itemFirstClickChild);
        if(can_af!=0) EnableMenuItem(propmenu,itemFirstClickDrag);
    } else {
        CheckMenuItem(propmenu,itemIgnoreClicks,false);
        if(first_click & 3) {
            DisableMenuItem(propmenu,itemFirstClickDrag);
            DisableMenuItem(propmenu,itemIgnoreClicks);
            if(first_click==2)
                DisableMenuItem(propmenu,itemFirstClick);
            else if(first_click==1)
                DisableMenuItem(propmenu,itemFirstClickChild);
        }
    }
    
    if((first_click & 16)==16)
        CheckMenuItem(propmenu,itemIgnoreKeys,true);
    else
        CheckMenuItem(propmenu,itemIgnoreKeys,false);
    
    if( Is_Whole_Screen_Mode()==0 ) DrawMenuBar();
}

void Launch_Startup_Items()
{
    char *script=NULL;
    char buf[255];
    FILE *file;
    int buflen, err, gotbad=0;

    sprintf(lastsub,"TRACE1: Launch_Startup_Items\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: Launch_Startup_Items\n");
#endif

#ifdef DEBUG
    printf("About to malloc script in Launch_Startup_Items: 60\n");
#endif
    script=malloc(60);
#ifdef DEBUG
    printf("script: %#lx\n",(long) script);
#endif
    if(script==NULL)
    {
	if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in Launch_Startup_Items.\n", PROGNAME);
            return;
        } else {
            fprintf(stderr, "%s: Memory allocation failed in Launch_Startup_Items.\n", PROGNAME);
            ShowAlert(WARNMEM_NAME,1,0);
            return;
        }
    }

// Tell the EditStartups script to make a temporary copy of the StartupItems file
// (This lets the script do the work of finding out if it's the user's personal one)
    sprintf(script,"Contents/Scripts/EditStartups 'copy' '%s'",randomID);
    printf("%s info: Running: %s\n",PROGNAME,script);
    err=system(script);
    if(err==NOLMIFLDR_STAT)
        ShowAlert(NOLMIFLDR_NAME,1,0);
    else if(err==NOTEXIST_STAT)
        ShowAlert(NOTEXISTLAUNCH_NAME,1,0);
    else if(err!=0) {
        printf("Script returned with error: %d\n",err);
        ShowAlert(WARNLAUNCH_NAME,1,0);
    }
    
// This is the name of the temporary file
    sprintf(script,"/tmp/StartupItems.%s",randomID);
#ifdef DEBUG
    printf("Temporary StartupItems file: %s\n",script);
#endif
    
// Check to see if the file exists
    if((file=fopen(script, "r"))) 
    {        
// and extract each line from the file (max 254 characters for a line)
        while(fgets(buf, 254, file)) 
        {
// only use a line if it does not start with a hash and it is not blank
// (remember that a line includes the newline at the end of it)
            buflen=strlen(buf);
            if(buflen<2) {
#ifdef DEBUG
                printf("Short line: (%d)  %s",buflen,buf);
#endif
            } else if(strncmp(buf,"#",1)) {
// Now try to launch it (must be the same as an item in the Launch menu)
                printf("%s info: Got an item: (%d)  %s",PROGNAME,buflen,buf);
                if(LaunchMenuItem(buf,0,1,1)!=0) gotbad++;
#ifdef DEBUG
            } else {
                printf("Comment: %s",buf);
#endif
            }
        } 
// close the file
        fclose(file);
// and remove it
        sprintf(script,"/bin/rm -rf /tmp/StartupItems.%s",randomID);
//#ifdef DEBUG
        printf("%s info: removing temporary file: %s\n",PROGNAME,script);
//#endif
        system(script);
        if(gotbad==1)
            ShowAlert(LAUNCHFAIL1_NAME,1,0);
        else if(gotbad>1)
            ShowAlert(LAUNCHFAILS_NAME,1,0);
    } else
        printf("%s warning: temporary StartupItems file not found: '%s'\n",PROGNAME,script);
    
#ifdef DEBUG
    printf("About to free script in Launch_Startup_Items...\n");
#endif
    free(script);
#ifdef DEBUG
    printf("script freed OK\n");
#endif
}

void Edit_Rules_File()
{
    char	*script=NULL;
    int		err;
    
    script=malloc(75);
    if(script==NULL) {
        fprintf(stderr, "%s: Memory allocation failed in Edit_Rules_File.\n", PROGNAME);
        if(!on_cmd_line) ShowAlert(WARNMEM_NAME,1,0);
        return;
    }
    
    sprintf(script,"/usr/bin/open -a TextEdit ~/Library/Preferences/OroborOSX/default.rules");
    
    err=system(script);
    if(err) fprintf(stderr,"%s : error %d trying to open rules file\n",PROGNAME,err);
    
    free(script);
}

void Read_Rules_File()
{
    char	*script=NULL;
    int		err;
    
    script=malloc(90);
    if(script==NULL) {
        fprintf(stderr, "%s: Memory allocation failed in Read_Rules_List.\n", PROGNAME);
        if(!on_cmd_line) ShowAlert(WARNMEM_NAME,1,0);
        return;
    }
    
    sprintf(script,"/bin/cp ~/Library/Preferences/OroborOSX/default.rules /tmp/oroborosx-rules.%s",randomID);
    
    err=system(script);
    if(err)
        fprintf(stderr,"%s : error %d trying to copy rules file\n",PROGNAME,err);
    else
        init_rules_list();
        
    free(script);
}

void Edit_Menu_Items(char *scriptname)
{
    char	*script=NULL;
    int		err;

    sprintf(lastsub,"TRACE1: Edit_Menu_Items: %s\n",scriptname);
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: Edit_Menu_Items: %s\n",scriptname);
#endif

#ifdef DEBUG
    printf("About to malloc script in Edit_Menu_Items: 45\n");
#endif
    script=malloc(25+strlen(scriptname));
#ifdef DEBUG
    printf("script: %#lx\n",(long) script);
#endif
    if(script==NULL) {
        fprintf(stderr, "%s: Memory allocation failed in Edit_Menu_Items.\n", PROGNAME);
        if(!on_cmd_line) ShowAlert(WARNMEM_NAME,1,0);
        return;
    }

// crude editing method - the LaunchList/ThemeList/FindHelper script just opens up the
// folder for the user to fiddle with...
    sprintf(script,"Contents/Scripts/%s open open open",scriptname);
    printf("%s info: Running: %s\n",PROGNAME,script);
    err=system(script);
    if(err==NOLMIFLDR_STAT)
        ShowAlert(NOLMIFLDR_NAME,1,0);
    else if(err==NOTHEMEFLDR_STAT)
        ShowAlert(NOTHEMEFLDR_NAME,1,0);
    
#ifdef DEBUG
    printf("About to free script in Edit_Menu_Items...\n");
#endif
    free(script);
#ifdef DEBUG
    printf("script freed OK\n");
#endif

// Once OroborOSX is brought in front, it will recheck the Launch Menu Items folder
// and continue to do so every ten seconds for about a minute - hopefully this will
// allow it to catch any changes that are made to the items in the folder...
// switched off -use "Rebuild Launch Menu" item in the Launch menu...
// or the "Rebuild Themes Menu" item in the Themes menu...
//    reReadOnNextFront=-6;
    lastOKTicks=TickCount();
}

int LaunchItem(int item, int inhibit)
{
    char argstr[128];
    int arglen=0, err=0;
    
    sprintf(lastsub,"TRACE2: LaunchItem (item: %d)\n",item);
    subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: LaunchItem (item: %d)\n",item);
#endif

    sprintf(argstr,"");
    
    switch (item)
    {
        case itemEditStart:
            Edit_Startup_Items();
            break;
            
        case itemNewItem:
            printf("How did you do that!! -not yet available.\n");
            break;
            
        case itemEditItems:
            Edit_Menu_Items("LaunchList");
            break;
            
        case itemEditHelpers:
            Edit_Menu_Items("FindHelper");
            break;
            
        case itemRebuildMenu:
            Create_Launch_Menu();
            break;
            
        case itemxinitrc:
            err=RunX11App("xinitrc",7,"",0,inhibit);
            break;
            
// These have all been replaced by x11app items in the Launch Menu Items folder...
/*
        case itemxterm:
            if(launch_count_xterm==1)
                sprintf(argstr,"-ls -sb -sl 5000 -name \"`/bin/hostname`\"");
            else
                sprintf(argstr,"-ls -sb -sl -name \"`/bin/hostname` (%d)\"",launch_count_xterm);
            arglen=strlen(argstr);
            RunX11App("xterm",5,argstr,arglen);
            launch_count_xterm++;
            break;
        
        case itemxedit:
            if(launch_count_xedit==1)
                sprintf(argstr,"-name \"xedit\"");
            else
                sprintf(argstr,"-name \"xedit (%d)\"",launch_count_xedit);
            arglen=strlen(argstr);
            RunX11App("xedit",5,argstr,arglen);
            launch_count_xedit++;
            break;
        
        case itemxbiff:
            if(launch_count_xbiff==1)
                sprintf(argstr,"-geometry 50x50-0+0 -file /var/mail/`/usr/bin/whoami` -title \"mail:`/usr/bin/whomai`\"");
            else
                sprintf(argstr,"-geometry 50x50-0+0 -file /var/mail/`/usr/bin/whoami` -title \"mail:`/usr/bin/whomai` (%d)\"",launch_count_xbiff);
            arglen=strlen(argstr);
            RunX11App("xbiff",5,argstr,arglen);
            launch_count_xbiff++;
            break;
        
        case itemxclock:
            if(launch_count_xclock==1)
                sprintf(argstr,"-geometry 70x70-0-0 -title \"xclock\"");
            else
                sprintf(argstr,"-geometry 70x70-0-0 -title \"xclock (%d)\"",launch_count_xclock);
            arglen=strlen(argstr);
            RunX11App("xclock",6,argstr,arglen);
            launch_count_xclock++;
            break;
        
        case itemxeyes:
            if(launch_count_xeyes==1)
                sprintf(argstr,"-geometry 100x60-50+0 +shape -fg black -bg skyblue -outline black -center white -title \"xeyes\"");
            else
                sprintf(argstr,"-geometry 100x60-50+0 +shape -fg black -bg skyblue -outline black -center white -title \"xeyes (%d)\"",launch_count_xeyes);
            arglen=strlen(argstr);
            RunX11App("xeyes",5,argstr,arglen);
            launch_count_xeyes++;
            break;
        
        case itemxcalc:
            if(launch_count_xcalc==1)
                sprintf(argstr,"-title \"Calculator\"");
            else
                sprintf(argstr,"-title \"Calculator (%d)\"",launch_count_xcalc);
            arglen=strlen(argstr);
            RunX11App("xcalc",5,argstr,arglen);
            launch_count_xcalc++;
            break;
*/            
        default:
            if(item>itemUserItems)
                err=RunX11Script(item,inhibit,NULL);
            else {
                printf("What menu item was that? %d\n",item);
                err=-1234;
            }
            break;
    }
    
    return err;
}

/*
 * Launch an item from the Launch menu given its name
 */
 
int LaunchMenuItem(char *itemname, int hasext, int hasnl, int inhibit)
{
    MenuRef	menu;
    Str255	itemStr;
    int		namelen, item, itemlen, err=0;
    pstr_t	*pStr;
    char	*cStr;

    itemlen=strlen(itemname);
    
    sprintf(lastsub,"TRACE2: LaunchMenuItem (itemname,itemlen,hasext: %s , %d,%d)\n",itemname,itemlen,hasext);
    subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: LaunchMenuItem (itemname,itemlen,hasext: %s , %d,%d)\n",itemname,itemlen,hasext);
#endif

// Strip off the .x11app extension, if necessary
    if(hasext!=0) {
        if(itemlen<8) return -2;
        itemlen-=7; }
        
// And strip off the newline, if necessary
    if(hasnl!=0) {
        if(itemlen<2) return -2;
        itemlen-=1; }
    
    menu = GetMenuHandle( menuLaunch );

// Loop through the items in the Launch menu, looking for a match
    for(item=itemxinitrc+1;item<=launchItemsCount;item++) {
        GetMenuItemText(menu,item,itemStr);
    
// use that nasty hack again, to get the string out of the pascal string
// - there's probably a macro to do this 'legally' somewhere!
        pStr=&itemStr;
        namelen=pStr->len;
        cStr=&(pStr->thestr);
        cStr[namelen]=0;
                
// Now compare the strings, and if they match Launch that item
        if(namelen==itemlen)
            if(!strncasecmp(cStr,itemname,itemlen)) {
                err=LaunchItem(item,inhibit);
                return err; }
    }
    
// failed to find item in user-editable section of Launch menu- check for xinitrc
    if(strstr(itemname, "xinitrc")) {
    	err=LaunchItem(itemxinitrc,inhibit);
        return err;
    }
    
// not xinitrc, so failed to find it...
    return -1;
}

void Change_Theme(int toggleShape, int themeNum)
{
    if(themeNum==0)
        gUseShapedTheme=toggleShape;
    Setup_Themes_Menu(themeNum);
    reload_config();
}

void Put_Into_Group(int which)
{
}

void Create_New_Group(void)
{
}

void DoMenuCommand(long menuResult, short menuID, short menuItem)
{
//    short	menuID;		/* the resource ID of the selected menu */
//    short	menuItem;	/* the item number of the selected menu */
    
    sprintf(lastsub,"TRACE1: DoMenuCommand (menuResult: %#lx)\n",menuResult);
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: DoMenuCommand (menuResult: %#lx)\n",menuResult);
#endif

#ifdef SHOW_EVENTS
    printf("Got menuID, menuItem: %d, %d (%#lx)\n",menuID,menuItem,menuResult);
#else
    if(gVerboseConsole)
      printf("Got menuID, menuItem: %d, %d (%#lx)\n",menuID,menuItem,menuResult);
#endif

    if(menuID==0 || menuItem==0) {
        if(menuResult==0) return;
        menuID = HiWord(menuResult);    /* use macros to get item & menu number */
        menuItem = LoWord(menuResult);
    }
    
    switch (menuID) 
    {
        case menuApplication:
            switch (menuItem) 
            {
                case itemAbout:
                    DoAboutBox();
                    break;
                    
/*                case itemQuit:
                    quit();
                    system("/bin/sleep 1");
                    ExitToShell();
                    break;
*/				
                default:
                    break;
            }
            break;
        
        case menuLaunch:
            LaunchItem(menuItem,0);
            break;
            
        case menuWindow:
            Restart_XDarwin();
            manipulate_focus_client(menuItem);
            Stop_XDarwin();
            break;
            
        case menuProperties:
            if(menuItem==itemEditRules) {
                Edit_Rules_File();
            } else if(menuItem==itemReadRules) {
                Read_Rules_File();
            } else {
                Restart_XDarwin();
                manipulate_focus_client(menuItem+100);
                Stop_XDarwin();
            }
            break;
		
        case menuEdit:
            break;

        case menuOptions:
            if(menuItem==itemQuietStartup) {
                gQuietStartup = 1-gQuietStartup;
                Setup_Options_Menu(0);
                SavePrefs();
            } else if(menuItem==itemAllowOffScreen) {
                gIncludeOffScreen = 1-gIncludeOffScreen;
                Setup_Options_Menu(0);
                SavePrefs();
                if(gIncludeOffScreen==1)
                    ShowAlert(WARNCVD_NAME,1,0);
            } else if(menuItem==itemInterleave) {
                gInterleaving_Windows=1-gInterleaving_Windows;
                Setup_Options_Menu(0);
/*                passwordWin=MakeWindow(passwordWintype);
                if(passwordWin==NULL)
                    printf("%s warning: Failed to show password window!\n",PROGNAME);
                else
                    DoPasswordWindow(1);*/
                SavePrefs();
            } else if(menuItem==itemWholeScreenMode) {
                if( wasWholeScreenMode == 0 ) {
                    // this automatically sets up the tick in the menu item
                    Toggle_Whole_Screen_Mode();
                } else {
                    if( gVerboseConsole )
                        printf("Not toggling whole screen mode because of wasWholeScreenMode...\n");
                    wasWholeScreenMode = 0;
                    // remove tick from menu item
                    Setup_FullScreen_MenuOption(0);
                }
            } else if(menuItem==itemXDarwinPrefs) {
                Restart_XDarwin();
                OpenXDarwinPrefs();
                Stop_XDarwin();
            } else {
                Restart_XDarwin();
                change_options(menuItem);
                Stop_XDarwin();
                SavePrefs();
            }
            break;

        case menuKeymaps:
            if(menuItem==itemRebuildKeymaps)
                Create_Keymaps_Menu(1);
            else {
                int err=Select_Keymaps_Menu(menuItem,1);
                if(menuItem==itemSaveKeymap && err==-3) Create_Keymaps_Menu(1);
            }
            break;
            
        case menuHelp:
            if(menuItem==itemHelp)
                ShowOroborOSXHelp();
            else if(menuItem==itemVers)
                ShowHistoryFile();
                //ShowGeneralFile("Contents/Resources/HISTORY.rtf");
            else if(menuItem==itemLicense)
                ShowLicenseFile();
                //ShowGeneralFile("Contents/Resources/LICENSE.rtf");
            else if(menuItem==itemHelpThemes)
                ShowThemesInfoFile();
            else if(menuItem==itemHelpLMI)
                ShowLMIInfoFile();
            else if(menuItem==itemHelpHelpers)
                ShowHelpersInfoFile();
            else if(menuItem==itemHelpRules)
                ShowRulesInfoFile();
            else if(menuItem==itemVerbose) {
                    gVerboseConsole=1-gVerboseConsole;
                    SendSimpleHackCommand(HACKCMD_VERBOSECONSOLE,gVerboseConsole);
                    Setup_Help_Menu();
                }
            break;

        case menuThemes:
            switch (menuItem)
            {
                case itemAboutThemes:
                    ShowThemesInfoFile();
                    break;
                    
                case itemAddTheme:
                    break;
                    
                case itemEditThemeList:
                    Edit_Menu_Items("ThemeList");
                    break;
                    
                case itemRebuildThemes:
                    Create_Themes_Menu(1);
                    break;
                    
                case itemToggleShaped:
                    Change_Theme(1-gUseShapedTheme,0);
                    SavePrefs();
                    break;
                
                case itemAboutThisTheme:
                    ShowAboutThisTheme();
                    break;
                    
                default:
                    if(menuItem>itemUserThemes) {
                        Change_Theme(gUseShapedTheme,menuItem);
                        SavePrefs();
                    } else
                        printf("How on earth did you get that Themes menu item?! %d\n",menuItem);
                    break;
            }
            break;
            
        case menuGroup:
            switch (menuItem)
            {
                case itemNewGroup:
                    Create_New_Group();
                    break;
                    
                case itemOriginalGroup:
                    original_group();
                    break;
                    
                default:
                    if(menuItem>=itemStartGroups) {
                        Put_Into_Group(menuItem);
                    } else
                        printf("How on earth did you get that Change Group menu item?! %d\n",menuItem);
                    break;
            }
            break;
    }
    if( Is_Whole_Screen_Mode()==0 ) HiliteMenu(0);	// unhighlight what MenuSelect (or MenuKey) hilited
    Restart_XDarwin();
    force_ungrab();	// Try to force ungrabbing of cursor (sometimes gets grabbed on menu select)
    Stop_XDarwin();
}

