/******************************************/ 
/* Carbon-compatible section of OroborOSX */
/* (c) Adrian Umpleby Jan 2002            */
/*                                        */
/* Original oroborus window manager code  */
/* (c) Ken Lynch Jan 2001                 */
/*                                        */
/* Distributed under the GPL              */ 
/* See LICENSE for more details           */ 
/******************************************/ 
 
/* osxwindows.c */

#include "main.h"

typedef struct _BadWin BadWin; 
struct _BadWin
{
    long	winref;
    BadWin	*next;
}; 

BadWin	*badwin_list=NULL;
int	badwin_count=0;


void SpecialPositionWindow(WindowRef myWindow, int which)
{
    Rect	winrct;
    OSStatus	myStat;

    myStat = GetWindowBounds( myWindow, kWindowStructureRgn, &winrct );
    winrct.right -= winrct.left;
    winrct.bottom -= winrct.top;
    winrct.left = (gMainRect.right - gMainRect.left - winrct.right) / 2;
    if(which==0)	// Centred vertically
        winrct.top = (gMainRect.bottom - gMainRect.top - winrct.bottom) / 2;
    else if(which==1)	// slightly more towards the top
        winrct.top = (gMainRect.bottom - gMainRect.top - winrct.bottom) / 3;
    else		// more towards the top, like an alert
        winrct.top = (gMainRect.bottom - gMainRect.top - winrct.bottom) / 4;
    winrct.right += winrct.left;
    winrct.bottom += winrct.top;
    myStat = SetWindowBounds( myWindow, kWindowStructureRgn, &winrct );
}


WindowRef MakeAlertWindow(const char *name)
{
    IBNibRef	theNIBRef;
    OSErr	err;
    OSStatus	myStat;
    WindowRef	myWindow=NULL;

    sprintf(lastsub,"TRACE3: MakeAlertWindow (%s)\n",name);
    subtrace();
//#ifdef SHOW_TRACE3
    printf("TRACE3: MakeAlertWindow (%s)\n",name);
//#endif

    if(name==NULL) {
        printf("%s warning: got NULL for name in MakeAlertWindow! (%s)\n",PROGNAME,name);
        return NULL;
    }

    myStat = CreateNibReference(CFSTR("alerts"),&theNIBRef);
    myStat = CreateWindowFromNib(theNIBRef,CFStringCreateWithCString(NULL,name,CFStringGetSystemEncoding()),&myWindow);

    if(myWindow==NULL) return NULL;

    SpecialPositionWindow(myWindow,2);
    
    ShowWindow(myWindow);    
    DisposeNibReference(theNIBRef);

    return myWindow;
}

WindowRef MakeWindow(int windowtype)
{
    WindowRef	myWindow=NULL;
    OSErr	err;
    OSStatus	myStat;
    IBNibRef	theNIBRef;
    Rect	winrct;
    
    sprintf(lastsub,"TRACE3: MakeWindow (type: %d)\n",windowtype);
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: MakeWindow (type: %d)\n",windowtype);
#endif

    if(windowtype==x11Wintype || windowtype==noFocusWintype) {
/* Only set size for the x11 windows and the 'no active window' window */    
        winrct.left=1;
        winrct.right=2;
        winrct.top=21;
        winrct.bottom=22;
    }
    
    myStat = CreateNibReference(CFSTR("window"),&theNIBRef);
    if(myStat!=0) {
        fprintf(stderr,"%s: Failed to find localised window resources!\n",PROGNAME);
        ShowAlert(BADWINDOWNIB_NAME,0,0);
        gQuitFlag=-1002;
        return NULL;
    }
    
    switch(windowtype)
    {
/* put up the information window */
    case waiting1Wintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("Waiting1"),&myWindow);
        break;
    case waiting2Wintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("Waiting2"),&myWindow);
        break;
    case waiting3Wintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("Waiting3"),&myWindow);
        break;
/* special window for when all X11 windows are iconbars */    
    case noFocusWintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("NoFocusWin"),&myWindow);
        break;
/* put up the about window */
    case aboutWintype:
        {
        long retval=0;
        Gestalt(gestaltSystemVersion,&retval);
        if( retval<0x1020 )
            myStat = CreateWindowFromNib(theNIBRef,CFSTR("About-10.1"),&myWindow);
        else
            myStat = CreateWindowFromNib(theNIBRef,CFSTR("About"),&myWindow);
        }
        break;
/* put up the interleaving dialog */
    case passwordWintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("Password"),&myWindow);
        break;
/* put up the 'file exists' dialog */
    case keymapNameWintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("Keymap Name"),&myWindow);
        break;
/* make a 'dummy' window, so that it appears in dock icon's window list */
    default:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("Window"),&myWindow);
        HideWindow(noFocusWin);
        showingNoFocusWin=0;
        break;
    }
    
    DisposeNibReference(theNIBRef);

    if(myWindow==NULL || myStat!=0) {
        fprintf(stderr,"%s warning: Failed to create a window of type %d!\n",PROGNAME,windowtype);
//        gQuitFlag=-1003;
        return NULL;
    }
    
/* Only set title if it's an x11 window */    
    if(windowtype==x11Wintype)
        myStat=SetWindowBounds(myWindow,kWindowContentRgn,&winrct);
    else if(windowtype==noFocusWintype)
        myStat=SetWindowBounds(myWindow,kWindowContentRgn,&winrct);
    else if(windowtype==waiting1Wintype || windowtype==waiting2Wintype || windowtype==waiting3Wintype)
        SpecialPositionWindow(myWindow,1);
        
    ShowWindow(myWindow);
    
/* make sure it is not on the screen (if it's one of the x11 windows etc.) */    
    if(windowtype==x11Wintype || windowtype==noFocusWintype)
        MoveWindow(myWindow,-5000,-5000,false);
    
    return myWindow;
}

void MakeGroupWindow(OSXWinInfo *winfo, char *group)
{
    WindowRef	winref=0;
    char *title;
    
    sprintf(lastsub,"TRACE2: MakeGroupWindow (winfo,group: %#lx,  %s)\n",winfo,group);
    subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: MakeGroupWindow (winfo,title,group: %#lx,  %s)\n",winfo,group);
#endif

    winref=MakeWindow(x11Wintype);
    
    if(winref==0) {
        printf("%s warning: Failed to create group window: %s  (%#lx)\n",PROGNAME,group,winref);
        return;
    }
    
#ifdef DEBUG
    printf("Created group window...  %s  (%#lx)\n",group,winref);
#endif

    if(winfo) winfo->group_winref=(long) winref;

    title = malloc(strlen(group)+10);
    sprintf(title,"%s:",group);
    CheckWinTitle((long) winref,title,NULL,VIEWSTATE_GROUPWIN,0);
}

void MakeX11Window(OSXWinInfo *winfo, char *title, char *group)
{
    WindowRef	winref=0;
    
    sprintf(lastsub,"TRACE2: MakeX11Window (winfo,title,group: %#lx,  %s  %s)\n",winfo,title,group);
    subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: MakeX11Window (winfo,title,group: %#lx,  %s  %s)\n",winfo,title,group);
#endif

    winref=MakeWindow(x11Wintype);
    
    if(winref==0) {
        printf("%s warning: Failed to create X11 window: %s  (%#lx)\n",PROGNAME,title,winref);
        return;
    }
    
#ifdef DEBUG
    printf("Created X11 window...  %s:  %s  (%#lx)\n",group,title,winref);
#endif

    winfo->winref=(long) winref;

    CheckWinTitle((long) winref,title,group,winfo->view_state,winfo->group_winref);
}

int DestroyBadWindows(WindowRef checkinlist)
{
    BadWin	*w, *wo;
    int		i;
    int		wasinlist=0;

    if(badwin_count==0) return 0;
    if(badwin_list==NULL) return 0;
    
    w=badwin_list;
    
    for(i=0;i<badwin_count;i++) {
#ifdef DEBUG
        printf("destroying window %d: %#lx\n",i,w->winref);
#endif
        if(w->winref==(long) checkinlist) wasinlist=1;
        ShowWindow((WindowRef) w->winref);
        SelectWindow((WindowRef) w->winref);
        DisposeWindow((WindowRef) w->winref);
        wo=w;
        w=w->next;
        free(wo);
    }
    
    badwin_list=NULL;
    badwin_count=0;
    
    return wasinlist;
}

void AddToBadWindowList(WindowRef win)
{
    BadWin	*w;
    
    w=malloc(sizeof(BadWin)+2);
    
#ifdef DEBUG
    printf("bad window %d: %#lx, %#lx\n",badwin_count+1,w,win);
#endif

    if(badwin_list==NULL) {
        badwin_list=w;
        badwin_count=0;
        w->winref=(long) win;
    } else {
        w->next=badwin_list;
        badwin_list=w;
        w->winref=(long) win;
    }
    
    badwin_count++;
}

void Check_OSX_Window_Bounds(OSXWinInfo *winfo)
{
    long	winref=0;
    
    if(winfo!=NULL);
        winref=winfo->winref;
    if(winref==0) return;

// don't bother checking bounds -just make sure it's offscreen...
    MoveWindow((WindowRef) winref,-5000,-5000,false);

// check the group window too
    winref=winfo->group_winref;
    if(winref==0) return;

// don't bother checking bounds -just make sure it's offscreen...
    MoveWindow((WindowRef) winref,-5000,-5000,false);
}

void ReorderGroupWin(long winref)	/* Put a window at the bottom of the dock icon window list */
{
    sprintf(lastsub,"TRACE2: ReorderGroupWin (winref: %d)\n",winref);
    subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: ReorderGroupWin (winref: %d)\n",winref);
#endif

    if(winref==0) return;
    
// this is enough to put the window at the bottom of the list
    HideWindow((WindowRef) winref);
    ShowWindow((WindowRef) winref);
    
// just to make sure it's offscreen...
    MoveWindow((WindowRef) winref,-5000,-5000,false);
}

void ReorderOSXWin(OSXWinInfo *winfo)
{    
    if(winfo==NULL) return;
    ReorderGroupWin(winfo->winref);
}

// Check title in dock icon window list
void CheckWinTitle(long winref, char *title, char *group, int state, long groupwinref)
{
    char *osxtitle, *offset_title, *tmp_title, *lastpos_title;
    pstr_t ptitle;
    int copyit=0, gotcvdhidden=0;

//    sprintf(lastsub,"TRACE3: CheckWinTitle (winref: %d)\n",winref);
//    subtrace();
//#ifdef SHOW_TRACE3
//    printf("TRACE3: CheckWinTitle (winref: %d)\n",winref);
//#endif

    if(winref==0) return;
    
    osxtitle=&ptitle.thestr;
    offset_title=title;
    copyit = 0;
    
    // check if the title already includes the group (Netscape & OpenOffice both do this, for example)
    if(group!=NULL && title!=NULL)
        if(strlen(group)<strlen(title)) {
            int gotatstart=0;
            if(!strncasecmp(group,title,strlen(group))) {
                offset_title+=strlen(group);
                // strip out spaces after class name
                while(strlen(offset_title)>1 && !strncmp(offset_title," ",1))
                    offset_title++;
                // strip out colon or dash
                if(strlen(offset_title)>1 && (!strncmp(offset_title,":",1) || !strncmp(offset_title,"-",1))) {
                    gotatstart = 1;
                    offset_title++;
                    // strip out any spaces after colon/dash
                    while(strlen(offset_title)>1 && !strncmp(offset_title," ",1))
                        offset_title++;
                } else	// didn't get colon or dash, so set it back
                    offset_title = title;
            }
            
            // Check for the group name in brackets at the start
            if(gotatstart==0 && strlen(group)+3<strlen(title)) {
                tmp_title = malloc(strlen(group)+5);
                sprintf(tmp_title,"[%s]",group);
                if(!strncasecmp(tmp_title,title,strlen(group)+2)) {
                    offset_title += strlen(group)+2;
                    gotatstart = 1;
                }
                if(gotatstart==0) {
                    sprintf(tmp_title,"(%s)",group);
                    if(!strncasecmp(tmp_title,title,strlen(group)+2)) {
                        offset_title += strlen(group)+2;
                        gotatstart = 1;
                    }
                }
                if(gotatstart==0) {
                    sprintf(tmp_title,"{%s}",group);
                    if(!strncasecmp(tmp_title,title,strlen(group)+2)) {
                        offset_title += strlen(group)+2;
                        gotatstart = 1;
                    }
                }
                if(gotatstart==0) {
                    sprintf(tmp_title,"<%s>",group);
                    if(!strncasecmp(tmp_title,title,strlen(group)+2)) {
                        offset_title += strlen(group)+2;
                        gotatstart = 1;
                    }
                }
                if(gotatstart!=0) {
                    // strip off spaces
                    while(strlen(offset_title)>1 && !strncmp(offset_title," ",1))
                        offset_title++;
                }
                free(tmp_title);
            }
            
            if(gotatstart==0 && strlen(group)+3<strlen(title)) {
                tmp_title = malloc(strlen(group)+5);
                lastpos_title = title;
                
                while(lastpos_title!=NULL) {
                    sprintf(tmp_title," - %s",group);
                    lastpos_title += 2;
                    offset_title = strstr(lastpos_title,tmp_title);
                    if(offset_title==NULL) {
                        sprintf(tmp_title," : %s",group);
                        offset_title = strstr(lastpos_title,tmp_title);
                    }
                    if(offset_title==NULL) {
                        sprintf(tmp_title," (%s)",group);
                        offset_title = strstr(lastpos_title,tmp_title);
                    }
                    if(offset_title==NULL) {
                        sprintf(tmp_title," [%s]",group);
                        offset_title = strstr(lastpos_title,tmp_title);
                    }
                    if(offset_title==NULL) {
                        sprintf(tmp_title," {%s}",group);
                        offset_title = strstr(lastpos_title,tmp_title);
                    }
                    if(offset_title==NULL) {
                        sprintf(tmp_title," <%s>",group);
                        offset_title = strstr(lastpos_title,tmp_title);
                    }
                    if(offset_title==NULL) {
                        sprintf(tmp_title,"- %s",group);
                        offset_title = strstr(lastpos_title,tmp_title);
                    }
                    if(offset_title==NULL) {
                        sprintf(tmp_title,": %s",group);
                        offset_title = strstr(lastpos_title,tmp_title);
                    }
                    if(offset_title==NULL) {
                        sprintf(tmp_title," :%s",group);
                        offset_title = strstr(lastpos_title,tmp_title);
                    }
                    if(offset_title==NULL) {
                        sprintf(tmp_title," -%s",group);
                        offset_title = strstr(lastpos_title,tmp_title);
                    }
                    lastpos_title -= 2;
                    if(offset_title!=NULL)
                        lastpos_title = offset_title;
                    else if(lastpos_title!=title) {
                    	offset_title = lastpos_title;
                        lastpos_title = NULL;
                    } else
                        lastpos_title = NULL;
                }
                
                free(tmp_title);
                if(offset_title!=NULL) {
                    // found the group name tagged onto the end(ish) of the window title
                    // so cut it off and make a copy
                    tmp_title = offset_title;
                    copyit = tmp_title[0];
                    tmp_title[0]=0;
                    offset_title = strdup(title);
                    tmp_title[0]=copyit;
                    copyit = 1;
                } else
                    offset_title = title;
            }
        }
            
        if(state!=VIEWSTATE_GROUPWIN && (state & VIEWSTATE_CVDHIDDEN)!=0) {
            gotcvdhidden = 1;
            state -= VIEWSTATE_CVDHIDDEN;
        } else
            gotcvdhidden = 0;

    if(state==VIEWSTATE_GROUPWIN && groupwinref==0) {
    	sprintf(osxtitle,"%s",title);
    } else if(groupwinref!=0) {
        if(gotcvdhidden)
            sprintf(osxtitle,"  (  %s  )",offset_title);
        else if(state==VIEWSTATE_ICONBAR)
            sprintf(osxtitle,"    %s",offset_title);
        else if(state==VIEWSTATE_DOCKED)
            sprintf(osxtitle,"    %s",offset_title);
        else if(state==VIEWSTATE_SHADED)
            sprintf(osxtitle,"  =  %s",offset_title);
        else if(state==VIEWSTATE_HIDDEN)
            sprintf(osxtitle,"  >  %s",offset_title);
        else
            sprintf(osxtitle,"     %s",offset_title);
    } else if(title==NULL && group==NULL) {
        if(gotcvdhidden)
            sprintf(osxtitle,"( ?:  Untitled )");
        else
            sprintf(osxtitle,"?:  Untitled");
    } else if(title==NULL) {
        if(gotcvdhidden)
            sprintf(osxtitle,"( %s:  Untitled )",group);
        else
            sprintf(osxtitle,"%s:  Untitled",group);
    } else if(group==NULL) {
        if(gotcvdhidden)
            sprintf(osxtitle,"( %s )",offset_title);
        else
            sprintf(osxtitle,"%s",offset_title);
    } else {
        if(gotcvdhidden)
            sprintf(osxtitle,"( %s:  %s )",group,offset_title);
        else
            sprintf(osxtitle,"%s:  %s",group,offset_title);
    }
    
    if(copyit) free(offset_title);
    
    ptitle.len=strlen(osxtitle);
//    strncpy((char *) &ptitle.thestr,title,ptitle.len);
#ifdef DEBUG
    printf("CheckWinTitle: %s  (%d)\n",osxtitle,strlen(osxtitle));
#endif

    MoveWindow((WindowRef) winref,-5000,-5000,false);
    SetWTitle((WindowRef) winref, (char *) &ptitle);
}

void CheckOSXWinTitle(OSXWinInfo *winfo, char *title, char *group, int cvdhidden)
{
    if(winfo==NULL) return;
    if(cvdhidden && winfo->view_state!=VIEWSTATE_GROUPWIN)
        CheckWinTitle(winfo->winref,title,group,VIEWSTATE_CVDHIDDEN,winfo->group_winref);
    else
        CheckWinTitle(winfo->winref,title,group,winfo->view_state,winfo->group_winref);    
}

void DestroyGroupWindow(long winref)
{
    sprintf(lastsub,"TRACE3: DestroyGroupWindow (winref: %d)\n",winref);
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: DestroyGroupWindow (winref: %d)\n",winref);
#endif
    
    if(winref==0) return;

    if(gInvisible==0) {
        // This disposes of the OroborOSX window associated with the X11 window
        DisposeWindow((WindowRef) winref);
    } else {
        // This is to work around what appears to be a bug in the OSX (Carbon) window manager
        // -windows can't be disposed of (or hidden) while the app itself is hidden; they don't
        // disappear from the Window menu, and trying to bring up the list of windows from the
        // dock icon crashes the app... (YUK!)
        CheckWinTitle(winref,"<dead window>",NULL,VIEWSTATE_GROUPWIN,0);
        AddToBadWindowList((WindowRef) winref);
    }
}

void DestroyOSXWindow(OSXWinInfo *winfo)
{    
    if(winfo==NULL) return;
    DestroyGroupWindow(winfo->winref);
}

/*
 * X11 window has moved or changed size -update info
 * input...
 * winfo: pointer to OSWinInfo structure, which is filled in with the relevant information
 * x,y: X11 co-ords of top-left of frame
 * w,h: width and height of frame
 * wox,woy: top-left widget co-ords relative to top-left of frame
 * wendox,wendoy: bottom-right widget co-ords relative to top-left of frame
 */
void SetOSXWinPos(OSXWinInfo *winfo,int x, int y, int w, int h, int wox, int woy, int wendox, int wendoy) 
{
    long	winref=0;
    
    if(winfo!=NULL)
        winref=winfo->winref;

//    sprintf(lastsub,"TRACE2: SetOSXWinPos (winref: %d x,y: %d,%d w,h: %d,%d wsx,wsy: %d,%d wex,wey: %d,%d)\n",winref,x,y,w,h,wox,woy,wendox,wendoy);
//    subtrace();
//#ifdef SHOW_TRACE2
//    printf("TRACE2: SetOSXWinPos (winref: %d x,y: %d,%d w,h: %d,%d wsx,wsy: %d,%d wex,wey: //%d,%d)\n",winref,x,y,w,h,wox,woy,wendox,wendoy);
//#endif

    if(winref==0) return;
        
    winfo->x=x;
    winfo->y=y;
    winfo->width=w;
    winfo->height=h;
    
    winfo->widgets_x=x+wox;
    winfo->widgets_y=y+woy;
    winfo->widgets_w=x+wendox;
    winfo->widgets_h=y+wendoy;
}

// called when an X11 window is minimised/restored/docked/undocked
void ChangedOSXWinState(OSXWinInfo *winfo, char * title, char * group, int cvdhidden)
{
    long	winref=0;
    int		state=0;
    
    if(winfo!=NULL);
        winref=winfo->winref;
    
//    sprintf(lastsub,"TRACE2: SetOSXWinState (winref,state: %d %d)\n",winref,state);
//    subtrace();
//#ifdef SHOW_TRACE2
//    printf("TRACE2: SetOSXWinState (winref,state: %d %d)\n",winref,state);
//#endif

    if(winref==0) return;
    
    state = winfo->view_state;
    
//    if(state==VIEWSTATE_ICONBAR)
// note: don't actually use this since it removes it from the window list in the dock icon
        //HideWindow((WindowRef) winref);
    CheckOSXWinTitle(winfo,title,group,cvdhidden);
}

void FocusOSXWin(OSXWinInfo *winfo)
{
    long	winref=0;
    
    sprintf(lastsub,"TRACE2: FocusOSXWin (winfo: %#lx)\n",winfo);
    subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: FocusOSXWin (winfo,winref: %#lx,%d)\n",winfo,winref);
#endif

    if(winfo==NULL) {
#ifdef SHOW_TRACE2
        printf("  FocusOSXWin got NULL\n");
#endif
        if(alertWin!=NULL) {
            Bring_OOSX_To_Front();
            ShowWindow(alertWin);
            SelectWindow(alertWin);
            HideWindow(noFocusWin);
            lastFrontWin=alertWin;
            showingNoFocusWin=1;
            Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
            Setup_Properties_Menu_Translucency(NULL);
        } else if(showingAboutWin) {
            ShowWindow(aboutWin);
            SelectWindow(aboutWin);
            HideWindow(noFocusWin);
            lastFrontWin=aboutWin;
            showingNoFocusWin=1;
#ifdef DEBUG
            printf("Selected window: %#lx\n",aboutWin);
#endif
            Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
            Setup_Properties_Menu_Translucency(NULL);
        } else {
            ShowWindow(noFocusWin);
            showingNoFocusWin=1;
            MoveWindow(noFocusWin,-5000,-5000,false);
            SelectWindow(noFocusWin);
#ifdef DEBUG
            printf("Selected window: %#lx\n",noFocusWin);
#endif
            lastFrontWin=noFocusWin;
            Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
            Setup_Properties_Menu_Translucency(NULL);
        }
        return;
    }
    
    HideWindow(noFocusWin);
    showingNoFocusWin=0;
    
    winref=winfo->winref;
#ifdef SHOW_TRACE2
    printf("  FocusOSXWin (winfo,winref: %#lx,%d)\n",winfo,winref);
#endif

    Setup_Properties_Menu_Translucency(winfo);
    
    ShowWindow((WindowRef) winref);
    SelectWindow((WindowRef) winref);
    lastFrontWin=(WindowRef) winref;
#ifdef DEBUG
    printf("Selected window: %#lx\n",winref);
#endif
}

void DrawWindow(WindowRef window)
{
    Rect		tempRect;
    GrafPtr		curPort;
	
    sprintf(lastsub,"TRACE3: DrawWindow (window: %#lx)\n",window);
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: DrawWindow (window: %#lx)\n",window);
#endif

    GetPort(&curPort);
    if(window!=aboutWin && window!=textInputWin && window!=alertWin)
        MoveWindow(window,-5000,-5000,false);
    SetPort(GetWindowPort(window));
    BeginUpdate(window);
    EraseRect(GetWindowPortBounds(window, &tempRect));
    DrawControls(window);
    DrawGrowIcon(window);
    EndUpdate(window);
    SetPort(curPort);
}

pascal OSStatus OKEventHandler(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData)
{
//        printf("Got into OKEventHandler\n");
        gotControlEvent=OK_CTRL_ID;
        return noErr;
}

pascal OSStatus CNCLEventHandler(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData)
{
//        printf("Got into CNCLEventHandler\n");
        gotControlEvent=CNCL_CTRL_ID;
        return noErr;
}

pascal OSStatus RDMEEventHandler(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData)
{
//        printf("Got into RDMEEventHandler\n");
        //gotControlEvent=RDME_CTRL_ID;
        ShowOroborOSXHelp();
        return noErr;
}

pascal OSStatus LCNCEventHandler(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData)
{
//        printf("Got into lcncEventHandler\n");
        //gotControlEvent=LCNC_CTRL_ID;
        ShowLicenseFile();
//        ShowGeneralFile("Contents/Resources/LICENSE.rtf");
        return noErr;
}

pascal OSStatus HISTEventHandler(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData)
{
//        printf("Got into HISTEventHandler\n");
        //gotControlEvent=HIST_CTRl_ID;
        ShowHistoryFile();
//        ShowGeneralFile("Contents/Resources/History.rtf");
        return noErr;
}

int ShowAlert(const char *name, int useidle, int extrabuts)
{
    Point		mspt;
    OSStatus		myStat;
    EventRef		outEvent;
    EventTargetRef 	theTarget;
    long		startPWTicks;
//    EventTypeSpec	controlSpec = { kEventClassCommand, kEventCommandProcess }; // event class, event kind
    EventTypeSpec	controlSpec = { kEventClassControl, kEventControlHit }; // event class, event kind
    ControlRef 		okControl, rdmeControl, histControl, lcncControl, cnclControl;
    ControlID 		controlID;
    int			oldQuitFlag;
    EventHandlerRef	okRef, rdmeRef, histRef, lcncRef, cnclRef;
    
    if(alertWin!=NULL) {
        printf("%s warning: Alert already being shown?!? (%s)\n",PROGNAME,name);
        return 0;
    }
    
    alertWin=MakeAlertWindow(name);
    if(alertWin==NULL) {
        printf("%s warning: Failed to create Alert window! (%s)",PROGNAME,name);
        return 0;
    }
#ifdef DEBUG
    printf("Created Alert window: %#lx (%s)",alertWin,name);
#endif

    Allow_HUPs(0);

    //Find the ok control and install an event handler on it.
    controlID.signature = 'OK  ';
    controlID.id = OK_CTRL_ID;
    GetControlByID(alertWin, &controlID, &okControl);

    InstallEventHandler(GetControlEventTarget(okControl),
                        NewEventHandlerUPP( OKEventHandler ),
                        1, &controlSpec, 0, &okRef);
//    InstallWindowEventHandler(alertWin,NewEventHandlerUPP(okEventHandler),1,&controlSpec,0,&outRef);

    if(extrabuts==3) {
        // install handlers for the three extra buttons
        controlID.signature = 'RDME';
        controlID.id = RDME_CTRL_ID;
        GetControlByID(alertWin, &controlID, &rdmeControl);
        InstallEventHandler(GetControlEventTarget(rdmeControl),
                            NewEventHandlerUPP( RDMEEventHandler ),
                            1, &controlSpec, 0, &rdmeRef);
                            
        controlID.signature = 'LCNC';
        controlID.id = LCNC_CTRL_ID;
        GetControlByID(alertWin, &controlID, &lcncControl);
        InstallEventHandler(GetControlEventTarget(lcncControl),
                            NewEventHandlerUPP( LCNCEventHandler ),
                            1, &controlSpec, 0, &lcncRef);

        controlID.signature = 'HIST';
        controlID.id = HIST_CTRL_ID;
        GetControlByID(alertWin, &controlID, &histControl);
        InstallEventHandler(GetControlEventTarget(histControl),
                            NewEventHandlerUPP( HISTEventHandler ),
                            1, &controlSpec, 0, &histRef);
    } else if(extrabuts==1) {
        // install handler for the extra cancel button
        controlID.signature = 'CNCL';
        controlID.id = CNCL_CTRL_ID;
        GetControlByID(alertWin, &controlID, &cnclControl);
        InstallEventHandler(GetControlEventTarget(cnclControl),
                            NewEventHandlerUPP( CNCLEventHandler ),
                            1, &controlSpec, 0, &cnclRef);
    }

    gotControlEvent=0;
    oldQuitFlag=gQuitFlag;
    gQuitFlag=0;
    theTarget = GetEventDispatcherTarget();
    myStat=noErr;
    startPWTicks=TickCount();
    if(gInited==0) useidle=0;
#ifdef DEBUG
    printf("Entering event loop... %d\n",startPWTicks);
#endif
    while(myStat==noErr) {
        GetGlobalMouse(&mspt);
        if(useidle!=0) DoIdle(mspt,1);
        myStat=ReceiveNextEvent(0,NULL,kEventDurationForever,TRUE,&outEvent);
        if(myStat==noErr) {
//            printf("Got event... %d\n",TickCount()-startPWTicks);
            SendEventToEventTarget(outEvent, theTarget);
            ReleaseEvent(outEvent);
        } else
            printf("Got an error: %d\n",myStat);
        if(gotControlEvent!=0 || gQuitFlag!=0) myStat=1;
//        if(TickCount()>startPWTicks+3600) myStat=1;
    }
    RemoveEventHandler(okRef);
    if(extrabuts==3) {
        RemoveEventHandler(lcncRef);
        RemoveEventHandler(histRef);
        RemoveEventHandler(rdmeRef);
    } else if(extrabuts==1) {
        RemoveEventHandler(cnclRef);
    }
#ifdef DEBUG
    printf("Removing alert window...\n");
#endif
    DisposeWindow(alertWin);
    alertWin=NULL;
    Allow_HUPs(1);
    if(gQuitFlag==0) gQuitFlag=oldQuitFlag;
    // The button IDs, returned in gotControlEvent, start at 100.
    // 100 is the OK/default button (so return zero), 101 is the other button (usually cancel), so return 1.
    // The extra three buttons do not remove the alert window, so don't worry about them... 
    return gotControlEvent-100;
}

char *DoTextInputWindow(int isPassword)
{
    Point		mspt;
    OSStatus		myStat;
    EventRef		outEvent;
    EventTargetRef 	theTarget;
    long		startPWTicks;
//    EventTypeSpec	controlSpec = { kEventClassCommand, kEventCommandProcess }; // event class, event kind
    EventTypeSpec	controlSpec = { kEventClassControl, kEventControlHit }; // event class, event kind
    ControlRef 		okControl, cnclControl;
    ControlID 		controlID;
    int			oldQuitFlag, err, controlTag;
    EventHandlerRef	okRef, cnclRef;
    char		*inputText=NULL;
    Size		inputSize;
    
    Allow_HUPs(0);

    //Find the ok control and install an event handler on it.
    controlID.signature = 'OK  ';
    controlID.id = OK_CTRL_ID;
    GetControlByID(textInputWin, &controlID, &okControl);

    InstallEventHandler(GetControlEventTarget(okControl),
                        NewEventHandlerUPP( OKEventHandler ),
                        1, &controlSpec, 0, &okRef);
//    InstallWindowEventHandler(alertWin,NewEventHandlerUPP(okEventHandler),1,&controlSpec,0,&outRef);

    //Find the cancel control and install an event handler on it.
    controlID.signature = 'CNCL';
    controlID.id = CNCL_CTRL_ID;
    GetControlByID(textInputWin, &controlID, &cnclControl);

    InstallEventHandler(GetControlEventTarget(cnclControl),
                        NewEventHandlerUPP( CNCLEventHandler ),
                        1, &controlSpec, 0, &cnclRef);
//    InstallWindowEventHandler(alertWin,NewEventHandlerUPP(okEventHandler),1,&controlSpec,0,&outRef);

    gotControlEvent=0;
    oldQuitFlag=gQuitFlag;
    gQuitFlag=0;
    if(alertWin==NULL)
        alertWin=textInputWin;
    theTarget = GetEventDispatcherTarget();
    myStat=noErr;
    startPWTicks=TickCount();
    
// get ready to read the text
    if(isPassword) {
        controlTag = kControlEditTextPasswordTag;
        controlID.signature = 'PWTX';
    } else {
        controlTag = kControlEditTextTextTag;
        controlID.signature = 'TEXT';
    }
    controlID.id = TEXT_CTRL_ID;
    GetControlByID(textInputWin, &controlID, &cnclControl);
    
//    printf("Entering event loop... %d\n",startPWTicks);
    while(myStat==noErr) {

        GetGlobalMouse(&mspt);
        if(gInited!=0) DoIdle(mspt,1);
        myStat=ReceiveNextEvent(0,NULL,kEventDurationForever,TRUE,&outEvent);

        if(myStat==noErr) {
//            printf("Got event... %d\n",TickCount()-startPWTicks);
            SendEventToEventTarget(outEvent, theTarget);
            ReleaseEvent(outEvent);
        } else
            printf("Got an error: %d\n",myStat);

        // check for OK button
        if(gotControlEvent==OK_CTRL_ID) {
            GetControlDataSize(cnclControl,0,controlTag,&inputSize);
//#ifdef DEBUG
            printf("size of data: %#lx\n",inputSize);
//#endif
            // carry on if there's nothing in the edit box
            if(inputSize==0) {
                gotControlEvent=0;
            }
        }
        
        if(gotControlEvent!=0 || gQuitFlag!=0) myStat=1;
//        if(TickCount()>startPWTicks+3600) myStat=1;
    }
    
    RemoveEventHandler(okRef);
    RemoveEventHandler(cnclRef);

    if(gotControlEvent!=OK_CTRL_ID)
    {
        printf("Removing text input window (got cancel? %d)...\n",gotControlEvent);
        if(alertWin==textInputWin)
            alertWin=NULL;
        DisposeWindow(textInputWin);
        textInputWin=NULL;
        if(gotControlEvent!=CNCL_CTRL_ID) {
            fprintf(stderr,"%s warning: failed text input window: %d (%d)n",
                                                PROGNAME,myStat,gotControlEvent);
            ShowAlert(PASSWINFAIL_NAME,1,0);
        }
        Allow_HUPs(1);

    }
    else
    {
                
// Get password from text-edit field
        inputText=malloc(inputSize+5);    
        if(inputText!=NULL)
            GetControlData(cnclControl,0,controlTag,inputSize+4,inputText,&inputSize);
//    success = CFStringGetCString(passwdCFStr, passwdText, 100, CFStringGetSystemEncoding()); 
     
//#ifdef DEBUG
        printf("Removing text input window (got OK)...\n");
//#endif
        if(alertWin==textInputWin)
            alertWin=NULL;
        DisposeWindow(textInputWin);
        textInputWin=NULL;

        if(inputText==NULL) {
            fprintf(stderr,"%s warning: could not allocate memory for input text\n",PROGNAME);
            ShowAlert(WARNMEM_NAME,1,0);
        } else if(inputSize<1 || inputSize>100) {
            fprintf(stderr,"%s warning: found strange input text length: %d\n",PROGNAME,inputSize);
            ShowAlert(PASSWINFAIL_NAME,1,0);
        } else {
            inputText[inputSize]=0;
#ifdef DEBUG
            printf("Got OK in input text window... %s\n",inputText);
#endif
            /* not used any more...
            if(isPassword) {
            if(setenv("OOSX_TMP_PASSWD",inputText,1)!=0) {
                fprintf(stderr,"%s warning: failed to set passwd environment variable\n",PROGNAME);
                ShowAlert(PASSWINFAIL_NAME,1,0);
            } else {
                if(gInterleaving_Windows==0)
                    err = system("Contents/Scripts/GetXpid interleave");
                else
                    err = system("Contents/Scripts/GetXpid uninterleave");
                unsetenv("OOSX_TMP_PASSWD");
                if(err==0) {
                    gInterleaving_Windows=1-gInterleaving_Windows;
                    Setup_Options_Menu(0);
                    if(gInterleaving_Windows==0)
                        ShowAlert(INTLVOFFOK_NAME,1,0);
                    else
                        ShowAlert(INTLVONOK_NAME,1,0);
                } else if(err==PASSWDFAIL_ERRNO) {
                    ShowAlert(PASSWDFAIL_NAME,1,0);
                } else if(err==LOSTXDARWIN_ERRNO) {
                    ShowAlert(LOSTXDARWIN_NAME,1,0);
                } else if(err==INTLVON_ERRNO) {
                    gInterleaving_Windows=1;
                    Setup_Options_Menu(0);
                    ShowAlert(INTLVON_NAME,1,0);
                } else if(err==INTLVOFF_ERRNO) {
                    gInterleaving_Windows=0;
                    Setup_Options_Menu(0);
                    ShowAlert(INTLVOFF_NAME,1,0);
                } else {
                    fprintf(stderr,"%s warning: interleaving script failed (err: %d)\n",PROGNAME,err);
                    ShowAlert(INTLVFAIL_NAME,1,0);
                }
            }
            }*/
        }
        
        /*if(inputText!=NULL && isPassword) {
            free(inputText);
            inputText=NULL;
        }*/
    }
    
    if(gQuitFlag==0) gQuitFlag=oldQuitFlag;
    Allow_HUPs(1);
    
    return inputText;
}

void DoAboutBox(void)
{
    sprintf(lastsub,"TRACE3: DoAboutBox\n");
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: DoAboutBox\n");
#endif

//   (void) Alert(kAboutBox, nil);  / simple alert dialog box
    aboutWin=(WindowRef) MakeWindow(aboutWintype);
    if(aboutWin==NULL) return;
    
    gX11InFront=0;
    Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    gone_to_back();

    if(lastFrontWin==noFocusWin) {
        HideWindow(noFocusWin);
        showingNoFocusWin=1;
    }
        
    ShowWindow(aboutWin);
    SelectWindow(aboutWin);
#ifdef DEBUG
    printf("Showing about window: %#lx\n",aboutWin);
#endif
    showingAboutWin=1;
    lastFrontWin=aboutWin;
}


int Ask_Quit()
{
    int doquit=0;
    
    if(client_count==0)	// don't bother with confirmation if there's no clients
        return 1;
    else {
        Bring_OOSX_To_Front();	// should put a bouncing dock icon type notification here...
        doquit=ShowAlert(QUITX11_NAME,1,1);	// confirmation dialog if we started XDarwin
    }
    
    if(doquit)
        return 1;
    
    return noErr;
}
