/******************************************/ 
/* Carbon-compatible section of OroborOSX */
/* (c) Adrian Umpleby Jan 2002            */
/*                                        */
/* Original oroborus window manager code  */
/* (c) Ken Lynch Jan 2001                 */
/*                                        */
/* Distributed under the GPL              */ 
/* See LICENSE for more details           */ 
/******************************************/ 
 
/* screens.c */

#include "main.h"
#include "hackcmds.h"
#include <QuickTime/QuickTime.h>

GDHandle	gMainScreen;
Rect		gMainRect;
int		initialMBarHeight=-1, willEndWholeScreenMode=0;
int		initial_offx=-30000,initial_offy=-30000;
Ptr		restoreState=NULL;
long		keepInFrontUntil = 0;
static Rect	screenRects[100];	// 100 screens should be enough!
static int	screenCount=0;

// http://developer.apple.com/documentation/Carbon/Reference/Window_Manager/wind_mgr_ref/function_group_16.html

void Set_End_Whole_Screen_Mode()
{
    if( restoreState )
        willEndWholeScreenMode = 1;
    else
        willEndWholeScreenMode = 0;
}

void Check_End_Whole_Screen_Mode()
{
    if( willEndWholeScreenMode )
        End_Whole_Screen_Mode();
    willEndWholeScreenMode = 0;
}

int Is_Whole_Screen_Mode()
{
    if( restoreState ) return 1;
    return 0;
}

void Start_Whole_Screen_Mode()
{
    
    if( !restoreState ) {
        short x=0, y=0;
        OSStatus err;
        if( gVerboseConsole ) printf("%s info: Starting Whole Screen Mode\n",PROGNAME);
        //DrawMenuBar();
        HiliteMenu(0);
        err = BeginFullScreen( &restoreState, NULL, &x, &y, NULL, NULL, 0 );
        SendSimpleHackCommand(HACKCMD_SETWHOLESCREEN,1);
        //restoreState = (Ptr)-1;
    }
}

/* bring OroborOSX in front and wait for the Dock to go away if in whole screen mode */
void Wait_OOSX_To_Front()
{
    OSErr		myErr;
    Boolean		same;
    int			i;
    ProcessSerialNumber	frontPSN;
    
    sprintf(lastsub,"TRACE2: Wait_OOSX_To_Front\n");
    subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: Wait_OOSX_To_Front\n");
#endif

    //if( Is_Whole_Screen_Mode()==0 ) return;

    myErr=GetFrontProcess(&frontPSN);
    if(myErr==noErr)
        myErr=SameProcess(&frontPSN,&myPSN,&same);

    if( same ) return;

//    show_last_fifteen();
    myErr=SetFrontProcess(&myPSN);
//    myErr=SetFrontProcess(&xPSN);
//    gX11InFront=1;

    i = 0;
    while(!same) {
        i++;
        if(i>1000) {
#ifdef SHOW_WEIRD
            printf("%s warning: Failed to come in front for 1 sec!\n",PROGNAME);
#endif
            same=TRUE;
            break;
        }
        usleep(1000);
        myErr=GetFrontProcess(&frontPSN);
        if(myErr==noErr)
            myErr=SameProcess(&frontPSN,&myPSN,&same);
        if(myErr!=noErr) {
#ifdef SHOW_WEIRD
            printf("%s warning: Got an error trying to check if OOSX is in front?!\n",PROGNAME);
#endif
            same=TRUE;
            forceOOSXInFront=TRUE;
        }
    }
    if(keepInFrontUntil!=-1) {
    SendSimpleHackCommand(HACKCMD_COMEINFRONT,1);
//    myErr=SetFrontProcess(&xPSN);
    if( gVerboseConsole ) printf("Wait_OOSX_To_Front: fiddle with front process so Dock goes away...\n");
    keepInFrontUntil = TickCount() + 1000;
    if( keepInFrontUntil==-1 || keepInFrontUntil==0 ) keepInFrontUntil = 1;
    }
}

void End_Whole_Screen_Mode()
{
    if( restoreState ) {
        int isFront = gX11InFront;
        OSStatus err;
        Wait_OOSX_To_Front();
        if( gVerboseConsole ) printf("%s info: Ending Whole Screen Mode\n",PROGNAME);
        err = EndFullScreen( restoreState, NULL );
        SendSimpleHackCommand(HACKCMD_SETWHOLESCREEN,0);
        //DrawMenuBar();
        restoreState = NULL;
        //if( isFront ) Bring_X_To_Front(1);
    }
}

void Toggle_Whole_Screen_Mode()
{
    if( restoreState ) {
        Setup_FullScreen_MenuOption(0);
        End_Whole_Screen_Mode();
    } else {
        Setup_FullScreen_MenuOption(1);
        Start_Whole_Screen_Mode();
    }
}

int Count_Carbon_Screens(int update)
{
    int count=0, need_update=0;
    GDHandle aScreen;
    long startTick=TickCount();
    
    if( initialMBarHeight<0 ) {
        initialMBarHeight = GetMBarHeight();
        need_update = 1;
        printf("%s info: First time in Count_Carbon_Screens - will need update...\n",PROGNAME);
    } else if( screenCount<=0 )
        need_update = 1;
    else
        need_update = 0;
    
    gMainScreen=GetMainDevice();
    aScreen = GetDeviceList();
    while(aScreen) {
        Rect r=(*aScreen)->gdRect;
        if( need_update ||
            screenRects[count].left != r.left ||
            screenRects[count].right != r.right ||
            screenRects[count].top != r.top ||
            screenRects[count].bottom != r.bottom ) {
            if( update==0 && need_update==0 ) return 0;
            need_update = 1;
            screenRects[count] = r;
        }
        count++;
        aScreen=GetNextDevice(aScreen);
    }
    
    if( TickCount()>startTick+30 ) {
        printf("%s warning: getting device list took a long time! (%#lx)\n",PROGNAME,TickCount()-startTick);
        Allow_HUPs(1);
    }
    
    if( count!=screenCount ) need_update = 1;
    screenCount = count;

    if( need_update ) return count;
    
    if(gVerboseConsole) printf("%s info: Screen configuration not changed (%d)...\n",PROGNAME,count);
    return -count;
}

void Carbon_To_X11(int *x, int *y)
{
    int leftx=30000, topy=30000;
    Rect aRect;
    GDHandle aScreen;
    
    if( initialMBarHeight<0 ) initialMBarHeight = GetMBarHeight();
    gMainScreen=GetMainDevice();
    aScreen=GetDeviceList();

    if(aScreen==NULL) {
        printf("%s warning: Got NULL for GetDeviceList?!\n",PROGNAME);
        return;
    } else if(*aScreen==NULL) {
        printf("%s warning: Got NULL pointer for GetDeviceList?!\n",PROGNAME);
        return;
    }

    if( initial_offx==-30000 && initial_offy==-30000 ) {
    while(aScreen) {
        aRect=(*aScreen)->gdRect;
        if(aScreen==gMainScreen) aRect.top+=initialMBarHeight-1;
        if(aRect.top<topy) topy=aRect.top;
        if(aRect.left<leftx) leftx=aRect.left;
        aScreen=GetNextDevice(aScreen);
    }
        // first time through...
        initial_offx = leftx;
        initial_offy = topy;
    } else {
        // offset doesn't change in XDarwin at the moment, even when screen configuration changes
        leftx = initial_offx;
        topy = initial_offy;
    }
    *x -= leftx;
    *y -= topy;
    //if(topy<0) *y += 22;
}

void Cocoa_To_Carbon(int *x, int *y)
{
    gMainScreen=GetMainDevice();
    if(gMainScreen==NULL) {
        printf("%s warning: Got NULL for GetMainDevice?!\n",PROGNAME);
        return;
    } else if(*gMainScreen==NULL) {
        printf("%s warning: Got NULL pointer for GetMainDevice?!\n",PROGNAME);
        return;
    }
    gMainRect=(*gMainScreen)->gdRect;
    *y = gMainRect.bottom - *y;
}

void Get_Best_Window_Zoom(int *x, int *y, int *w, int *h, int which, int lb, int rb, int tb, int bb)
{
    Rect rct;
    int ix,iy,iw,ih,offset_x=0,offset_y=0;
    long area=0, best=0, dist=0;
    GDHandle aScreen,bestScreen=NULL;
    
    Carbon_To_X11(&offset_x,&offset_y);
    bestScreen = gMainScreen;
    rct.left = *x - offset_x;
    rct.top = *y - offset_y;
    rct.right = rct.left + *w;
    rct.bottom = rct.top + *h;
    aScreen = GetDeviceList();
    while(aScreen) {
        Rect aRect=(*aScreen)->gdRect, sr;
        if(aScreen==gMainScreen) aRect.top+=GetMBarHeight()-1;
        if(SectRect(&rct,&aRect,&sr)) {
            area = (sr.right-sr.left)*(sr.bottom-sr.top);
            if( area > best ) {
                best = area;
                bestScreen = aScreen;
            }
        } else if( best==0 ) {
            area = -1;
            if( rct.right>=aRect.left && rct.left<=aRect.right ) {
                if( rct.top>=aRect.bottom )
                    area = rct.top - aRect.bottom;
                else if( rct.bottom<=aRect.top )
                    area = aRect.top - rct.bottom;
            } else if( rct.bottom>=aRect.top && rct.top<=aRect.bottom ) {
                if( rct.left>=aRect.right )
                    area = rct.left - aRect.right;
                else if( rct.right<=aRect.left )
                    area = aRect.left - rct.right;
            }
            if( area > 0 ) {
                if( area < dist ) {
                    dist = area;
                    bestScreen = aScreen;
                }
            }
        }
        aScreen=GetNextDevice(aScreen);
    }
    if(bestScreen) {
        Rect aRect;
        long startTick=TickCount();
        int err=GetAvailableWindowPositioningBounds( bestScreen, &aRect );
        if( TickCount()>startTick+25 ) {
            printf("Get_Best_Window_Zoom warning: GetAvailableWindowPositioningBounds took a long time (%#xl)!\n",TickCount()-startTick);
            Allow_HUPs(1);
        }
        if( err==0 ) {
            int zx = aRect.left + offset_x + lb;
            int zy = aRect.top + offset_y + tb;
            int zw = aRect.right - aRect.left - lb - rb;
            int zh = aRect.bottom - aRect.top - tb - bb;
            if( which==1 ) {
                if(*x>=zx && *x+*w<=zx+zw)
                    zx = *x;
                else if(*x+*w>zx+zw)
                    zx = zx + zw - *w;
                zw = *w;
            } else if( which==2 ) {
                if(*y>=zy && *y+*h<=zy+zh)
                    zy = *y;
                else if(*y+*h>zy+zh)
                    zy = zy + zh - *h;
                zh = *h;
            }
            *x = zx;
            *y = zy;
            *w = zw;
            *h = zh;
        } else {
            printf("Get_Best_Window_Zoom warning: GetAvailableWindowPositioningBounds failed! (err: %d)\n",err);
            Allow_HUPs(1);
        }
    }
}

void Get_Carbon_Screen_Bounds(int screen, int *x, int *y, int *w, int *h)
{
    int count=0, err=0;
    GDHandle aScreen;
    long startTick;
    
    *x = 0;
    *y = 0;
    *w = 0;
    *h = 0;
    
    gMainScreen=GetMainDevice();
    aScreen = GetDeviceList();
    while(aScreen) {
        if(count==screen) {
            Rect aRect;//=(*aScreen)->gdRect;
            //if(aScreen==gMainScreen) aRect.top+=GetMBarHeight()-1;
            startTick = TickCount();
            err = GetAvailableWindowPositioningBounds( aScreen, &aRect );
            if( TickCount()>startTick+15 ) {
                printf("Get_Carbon_Screen_Bounds warning: GetAvailableWindowPositioningBounds took a long time! (%d)\n",count);
                Allow_HUPs(1);	// make sure keepalive knows we're OK
            }
            if( err==0 ) {
                *x = aRect.left;
                *y = aRect.top;
                *w = aRect.right - aRect.left;
                *h = aRect.bottom - aRect.top;
            } else {
                printf("Get_Carbon_Screen_Bounds warning: Screen %d GetAvailableWindowPositioningBounds failed! (err: %d)\n",count,err);
                Allow_HUPs(1);	// make sure keepalive knows we're OK
            }
            return;
        }
        count++;
        aScreen=GetNextDevice(aScreen);
    }
    
    // Failed to find screen!!
    printf("%s warning: Get_Carbon_Screen_Bounds failed to find screen %d! (%d)\n",
                                                                PROGNAME,screen,count);
    Allow_HUPs(1);	// make sure keepalive knows we're OK
}
