/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 

/* common.h */

/*
 * This contains the info that both the 'oroborus part' and thhe 'Carbon part'
 * of OroborOSX need to know about
 */
 
#ifndef OROB_COMMON_H 
#define OROB_COMMON_H 

#include <unistd.h>

#include "definitions.h"
#include "errors.h"
#include "trace.h"

#define PROGNAME		"OroborOSX"

#define BGWIN_NOTHING		0
#define BGWIN_DIMMED		1
#define BGWIN_TRANSLUCENT	2

// Translucency given as a percentage (i.e. 100 is opaque)
#define	ACTIVE_TRANSLUCENT0	100
#define ACTIVE_TRANSLUCENT1	85
#define ACTIVE_TRANSLUCENT2	65

#define INACTIVE_TRANSLUCENT0	85
#define INACTIVE_TRANSLUCENT1	80
#define INACTIVE_TRANSLUCENT2	60

#define MOUSE_STOPPED	-1
#define MOUSE_NORMAL	0
#define MOUSE_MOVE	1
#define MOUSE_RESIZE	2
#define MOUSE_CLOSE	3
#define MOUSE_ZOOM	4
#define MOUSE_CLOSE_OUT	5
#define MOUSE_ZOOM_OUT	6
#define MOUSE_OVERCLOSE	7
#define MOUSE_OVERZOOM	8
#define MOUSE_ICON	9
#define MOUSE_ICON_OUT	10
#define MOUSE3_TITLE	20

#define ICONBARPOS_TOPLEFT	0
#define ICONBARPOS_TOPRIGHT	1
#define ICONBARPOS_BOTLEFT	2
#define ICONBARPOS_BOTRIGHT	3

// number of steps for iconbars
#define ICONSTEPS_FAST 4
#define ICONSTEPS_MED 7
#define ICONSTEPS_SLOW 20

#define VIEWSTATE_GROUPWIN	-1
#define VIEWSTATE_HIDDEN	0
#define VIEWSTATE_VISIBLE	1
#define VIEWSTATE_SHADED	2
#define VIEWSTATE_ICONBAR	3
#define VIEWSTATE_DOCKED	4
#define VIEWSTATE_CVDHIDDEN	128

#define	menuApplication		128	/* Application menu ID */
#define	itemAbout		1

#define menuLaunch		129	/* Launch menu ID */
#define itemNewItem		1
#define itemRebuildMenu		2
#define itemEditItems		4
#define itemEditHelpers		5
#define itemEditStart		7
#define itemxinitrc		9
// points to divider item in Launch menu before the items added from the Launch Menu Items folder
#define itemUserItems		10
// These have now been replaced by x11app items in the Launch menu Items folder...
/*
#define itemxterm		11
#define itemxedit		12
#define itemxbiff		13
#define itemxclock		14
#define itemxeyes		15
#define itemxcalc		16
#define itemUserItems		17
// The rest of these are not actually used - they are just to remind of the default for the user part
#define itemETerm		18
#define itemGimp		19
#define itemNEdit		20
*/

#define	menuEdit		130	/* Edit menu ID */

#define	menuOptions		131	/* Options menu ID */
#define itemQuietStartup	1
#define itemXDarwinPrefs	2
#define itemInterleave		3
#define itemWholeScreenMode	4
#define itemAllowOffScreen	6
#define itemBGWinDimmed		7
#define itemBGWinTranslucent	8
#define itemThemeSubmenu	10
#define itemKeymapsSubmenu	11
//#define itemLeaveMenubar	10
#define itemMinToDock		13
#define itemFancyIconbars	14
#define itemHideIconbars	15
#define itemSlowIconbars	17
#define itemMedIconbars		18
#define itemFastIconbars	19
#define itemTopLeft		21
#define itemTopRight		22
#define itemBotLeft		23
#define itemBotRight		24
#define itemScreenListStart	26

#define menuThemes		132	/* Themes submenu ID */
#define itemAboutThemes		1
#define itemAddTheme		3
#define itemEditThemeList	4
#define itemRebuildThemes	5
#define itemToggleShaped	7
#define itemAboutThisTheme	9
// points to the divider in the Themes submenu before the items added from the Themes folder
#define itemUserThemes		10
// These items are not actually used -they are replaced by the themes that are actually found...
#define itemEau			11
#define itemGreyphite		12
#define itemNextish		13

#define	menuWindow		133	/* Window menu ID */
#define itemWindowshade		1
#define	itemMinimise		2
#define	itemZoom		3
#define	itemClose		5
#define	itemForceClose		6
#define itemProperties		8
#define itemGroupSubmenu	10
#define itemHideGroup		11
#define itemLowerGroup		13
#define itemNextGroup		14
#define itemLower		16
#define itemNext		17

#define menuProperties		135	/* Properties submenu ID */
#define itemFirstClick		1
#define itemFirstClickChild	2
#define itemFirstClickDrag      3
#define itemAutoFocus		5
#define itemAutoRaise		6
#define itemAutoLower		7
#define itemClickRaise		8
#define itemNoGrabFocus		10
#define itemRetainFocus		11
#define itemNoframes		13
#define itemLiveResize		14
#define itemFloat		16
#define itemPopup		17
#define itemModal		18
#define itemIgnoreKeys		20
#define itemIgnoreClicks	21
#define itemTranslucent0	23
#define itemTranslucent1	24
#define itemTranslucent2	25
#define itemBGTranslucency	27
#define itemDimWhenInactive	29
#define itemTransWhenInactive	30
#define itemNotDimWhenInactive	31
#define itemNotTransWhenInactive 32
#define itemEditRules		34
#define itemReadRules		35

#define menuGroup		136	/* Group submenu ID */
#define itemNewGroup		1
#define itemOriginalGroup	2
#define itemStartGroups		4

#define	menuHelp		134	/* Help menu ID */
#define itemHelp		1
#define itemHelpLMI		3
#define itemHelpHelpers		4
#define itemHelpThemes		5
#define itemHelpRules		6
#define itemVers		8
#define itemLicense		9
#define itemVerbose		11

#define menuKeymaps		137	/* X11 Keymaps submenu ID */
#define itemMergeXmodmap	1
#define itemSaveKeymap		2
#define itemRebuildKeymaps	3
#define itemGlobalKeymaps	5
#define initPersonalKeymaps	7

#define minmethod_iconbars	0
#define minmethod_dock		1
#define minmethod_icons		2

/***** Preferences *****/

/* Iconbar */
#define kPrefIconbarPos			"IconbarPos"
#define kPrefIconbarScreen		"IconbarScreen"
#define kPrefIconbarSteps		"IconbarSteps"
#define kPrefIconbarShrinking		"IconbarShrinking"
#define kPrefIconbarHiding		"IconbarHiding"
#define kPrefMinMethod			"MinimiseMethod"
/* Themes */
#define kPrefThemeName			"ThemeName"
#define kPrefThemeShaped		"ThemeShaped"
/* Misc */
#define kPrefInterleavingWindows	"InterleavingWindows"
#define kPrefBGWinSetting		"InactiveWindowBehaviour"
#define kPrefQuietStartup		"QuietStartup"
#define kIncludeOffScreen		"IncludeOffScreen"
#define kPrefKeymapName			"KeymapName"
#define kPrefKeymapGlobal		"KeymapIsGlobal"

// Extra window info structure
typedef struct 
{
    long winref, group_winref;
    int x, y, width, height;
    int widgets_x, widgets_y, widgets_w, widgets_h;
    int view_state, translucency, inactive_translucency, drawalpha;
} OSXWinInfo; 

/* globals */
extern OSXWinInfo *group_list; 
extern int group_count, realPathLen, wasWholeScreenMode, gotCommandDown, itemPersonalKeymaps;
extern char realPath[512];
extern int gInvisible, on_cmd_line, shrinking_iconbars, iconbar_screen, live_resizing;
extern int gQuitFlag, gX11InFront, gInterleaving_Windows, gMinMethod, gVerboseConsole;
extern int gIncludeOffScreen, gGotCVDHidden, gKeymapGlobal, gWaitingForSIGCHLD;
extern int gMenubarSpace, gUseShapedTheme, gBGWinSetting, gX11MouseStuck, gHiding_Iconbars;
extern int mouse_mode, real_x, real_y, iconbarpos, diffscrn_x, diffscrn_y, launched_xdarwin;
extern long lastBackTicks, lastIconbarOutTicks, client_count, lastsentmods, keepInFrontUntil;
extern char randomID[20];
extern char *Display_Name, *xpidstr, *xverstr, *gThemeDir, *gKeymapName;
extern char *lastsub15, *lastsub14, *lastsub13, *lastsub12, *lastsub11;
extern char *lastsub10, *lastsub9, *lastsub8, *lastsub7, *lastsub6;
extern char *lastsub5, *lastsub4, *lastsub3, *lastsub2, *lastsub1;
extern long check_clients_onscreen, delayed_enter_ticks;

/* main.c */
extern void ClearUpSockets();
extern char *Escape_Quotes(char *instr, int which);
extern void Bring_To_Front(int quick);
extern void Bring_X_To_Front(int quick);
extern void Bring_OOSX_To_Front();
extern void will_hide_all_iconbars();
extern void Compare_Screens(int scrnx, int scrny, int x11x, int x11y, int allowset);
extern void Stop_XDarwin();
extern void Restart_XDarwin();
extern int Carbon_Button();
/* Prefs */
extern int	getIntPref(const char* id, int defaultVal);
extern void	setIntPref(const char* id, int newVal);
extern char*	getStringPref(const char* id, const char* defaultVal);	/* Returns newly malloc'ed memory; caller should free. */
extern void	setStringPref(const char* id, const char* newVal);

/* screens.c */
extern void Carbon_To_X11(int *x, int *y);
extern void Cocoa_To_Carbon(int *x, int *y);
extern int Count_Carbon_Screens();
extern void Get_Carbon_Screen_Bounds(int screen, int *x, int *y, int *w, int *h);
extern void Get_Best_Window_Zoom(int *x, int *y, int *w, int *h, int which, int lb, int rb, int tb, int bb);
extern void Start_Whole_Screen_Mode();
extern void End_Whole_Screen_Mode();
extern void Toggle_Whole_Screen_Mode();
extern int Is_Whole_Screen_Mode();
extern void Set_End_Whole_Screen_Mode();
extern void Check_End_Whole_Screen_Mode();
extern void Wait_OOSX_To_Front();

/* macevents.c */
void Start_Time_Check(long);
int Is_Time_Up();
void Allow_HUPs(int);
void SetCmdKeyCursor(int);

/* menus.c */
extern void Setup_FullScreen_MenuOption(int ticked);
extern void Setup_Options_Menu(int which);
extern void Setup_Window_Menu(int all_off, int has_frame, int has_close, int has_min, int has_zoom, int is_float, int can_af, int can_ar, int can_al, int first_click, int click_raise, int live_resize, int no_grab_focus, int retain_focus, int client_modality);
extern void Setup_Properties_Menu_Translucency(OSXWinInfo *winfo);
extern long GetCarbonTickCount();
extern void Add_Group_to_Menu(char *name);
extern void Remove_Group_from_Menu(char *name);

/* osxwindows.c */
extern int Ask_Quit();
extern int ShowAlert(const char *name, int useidle, int extrabuts);
extern void Check_OSX_Window_Bounds(OSXWinInfo *winfo);
extern void MakeGroupWindow(OSXWinInfo *winfo, char *group);
extern void MakeX11Window(OSXWinInfo *winfo, char *title, char *group);
extern void FocusOSXWin(OSXWinInfo *winfo);
extern void DestroyGroupWindow(long winref);
extern void DestroyOSXWindow(OSXWinInfo *winfo);
extern void CheckWinTitle(long winref, char *title, char *group, int state, long groupwinref);
extern void CheckOSXWinTitle(OSXWinInfo *winfo, char *title, char *group, int cvdhidden);
extern void ChangedOSXWinState(OSXWinInfo *winfo, char * title, char * group, int cvdhidden);
extern void ReorderGroupWin(long winref);
extern void ReorderOSXWin(OSXWinInfo *winfo);
extern void SetOSXWinPos(OSXWinInfo *winfo,int x, int y, int w, int h, int wox, int woy, int wendox, int wendoy);

/* oroborus.c */
extern void quit();
extern int orob_main(int argc, char *argv[], char *dpy_str);
extern void init_rules_list();

/* x11events.c */ 
extern void do_event_loop(int sleepus,int trackMouse);
extern void force_ungrab();

/* client.c */
extern void left_client();
extern void gone_to_back();
extern void come_to_front();
extern void change_focus(long winref);
extern void hide_all_iconbars();
extern void manipulate_focus_client(int action);
extern void change_options(int newattr);
extern void raise_all_clients();
extern void map_all_clients();
extern void unmap_all_clients();
extern void check_osx_windows_bounds();
extern long get_next_osxwinref(int);
extern char *get_next_win_title(int);
extern void init_nearest_iconbar();
extern void list_all_windows(); 
extern void list_all_clients(); 

/* group.c */
extern char *get_next_group_name();
extern char *get_next_title_in_group(int);
extern long get_next_osxwinref_in_group(int, int);
extern char *get_focus_client_grouporg_name();
extern char *get_focus_client_group_name();
extern void mark_group(long, int);
extern int count_and_mark_groups();
extern char *get_marked_group_name(int);
extern void original_group();

/* config.c */ 
extern void reload_config(); 

/* misc.c */
extern int count_xinerama_heads();
extern void update_xinerama_screens();
extern int MyXServerGrabLevel();

/* xdarwinhack.c */
extern void SendSimpleHackCommand(int cmd, int param);
extern void SendModsToXDarwin(long new, long old);
extern void SendKeyToXDarwin(int key, int down);
extern void SendKeycodeModsToXDarwin(int *modlist);

#endif /* OROB_COMMON_H */

