/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 
 
#include "oroborus.h" 
#include "hackcmds.h"
#include "iconbars.h"

Client *focus_before_hide;

/*
 * Remove this client from parent's sibling list
 */

void remove_sibling_client(Client *c)
{
  Client *cp, *cnxt, *cprv;
  
  sprintf(lastsub,"TRACE3: remove_sibling_client (window: %#lx)\n",C_WINDOW);
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: remove_sibling_client (window: %#lx)\n",C_WINDOW);
#endif

  cp=c->parent;
  if(!cp) return;
  
  cnxt=cp->first_child;

/* This should never really happen! */
  if(cnxt==NULL) {
#ifdef SHOW_WEIRD
    fprintf(stderr,"%s warning: parent lost children!\n",PROGNAME);
#endif
    cp->child_count=0;
    return;
  }
  
/* This should never really happen! */
  if(cp->child_count!=0)
    cp->child_count--;
#ifdef SHOW_WEIRD
  else
    fprintf(stderr,"%s warning: parent has inconsistent child_count!\n",PROGNAME);
#endif

  if(cnxt==c)
    cp->first_child=c->next_sibling;
    
  cnxt=c->next_sibling;
  cprv=c->prev_sibling;
  
  if(cnxt)
    cnxt->prev_sibling=cprv;
  if(cprv)
    cprv->next_sibling=cnxt;
}

int undock_client(Client *c)
{
    long nowTime;
    
    if(c!=NULL) {
      nowTime=GetCarbonTickCount();
      if(c->dock_timer==0 && c->undock_timer==0)
        nowTime=10;
      if(nowTime<c->dock_timer || nowTime<c->undock_timer) {
        return 1;
      } else {
        if(c->osxwinfo.view_state==VIEWSTATE_HIDDEN) {
            gInvisible = 0;
            map_group(get_group_for_client(c));
//          map_all_clients();
//          Bring_OOSX_To_Front();
        }
        c->dock_timer=0;
        c->undock_timer=0;
        if(gVerboseConsole)
          printf("%s info: getting client from dock (%#lx)\n",PROGNAME,C_WINDOW);
        SendHackCommand(c,1,HACKCMD_DOCKWIN,0,0,0,0,0);
//        XMapWindow(dpy,c->frame);
//        XMapWindow(dpy,c->window);
        set_wm_state(c->window, NormalState);
        XMoveWindow(dpy,c->frame,FX,FY);
        // check to see if window was docked during a theme change
        if(c->is_e_m_r & THEME_CHANGED) {
          // fool it into doing an unwindowshade, even though it is already unshaded
          set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_SHADED);
          toggle_client_shaded(c);
          c->is_e_m_r ^= THEME_CHANGED;
        }
        if(c->cover && c->height-TH-BH>0 && c->width-LW-RW>0)
          XMoveResizeWindow(dpy,c->cover,c->x+LW,c->y+TH,c->width-LW-RW,c->height-TH-BH);
//        show_last_fifteen();
        //sleep(3);
        if( c==lastfocus_client ) {
            // have to make sure frame is redrawn in the case where the window was docked
            // while not frontmost app, but while it had focus
            lastfocus_client=NULL;
            lastfocus=0;
        }
        gX11InFront = 1;
        set_focus_client(c);
        if((c->auto_properties & AUTO_FOCUS)==AUTO_FOCUS)
          raise_client(c);	// just in case set_focus_client did not raise it
        c->osxwinfo.view_state = VIEWSTATE_VISIBLE;
        if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
        return 0;
      }
    }
    return 2;
}

void will_undock_client(Client *c)
{
    if(c!=NULL) {
      if(c->undock_timer==0) {
        c->undock_timer=GetCarbonTickCount()+300;
        if(gVerboseConsole)
          printf("%s info: will get client from dock (%#lx)\n",PROGNAME,C_WINDOW);
        SendHackCommand(c,1,HACKCMD_DOCKWIN,2,0,0,0,0);
      } else if(GetCarbonTickCount()>=c->undock_timer) {
        c->dock_timer=0;
        c->undock_timer=0;
        undock_client(c);
      }
    }
}

Client *check_for_undock(int hackid)
{
    Client *c;
    int i;
    
    c=client_list;
    if(c==NULL) return NULL;
    if(hackid==0) return NULL;
    
    for(i=0;i<client_count;i++) {
        if(c->is_docked && hackid==c->hack_id)
            return c;
        c=c->next;
    }
    return NULL;
}

Client *get_client_from_hackid(int hackid)
{
    Client *c;
    int i;
    
    c=client_list;
    if(c==NULL) return NULL;
    
    for(i=0;i<client_count;i++) {
        if(hackid==c->hack_id)
            return c;
        c=c->next;
    }
    return NULL;
}

void dock_client(Client *c)
{
    Client *tc;
    
    if(c!=NULL) {
      if(c->dock_timer!=0) return;
        c->is_docked=10;
        tc=get_top_most_client(get_gnome_hint(c->window,WIN_LAYER),False);
        if(tc!=NULL) {
            if(check_client_modality(tc,0)==c) {
                tc=NULL;
                if(gVerboseConsole)
                    printf("Got to set top to NULL, otherwise window will come back from dock\n");
            }
        }
        c->is_docked=0;
      c->dock_timer=GetCarbonTickCount()+300;
//      tc=get_bottom_most_client(get_gnome_hint(c->window,WIN_LAYER),False);
      if(gVerboseConsole)
        printf("sending window to dock... %#lx (cl,fr: %#lx,%#lx)\n",c->window,c,c->frame);
//      if(tc!=c)
//        set_focus_client(tc);
//      else
//        set_focus_client(NULL);
      SendHackTitleToDockCommand(c,1);
      //SendHackCommand(c,1,HACKCMD_DOCKWIN,1,0,0,0,0);
      //XMoveWindow(dpy,c->frame,-FW-10,FY);
//      c->ignore_unmap++; 
//      XUnmapWindow(dpy,c->frame);
//      XUnmapWindow(dpy,c->window);
      set_wm_state(c->window, IconicState);
      c->osxwinfo.view_state = VIEWSTATE_DOCKED;
      if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
      //tc=get_top_most_client(get_gnome_hint(c->window,WIN_LAYER),False);
      lower_client(c);
      usleep(300000); // give it 0.3 seconds, so it gets a chance to start sending to the dock
      if(gX11InFront==0) {
        if(tc==NULL)
            set_focus_client(NULL);
      } else if(tc)
        set_focus_client(tc);
      else
        set_focus_client(NULL);
//      SendHackCommand(c,1,HACKCMD_DOCKWIN,1,0,0,0,0);
      //sleep(3);
    }
}

/*  
 *  
 * Get client frame from window ID, from window frame ID, title window ID, or from OS X WindowRef
 * New option - send in win with negative value to avoid check on lastenter (should speed it up a bit)
 */ 

Client *get_client_from_window(Window w, int win) 
{ 
  Client *c=NULL; 
  Client *ct=NULL;
  Client *cf=NULL;
  Client *cw=NULL;
  long i;
  int docheck=1; 
 

  if(w==0 || client_count==0 || w==root || w==check_win ) {
    sprintf(lastsub,"TRACE3: got 0 for client_from_window (win,what,client_count: %#lx,%d,%d)\n",w,win,client_count);
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: got 0 for client_from_window (win,what,client_count: %#lx,%d,%d)\n",w,win,client_count); 
#endif 
    return NULL;
  }

  if(win<0) {
    win=-win;
    docheck=0; }
    
  c=client_list; 
//#ifdef DEBUG
//   printf("   checking client [%#lx] %#lx _%#lx_\n",c->frame,c->window,c->title);
//#endif

// Just putting the entire loop into these if blocks seems to have brought down the
// CPU time quite noticeably (according to 'top', anyway)...
    if(win==TITLE) 
    { 
      for(i=0;i<client_count;i++) 
      { 
      // No lastenter check here
      if(c->title==w) {
//        ct=c;
  sprintf(lastsub,"TRACE3: got client_from_title (%#lx,%#lx %d)\n",c,C_WINDOW,client_count); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: got client_from_title (%#lx,%#lx %d)\n",c,C_WINDOW,client_count); 
#endif 
        return c; 
      }
      c=c->next; 
      } 
    } 
    else if(win==FRAME_LE) 
    { 
      for(i=0;i<client_count;i++) 
      { 
      if(c->frame==w) {
        cf=c;
        // May want to continue and do a lastenter check...
        if(c->frame!=lastenter || docheck==0) {
  sprintf(lastsub,"TRACE3: got client_from_frame (%#lx,%#lx %d)\n",c,C_WINDOW,client_count); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: got client_from_frame (%#lx,%#lx %d)\n",c,C_WINDOW,client_count); 
#endif 
          return c;  }
      }
      c=c->next; 
      } 
    }  
    else if(win==FRAME_OR_COVER_LE) 
    { 
      for(i=0;i<client_count;i++) 
      { 
      if(c->frame==w || c->cover==w) {
        cf=c;
        // May want to continue and do a lastenter check...
        if(c->frame!=lastenter || docheck==0) {
  sprintf(lastsub,"TRACE3: got client_from_frame/cover (%#lx,%#lx %d)\n",c,C_WINDOW,client_count); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: got client_from_frame/cover (%#lx,%#lx %d)\n",c,C_WINDOW,client_count); 
#endif 
          return c;  }
      }
      c=c->next; 
      } 
    }  
    else  if(win==OSXWINREF)
    { 
      for(i=0;i<client_count;i++) 
      { 
      // No lastenter check here
      if(c->osxwinfo.winref==(long) w) { 
//        cw=c;
  sprintf(lastsub,"TRACE3: got client_from_osxwinref (%#lx,%#lx %d)\n",c,C_WINDOW,client_count); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: got client_from_osxwinref (%#lx,%#lx %d)\n",c,C_WINDOW,client_count); 
#endif 
        return c; 
      }
      c=c->next; 
      } 
    }
    // default is to look for the client that has the actual X window
    else 
    { 
      for(i=0;i<client_count;i++) 
      { 
      if(c->window==w) { 
//        cw=c;
        // May want to continue and do a lastenter check...
//        if(c->frame!=lastenter || docheck==0)
  sprintf(lastsub,"TRACE3: got client_from_window (%#lx,%#lx %d)\n",c,C_WINDOW,client_count); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: got client_from_window (%#lx,%#lx %d)\n",c,C_WINDOW,client_count); 
#endif 
          return c; 
      }
      c=c->next; 
      } 
    } 

// do a check for the sanity of lastenter
  if(w==lastenter) {
//    if(win==TITLE) {
//      if(ct==NULL) {
//#ifdef SHOW_WEIRD
//         printf("  failed to find title for lastenter - setting to zero\n");
//#endif
//         lastenter=0; }
//      return ct;
//    } else if(win==FRAME) {
      if(cf==NULL) {
#ifdef SHOW_WEIRD
         printf("%s oddity: failed to find frame for lastenter - setting to zero\n",PROGNAME);
#endif
         lastenter=0; }
//      return cf;
//    } else {
//      if(cw==NULL) {
//#ifdef SHOW_WEIRD
//         printf("  failed to find window for lastenter - setting to zero\n");
//#endif
//         lastenter=0; }
//    }
  }

  c=cf;
  
  sprintf(lastsub,"TRACE3: got client_from_window (%#lx,%#lx %d)\n",c,C_WINDOW,client_count); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: got client_from_window (%#lx,%#lx %d)\n",c,C_WINDOW,client_count); 
#endif

  return cf; 
} 
 
/* 
 * 
 * Get client with focus 
 * 
 */ 

Client *get_focus_client() 
{ 
  Client *c; 
  Window focus_win; 
  int dummy; 
 
  sprintf(lastsub,"TRACE3: get_focus_client (last: %#lx,%#lx)\n",FC_WINDOW,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_focus_client (last: %#lx,%#lx)\n",FC_WINDOW,lastfocus); 
#endif 
 
  XGetInputFocus(dpy, &focus_win, &dummy); 
  c=get_client_from_window(focus_win, WINDOW);

#ifdef DEBUG
  if(c==NULL)
    printf("  got no client for focus_win %#lx (%#lx)\n",focus_win,lastfocus);
  else
    printf("  got focus_win: %#lx (%#lx) [%#lx] %#lx\n",focus_win,lastfocus,c->frame,c->window);
#endif

// hack the focus for when menus are shown from a window -oroborus used to make
// an X11 nedit window 'inactive' when a menu was pulled down from it
  hacking_focus=False;
//  if(gX11InFront==0) return;
  
  if(c!=NULL)
    lastfocus=focus_win;
  else if(focus_win==check_win)
    lastfocus=0;
  else if(lastfocus!=0) {
    c=get_client_from_window(lastfocus, WINDOW);
    if(c!=NULL)
      hacking_focus=True;
    else
      lastfocus=0;
  }

  sprintf(lastsub,"TRACE3: leaving get_focus_client (win: %#lx  hf: %d)\n",focus_win,hacking_focus); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: leaving get_focus_client (win: %#lx  hf: %d)\n",focus_win,hacking_focus); 
#endif 

  lastfocus_client=c;
  return c;   
} 
 
/* 
 * 
 * Get next client in client list 
 * 
 */ 

Client *get_next_client(Client *c) 
{ 
  long i; 
 
  sprintf(lastsub,"TRACE3: get_next_client (win: %#lx)\n",C_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_next_client (win: %#lx)\n",C_WINDOW); 
#endif 
 
  if(c!=NULL) 
  { 
    for(i=0;i<client_count;i++) 
    { 
      c=c->next; 
#ifdef WORKSPACE_SUPPORT 
      if(!get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS && get_gnome_hint(c->window, WIN_WORKSPACE)==workspace) 
#else 
      if(!get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS) 
#endif /* WORKSPACE */ 
        return c; 
    } 
  } 
  return NULL; 
} 

/* 
 * 
 * Get previous client in client list 
 * 
 */ 

Client *get_prev_client(Client *c) 
{ 
  long i; 
 
  sprintf(lastsub,"TRACE3: get_prev_client (win: %#lx)\n",C_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_prev_client (win: %#lx)\n",C_WINDOW); 
#endif 
 
  if(c!=NULL) 
  { 
    for(i=0;i<client_count;i++) 
    { 
      c=c->prev; 
#ifdef WORKSPACE_SUPPORT 
      if(!get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS && get_gnome_hint(c->window, WIN_WORKSPACE)==workspace) 
#else 
      if(!get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS) 
#endif /* WORKSPACE */ 
        return c; 
    } 
  } 
  return NULL;
}
 
/* 
 * 
 * Get top-most client (and say whether iconbars are to be included in search)
 * 
 */ 

Client *get_top_most_client(int layer, int get_icons) 
{ 
  Window *windows, w1, w2; 
  int count, i; 
  Client *top=NULL; 
  Client *c=NULL; 
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE3: get_top_most_client (layer,get_icons: %d,%d)\n",layer,get_icons); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_top_most_client (layer,get_icons: %d,%d)\n",layer,get_icons); 
#endif 
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
#ifdef LAYER_SUPPORT 
  for(i=0;i<count;i++) 
  { 
    XGetWindowAttributes(dpy, windows[i], &attr); 
    if(attr.map_state==IsViewable) 
    { 
      c=get_client_from_window(windows[i], FRAME); 
      if(c!=NULL && (c->isoffscreen==0 || gIncludeOffScreen) &&
            (c->is_icon==NOT_ICONBAR || get_icons) && c->is_docked==0) {
        if(get_gnome_hint(c->window, WIN_LAYER)<=layer) 
          top=c; 
        else 
          break; 
      } 
    } 
  } 
#else 
  for(i=count-1;i>=0;i--) 
  { 
    XGetWindowAttributes(dpy, windows[i], &attr); 
    if(attr.map_state==IsViewable) { 
      c=get_client_from_window(windows[i], FRAME); 
      if(c!=NULL && (c->isoffscreen==0 || gIncludeOffScreen))
        if((c->modal_state & ANY_FLOATING)==0 && (c->is_icon==NOT_ICONBAR || get_icons) && c->is_docked==0) {
          top=c;
          break;
        }
    } 
  } 
#endif /* LAYER */ 
  if(windows) XFree(windows);
  
  if(top==NULL)
    FocusOSXWin(NULL);
  else if( top->is_icon!=NOT_ICONBAR )
    FocusOSXWin(NULL);

  return top; 
} 

/* 
 * 
 * Get top_most client which contains point x,y (and say whether iconbars are to be included)
 * 
 */ 

Client *get_client_container(int x, int y, int get_icons)
{
  Window *windows, w1, w2; 
  int count, i; 
  Client *top=NULL; 
  Client *c=NULL; 
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE2: get_client_container (x,y,get_icons: %d,%d %d)\n",x,y,get_icons); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: get_client_container (x,y,get_icons: %d,%d %d)\n",x,y,get_icons); 
#endif 
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
  
  for(i=count-1;i<=0;i--) 
  { 
    XGetWindowAttributes(dpy, windows[i], &attr); 
    if(attr.map_state==IsViewable) { 
      c=get_client_from_window(windows[i], FRAME); 
      if(c!=NULL && (c->is_icon==NOT_ICONBAR || get_icons) && c->is_docked==0) 
      { 
        if(x>=attr.x)
          if(x<attr.x+attr.width)
            if(y>=attr.y)
              if(y<attr.y+attr.height) {
                top=c;
                break; }
      }
    } 
  } 

  if(windows) XFree(windows);
  
  return top; 
}

Client *get_lowest_entered_client(int get_icons) 
{ 
  Window *windows=NULL, w1, w2; 
  int count, i; 
  Client *c=NULL; 
  Client *bot=NULL;
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE3: get_lowest_entered_client (get_icons: %d)\n",get_icons); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_lowest_entered_client (get_icons: %d)\n",get_icons); 
#endif 
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
  for(i=0;i<count;i++) 
  { 
    XGetWindowAttributes(dpy, windows[i], &attr); 
    if(attr.map_state==IsViewable) 
    { 
      c=get_client_from_window(windows[i], FRAME); 
      if(c!=NULL && c->isoffscreen==0 && ((c->is_icon==NOT_ICONBAR || get_icons)
                 && c->is_docked==0) && (c->is_e_m_r & IS_ENTERED)!=0) 
      { 
        bot=c;
        break; 
      } 
    } 
  } 
  if(windows) XFree(windows);  
  return bot; 
} 

/* 
 * 
 * Get bottom-most client (and say whether iconbars are to be included)
 * 
 */ 

Client *get_bottom_most_client(int layer, int get_icons) 
{ 
  Window *windows=NULL, w1, w2; 
  int count, i; 
  Client *c=NULL; 
  Client *bot=NULL;
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE3: get_bottom_most_client (layer,get_icons: %d,%d)\n",layer,get_icons); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_bottom_most_client (layer,get_icons: %d,%d)\n",layer,get_icons); 
#endif 
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
  for(i=0;i<count;i++) 
  { 
    XGetWindowAttributes(dpy, windows[i], &attr); 
    if(attr.map_state==IsViewable) 
    { 
      c=get_client_from_window(windows[i], FRAME); 
      if(c!=NULL && (c->isoffscreen==0 || gIncludeOffScreen) &&
                ((c->is_icon==NOT_ICONBAR || get_icons) && c->is_docked==0)) 
      { 
#ifdef LAYER_SUPPORT 
        if(get_gnome_hint(c->window, WIN_LAYER)>=layer) { 
          bot=c;
          break; }
#else 
        bot=c;
        break; 
#endif /* LAYER */ 
      } 
    } 
  } 
  if(windows) XFree(windows);  
  return bot; 
} 

/*
 * A new window has been chosen from the dock icon's window list, so change the focus in X11
 */
 
void change_focus(long winref)
{
  Client *c=NULL, *lc=NULL;
  int need_ungrab=0;
  
  sprintf(lastsub,"TRACE1: change_focus (ref: %#lx)\n",winref);
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: change_focus (ref: %#lx)\n",winref);
#endif

  need_ungrab = check_need_ungrab();
/*
  if(lastfocus_client!=NULL && pointer_grabbed==0) {
    retval=XGrabPointer(dpy, lastfocus_client->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeSync, GrabModeSync, None, normal_curs, CurrentTime);
    if(retval==GrabSuccess) {
        XUngrabPointer(dpy, CurrentTime);
        XAllowEvents(dpy, SyncPointer, CurrentTime);
        printf("Got %d from XGrabPointer -no need for fake button click.\n",retval);
    } else {
        need_ungrab = 1;
    }
  }*/
//  if(gIncludeOffScreen!=0)
//    lc=get_top_most_client(WIN_LAYER_NORMAL,False);
    
  c=get_client_from_window((Window) winref,OSXWINREF);
  
  if(c==NULL) { // not found, so it must be a group window
    raise_group(winref);
//    c=get_top_most_client(WIN_LAYER_NORMAL,False);
  } else {
    if(c->is_docked)
      will_undock_client(c);
    else if(c->is_icon==NOT_ICONBAR) {
      raise_client(c);
      set_focus_client(c);
      Bring_To_Front(0);
    } else {
      c->is_icon=ICONBAR_RAISED;
      raise_client(c);
      minmax_client(c);
      Bring_To_Front(0);
    }
  }
  
  if(need_ungrab) {
    printf("%s info: Got %d from XGrabPointer - asking XDarwin for fake button click.\n",PROGNAME,need_ungrab);
    //SendSimpleHackCommand(HACKCMD_UNGRABPTR,0);
    force_grab_release();
  } else if(gVerboseConsole)
    printf("%s info: Got %d from XGrabPointer -no need for fake button click.\n",PROGNAME,need_ungrab);
/*
  if(c!=NULL && lc!=NULL && gIncludeOffScreen!=0 && c->isoffscreen!=0) {
    long endTime;
    Bring_OOSX_To_Front();
    usleep(100000);
    printf("raising old client win %#lx\n",lc->window);
    raise_client(lc);
    Bring_To_Front(1);
    printf("raising new client win %#lx\n",c->window);
    raise_client(c);
    usleep(500000);
    endTime=GetCarbonTickCount()+60;
    while( c->isoffscreen!=0 && GetCarbonTickCount()<endTime) {
      Bring_OOSX_To_Front();
      usleep(100000);
      printf("raising old client win %#lx\n",lc->window);
      raise_client(lc);
      Bring_To_Front(1);
      printf("raising new client win %#lx\n",c->window);
      raise_client(c);
      usleep(500000);
    }
  }*/
}

/*void stack_cover(Client *c, int which)
{
  XWindowChanges wc; 
  
  return;
  if(c==NULL) return;
  if(c->cover==0) return;
  
  wc.stack_mode=which;
  wc.sibling=c->frame;
  XConfigureWindow(dpy,c->cover,CWStackMode|CWSibling,&wc);
}*/

void stack_client_re_client(Client *c, Client *ac, int which) 
{ 
  XWindowChanges wc; 
 
  sprintf(lastsub,"TRACE3: stack_client_below_client (wins,last,which: %#lx,%#lx,%#lx,%d)\n",C_WINDOW,AC_WINDOW,FC_WINDOW,which); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: stack_client_below_client (wins,last,which: %#lx,%#lx,%#lx,%d)\n",C_WINDOW,AC_WINDOW,FC_WINDOW,which); 
#endif 
 
  if(c==NULL || ac==NULL) {
#ifdef SHOW_WEIRD
    printf("%s oddity: a client is NULL in stack_client_below_client (%#lx %#lx)\n",PROGNAME,C_WINDOW,AC_WINDOW);
#endif
    return;
  }
  
  if(c->is_docked) return;
  
  wc.stack_mode=which;
  wc.sibling=ac->frame;
//  XConfigureWindow(dpy, c->frame, CWStackMode, &wc);
//  redraw_client_frame(c);
  configure_client(c, CWStackMode|CWSibling, &wc, c->is_icon!=NOT_ICONBAR, 0); 
} 

void restack_autolower_client(Client *c)
{
  Window w1, w2, *windows; 
  int count, i, j, k, cpos=-1, cposn=-1, minpos;
  long sum, minsum;
  Client *ac;
  XWindowAttributes attr;

  if(c==NULL) return;
  if(c->al_list==NULL) return;
  if(c->al_count==0) return;
  if((c->auto_properties & (AUTO_FOCUS|AUTO_LOWER))!=(AUTO_FOCUS|AUTO_LOWER)) return;
  
  MyIncreaseXServerGrab();
  
  XQueryTree(dpy, root, &w1, &w2, &windows, &count);
  
  j = 0;
  for(i=0;i<count;i++) {
    if(windows[i] != check_win ) {
      XGetWindowAttributes(dpy, windows[i], &attr);
      if(attr.map_state==IsViewable) {
        windows[j] = windows[i];
        j++;
      }
    }
  }
  count = j;
  
  if(gVerboseConsole)
    printf("restack %#lx = \"%s\"  count: %d (was %d)\n",C_WINDOW,c->name,count,c->al_count);
  
  if(count>0) {
    for(i=0;i<c->al_count;i++) {
      if(c->al_list[i]==c->frame) {
        cpos = i;
        c->al_list[i] = -1;
      } else {
        for(j=0;j<count;j++) {
          if(windows[j]==c->al_list[i]) {
            c->al_list[i] = (Window) j;
            break;
          } else if(j==count-1) {
            c->al_list[i] = -1;
          }
        }
      }
    }
    
    for(i=0;i<count;i++) {
      if(windows[i]==c->frame) {
        cposn = i;
        break;
      }
    }
    
    if(gVerboseConsole)
      printf("Found posn: %d (was %d)\n",cposn,cpos);

    if(cpos<0 || cposn<0) count = 0;
  }
  
  if(count>0) {
    minsum = 0;
    minpos = cposn;
    for(j=0;j<c->al_count;j++) {
      k = (int) c->al_list[j];
      if(k>=0) {
        k = (k-minpos)-(j-cpos);
        minsum = minsum + k*k;
      }
    }
    for(i=0;i<cposn;i++) {
      sum = 0;
      for(j=0;j<c->al_count;j++) {
        k = (int) c->al_list[j];
        if(k>=0) {
          if(k>=i && k<cposn) k++;
          k = (k-i)-(j-cpos);
          sum = sum + k*k;
        }
      }
      if(sum<minsum) {
        minsum = sum;
        minpos = i;
      }
    }
    if(minpos<cposn) {
      ac = NULL;
      for(i=minpos;i<count;i++) {
        ac = get_client_from_window(windows[i],FRAME);
        if(ac!=NULL && ac!=c)
          break;
      }

      if(gVerboseConsole)
        printf("lowering to posn %d (%#lx) below win: %#lx = \"%s\")\n",minpos,minsum,AC_WINDOW,ac==NULL?"-":ac->name);

      if(ac!=NULL)
        stack_client_re_client(c,ac,Below);

/*      k=0;
      force_raise_client(c);
      for(i=minpos;i<count;i++) {
        if(k==1) {
          if(windows[i]!=c->frame)
            XRaiseWindow(dpy,windows[i]);
        } else {
          ac = get_client_from_window(windows[i],FRAME);
          if(ac!=NULL && ac!=c) {
            force_raise_client(ac);
            k=1;
          }
        }
      }*/
    }
  }
  
  MyDecreaseXServerGrab();

  if(windows!=NULL) XFree(windows);
  if(c->al_list!=NULL) XFree(c->al_list);
  c->al_list=NULL;
  c->al_count=0;
}

/*
 * raise_non_clients is used to raise a client in such a way that any menus of other clients do not
 * end up being obscured by the raised client - it might even obscure its own menus otherwise!
 */
int raise_non_clients(Client *c, int get_icons) 
{ 
  Window *windows, w1, w2;
  Window xcid; 
  XWindowChanges wc;
  int count, i, j, k; 
  Client *top=NULL; 
  Client *ac=NULL; 
  XWindowAttributes attr; 
 
  if(c==NULL) return 0;
  
  xcid = (C_WINDOW) & X_CLIENTID_MASK;

  sprintf(lastsub,"TRACE3: raise_non_clients (win,get_icons: %#lx,%d,%#lx)\n",
                                                            C_WINDOW,get_icons,xcid); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: raise_non_clients (win,get_icons: %#lx,%d,%#lx)\n",
                                                            C_WINDOW,get_icons,xcid); 
#else
  if(gVerboseConsole)
    printf("%s info: into raise_non_clients (win,get_icons: %#lx,%d,%#lx)\n",
                                                        PROGNAME,C_WINDOW,get_icons,xcid);
#endif 
 
  XLowerWindow(dpy,check_win);
  XQueryTree(dpy, root, &w1, &w2, &windows, &count);
  j = count;
  for(i=count-1;i>=0;i--) 
  { 
    if( windows[i]!=check_win ) {
    XGetWindowAttributes(dpy, windows[i], &attr);
    // make sure this potential menu is visible, not offscreen, nor very small
    if(attr.map_state==IsViewable) {
    if(attr.x+attr.width<=0  || attr.x>=XDisplayWidth(dpy,screen) ||
        attr.y+attr.height<=0 || attr.y>=XDisplayHeight(dpy,screen) ||
        attr.width<6 || attr.height<6 || attr.width*attr.height<200 ) {
      // don't know if it's a good idea to mess with these offscreen windows, but it might
      // help prevent this occasional problem with windows not coming completely in front...
      if(get_client_from_window(windows[i],FRAME)==NULL)
        XLowerWindow(dpy,windows[i]);
    } else {
      j--;
      windows[j] = windows[i];
      ac=get_client_from_window(windows[j], FRAME); 
      if(ac!=NULL) {
        if((ac->modal_state & ANY_FLOATING)==0 && (ac->is_icon==NOT_ICONBAR || get_icons) && ac->is_docked==0) 
        { 
//          if(c==ac) {
            // a client is being raised, so raise all (non-client) windows that are above top client
            // -this ensures any active menus are also kept in front of the client
            if(j<count-1) {
              XRaiseWindow(dpy,check_win);
              if((windows[count-1] & X_CLIENTID_MASK)==xcid) {
                XRaiseWindow(dpy,windows[count-1]);
                if(gVerboseConsole) printf("   raised window: %#lx %#lx\n",
                                    windows[count-1],windows[count-1] & X_CLIENTID_MASK);
              }
              for(k=count-2;k>j;k--) {
                wc.stack_mode=Above; 
                wc.sibling=check_win;
                if((windows[k] & X_CLIENTID_MASK)==xcid) {
                    XConfigureWindow(dpy,windows[k],CWStackMode|CWSibling,&wc);
                    if(gVerboseConsole) printf("   put window below: %#lx %#lx\n",
                                                windows[k],windows[k] & X_CLIENTID_MASK);
                }
              }
              wc.stack_mode=Above; 
              wc.sibling=check_win;
              configure_client(c, CWStackMode|CWSibling, &wc, c->is_icon!=NOT_ICONBAR, 0); 
            } else {
              wc.stack_mode=Above; 
              configure_client(c, CWStackMode, &wc, c->is_icon!=NOT_ICONBAR, 0);
            }
            if(windows) XFree(windows);
            return 1;
//          }
        }
        // want to get out after top client is found
        break;
      }
    }
    }
    }
  } 

  if(windows) XFree(windows);
  
  return 0; 
} 

/* 
 *  
 * Raise client window 
 * 
 */ 

void raise_client(Client *c) 
{ 
  Client *ac;
  XWindowChanges wc;

  sprintf(lastsub,"TRACE2: raise_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW); 
  subtrace();
//  show_last_fifteen();
#ifdef SHOW_TRACE2 
  printf("TRACE2: raise_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW);
#else
  if(gVerboseConsole)
    printf("%s info: into raise_client (win,last: %#lx,%#lx)\n",PROGNAME,C_WINDOW,FC_WINDOW); 
#endif 
 
  if(c==NULL) 
  { 
#ifdef SHOW_WEIRD 
    printf("%s oddity: client is NULL in raise_client\n",PROGNAME); 
#endif 
    return;
  } 

//  if(c->osxwinfo.view_state==VIEWSTATE_HIDDEN) {
//    gInvisible = 0;
//    map_all_clients();
//    Bring_OOSX_To_Front();
//  }

  if(lastfocus_client!=c)
    restack_autolower_client(lastfocus_client);
    
  if(c->osxwinfo.view_state==VIEWSTATE_HIDDEN) {
    gInvisible = 0;
    if(gVerboseConsole)
      printf("WINDOW TO RAISE IS HIDDEN\n");
    map_group(get_group_for_client(c));
  }
  
  if(c->is_docked) {
     will_undock_client(c);
     return;
  }

  if(c->is_icon==NOT_ICONBAR)
    show_client(c);
    
  ac=c;
  if(c->is_icon==NOT_ICONBAR)
    c=check_client_modality(ac,1);
  sprintf(lastsub,"  after check_client_modality: new,old: %#lx,%#lx)\n",C_WINDOW,AC_WINDOW); 
  subtrace();
 
  if(ac=c) {
    //if(c==lastfocus_client)
      if(raise_non_clients(c,0)) return;
  } else if(c->is_icon==NOT_ICONBAR)
    show_client(c);
        
//  wc.stack_mode=Above; 
//  XConfigureWindow(dpy, check_win, CWStackMode, &wc);
  wc.stack_mode=Above; 
//  XConfigureWindow(dpy, c->frame, CWStackMode, &wc);
  configure_client(c, CWStackMode, &wc, c->is_icon!=NOT_ICONBAR, 0); 
} 

/*
 * Force a particular client to be raised, with no regard for modal children etc.
 */

void force_raise_client(Client *c) 
{ 
//  Client *orgc;
  XWindowChanges wc; 
 
  sprintf(lastsub,"TRACE3: force_raise_client (window,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: force_raise_client (window,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW); 
#else
  if(gVerboseConsole)
    printf("%s info: into force_raise_client (window,last: %#lx,%#lx)\n",PROGNAME,C_WINDOW,FC_WINDOW); 
#endif 
 
  if(c==NULL) 
  { 
#ifdef SHOW_WEIRD 
    printf("%s oddity: client is NULL in force_raise_client\n",PROGNAME); 
#endif 
    return;
  } 
 
  if(c->osxwinfo.view_state==VIEWSTATE_HIDDEN) {
    printf("WINDOW TO FORCE-RAISE IS HIDDEN\n");
    map_group(get_group_for_client(c));
  }
  
//  if(c->osxwinfo.view_state==VIEWSTATE_HIDDEN) {
//    gInvisible = 0;
//    map_all_clients();
//    Bring_OOSX_To_Front();
//  }

  if(c->is_docked) {
    if(gVerboseConsole)
        printf("  docked client in force_raise_client: %#lx (%d)\n",C_WINDOW,c->undock_timer);
    will_undock_client(c);
    return;
  }

  if(c->is_icon==NOT_ICONBAR)
    show_client(c);
  else {
    minmax_client(c);
  }
  
  //if(c==lastfocus_client)
  if(raise_non_clients(c,0)) return;
  
//  wc.stack_mode=Above; 
//  XConfigureWindow(dpy, check_win, CWStackMode, &wc);
  wc.stack_mode=Above; 
//  XConfigureWindow(dpy, c->frame, CWStackMode, &wc);
  configure_client(c, CWStackMode, &wc, c->is_icon!=NOT_ICONBAR, 0); 
} 
 
void raise_all_clients()
{
  Window *windows=NULL, w1, w2; 
  int count, i; 
  Client *c=NULL; 
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE1: raise_all_clients\n"); 
  subtrace();
//#ifdef SHOW_TRACE1 
  printf("TRACE1: raise_all_clients\n"); 
//#endif 
  if(can_send_hackcmds) {
    SendSimpleHackCommand(HACKCMD_ALLTOFRONT,1);
    return;
  }
  
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
  for(i=0;i<count;i++) 
  { 
    XGetWindowAttributes(dpy, windows[i], &attr); 
    if(attr.map_state==IsViewable) {
      c=get_client_from_window(windows[i], FRAME); 
      if(c!=NULL && (c->is_icon==NOT_ICONBAR)) 
        force_raise_client(c);
    }
  } 
  if(windows) XFree(windows);  
}

/*
 *
 * Entered/Left a client
 *
 */

void entered_client(Client *c, int can_focus)
{
  if( delayed_enter_ticks && gVerboseConsole )
    printf("OroborOSX info: new enter/leave (%#lx) overriding old (%#lx)\n",c,delayed_enter_client);
  
  delayed_enter_client=c;
  delayed_can_focus=can_focus;
  delayed_enter_ticks=GetCarbonTickCount();
  if( delayed_enter_ticks==0 ) delayed_enter_ticks=1;
}

void delayed_entered_client(Client *c, int can_focus)
{
  Client *fc, *ac;
  
  ac = overicon;
  sprintf(lastsub,"TRACE1: delayed_entered_client  win: %#lx (can_focus: %d) icon: %#lx\n",
                                                                C_WINDOW,can_focus,AC_WINDOW); 
  subtrace();
#ifdef SHOW_ENTER_LEAVE 
  printf("delayed_entered_client  win: %#lx (can_focus: %d) icon: %#lx\n",
                                                    C_WINDOW,can_focus,AC_WINDOW);
#else
#ifdef SHOW_TRACE1 
  printf("TRACE1: delayed_entered_client  win: %#lx (can_focus: %d) icon: %#lx\n",
                                                        C_WINDOW,can_focus,AC_WINDOW); 
#endif 
#endif

  delayed_enter_client=NULL;
  delayed_enter_ticks=0;
  
/*
  if(overicon==NULL && iconbar_state==ICONBAR_VISIBLE) {
#ifdef SHOW_WEIRD
    printf("Overicon is NULL, but iconbar visible? %d\n",iconbar_state);
#endif
    if(c==NULL)
      will_hide_all_iconbars();
    else if(c->is_icon!=NOT_ICONBAR)
      overicon=c;
    else
      overicon=client_list;
  }
*/ 
  if(c!=NULL) {
    if(c->is_docked) return;
    if(c->isoffscreen!=0) return;
  }
//      if(undock_client(c)) return;

  if(overicon!=NULL) {
#ifdef SHOW_WEIRD
// we were over an iconbar, but the bars were not visible??
     if(iconbar_state!=ICONBAR_VISIBLE)
       printf("%s oddity:  icon: %#lx , but not visible? %d\n",PROGNAME,AC_WINDOW,iconbar_state);
#endif
     if(overicon->is_icon!=ICONBAR_RAISED) {
// we were over an iconbar, but it was not lit up??
#ifdef SHOW_WEIRD
       printf("%s oddity:  icon: %#lx , but not raised? %d\n",PROGNAME,AC_WINDOW,overicon->is_icon);
#endif
// if we're not over a new one then hide the iconbars
       if(c==NULL)
         will_hide_all_iconbars();
       else if(c->is_icon==NOT_ICONBAR)
         will_hide_all_iconbars();
       overicon->is_e_m_r &= NOT_ENTERED; // make sure entered bit is not set
       overicon=NULL;
     } else if(overicon!=c) {
// over a different iconbar (or none) to what we were
       if(iconbar_state!=ICONBAR_VISIBLE) {
#ifdef SHOW_WEIRD
          printf("%s oddity: How did that happen? iconbar_state=%d, icon=%#lx\n",PROGNAME,iconbar_state,AC_WINDOW);
#endif
          slide_all_iconbars(overicon);
       }
       if(c==NULL) {
// not over a new iconbar, so prepare to hide them
         overicon->is_icon=ICONBAR_VISIBLE;
         will_hide_all_iconbars();
         if( overicon->icon_w!=iconbar_width ) {
           resize_client(overicon,iconbar_width,overicon->icon_h,3);
           raise_client(overicon);
         }
         overicon->is_e_m_r &= NOT_ENTERED;
         redraw_client_frame(overicon);
       } else if(c->is_icon==NOT_ICONBAR) {
// not over a new iconbar, so prepare to hide them
         overicon->is_icon=ICONBAR_VISIBLE;
         will_hide_all_iconbars();
         if( overicon->icon_w!=iconbar_width ) {
           resize_client(overicon,iconbar_width,overicon->icon_h,3);
           raise_client(overicon);
         }
         overicon->is_e_m_r &= NOT_ENTERED;
         redraw_client_frame(overicon);
       } else {
// still over an iconbar, so don't hide them, but make sure last one is redrawn correctly
         overicon->is_icon=ICONBAR_VISIBLE;
         if( overicon->icon_w!=iconbar_width ) {
           resize_client(overicon,iconbar_width,overicon->icon_h,3);
           raise_client(overicon);
         }
         overicon->is_e_m_r &= NOT_ENTERED;
         redraw_client_frame(overicon);
       }
//       lower_client(overicon);
       overicon=NULL;
     }
  } else if(c!=NULL) {
     if(iconbar_state!=ICONBAR_HIDDEN && lastIconbarOutTicks==0 && can_send_hackcmds==0) {
#ifdef SHOW_WEIRD
       printf("Iconbars not ready to be hidden, but overicon NULL?!\n");
#endif
       will_hide_all_iconbars();
     }
  }

  if(c==NULL)
  {
    fc=lastfocus_client;
//    fc=get_focus_client();
    if(fc!=NULL) {
        restack_autolower_client(fc);
        ac=get_top_most_client(WIN_LAYER_NORMAL,False);
        if(gX11InFront==1 && ac!=fc && ac!=NULL) set_focus_client(ac);
    }
  }
  else if(c->is_icon!=NOT_ICONBAR)
  {
    if(overicon!=c)
      if(c->icon_w<iconbar_width)
        c->is_e_m_r |= LOCK_ENTERED;
// we are over a new iconbar, show them if not already
    slide_all_iconbars(c);
// prepare for new iconbar by changing width (if necessary) and 'light it up'
    c->is_icon=ICONBAR_RAISED;
    if(overicon!=c) {
      if( c->icontxt_w>=TW-6 ) {
        resize_client(c,c->icontxt_w+6-LW-RW+TLW+TRW,c->icon_h,3);
        raise_client(c);
        redraw_client_frame(c);
        c->is_e_m_r |= LOCK_ENTERED;
      } else {
        raise_client(c);
        redraw_client_frame(c);
      }
    }
    overicon=c;
  }
  else if(can_focus) {
    if((c->auto_properties & AUTO_FOCUS)==AUTO_FOCUS) {
      if(gX11InFront==1 && lastfocus_client!=c) {
        restack_autolower_client(lastfocus_client);
        set_focus_client(c);
      }
    } else {
      fc=lastfocus_client;
      if(fc!=NULL) {
        restack_autolower_client(fc);
        ac=get_top_most_client(WIN_LAYER_NORMAL,False);
        if(gX11InFront==1 && ac!=fc && ac!=NULL) set_focus_client(ac);
      }
    }
  }

  if(can_send_hackcmds!=0) {
    if(c!=NULL) {
      if(gX11InFront!=0 && (c==lastfocus_client || (c->auto_properties & (AUTO_FOCUS|NONAUTO_FIRSTCLICK))
                                                || client_modal_firstclick(c)==lastfocus_client )) {
        pointer_grabbing(0);
      } else {
        pointer_grabbing(c->frame);
      }
    }
  }
}

/*
 * Glue routine to be called from main.c to say we've left a client
 */
void left_client()
{
    if(can_send_hackcmds==0) entered_client(NULL,0);
}

/*
 * Want to check if a client has modal children, and returns the 'newest' child
 * Also, raises modal windows as it finds them if raiseit is non-zero
 */
 
Client *check_client_modality(Client *orgc, int raiseit)
{
  Client *newc,*c,*tmpc;
  
  c=orgc;
  newc=c;
  if(c->first_child) {
    if(c->modal_state & HAS_ANY_MODAL_CHILD) {
      if( raiseit ) force_raise_client(c);
      c=c->first_child;
      if((c->modal_state & ANY_MODAL)>DONT_IGNORE_MODAL) {
        newc=c;
        if( raiseit ) force_raise_client(c);
      }
      if(c->modal_state & HAS_ANY_MODAL_CHILD)
        newc=check_client_modality(c,raiseit);
      while(c->next_sibling) {
        c=c->next_sibling; 
        if((c->modal_state & ANY_MODAL)>DONT_IGNORE_MODAL) {
          newc=c;
          if( raiseit ) force_raise_client(c);
        }
        if(c->modal_state & HAS_ANY_MODAL_CHILD)
          newc=check_client_modality(c,raiseit);
      }
    }
  }
  return newc;
}

/*
 * - checks if a client is set to accept first click when modal child is topmost
 * - and follows through parent-child chain to return topmost modal child (so it
 * - can be compared against the current focus client)
 */
Client *client_modal_firstclick(Client *c)
{
    Client *ac=NULL;
    
    if(c==NULL) {
        // make sure we don't return lastfocus_client
        if(lastfocus_client==NULL) return client_list;
        return NULL;
    }
    
    if((c->auto_properties | NONAUTO_FIRSTCLICKCHILD)==0) {
        // make sure we don't return lastfocus_client
        if(lastfocus_client==NULL) return client_list;
        return NULL;
    }

    ac=check_client_modality(c,0);

    if(ac==c) {
        // make sure we don't return lastfocus_client
        if(lastfocus_client==ac) return NULL;
        return ac;
    }
    
    if(ac==NULL) {
        // make sure we don't return lastfocus_client
        if(lastfocus_client==NULL) return client_list;
        return NULL;
    }
    
    return ac;
}

/* 
 *  
 * Give client window keyboard focus 
 * 
 */ 

void quick_set_focus_client(Client *c)
{
  Window focus_win; 
  int dummy;
  
  if(c==NULL) return;
  if(c->is_icon!=NOT_ICONBAR) return;

  if(gVerboseConsole) {
    XGetInputFocus(dpy, &focus_win, &dummy); 
    printf("Quick focus for win: %#lx (now: %#lx %d [rtp,rtr,rtn: %d %d %d])\n",C_WINDOW,focus_win,dummy,
            RevertToParent,RevertToPointerRoot,RevertToNone);
    //XSetInputFocus(dpy, check_win, RevertToNone, CurrentTime);
    //MyXSync(dpy,0);
  }
  if(c!=lastfocus_client) {
    printf("%s Warning: got quick focus for win,last = %#lx,%#lx ???\n",PROGNAME,C_WINDOW,FC_WINDOW); 
    lastfocus_client=c;
    lastfocus=c->window;
  }
//  } else
//    printf("%s info: quick_set_focus_client: %#lx\n",PROGNAME,c->window);
  hacking_focus = 0;
  XSetInputFocus(dpy, c->window, RevertToNone, CurrentTime);
}

void set_focus_client(Client *c) 
{ 
  Client *ac;
  int can_af=-1, can_ar=-1, can_al=-1, should_raise=0, first_click=0;
  int click_raise=0, live_resize=0, retain_focus=0, no_grab_focus=0, client_modality=0;
  Window w1, w2; 
  int count, i, j; 
  XWindowAttributes attr;
  
  sprintf(lastsub,"TRACE1: set_focus_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: set_focus_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW); 
#else
  if(gVerboseConsole)
    printf("%s info: into set_focus_client (win,last: %#lx,%#lx)\n",PROGNAME,C_WINDOW,FC_WINDOW); 
#endif 
 
  hacking_focus = 0;
  
  if(c==NULL) 
  { 
#ifdef DEBUG
    printf("  got NULL client in set_focus_client\n"); 
#endif
//    XUngrabPointer(dpy,CurrentTime);
    XSetInputFocus(dpy, check_win, RevertToNone, CurrentTime);
//    XGrabPointer(dpy, check_win, True, ButtonPressMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, resize_curs, CurrentTime); 
//    XAllowEvents(dpy, SyncPointer, CurrentTime);
    if(lastfocus_client!=NULL) {
      c=lastfocus_client;
      if((gBGWinSetting & BGWIN_TRANSLUCENT)==BGWIN_TRANSLUCENT)
        SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.inactive_translucency,0,0,0,0);
      lastfocus_client=NULL;
      lastfocus=0;
      if((gBGWinSetting & BGWIN_DIMMED)==BGWIN_DIMMED && c->is_dimmed==0)
        SendHackCommand(c,1,HACKCMD_DIMWIN,1,0,0,0,0);
      if((c->has_frame) || (c->is_icon!=NOT_ICONBAR)) redraw_client_frame(c);
      SendHackCommand(c,1,HACKCMD_SETSHADOW,0,0,0,0,0);
      if(gVerboseConsole) printf("++++ never accept first click...\n");
      SendHackCommand(c,1,HACKCMD_SETFIRSTCLICK,0,0,0,0,0);
      restack_autolower_client(c);
//      if((c->auto_properties & AUTO_FOCUS|NONAUTO_FIRSTCLICK)==0 || gX11InFront==0)
//        stack_cover(c,Above);
//      else
//        stack_cover(c,Below);
    }
    lastfocus=0; // just to make sure ;-)
    Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0); // No active client, so disable Window & Properties items
    Setup_Properties_Menu_Translucency(NULL); // No active client, so disable translucency items in Properties
    return;
  } 

  Set_End_Whole_Screen_Mode();
  
  if(c->osxwinfo.view_state==VIEWSTATE_HIDDEN) {
    gInvisible = 0;
    map_group(get_group_for_client(c));
//    map_all_clients();
//    Bring_OOSX_To_Front();
  }

  ac=c;
  c=check_client_modality(ac,0);
  sprintf(lastsub,"  after check_client_modality:  new,old: %#lx,%#lx\n",C_WINDOW,AC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("  after check_client_modality:  new,old: %#lx,%#lx\n",C_WINDOW,AC_WINDOW); 
#endif

  if(c->is_docked) {
//    will_undock_client(c);
    return;
  }
  
  ac=c;
    if(lastfocus_client!=NULL && lastfocus_client!=c) {
      c=lastfocus_client;
      if((gBGWinSetting & BGWIN_TRANSLUCENT)==BGWIN_TRANSLUCENT
                && client_modal_firstclick(c)!=lastfocus_client
                && (c->auto_properties & (NONAUTO_FIRSTCLICK | AUTO_FOCUS))!=NONAUTO_FIRSTCLICK)
        SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.inactive_translucency,0,0,0,0);
      lastfocus_client=NULL;
      if((gBGWinSetting & BGWIN_DIMMED)==BGWIN_DIMMED && c->is_dimmed==0
                && client_modal_firstclick(c)!=lastfocus_client
                && (c->auto_properties & (NONAUTO_FIRSTCLICK | AUTO_FOCUS))!=NONAUTO_FIRSTCLICK)
        SendHackCommand(c,1,HACKCMD_DIMWIN,1,0,0,0,0);
      if((c->has_frame) || (c->is_icon!=NOT_ICONBAR)) redraw_client_frame(c);
      restack_autolower_client(c);
      lastfocus_client=c;
      SendHackCommand(c,1,HACKCMD_SETSHADOW,0,0,0,0,0);
//      if((c->auto_properties & AUTO_FOCUS|NONAUTO_FIRSTCLICK)==0 || gX11InFront==0)
//        stack_cover(c,Above);
//      else
//        stack_cover(c,Below);
      if( (c->auto_properties & (NONAUTO_FIRSTCLICK | AUTO_FOCUS)) &&
          (c->auto_properties & NONAUTO_IGNORECLICKS)==0 ) {
        if(gVerboseConsole) printf("++++ lastfocus does accept first click...\n");
        SendHackCommand(c,1,HACKCMD_SETFIRSTCLICK,1,0,0,0,0);
      } else {
        if(gVerboseConsole) printf("++++ lastfocus does not accept first click...\n");
        SendHackCommand(c,1,HACKCMD_SETFIRSTCLICK,0,0,0,0,0);
      }
      c=ac;
    }
      
  if(c->is_icon==NOT_ICONBAR)
  { 
#ifdef DEBUG 
    printf("  focusing client [frame: %#lx] win: %#lx\n",c->frame,c->window); 
#endif 
//    XUngrabPointer(dpy,CurrentTime);
    XSetInputFocus(dpy, c->window, RevertToNone, CurrentTime);
//    XGrabPointer(dpy, c->frame, True, ButtonPressMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, resize_curs, CurrentTime); 
//    XAllowEvents(dpy, SyncPointer, CurrentTime);

    if(lastfocus_client!=c || lastfocus!=c->window) {
       lastfocus_client=c;
       lastfocus=c->window;
       if((gBGWinSetting & BGWIN_TRANSLUCENT)==BGWIN_TRANSLUCENT && gX11InFront!=0)
         SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.translucency,0,0,0,0);
       if((gBGWinSetting & BGWIN_DIMMED)==BGWIN_DIMMED && c->is_dimmed!=0 && gX11InFront!=0)
         SendHackCommand(c,1,HACKCMD_DIMWIN,0,0,0,0,0);
       if((c->has_frame) || (c->is_icon!=NOT_ICONBAR)) redraw_client_frame(c);
       SendHackCommand(c,1,HACKCMD_SETSHADOW,1,0,0,0,0);
    }
    if((c->auto_properties & NO_GRAB_FOCUS)==NO_GRAB_FOCUS)
        no_grab_focus=1;
    if((c->auto_properties & RETAIN_FOCUS)==RETAIN_FOCUS)
        retain_focus=1;
    if((c->auto_properties & LIVE_RESIZING)==LIVE_RESIZING)
        live_resize=1;
    if((c->auto_properties & NONAUTO_FIRSTCLICKCHILD)==NONAUTO_FIRSTCLICKCHILD)
        first_click=2;
    if((c->auto_properties & NONAUTO_FIRSTCLICK)==NONAUTO_FIRSTCLICK)
        first_click+=1;
    if((c->auto_properties & NONAUTO_DRAGFIRSTCLICK)==NONAUTO_DRAGFIRSTCLICK)
        first_click+=4;
    if((c->auto_properties & NONAUTO_IGNORECLICKS)==NONAUTO_IGNORECLICKS)
        first_click+=8;
    // Hmmmm... how does this work?!? (i.e. how can a client which ignores focus ever
    // get into set_focus_client?!) - so how does one change the properties for it...???!
    if((c->auto_properties & NONAUTO_IGNOREFOCUS)==NONAUTO_IGNOREFOCUS)
        first_click+=16;
    if((c->auto_properties & AUTO_CLICKRAISE)==AUTO_CLICKRAISE)
        click_raise=1;
    if((c->auto_properties & AUTO_FOCUS)==AUTO_FOCUS)
       can_af=1;
    if((c->auto_properties & AUTO_LOWER)==AUTO_LOWER)
       can_al=1;
    if((c->auto_properties & AUTO_RAISE)==AUTO_RAISE) {
       should_raise=1;
       can_ar=1;
       if(can_al==1) {
         if(c->al_list!=NULL) {
           printf("Found autolower list already! (cnt: %d)\n",c->al_count);
//            XFree(c->al_list);
         } else {
           XQueryTree(dpy, root, &w1, &w2, &c->al_list, &c->al_count);
           if(c->al_list!=NULL) {
             j = 0;
             for(i=0;i<c->al_count;i++) {
               if(c->al_list[i] != check_win ) {
                 XGetWindowAttributes(dpy, c->al_list[i], &attr);
                 if(attr.map_state==IsViewable) {
                   c->al_list[j] = c->al_list[i];
                   j++;
                 }
               }
             }
             c->al_count = j;
             if(gVerboseConsole)
               printf("Ready for lower count: %d\n",j);
           }
         }
       }
    } else if(can_af<1)
       should_raise=1;
    else if(can_af>0) {
       should_raise=0;
    }
    if(should_raise==1)
        raise_client(c);

//      if(c!=get_top_most_client(WIN_LAYER_NORMAL,False))
    ac = c;
    while(ac->osxwinfo.winref==0 && ac->parent!=NULL)
      ac = ac->parent;
    if(ac->osxwinfo.winref!=0)
      FocusOSXWin(&(ac->osxwinfo));

    if(c->modal_state & STRICT_ANY_MODAL) {
      if(c->modal_state & DONT_IGNORE_MODAL)
        client_modality=1;
      else
        client_modality=-1;
    } else
      client_modality=0;
      
    if( first_click & 8 )
        SendHackCommand(c,1,HACKCMD_SETFIRSTCLICK,0,0,0,0,0);
    else
        SendHackCommand(c,1,HACKCMD_SETFIRSTCLICK,1,0,0,0,0);

    Setup_Window_Menu(1, c->has_frame, 1, 1, c->can_max, 0, can_af, can_ar, can_al, first_click, click_raise, live_resize, no_grab_focus, retain_focus, client_modality);
  } 

  sprintf(lastsub,"TRACE2: leaving set_focus_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: leaving set_focus_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW); 
#endif 
} 

/*
 * When XDarwin is brought in front, we have to also (unfortunately) bring all
 * clients that have auto-focus set (to work around mouse tracking issue in XDarwin)
 * -not any more! (If the 'hacked' version of XDarwin is used this routine is not called)
 */
void raise_auto_clients(Client *skipc)
{
  Client *c;
  int i;
  
  sprintf(lastsub,"TRACE2: raise_auto_clients \n");
  subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: raise_auto_clients \n");
#endif

  c=client_list;
  if(c==NULL) return;
  
  for(i=0;i<client_count;i++) {
    if(c->is_icon==NOT_ICONBAR)
      if((c->auto_properties & AUTO_FOCUS)!=0)
        if(c->is_docked==0)
          if(c!=skipc)
            raise_client(c);
    c=c->next;
  }
}

/*
 * OroborOSX has come in front, so reactivate the focus window
 */
void come_to_front()
{
    Client *c=NULL, *ac=NULL;
    int i=0;

  sprintf(lastsub,"TRACE2: come_to_front \n");
  subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: come_to_front \n");
#endif

    Set_End_Whole_Screen_Mode();
    if(can_send_hackcmds) MyIncreaseXServerGrab();
    
    c=get_focus_client();
    if( c ) {
      if((gBGWinSetting & BGWIN_TRANSLUCENT)==BGWIN_TRANSLUCENT)
        SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.translucency,0,0,0,0);
      if((gBGWinSetting & BGWIN_DIMMED)==BGWIN_DIMMED && c->is_dimmed!=0)
        SendHackCommand(c,1,HACKCMD_DIMWIN,0,0,0,0,0);
      if((c->has_frame) || (c->is_icon!=NOT_ICONBAR)) redraw_client_frame(c);
      SendHackCommand(c,1,HACKCMD_SETSHADOW,1,0,0,0,0);
      if(gVerboseConsole) printf("++++ accept first click for focus in come_to_front...\n");
      SendHackCommand(c,1,HACKCMD_SETFIRSTCLICK,1,0,0,0,0);
      pointer_grabbing(0);
//      stack_cover(c,Below);
    }
    
    ac = client_list;
    for(i=0;i<client_count;i++) {
      if( !ac )
        i = client_count;
      else if( (ac->auto_properties & (AUTO_FOCUS|NONAUTO_FIRSTCLICK)) &&
                (ac->auto_properties & NONAUTO_IGNORECLICKS)==0 &&
                 ac->is_icon==NOT_ICONBAR && ac->is_docked==0)
        SendHackCommand(ac,1,HACKCMD_SETFIRSTCLICK,1,0,0,0,0);
      ac = ac->next;
    }
    
    ac=get_top_most_client(WIN_LAYER_NORMAL,False);
    if(can_send_hackcmds==0)
      raise_auto_clients(ac);
    if( ac ) {
      //printf("come_to_front, so raising check_win first before client...\n");
      //XRaiseWindow(dpy,check_win);
      raise_client(ac);
    }

    if(can_send_hackcmds) MyDecreaseXServerGrab();

#ifdef SHOW_EVENTS
    printf("come_to_front (focus,top: %#lx,%#lx)\n",C_WINDOW,AC_WINDOW);
#else
    if(gVerboseConsole)
        printf("come_to_front (focus,top: %#lx,%#lx)\n",C_WINDOW,AC_WINDOW);
#endif
}

/*
 * Another OS X app has come in front, so deactivate the focus window
 */

void gone_to_back()
{
    Client *c=NULL, *ac=NULL;
    int need_ungrab=0, i=0;
    
  sprintf(lastsub,"TRACE2: gone_to_back \n");
  subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: gone_to_back \n");
#endif

    // need to make sure any pull-down menus are removed
    // this is done by telling XDarwin to send a mouse-release event (using button 5 -hope that's ok!)
    need_ungrab = check_need_ungrab();
/*
  if(lastfocus_client!=NULL && pointer_grabbed==0) {
    retval=XGrabPointer(dpy, lastfocus_client->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeSync, GrabModeSync, None, normal_curs, CurrentTime);
    if(retval==GrabSuccess) {
        XUngrabPointer(dpy, CurrentTime);
        XAllowEvents(dpy, SyncPointer, CurrentTime);
        printf("Got %d from XGrabPointer -no need for fake button click.\n",retval);
    } else {
        need_ungrab = 1;
    }
  }*/
  
  Set_End_Whole_Screen_Mode();
  if(can_send_hackcmds) MyIncreaseXServerGrab();

    c=get_focus_client();
    if( c!=NULL ) {
      if((gBGWinSetting & BGWIN_TRANSLUCENT)==BGWIN_TRANSLUCENT)
        SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.inactive_translucency,0,0,0,0);
      if((gBGWinSetting & BGWIN_DIMMED)==BGWIN_DIMMED && c->is_dimmed==0)
        SendHackCommand(c,1,HACKCMD_DIMWIN,1,0,0,0,0);
      if((c->has_frame) || (c->is_icon!=NOT_ICONBAR)) redraw_client_frame(c);
      if(gVerboseConsole) printf("++++ never accept first click in gone_to_back...\n");
      SendHackCommand(c,1,HACKCMD_SETFIRSTCLICK,0,0,0,0,0);
//      if((c->auto_properties & AUTO_FOCUS|NONAUTO_FIRSTCLICK)==0 || gX11InFront==0)
//        stack_cover(c,Above);
//      else
//        stack_cover(c,Below);
    }
    SendSimpleHackCommand(HACKCMD_SETSHADOW,1);
    
    ac = client_list;
    for(i=0;i<client_count;i++) {
      if( !ac )
        i = client_count;
      else if( ac->is_icon==NOT_ICONBAR && ac->is_docked==0 )
        SendHackCommand(ac,1,HACKCMD_SETFIRSTCLICK,0,0,0,0,0);
      ac = ac->next;
    }

    if(can_send_hackcmds) MyDecreaseXServerGrab();

    if(need_ungrab) {
        printf("%s info: Got %d from XGrabPointer - asking XDarwin for fake button click.\n",PROGNAME,need_ungrab);
        //SendSimpleHackCommand(HACKCMD_UNGRABPTR,0);
        force_grab_release();
    } else if(gVerboseConsole)
        printf("%s info: Got %d from XGrabPointer -no need for fake button click.\n",PROGNAME,need_ungrab);
    
    if(c!=NULL)
      pointer_grabbing(c->frame);
    
//    XUngrabPointer(dpy, CurrentTime); 
//    XGrabPointer(dpy, check_win, True, ButtonPressMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, resize_curs, CurrentTime);
//    XAllowEvents(dpy, SyncPointer, CurrentTime);

#ifdef SHOW_EVENTS
    printf("gone_to_back (focus: %#lx)\n",C_WINDOW);
#else
    if(gVerboseConsole)
        printf("gone_to_back (focus: %#lx)\n",C_WINDOW);
#endif
}

/* 
 * 
 * Lower client window 
 * 
 */ 

void lower_client(Client *c) 
{ 
  XWindowChanges wc; 
 
  sprintf(lastsub,"TRACE3: lower_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: lower_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW); 
#endif 
 
  if(c==NULL) {
#ifdef SHOW_WEIRD
    printf("%s oddity: client is NULL in lower_client\n",PROGNAME);
#endif
    return;
  }
  
  if(c->is_docked) return;
//    if(undock_client(c)) return;

  if(c->is_icon==NOT_ICONBAR) {
    if((c->auto_properties & (AUTO_FOCUS|AUTO_RAISE))!=0) {
//#ifdef SHOW_WEIRD
//      printf("  %s question... should an auto-focus/auto-raise client really be lowered?  -hmmmm...\n",PROGNAME);
//#endif
//    return;
    }
  }
  
  wc.stack_mode=Below; 
//  XConfigureWindow(dpy, c->frame, CWStackMode, &wc);
//  redraw_client_frame(c);
  configure_client(c, CWStackMode, &wc, c->is_icon!=NOT_ICONBAR, 0);
  
  if(c->is_icon==NOT_ICONBAR) {
    while(c->parent!=NULL && (c->modal_state & ANY_MODAL)>DONT_IGNORE_MODAL) {
      c = c->parent;
      if(c->is_icon==NOT_ICONBAR)
          configure_client(c, CWStackMode, &wc, 0, 0);
    }
  }
} 
 
/* 
 *  
 * Show client window 
 * 
 */
 
void show_client(Client *c) 
{ 
  sprintf(lastsub,"TRACE3: show_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: show_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW); 
#endif 
 
  if(c==NULL) {
#ifdef SHOW_WEIRD
    printf("%s oddity: client is NULL in show_client\n",PROGNAME);
#endif
    return;
  } 

  if(c->is_docked) {
    will_undock_client(c);
    return;
  }
  
  c->is_icon=NOT_ICONBAR;
  XMapWindow(dpy, c->frame); 
  XMapWindow(dpy, c->window);
  if(c->cover) XMapWindow(dpy, c->cover);
  set_wm_state(c->window, NormalState); 
} 
 
/* 
 * 
 * Hide client window - this is the original iconisation from oroborus which is no longer used
 * 
 */ 

void old_hide_client(Client *c) 
{ 
  sprintf(lastsub,"TRACE3: hide_client (old)  %#lx\n",C_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: hide_client (old)  %#lx\n",C_WINDOW); 
#endif 

  if(c==NULL) {
#ifdef SHOW_WEIRD
    printf("%s oddity: client is NULL in old_hide_client\n",PROGNAME);
#endif
    return;
  }

  /* We need to ignore one of the unmaps, otherwise our client gets deleted */ 
  c->ignore_unmap++; 
  XUnmapWindow(dpy, c->frame); 
  XUnmapWindow(dpy, c->window); 
  set_wm_state(c->window, IconicState); 
} 
 
/*  
 * 
 * Configure client - the main resizing, raising, lowering function 
 * 
 */ 

void configure_client(Client *c, unsigned int mask, XWindowChanges *wc, int frame_only, int config_win) 
{ 
  XConfigureEvent ce; 
  int w, w1, h, h1, orig_is_icon, was_same_size=1, really_was_same_size=1;
  unsigned int orgmask;
  Client *topc;
 
  sprintf(lastsub,"TRACE2: configure_client (win,mask,frame_only,lastfocus: %#lx,%d,%d,%#lx)\n",
                                                                C_WINDOW,mask,frame_only,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: configure_client (win,mask,frame_only,lastfocus: %#lx,%d,%d,%#lx) %d\n",
                                                C_WINDOW,mask,frame_only,lastfocus,CWStackMode); 
#else
  if(gVerboseConsole)
    printf("%s info: into configure_client (win,mask,frame_only,lastfocus: %#lx,%d,%d,%#lx) %d\n",
                                            PROGNAME,C_WINDOW,mask,frame_only,lastfocus,CWStackMode);
  /*printf("configure_client: ");
  if(mask&(CWWidth|CWHeight)) printf("%d x %d ",wc->width,wc->height);
  if(mask&(CWX|CWY)) printf("%d x %d ",wc->x,wc->y);
  if(mask&CWStackMode) printf(" reorder");
  printf("\n");*/
#endif 

  //show_last_fifteen();
  
  if(c==NULL) {
#ifdef SHOW_WEIRD
  printf("%s oddity: client is NULL in configure_client\n",PROGNAME);
#endif
    return;
  }
  
//  if(c->is_docked) return;
//    if(undock_client(c)) return;

  if(can_send_hackcmds) MyIncreaseXServerGrab();
  
#ifdef XINERAMA_SUPPORT
  if(xinerama) update_xinerama_screens();
#endif
  
  if(config_win==3 || config_win==4 || c->is_icon!=NOT_ICONBAR) was_same_size=0;
  
  orig_is_icon=c->is_icon;

// if the application itself has requested a window move (i.e. not being dragged by user)...
  if(config_win==100) {
    config_win=1;
#ifdef XINERAMA_SUPPORT
    if(xinerama && (mask & (CWX|CWY))!=0 && (c->is_e_m_r & (IS_MOVING|SHOWING_SIZE))==0 && orig_is_icon==NOT_ICONBAR) {
      int head=0;
      int x=c->x;
      int y=c->y;
      h=c->height;
      w=c->width;
      if(mask & CWX) x=wc->x;    
      if(mask & CWY) y=wc->y;
      if(mask & CWHeight) h=wc->height;    
      if(mask & CWWidth) w=wc->width;
      if((c->modal_state & STRICT_ANY_MODAL)==0 || c->parent==NULL)
        head=-2;
      else
        head=find_rect_head(c->parent->x,c->parent->y,c->parent->width,c->parent->height);
      check_window_pos(x,y,&w,&h,&w1,&h1,&head);
      if(wc->x!=w1 || wc->y!=h1) {
        mask |= (CWX|CWY);
        wc->x = w1;
        wc->y = h1;
      }
    }
#endif /* XINERAMA */
// make sure the client will get an event, even if the size has not actually changed
    was_same_size=0;
  }
  
// choose the height and width depending on whether the client is an iconbar or not
  if(c->is_icon!=NOT_ICONBAR) {
    h=c->icon_h;
    w=c->icon_w;
    if(mask & CWX) c->icon_x=wc->x;
    if(mask & CWY) c->icon_y=wc->y;
  } else {
    h=c->height;
    w=c->width;
    if(mask & CWX) c->x=wc->x;
    if(mask & CWY) c->y=wc->y;
  }

/*
  if(mask & CWWidth) {
    if(w==wc->width)
      mask=mask ^ CWWidth;
  }

  if(mask & CWHeight) {
    if(h==wc->height)
      mask=mask ^ CWHeight;
  }
*/

  if(!frame_only) {
// if it wants the whole frame, then it should not be an iconbar!
#ifdef SHOW_WEIRD
    if(c->is_icon!=NOT_ICONBAR)
      printf("%s oddity: Asked to configure an iconbar with more than just frame!\n",PROGNAME);
#endif
    c->is_icon=NOT_ICONBAR;
  }

  if(mask & CWWidth) 
  { 
    w=wc->width; 
 
    /*printf("Setting width: %d , %d , %d , %d , %d\n",w,c->icontxt_x,c->name_w,c->width,c->icon_w);*/

    if( c->is_icon!=NOT_ICONBAR && frame_only) {
      // reset the iconbar width
      if(w<1) w=1; 
      c->icon_w=w;
      was_same_size=0;
    } else {

      /* Get resize increments */ 
      if(c->size->flags & PResizeInc) 
      { 
        w1=(w-c->size->min_width)/c->size->width_inc; 
        w=c->size->min_width+(w1*c->size->width_inc); 
      } 
      /* Get min size */ 
      if(c->size->flags & PMinSize) 
        if(w<c->size->min_width) w=c->size->min_width; 
      /* Get max size */ 
      if((c->size->flags & PMaxSize) && c->size->max_width>0)
        if(w>c->size->max_width) w=c->size->max_width; 
 
      /* Make sure we never set width to <1 */ 
      if(w<1) w=1; 

      if(c->width!=w) {
        was_same_size = 0;
      	really_was_same_size = 0;
      }
      c->width=w;
    }
  } 

  if(mask & CWHeight) 
  { 
    h=wc->height; 
 
    if( c->is_icon!=NOT_ICONBAR && frame_only ) {
      // reset the iconbar height
      if(h<1) h=1; 
      c->icon_h=h;
      was_same_size=0;
    } else {

      /* Get resize increments */ 
      if(c->size->flags & PResizeInc) 
      { 
        h1=(h-c->size->min_height)/c->size->height_inc; 
        h=c->size->min_height+(h1*c->size->height_inc); 
      } 
      /* Get min size */ 
      if(c->size->flags & PMinSize) 
        if(h<c->size->min_height) h=c->size->min_height; 
      /* Get max size */ 
      if((c->size->flags & PMaxSize) && c->size->max_height>0)
        if(h>c->size->max_height) h=c->size->max_height; 
 
      /* Make sure we never set height to <1 */ 
      if(h<1) h=1; 

      if(c->height!=h) {
        was_same_size = 0;
        really_was_same_size = 0;
      }
      c->height=h;
    }
  } 
 
  // if the only change was a resize, and no resize has actually happened, get out
//  if((mask|CWHeight|CWWidth)==(CWHeight|CWWidth) && was_same_size!=0) {
//    printf("getting out of configure_client since no resize happened...\n");
//    return;
//  }
  
#ifdef LAYER_SUPPORT 
  /* Deal with window layers */ 
  if(mask & CWStackMode) 
  { 
    Client *sibling=NULL; 
 
    /* Takes care of layers */ 
    switch(wc->stack_mode) 
    { 
    case Above: 
      /* Window wants to be raised to top of its layer */ 
      sibling=get_top_most_client(get_gnome_hint(c->window, WIN_LAYER),True); 
      if(sibling==NULL) 
        wc->stack_mode=Below; 
      break; 
    case Below: 
      /* Window wants to be lowered to bottom of its layer */ 
      sibling=get_bottom_most_client(get_gnome_hint(c->window, WIN_LAYER),True); 
      break; 
    } 
    if(sibling==NULL) 
    { 
      if(mask & CWSibling) mask=mask ^ CWSibling; 
    } 
    else if(sibling==c) 
    { 
      if(mask & CWStackMode) mask=mask ^ CWStackMode; 
      if(mask & CWSibling) mask=mask ^ CWSibling; 
    } 
    else 
    { 
      mask=mask | CWSibling; 
      wc->sibling=sibling->frame; 
    } 
#ifdef WORKSPACE_SUPPORT 
    set_gnome_hint(c->window, WIN_WORKSPACE, workspace); 
#endif /* WORKSPACE */ 
    if(c->is_docked) {
        printf("%s warning: Tried to change stacking for docked window (%#lx)\n",PROGNAME,C_WINDOW);
        mask = mask ^ CWStackMode;
        wc->sibling=0;
        if(mask & CWSibling) mask=mask ^ CWSibling; 
    }
  } 
#endif /* LAYER */ 
 

#ifdef SHAPE_SUPPORT
//  if(shape) 
//    XShapeSelectInput(dpy, c->window, ShapeNotifyMask); 
#endif

  if((mask & (CWHeight|CWWidth))!=0 && was_same_size==0) {
    //if((mask & (CWX|CWY))!=0)
        //SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,201,0,0,0,0);
    SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,HACKARG_INHIBITDRAW,0,0,0,0);
//    SendHackCommand(c,1,HACKCMD_INHIBITREFRESH,1,0,0,0,0);
//    SendHackCommand(c,1,HACKCMD_INHIBITFULL,1,0,0,0,0);
  }
  
  orgmask=mask;
//  if(!frame_only && mouse_mode==MOUSE_NORMAL)
  if(!frame_only)
  {
    if((c->is_e_m_r & IS_MOVING)==0 || mouse_mode==MOUSE_NORMAL) {
      if(c->is_docked==0) {
        XMapWindow(dpy,c->window);	// just to make sure it's back
        if(c->cover) XMapWindow(dpy,c->cover);
      }
/* We don't want any stacking order changes on our actual client */ 
      if(mask & CWStackMode) mask=mask ^ CWStackMode; 
      if(mask & CWSibling) mask=mask ^ CWSibling;
    
//      if((mask & (CWHeight|CWWidth|CWX|CWY))!=0 && ((c->auto_properties & LIVE_RESIZING)==LIVE_RESIZING || (c->is_e_m_r & IS_RESIZING)==0)) {
      if((c->auto_properties & LIVE_RESIZING)==LIVE_RESIZING || (c->is_e_m_r & IS_RESIZING)==0) {
//      c->resize_timer=GetCarbonTickCount()+20;
//    {} else if(config_win!=0) {}
        if(config_win!=0) {
//          if(lastfocus_client==c) {
            wc->x=LW;
            wc->y=TH;
//          } else {
//            wc->x=-FW-10;
//            wc->y=-FH-10;
//          }
          wc->width=c->width;
          wc->height=c->height;
          XConfigureWindow(dpy, c->window, mask, wc);
//          XMoveResizeWindow(dpy,c->cover,FX+LW,FY+TH,c->width,c->height);
          //printf("Done a configure window on this client's window\n");
        } else if((mask & (CWHeight|CWWidth))!=0) {
//          if(lastfocus_client==c) {
            wc->x=LW;
            wc->y=TH;
//          } else {
//            wc->x=-FW-10;
//            wc->y=-FH-10;
//          }
          XMoveResizeWindow(dpy,c->window, wc->x, wc->y, c->width, c->height);
//          XMoveResizeWindow(dpy,c->cover,FX+LW,FY+TH,c->width,c->height);
        }
      }
    }
  } else if(c->is_icon!=NOT_ICONBAR ||
            (get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED==WIN_STATE_SHADED)) {
    XUnmapWindow(dpy,c->window);	// just to make sure it's gone
    if(c->cover) XMoveResizeWindow(dpy,c->cover,-6000,-6000,5,5);
    //XUnmapWindow(dpy,c->cover);
  }
//  else if(c->is_docked==0)
//    XMapWindow(dpy,c->window);	// just to make sure it's back

  mask=orgmask;

  if((mask & (CWStackMode|CWSibling))==CWStackMode) {
    check_top_client=1;
    if(wc->stack_mode==Above) {
      // This makes sure that the client is brought in front by the XDarwin application, if it wasn't
      // already. Otherwise it can still remain behind other OSX windows if XDarwin was not in front at
      // the time the window was clicked.
      XConfigureWindow(dpy, check_win, CWStackMode, wc);
      wc->stack_mode=Above;
    } else if(wc->stack_mode==Below) {
      wc->stack_mode=Below;
      // Probably no need to do this...
      XConfigureWindow(dpy, check_win, CWStackMode, wc);
      wc->stack_mode=Below;
    }
  }
  
  check_ptrwin=1;
  
  if(c->is_docked!=0) {	// make sure X server takes docked clients out of harm's way...
    wc->x=-FW-10; 
    wc->y=-FH-10;
    mask |= (CWX|CWY);
  } else {
    wc->x=FX; 
    wc->y=FY;
  }
  
/*  if((c->is_e_m_r & IS_RESIZING)!=0 && mouse_mode!=MOUSE_NORMAL && c->real_w>0 && c->real_h>0) {
    if(c->real_w>=FW && c->real_h>=FH) {
      mask = mask ^ (mask & (CWHeight|CWWidth));
      wc->width=c->real_w;
      wc->height=c->real_h;
      printf("Not changing frame: %d %d  vs  %d %d\n",c->real_w,c->real_h,FW,FH);
    } else {
      printf("Frame too small: %d %d  vs  %d %d\n",c->real_w,c->real_h,FW,FH);
      c->real_w=FW*2;
      c->real_h=FH*2;
    }
  } else {*/
    if(mask|CWWidth) c->real_w=FW;
    if(mask|CWHeight) c->real_h=FH;
//  }
  
  wc->width=c->real_w;
  wc->height=c->real_h;
  if(mask) {
    if((c->is_e_m_r & IS_MOVING)==IS_MOVING && can_send_hackcmds) {
      SendHackCommand(c, 2, HACKCMD_SETWINPOS, wc->x, wc->y, 0, 0, 0 );
    } else {
//    printf("Configuring frame: (%d,%d) [%d x %d]\n",wc->x,wc->y,wc->width,wc->height);
      if((c->is_e_m_r & IS_RESIZING)==0 && c->cover) {
        if(wc->height<=TH+BH) {
          XMoveResizeWindow(dpy,c->cover,-6000,-6000,5,5);
        } else {
          if(mask & (CWHeight|CWWidth))
            XResizeWindow(dpy,c->cover,wc->width-LW-RW,wc->height-TH-BH);
          if(mask & (CWX|CWY))
            XMoveWindow(dpy,c->cover,wc->x+LW,wc->y+TH);
        }
      }
      if(can_send_hackcmds && (mask & (CWX|CWY)) && (mask & (CWHeight|CWWidth)) && really_was_same_size==0 )
        SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,201,0,0,0,0);
      if((mask & (CWHeight|CWWidth))!=0 && (was_same_size==0 || config_win==5)) {
        SendHackCommand(c,1,HACKCMD_DOSHADOW,-198,0,0,0,0);
        if((c->is_e_m_r & IS_RESIZING)!=0 && mouse_mode!=MOUSE_NORMAL) {
          if(gVerboseConsole)
            printf("%s info: sending noshadow\n",PROGNAME);
          SendHackCommand(c,1,HACKCMD_DOSHADOW,-4,0,0,0,0);
        }
      }
      SendHackCommand(c,1,HACKCMD_DOSHADOW,-998,0,0,0,0);
      /*if(mask&(CWX|CWY)) printf("moving frame to ( %d , %d )\n",wc->x,wc->y);
      if(mask&(CWHeight|CWWidth)) printf("resizing frame to ( %d , %d )\n",wc->width,wc->height);*/
      XConfigureWindow(dpy, c->frame, mask, wc);
      if(can_send_hackcmds && (mask & (CWX|CWY)) && (mask & (CWHeight|CWWidth)) && really_was_same_size==0)
        SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,200,0,0,0,0);
//      if(mask & CWStackMode) {
//        if(c==lastfocus_client || (c->auto_properties & AUTO_FOCUS|NONAUTO_FIRSTCLICK))
//          stack_cover(c,Below);
//        else
//          stack_cover(c,Above);
//      }
      MyXSync(dpy,False);
    }
  }
  mask = orgmask;
  
  /* If size has changed */ 
  if((mask & (CWHeight|CWWidth))!=0 && was_same_size==0) 
  { 
    //SendHackCommand(c,1,HACKCMD_INHIBITREFRESH,1,0,0,0,0);
    if( c->has_frame || (c->is_icon!=NOT_ICONBAR) ) 
    { 
      /* Reposition and resize frame parts */ 

      if(c->is_icon!=NOT_ICONBAR && TW<1) {
        if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT) {
           if(c->top_left) {
              XMapWindow(dpy,c->top_left);
              XMoveResizeWindow(dpy, c->top_left, TLX, TLY, TLW, TLH);
//              XMoveResizeWindow(dpy, c->top_left, TLX, TLY, icontab_width, TLH);
              if(c->top_right) XUnmapWindow(dpy, c->top_right);
              if(c->title) XUnmapWindow(dpy, c->title); 
//              if(c->top_right) XMoveResizeWindow(dpy, c->top_right, TRX+TRW-1, TRY, 1, TRH);
//              if(c->title) XMoveResizeWindow(dpy, c->title, 1, TY, 1, TH); 
           } else if(c->top_right) {
              XMapWindow(dpy,c->top_right);
              XMoveResizeWindow(dpy, c->top_right, TRX, TRY, TRW, TRH);
//              XMoveResizeWindow(dpy, c->top_right, TRX, TRY, icontab_width, TRH);
              if(c->top_left) XUnmapWindow(dpy, c->top_left);
              if(c->title) XUnmapWindow(dpy, c->title); 
//              if(c->top_left) XMoveResizeWindow(dpy, c->top_left, TLX, TLY, 1, TLH);
//              if(c->title) XMoveResizeWindow(dpy, c->title, 1, TY, 1, TH); 
           }
        } else {
           if(c->top_right) {
              XMapWindow(dpy,c->top_right);
              XMoveResizeWindow(dpy, c->top_right, TRX, TRY, TRW, TRH);
              if(c->top_left) XUnmapWindow(dpy, c->top_left);
              if(c->title) XUnmapWindow(dpy, c->title); 
//              if(c->top_left) XMoveResizeWindow(dpy, c->top_left, TLX, TLY, 1, TLH);
//              if(c->title) XMoveResizeWindow(dpy, c->title, 1, TY, 1, TH); 
           } else if(c->top_left) {
              XMapWindow(dpy,c->top_left);
              XMoveResizeWindow(dpy, c->top_left, TLX, TLY, TLW, TLH);
              if(c->title) XUnmapWindow(dpy,c->title);
//              if(c->title) XMoveResizeWindow(dpy, c->title, 1, TY, 1, TH); 
           }
        }
      } else {
        if(c->title) XMapWindow(dpy,c->title);
        if(c->top_left) XMapWindow(dpy,c->top_left);
        if(c->top_right) XMapWindow(dpy,c->top_right);
        if(c->title) XMoveResizeWindow(dpy, c->title, TX, TY, TW<1?1:TW, config.top.height); 
        if(c->top_left) XMoveResizeWindow(dpy, c->top_left, TLX, TLY, TLW, TLH); 
        if(c->top_right) XMoveResizeWindow(dpy, c->top_right, TRX, TRY, TRW, TRH);
      }
//      XMoveResizeWindow(dpy, c->top_right, w+LW+RW-TRW, TRY, TRW, TRH);

      if( c->is_icon!=NOT_ICONBAR ) {
        XUnmapWindow(dpy,c->window);
        // window buttons need to be nowehere in sight for iconbars! (These are unmapped now...)
//        XMoveResizeWindow(dpy, c->close, -100, -100, CW, CH);
//        XMoveResizeWindow(dpy, c->icon, -100, -100, IW, IH);
//        XMoveResizeWindow(dpy, c->zoom, -100, -100, HW, HH);
      } else if( frame_only) {
        // only doing frame...
        if(c->close) XMoveResizeWindow(dpy, c->close, CX, CY, CW, CH);
        if(c->icon) XMoveResizeWindow(dpy, c->icon, IX, IY, IW, IH);
        if(c->zoom) XMoveResizeWindow(dpy, c->zoom, HX, HY, HW, HH);
      } else {
        // doing everything...
        if(c->left) XMoveResizeWindow(dpy, c->left, LX, LY, LW, LH);
        if(c->right) XMoveResizeWindow(dpy, c->right, RX, RY, RW, RH);
        if(c->bottom_left) XMoveResizeWindow(dpy, c->bottom_left, BLX, BLY, BLW, BLH);
        if(c->bottom) XMoveResizeWindow(dpy, c->bottom, BX, BY, BW, BH);
        if(c->resize) XMoveResizeWindow(dpy, c->resize, RSX, RSY, RSW, RSH);
        if(c->close) XMoveResizeWindow(dpy, c->close, CX, CY, CW, CH);
        if(c->icon) XMoveResizeWindow(dpy, c->icon, IX, IY, IW, IH);
        if(c->zoom) XMoveResizeWindow(dpy, c->zoom, HX, HY, HW, HH);
      }
//      redraw_client_frame(c);
    }
  }
  
//  c->is_icon=orig_is_icon;

  if((mask & (CWHeight|CWWidth))!=0 && was_same_size==0) {
      if(c->is_icon!=ICONBAR_RAISED && c->is_icon!=NOT_ICONBAR)
        c->is_e_m_r &= NOT_ENTERED;
      else if(can_send_hackcmds)
        c->is_e_m_r |= IS_ENTERED;
//      SendHackCommand(c,1,HACKCMD_INHIBITREFRESH,0,0,0,0,0);
//      redraw_client_frame(c);
  } else if((mask & (CWX|CWY))!=0) {
    if((c->is_e_m_r & IS_MOVING)==IS_MOVING)	// need to show new position in titlebar
      redraw_client_title(c);
    else if((c->is_e_m_r & SHOWING_POSITION)==SHOWING_POSITION)  // put back original titlebar after move
      redraw_client_title(c);
  }
  
#ifdef SHAPE_SUPPORT
  /* Select shape events */ 
//  if(shape) 
//    XShapeSelectInput(dpy, c->window, 0); 
    if((mask & (CWHeight|CWWidth))!=0 && was_same_size==0) { 
      if(can_send_hackcmds==0) {
        // make sure it's all opaque at the start (shape is inhibited using the second parameter)
        set_client_shape(c,1);
/*      } else {
        SendHackCommand(c,1,HACKCMD_INHIBITREFRESH,0,0,0,0,0);
        set_client_shape(c,0);
        if((c->is_e_m_r & IS_RESIZING)==0 || mouse_mode==MOUSE_NORMAL)
            SendHackCommand(c,1,HACKCMD_INHIBITFULL,2,0,0,0,0);
        else
            SendHackCommand(c,1,HACKCMD_INHIBITREFRESH,3,0,0,0,0);*/
      }
    }
//      set_client_shape(c,c->is_shaped);
#endif /* SHAPE */ 
//    XConfigureWindow(dpy, c->frame, mask, wc); 

//    if( c->has_frame || (c->is_icon!=NOT_ICONBAR) ) 
//      redraw_client_frame(c);

//#ifdef SHAPE_SUPPORT 
//    set_client_shape(c);
//#endif /* SHAPE */ 
//  }
//  else
//    XConfigureWindow(dpy, c->frame, mask, wc); 
  
//   MyIncreaseXServerGrab();

  /* Notify client of window changes if changes to position or size */ 
//  if(mouse_mode==MOUSE_NORMAL && (mask & (CWX|CWY|CWHeight|CWWidth))!=0 && !frame_only && (c->is_icon==NOT_ICONBAR) ) 
//  if(config_win!=2 && ((mask & (CWX|CWY))!=0 || ((mask & (CWHeight|CWWidth))!=0 && was_same_size==0)))
  if(config_win!=2 && ((mask & (CWX|CWY|CWHeight|CWWidth))!=0))
  if(((c->is_e_m_r & (IS_MOVING | IS_RESIZING))==0 || mouse_mode==MOUSE_NORMAL)
                                        && !frame_only && (c->is_icon==NOT_ICONBAR) ) 
  { 
#ifdef SHOW_EVENTS
    printf("Sending configurenotify event to window: %#lx\n",c->window);
#else
    if(gVerboseConsole)
        printf("Sending configurenotify event to window: %#lx\n",c->window);
#endif
    ce.type=ConfigureNotify;
    ce.display=dpy;
    ce.event=c->window; 
    ce.window=c->window; 
//    if(lastfocus_client==c) {
      ce.x=c->x; 
      ce.y=c->y; 
//    } else {
//      ce.x=-FW-10;
//      ce.y=-FH-10;
//    }
    ce.width=c->width; 
    ce.height=c->height;
    ce.border_width=0; 
//    ce.above=c->frame; 
    ce.above=0; 
    ce.override_redirect=0; 
    XSendEvent(dpy, c->window, False, StructureNotifyMask, (XEvent *)&ce);
  }
//#ifdef SHAPE_SUPPORT
//  } else if(c->is_shaped==0) {
  // now set the shape (it was made all opaque before, so that it looks a bit better) 
  if((mask & (CWHeight|CWWidth))!=0 && (was_same_size==0 || config_win==5)) {
//    if(c->has_frame || c->is_icon!=NOT_ICONBAR) redraw_client_frame(c);
//    if(config_win==4 && can_send_hackcmds!=0)
//        SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,-1,0,0,0,0);
#ifdef SHAPE_SUPPORT
    //if((c->is_e_m_r & (IS_MOVING | IS_RESIZING))==0)
    SendHackCommand(c,1,HACKCMD_DOSHADOW,-999,0,0,0,0);
    set_client_shape(c,0);
    SendHackCommand(c,1,HACKCMD_DOSHADOW,-998,0,0,0,0);
#endif
    if(c->has_frame || c->is_icon!=NOT_ICONBAR) redraw_client_frame(c);
    SendHackCommand(c,1,HACKCMD_DOSHADOW,-199,0,0,0,0);
#if 0
    if(can_send_hackcmds!=0) {
//      SendHackCommand(c,1,HACKCMD_INHIBITREFRESH,0,0,0,0,0);
//      XSync(dpy,False);
      if((c->is_e_m_r & IS_RESIZING)==0 || mouse_mode==MOUSE_NORMAL) {
        if(gVerboseConsole)
            printf("%s info: sending inhibitnone, updateall and redoshadow\n",PROGNAME);
        SendHackCommand(c,2,HACKCMD_INHIBITLEVEL,HACKARG_INHIBITNONE,1,0,0,0);
//        SendHackCommand(c,1,HACKCMD_UPDATEALL,1,0,0,0,0);
        SendHackCommand(c,1,HACKCMD_DOSHADOW,-3,0,0,0,0);
      } else {
        if(gVerboseConsole)
            printf("%s info: sending inhibitscreen and updateall\n",PROGNAME);
        //SendHackCommand(c,1,HACKCMD_DOSHADOW,-4,0,0,0,0);
//        SendHackCommand(c,2,HACKCMD_INHIBITLEVEL,HACKARG_INHIBITSCREEN,1,0,0,0);
        SendHackCommand(c,2,HACKCMD_INHIBITLEVEL,HACKARG_INHIBITNONE,1,0,0,0);
      }
      if((mask & (CWX|CWY))!=0)
        SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,200,0,0,0,0);
//      if((c->is_e_m_r & IS_RESIZING)==0 || mouse_mode==MOUSE_NORMAL)
//        SendHackCommand(c,1,HACKCMD_INHIBITFULL,2,0,0,0,0);
//      else
//        SendHackCommand(c,1,HACKCMD_INHIBITREFRESH,3,0,0,0,0);
    }
#endif
  }

    if(c->osxwinfo.winref!=0)
        SetOSXWinPos(&(c->osxwinfo),FX,FY,FW,FH,CX,CY,HX+HW,HY+HH);
        
//  if(c!=lastfocus_client && !frame_only)
//    XMoveResizeWindow(dpy,c->window, -c->width-10, -c->height-10, c->width, c->height);
//#endif /* SHAPE */ 
//  }

//          if(lastfocus_client!=c) {
//            wc->x=-FW-10;
//            wc->y=-FH-10;
//            XMoveResizeWindow(dpy,c->window, wc->x, wc->y, c->width, c->height);
//          }

//  if((mask & CWStackMode)==0) {
//    if(get_top_most_client(WIN_LAYER_NORMAL,True)==c) {
//      wc->stack_mode=Above;
//      XConfigureWindow(dpy, c->frame, CWStackMode, wc);
//    }
//  }
//  if((mask & (CWHeight|CWWidth))==0)
//    XResizeWindow(dpy, c->frame, wc->width, wc->height);
  
//  MyDecreaseXServerGrab();
  if(can_send_hackcmds) MyDecreaseXServerGrab();
//  printf("%s info: Finished configure_client\n",PROGNAME);
} 

/*
 * Check that a client is visible after screen change
 */
void check_client_ok(Client *c)
{
  int head=find_rect_head(c->x,c->y,c->width,c->height);
  
  if( head<0 ) head = find_head(c->x+c->width/2,c->y+c->height/2);
    
  if( c->x+c->width+RW>xinerama_screens[head].x_org+xinerama_screens[head].width )
    c->x=xinerama_screens[head].x_org+xinerama_screens[head].width-c->width-RW;
  if( c->y+c->height+BH>xinerama_screens[head].y_org+xinerama_screens[head].height )
    c->y=xinerama_screens[head].y_org+xinerama_screens[head].height-c->height-BH;
  // more important that the titlebar is onscreen, so do that last
  if( c->x-LW<xinerama_screens[head].x_org ) c->x=xinerama_screens[head].x_org+LW;
  if( c->y-TH<xinerama_screens[head].y_org ) c->y=xinerama_screens[head].y_org+TH;
  
  // Also have to check the zoom position
  if( c->old_x+c->old_width+RW>xinerama_screens[head].x_org+xinerama_screens[head].width )
    c->old_x=xinerama_screens[head].x_org+xinerama_screens[head].width-c->old_width-RW;
  if( c->old_y+c->old_height+BH>xinerama_screens[head].y_org+xinerama_screens[head].height )
    c->old_y=xinerama_screens[head].y_org+xinerama_screens[head].height-c->old_height-BH;
  // more important that the titlebar is onscreen, so do that last
  if( c->old_x-LW<xinerama_screens[head].x_org ) c->old_x=xinerama_screens[head].x_org+LW;
  if( c->old_y-TH<xinerama_screens[head].y_org ) c->old_y=xinerama_screens[head].y_org+TH;
  
  // only update the window if it's visible
  if(c->is_icon==NOT_ICONBAR && c->isoffscreen==0 && 
     c->osxwinfo.view_state==VIEWSTATE_VISIBLE && c->is_docked==0 )
    move_client(c,c->x,c->y,0);
}

/* 
 * 
 * Move client to new position 
 * 
 */ 

void move_client(Client *c, int x, int y, int redraw) 
{ 
  XWindowChanges wc;
  int was_entered;
 
  sprintf(lastsub,"TRACE2: move_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW);
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: move_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW);
#endif
 
  if(c==NULL) return;
  
  if(c->is_docked) return;
//    if(undock_client(c)) return;

  was_entered=c->is_e_m_r;
  if((was_entered & IS_MOVING)==0)
    was_entered &= NOCHANGE_SHOWFLAGS;
  
  wc.x=x; 
  wc.y=y;
  
  if(redraw==100)
    configure_client(c, CWX|CWY, &wc, c->is_icon!=NOT_ICONBAR, 100);
  else if(redraw!=0) {
    if(c->is_icon==NOT_ICONBAR) {
      wc.width=c->width;
      wc.height=c->height;
    } else {
      wc.width=c->icon_w;
      wc.height=c->icon_h;
    }
    configure_client(c, CWX|CWY|CWWidth|CWHeight, &wc, c->is_icon!=NOT_ICONBAR, 0); 
  } else
    configure_client(c, CWX|CWY, &wc, c->is_icon!=NOT_ICONBAR, 0);
    
  c->is_e_m_r = (c->is_e_m_r & CHANGE_SHOWFLAGS) | was_entered;
} 
 
/* 
 * 
 * Resize client 
 * 
 */ 
void resize_client(Client *c, int w, int h, int frame_force) 
{ 
  XWindowChanges wc;
  int mask, was_entered, frame_only, config_win;
 
  sprintf(lastsub,"TRACE2: resize_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW);
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: resize_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW);
#endif
 
  if(c==NULL) return;
  
  if(c->is_docked) return;
//    if(undock_client(c)) return;

  was_entered=c->is_e_m_r;
  if((was_entered & IS_RESIZING)==0)
    was_entered &= NOCHANGE_SHOWFLAGS;

  wc.width=w; 
  wc.height=h;
  mask = CWWidth|CWHeight|CWX|CWY;
  
  if(c->is_icon!=NOT_ICONBAR) {
    // watch out for iconbars on the right!
    if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT) {
//      mask=mask|CWX;
      wc.x=c->icon_x+c->icon_w-w;
      wc.y=c->icon_y;
    } else {
      wc.x=c->icon_x;
      wc.y=c->icon_y;
    }
  } else {
    wc.x=c->x;
    wc.y=c->y;
  }
  
  frame_only = frame_force & 1;
  if(c->is_icon!=NOT_ICONBAR)
    frame_only = 1;
    
  if((frame_force & 2)==2)
    config_win=5;
  else
    config_win=0;

  configure_client(c, mask, &wc, frame_only, config_win);
//  XMoveResizeWindow(dpy,c->frame,wc.x,wc.y,wc.width,wc.height);

  c->is_e_m_r = (c->is_e_m_r & CHANGE_SHOWFLAGS) | was_entered;
} 
 
/* 
 * 
 * Move & Resize client 
 * 
 */ 
void move_resize_client(Client *c, int x, int y, int w, int h, int frame_only) 
{ 
  XWindowChanges wc;
  int mask, was_entered;
 
  sprintf(lastsub,"TRACE2: move_resize_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW);
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: move_resize_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW);
#endif
 
  if(c==NULL) return;
  
  if(c->is_docked) return;
//    if(undock_client(c)) return;
    
  was_entered=c->is_e_m_r;

  wc.x=x; 
  wc.y=y;
  wc.width=w; 
  wc.height=h;
  mask = CWX|CWY|CWWidth|CWHeight;
/*  
  if(c->is_icon!=NOT_ICONBAR) {
    // watch out for iconbars on the right!
    if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT) {
      wc.x=c->icon_x+c->icon_w-w;
      wc.y=c->icon_y ;
    } else {
      wc.x=c->icon_x;
      wc.y=c->icon_y;
    }
  } else {
    wc.x=c->x;
    wc.y=c->y;
  }
*/  
  configure_client(c, mask, &wc, frame_only || (c->is_icon!=NOT_ICONBAR), 0);
//  XMoveResizeWindow(dpy,c->frame,wc.x,wc.y,wc.width,wc.height);

  c->is_e_m_r=was_entered;
} 
 
/* 
 * 
 * Toggle client maximized 
 * 
 */ 

void toggle_client_maximized(Client *c, int which) 
{ 
  XWindowChanges wc;
  int org_hint; 
 
  if(c==NULL) {
    printf("%s warning: got into toggle_client_maximized with NULL client\n",PROGNAME);
    return;
  }

  if( c->is_icon!=NOT_ICONBAR )
    return;

  sprintf(lastsub,"TRACE1: toggle_client_maximized (win,can_max,which: %#lx,%d,%d)\n",C_WINDOW,c->can_max,which); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: toggle_client_maximized (win,can_max,which: %#lx,%d,%d)\n",C_WINDOW,c->can_max,which); 
#endif 
 
  if(c->can_max) 
  { 
    unsigned int mask=0;
    org_hint = get_gnome_hint(c->window, WIN_STATE);
    if((org_hint & WIN_STATE_MAXIMIZED)==WIN_STATE_MAXIMIZED) 
    { 
      wc.x=c->old_x; 
      wc.y=c->old_y; 
      wc.width=c->old_width; 
      wc.height=c->old_height; 
      if(wc.x!=c->x) mask |= CWX;
      if(wc.y!=c->y) mask |= CWY;
      if(wc.width!=c->width) mask |= CWWidth;
      if(wc.height!=c->height) mask |= CWHeight;
    } 
    else 
    { 
      int head=0; 
 
#ifdef XINERAMA_SUPPORT 
      if(xinerama)
        head=find_head(c->x+c->width/2, c->y+c->height/2); 
#endif /* XINERAMA */ 
 
      c->old_x=c->x; 
      c->old_y=c->y; 
      c->old_width=c->width; 
      c->old_height=c->height; 
      Get_Best_Window_Zoom(&c->x,&c->y,&c->width,&c->height,which,
            LW+config.left_margin,RW+config.right_margin,TH+config.top_margin,BH+config.bottom_margin);
/*      c->x=screen_x(head)+LW+config.left_margin; 
      c->y=screen_y(head)+TH+config.top_margin; 
      c->width=screen_width(head)-LW-RW-config.left_margin-config.right_margin; 
      c->height=screen_height(head)-TH-BH-config.top_margin-config.bottom_margin;
      if(which==1) {
        c->x = c->old_x;
        c->width = c->old_width;
      } else if (which==2) {
        c->y = c->old_y;
        c->height = c->old_height;
      }*/
      if(c->size->flags & PMaxSize) {
        if(c->size->max_width>0 && c->size->max_width<c->width) {
          if( c->old_x+c->size->max_width > c->x+c->width )
            c->x = c->x+c->width-c->size->max_width;
          else
            c->x = c->old_x;
          c->width = c->size->max_width;
        }
        if(c->size->max_height>0 && c->size->max_height<c->height) {
          if( c->old_y+c->size->max_height > c->y+c->height )
            c->y = c->y+c->height-c->size->max_height;
          else
            c->y = c->old_y;
          c->height = c->size->max_height;
        }
      }
      wc.x=c->x;
      wc.y=c->y;
      wc.width=c->width;
      wc.height=c->height; 
      if(wc.x!=c->old_x) mask |= CWX;
      if(wc.y!=c->old_y) mask |= CWY;
      if(wc.width!=c->old_width) mask |= CWWidth;
      if(wc.height!=c->old_height) mask |= CWHeight;
    } 
    if((org_hint & WIN_STATE_MAXIMIZED)==WIN_STATE_MAXIMIZED) 
      set_gnome_hint(c->window, WIN_STATE, org_hint^WIN_STATE_MAXIMIZED);
    configure_client(c, mask, &wc, c->is_icon!=NOT_ICONBAR, 3); 
    if((org_hint & WIN_STATE_MAXIMIZED)==0) 
      set_gnome_hint(c->window, WIN_STATE, org_hint^WIN_STATE_MAXIMIZED); 
  } 
} 
 
/* 
 *  
 * Toggle client shaded 
 * 
 */ 

void toggle_client_shaded(Client *c) 
{ 
  int head;
  XWindowChanges wc;
  int mask;

  if(c->is_icon!=NOT_ICONBAR)
    return;

  sprintf(lastsub,"TRACE1: toggle_client_shaded (win,last: %#lx %#lx)\n",C_WINDOW,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: toggle_client_shaded (win,last: %#lx %#lx)\n",C_WINDOW,FC_WINDOW); 
#endif 
 
//  if(c->is_icon==NOT_ICONBAR) 
//  { 
    set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_SHADED); 

    wc.width=c->width; 
    wc.height=c->height;
    mask = CWWidth|CWHeight;
  
    if(c->is_icon!=NOT_ICONBAR) {
      // watch out for iconbars on the right!
      if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT) {
        mask=mask|CWX;
        wc.x=c->icon_x+c->icon_w-c->width;
      }
    }
    
    if((get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)==WIN_STATE_SHADED) {
      if(c->left) XUnmapWindow(dpy,c->left);
      if(c->right) XUnmapWindow(dpy,c->right);
      if(c->bottom) XUnmapWindow(dpy,c->bottom);
      if(c->bottom_left) XUnmapWindow(dpy,c->bottom_left);
      if(c->resize) XUnmapWindow(dpy,c->resize);
      if(c->title) XMapWindow(dpy,c->title);
      if(c->top_right) XMapWindow(dpy,c->top_right);
      if(c->top_left) XMapWindow(dpy,c->top_left);
    } else if(c->has_frame) {
      if(c->left) XMapWindow(dpy,c->left);
      if(c->right) XMapWindow(dpy,c->right);
      if(c->bottom) XMapWindow(dpy,c->bottom);
      if(c->bottom_left) XMapWindow(dpy,c->bottom_left);
      if(c->resize) XMapWindow(dpy,c->resize);
    } else {
      if(c->title) XUnmapWindow(dpy,c->title);
      if(c->top_right) XUnmapWindow(dpy,c->top_right);
      if(c->top_left) XUnmapWindow(dpy,c->top_left);
    }
  
    configure_client(c, mask, &wc, (get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)==WIN_STATE_SHADED, 4);
    if(c->has_frame!=0 || c->is_icon!=NOT_ICONBAR) {
//#ifdef SHAPE_SUPPORT
//        set_client_shape(c,0);
//#endif
        redraw_client_frame(c);
    }
//    resize_client(c, c->width, c->height, True);

// Don't want the titlebar to be off the screen!
    if( xinerama ) {
       head=find_head(c->x,c->y);
       if(c->y-TH<screen_y(head))
         move_client(c, c->x, screen_y(head)+TH, 0);
    } else {
       if(c->y-TH<0)
         move_client(c, c->x, TH, 0);
    }
    
//    redraw_client_frame(c); 
//  }

}

void toggle_minimise_client(Client *c, int which)
{
    if((gMinMethod==minmethod_iconbars && which==0) || (gMinMethod==minmethod_dock && which!=0)) {
        minmax_client(c);
    } else if((gMinMethod==minmethod_dock && which==0) || (gMinMethod==minmethod_iconbars && which!=0)) {
        if(c->is_docked==0)
            dock_client(c);
        else
            will_undock_client(c);
    }
}

#ifdef LAYER_SUPPORT 
/* 
 * 
 * Set client layer 
 * 
 */ 

void set_client_layer(Client *c, int layer) 
{ 
  sprintf(lastsub,"TRACE2: set_client_layer (win,layer: %#lx,%d)\n",C_WINDOW,layer); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: set_client_layer (win,layer: %#lx,%d)\n",C_WINDOW,layer); 
#endif 
 
  set_gnome_hint(c->window, WIN_LAYER, layer); 
  if(get_wm_state(c->window)==NormalState) 
    if(c->is_icon==NOT_ICONBAR)
      raise_client(c); 
} 
 
/* 
 * 
 * Get client layer 
 * 
 */ 

int get_client_layer(Client *c) 
{ 
  sprintf(lastsub,"TRACE2: get client layer (win,last: %#lx)\n",C_WINDOW,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: get client layer (win,last: %#lx)\n",C_WINDOW,FC_WINDOW); 
#endif 
 
  return get_gnome_hint(c->window, WIN_LAYER); 
} 
#endif /* LAYER */ 

/*
 * parent_client_gone - used when a parent client is deleted: removes references from children
 */
void parent_client_gone(Client *c)
{
    Client *nc, *pc=NULL;
    
    if(c!=NULL)
        pc = c->prev_sibling;
        
    while(c!=NULL) {
        c->parent = NULL;
        c->prev_sibling = NULL;
        nc = c->next_sibling;
        c->next_sibling = NULL;
        c = nc;
    }
}

/* 
 * 
 * Delete client 
 * 
 */ 

void delete_client(Client *c, int remap) 
{ 
  sprintf(lastsub,"TRACE1: delete_client (win,last,remap: %#lx,%#lx,%d)\n",C_WINDOW,FC_WINDOW,remap);
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: delete_client (win,last,remap: %#lx,%#lx,%d)\n",C_WINDOW,FC_WINDOW,remap);
#else
  if(gVerboseConsole)
    printf("%s info: delete_client (win,last,remap: %#lx,%#lx,%d)\n",PROGNAME,C_WINDOW,FC_WINDOW,remap);
#endif

#ifdef SHOW_WEIRD
  if(c==NULL)
    printf("%s oddity: got NULL client in delete_client!\n",PROGNAME);
#endif
  if(c==NULL) return;

  MyIncreaseXServerGrab();
  
// make sure it has the right state...
  if((c->is_icon!=NOT_ICONBAR || c->is_docked) && remap==1) {
    if((get_gnome_hint(c->window,WIN_STATE) | WIN_STATE_SHADED)==WIN_STATE_SHADED)
      set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_SHADED); 
    c->ignore_unmap++; 
    XUnmapWindow(dpy, c->frame); 
    XUnmapWindow(dpy, c->window); 
    set_wm_state(c->window, IconicState);
    printf("Set window (%#lx = \"%s\") to IconicState\n",c->window,c->name);
  }
  
  if(remap!=-1) {
    // The get_window_parent produces a BadWindow error, so inhibit it
    inhibit_errors=True;
    /* Check if frame is still the clients parent */
    if(get_window_parent(c->window)==c->frame) 
    { 
      inhibit_errors=False;
      /* Reparent our client back to root if it still exists */ 
      gravitate_client(c, REMOVE_GRAVITY); 
      if(launched_xdarwin!=0) // don't bother remapping if we launched XDarwin
        XUnmapWindow(dpy, c->window); 
      XReparentWindow(dpy, c->window, root, c->x, c->y);
      if(c->is_icon!=NOT_ICONBAR || c->is_docked) {
        set_wm_state(c->window, IconicState);
        printf("Set window (%#lx = \"%s\") to IconicState\n",c->window,c->name);
      } else {
        set_wm_state(c->window, WithdrawnState); 
        printf("Set window (%#lx = \"%s\") to WithdrawnState\n",c->window,c->name);
      }

#ifdef KEYBOARD_SUPPORT
      /* Remove any key grabs on this window */
      ungrab_keys(c->window);
#endif
 
      /* Remap is 1 if we are quitting OroborOSX (and -1 if window has vanished -but not here) */ 
      if(remap==0 || launched_xdarwin!=0) // don't bother remapping if we launched XDarwin, since it will quit
        XUnmapWindow(dpy, c->window); 
      else
        XMapWindow(dpy, c->window); 
    } else
      inhibit_errors=False;
  }
  
#ifdef DEBUG
  printf("  checking lastenter: [%#lx], [%#lx]\n",c->frame,lastenter);
  if( c->frame==lastenter ) printf("  window was last to be entered: [%#lx]\n",lastenter);
  printf("  checking lastc: [%#lx], [%#lx]\n",c,lastc);
  if( c==lastc ) printf("  window was lastc: [%#lx]\n",lastc);
  if(focus_before_hide==c) printf("  window was focus_before_hide: [%#lx]\n",focus_before_hide);
  if(lastfocus_client==c) printf("  window was lastfocus_client: [%#lx]\n",lastfocus_client);
#endif
  if( c->frame==lastenter ) lastenter=0;
  if( c==last_client_entered_hack ) last_client_entered_hack=NULL;
  if( c==lastc ) lastc=NULL;
  if(lastfocus_client==c) lastfocus_client=NULL;
  // special flag to check if the focus window has vanished while hidden (OK, bad coding, but it works...)
  if(focus_before_hide==c) focus_before_hide=-1;

  // make sure the group's window is also removed if necessary
  check_delete_group(c);
  
  if(c->osxwinfo.winref!=0)
    DestroyOSXWindow(&(c->osxwinfo));

  if(c->first_child)
    parent_client_gone(c->first_child);
  if(c->parent)
    remove_sibling_client(c);
    
  if(c->icon_y==nexticon_y-config.top.height)
    nexticon_y-=config.top.height;
	 
  /* Get rid of our frame window */ 
  XDestroyWindow(dpy, c->frame);
  if(c->cover) XDestroyWindow(dpy, c->cover);
 
  /* Remove reference from our client list */ 
  if(client_count==1) 
    client_list=NULL; 
  else 
  { 
    c->next->prev = c->prev; 
    c->prev->next = c->next; 
    if(c==client_list) 
      client_list=client_list->next; 
  } 
 
  if(c->name) XFree(c->name);
  if(c->clientcmd) XFree(c->clientcmd);
  if(c->clienthost) XFree(c->clienthost);
  if(c->clientclass) XFree(c->clientclass);
  if(c->clientname) XFree(c->clientname);
  if(c->clientgrouporg)
    if(c->clientgrouporg!=c->clientgroup)
      XFree(c->clientgrouporg);
  if(c->clientgroup) XFree(c->clientgroup); 
  if(c->size) XFree(c->size); 
  free(c); 
 
  client_count--; 
  update_gnome_client_list();

  MyDecreaseXServerGrab();

  sprintf(lastsub,"TRACE2: delete_client -client removed\n");
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: delete_client -client removed\n");
#endif
} 
 
/* 
 * 
 * Send close message to client window 
 * 
 */ 

void close_client(Client *c) 
{ 
  int i, n, found=False; 
  Atom *protocols=NULL; 
  XEvent e; 
 
  sprintf(lastsub,"TRACE1: close_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: close_client (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW); 
#endif 
 
#ifdef SHOW_WEIRD
  if(c==NULL) printf("%s oddity: got NULL client in close_client!\n",PROGNAME);
#endif
  if(c==NULL) return;
  
  if(XGetWMProtocols(dpy, c->window, &protocols, &n)) 
  { 
    for(i=0; i<n; i++) 
      if(protocols[i]==icccm[WM_DELETE]) found=True; 
    if(protocols) XFree(protocols);  
  }
  
  if(found) 
  { 
#ifdef SHOW_EVENTS
    printf("  sending delete event to window %#lx = \"%s\"\n",c->window,c->name);
#endif
    e.type=ClientMessage; 
    e.xclient.window=c->window; 
    e.xclient.message_type=icccm[WM_PROTOCOLS]; 
    e.xclient.format=32; 
    e.xclient.data.l[0]=icccm[WM_DELETE]; 
    e.xclient.data.l[1]=CurrentTime; 
    XSendEvent(dpy, c->window, False, NoEventMask, &e); 
  } 
  else  {
#ifdef DEBUG
    printf("  killing window %#lx = \"%s\"\n",c->window,c->name);
#endif
    XKillClient(dpy, c->window); 
  }
} 
 
#ifdef SHAPE_SUPPORT 
/* 
 * 
 * Set client window shape 
 * 
 */ 

void set_client_shape(Client *c, int inhibit) 
{ 
  Window temp,wintmp;
  int i, doneset=0;	// doneset makes sure that ShapeSet is applied the first time, then ShapeUnion
 
  sprintf(lastsub,"TRACE3: set_client_shape (win,last,inhibit: %#lx,%#lx,%d)\n",C_WINDOW,FC_WINDOW,inhibit);
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: set_client_shape (win,last,inhibit: %#lx,%#lx,%d)\n",C_WINDOW,FC_WINDOW,inhibit);
#else
  if(gVerboseConsole)
    printf("%s info: set_client_shape (win,last,inhibit: %#lx,%#lx,%d)\n",PROGNAME,C_WINDOW,FC_WINDOW,inhibit);
#endif

#ifdef SHOW_WEIRD
  if(c==NULL) printf("%s oddity: got NULL client in set_client_shape!\n",PROGNAME);
#endif
  if(c==NULL) return;

  if(c->is_docked) return;
//    if(undock_client(c)) return;
 
  if(shape) {
  
    MyIncreaseXServerGrab();

    if(inhibit==0 || inhibit==1)
    {

      temp=XCreateSimpleWindow(dpy, root, 0, 0, FW, FH, 0, 0, 0);
      doneset=ShapeSet;

      if(c->has_frame==0 && c->is_icon==NOT_ICONBAR)
      {
//        if(inhibit==0 && c->is_shaped) {
        if(inhibit==0) {
          XShapeCombineShape(dpy, temp, ShapeBounding, LW, TH, c->window, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
      }
      else if( c->is_icon!=NOT_ICONBAR)
      {
        // Just shape titlebar parts for an iconbar
        if(TW<1) { // this probably implies it's only an iconbar tab...
          if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT) {
            if(c->top_left) { // try to apply top-left shape
              XShapeCombineMask(dpy, c->top_left, ShapeBounding, 0, 0, config.top_left.mask[2], ShapeSet);
              XShapeCombineShape(dpy, temp, ShapeBounding, TLX, TLY, c->top_left, ShapeBounding, doneset);
              doneset=ShapeUnion;
            } else if(c->top_right) { // failed -try to use top-right
              XShapeCombineMask(dpy, c->top_right, ShapeBounding, 0, 0, config.top_right.mask[2], ShapeSet);
              XShapeCombineShape(dpy, temp, ShapeBounding, TRX, TRY, c->top_right, ShapeBounding, doneset);
              doneset=ShapeUnion;
            }
          } else {
            if(c->top_right) { // try to apply top-right shape
              XShapeCombineMask(dpy, c->top_right, ShapeBounding, 0, 0, config.top_right.mask[2], ShapeSet);
              XShapeCombineShape(dpy, temp, ShapeBounding, TRX, TRY, c->top_right, ShapeBounding, doneset);
              doneset=ShapeUnion;
            } else if(c->top_left) { // failed -try to use top-left
              XShapeCombineMask(dpy, c->top_left, ShapeBounding, 0, 0, config.top_left.mask[2], ShapeSet);
              XShapeCombineShape(dpy, temp, ShapeBounding, TLX, TLY, c->top_left, ShapeBounding, doneset);
              doneset=ShapeUnion;
            }
          }
        } else {
          if(c->top_left) { // try to apply top-left shape
            XShapeCombineMask(dpy, c->top_left, ShapeBounding, 0, 0, config.top_left.mask[2], ShapeSet);
            XShapeCombineShape(dpy, temp, ShapeBounding, TLX, TLY, c->top_left, ShapeBounding, doneset);
            doneset=ShapeUnion;
            if(c->top_right) { // and top-right shape
              XShapeCombineMask(dpy, c->top_right, ShapeBounding, 0, 0, config.top_right.mask[2], ShapeSet);
              XShapeCombineShape(dpy, temp, ShapeBounding, TRX, TRY, c->top_right, ShapeBounding, doneset);
              doneset=ShapeUnion;
            }
          } else if(c->top_right) { // no top-left, so just try to apply top-right shape
            XShapeCombineMask(dpy, c->top_right, ShapeBounding, 0, 0, config.top_right.mask[2], ShapeSet);
            XShapeCombineShape(dpy, temp, ShapeBounding, TRX, TRY, c->top_right, ShapeBounding, doneset);
            doneset=ShapeUnion; 
          }
          if(c->title && config.top.width>0 && config.top.height>0) { // try to apply title shape
            XShapeCombineMask(dpy, c->title, ShapeBounding, 0, 0, config.top.mask[2], ShapeSet);
            for(i=config.top.width;i<TRX-TX;i+=config.top.width)
              XShapeCombineMask(dpy, c->title, ShapeBounding, i, 0, config.top.mask[2], ShapeUnion);
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, TRX-TX, config.top.height, 0, 0, 0);
            XShapeCombineShape(dpy, c->title, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
            XDestroyWindow(dpy, wintmp);
            XShapeCombineShape(dpy, temp, ShapeBounding, TX, TY, c->title, ShapeBounding, doneset);
            doneset=ShapeUnion;
          }
        }
      }
      else if( (get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)==WIN_STATE_SHADED )
      { // window is shaded, so only shape titlebar, starting with the actual subwindows
        XShapeCombineMask(dpy, c->zoom, ShapeBounding, 0, 0, config.zoom.mask[0], ShapeSet);
        XShapeCombineMask(dpy, c->icon, ShapeBounding, 0, 0, config.icon.mask[0], ShapeSet);
        XShapeCombineMask(dpy, c->close, ShapeBounding, 0, 0, config.close.mask[0], ShapeSet);
        if(c->top_left) XShapeCombineMask(dpy, c->top_left, ShapeBounding, 0, 0, config.top_left.mask[1], ShapeSet);
        if(c->top_right) XShapeCombineMask(dpy, c->top_right, ShapeBounding, 0, 0, config.top_right.mask[1], ShapeSet);
        if(config.top.width>0 && config.top.height>0 && TRX>TX) {
          XShapeCombineMask(dpy, c->title, ShapeBounding, 0, 0, config.top.mask[1], ShapeSet);
          for(i=config.top.width;i<TRX-TX;i+=config.top.width)
            XShapeCombineMask(dpy, c->title, ShapeBounding, i, 0, config.top.mask[1], ShapeUnion);
          wintmp=XCreateSimpleWindow(dpy, root, 0, 0, TRX-TX, config.top.height, 0, 0, 0);
          XShapeCombineShape(dpy, c->title, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
          XDestroyWindow(dpy, wintmp);
        }
        // now try to apply subwindow shapes to the entire window
        if(c->top_left) {
          XShapeCombineShape(dpy, temp, ShapeBounding, TLX, TLY, c->top_left, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
        if(c->top_right) {
          XShapeCombineShape(dpy, temp, ShapeBounding, TRX, TRY, c->top_right, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
        if(c->close) {
          XShapeCombineShape(dpy, temp, ShapeBounding, CX, CY, c->close, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
        if(c->icon) {
          XShapeCombineShape(dpy, temp, ShapeBounding, IX, IY, c->icon, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
        if(c->zoom) {
          XShapeCombineShape(dpy, temp, ShapeBounding, HX, HY, c->zoom, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
        if(c->title) {
          XShapeCombineShape(dpy, temp, ShapeBounding, TX, TY, c->title, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
      }
      else if(inhibit==0 || inhibit==1)
      { // The entire window is visible (i.e. not shaded nor an iconbar)
        if(inhibit==0 && c->is_shaped) { // Make sure the client window's shape is applied
          if(gVerboseConsole)
            printf("------- Using window's shape for reshaping...\n");
          XShapeCombineShape(dpy, temp, ShapeBounding, LW, TH, c->window, ShapeBounding, doneset);
          doneset=ShapeUnion;
        } else { // don't apply the client window's shape, instead fill in the whole rect
          if(gVerboseConsole)
            printf("------- Using rectangular shape for reshaping...\n");
          wintmp=XCreateSimpleWindow(dpy, root, 0, 0, c->width, c->height, 0, 0, 0);
          XShapeCombineShape(dpy, temp, ShapeBounding, LW, TH, wintmp, ShapeBounding, doneset);
          XDestroyWindow(dpy, wintmp);
          doneset=ShapeUnion;
        }
        // Shape the corner subwindows in case there are 'internal' shapes (eg, triangular resize widget)
        if(c->bottom_left) XShapeCombineMask(dpy, c->bottom_left, ShapeBounding, 0, 0, config.bottom_left.mask[0], ShapeSet);
        if(c->resize) XShapeCombineMask(dpy, c->resize, ShapeBounding, 0, 0, config.resize.mask[0], ShapeSet);
        if(c->top_left) XShapeCombineMask(dpy, c->top_left, ShapeBounding, 0, 0, config.top_left.mask[0], ShapeSet);
        if(c->top_right) XShapeCombineMask(dpy, c->top_right, ShapeBounding, 0, 0, config.top_right.mask[0], ShapeSet);
        if(c->bottom_left) XShapeCombineMask(dpy, c->bottom_left, ShapeBounding, 0, 0, config.bottom_left.mask[0], ShapeSet);
        // Shape the contents of the widget subwindows
        XShapeCombineMask(dpy, c->zoom, ShapeBounding, 0, 0, config.zoom.mask[0], ShapeSet);
        XShapeCombineMask(dpy, c->icon, ShapeBounding, 0, 0, config.icon.mask[0], ShapeSet);
        XShapeCombineMask(dpy, c->close, ShapeBounding, 0, 0, config.close.mask[0], ShapeSet);
        // Only shape the frame if we're using a shaped theme...
        if(gUseShapedTheme!=0) {
          // Shape the left edge frame subwindow
          if(config.left.width>0 && config.left.height>0 && BLY>LY && c->left) {
            XShapeCombineMask(dpy, c->left, ShapeBounding, 0, 0, config.left.mask[0], ShapeSet);
            for(i=config.left.height;i<BLY-LY;i+=config.left.height)
              XShapeCombineMask(dpy, c->left, ShapeBounding, 0, i, config.left.mask[0], ShapeUnion);
            // make sure the above shape does not extend beyond the left edge (i.e. not into bottom-left)
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, LW, BLY-LY, 0, 0, 0);
            XShapeCombineShape(dpy, c->left, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
            XDestroyWindow(dpy, wintmp);
          }
          // Shape the right edge frame window
          if(config.right.width>0 && config.right.height>0 && RSY>RY && c->right) {
            XShapeCombineMask(dpy, c->right, ShapeBounding, 0, 0, config.right.mask[0], ShapeSet);
            for(i=config.right.height;i<RSY-RY;i+=config.right.height)
              XShapeCombineMask(dpy, c->right, ShapeBounding, 0, i, config.right.mask[0], ShapeUnion);
            // make sure the above shape does not extend beyond the right edge (i.e. not into resize)
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, RW, RSY-RY, 0, 0, 0);
            XShapeCombineShape(dpy, c->right, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
            XDestroyWindow(dpy, wintmp);
          }
          // Shape the bottom edge frame window
          if(config.bottom.width>0 && config.bottom.height>0 && RSX>BX && c->bottom) {
            XShapeCombineMask(dpy, c->bottom, ShapeBounding, 0, 0, config.bottom.mask[0], ShapeSet);
            for(i=config.bottom.width;i<RSX-BX;i+=config.bottom.width)
              XShapeCombineMask(dpy, c->bottom, ShapeBounding, i, 0, config.bottom.mask[0], ShapeUnion);
            // make sure the above shape does not extend beyond the bottom edge (i.e. not into resize)
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, RSX-BX, config.bottom.height, 0, 0, 0);
            XShapeCombineShape(dpy, c->bottom, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
            XDestroyWindow(dpy, wintmp);
          }
          // Shape the title frame window
          if(config.top.width>0 && config.top.height>0 && TRX>TX) {
            XShapeCombineMask(dpy, c->title, ShapeBounding, 0, 0, config.top.mask[0], ShapeSet);
            for(i=config.top.width;i<TRX-TX;i+=config.top.width)
              XShapeCombineMask(dpy, c->title, ShapeBounding, i, 0, config.top.mask[0], ShapeUnion);
            // make sure the above shape does not extend beyond the title (i.e. not into top-right)
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, TRX-TX, config.top.height, 0, 0, 0);
            XShapeCombineShape(dpy, c->title, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
            XDestroyWindow(dpy, wintmp);
          }
          // try to apply the widget shapes to the actual window
          if(c->close) {
            XShapeCombineShape(dpy, temp, ShapeBounding, CX, CY, c->close, ShapeBounding, doneset);
            doneset=ShapeUnion;
          }
          if(c->icon) {
            XShapeCombineShape(dpy, temp, ShapeBounding, IX, IY, c->icon, ShapeBounding, doneset);
            doneset=ShapeUnion;
          }
          if(c->zoom) {
            XShapeCombineShape(dpy, temp, ShapeBounding, HX, HY, c->zoom, ShapeBounding, doneset);
            doneset=ShapeUnion;
          }
        }
        else
        {
          if(config.left.width>0 && config.left.height>0 && BLY>LY && c->left) {
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, LW, BLY-LY, 0, 0, 0);
            XShapeCombineShape(dpy, c->left, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeSet); 
            XDestroyWindow(dpy, wintmp);
          }
          if(config.right.width>0 && config.right.height>0 && RSY>RY && c->right) {
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, RW, RSY-RY, 0, 0, 0);
            XShapeCombineShape(dpy, c->right, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeSet); 
            XDestroyWindow(dpy, wintmp);
          }
          if(config.bottom.width>0 && config.bottom.height>0 && RSX>BX && c->bottom) {
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, RSX-BX, config.bottom.height, 0, 0, 0);
            XShapeCombineShape(dpy, c->bottom, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeSet); 
            XDestroyWindow(dpy, wintmp);
          }
          if(config.top.width>0 && config.top.height>0 && TRX>TX) {
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, TRX-TX, config.top.height, 0, 0, 0);
            XShapeCombineShape(dpy, c->title, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeSet); 
            XDestroyWindow(dpy, wintmp);
          }
        }
        // Now apply all of the shaped frame subwindows to the actual window
        if(c->title) {
          XShapeCombineShape(dpy, temp, ShapeBounding, TX, TY, c->title, ShapeBounding, doneset);  
          doneset=ShapeUnion;
        }
        if(c->left) {
          XShapeCombineShape(dpy, temp, ShapeBounding, LX, LY, c->left, ShapeBounding, doneset); 
          doneset=ShapeUnion;
        }
        if(c->right) {
          XShapeCombineShape(dpy, temp, ShapeBounding, RX, RY, c->right, ShapeBounding, doneset); 
          doneset=ShapeUnion;
        }
        if(c->bottom) {
          XShapeCombineShape(dpy, temp, ShapeBounding, BX, BY, c->bottom, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
        if(c->top_left) {
          XShapeCombineShape(dpy, temp, ShapeBounding, TLX, TLY, c->top_left, ShapeBounding, doneset);    
          doneset=ShapeUnion;
        }
        if(c->top_right) {
          XShapeCombineShape(dpy, temp, ShapeBounding, TRX, TRY, c->top_right, ShapeBounding, doneset);    
          doneset=ShapeUnion;
        }
        if(c->resize) {
          XShapeCombineShape(dpy, temp, ShapeBounding, RSX, RSY, c->resize, ShapeBounding, doneset);    
          doneset=ShapeUnion;
        }
        if(c->bottom_left) {
          XShapeCombineShape(dpy, temp, ShapeBounding, BLX, BLY, c->bottom_left, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
      } 
    }
    else
      temp=XCreateSimpleWindow(dpy, root, 0, 0, XDisplayWidth(dpy, screen), XDisplayHeight(dpy, screen), 0, 0, 0);
//      temp=XCreateSimpleWindow(dpy, root, 0, 0, FW, FH, 0, 0, 0);

    XShapeCombineShape(dpy, c->frame, ShapeBounding, 0, 0, temp, ShapeBounding, ShapeSet); 
//    XShapeCombineMask(dpy, c->frame, ShapeBounding, 0, 0, temp, ShapeSet); 
    XDestroyWindow(dpy, temp);

    MyDecreaseXServerGrab();
  }
  if(gVerboseConsole)
    printf("%s info: Done set_client_shape\n",PROGNAME);
} 
#endif /* SHAPE */ 
 
#ifdef WORKSPACE_SUPPORT 
/* 
 * 
 * Toggle sticky/unsticky 
 * 
 */ 

void toggle_client_sticky(Client *c) 
{ 
  sprintf(lastsub,"TRACE2: toggle_client_sticky (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: toggle_client_sticky (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW); 
#endif 
 
  set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_STICKY); 
} 
 
/* 
 * 
 * Move client to new workspace 
 * 
 */ 

void move_client_to_workspace(Client *c, int ws) 
{ 
  sprintf(lastsub,"TRACE2: move_client_to_workspace (win,ws: %#lx,%d)\n",C_WINDOW,ws); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: move_client_to_workspace (win,ws: %#lx,%d)\n",C_WINDOW,ws); 
#endif 

#ifdef SHOW_WEIRD
  if(ws==workspace || ws<0 || ws>=config.workspace_count)
    printf("%s oddity: got invalid workspace in move_client_to_workspace: %d\n",PROGNAME,ws);
#endif

  if(ws==workspace || ws<0 || ws>=config.workspace_count) return; 
  set_gnome_hint(c->window, WIN_WORKSPACE, ws); 
  change_to_workspace(ws); 
  if(c->is_icon==NOT_ICONBAR) {
    set_focus_client(c); 
    raise_client(c); }
} 
#endif /* WORKSPACE */ 

/*
 * manipulate_focus_client - perform various changes related to the currect focus client
 * (many are property changes from the Property submenu, in which case the action>100;
 * while others are window/group cycling)
 */
void manipulate_focus_client(int action)
{
  Client *c, *tc;
  XWindowChanges wc;
  int need_ungrab=0;
    
  sprintf(lastsub,"TRACE1: manipulate_focus_client %#lx (action: %d)\n",FC_WINDOW,action); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: manipulate_focus_client %#lx (action: %d)\n",FC_WINDOW,action); 
#endif 

//  c=get_focus_client();
  c=lastfocus_client;
  if(c==NULL) {
#ifdef SHOW_WEIRD
    printf("%s oddity: Got NULL on Window menu: %d\n",PROGNAME,action);
#endif
    return; }
  
  switch(action)
  {
    case itemWindowshade:
      need_ungrab = check_need_ungrab();
      toggle_client_shaded(c);
      if(need_ungrab) force_grab_release();
      break;
      
    case itemMinimise:
      if(c->is_icon==NOT_ICONBAR) {
        need_ungrab = check_need_ungrab();
        toggle_minimise_client(c,0);
        if(need_ungrab) force_grab_release();
#ifdef SHOW_WEIRD
      } else {
        printf("%s oddity: Tried to minimise an iconbar! (%#lx)\n",PROGNAME,C_WINDOW);
#endif
      }
      break;
      
    case itemZoom:
      need_ungrab = check_need_ungrab();
      toggle_client_maximized(c,0);
      if(need_ungrab) force_grab_release();
      break;
      
    case itemClose:
      need_ungrab = check_need_ungrab();
      if(need_ungrab) force_grab_release();
      close_client(c);
      break;
      
    case itemForceClose:
      {
        Window win=c->window;
        // first, delete our client (also unmaps c->window, and removes c from client list, so
        // we MUST NOT REFERENCE C AGAIN after this delete_client!)
        delete_client(c,0);
        // Finally, do something really nasty to the actual client's window...
        XDestroyWindow(dpy, win);
      }
      break;
      
    case itemLower:
      tc=get_bottom_most_client(WIN_LAYER_NORMAL,False);
      if(tc!=c) {
        need_ungrab = check_need_ungrab();
        if(get_group_for_client(c)==0)
          tc = NULL;
        else
          tc = get_top_not_this_in_group(c,WIN_LAYER_NORMAL,0,get_group_for_client(c));
        if(tc==NULL) {
          lower_client(c);
          tc=get_top_most_client(WIN_LAYER_NORMAL,False);
        } else {
          // trick to prevent OSX window manager from bringing next X11 window in front
          XRaiseWindow(dpy,check_win);
          XLowerWindow(dpy,c->frame);
          raise_client(tc);
          lower_client(c);
        }
        if(tc!=c && tc!=NULL) {
          if(tc->is_docked)
            will_undock_client(tc);
          else
            set_focus_client(tc);
        } else
          XLowerWindow(dpy,check_win);
        if(need_ungrab) force_grab_release();
        check_ptrwin=1;
        check_top_client=1;
      }
      break;

    case itemHideGroup:
      need_ungrab = check_need_ungrab();
      unmap_group(get_group_for_client(c));
      tc = get_top_most_client(WIN_LAYER_NORMAL,0);
      if(gVerboseConsole && tc!=NULL && tc->osxwinfo.view_state==VIEWSTATE_HIDDEN) {
        printf("TOP WINDOW HIDDEN AFTER HIDE GROUP\n");
      }
      if(need_ungrab) force_grab_release();
      check_ptrwin=1;
      check_top_client=1;
      break;
      
    case itemNext:
      need_ungrab = check_need_ungrab();
      get_next_osxwinref_in_group(1,0);
      if(need_ungrab) force_grab_release();
      check_ptrwin=1;
      check_top_client=1;
      break;
      
    case itemNextGroup:
      need_ungrab = check_need_ungrab();
      raise_next_group();
      if(need_ungrab) force_grab_release();
      check_ptrwin=1;
      check_top_client=1;
//      printf("Next Group is not yet functional\n");
      break;

    case itemLowerGroup:
      if(c!=NULL) {
        need_ungrab = check_need_ungrab();
        if(get_group_for_client(c)==0)
          lower_client(c);
        else
          lower_group(get_group_for_client(c));
        tc = get_top_most_client(WIN_LAYER_NORMAL,0);
        if(tc!=NULL && tc->osxwinfo.view_state==VIEWSTATE_HIDDEN) {
            if(gVerboseConsole)
                printf("TOP WINDOW HIDDEN AFTER LOWER GROUP\n");
            map_group(get_group_for_client(tc));
        }
        if(need_ungrab) force_grab_release();
        check_ptrwin=1;
        check_top_client=1;
      }
//      printf("Send Group to Back is not yet functional\n");
      break;

    case itemFirstClick+100:
        c->auto_properties ^= NONAUTO_FIRSTCLICK;
        set_focus_client(c);
        break;
        
    case itemFirstClickChild+100:
        c->auto_properties ^= NONAUTO_FIRSTCLICKCHILD;
        set_focus_client(c);
        break;
        
    case itemFirstClickDrag+100:
        c->auto_properties ^= NONAUTO_DRAGFIRSTCLICK;
        set_focus_client(c);
        break;
        
    case itemAutoFocus+100:
      c->auto_properties ^= AUTO_FOCUS;
      set_focus_client(c);
      break;
      
    case itemAutoRaise+100:
      c->auto_properties ^= AUTO_RAISE;
      set_focus_client(c);
      break;
      
    case itemAutoLower+100:
      c->auto_properties ^= AUTO_LOWER;
      set_focus_client(c);
      break;
      
    case itemClickRaise+100:
        c->auto_properties ^= AUTO_CLICKRAISE;
        set_focus_client(c);
        break;
    
    case itemNoGrabFocus+100:
        c->auto_properties ^= NO_GRAB_FOCUS;
        set_focus_client(c);
        break;
        
    case itemRetainFocus+100:
        c->auto_properties ^= RETAIN_FOCUS;
        set_focus_client(c);
        break;
        
    case itemIgnoreClicks+100:
        c->auto_properties ^= NONAUTO_IGNORECLICKS;
        set_focus_client(c);
        break;
    
    // Hmmmmm... dunno how this is going to work - if you set this, then you can't
    // focus the client again, so you can't get the Property menu to set it back!!
    case itemIgnoreKeys+100:
        c->auto_properties ^= NONAUTO_IGNOREFOCUS;
        set_focus_client(c);
        break;
        
    case itemFloat+100:
#ifdef SHOW_WEIRD
      printf("Floating palettes not yet available\n");
#endif
      break;

    case itemPopup+100:
#ifdef SHOW_WEIRD
      printf("Popup windows not yet available\n");
#endif
      break;
      
    case itemModal+100:
      if(c->modal_state & STRICT_ANY_MODAL) {
        c->modal_state ^= DONT_IGNORE_MODAL;
        set_focus_client(c);
      }
      break;
      
    case itemLiveResize+100:
      c->auto_properties ^= LIVE_RESIZING;
      set_focus_client(c);
      break;

    case itemTranslucent0+100:
      if(c->osxwinfo.translucency!=ACTIVE_TRANSLUCENT0) {
        if(gVerboseConsole)
          printf("Making window %#lx non-translucent...\n",C_WINDOW);
        c->osxwinfo.translucency=ACTIVE_TRANSLUCENT0;
        c->osxwinfo.inactive_translucency=INACTIVE_TRANSLUCENT0;
        SendHackCommand(c,1,HACKCMD_TRANSLUCENT,ACTIVE_TRANSLUCENT0,0,0,0,0);
        Setup_Properties_Menu_Translucency(&c->osxwinfo);
      }
      break;

    case itemTranslucent1+100:
      if(c->osxwinfo.translucency!=ACTIVE_TRANSLUCENT1) {
        if(gVerboseConsole)
          printf("Making %#lx window translucent...\n",C_WINDOW);
        c->osxwinfo.translucency=ACTIVE_TRANSLUCENT1;
        c->osxwinfo.inactive_translucency=INACTIVE_TRANSLUCENT1;
        SendHackCommand(c,1,HACKCMD_TRANSLUCENT,ACTIVE_TRANSLUCENT1,0,0,0,0);
        Setup_Properties_Menu_Translucency(&c->osxwinfo);
      }
      break;

    case itemTranslucent2+100:
      if(c->osxwinfo.translucency!=ACTIVE_TRANSLUCENT2) {
        if(gVerboseConsole)
          printf("Making window %#lx very translucent...\n",C_WINDOW);
        c->osxwinfo.translucency=ACTIVE_TRANSLUCENT2;
        c->osxwinfo.inactive_translucency=INACTIVE_TRANSLUCENT2;
        SendHackCommand(c,1,HACKCMD_TRANSLUCENT,ACTIVE_TRANSLUCENT2,0,0,0,0);
        Setup_Properties_Menu_Translucency(&c->osxwinfo);
      }
      break;

/*    case itemSendToDock+100:
      dock_client(c);
      break;

    case itemGetFromDock+100:
      will_undock_client(c);
      break;
*/
    case itemNoframes+100:
#ifdef DEBUG
      printf("No frames toggle for win %#lx (%d)\n",C_WINDOW,c->has_frame);
#endif
      wc.height=c->height; 
      wc.width=c->width; 
      wc.x=c->x; 
      wc.y=c->y;
      if(c->has_frame==0) {

        if(c->title)
            XMapWindow(dpy,c->title);
        else if(config.top.height!=0 && config.top.width!=0)
            c->title=create_frame_part(c->frame, TX, TY, TW<1?1:TW, config.top.height, 1);

        if(c->top_left)
            XMapWindow(dpy,c->top_left);
        else if(config.top_left.height!=0 && config.top_left.width!=0)
            c->top_left=create_frame_part(c->frame, TLX, TLY, TLW, TLH, 1);
            
        if(c->top_right)
            XMapWindow(dpy,c->top_right);
        else if(config.top_right.height!=0 && config.top_right.width!=0)
            c->top_right=create_frame_part(c->frame, TRX, TRY, TRW, TRH, 1);
            
        if(c->zoom)
            XMapWindow(dpy,c->zoom);
            
        if(c->icon)
            XMapWindow(dpy,c->icon);
            
        if(c->close)
            XMapWindow(dpy,c->close);
            
        if(config.left.height==0 || config.left.width==0) {
            if(c->left) XDestroyWindow(dpy,c->left);
            c->left=0;
        } else if(c->left) XMapWindow(dpy,c->left);
        else c->left=create_frame_part(c->frame, LX, LY, LW, LH, 1);
            
        if(c->right)
            XMapWindow(dpy,c->right);
        else if(config.right.height!=0 && config.right.width!=0)
            c->right=create_frame_part(c->frame, RX, RY, RW, RH, 1);
            
        if(c->bottom)
            XMapWindow(dpy,c->bottom);
        else if(config.bottom.height!=0 && config.bottom.width!=0)
            c->bottom=create_frame_part(c->frame, BX, BY, BW<1?1:BW, config.bottom.height, 1);
            
        if(c->bottom_left)
            XMapWindow(dpy,c->bottom_left);
        else if(config.bottom_left.height!=0 && config.bottom_left.width!=0)
            c->bottom_left=create_frame_part(c->frame, BLX, BLY, BLW, BLH, 1);
            
        if(c->resize)
            XMapWindow(dpy,c->resize);
        else if(config.resize.width!=0 && config.resize.height!=0)
            c->resize=create_frame_part(c->frame, RSX, RSY, RSW, RSH, 1);
            
        gravitate_client(c, REMOVE_GRAVITY);
        c->has_frame=1;
        gravitate_client(c, APPLY_GRAVITY);
        
        configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, False, 3);
      } else {
        if((get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)==WIN_STATE_SHADED)
            toggle_client_shaded(c);
        if(c->title) XUnmapWindow(dpy,c->title);
        if(c->top_left) XUnmapWindow(dpy,c->top_left);
        if(c->top_right) XUnmapWindow(dpy,c->top_right);
        if(c->zoom) XUnmapWindow(dpy,c->zoom);
        if(c->icon) XUnmapWindow(dpy,c->icon);
        if(c->close) XUnmapWindow(dpy,c->close);
        if(c->left) XUnmapWindow(dpy,c->left);
        if(c->right) XUnmapWindow(dpy,c->right);
        if(c->bottom) XUnmapWindow(dpy,c->bottom);
        if(c->bottom_left) XUnmapWindow(dpy,c->bottom_left);
        if(c->resize) XUnmapWindow(dpy,c->resize);
        gravitate_client(c, REMOVE_GRAVITY);
        c->has_frame=0;
        gravitate_client(c, APPLY_GRAVITY);
        configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, False, 3);
      }
      resize_client(c,c->width,c->height,3);
      set_focus_client(c);
      break;

    default:
#ifdef SHOW_WEIRD
      printf("%s oddity: Got funny action in manipulate_focus_client: %d (%#lx)\n",PROGNAME,action,C_WINDOW);
#endif
      break;
  }
}

void reset_all_inactive_clients(int newset, int which)
{
    Client *c;
    int i;
    
    c=client_list;
    if(c==NULL) return;
    
    for(i=0;i<client_count;i++) {
      if(c!=lastfocus_client) {
        if(which==BGWIN_DIMMED) {
          // check if client is becoming dimmed
          if((newset & which)!=0 && client_modal_firstclick(c)!=lastfocus_client
                    && (c->auto_properties & (NONAUTO_FIRSTCLICK | AUTO_FOCUS))!=NONAUTO_FIRSTCLICK)
            SendHackCommand(c,1,HACKCMD_DIMWIN,1,0,0,0,0);
          else
            SendHackCommand(c,1,HACKCMD_DIMWIN,0,0,0,0,0);
          if((c->has_frame) || (c->is_icon!=NOT_ICONBAR)) redraw_client_frame(c);
        } else if(which==BGWIN_TRANSLUCENT) {
          // check if client is becoming translucent
          if((newset & which)!=0 && client_modal_firstclick(c)!=lastfocus_client
                    && (c->auto_properties & (NONAUTO_FIRSTCLICK | AUTO_FOCUS))!=NONAUTO_FIRSTCLICK)
            SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.inactive_translucency,0,0,0,0);
          else
            SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.translucency,0,0,0,0);
        }
      }
      c=c->next;
    }
}

void change_options(int newattr)
{
  switch(newattr)
  {
    case itemBGWinDimmed:
      gBGWinSetting ^= BGWIN_DIMMED;
      Setup_Options_Menu(0);
      reset_all_inactive_clients(gBGWinSetting,BGWIN_DIMMED); // toggle dimness of inactive clients
      break;
      
    case itemBGWinTranslucent:
      gBGWinSetting ^= BGWIN_TRANSLUCENT;
      Setup_Options_Menu(0);
      reset_all_inactive_clients(gBGWinSetting,BGWIN_TRANSLUCENT); // toggle translucency of inactive clients
      break;
      
    case itemMinToDock:
#ifdef SHOW_WEIRD
      if(can_send_hackcmds==0)
        printf("Should not get itemMinToDock!\n");
#endif
      if(gMinMethod==minmethod_dock)
        gMinMethod=minmethod_iconbars;
      else if(gMinMethod==minmethod_iconbars)
        gMinMethod=minmethod_dock;
      Setup_Options_Menu(0);
      break;        

    case itemHideIconbars:
#ifdef SHOW_WEIRD
      if(can_send_hackcmds==0)
        printf("Should not get itemHideIconbars!\n");
#endif
      if(gHiding_Iconbars==0) {
        gHiding_Iconbars=1;
        will_hide_all_iconbars();
      } else {
        unhide_all_iconbars();
        gHiding_Iconbars=0;
      }
      Setup_Options_Menu(0);
      break;
      
    case itemTopLeft:
      change_iconbarpos(ICONBARPOS_TOPLEFT);
      Setup_Options_Menu(newattr);
      break;

    case itemTopRight:
      change_iconbarpos(ICONBARPOS_TOPRIGHT);
      Setup_Options_Menu(newattr);
      break;

    case itemBotLeft:
      change_iconbarpos(ICONBARPOS_BOTLEFT);
      Setup_Options_Menu(newattr);
      break;

    case itemBotRight:
      change_iconbarpos(ICONBARPOS_BOTRIGHT);
      Setup_Options_Menu(newattr);
      break;
    
    case itemSlowIconbars:
      iconbar_steps=ICONSTEPS_SLOW;
      Setup_Options_Menu(newattr);
      break;
      
    case itemMedIconbars:
      iconbar_steps=ICONSTEPS_MED;
      Setup_Options_Menu(newattr);
      break;
      
    case itemFastIconbars:
      iconbar_steps=ICONSTEPS_FAST;
      Setup_Options_Menu(newattr);
      break;
      
    case itemFancyIconbars:
      if(shrinking_iconbars==0)
        shrinking_iconbars=1;
      else
        shrinking_iconbars=0;
      Setup_Options_Menu(0);
      break;
      
    case itemInterleave:
//#ifdef SHOW_WEIRD
//      printf("Automatic interleave control not yet available...\n");
//#endif
      Setup_Options_Menu(0);
      break;

/*    case itemLeaveMenubar:
      change_iconbar_menuspace();
      Setup_Options_Menu(0);
      break;*/
    
    default:
      if(newattr>=itemScreenListStart) {
        change_iconbar_screen(newattr-itemScreenListStart);
        Setup_Options_Menu(newattr);
      } else {
#ifdef SHOW_WEIRD
      printf("%s oddity: Got unknown attr in change_options\n",PROGNAME);
#endif
      }
      break;
  }

}

void check_missing_clients()
{
  Client *c, *orgc;
  int i, done=0;
  
  while(done==0) {
    c=client_list;
    if(c==NULL || client_count==0) return;
    for(i=0;i<client_count;i++) {
      if(c->osxwinfo.view_state==VIEWSTATE_HIDDEN) {
        printf("%s info: window has disappeared: %#lx (%s)\n",PROGNAME,C_WINDOW,c->name);
        delete_client(c,-1);
//        done=1;
        break;
      } else
        c=c->next;
        if(c==NULL) {
          done=1;
          break;
        } else if(i==client_count-1)
          done=1;
    }
  }
  
}

void map_all_clients() {
  Window *windows=NULL, w1, w2; 
  int count, i, x, y; 
  Client *c;
  XWindowAttributes attr;

//  if(can_send_hackcmds!=0) return;
  
  sprintf(lastsub,"TRACE1: map_all_clients\n"); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: map_all_clients\n"); 
#endif
 
  MyIncreaseXServerGrab(); // grab the server, so nothing happens while we're messing around
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
#ifdef DEBUG
  printf("  children of root: %d\n",count);
#endif

  // Search through the windows that actually exist. (It's possible that some might have
  // disappeared while hidden, but they remain in the client list. This is because the window
  // manager had no notification they were unmapped as they already were unmapped when the
  // user chose to hide OroborOSX.)
  for(i=0;i<count;i++) { 
#ifdef DEBUG
    printf("  child %d (%#lx)\n",i,windows[i]);
#endif
    c=get_client_from_window(windows[i],FRAME);
    if(can_send_hackcmds==0) {
      if(c!=NULL) {
        if(c->is_docked==0) {
          XMapWindow(dpy, c->frame); 
          if(c->is_icon!=NOT_ICONBAR)
            c->osxwinfo.view_state=VIEWSTATE_ICONBAR;
          else
            c->osxwinfo.view_state=VIEWSTATE_VISIBLE;
          if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
          if(c->is_icon==NOT_ICONBAR) {
            XMapWindow(dpy, c->window);
            set_wm_state(c->window, NormalState);
          }
          if((c->has_frame) || (c->is_icon!=NOT_ICONBAR)) redraw_client_frame(c);
        } else {
          c->osxwinfo.view_state=VIEWSTATE_DOCKED;
          if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
        }
      }
    } else {
      XGetWindowAttributes(dpy, windows[i], &attr);
//      if(attr.map_state==IsViewable) {
        x = attr.x;
        y = attr.y;
//      }
      XMoveWindow(dpy, windows[i], HACKCMD_STARTX, HACKCMD_STARTY);
      XMoveWindow(dpy, windows[i], HACKCMD_HIDEWINDOW, 0);
      if(c!=NULL) {
        if(c->is_docked!=0) {
          c->osxwinfo.view_state=VIEWSTATE_DOCKED;
        } else if(c->is_icon!=NOT_ICONBAR) {
          c->osxwinfo.view_state=VIEWSTATE_ICONBAR;
        } else {
          set_wm_state(windows[i], NormalState);
          c->osxwinfo.view_state=VIEWSTATE_VISIBLE;
        }
        if(c->isoffscreen==0) {
          if(c->has_frame) redraw_client_frame(c);
          XMoveWindow(dpy,windows[i],FX,FY);
          MyXSync(dpy,0);
        }
        if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
      } else {
        set_wm_state(windows[i], NormalState);
        XMoveWindow(dpy, windows[i], x, y);
        MyXSync(dpy,0);
      }
    }
  } 

#ifdef DEBUG
  printf("focus client before: %#lx\n",focus_before_hide);
#endif

  if(windows) XFree(windows);

  // loop through possibilities until there is nothing left (got NULL) or find OK client (then set NULL)
  while(focus_before_hide!=NULL) {

    if(focus_before_hide==-1) {	// focus client disappeared while hidden, so get topmost client and try again
      printf("%s warning: focus win disappeared while hidden\n",PROGNAME);
      focus_before_hide=get_top_most_client(WIN_LAYER_NORMAL,False);
    } else if(focus_before_hide->osxwinfo.view_state==VIEWSTATE_HIDDEN) {	// client is still hidden?!
#ifdef SHOW_WEIRD
      printf("focus win is hidden!! %#lx (%s)\n",focus_before_hide->window,focus_before_hide->name);
#endif
      delete_client(focus_before_hide,-1);
      focus_before_hide=get_top_most_client(WIN_LAYER_NORMAL,False);
    } else if(focus_before_hide->osxwinfo.view_state==VIEWSTATE_ICONBAR) {	// focus client is an iconbar!?
#ifdef SHOW_WEIRD
      printf("focus win is an iconbar!! %#lx (%s)\n",focus_before_hide->window,focus_before_hide->name);
#endif
      focus_before_hide=get_top_most_client(WIN_LAYER_NORMAL,False);		// go for topmost instead
    } else {	// everything is OK now, so set the focus client and get out
#ifdef DEBUG
      printf("focus win OK: %#lx\n",focus_before_hide->window);
#endif
      if(!grabstat_before_hide) set_focus_client(focus_before_hide);
      grabstat_before_hide=0;
      focus_before_hide=NULL;
    }

  }

  MyDecreaseXServerGrab(); // release the X server, if it was before
  check_missing_clients();

}

void unmap_all_clients() {
  Window *windows=NULL, w1, w2; 
  int count, i, x, y; 
  Client *c;
  XWindowAttributes attr;
  
  focus_before_hide=get_focus_client();
  
//  if(can_send_hackcmds!=0) {
//    SendSimpleHackCommand(HACKCMD_HIDE_ALLXWIN,1);
//    return;
//  }
  
  sprintf(lastsub,"TRACE1: unmap_all_clients\n"); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: unmap_all_clients\n"); 
#endif
 
  MyIncreaseXServerGrab(); // grab the server, so nothing happens while we're messing around
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
#ifdef DEBUG
  printf("  children of root: %d\n",count);
#endif

  for(i=0;i<count;i++) 
  { 
#ifdef DEBUG
    printf("  child %d (%#lx)\n",i,windows[i]);
#endif
    c=get_client_from_window(windows[i],FRAME);
    if(can_send_hackcmds==0) {
      if(c!=NULL) {
        if(c->is_docked==0) {
          c->ignore_unmap++; 
          XUnmapWindow(dpy, c->frame);
          c->osxwinfo.view_state=VIEWSTATE_HIDDEN;
          if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
          XUnmapWindow(dpy, c->window); 
          set_wm_state(c->window, WithdrawnState); 
        }
      }
    } else {
      if(c!=NULL) {
        c->osxwinfo.view_state=VIEWSTATE_HIDDEN;
        if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
      }
      XGetWindowAttributes(dpy, windows[i], &attr);
      if(attr.map_state==IsViewable) {
        x = attr.x;
        y = attr.y;
        set_wm_state(windows[i], WithdrawnState);
        XMoveWindow(dpy, windows[i], HACKCMD_STARTX, HACKCMD_STARTY);
        XMoveWindow(dpy, windows[i], HACKCMD_HIDEWINDOW, 1);
        /*if(c!=NULL && c->isoffscreen==0)
            XMoveWindow(dpy, windows[i], x, y);
        else*/
            XMoveWindow(dpy, windows[i], -5000, -5000);
        MyXSync(dpy,0);
      }
    }
  }

  MyDecreaseXServerGrab(); // release the X server, if it was before
  
  if(windows) XFree(windows);
  
//  c=client_list;
//  for(i=0;i<client_count;i++) {
//    c=c->next;
//  }
}

/*
 * For debugging purposes
 */
void list_all_windows()
{
  Window *windows=NULL, w1, w2; 
  int count, i, hints, argc; 
  char **argv;
  Client *c;
  XWindowAttributes attr; 
  XClassHint xch;

  XGetWindowAttributes(dpy, root, &attr); 

  XQueryTree(dpy, root, &w1, &w2, &windows, &count);
  if(count==1)
    printf("%s info: there is %d top level X11 window...\n",PROGNAME,count);
  else
    printf("%s info: there are %d top level X11 windows...\n",PROGNAME,count);

  for(i=count-1;i>=0;i--) 
  { 
    char *myname=NULL;
    XGetWindowAttributes(dpy, windows[i], &attr);
    if(attr.map_state==IsViewable)
      printf("  window %d (%#lx) viewable, (X,Y): %d,%d (W,H): %d,%d\n",i,windows[i],attr.x,attr.y,attr.width,attr.height);
    else
      printf("  window %d (%#lx) not viewable, (X,Y): %d,%d (W,H): %d,%d\n",i,windows[i],attr.x,attr.y,attr.width,attr.height);
    c=get_client_from_window(windows[i], FRAME);
    if(c==NULL) {
      if(windows[i]==check_win)
        printf("     this is the oroborus 'check_win' created for use with gnome...\n");
//      else if(windows[i]==cursor_follow_win)
//        printf("     this is the 'cursor_follow_win' created to avoid inconsistent mouseovers...\n");
      else
        printf("     this window is not recognised as a client by OroborOSX\n");
    } else {
      printf("     client ptr,org-grp,name,now-grp: %#lx ,'%s','%s','%s'\n",c,c->clientgrouporg,c->clientname,c->clientgroup);
      printf("     client frame,window,osxwinref,iconbarstate,hackID: %#lx,%#lx,%#lx,%d,%d\n",c->frame,c->window,c->osxwinfo.winref,c->is_icon,c->hack_id);
      windows[i] = c->window;
    }
#ifdef I18N_SUPPORT 
    if(use_fontset)
      // call wrapper function for I18N 
      XFetchNameWrapper(dpy, windows[i], &myname); 
    else
      XFetchName(dpy, windows[i], &myname);
#else
    XFetchName(dpy, windows[i], &myname); 
#endif /* I18N */
    if(myname!=NULL) {
      printf("     window title: %s\n",myname);
      XFree(myname);
    } else
      printf("     got NULL title...\n");
    hints=XGetClassHint(dpy,windows[i],&xch);
    if(hints!=0) {
      printf("     window class & name: '%s', '%s'\n",xch.res_class,xch.res_name);
      XFree(xch.res_name);
      XFree(xch.res_class);
    }
    if(XGetCommand(dpy,windows[i],&argv,&argc)!=0) {
      printf("     window command (%d args): %s\n",argc,*argv);
      if(*argv!=NULL) XFreeStringList(argv);
    } else
      printf("     got NULL command string...\n");
  } 

  if(windows) XFree(windows);

  c=client_list;
  if(c==NULL) {
    printf("%s info: no clients\n",PROGNAME);
    return;
  }
  if(client_count==0) {
    printf("%s oddity: No clients in list, but client_list is %#lx !?\n",PROGNAME,client_list);
    return;
  }
  
}

void list_all_clients()
{
  Client *c;
  int i;
  
  c=client_list;
  if(c==NULL) {
    printf("%s info: no clients\n",PROGNAME);
    return;
  }
  if(client_count==0) {
    printf("%s oddity: No clients in list, but client_list is %#lx !?\n",PROGNAME,client_list);
    return;
  }

  printf("%s info: List of clients:\n",PROGNAME);
  for(i=0;i<client_count;i++) {
    printf("  %#lx: title,org-grp,name,now-grp: '%s','%s','%s','%s'\n",c,c->name,c->clientgrouporg,c->clientname,c->clientgroup);
    printf("            frame,window,osxwinref,iconbarstate,hackID: %#lx,%#lx,%#lx,%d,%d\n",c->frame,c->window,c->osxwinfo.winref,c->is_icon,c->hack_id);
    c=c->next;
  }

}

/*
 * Check that OSX windows are off-screen
 */
void check_osx_windows_bounds() {
  Window *windows=NULL; 
  int count, i; 
  Client *c;

    sprintf(lastsub,"TRACE3: check_osx_windows_bounds\n");
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: check_osx_windows_bounds\n");
#endif

  c=client_list;
  if(c==NULL) return;
  if(client_count==0) {
    printf("%s oddity: No clients in list, but client_list is %#lx !?\n",PROGNAME,client_list);
    return;
  }
  
  for(i=0;i<client_count;i++) {
    if(c->osxwinfo.winref!=0)
        Check_OSX_Window_Bounds(&c->osxwinfo);
    c=c->next;
  }
}

/*
 * Find the OSX WindowRef (returned as a long) for the bottom-most client (for the "Next" menu item)
 */
long get_next_osxwinref(int raiseit)
{
    Client *c, *fc;
    long osxwr=0;
    
    fc=get_focus_client();
    
//#ifdef DEBUG
    if(fc==NULL && gVerboseConsole)
      printf("Got NULL focus client in get_next_osxwinref.\n");
//#endif
    if(fc==NULL) return 0;
    
    c=get_bottom_most_client(get_gnome_hint(fc->window,WIN_LAYER),False);
    
//#ifdef DEBUG
    if(c==fc && gVerboseConsole)
      printf("Got focus client is next client in get_next_osxwinref.\n");
//#endif
    if(c==fc) return 0;
    
    if(c!=NULL) {
      if(raiseit!=0) {
        if(c->is_docked)
          will_undock_client(c);
        else {
          set_focus_client(c);
          if((c->auto_properties & (AUTO_FOCUS|AUTO_RAISE))==AUTO_FOCUS)
            raise_client(c);
          if(c->al_list!=NULL) XFree(c->al_list);
          c->al_list = NULL;
          c->al_count = 0;
        }
      }
      osxwr=c->osxwinfo.winref;
    }
    
    sprintf(lastsub,"TRACE2: get_next_osxwinref -got next win: %#lx %d (winref: %d)\n",C_WINDOW,raiseit,osxwr);
    subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE3: get_next_osxwinref -got next win: %#lx %d (winref: %d)\n",C_WINDOW,raiseit,osxwr);
#else
    if(gVerboseConsole)
        printf("%s info: got next win: %#lx %d (winref: %d)\n",PROGNAME,C_WINDOW,raiseit,osxwr);
#endif
  
    return osxwr;
}

/*
 * Find the full title for the bottom-most client (for the "Next" menu item)
 */
char *get_next_win_title(int raiseit)
{
    Client *c, *fc;
    char *wintitle=NULL, *offset_title=NULL;
    
    fc=get_focus_client();
    
#ifdef DEBUG
    if(fc==NULL)
      printf("Got NULL focus client in get_next_win_title.\n");
#endif
    if(fc==NULL) return NULL;
    
    c=get_bottom_most_client(get_gnome_hint(fc->window,WIN_LAYER),False);
    
#ifdef DEBUG
    if(c==fc)
      printf("Got focus client is next client in get_next_win_title.\n");
#endif
    if(c==fc) return NULL;
    
    if(c!=NULL) {
      if(raiseit!=0) {
        if(c->is_docked)
          will_undock_client(c);
        else {
          set_focus_client(c);
          raise_client(c);
        }
      }
      if(c->name==NULL && c->clientgroup==NULL) {
        wintitle = malloc(15);
        sprintf(wintitle,"?: Untitled");
      } else if(c->name==NULL) {
        wintitle = malloc(strlen(c->clientgroup)+15);
        sprintf(wintitle,"%s: Untitled",c->clientgroup);
      } else if(c->clientgroup==NULL) {
        wintitle = malloc(strlen(c->name)+5);
        sprintf(wintitle,"?: %s",c->name);
      } else {
        offset_title = c->name;
        if(strlen(c->clientgroup)<strlen(c->name))
            if(!strncmp(c->clientgroup,c->name,strlen(c->clientgroup))) {
                offset_title+=strlen(c->clientgroup);
                // strip out spaces after class name
                while(strlen(offset_title)>1 && !strncmp(offset_title," ",1))
                    offset_title++;
                // strip out colon
                if(strlen(offset_title)>1 && !strncmp(offset_title,":",1))
                    offset_title++;
                // strip out any spaces after colon
                while(strlen(offset_title)>1 && !strncmp(offset_title," ",1))
                    offset_title++;
            }
        wintitle = malloc(strlen(c->clientgroup)+strlen(offset_title)+5);
        sprintf(wintitle,"%s:  %s",c->clientgroup,offset_title);
      }
    }
    
    sprintf(lastsub,"TRACE2: get_next_win_title -got next win: %#lx (title: %s)\n",C_WINDOW,wintitle);
    subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: get_next_win_title -got next win: %#lx (title: %s)\n",C_WINDOW,wintitle);
#endif
  
    return wintitle;
}

