/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 
 
#include "oroborus.h" 
#include "hackcmds.h"
#include "iconbars.h"

/* 
 * 
 * Redraw client borders 
 * 
 */ 

void redraw_client_frame(Client *c) 
{ 
  int state=FRAME_INACTIVE, shaded=0, wasalpha, wasinhibit; 
  
  sprintf(lastsub,"TRACE3: redraw_client_frame (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW);
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: redraw_client_frame (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW);
#endif 

#ifdef SHOW_WEIRD
  if(c==NULL) printf("%s oddity: got NULL client in redraw_client_frame!\n",PROGNAME);
#endif
  if(c==NULL) return;

  if(c->has_frame==0 && c->is_icon==NOT_ICONBAR) {
    redraw_client_title(c);	// need to make sure title of OSX window is set
    if(gVerboseConsole)
        printf("%s info: window %#lx has no frame - no frame drawing done\n",PROGNAME,c->window); 
    return; 
  }
  
  MyIncreaseXServerGrab();
//  if(c==get_focus_client() && gX11InFront!=0)
  wasalpha=c->osxwinfo.drawalpha;
  wasinhibit=c->inhibit_level;
#ifdef DEBUG
  printf("wasalpha at start of frame: %d\n",wasalpha);
#endif
  if(c==lastfocus_client && gX11InFront!=0) {
    state=FRAME_ACTIVE;
#ifdef DEBUG
    printf("redraw_client_frame telling XDarwin to set alpha to %d\n",config.active_titlebar_alpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,config.active_titlebar_alpha,0,0,0,0);
  } else if(c->is_icon==NOT_ICONBAR) {
#ifdef DEBUG
    printf("redraw_client_frame telling XDarwin to set alpha to %d\n",config.inactive_titlebar_alpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,config.inactive_titlebar_alpha,0,0,0,0);
    if(wasinhibit>HACKARG_INHIBITSCREEN)
        SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,HACKARG_INHIBITSCREEN,0,0,0,0);
  }
  
/*  if(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED)*/
  if( c->is_icon==ICONBAR_RAISED ) {
    shaded=4;
    state=FRAME_ACTIVE;
  } else if( c->is_icon==ICONBAR_HIDDEN || c->is_icon==ICONBAR_VISIBLE ) {
    shaded=4;
    state=FRAME_INACTIVE;
  } else if(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED) {
    shaded=2;
  }
 
  //if(c->is_dimmed==1)
  //  SendHackCommand(c,1,HACKCMD_DIMWIN,2,0,0,0,0); // temporarily switch off dimming -no reshape nor redraw!
  
  if(c->is_icon!=NOT_ICONBAR && TW<1) {
    // this iconbar is 'hidden' (i.e. it is shrunk to a small tab on one side of the screen)
    if(iconbarpos==ICONBARPOS_TOPLEFT || iconbarpos==ICONBARPOS_BOTLEFT) {
      // iconbars on left, so try to only do the right side of the iconbar tab
      if(c->top_right) {
        XSetWindowBackgroundPixmap(dpy, c->top_right, config.top_right.pixmap[state+shaded]);
        XClearWindow(dpy, c->top_right);
      } else if(c->top_left) {
        XSetWindowBackgroundPixmap(dpy, c->top_left, config.top_left.pixmap[state+shaded]);
        XClearWindow(dpy, c->top_left);
      }
    } else {
      // iconbars on right, so try to only do the left side of the iconbar tab
      if(c->top_left) {
        XSetWindowBackgroundPixmap(dpy, c->top_left, config.top_left.pixmap[state+shaded]);
        XClearWindow(dpy, c->top_left);
      } else if(c->top_right) {
        XSetWindowBackgroundPixmap(dpy, c->top_right, config.top_right.pixmap[state+shaded]);
        XClearWindow(dpy, c->top_right);
      }
    }
  } else {
    if(c->top_left) {
      XSetWindowBackgroundPixmap(dpy, c->top_left, config.top_left.pixmap[state+shaded]); 
      XClearWindow(dpy, c->top_left); }
    if(c->top_right) {
      XSetWindowBackgroundPixmap(dpy, c->top_right, config.top_right.pixmap[state+shaded]);
      XClearWindow(dpy, c->top_right); }
  }
  
  if( c->is_icon==NOT_ICONBAR ) {

    if(c->left) { XSetWindowBackgroundPixmap(dpy, c->left, config.left.pixmap[state]); 
    XClearWindow(dpy, c->left); }
    if(c->right) { XSetWindowBackgroundPixmap(dpy, c->right, config.right.pixmap[state]); 
    XClearWindow(dpy, c->right); }
    if(c->bottom_left) { XSetWindowBackgroundPixmap(dpy, c->bottom_left, config.bottom_left.pixmap[state]); 
    XClearWindow(dpy, c->bottom_left); }
    if(c->bottom) { XSetWindowBackgroundPixmap(dpy, c->bottom, config.bottom.pixmap[state]); 
    XClearWindow(dpy, c->bottom); }
    if(c->resize) { XSetWindowBackgroundPixmap(dpy, c->resize, config.resize.pixmap[state]); 
    XClearWindow(dpy, c->resize); }
 
    redraw_client_buttons(c); 
  }

  if( TW>0 ) redraw_client_title(c);
  
//  XFlush(dpy);
//  XSync(dpy,0);
  
  //if(c->is_dimmed==1)
  //  SendHackCommand(c,1,HACKCMD_DIMWIN,3,0,0,0,0); // switch on dimming again -no reshape nor redraw!

//  if(c->osxwinfo.drawalpha!=wasalpha)
#ifdef DEBUG
  printf("wasalpha in frame: %d\n",wasalpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,wasalpha,0,0,0,0);

  if(wasinhibit!=c->inhibit_level)
    SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,wasinhibit,0,0,0,0);

  MyDecreaseXServerGrab();

  sprintf(lastsub,"TRACE3: leaving redraw_client_frame\n");
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: leaving redraw_client_frame\n");
#endif 
} 
 
int check_correct_button_state(int but_state, int frame_state)
{
    if(but_state==BUTTON_OVER) {
        if(frame_state==FRAME_INACTIVE)
            return BUTTON_OVERBACK;
    } else if(but_state==BUTTON_OVERBACK) {
        if(frame_state==FRAME_ACTIVE)
            return BUTTON_OVER;
    } else if(but_state==BUTTON_PRESSED) {
        if(frame_state==FRAME_INACTIVE)
            return BUTTON_PRESSEDBACK;
    } else if(but_state==BUTTON_PRESSEDBACK) {
        if(frame_state==FRAME_ACTIVE)
            return BUTTON_PRESSED;
    } else if(but_state==BUTTON_ACTIVEOFF) {
        if(frame_state==FRAME_INACTIVE)
            return BUTTON_INACTIVEOFF;
    } else if(but_state==BUTTON_INACTIVEOFF) {
        if(frame_state==FRAME_ACTIVE)
            return BUTTON_ACTIVEOFF;
    }
    return but_state;
}

/* 
 *  
 * Redraw client buttons 
 * 
 */ 

void redraw_client_buttons(Client *c) 
{ 
  int state=FRAME_INACTIVE, mainstate, wasalpha, wasinhibit; 
 
//  sprintf(lastsub,"TRACE3: redraw_client_buttons (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW);
//  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: redraw_client_buttons (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW);
#endif

#ifdef SHOW_WEIRD
  if(c==NULL) printf("%s oddity: got NULL client in redraw_client_buttons!\n",PROGNAME);
#endif
  if(c==NULL) return;

#ifdef SHOW_WEIRD 
  if(c->is_icon!=NOT_ICONBAR) printf("%s oddity: window %#lx is an iconbar - no button drawing done\n",PROGNAME,c->window);
//  if(c->has_frame==0 && c->is_icon==NOT_ICONBAR) printf("%s oddity: client/window (%#lx,%#lx) has no frame - no button drawing done\n",PROGNAME,c,c->window); 
#endif 
  if( c->is_icon!=NOT_ICONBAR ) return;
  if(c->has_frame==0 && c->is_icon==NOT_ICONBAR) return;
 
  MyIncreaseXServerGrab();

  //if(c->is_dimmed==1)
  //  SendHackCommand(c,1,HACKCMD_DIMWIN,2,0,0,0,0); // temporarily switch off dimming -no reshape nor redraw!

//  if(c==get_focus_client() && gX11InFront!=0) 
  wasalpha=c->osxwinfo.drawalpha;
  wasinhibit=c->inhibit_level;
  if(c==lastfocus_client && gX11InFront!=0) {
    state=FRAME_ACTIVE;
#ifdef DEBUG
    printf("redraw_client_buttons telling XDarwin to set alpha to %d\n", config.active_titlebar_alpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,config.active_titlebar_alpha,0,0,0,0);
  } else {
#ifdef DEBUG
    printf("redraw_client_buttons telling XDarwin to set alpha to %d\n", config.inactive_titlebar_alpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,config.inactive_titlebar_alpha,0,0,0,0);
    if(wasinhibit>HACKARG_INHIBITSCREEN)
        SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,HACKARG_INHIBITSCREEN,0,0,0,0);
  }
 
  mainstate=state;
  
  if(c->icon) {
    c->icon_state=check_correct_button_state(c->icon_state,state);
    XSetWindowBackgroundPixmap(dpy, c->icon,
            config.icon.pixmap[c->icon_state==BUTTON_UNPRESSED?state:c->icon_state]); 
    XClearWindow(dpy, c->icon);
  }
  if(c->close) {
    c->close_state=check_correct_button_state(c->close_state,state);
    XSetWindowBackgroundPixmap(dpy, c->close,
            config.close.pixmap[c->close_state==BUTTON_UNPRESSED?state:c->close_state]); 
    XClearWindow(dpy, c->close);
  }
  if(c->can_max) {
//    if(c->zoom_state!=BUTTON_UNPRESSED) state=c->zoom_state;
    if(c->zoom_state!=BUTTON_UNPRESSED)
        state=check_correct_button_state(c->zoom_state,state);
  } else if(state==FRAME_ACTIVE)
    state=BUTTON_ACTIVEOFF;
  else if(state==FRAME_INACTIVE)
    state=BUTTON_INACTIVEOFF;
  if(c->zoom) {
    XSetWindowBackgroundPixmap(dpy, c->zoom, config.zoom.pixmap[state]);
    XClearWindow(dpy, c->zoom);
  }

//  XFlush(dpy);
//  XSync(dpy,0);

  //if(c->is_dimmed==1)
  //  SendHackCommand(c,1,HACKCMD_DIMWIN,3,0,0,0,0); // switch on dimming again -no reshape nor redraw!

  if(wasinhibit!=c->inhibit_level)
    SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,wasinhibit,0,0,0,0);

//  if(wasalpha!=c->osxwinfo.drawalpha)
#ifdef DEBUG
  printf("wasalpha in buttons: %d\n",wasalpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,wasalpha,0,0,0,0);

  MyDecreaseXServerGrab();
} 
 
/* 
 * 
 * Redraw client name in the title bar 
 * 
 */ 

void redraw_client_title(Client *c) 
{ 
  int state=FRAME_INACTIVE, shaded=0; 
  Pixmap temp; 
  GC gc; 
  XGCValues gv;
  char *name_string=NULL;
#ifdef I18N_SUPPORT 
  XRectangle ink, logical; 
  int ascent=0, descent=0; 
#endif /* I18N */ 
  int width=0, leftext, rightext, left_add=0, right_add=0, wasalpha, wasinhibit; 
 
//  sprintf(lastsub,"TRACE3: redraw_client_title (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW);
//  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: redraw_client_title (win,last: %#lx,%#lx)\n",C_WINDOW,FC_WINDOW);
#endif

#ifdef SHOW_WEIRD
  if(c==NULL)
    printf("%s oddity: got NULL client in redraw_client_title!\n",PROGNAME);
#endif
  if(c==NULL) return;

#ifdef SHOW_WEIRD
  if(c->title==0) printf("%s oddity: got zero title window in redraw_client_title! (%#lx)\n",PROGNAME,c->window);
#endif
  if(c->title==0) return;
  
  if(c->has_frame==0 && c->is_icon==NOT_ICONBAR) {
//#ifdef SHOW_WEIRD 
//    printf("%s oddity: client/window (%#lx,%#lx) has no frame - no title drawing done\n",PROGNAME,c,c->window); 
//#endif 
    if(c->osxwinfo.winref!=0)
      CheckOSXWinTitle(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
    return;
  }
  
  if(TW<1) printf("%s warning: Not drawn title of window %#lx: %d\n",PROGNAME,c->window,TW);
  if(TW<1) return;
  
  MyIncreaseXServerGrab();

  //if(c->is_dimmed==1)
  //  SendHackCommand(c,1,HACKCMD_DIMWIN,2,0,0,0,0); // temporarily switch off dimming -no reshape nor redraw!

//  if(c==get_focus_client() && gX11InFront!=0) 
  wasalpha=c->osxwinfo.drawalpha;
  wasinhibit=c->inhibit_level;
  if(c==lastfocus_client && gX11InFront!=0) {
    state=FRAME_ACTIVE;
#ifdef DEBUG
    printf("redraw_client_title telling XDarwin to set alpha to %d\n", config.active_titlebar_alpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,config.active_titlebar_alpha,0,0,0,0);
  } else if(c->is_icon==NOT_ICONBAR) {
#ifdef DEBUG
    printf("redraw_client_title telling XDarwin to set alpha to %d\n", config.inactive_titlebar_alpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,config.inactive_titlebar_alpha,0,0,0,0);
    if(wasinhibit>HACKARG_INHIBITSCREEN)
        SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,HACKARG_INHIBITSCREEN,0,0,0,0);
  }
  
/*  if(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED)*/
  if( c->is_icon==ICONBAR_VISIBLE || c->is_icon==ICONBAR_HIDDEN ) {
    shaded=4;
    state=FRAME_INACTIVE;
  } else if( c->is_icon==ICONBAR_RAISED ) {
    shaded=4;
    state=FRAME_ACTIVE;
  } else if(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED) {
    shaded=2;
  }
 
  gv.function=GXcopy; 
  gv.tile=config.top.pixmap[state+shaded]; 
  gv.fill_style=FillTiled; 
  gc=XCreateGC(dpy, root, GCFunction|GCFillStyle|GCTile, &gv); 
  temp=XCreatePixmap(dpy, root, TW, config.top.height, DefaultDepth(dpy, screen)); 
  XFillRectangle(dpy, temp, gc, 0, 0, TW, config.top.height); 
 
  if((c->is_e_m_r & IS_RESIZING)==IS_RESIZING) {
    if(c->name) {
      name_string=malloc(strlen(c->name)+20);
      sprintf(name_string,"%s [%dx%d]",c->name,c->width,c->height);
    } else {
      name_string=malloc(20);
      sprintf(name_string,"[%dx%d]",c->width,c->height);
    }
    c->is_e_m_r |= SHOWING_SIZE;
  } else if((c->is_e_m_r & IS_MOVING)==IS_MOVING) {
    if(c->name) {
      name_string=malloc(strlen(c->name)+20);
      sprintf(name_string,"%s (%d,%d)",c->name,FX,FY);
    } else {
      name_string=malloc(20);
      sprintf(name_string,"(%d,%d)",FX,FY);
    }
    c->is_e_m_r |= SHOWING_POSITION;
  } else if(c->name) {
    if(strlen(c->name)>0)
      name_string=strdup(c->name);
    else
      name_string=NULL;
    c->is_e_m_r &= NOT_SHOWING_SIZEPOS;
  }
   
  if(name_string) 
  { 
    int posx=0;
    int posy=0; 
    
    while(width==0) {
    
#ifdef I18N_SUPPORT
    if(use_fontset)
    {
      XmbTextExtents(config.fontset, name_string, strlen(name_string), &ink, &logical); 
      ascent=-logical.y; 
      descent=logical.height+logical.y; 
      width=logical.width; 
    }
    else
      width=XTextWidth(config.font, name_string, strlen(name_string));
#else  
    width=XTextWidth(config.font, name_string, strlen(name_string)); 
#endif /* I18N */ 
 
//    if( c->is_icon==NOT_ICONBAR ) {
      left_add=TLW;
      right_add=TRW;
//    }
    
    switch(config.text_align) 
    { 
    case ALIGNMENT_MIDDLE: 
//      posx=width>TW?0:(FW-width)/2; 
      posx=(FW-width)/2; 
      break; 
    case ALIGNMENT_RIGHT: 
//      posx=width>TW?0:FW-right_add-width; 
      posx=FW-right_add-width; 
      break; 
    case ALIGNMENT_LEFT: 
      posx=2+left_add; 
      break; 
    }
    
    leftext=left_add;
    rightext=FW-right_add;
    if(config.close.x>0) {
      if(config.close.x+config.close.width>=leftext)
        leftext=config.close.x+config.close.width-1;
    } else if(CX<rightext)
      rightext=CX;
    if(config.icon.x>0) {
      if(config.icon.x+config.icon.width>=leftext)
        leftext=config.icon.x+config.icon.width-1;
    } else if(IX<rightext)
      rightext=IX;
    if(config.zoom.x>0) {
      if(config.zoom.x+config.zoom.width>=leftext)
        leftext=config.zoom.x+config.zoom.width-1;
    } else if(HX<rightext)
      rightext=HX;
    
//    leftext=leftext-TLW;
//    rightext=rightext-TLW;
    
    // iconbars have text on the left - also, deal with case where there is not enough space
    if( c->is_icon!=NOT_ICONBAR )
       posx=2+left_add;
    else {
      if(posx>rightext-width)
        posx=rightext-width;
      if( posx<leftext )
        posx=leftext;
    }
        
    if(strlen(name_string)>12 && posx==leftext && posx+width>rightext) {
      if((c->is_e_m_r & IS_RESIZING)==IS_RESIZING) {
        free(name_string);
        name_string=malloc(20);
        sprintf(name_string,"[%dx%d]",c->width,c->height);
        width=0; // make it go back and check width again
      } else if((c->is_e_m_r & IS_MOVING)==IS_MOVING) {
        free(name_string);
        name_string=malloc(20);
        sprintf(name_string,"(%d,%d)",FX,FY);
        width=0; // make it go back and check width again
      }
    }
    
    } /* while(width) */
    
#ifdef I18N_SUPPORT
    if(use_fontset)
      posy=(config.top.height-descent)-((config.top.height-ascent-descent)/2);
    else
      posy=(config.top.height-config.font->descent)-((config.top.height-config.font->ascent-config.font->descent)/2);
#else 
    posy=(config.top.height-config.font->descent)-((config.top.height-config.font->ascent-config.font->descent)/2); 
#endif /* I18N */ 

    if( c->is_icon!=NOT_ICONBAR ) {
      if( c->icontxt_w==width ) {
        // text width has not changed, so leave position as it was
         posx=c->icontxt_x;
         posy=c->icontxt_y;
      } else {
         c->icontxt_x=posx;
         c->icontxt_y=posy;
         c->icontxt_w=width; }
    } else if( c->name_w==width && c->name_x==posx && c->name_y==posy ) {
        // text width has not changed, so leave position as it was
         posx=c->name_x;
         posy=c->name_y;
    } else {
         c->name_x=posx;
         c->name_y=posy;
         c->name_w=width;
    }

    posx -= left_add;
//    if(width+3<TW) {

//      posx -= TLW;
      
      if(config.title.height>0 && config.title.width>0) {
        if(config.title_left.width==0 || config.title_right.width==0 || config.title_left.height==0 || config.title_right.height==0) {
          XFreeGC(dpy, gc); 
          gv.function=GXcopy; 
          gv.tile=config.title.pixmap[state+shaded]; 
          gv.fill_style=FillTiled; 
          gc=XCreateGC(dpy, root, GCFunction|GCFillStyle|GCTile, &gv);
          XFillRectangle(dpy, temp, gc, posx-2, (config.top.height-config.title.height)/2, width+4, config.title.height);
        } else {
          XFreeGC(dpy, gc); 
          gv.function=GXcopy; 
          gv.tile=config.title_left.pixmap[state+shaded]; 
          gv.fill_style=FillTiled;
          gv.ts_x_origin=posx-config.title_left.width;
          gv.ts_y_origin=(config.top.height-config.title_left.height)/2;
          gc=XCreateGC(dpy, root, GCFunction|GCFillStyle|GCTile|GCTileStipXOrigin|GCTileStipYOrigin, &gv);
          XFillRectangle(dpy, temp, gc, gv.ts_x_origin, gv.ts_y_origin, config.title_left.width, config.title_left.height);          
          XFreeGC(dpy, gc); 
          gv.function=GXcopy; 
          gv.tile=config.title.pixmap[state+shaded]; 
          gv.fill_style=FillTiled; 
          gv.ts_x_origin=posx;
          gv.ts_y_origin=(config.top.height-config.title.height)/2;
          gc=XCreateGC(dpy, root, GCFunction|GCFillStyle|GCTile|GCTileStipXOrigin|GCTileStipYOrigin, &gv);
          XFillRectangle(dpy, temp, gc, gv.ts_x_origin, gv.ts_y_origin, width, config.title.height);          
          XFreeGC(dpy, gc); 
          gv.function=GXcopy; 
          gv.tile=config.title_right.pixmap[state+shaded]; 
          gv.fill_style=FillTiled; 
          gv.ts_x_origin=posx+width;
          gv.ts_y_origin=(config.top.height-config.title_right.height)/2;
          gc=XCreateGC(dpy, root, GCFunction|GCFillStyle|GCTile|GCTileStipXOrigin|GCTileStipYOrigin, &gv);
          XFillRectangle(dpy, temp, gc, gv.ts_x_origin, gv.ts_y_origin, config.title_right.width, config.title_right.height);          
        }
      }
    
#ifdef I18N_SUPPORT
      if(use_fontset)
        XmbDrawString(dpy, temp, config.fontset, config.font_gc[state], posx, posy, name_string, strlen(name_string));
      else
        XDrawString(dpy, temp, config.font_gc[state], posx, posy, name_string, strlen(name_string));
#else 
      XDrawString(dpy, temp, config.font_gc[state], posx, posy, name_string, strlen(name_string)); 
#endif /* I18N */ 

//    }
    
 // need to make sure dock icon's window list is up to date
//    if(c->osxwinfo.winref!=0 && mouse_mode!=MOUSE_MOVE && mouse_mode!=MOUSE_RESIZE)
    if(c->osxwinfo.winref!=0)
      CheckOSXWinTitle(&(c->osxwinfo),name_string,c->clientgroup,c->isoffscreen);
      
    free(name_string);
  }
  
  XSetWindowBackgroundPixmap(dpy, c->title, temp); 
  XClearWindow(dpy, c->title); 
 
  XFreeGC(dpy, gc); 
  XFreePixmap(dpy, temp); 

//  XFlush(dpy);
//  XSync(dpy,0);

  //if(c->is_dimmed==1)
  //  SendHackCommand(c,1,HACKCMD_DIMWIN,3,0,0,0,0); // switch on dimming again -no reshape nor redraw!

  if(wasinhibit!=c->inhibit_level)
    SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,wasinhibit,0,0,0,0);

//  if(wasalpha!=c->osxwinfo.drawalpha)
#ifdef DEBUG
  printf("wasalpha in title: %d\n",wasalpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,wasalpha,0,0,0,0);

  MyDecreaseXServerGrab();
} 
 
