/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 

#include "oroborus.h"
#include "iconbars.h"
#include "client_properties.h"

#define COMMENT_CHARS		"!#$@/"
#define NUM_COMMENT_CHARS	5

#define KEYWD_STARTRULE		"NewRule"
#define KEYWD_DESCRIPTION	"Description"
#define KEYWD_STARTMERGE	"StartMergeProperties"
#define KEYWD_STARTFINAL	"StartFinalProperties"
#define KEYWD_ENDRULES		"EndRules"
#define KEYWD_PARSEDEBUG	"ToggleVerboseParse"

#define XPROP_CLASS		"Class"
#define XPROP_NAME		"Name"
#define XPROP_TITLE		"Title"
#define XPROP_HOST		"Host"
#define XPROP_COMMAND		"Command"
#define XPROP_TRANSIENT		"IsTransient"
#define XPROP_RESIZABLE		"IsResizable"
#define XPROP_FRAMED		"IsFramed"
#define XPROP_CHILD		"IsChild"
#define XPROP_ICON		"IsIcon"
#define XPROP_MODAL		"IsModal"
#define XPROP_CANMAX		"CanZoom"

#define XPROPNUM_CLASS		0	// "Class"
#define XPROPNUM_NAME		1	// "Name"
#define XPROPNUM_TITLE		2	// "Title"
#define XPROPNUM_HOST		3	// "Host"
#define XPROPNUM_COMMAND	4	// "Command"
#define XPROPNUM_LASTSTRING	4	// This is the last X property that has a string value
#define XPROPNUM_TRANSIENT	5	// "IsTransient"
#define XPROPNUM_RESIZABLE	6	// "IsResizable"
#define XPROPNUM_FRAMED		7	// "IsFramed"
#define XPROPNUM_CHILD		8	// "IsChild"
#define XPROPNUM_ICON		9	// "IsIcon"
#define XPROPNUM_MODAL		10	// "IsModal"
#define XPROPNUM_CANMAX		11	// "CanZoom"
// Remember to change XPROPNUM_MAX in client_properties.h!!
// It should be one more than the maximum XPROPNUM value above
// Also, be sure to add new keywords into init_keywords()

#define MATCH_ANYTHING		0	//	= *
#define MATCH_EMPTY		1	//	= ""
#define MATCH_EXACTLY		2	//	= "..."
#define MATCH_STARTSWITH	3	//	= "..."*
#define MATCH_ENDSWITH		4	//	= *"..."
#define MATCH_CONTAINS		5	//	= *"..."*
#define MATCH_NOTEMPTY		6	//	~ ""
#define MATCH_NOTEXACTLY	7	//	~ "..."
#define MATCH_NOTSTARTSWITH	8	//	~ "..."*
#define MATCH_NOTENDSWITH	9	//	~ *"..."
#define MATCH_NOTCONTAINS	10	//	~ *"..."*

#define PROP_AUTOFOCUS		"AutoFocus"
#define PROP_AUTORAISE		"AutoRaise"
#define PROP_AUTOLOWER		"AutoLower"
#define PROP_FIRSTCLICK		"AcceptFirstClick"
#define PROP_CLICKRAISE		"RaiseOnClick"
#define PROP_CLICKDRAG          "DragOnFirstClick"
#define PROP_GRABFOCUS		"GrabFocus"
#define PROP_RETAINFOCUS	"RetainFocus"
#define PROP_MODAL		"ModalWindow"
#define PROP_POPUP		"PopUpWindow"
#define PROP_GROUPFLOATING	"GroupFloating"
#define PROP_GLOBALFLOATING	"GlobalFloating"
#define PROP_IGNOREFOCUS	"IgnoreFocus"
#define PROP_IGNORECLICKS	"IgnoreClicks"
#define PROP_FRAMES		"ShowFrames"
#define PROP_SHOWCLOSE		"ShowClose"
#define PROP_SHOWZOOM		"ShowZoom"
#define PROP_SHOWMIN		"ShowMin"
#define PROP_SHOWRESIZE		"ShowResize"
#define PROP_LIVERESIZE		"LiveResize"
#define PROP_TRANSLUCENT	"Translucency"
#define PROP_ADDTOWINMENU	"AddToWindowMenu"
#define PROP_FULLSCREEN		"FullScreen"
#define PROP_STARTICON		"InitAsIcon"
#define PROP_STARTSCREEN	"InitialScreen"
#define PROP_PLACEMENT		"Placement"
#define PROP_UNGROUPCLIENTS	"UngroupClients"
#define PROP_HIDEDOCKANDMENU	"HideDockAndMenus"
#define PROP_STARTGROUP		"InitialGroup"

#define PROPNUM_AUTOFOCUS	0	// "AutoFocus"
#define PROPNUM_AUTORAISE	1	// "AutoRaise"
#define PROPNUM_AUTOLOWER	2	// "AutoLower"
#define PROPNUM_FIRSTCLICK	3	// "AcceptFirstClick"
#define PROPNUM_CLICKRAISE	4	// "RaiseOnClick"
#define PROPNUM_CLICKDRAG       5       // "DragOnFirstClick"
#define PROPNUM_GRABFOCUS	6	// "GrabFocus"
#define PROPNUM_RETAINFOCUS	7	// "RetainFocus"
#define PROPNUM_MODAL		8	// "ModalWindow"
#define PROPNUM_POPUP		9	// "PopUpWindow"
#define PROPNUM_GROUPFLOATING	10	// "GroupFloating"
#define PROPNUM_GLOBALFLOATING	11	// "GlobalFloating"
#define PROPNUM_IGNOREFOCUS	12	// "IgnoreFocus"
#define PROPNUM_IGNORECLICKS	13	// "IgnoreClicks"
#define PROPNUM_FRAMES		14	// "ShowFrames"
#define PROPNUM_SHOWCLOSE	15	// "ShowClose"
#define PROPNUM_SHOWZOOM	16	// "ShowZoom"
#define PROPNUM_SHOWMIN		17	// "ShowMin"
#define PROPNUM_SHOWRESIZE	18	// "ShowResize"
#define PROPNUM_LIVERESIZE	19	// "LiveResize
#define PROPNUM_TRANSLUCENT	20	// "Translucency"
#define PROPNUM_ADDTOWINMENU	21	// "AddToWindowMenu"
#define PROPNUM_FULLSCREEN	22	// "FullScreen"
#define PROPNUM_STARTICON	23	// "InitAsIcon"
#define PROPNUM_STARTSCREEN	24	// "InitialScreen"
#define PROPNUM_PLACEMENT	25	// "Placement"
#define PROPNUM_UNGROUPCLIENTS	26	// "UngroupClients"
#define PROPNUM_HIDEDOCKANDMENU	27	// "HideDockAndMenus"
#define PROPNUM_LASTINT		27	// This is the last client property that is an integer/boolean
#define PROPNUM_STARTGROUP	28	// "InitialGroup"
// Remember to change PROPNUM_MAX in client_properties.h!!
// It should be one more than the maximum PROPNUM value above
// Also, be sure to add new keywords into init_keywords()

#define STARTSCREEN_CURSOR	-7
#define STARTSCREEN_BIGGEST	-6
#define STARTSCREEN_SMALLEST	-5
#define STARTSCREEN_RIGHTMOST	-4
#define STARTSCREEN_LEFTMOST	-3
#define STARTSCREEN_MIDDLE	-2
#define STARTSCREEN_PARENT	-1

// These can also be negative, in which case it overrides any placement hints given by the X11 app
#define PLACEMENT_NONE		0
#define PLACEMENT_MIDDLE	1
#define PLACEMENT_CURSOR	2
#define PLACEMENT_STAGGER	3
#define PLACEMENT_TILE		4
#define PLACEMENT_ALERT		5
#define PLACEMENT_CANSPAN	6

#define FIRSTCLICK_NEVER	 0
#define FIRSTCLICK_ALWAYS	 1
#define FIRSTCLICK_TOPMODALCHILD 2

#define TRANSLUCENT_NONE	 0
#define TRANSLUCENT_NORMAL	 1
#define TRANSLUCENT_VERY	 2

typedef struct 
{ 
    char *name; 
} MatchKWStruct; 

typedef struct 
{ 
    char *name; 
} PropKWStruct; 

char	*gMatchKeywords[XPROPNUM_MAX];
char	*gPropKeywords[PROPNUM_MAX];

void append_match_to_rule(Match_Rec *match, Rule_Rec *rule)
{
    if(rule==NULL) return;
    if(match==NULL) return;
    
    if(rule->lastmatch==NULL || rule->matchlist==NULL) {
        rule->matchlist = match;
        rule->lastmatch = match;
    } else {
        match->prev = rule->lastmatch;
        match->next = NULL;
        rule->lastmatch->next = match;
        rule->lastmatch = match;
    }
}

void append_prop_to_rule(Property_Rec *prop, Rule_Rec *rule)
{
    if(rule==NULL) return;
    if(prop==NULL) return;
    
    if(rule->lastprop==NULL || rule->proplist==NULL) {
        rule->proplist = prop;
        rule->lastprop = prop;
    } else {
        prop->prev = rule->lastprop;
        prop->next = NULL;
        rule->lastprop->next = prop;
        rule->lastprop = prop;
    }
}

void append_rule_to_list(Rule_Rec *rule, RulesList *rules)
{    
    if(rules==NULL) return;
    if(rule==NULL) return;
    
    // If there are no properties to set, it's not much of a rule!
    if(rule->proplist==NULL) return;
    
    rules->rulecount++;
    if(rules->rulecount<=1 || rules->startrule==NULL || rules->lastrule==NULL) {
        rules->rulecount = 1;
        rules->startrule = rule;
        rules->lastrule = rule;
        rule->prev = NULL;
        rule->next = NULL;
    } else {
        rule->prev = rules->lastrule;
        rule->next = NULL;
        rules->lastrule->next = rule;
        rules->lastrule = rule;
    }
}

void insert_rule_in_list(Rule_Rec *rule, Rule_Rec *after, RulesList *rules)
{
    if(rules==NULL) return;
    if(rule==NULL) return;
    
    // put at start of list if after is NULL
    if(after==NULL) {
        rule->next = rules->startrule;
        rule->prev = NULL;
        rules->startrule = rule;
    } else {
        rule->prev = after;
        rule->next = after->next;
        after->next = rule;
    }
    rules->rulecount++;
}

void remove_rule_from_list(Rule_Rec *rule, RulesList *rules)
{
    if(rules==NULL) return;
    if(rule==NULL) return;
    
    if(rule->next==NULL && rule->prev==NULL && rules->rulecount>1) return;
    
    if(rule->next) rule->next->prev = rule->prev;
    if(rule->prev) rule->prev->next = rule->next;
    
    if(rules->startrule==rule) rules->startrule = rule->next;
    if(rules->lastrule==rule) rules->lastrule = rule->prev;
    
    rules->rulecount--;
    
    rule->next = NULL;
    rule->prev = NULL;
}

void free_one_rule(Rule_Rec *rule)
{
    Match_Rec *match, *nextmatch;
    Property_Rec *prop, *nextprop;
    
    if(rule==NULL) return;
    
    //printf("Freeing name and description...\n");
    if(rule->name) free(rule->name);
    if(rule->description) free(rule->description);
    //printf("Done.\n");

    match = rule->matchlist;
    while(match) {
        nextmatch = match->next;
        //printf("Freeing next match...\n");
        if(match->matchstr) free(match->matchstr);
        free(match);
        //printf("Done.\n");
        match = nextmatch;
    }
    rule->matchlist = NULL;
    rule->lastmatch = NULL;
    rule->matchcount = 0;

    prop = rule->proplist;
    while(prop) {
        nextprop = prop->next;
        //printf("Freeing next property...\n");
        if(prop->propstr) free(prop->propstr);
        free(prop);
        //printf("Done.\n");
        prop = nextprop;
    }
    rule->proplist = NULL;
    rule->lastprop = NULL;
    rule->cont = 0;
    free(rule);
    //printf("Finished freeing rule.\n");
}

void clear_rules_list(RulesList *rules)
{
    Rule_Rec *rule, *next;
    
    if(rules==NULL) return;
    
    rule = rules->startrule;
    
    while(rule) {
        next = rule->next;
        free_one_rule(rule);
        rule = next;
    }
    
    rules->startrule = NULL;
    rules->lastrule = NULL;
    rules->rulecount = 0;
}

RulesList *new_rules_list()
{
    RulesList *rules=malloc(sizeof(RulesList));
    
    if(rules==NULL) return NULL;
    
    rules->rulecount = 0;
    rules->startrule = NULL;
    rules->lastrule = NULL;
    
    return rules;
}

void init_keywords()
{
    gMatchKeywords[XPROPNUM_CLASS] = strdup(XPROP_CLASS);
    gMatchKeywords[XPROPNUM_NAME] = strdup(XPROP_NAME);
    gMatchKeywords[XPROPNUM_TITLE] = strdup(XPROP_TITLE);
    gMatchKeywords[XPROPNUM_HOST] = strdup(XPROP_HOST);
    gMatchKeywords[XPROPNUM_COMMAND] = strdup(XPROP_COMMAND);
    gMatchKeywords[XPROPNUM_TRANSIENT] = strdup(XPROP_TRANSIENT);
    gMatchKeywords[XPROPNUM_RESIZABLE] = strdup(XPROP_RESIZABLE);
    gMatchKeywords[XPROPNUM_FRAMED] = strdup(XPROP_FRAMED);
    gMatchKeywords[XPROPNUM_CHILD] = strdup(XPROP_CHILD);
    gMatchKeywords[XPROPNUM_ICON] = strdup(XPROP_ICON);
    gMatchKeywords[XPROPNUM_MODAL] = strdup(XPROP_MODAL);
    gMatchKeywords[XPROPNUM_CANMAX] = strdup(XPROP_CANMAX);
    
    gPropKeywords[PROPNUM_AUTOFOCUS] = strdup(PROP_AUTOFOCUS);
    gPropKeywords[PROPNUM_AUTORAISE] = strdup(PROP_AUTORAISE);
    gPropKeywords[PROPNUM_AUTOLOWER] = strdup(PROP_AUTOLOWER);
    gPropKeywords[PROPNUM_FIRSTCLICK] = strdup(PROP_FIRSTCLICK);
    gPropKeywords[PROPNUM_CLICKRAISE] = strdup(PROP_CLICKRAISE);
    gPropKeywords[PROPNUM_CLICKDRAG] = strdup(PROP_CLICKDRAG);
    gPropKeywords[PROPNUM_GRABFOCUS] = strdup(PROP_GRABFOCUS);
    gPropKeywords[PROPNUM_RETAINFOCUS] = strdup(PROP_RETAINFOCUS);
    gPropKeywords[PROPNUM_MODAL] = strdup(PROP_MODAL);
    gPropKeywords[PROPNUM_POPUP] = strdup(PROP_POPUP);
    gPropKeywords[PROPNUM_GROUPFLOATING] = strdup(PROP_GROUPFLOATING);
    gPropKeywords[PROPNUM_GLOBALFLOATING] = strdup(PROP_GLOBALFLOATING);
    gPropKeywords[PROPNUM_IGNOREFOCUS] = strdup(PROP_IGNOREFOCUS);
    gPropKeywords[PROPNUM_IGNORECLICKS] = strdup(PROP_IGNORECLICKS);
    gPropKeywords[PROPNUM_FRAMES] = strdup(PROP_FRAMES);
    gPropKeywords[PROPNUM_SHOWCLOSE] = strdup(PROP_SHOWCLOSE);
    gPropKeywords[PROPNUM_SHOWZOOM] = strdup(PROP_SHOWZOOM);
    gPropKeywords[PROPNUM_SHOWMIN] = strdup(PROP_SHOWMIN);
    gPropKeywords[PROPNUM_SHOWRESIZE] = strdup(PROP_SHOWRESIZE);
    gPropKeywords[PROPNUM_LIVERESIZE] = strdup(PROP_LIVERESIZE);
    gPropKeywords[PROPNUM_TRANSLUCENT] = strdup(PROP_TRANSLUCENT);
    gPropKeywords[PROPNUM_ADDTOWINMENU] = strdup(PROP_ADDTOWINMENU);
    gPropKeywords[PROPNUM_FULLSCREEN] = strdup(PROP_FULLSCREEN);
    gPropKeywords[PROPNUM_STARTICON] = strdup(PROP_STARTICON);
    gPropKeywords[PROPNUM_STARTSCREEN] = strdup(PROP_STARTSCREEN);
    gPropKeywords[PROPNUM_PLACEMENT] = strdup(PROP_PLACEMENT);
    gPropKeywords[PROPNUM_UNGROUPCLIENTS] = strdup(PROP_UNGROUPCLIENTS);
    gPropKeywords[PROPNUM_HIDEDOCKANDMENU] = strdup(PROP_HIDEDOCKANDMENU);
    gPropKeywords[PROPNUM_STARTGROUP] = strdup(PROP_STARTGROUP);
}

void apply_client_properties(ClientInitProps *p, Property_Rec *proplist)
{
    Property_Rec *prop=proplist;
    if(proplist==NULL) return;
    if(p==NULL) return;
    
    while(prop) {
        if(prop->which >= 0 && prop->which < PROPNUM_MAX) {
            // Note that for a string property, the ClientInitProps is made to point to the same
            // string that prop->propstr points to - so we must NOT free this later on when we free
            // the ClientInitProps (the string remains until the rule containing this property is
            // freed - so we must also ensure that the rule stays around at least until after
            // this ClientInitProps has been used).
            if(prop->propstr) {
                (char *) p->props[prop->which] = prop->propstr;
                if(gVerboseConsole)
                    printf("Setting property '%s' to '%s'\n",
                                gPropKeywords[prop->which],prop->propstr);
            } else {
                p->props[prop->which] = prop->propval;
                if(gVerboseConsole)
                    printf("Setting property '%s' to %d\n",
                                gPropKeywords[prop->which],prop->propval);
            }
        }
        prop = prop->next;
    }
}

int check_client_match(ClientInitProps *p, Match_Rec *matchlist)
{
    Match_Rec *match=matchlist;
    if(matchlist==NULL) return 0;
    if(p==NULL) return 0;
    
    while(match) {
        if(match->which >= 0 && match->which <= XPROPNUM_LASTSTRING) {
            char *matchstr = match->matchstr;
            char *xpropstr = (char *)(p->xprops[match->which]);
            if(gVerboseConsole) printf("Match type %d : %s (%d) cf %s (%d)\n",
                    match->matchtype,matchstr,mystrlen(matchstr),xpropstr,mystrlen(matchstr));
            // For a string, there are several possible tests
            switch(match->matchtype) {
                case MATCH_ANYTHING:
                    // effectively ignores this property - anything matches
                    break;
                case MATCH_EMPTY:
                    // matches with NULL or empty string (length zero)
                    if(xpropstr!=NULL && strlen(xpropstr)>0) return 0;
                    break;
                case MATCH_EXACTLY:
                    // exact match can include both being NULL
                    if(matchstr==NULL && xpropstr!=NULL) return 0;
                    if(matchstr!=NULL && xpropstr==NULL) return 0;
                    if(matchstr!=NULL && xpropstr!=NULL) {
                        // must be exactly the same length, and characters all match
                        if(strlen(matchstr)!=strlen(xpropstr)) return 0;
                        if(strncmp(matchstr,xpropstr,strlen(matchstr))) return 0;
                    }
                    break;
                case MATCH_STARTSWITH:
                    // don't allow NULL to match
                    if(xpropstr==NULL || matchstr==NULL) return 0;
                    // length of matchstr can't be greater than length of xpropstr
                    if(strlen(matchstr)>strlen(xpropstr)) return 0;
                    // test the first strlen(matchstr) characters
                    if(strncmp(matchstr,xpropstr,strlen(matchstr))) return 0;
                    break;
                case MATCH_ENDSWITH:
                    // don't allow NULL to match
                    if(xpropstr==NULL || matchstr==NULL) return 0;
                    // length of matchstr can't be greater than length of xpropstr
                    if(strlen(matchstr)>strlen(xpropstr)) return 0;
                    // offset by difference...
                    xpropstr+=(strlen(xpropstr)-strlen(matchstr));
                    // so we can test the last strlen(matchstr) characters
                    if(strncmp(matchstr,xpropstr,strlen(matchstr))) return 0;
                    break;
                case MATCH_CONTAINS:
                    // don't allow NULL to match
                    if(xpropstr==NULL || matchstr==NULL) return 0;
                    // matchstr must be contained within xpropstr
                    if(strstr(xpropstr,matchstr)==NULL) return 0;
                    break;
                case MATCH_NOTEMPTY:
                    // NULL is empty
                    if(xpropstr==NULL) return 0;
                    // must have more than zero length
                    if(strlen(xpropstr)>0) return 0;
                    break;
                case MATCH_NOTEXACTLY:
                    // if they are both NULL, then they are considered identical
                    if(matchstr==NULL && xpropstr==NULL) return 0;
                    // but just one of them being NULL passes the test
                    if(matchstr!=NULL && xpropstr!=NULL) {
                        if(strlen(matchstr)==strlen(xpropstr))
                            if(!strncmp(matchstr,xpropstr,strlen(xpropstr))) return 0;
                    }
                    break;
                case MATCH_NOTSTARTSWITH:
                    if(matchstr==NULL) return 0;
                    if(xpropstr!=NULL && strlen(matchstr)<=strlen(xpropstr))
                        if(!strncmp(matchstr,xpropstr,strlen(matchstr))) return 0;
                    break;
                case MATCH_NOTENDSWITH:
                    if(matchstr==NULL) return 0;
                    if(xpropstr!=NULL && strlen(matchstr)<=strlen(xpropstr)) {
                        xpropstr+=(strlen(xpropstr)-strlen(matchstr));
                        if(!strncmp(matchstr,xpropstr,strlen(matchstr))) return 0;
                    }
                    break;
                case MATCH_NOTCONTAINS:
                    if(matchstr==NULL) return 0;
                    if(xpropstr!=NULL && strstr(xpropstr,matchstr)!=NULL) return 0;
                    break;
                default:
                    // don't understand...
                    printf("%s warning: got unknown matchtype (%d) in check_client_match\n",
                                                                    PROGNAME,match->matchtype);
                    break;
            }
        } else if(match->which < XPROPNUM_MAX) {
            if(gVerboseConsole) printf("Match %d cf %d\n",p->xprops[match->which],match->matchval);
            // This is straightforward matching - the match value is either zero or one,
            // and the client's corresponding xprops value must match.
            if(p->xprops[match->which]!=match->matchval) return 0;
        }
        match = match->next;
    }
    return 1;
}

Match_Rec *new_match()
{
    Match_Rec *newmatch = malloc(sizeof(Match_Rec));
    if(newmatch) {
        newmatch->which = -1;
        newmatch->matchval = 0;
        newmatch->matchstr = NULL;
        newmatch->matchtype = 0;
        newmatch->next = NULL;
        newmatch->prev = NULL;
    }
    return newmatch;
}

Property_Rec *new_property()
{
    Property_Rec *newprop = malloc(sizeof(Property_Rec));
    if(newprop) {
        newprop->which = -1;
        newprop->propval = 0;
        newprop->propstr = NULL;
        newprop->next = NULL;
        newprop->prev = NULL;
    }
    return newprop;
}

Rule_Rec *new_rule()
{
    Rule_Rec *newrule = malloc(sizeof(Rule_Rec));
    if(newrule) {
        newrule->name = NULL;
        newrule->description = NULL;
        newrule->matchlist = NULL;
        newrule->lastmatch = NULL;
        newrule->matchcount = 0;
        newrule->proplist = NULL;
        newrule->lastprop = NULL;
        newrule->cont = 1;
        newrule->next = NULL;
        newrule->prev = NULL;
    }
    return newrule;
}

void strip_left_spaces(char *str)
{
    int i=0, j=0, l, gotequals=0;
    if(str==NULL) return;
    l = strlen(str);
    
    while(i<l) {
        if(str[i]!=32) {
            if(str[i]=='=') gotequals = 1;
            if(str[i]=='~') gotequals = 1;
            str[j++] = str[i++];
        } else if(gotequals) {
            str[j++] = str[i++];
        } else
            i++;
    }
    // move termination to correct place
    str[l-i+j] = 0;
}

void strip_quotes(char *str)
{
    int l, i=0, j=0, k=-1;
    if(str==NULL) return;
    l = strlen(str);
    // check if this string does not have a quote character
    if(!strstr(str,"\"")) {
        // no quotes, so strip out spaces, starting from the beginning
        while(i<l) {
            if(str[i]!=32) break;
            i++;
        }
        // found first non-space, so start shifting characters, noting where last non-space is
        while(i<l) {
            str[j] = str[i];
            if(str[j]!=32) k = j;
            j++;
            i++;
        }
        // terminate string after last non-space
        str[k+1] = 0;
        return;
    }
    // look for the first quote
    while(i<l)
        if(str[i++]=='"') break;
    // shift back characters after first quote, noting where final quote is
    while(i<l) {
        str[j] = str[i];
        if(str[j]=='"') k = j;
        i++;
        j++;
    }
    // terminate on the final quote (if there was one)
    if(k>0)
        str[k] = 0;
    else
        str[j] = 0;
}

int parse_rightside_matchstring(char *str,int notequals)
{
    int l, i=0, j=0, result=-10;
    int firststar=-1, firstquote=-1, lastquote=-1, laststar=-1, firstnonspc=-1, lastnonspc=-1;
    
    if(str==NULL) return -1;
    
    //printf("Parsing rightside string: %s\n",str);
    
    l = strlen(str);
    while(i<l) {
        if(str[i]=='*') {
            if(firststar<0 && firstnonspc<0 && firstquote<0) {
                // This is the first non-space character, and it's an asterisk, and it's before a quote
                firststar=i;
            } else if(firststar>=0) {
                // had asterisk at start, so note this one as a possible last asterisk,
                // and make sure lastnonspc is up-to-date if there was a previous last asterisk
                if(laststar>lastnonspc) lastnonspc = laststar;
                laststar=i;
            } else {
                // got an asterisk, but was not one at start, so note this as possible last asterisk,
                // also make sure lastnonspc is up-to-date if there already has been one
                if(laststar>lastnonspc) lastnonspc = laststar;
                laststar=i;
            }
        } else if(str[i]=='"') {
            if(firstquote<0 && firstnonspc<0) {
                // this is the first non-space character (apart from asterisk), and it's a quote
                firstquote=i;
            } else if(firstquote>=0) {
                // got a quote at start, so note this one as a possible last quote,
                // as well as making sure lastnonspc is up-to-date if there already was a lastquote
                if(lastquote>lastnonspc) lastnonspc = lastquote;
                lastquote=i;
                // if we've had a previous potential last asterisk, remove it since it's actually within quotes
                if(laststar>=0) laststar = -1;
            } else {
                // had no quote at start, so assume this is part of string - i.e. a normal non-space character
                if(firstnonspc<0) firstnonspc = i;
                lastnonspc = i;
            }
        } else if(str[i]!=32) {
            // not a space, so update lastnonspc
            if(firstnonspc<0) firstnonspc=i;
            lastnonspc = i;
            // also note if this is after current last asterisk and/or quote
            if(lastnonspc>laststar) laststar = -1;
            if(lastnonspc>lastquote) lastquote = -1;
        }
        i++;
    }
    
    if(firstquote>=0 && lastquote<0) {
        // only had a single quote at the start - assume part of string with no quotes around it...
        firstnonspc = firstquote;
        firstquote = -1;
    }
    
    if(firstnonspc<0) {
        // had no string at all
        str[0] = 0;
        // so find out if it's empty (i.e. two quotes or nothing) or if it's just an asterisk
        if(firststar<0 && laststar<0) {
            if(notequals)
                return MATCH_NOTEMPTY;
            else
                return MATCH_EMPTY;
        } else {
            // notequals does not make sense here...
            return MATCH_ANYTHING;
        }
    } else if(firstquote<0 && lastquote<0) {
        // had no quotes, so make sure string lies from first non-space to last non-space
        firstquote = firstnonspc-1;
        lastquote = lastnonspc+1;
    }
    
    if(firststar<0 && laststar<0) {
        // had no asterisks, so exact match required
        if(notequals)
            result = MATCH_NOTEXACTLY;
        else
            result = MATCH_EXACTLY;
    } else if(firststar>=0 && laststar>=0) {
        // had an asterisk at start and end, so 'contains' match required
        if(notequals)
            result = MATCH_NOTCONTAINS;
        else
            result = MATCH_CONTAINS;
    } else if(firststar>=0) {
        // had an asterisk at start, so 'ends with' match required
        if(notequals)
            result = MATCH_NOTENDSWITH;
        else
            result = MATCH_ENDSWITH;
    } else if(laststar>=0) {
        // had an asterisk at end, so 'starts with' match required
        if(notequals)
            result = MATCH_NOTSTARTSWITH;
        else
            result = MATCH_STARTSWITH;
    }
    
    // correct the string so it only returns parts between quotes (or asterisks)
    j = 0;
    i = firstquote+1;
    while(i<lastquote)
        str[j++] = str[i++];

    // make sure it's terminated correctly
    str[j] = 0;
    
    return result;
}

int parse_rules_file(char *filename, RulesList *rules)
{
#define RULEMODE_NEEDRULE	0
#define RULEMODE_NEEDDESC	1
#define RULEMODE_NEEDMATCH	2
#define RULEMODE_NEEDPROPS	3

    int mode=0, kwlen, i, notequals=0, verbose=0;
    Rule_Rec	 *newrule=NULL;
    Match_Rec	 *match=NULL;
    Property_Rec *prop=NULL;
    FILE *file;
    char buf[1024], keyword[100], *rightside;
    
    if(rules==NULL) return -1;
    if(filename==NULL) return -2;
    
    if(!(file=fopen(filename,"r"))) {
        fprintf(stderr,"%s error: could not open rules file '%s'\n",PROGNAME,filename);
        return -3;
    }
    
    if(gMatchKeywords[0]==NULL) init_keywords();
    
    // loop through lines of file
    while(fgets(buf,sizeof buf,file)) {
    
        if(verbose) printf("LINE TO PARSE: %s",buf);
        strip_left_spaces(buf);
        if(buf==NULL) continue;
        
        kwlen = strlen(buf)-1;
        notequals = 0;
        rightside = strstr(buf,"=");
        if(rightside) {
            kwlen = rightside-buf;
            rightside++;
        } else {
            rightside = strstr(buf,"~");
            if(rightside) {
                kwlen = rightside-buf;
                rightside++;
                notequals = 1;
            }
        }
        
        // ignore short lines (in particular, lines that are blank!)
        // also, ignore lines that cant possibly have a keyword, because the first part is too long
        if(kwlen<2 || kwlen>98) continue;
        
        // copy the keyword from buffer into keyword area
        for(i=0;i<kwlen;i++)
            keyword[i] = buf[i];
        keyword[kwlen] = 0;
        // make sure right side is terminated before newline character
        if(rightside) {
            if(rightside[strlen(rightside)-1]=='\n')
                rightside[strlen(rightside)-1] = 0;
        }

        // Check for comments characters at start
        {
            char bufchar[2];
            bufchar[0] = buf[0];
            bufchar[1] = 0;
            if(strstr(COMMENT_CHARS,bufchar)) {
                //printf("Got comment line: %s\n",buf);
                continue;
            }
        }
        
        /*printf("Got keyword: %s  ",keyword);
        if(!rightside)
            printf("\n");
        else if(notequals)
            printf("~  rightside: %s\n",rightside);
        else
            printf("=  rightside: %s\n",rightside);*/
        
        if(verbose) printf("Comparing with end-rules keyword\n");        
        if(!strncasecmp(keyword,KEYWD_ENDRULES,kwlen)) {
            // we got the end of the rules file, so make sure pending rule is added to rules list
            if(newrule) {
                if(mode==RULEMODE_NEEDPROPS) {
                    if(verbose) printf("Appending final rule to rules list...\n");
                    append_rule_to_list(newrule,rules);
                    newrule = NULL;
                } else {
                    printf("%s warning: final rule ended without any properties -ignoring...\n",PROGNAME);
                    free_one_rule(newrule);
                    newrule = NULL;
                }
            }
            break;
        }

        if(verbose) printf("Comparing with debug keyword\n");        
        if(!strncasecmp(keyword,KEYWD_PARSEDEBUG,kwlen)) {
            verbose = 1-verbose;
            if(verbose)
                printf("========== VERBOSE PARSING ON ==========\n");
            else
                printf("========== VERBOSE PARSING OFF =========\n");
            continue;
        }
        
        if(verbose) printf("Comparing with start-rule keyword\n");        
        if(!strncasecmp(keyword,KEYWD_STARTRULE,kwlen)) {
            // we got the start of a new rule, so close off previous...
            if(newrule) {
                if(mode==RULEMODE_NEEDPROPS) {
                    if(verbose) printf("Starting new rule - appending previous to rules list...\n");
                    append_rule_to_list(newrule,rules);
                } else {
                    printf("%s warning: rule ended without any properties -ignoring...\n",PROGNAME);
                    free_one_rule(newrule);
                    newrule = NULL;
                }
            }
            // start a new one
            newrule = new_rule();
            if(!newrule) {
                fprintf(stderr,"%s error: could not create a new rule!\n",PROGNAME);
                return -4;
            }
            // and fill in the name
            if(rightside==NULL)
                newrule->name = strdup("Untitled");
            else {
                strip_quotes(rightside);
                newrule->name = strdup(rightside);
            }
            mode = RULEMODE_NEEDDESC;
            if(verbose) printf("Added rule name: '%s'\n",newrule->name);
            continue;
        }
        
        if(verbose) printf("Comparing with rule-description keyword\n");        
        if(!strncasecmp(keyword,KEYWD_DESCRIPTION,kwlen)) {
            if(mode!=RULEMODE_NEEDDESC) {
                printf("%s warning: got '%s' in wrong place - ignoring\n",PROGNAME,KEYWD_DESCRIPTION);
            } else {
                // we got the description
                if(rightside==NULL)
                    newrule->description = strdup("No description");
                else {
                    strip_quotes(rightside);
                    newrule->description = strdup(rightside);
                }
                mode = RULEMODE_NEEDMATCH;
                if(verbose) printf("Added rule description: '%s'\n",newrule->description);
            }
            continue;
        }
        
        if(verbose) printf("Comparing with property-start keywords\n");        
        // Check for start of properties
        if(!strncasecmp(keyword,KEYWD_STARTMERGE,kwlen)) {
            if(mode==RULEMODE_NEEDDESC) {
                printf("%s info: missing description for rule '%s'\n",PROGNAME,newrule->name);
                newrule->description = strdup("No description");
                mode = RULEMODE_NEEDMATCH;
            }
            if(mode!=RULEMODE_NEEDMATCH) {
                printf("%s warning: got '%s' in wrong place - ignoring\n",PROGNAME,KEYWD_STARTMERGE);
            } else {
                newrule->cont = 1;	// with merge
                mode = RULEMODE_NEEDPROPS;
                if(verbose) printf("Starting properties merge...\n");
            }
            continue;
        } else if(!strncasecmp(keyword,KEYWD_STARTFINAL,kwlen)) {
            if(mode!=RULEMODE_NEEDMATCH) {
                printf("%s warning: got '%s' in wrong place - ignoring\n",PROGNAME,KEYWD_STARTFINAL);
            } else {
                newrule->cont = 0;	// final properties for this match rule
                mode = RULEMODE_NEEDPROPS;
                if(verbose) printf("Starting final properties...\n");
            }
            continue;
        }
            
        if(mode==RULEMODE_NEEDMATCH) {
            // check for a match keyword
            int l=XPROPNUM_MAX,j=-1;
            if(verbose) printf("Comparing with match keyword list\n");        
            while(l>0) {
                l--;
                if(gMatchKeywords[l]==NULL) {
                    fprintf(stderr,"%s error: got null keyword no. %d !\n",PROGNAME,l);
                } else if(strlen(gMatchKeywords[l])==kwlen) {
                    if(!strncasecmp(gMatchKeywords[l],keyword,kwlen)) {
                        match=new_match();
                        if(match==NULL) return -10;
                        match->which = l;
                        if(l>XPROPNUM_LASTSTRING) {
                            if(rightside==NULL)
                                match->matchval = 1;	// consider NULL to mean "yes"
                            else {
                                // take away any quotes around right side string
                                if(verbose) printf("Stripping quotes: \n");
                                strip_quotes(rightside);
                                if(verbose) {
                                    if(rightside==NULL)
                                        printf("  ...got NULL!\n");
                                    else
                                        printf("  ...  '%s'\n",rightside);
                                }
                                // be fairly flexible in what can appear on the right side
                                if(rightside==NULL)
                                    match->matchval = 1;
                                else if(strlen(rightside)==1) {
                                    if(!strncasecmp(rightside,"y",1))
                                        match->matchval = 1;
                                    else if(!strncasecmp(rightside,"1",1))
                                        match->matchval = 1;
                                } else if(!strncasecmp(rightside,"on",2)) {
                                    // hmmmm - guess "on" is quite close to "no"...
                                    match->matchval = 1;
                                } else if(!strncasecmp(rightside,"yes",3))
                                    match->matchval = 1;
                                else // anything but "y", "yes", "on", or "1" is considered to be "no"
                                    match->matchval = 0;
                            }
                            if(verbose) printf("Got boolean match keyword: '%s'  value: %d\n",
                                                                keyword,match->matchval);
                            j = l;
                            l = 0;
                        } else if(match) {
                            match->matchtype = parse_rightside_matchstring(rightside,notequals);
                            if(match->matchtype<0) {
                                free(match);
                                match = NULL;
                                printf("%s warning: can't parse match line after keyword '%s'...\n",
                                                                                    PROGNAME,keyword);
                                printf("%s\n",buf);
                            } else {
                                // rightside has been stripped of leading/trailing quotes/stars
                                if(rightside[0]==0)
                                    match->matchstr = NULL;
                                else
                                    match->matchstr = strdup(rightside);
                                if(verbose) {
                                    printf("Got string match keyword: '%s'  type: ",keyword);
                                    switch(match->matchtype) {
                                        case MATCH_ANYTHING:
                                            printf("'is anything'");
                                            break;
                                        case MATCH_EMPTY:
                                            printf("'is empty'");
                                            break;
                                        case MATCH_EXACTLY:
                                            printf("'is exactly'");
                                            break;
                                        case MATCH_STARTSWITH:
                                            printf("'starts with'");
                                            break;
                                        case MATCH_ENDSWITH:
                                            printf("'ends with'");
                                            break;
                                        case MATCH_CONTAINS:
                                            printf("'contains'");
                                            break;
                                        case MATCH_NOTEMPTY:
                                            printf("'is not empty'");
                                            break;
                                        case MATCH_NOTEXACTLY:
                                            printf("'is not exactly'");
                                            break;
                                        case MATCH_NOTSTARTSWITH:
                                            printf("'does not start with'");
                                            break;
                                        case MATCH_NOTENDSWITH:
                                            printf("'does not end with'");
                                            break;
                                        case MATCH_NOTCONTAINS:
                                            printf("'does not contain'");
                                            break;
                                        default:
                                            printf("UNKNOWN!?! (%d)",match->matchtype);
                                            break;
                                    }
                                    if(match->matchstr)
                                        printf("  string: '%s'\n",match->matchstr);
                                    else
                                        printf("\n");
                                }
                            }
                        }
                        j = l;
                        l = 0;
                    }
                }
            }
            if(match) {
                if(verbose) printf("Appending match to rule...\n");
                append_match_to_rule(match,newrule);
                match = NULL;
            } else if(j==-1) {
                printf("%s warning: ignoring line with unknown match keyword: '%s'\n",PROGNAME,keyword);
                printf("%s\n",buf);
            }
        } else if(mode==RULEMODE_NEEDPROPS) {
            // check for a property keyword
            int l=PROPNUM_MAX,j=-1;
            if(verbose) printf("Comparing with properties keyword list\n");        
            while(l>0) {
                l--;
                if(gPropKeywords[l]==NULL) {
                    fprintf(stderr,"%s error: got NULL property keyword no. %d !\n",PROGNAME,l);
                } else if(strlen(gPropKeywords[l])==kwlen) {
                    if(!strncasecmp(gPropKeywords[l],keyword,kwlen)) {
                        prop=new_property();
                        if(prop==NULL) return -20;
                        prop->which = l;
                        if(rightside==NULL) {
                            free(prop);
                            prop = NULL;
                            printf("%s warning: could not parse value for property %s...\n",
                                                                                PROGNAME,keyword);
                            printf("%s\n",buf);
                        } else if(l>PROPNUM_LASTINT) {
                                if(verbose) printf("Stripping quotes: \n");
                                strip_quotes(rightside);
                                if(verbose) {
                                    if(rightside==NULL)
                                        printf("  ...got NULL!\n");
                                    else
                                        printf("  ...  '%s'\n",rightside);
                                }
//                            strip_quotes(rightside);
                            prop->propstr = strdup(rightside);
                            if(verbose) printf("Got property keyword: '%s'  string: '%s'\n",keyword,rightside);
                        } else {
                                if(verbose) printf("Stripping quotes: \n");
                                strip_quotes(rightside);
                                if(verbose) {
                                    if(rightside==NULL)
                                        printf("  ...got NULL!\n");
                                    else
                                        printf("  ...  '%s'\n",rightside);
                                }
//                            strip_quotes(rightside);
                            // allow various boolean strings on right side, besides an integer
                            if(rightside==NULL)
                                prop->propval = 1;
                            else if(!strncasecmp(rightside,"yes",3))
                                prop->propval = 1;
                            else if(!strncasecmp(rightside,"off",3))
                                prop->propval = 0;
                            else if(!strncasecmp(rightside,"on",2))
                                prop->propval = 1;
                            else if(!strncasecmp(rightside,"no",2))
                                prop->propval = 0;
                            else if(!strncasecmp(rightside,"y",1))
                                prop->propval = 1;
                            else if(!strncasecmp(rightside,"n",1))
                                prop->propval = 0;
                            else
                                prop->propval = atoi(rightside);
                            if(verbose) printf("Got property keyword: '%s'  value: %d\n",keyword,prop->propval);
                        }
                        j = l;
                        l = 0;
                    }
                }
            }
            if(prop) {
                if(verbose) printf("Appending property to rule...\n");
                append_prop_to_rule(prop,newrule);
                prop = NULL;
            } else if(j==-1) {
                printf("%s warning: ignoring line with unknown property keyword: %s\n",PROGNAME,keyword);
                printf("%s\n",buf);
            }
        }
    }
    
    if(newrule) {
        if(mode==RULEMODE_NEEDPROPS) {
            if(verbose) printf("Appending final rule to rules list...\n");
            append_rule_to_list(newrule,rules);
        } else {
            printf("%s warning: reached end of rules file, but last rule has no properties -ignoring\n",PROGNAME);
            free_one_rule(newrule);
        }
    }
    
    fclose(file);
    // anything else to do...?
    printf("%s info: parsed %d rules successfully from file %s\n",PROGNAME,rules->rulecount,filename);
    return 0;
}

void write_rules_file(char *filename, RulesList *rules)
{
    int mode=0;
    Rule_Rec *rule;
    
    if(rules==NULL) return;
    if(filename==NULL) return;
    
    // try to open the file
    //...
    rule = rules->startrule;
    while(rule) {
        Match_Rec *match = rule->matchlist;
        Property_Rec *prop = rule->proplist;
        // Write out rule starter line KEYWD_STARTRULE, with name
        //...
        // Write out rule description line KEYWD_DESCRIPTION
        //...
        // Go through matches
        while(match) {
            // Write this match line gMatchKeywords[match->which]
            // finish this...
            match = match->next;
        }
        // Write out new start properties line
        if(rule->cont)
            ;	//...KEYWD_STARTMERGE
        else
            ;	//...KEYWD_STARTFINAL
        // Go through properties
        while(prop) {
            // Write this property line gPropKeywords[prop->which]
            // finish this...
            prop = prop->next;
        }
        // finish this...
        rule = rule->next;
    }
    // Write out end-of-rules line KEYWD_ENDRULES
    // Close file
}

ClientInitProps *extract_win_props(Window win, int map_state)
{
    // Looks at client window's properties and sets up ClientInitProps xprops with
    // corresponding values. Also puts default props into ClientInitProps, so that
    // any not changed by new_client_properties are already set to sensible values.
    Window leadwin, tfwin, groupwin, clientwin;
    long dummy=0, readleft, nitems, *propret;
    int retval, argc=0;
    ClientInitProps *p=NULL;
    PropMwmHints *mhints=NULL; 
    XWMHints *hints=NULL; 
    XClassHint xch;
    XID tmpwin;
    Atom clientleader,clret;
    char **argv=NULL;
    XTextProperty txtprop;
    
    if(win) p = malloc(sizeof(ClientInitProps));

    if(p) {
        // set up initial values if we are initialising this client, otherwise
        // set to -999 so we know if anything has changed
        if(map_state==-999) {
            p->xprops[XPROPNUM_CLASS] = 0;
            p->xprops[XPROPNUM_NAME] = 0;
            p->xprops[XPROPNUM_TITLE] = 0;
            p->xprops[XPROPNUM_HOST] = 0;
            p->xprops[XPROPNUM_COMMAND] = 0;
            p->xprops[XPROPNUM_TRANSIENT] = 0;
            p->xprops[XPROPNUM_RESIZABLE] = 1;
            p->xprops[XPROPNUM_FRAMED] = 1;
            p->xprops[XPROPNUM_CHILD] = 0;
            p->xprops[XPROPNUM_ICON] = 0;
            p->xprops[XPROPNUM_MODAL] = 0;
            p->xprops[XPROPNUM_CANMAX] = 1;
            p->props[PROPNUM_AUTOFOCUS] = -999;
            p->props[PROPNUM_AUTORAISE] = -999;
            p->props[PROPNUM_AUTOLOWER] = -999;
            p->props[PROPNUM_FIRSTCLICK] = -999;
            p->props[PROPNUM_CLICKRAISE] = -999;
            p->props[PROPNUM_CLICKDRAG] = -999;
            p->props[PROPNUM_GRABFOCUS] = -999;
            p->props[PROPNUM_RETAINFOCUS] = -999;
            p->props[PROPNUM_MODAL] = -999;
            p->props[PROPNUM_POPUP] = -999;
            p->props[PROPNUM_GROUPFLOATING] = -999;
            p->props[PROPNUM_GLOBALFLOATING] = -999;
            p->props[PROPNUM_IGNOREFOCUS] = -999;
            p->props[PROPNUM_IGNORECLICKS] = -999;
            p->props[PROPNUM_FRAMES] = -999;
            p->props[PROPNUM_SHOWCLOSE] = -999;
            p->props[PROPNUM_SHOWMIN] = -999;
            p->props[PROPNUM_SHOWZOOM] = -999;
            p->props[PROPNUM_SHOWRESIZE] = -999;
            p->props[PROPNUM_LIVERESIZE] = -999;
            p->props[PROPNUM_TRANSLUCENT] = -999;
            p->props[PROPNUM_FULLSCREEN] = -999;
            p->props[PROPNUM_STARTICON] = -999;
            p->props[PROPNUM_ADDTOWINMENU] = -999;
            p->props[PROPNUM_STARTSCREEN] = -999;
            p->props[PROPNUM_PLACEMENT] = -999;
            p->props[PROPNUM_UNGROUPCLIENTS] = -999;
            p->props[PROPNUM_HIDEDOCKANDMENU] = -999;
            p->props[PROPNUM_STARTGROUP] = 0;
            p->parent_window = 0;
            p->parent_client = NULL;
        } else {
            p->xprops[XPROPNUM_CLASS] = 0;
            p->xprops[XPROPNUM_NAME] = 0;
            p->xprops[XPROPNUM_TITLE] = 0;
            p->xprops[XPROPNUM_HOST] = 0;
            p->xprops[XPROPNUM_COMMAND] = 0;
            p->xprops[XPROPNUM_TRANSIENT] = 0;
            p->xprops[XPROPNUM_RESIZABLE] = 1;
            p->xprops[XPROPNUM_FRAMED] = 1;
            p->xprops[XPROPNUM_CHILD] = 0;
            p->xprops[XPROPNUM_ICON] = 0;
            p->xprops[XPROPNUM_MODAL] = 0;
            p->xprops[XPROPNUM_CANMAX] = 1;
            p->props[PROPNUM_AUTOFOCUS] = 0;
            p->props[PROPNUM_AUTORAISE] = 0;
            p->props[PROPNUM_AUTOLOWER] = 0;
            p->props[PROPNUM_FIRSTCLICK] = 0;
            p->props[PROPNUM_CLICKRAISE] = 0;
            p->props[PROPNUM_CLICKDRAG] = 0;
            p->props[PROPNUM_GRABFOCUS] = 1;
            p->props[PROPNUM_RETAINFOCUS] = 0;
            p->props[PROPNUM_MODAL] = 0;
            p->props[PROPNUM_POPUP] = 0;
            p->props[PROPNUM_GROUPFLOATING] = 0;
            p->props[PROPNUM_GLOBALFLOATING] = 0;
            p->props[PROPNUM_IGNOREFOCUS] = 0;
            p->props[PROPNUM_IGNORECLICKS] = 0;
            p->props[PROPNUM_FRAMES] = 1;
            p->props[PROPNUM_SHOWCLOSE] = 1;
            p->props[PROPNUM_SHOWMIN] = 1;
            p->props[PROPNUM_SHOWZOOM] = 1;
            p->props[PROPNUM_SHOWRESIZE] = 1;
            p->props[PROPNUM_LIVERESIZE] = 1;
            p->props[PROPNUM_TRANSLUCENT] = 0;
            p->props[PROPNUM_FULLSCREEN] = 0;
            p->props[PROPNUM_STARTICON] = 0;
            p->props[PROPNUM_ADDTOWINMENU] = 1;
            p->props[PROPNUM_STARTSCREEN] = STARTSCREEN_CURSOR;
            p->props[PROPNUM_PLACEMENT] = PLACEMENT_MIDDLE;
            p->props[PROPNUM_UNGROUPCLIENTS] = 0;
            p->props[PROPNUM_HIDEDOCKANDMENU] = 0;
            p->props[PROPNUM_STARTGROUP] = 0;
            p->parent_window = 0;
            p->parent_client = NULL;
        }
        printf("Extracting X11 window properties...\n");
        if(get_wm_state(win)==IconicState) p->xprops[XPROPNUM_ICON] = 1;
        // check for WM_TRANSIENTFOR
        retval=XGetTransientForHint(dpy,win,&tfwin);
        if(retval && tfwin) {
            p->xprops[XPROPNUM_TRANSIENT] = 1;
            p->xprops[XPROPNUM_CHILD] = 1;
            // start by assuming it will be modal - mwm hints can change this...
            p->xprops[XPROPNUM_MODAL] = 1;
            p->parent_window = tfwin;
            if(gVerboseConsole) printf("TransientFor: %#lx\n",tfwin);
        } else {
            if(gVerboseConsole) printf("No TransientFor: %d (%#lx)\n",retval,tfwin);
        }
        
        // check for WM_CLIENT_LEADER
        leadwin=0;
        dummy=0;
        clientleader = XInternAtom(dpy,"WM_CLIENT_LEADER",True);
        if(clientleader!=None)
            dummy=XGetWindowProperty(dpy, win, clientleader, 0, 1, False, AnyPropertyType,
                                &clret, &retval, &nitems, &readleft, (unsigned char **) &propret);
        if(dummy) {
            //if(gVerboseConsole)
                printf("WM_CLIENT_LEADER atom: %#lx (%d %d %d %d %d)\n",
                                            propret,dummy,clret,retval,nitems,readleft);
            if(propret) {
                leadwin=(Window) *propret;
                XFree(propret);
                if(gVerboseConsole) printf("  value: %#lx\n",leadwin);
                // TransientFor will override WM_CLIENT_LEADER
                if(p->parent_window==0) p->parent_window = leadwin;
                p->xprops[XPROPNUM_CHILD] = 1;
            }
        } else //if(gVerboseConsole)
            printf("No WM_CLIENT_LEADER property\n");

        // Check for window_group from WM_HINTS
        hints=XGetWMHints(dpy, win); 
        if(hints) { 
            //if(gVerboseConsole) {
                printf("flags: %#lx\n",hints->flags);
                printf("input: %d\n",hints->input);
                printf("XID  : %#lx\n",hints->window_group);
                printf("state: %#lx\n",hints->initial_state);
            //}
            if(hints->flags & WindowGroupHint) {
                tmpwin=hints->window_group;
                // assuming this what the window_group is normally used for -i.e. it points to parent window ID
                groupwin = (Window) tmpwin;
                // TransientFor or WM_CLIENT_LEADER will override window_group
                if(p->parent_window==0) p->parent_window = groupwin;
                p->xprops[XPROPNUM_CHILD] = 1;
            }
            
            /* Sets the initial window state if set by client */ 
            if(map_state!=-999 && (hints->flags & StateHint)!=0 && map_state!=IsViewable) {
                //if(gVerboseConsole)
                    printf("Used initial state hints (%d)\n",hints->initial_state);
                set_wm_state(win, hints->initial_state); 
            }
    
            XFree(hints);
        } else {
// set initial client state
            if(map_state!=-999 && map_state!=IsViewable)
                set_wm_state(win, NormalState);
// Need to create a new OS X window, so name appears in dock icon's window list
            //MakeX11Window(&(c->osxwinfo),c->name,c->clientgroup);
// And bring OroborOSX/XDarwin in front so we don't have to go searching for the new window
            //c->osxwinfo.winref=-10;
        }
        
        /* Get MWM hints */ 
        mhints=get_mwm_hints(win); 
        if(mhints) { 
            //if(gVerboseConsole) {
                printf("---- MWM HINTS...\n");
                printf("flags      : %#lx\n",mhints->flags);
                printf("functions  : %#lx\n",mhints->functions);
                printf("decorations: %#lx\n",mhints->decorations);
                printf("inputMode  : %#lx\n",mhints->inputMode);
                printf("status     : %#lx\n",mhints->status);
                printf("DECOR_BORDER, DECOR_TITLE: %#lx , %#lx\n",MWM_DECOR_BORDER,MWM_DECOR_TITLE); 
            //}
            /*if(mhints->flags==0) {
                if(c->osxwinfo.winref==0)
                    c->osxwinfo.winref=-10;
                //MakeX11Window(&(c->osxwinfo),c->name,c->clientgroup);
                //c->osxwinref=MakeWindow(c->name,c->clientclass,0);
            }*/
            if((mhints->flags&MWM_HINTS_DECORATIONS)!=0 && (mhints->decorations&MWM_DECOR_ALL)==0) { 
                /* Window has no border if *neither* MWM_DECOR_BORDER *nor* MWM_DECOR_TITLE are set */ 
                if((mhints->decorations&(MWM_DECOR_BORDER|MWM_DECOR_TITLE))==0)
                    p->xprops[XPROPNUM_FRAMED] = 0;
                /* Set if client can be maximized or not */ 
                if((mhints->decorations&MWM_DECOR_MAXIMIZE)==0) 
                    p->xprops[XPROPNUM_CANMAX] = 0;
                /* Set if client can be resized or not */ 
                if((mhints->decorations&MWM_DECOR_RESIZEH)==0)
                    p->xprops[XPROPNUM_RESIZABLE] = 0;
            }
            p->xprops[XPROPNUM_MODAL] = 0;
            // make a note of modal windows...
            if((mhints->flags&MWM_HINTS_INPUT_MODE)!=0) {
                if((mhints->inputMode&MWM_INPUT_PRIMARY_APPLICATION_MODAL)!=0) {
                    p->xprops[XPROPNUM_MODAL] = 1;
                    /*c->osxwinfo.winref=-1;
                    c->modal_state|=APP_MODAL;
                    if(c->parent)
                        (c->parent)->modal_state|=HAS_APP_MODAL_CHILD;*/
                }
                if((mhints->inputMode&MWM_INPUT_SYSTEM_MODAL)!=0) {
                    p->xprops[XPROPNUM_MODAL] = 1;
                    /*c->osxwinfo.winref=-1;
                    c->modal_state|=SYS_MODAL;
                    if(c->parent)
                        (c->parent)->modal_state|=HAS_SYS_MODAL_CHILD;*/
                }
            } 
            /*if(c->osxwinfo.winref==0) {
                // create OS X window, and bring OroborOSX/XDarwin in front
                //MakeX11Window(&(c->osxwinfo),c->name,c->clientgroup);
                //c->osxwinref=MakeWindow(c->name,c->clientclass,0);
                c->osxwinfo.winref=-10;
            }
            XFree(mhints);*/
        }

        // Set the window title property
#ifdef I18N_SUPPORT 
        if(use_fontset)
            /* call wrapper function for I18N */ 
            XFetchNameWrapper(dpy, win, (char **) &(p->xprops[XPROPNUM_TITLE])); 
        else
            XFetchName(dpy, win,(char **) &(p->xprops[XPROPNUM_TITLE])); 
#else 
        XFetchName(dpy, win, (char **) &(p->xprops[XPROPNUM_TITLE])); 
#endif /* I18N */
        //if(gVerboseConsole) {
            if(p->xprops[XPROPNUM_TITLE])
                printf("  title: %s\n",(char *)(p->xprops[XPROPNUM_TITLE]));
            else
                printf("  No title for window!\n");
        //}
        
        // Set the window's class and name properties
        clientwin = p->parent_window;
        if(clientwin) {
            Client *cp = get_client_from_window(clientwin,WINDOW_LE);
            if(cp) {
                // inherit from parent client
                //if(gVerboseConsole)
                    printf("  group from parent client: %s\n",cp->clientgroup);
                (char *)(p->xprops[XPROPNUM_CLASS])=strdup(cp->clientgroup);
                p->parent_client = cp;
            } else
                p->parent_client = NULL;
        } else {
            p->parent_client = NULL;
            clientwin = win;
        }
        
        retval=XGetClassHint(dpy,clientwin,&xch);
        if(retval!=0) {
            //if(gVerboseConsole) {
                printf("  res_class: %s\n",xch.res_class);
                printf("  res_name: %s\n",xch.res_name);
            //}
            if(p->xprops[XPROPNUM_CLASS]==0) (char *)(p->xprops[XPROPNUM_CLASS])=strdup(xch.res_class);
            (char *)(p->xprops[XPROPNUM_NAME])=strdup(xch.res_name);
            if(xch.res_class) XFree(xch.res_class);
            if(xch.res_name) XFree(xch.res_name);
        } else {
            //if(gVerboseConsole)
                printf("  No class and name!\n");
            // Search for an existing managed window which belongs to the same X11 client application
            // Found one, so inherit its group
            // none found, so search for any window which belongs to the same X11 client app
            // Found one, but it also has no class, so continue search
            // Found one which has a class, so inherit it
        }

        // Set the window's originating host property
        retval = XGetWMClientMachine(dpy,win,&txtprop);
        //if(gVerboseConsole)
            printf("  ret,value,nitems: %d %#xl %d\n",retval,txtprop.value,txtprop.nitems);
        if(retval && txtprop.value) {
            //if(gVerboseConsole)
                printf("  host: %s\n",txtprop.value);
            (char *)(p->xprops[XPROPNUM_HOST]) = strdup(txtprop.value);
            XFree(txtprop.value);
        }
            
        // Set the window's command property
        retval=XGetCommand(dpy,win,&argv,&argc);
        //if(gVerboseConsole)
            printf("  ret,argc: %d %d\n",retval,argc);
        if(argv) {
            //if(gVerboseConsole)
                printf("  argv: %s\n",*argv);
            (char *) p->xprops[XPROPNUM_COMMAND] = strdup(*argv);
            XFreeStringList(argv);
        } else //if(gVerboseConsole)
            printf("  No command!\n");
        
    }
    
        printf("Done!\n");
    return p;
}

void new_client_properties(Client *c, RulesList *rules, int map_state, int reposition)
{
    /* Note: this subroutine assumes c->x,c->y,c->width.c->height have already been
     * set up with the window position hints... */
    ClientInitProps *p=NULL;
    int placement=0;
    Rule_Rec *rule;
    if(rules==NULL) return;
    if(c==NULL) return;
    
    p = extract_win_props(c->window, map_state);
    if(p==NULL) {
        fprintf(stderr,"%s error: could not allocate ClientInitProps!",PROGNAME);
        return;
    }
    
    rule = rules->startrule;
    while(rule) {
        int ruletype, carryon;
        Property_Rec *props;
        if(gVerboseConsole) printf("Checking next rule...\n");
        ruletype = check_client_match(p,rule->matchlist);
        props = rule->proplist;
        carryon = rule->cont;
        rule = rule->next;
        if(ruletype) {
            if(gVerboseConsole) {
                printf("Applying properties (%d)",ruletype);
            	if(carryon==0)
                    printf(" and terminating\n");
                else
                    printf("\n");
            }
            apply_client_properties(p,props);
            if(carryon==0)	// 0 means we got "StartFinalProperties",
                rule = NULL;	// so no more rules should be checked
	} else if(gVerboseConsole)
            printf("Match failed.\n");
    }
    
    // apply properties in ClientInitProps to client's data structure

    c->parent = p->parent_client;

    if( p->props[PROPNUM_SHOWZOOM]==1 )
        c->can_max = 1;
    else if( p->props[PROPNUM_SHOWZOOM]==0 )
        c->can_max = 0;
        
    if( p->props[PROPNUM_AUTOFOCUS]==1 )
        c->auto_properties |= AUTO_FOCUS;
    else if( p->props[PROPNUM_AUTOFOCUS]==0 && (c->auto_properties&AUTO_FOCUS)!=0 )
        c->auto_properties ^= AUTO_FOCUS;
        
    if( p->props[PROPNUM_AUTORAISE]==1 )
        c->auto_properties |= AUTO_RAISE;
    else if( p->props[PROPNUM_AUTORAISE]==0 && (c->auto_properties&AUTO_RAISE)!=0 )
        c->auto_properties ^= AUTO_RAISE;
        
    if( p->props[PROPNUM_AUTOLOWER]==1 )
        c->auto_properties |= AUTO_LOWER;
    else if( p->props[PROPNUM_AUTOLOWER]==0 && (c->auto_properties&AUTO_LOWER)!=0 )
        c->auto_properties ^= AUTO_LOWER;
        
    if( p->props[PROPNUM_FIRSTCLICK]==FIRSTCLICK_ALWAYS )
        c->auto_properties |= NONAUTO_FIRSTCLICK;
    else if( p->props[PROPNUM_FIRSTCLICK]==FIRSTCLICK_TOPMODALCHILD )
        c->auto_properties |= NONAUTO_FIRSTCLICKCHILD;
    
    if( p->props[PROPNUM_FIRSTCLICK]==FIRSTCLICK_NEVER && (c->auto_properties&NONAUTO_FIRSTCLICKCHILD)!=0 )
        c->auto_properties ^= NONAUTO_FIRSTCLICKCHILD;
    if( p->props[PROPNUM_FIRSTCLICK]==FIRSTCLICK_NEVER && (c->auto_properties&NONAUTO_FIRSTCLICK)!=0 )
        c->auto_properties ^= NONAUTO_FIRSTCLICK;
        
    if( p->props[PROPNUM_CLICKRAISE]==1 )
        c->auto_properties |= AUTO_CLICKRAISE;
    else if( p->props[PROPNUM_CLICKRAISE]==0 && (c->auto_properties&AUTO_CLICKRAISE)!=0 )
        c->auto_properties ^= AUTO_CLICKRAISE;
        
    if( p->props[PROPNUM_CLICKDRAG]==1 )
        c->auto_properties |= NONAUTO_DRAGFIRSTCLICK;
    else if( p->props[PROPNUM_CLICKDRAG]==0 && (c->auto_properties&NONAUTO_DRAGFIRSTCLICK)!=0 )
        c->auto_properties ^= NONAUTO_DRAGFIRSTCLICK;
        
    if( p->props[PROPNUM_GRABFOCUS]==0 )
        c->auto_properties |= NO_GRAB_FOCUS;
    else if( p->props[PROPNUM_GRABFOCUS]==1 && (c->auto_properties&NO_GRAB_FOCUS)!=0 )
        c->auto_properties ^= NO_GRAB_FOCUS;
        
    if( p->props[PROPNUM_RETAINFOCUS]==1 )
        c->auto_properties |= RETAIN_FOCUS;
    else if( p->props[PROPNUM_RETAINFOCUS]==0 && (c->auto_properties&RETAIN_FOCUS)!=0 )
        c->auto_properties ^= RETAIN_FOCUS;
        
    if( p->props[PROPNUM_IGNORECLICKS]==1 )
        c->auto_properties |= NONAUTO_IGNORECLICKS;
    else if( p->props[PROPNUM_IGNORECLICKS]==0 && (c->auto_properties&NONAUTO_IGNORECLICKS)!=0 )
        c->auto_properties ^= NONAUTO_IGNORECLICKS;
        
    if( p->props[PROPNUM_IGNORECLICKS]==1 )
        c->auto_properties |= NONAUTO_IGNOREFOCUS;
    else if( p->props[PROPNUM_IGNORECLICKS]==0 && (c->auto_properties&NONAUTO_IGNOREFOCUS)!=0 )
        c->auto_properties ^= NONAUTO_IGNOREFOCUS;
        
    if( p->props[PROPNUM_MODAL]==1 ) {
        c->modal_state |= (APP_MODAL|DONT_IGNORE_MODAL);
        if(c->parent)
            (c->parent)->modal_state|=HAS_APP_MODAL_CHILD;
    } else if( p->props[PROPNUM_MODAL]==0 &&
            (c->modal_state&STRICT_ANY_MODAL)!=0 && (c->modal_state&DONT_IGNORE_MODAL)!=0 ) {
        c->modal_state ^= DONT_IGNORE_MODAL;
        if(c->parent)
            (c->parent)->modal_state|=HAS_APP_MODAL_CHILD;
    }
        
    if( p->props[PROPNUM_FRAMES]==0 )
        c->has_frame = 0;
    else if( p->props[PROPNUM_FRAMES]==1 )
        c->has_frame = 1;
        
    if( p->props[PROPNUM_LIVERESIZE]==1 )
        c->auto_properties |= LIVE_RESIZING;
    else if( p->props[PROPNUM_LIVERESIZE]==0 && (c->auto_properties&LIVE_RESIZING)!=0 )
        c->auto_properties ^= LIVE_RESIZING;
    
    if( p->props[PROPNUM_TRANSLUCENT]==TRANSLUCENT_NONE ) {
        c->osxwinfo.translucency=ACTIVE_TRANSLUCENT0;
        c->osxwinfo.inactive_translucency=INACTIVE_TRANSLUCENT0;
    } else if( p->props[PROPNUM_TRANSLUCENT]==TRANSLUCENT_NORMAL ) {
        c->osxwinfo.translucency=ACTIVE_TRANSLUCENT1;
        c->osxwinfo.inactive_translucency=INACTIVE_TRANSLUCENT1;
    } else if( p->props[PROPNUM_TRANSLUCENT]==TRANSLUCENT_VERY ) {
        c->osxwinfo.translucency=ACTIVE_TRANSLUCENT2;
        c->osxwinfo.inactive_translucency=INACTIVE_TRANSLUCENT2;
    }
        
    // if we are initialising this client, then set up client structure and placement appropriately
    if( map_state!=-999 ) {
        c->clientgrouporg = c->clientgroup;
        // Point to existing strings in xprops for the rest - these are not freed with ClientInitProps
        c->clientclass = (char *)p->xprops[XPROPNUM_CLASS];
        c->clientname = (char *)p->xprops[XPROPNUM_NAME];
        c->name = (char *)p->xprops[XPROPNUM_TITLE];    // *FIXME* - need to deal with null title
        c->clienthost = (char *)p->xprops[XPROPNUM_HOST];
        c->clientcmd = (char *)p->xprops[XPROPNUM_COMMAND];
        // Need to copy the group name, since original is either in rules list (STARTGROUP) or
        // the class name (used by c->clientclass) and will be freed independently at some point
        if(p->props[PROPNUM_STARTGROUP]==0) {
            if( c->clientclass )
                c->clientgroup = strdup(c->clientclass);
            else
                c->clientgroup = NULL;
        } else
            c->clientgroup = strdup((char *)p->props[PROPNUM_STARTGROUP]);
        // special flags to tell create_new_client various things...
        c->dock_timer = 0;
        if(p->props[PROPNUM_UNGROUPCLIENTS]==1) c->dock_timer += 8;
        if(p->props[PROPNUM_HIDEDOCKANDMENU]==1) c->dock_timer += 4;
        if(p->props[PROPNUM_ADDTOWINMENU]==1) c->dock_timer += 2;
        if(p->props[PROPNUM_STARTICON]==1) c->dock_timer += 1;

  placement = p->props[PROPNUM_PLACEMENT];
  printf("Got placement: %d\n",placement);
    
  /* Move the client to position set by hints  -changed this to work more like twm 
   * note that placement<0 allows override of app's placement hints */
  if( placement>=0 && ((c->size->flags & USPosition)!=0 ||
        ((c->size->flags & PPosition)!=0 && (c->x>0 || c->y>0)) || !reposition)) 
  {
    int head=0, x=0, y=0, xmin, xmax, ymin, ymax;
    //if(gVerboseConsole)
        printf("Found hints for geometry...   %d x %d  @ %d , %d\n",c->width,c->height,c->x,c->y);
#ifdef XINERAMA_SUPPORT
    if(xinerama)
    {
      update_xinerama_screens();
      if((c->modal_state & STRICT_ANY_MODAL)==0 || c->parent==NULL)
        head=-1;
      else
        head=find_rect_head(c->parent->x,c->parent->y,c->parent->width,c->parent->height);
      if( placement == PLACEMENT_CANSPAN ) {
        x = c->x;
        y = c->y;
      } else
        check_window_pos(c->x,c->y,&c->width,&c->height,&x,&y,&head);
      /* If window positioned within 15 pixels of center of display, consider it to be centered on the current head */
/*      if(abs((c->x+(c->width/2))-(DisplayWidth(dpy, screen)/2))<15 && abs((c->y+(c->height/2))-(DisplayHeight(dpy, screen)/2))<15)
      {
        get_mouse_xy(&x, &y);
        head=find_head(x, y);
        c->x=screen_x(head)+(screen_width(head)/2)-(c->width/2);
        c->y=screen_y(head)+(screen_height(head)/2)-(c->height/2);
        printf("XINERAMA: Found cursor on head %d... set position to %d,%d\n",head,c->x,c->y);
      }*/
      c->x = x;
      c->y = y;
      //head=find_head(c->x, c->y);
      if(gVerboseConsole) printf("XINERAMA: Window placed on head %d\n",head);
    }
#endif /* XINERAMA */
 
    if( placement==PLACEMENT_CANSPAN ) {
      xmin = screen_x(0);
      ymin = screen_y(0);
      xmax = xmin+screen_width(0);
      xmin = ymin+screen_height(0);
      head = 1;
      while(head<xinerama_heads) {
        int xx = screen_x(head);
        int yy = screen_y(head);
        if( xx < xmin ) xmin = xx;
        if( yy < ymin ) ymin = yy;
        xx += screen_width(head);
        yy += screen_height(head);
        if( xx > xmax ) xmax = xx;
        if( yy > ymax ) ymax = yy;
        head++;
      }
      head = 0;
    } else {
      xmin=screen_x(head)+LW;
      ymin=screen_y(head)+TH;
      if( screen_y(head)==0 && gMenubarSpace!=0 ) ymin+=22;
      xmax=(screen_x(head)+screen_width(head))-RW; 
      ymax=(screen_y(head)+screen_height(head))-BH; 
    }
    //if(gVerboseConsole)
      printf("Got bounds (min/max): %d/%d , %d/%d\n",xmin,xmax,ymin,ymax);
    gravitate_client(c, APPLY_GRAVITY); 
    //if(gVerboseConsole)
        printf("Placing via hints: %d x %d @ %d,%d\n",c->width,c->height,c->x,c->y);
 
    /* Keep window fully inside desktop */ 
    if(c->width>xmax-xmin) {
        c->x=xmin;
        c->width=xmax-xmin;
    } else if(c->x+c->width>xmax) {
        c->x=xmax-c->width;
    } else if(c->x<xmin)
        c->x=xmin;
    if(c->height>ymax-ymin) {
        c->y=ymin;
        c->height=ymax-ymin;
    } else if(c->y+c->height>ymax) {
        c->y=ymax-c->height;
    } else if(c->y<ymin)
        c->y=ymin;
  } 
 
  /* Otherwise follow window placement policy */ 
  else 
  { 
    int head=0, x=0, y=0, xmin, xmax, ymin, ymax;
    if( placement < 0 ) placement = -placement;	// negative ones used to override X11 app's placement hints

#ifdef XINERAMA_SUPPORT
    if(xinerama)
    {
      head=-1;
      // Parent screen placement policy is followed, provided there is a parent!
      if(p->props[PROPNUM_STARTSCREEN]==STARTSCREEN_PARENT && c->parent)
        head=find_rect_head(c->parent->x,c->parent->y,c->parent->width,c->parent->height);
      if(head>=0) {
        //if(gVerboseConsole)
          printf("XINERAMA: Found parent on head %d...\n",head);
      } else if( p->props[PROPNUM_STARTSCREEN]>=0 ) {
        head = p->props[PROPNUM_STARTSCREEN];
        if( head >= xinerama_heads ) head = xinerama_heads-1;
          printf("XINERAMA: Placing on head %d...\n",head);
      } else {
        // *FIXME* - other screen placement policies are not followed yet...
        get_mouse_xy(&x, &y);
        head=find_head(x, y);
        //if(gVerboseConsole)
            printf("XINERAMA: Found cursor (%d,%d) on head %d...\n",x,y,head);
      }
    }
#endif /* XINERAMA */

    xmin=screen_x(head)+config.left_margin+LW;
    ymin=screen_y(head)+config.top_margin+TH;
    if(screen_y(head)==0 && gMenubarSpace!=0) ymin+=22;
    xmax=(screen_x(head)+screen_width(head))-config.right_margin-RW; 
    ymax=(screen_y(head)+screen_height(head))-config.bottom_margin-BH; 

    //if(gVerboseConsole)
      printf("Got bounds (min/max): %d/%d , %d/%d\n",xmin,xmax,ymin,ymax);

    if(placement==PLACEMENT_CURSOR) {
      get_mouse_xy(&x, &y); 
      //if(gVerboseConsole)
        printf("Placing with cursor: %d x %d @ %d,%d\n",c->width,c->height,x,y);
    } else if(placement==PLACEMENT_ALERT) {
      x=screen_x(head)+(screen_width(head)/2); 
      y=screen_y(head)+(screen_height(head)/3);
      //if(gVerboseConsole)
        printf("Placing as alert on head %d: %d x %d @ %d,%d\n",head,c->width,c->height,x,y);
    } else if(placement==PLACEMENT_NONE) {
      x = c->x + c->width/2;
      y = c->y + c->height/2;
        printf("Placing on head %d in original position: %d x %d @ %d,%d\n",head,c->width,c->height,c->x,c->y);
    } else {
      // *FIXME* - other placement policies are not followed yet
      x=screen_x(head)+(screen_width(head)/2); 
      y=screen_y(head)+(screen_height(head)/2);
      //if(gVerboseConsole)
        printf("Placing in center of head %d: %d x %d @ %d,%d\n",head,c->width,c->height,x,y);
    }
    c->x=x-(c->width/2); 
    c->y=y-(c->height/2); 
 
    /* Keep window fully inside desktop */ 
    if(c->width>xmax-xmin) {
        c->x=xmin;
        c->width=xmax-xmin;
    } else if(c->x+c->width>xmax) {
        c->x=xmax-c->width;
    } else if(c->x<xmin)
        c->x=xmin;
    if(c->height>ymax-ymin) {
        c->y=ymin;
        c->height=ymax-ymin;
    } else if(c->y+c->height>ymax) {
        c->y=ymax-c->height;
    } else if(c->y<ymin)
        c->y=ymin;
  } 
  
    }
        
    // finish this...
    
    // Don't need ClientInitProps any more - note that the strings referenced by the
    // xprops and props are still in use by the ruleslist and/or client, so don't free them
    free(p);
    p = NULL;
    printf("Finishing new_client_properties... %d x %d  @ %d , %d\n",c->width,c->height,c->x,c->y);
}

