/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 
 
#ifndef OROB_CLIENT_PROPS_H 
#define OROB_CLIENT_PROPS_H 

#include "oroborus.h" 

typedef struct _Property_Rec Property_Rec;

struct _Property_Rec
{
    int		 which;
    int		 propval;
    char	 *propstr;
    Property_Rec *next;
    Property_Rec *prev;
};

typedef struct _Match_Rec Match_Rec;

struct _Match_Rec
{
    int		which;
    int		matchval;
    char	*matchstr;
    int		matchtype;
    Match_Rec	*next;
    Match_Rec	*prev;
};

typedef struct _Rule_Rec Rule_Rec;

struct _Rule_Rec
{
    char		*name;
    char		*description;
    Match_Rec		*matchlist;
    Match_Rec		*lastmatch;
    int			matchcount;
    Property_Rec	*proplist;
    Property_Rec	*lastprop;
    int			cont;
    Rule_Rec		*next;
    Rule_Rec		*prev;
};

typedef struct
{
    int		rulecount;
    Rule_Rec	*startrule;
    Rule_Rec	*lastrule;
} RulesList;

#define XPROPNUM_MAX		12
#define PROPNUM_MAX		29

typedef struct
{
    int xprops[XPROPNUM_MAX];
    int props[PROPNUM_MAX];
    Window parent_window;
    Client *parent_client;
} ClientInitProps;

extern RulesList *gMainRules;

extern RulesList *new_rules_list();
extern void clear_rules_list(RulesList *rules);
extern void add_rule_to_list(Rule_Rec *rule, RulesList *rules);
extern void append_rule_to_list(Rule_Rec *rule, RulesList *rules);
extern void insert_rule_in_list(Rule_Rec *rule, Rule_Rec *after, RulesList *rules);
extern void remove_rule_from_list(Rule_Rec *rule, RulesList *rules);
extern void init_keywords();
extern int parse_rules_file(char *filename, RulesList *rules);
extern ClientInitProps *extract_win_props(Window win, int map_state);
extern void new_client_properties(Client *c, RulesList *rules, int map_state, int reposition);
extern void write_rules_file(char *filename, RulesList *rules);

#endif /* OROB_CLIENT_PROPS_H */
