/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 

#include "oroborus.h"
#include "iconbars.h"

#define RC_THEME			0 
#define RC_USE_THEMERC			1
#define RC_FONT				2 
#define RC_TEXT_ALIGNMENT		3 
#define RC_ACTIVE_TEXT_COLOR		4 
#define RC_INACTIVE_TEXT_COLOR		5 
#define RC_WINDOW_PLACEMENT		6 
#define RC_FOCUS_STYLE			7 
#define RC_RAISE_ON_FOCUS		8 
#define RC_FOCUS_NEW			9
#define RC_CONTINUOUS_RESIZE		10 
#define RC_ACTIVE_COLOR_1		11
#define RC_ACTIVE_SHADOW_1		12 
#define RC_ACTIVE_HILIGHT_1		13 
#define RC_INACTIVE_COLOR_1		14 
#define RC_INACTIVE_SHADOW_1		15 
#define RC_INACTIVE_HILIGHT_1		16 
#define RC_ACTIVE_COLOR_2		17 
#define RC_ACTIVE_SHADOW_2		18 
#define RC_ACTIVE_HILIGHT_2		19 
#define RC_INACTIVE_COLOR_2		20 
#define RC_INACTIVE_SHADOW_2		21 
#define RC_INACTIVE_HILIGHT_2		22 
#define RC_TOP_MARGIN			23 
#define RC_BOTTOM_MARGIN		24 
#define RC_LEFT_MARGIN			25 
#define RC_RIGHT_MARGIN			26 
#define RC_KEY_NEXT_WINDOW		27
#define RC_KEY_PREV_WINDOW		28 
#define RC_KEY_MAXIMIZE			29
#define RC_KEY_SHADE			30 
#define RC_KEY_HIDE			31 
#define RC_KEY_CLOSE			32 
#define RC_KEY_MOVE_UP			33 
#define RC_KEY_MOVE_DOWN		34 
#define RC_KEY_MOVE_LEFT		35 
#define RC_KEY_MOVE_RIGHT		36 
#define RC_KEY_RESIZE_UP		37 
#define RC_KEY_RESIZE_DOWN		38 
#define RC_KEY_RESIZE_LEFT		39 
#define RC_KEY_RESIZE_RIGHT		40 
#define RC_KEY_RAISE_LAYER		41 
#define RC_KEY_LOWER_LAYER		42 
#define RC_WORKSPACE_COUNT		43 
#define RC_KEY_STICKY			44 
#define RC_KEY_NEXT_WS			45 
#define RC_KEY_PREV_WS			46 
#define RC_KEY_ADD_WS			47 
#define RC_KEY_DEL_WS			48 
#define RC_KEY_WS_1			49 
#define RC_KEY_WS_2			50 
#define RC_KEY_WS_3			51 
#define RC_KEY_WS_4			52 
#define RC_KEY_WS_5			53 
#define RC_KEY_WS_6			54 
#define RC_KEY_WS_7			55 
#define RC_KEY_WS_8			56 
#define RC_KEY_WS_9			57 
#define RC_KEY_MOVE_NEXT_WS		58 
#define RC_KEY_MOVE_PREV_WS		59 
#define RC_KEY_MOVE_WS_1		60 
#define RC_KEY_MOVE_WS_2		61 
#define RC_KEY_MOVE_WS_3		62 
#define RC_KEY_MOVE_WS_4		63 
#define RC_KEY_MOVE_WS_5		64 
#define RC_KEY_MOVE_WS_6		65 
#define RC_KEY_MOVE_WS_7		66 
#define RC_KEY_MOVE_WS_8		67 
#define RC_KEY_MOVE_WS_9		68 
#define RC_CLOSE_X			69
#define RC_CLOSE_Y			70
#define RC_HIDE_X			71
#define RC_HIDE_Y			72
#define RC_ZOOM_X			73
#define RC_ZOOM_Y			74
#define RC_ACTIVE_TITLEBAR_ALPHA	75
#define RC_INACTIVE_TITLEBAR_ALPHA	76
#define RC_LAST				77
 
/* Config option structure */ 
typedef struct 
{ 
  char *option, *value; 
} RcStruct; 

/* 
 * 
 * Load font 
 * 
 */ 

/* 
 * 
 * Free memory used for rc structure 
 * 
 */ 

void free_rc(RcStruct *rc_values) 
{ 
  int i; 
 
#ifdef SHOW_TRACE1 
  printf("TRACE1: free_rc\n"); 
#endif 
 
  for(i=0;rc_values[i].option;i++) 
  { 
    if(rc_values[i].option) 
      free(rc_values[i].option); 
    if(rc_values[i].value) 
      free(rc_values[i].value); 
  } 
} 
 
/*
 * 
 * Simple routine to parse rc file into structure 
 * 
 */ 

int parse_rc(char *rc_dir, char *rc_file, RcStruct *rc_values) 
{ 
  char *full_file_name=NULL, buf[1024], *lvalue=NULL, *rvalue=NULL; 
  FILE *file; 
  int i; 
 
#ifdef SHOW_TRACE1 
  printf("TRACE1: parse_rc\n"); 
#endif 
 
  /* Build full file name from rc_dir and rc_file */ 
  if((full_file_name=malloc(strlen(rc_dir)+strlen(rc_file)+2))==NULL)
  { 
    if(on_cmd_line) {
      fprintf(stderr, "%s: Memory allocation failed in function parse_rc.\n", PROGNAME);
      ClearUpSockets();
      exit(1);
    } else {
      fprintf(stderr, "%s: Memory allocation failed in function parse_rc.\n", PROGNAME);
      gQuitFlag=-3;
      return 0;
    }
  } 

  sprintf(full_file_name, "%s/%s", rc_dir, rc_file); 

#ifdef DEBUG
  printf("  loading config file %s\n", full_file_name);
#endif
 
  /* Open file and parse into rc_values */ 
  if((file=fopen(full_file_name, "r"))) 
  { 
    while(fgets(buf, sizeof buf, file)) 
    { 
      lvalue=strtok(buf, "="); 
      if(lvalue) 
      { 
        for(i=0;rc_values[i].option;i++) 
        { 
          if(!strncasecmp(lvalue, rc_values[i].option, strlen(lvalue))) 
          { 
            rvalue=strtok(NULL, "\n"); 
            if(rvalue) 
            { 
              if(rc_values[i].value) free(rc_values[i].value); 
              rc_values[i].value=strdup(rvalue); 
            } 
          } 
        } 
      } 
    } 
    fclose(file);
    free(full_file_name);
    return True; 
  } 

#ifdef DEBUG
  printf("  file not found\n");
#endif

  free(full_file_name); 
  return False; 
} 

void load_font(char *font_name, char *def_font_name) 
{ 
#ifdef SHOW_TRACE2 
  printf("TRACE2: load_font\n"); 
#endif 
 
  /* Load font */ 
#ifdef I18N_SUPPORT
  if(use_fontset)
  {
    config.fontset=XLoadQueryFontWrapper(dpy, font_name); 
    if(!config.fontset)
    {
      fprintf(stderr, "%s: Can't find fontset %s, loading default fontset...\n", PROGNAME, font_name); 
      config.fontset=XLoadQueryFontWrapper(dpy, def_font_name);
    }
    if(!config.fontset)
    { 
      fprintf(stderr, "%s: Can't find default fontset (%s).\n", PROGNAME, font_name); 
      ClearUpSockets();
      exit(1);
    }
  }
  else
  {
    config.font=XLoadQueryFont(dpy, font_name); 
    if(!config.font)
    {
      fprintf(stderr, "%s: Can't find font %s, loading default font...\n", PROGNAME, font_name);
      config.font=XLoadQueryFont(dpy, def_font_name);
    }  
    if(!config.font)
    {
      fprintf(stderr, "%s: Can't find default font (%s).\n", PROGNAME, font_name); 
      ClearUpSockets();
      exit(1);
    }
  }
#else 
  config.font=XLoadQueryFont(dpy, font_name); 
  if(!config.font)
  {
    fprintf(stderr, "%s: Can't find font %s, loading default font...\n", PROGNAME, font_name);
    config.font=XLoadQueryFont(dpy, def_font_name);
  }  
  if(!config.font)
  {
    fprintf(stderr, "%s: Can't find default font (%s).\n", PROGNAME, font_name); 
    ClearUpSockets();
    exit(1);
  }
#endif /* I18N */ 
} 
 
/* 
 * 
 * Load config file 
 * 
 */ 

void load_config() 
{ 
  XGCValues gv; 
  XColor active_col, inactive_col, dummyc; 
  XpmColorSymbol color_table[14]; 
  RcStruct rc_values[RC_LAST+1]; 
  char *def_font=NULL;
  int i;
  char *reldir=NULL;
 
#ifdef SHOW_TRACE2 
  printf("TRACE2: load_config\n"); 
#endif 
 
  /* Initialize rc structure */ 
  rc_values[RC_THEME].option=strdup("theme"); 
  rc_values[RC_THEME].value=NULL; 
  rc_values[RC_USE_THEMERC].option=strdup("use_themerc");
  rc_values[RC_USE_THEMERC].value=NULL;
  rc_values[RC_FONT].option=strdup("font"); 
  rc_values[RC_FONT].value=NULL; 
  rc_values[RC_TEXT_ALIGNMENT].option=strdup("text_alignment"); 
  rc_values[RC_TEXT_ALIGNMENT].value=NULL;
  rc_values[RC_CONTINUOUS_RESIZE].option=strdup("continuous_resize");
  rc_values[RC_CONTINUOUS_RESIZE].value=NULL;
  rc_values[RC_ACTIVE_TEXT_COLOR].option=strdup("active_text_color"); 
  rc_values[RC_ACTIVE_TEXT_COLOR].value=NULL; 
  rc_values[RC_INACTIVE_TEXT_COLOR].option=strdup("inactive_text_color"); 
  rc_values[RC_INACTIVE_TEXT_COLOR].value=NULL; 
  rc_values[RC_WINDOW_PLACEMENT].option=strdup("window_placement"); 
  rc_values[RC_WINDOW_PLACEMENT].value=NULL; 
  rc_values[RC_FOCUS_STYLE].option=strdup("focus_style"); 
  rc_values[RC_FOCUS_STYLE].value=NULL; 
  rc_values[RC_RAISE_ON_FOCUS].option=strdup("raise_on_focus"); 
  rc_values[RC_RAISE_ON_FOCUS].value=NULL; 
  rc_values[RC_FOCUS_NEW].option=strdup("focus_new"); 
  rc_values[RC_FOCUS_NEW].value=NULL; 
  rc_values[RC_ACTIVE_COLOR_1].option=strdup("active_color_1"); 
  rc_values[RC_ACTIVE_COLOR_1].value=NULL; 
  rc_values[RC_ACTIVE_SHADOW_1].option=strdup("active_shadow_1"); 
  rc_values[RC_ACTIVE_SHADOW_1].value=NULL; 
  rc_values[RC_ACTIVE_HILIGHT_1].option=strdup("active_hilight_1"); 
  rc_values[RC_ACTIVE_HILIGHT_1].value=NULL; 
  rc_values[RC_INACTIVE_COLOR_1].option=strdup("inactive_color_1"); 
  rc_values[RC_INACTIVE_COLOR_1].value=NULL; 
  rc_values[RC_INACTIVE_SHADOW_1].option=strdup("inactive_shadow_1"); 
  rc_values[RC_INACTIVE_SHADOW_1].value=NULL; 
  rc_values[RC_INACTIVE_HILIGHT_1].option=strdup("inactive_hilight_1"); 
  rc_values[RC_INACTIVE_HILIGHT_1].value=NULL; 
  rc_values[RC_ACTIVE_COLOR_2].option=strdup("active_color_2");  
  rc_values[RC_ACTIVE_COLOR_2].value=NULL; 
  rc_values[RC_ACTIVE_SHADOW_2].option=strdup("active_shadow_2");  
  rc_values[RC_ACTIVE_SHADOW_2].value=NULL; 
  rc_values[RC_ACTIVE_HILIGHT_2].option=strdup("active_hilight_2"); 
  rc_values[RC_ACTIVE_HILIGHT_2].value=NULL; 
  rc_values[RC_INACTIVE_COLOR_2].option=strdup("inactive_color_2"); 
  rc_values[RC_INACTIVE_COLOR_2].value=NULL; 
  rc_values[RC_INACTIVE_SHADOW_2].option=strdup("inactive_shadow_2"); 
  rc_values[RC_INACTIVE_SHADOW_2].value=NULL; 
  rc_values[RC_INACTIVE_HILIGHT_2].option=strdup("inactive_hilight_2"); 
  rc_values[RC_INACTIVE_HILIGHT_2].value=NULL; 
  rc_values[RC_TOP_MARGIN].option=strdup("top_margin"); 
  rc_values[RC_TOP_MARGIN].value=NULL; 
  rc_values[RC_BOTTOM_MARGIN].option=strdup("bottom_margin"); 
  rc_values[RC_BOTTOM_MARGIN].value=NULL; 
  rc_values[RC_LEFT_MARGIN].option=strdup("left_margin"); 
  rc_values[RC_LEFT_MARGIN].value=NULL; 
  rc_values[RC_RIGHT_MARGIN].option=strdup("right_margin"); 
  rc_values[RC_RIGHT_MARGIN].value=NULL; 
  rc_values[RC_WORKSPACE_COUNT].option=strdup("workspace_count"); 
  rc_values[RC_WORKSPACE_COUNT].value=NULL; 
  rc_values[RC_KEY_NEXT_WINDOW].option=strdup("key_next_window"); 
  rc_values[RC_KEY_NEXT_WINDOW].value=NULL;
  rc_values[RC_KEY_PREV_WINDOW].option=strdup("key_prev_window");
  rc_values[RC_KEY_PREV_WINDOW].value=NULL; 
  rc_values[RC_KEY_MAXIMIZE].option=strdup("key_maximize"); 
  rc_values[RC_KEY_MAXIMIZE].value=NULL; 
  rc_values[RC_KEY_SHADE].option=strdup("key_shade"); 
  rc_values[RC_KEY_SHADE].value=NULL; 
  rc_values[RC_KEY_HIDE].option=strdup("key_hide"); 
  rc_values[RC_KEY_HIDE].value=NULL; 
  rc_values[RC_KEY_CLOSE].option=strdup("key_close"); 
  rc_values[RC_KEY_CLOSE].value=NULL; 
  rc_values[RC_KEY_MOVE_UP].option=strdup("key_move_up"); 
  rc_values[RC_KEY_MOVE_UP].value=NULL; 
  rc_values[RC_KEY_MOVE_DOWN].option=strdup("key_move_down"); 
  rc_values[RC_KEY_MOVE_DOWN].value=NULL; 
  rc_values[RC_KEY_MOVE_LEFT].option=strdup("key_move_left"); 
  rc_values[RC_KEY_MOVE_LEFT].value=NULL; 
  rc_values[RC_KEY_MOVE_RIGHT].option=strdup("key_move_right"); 
  rc_values[RC_KEY_MOVE_RIGHT].value=NULL; 
  rc_values[RC_KEY_RESIZE_UP].option=strdup("key_resize_up"); 
  rc_values[RC_KEY_RESIZE_UP].value=NULL; 
  rc_values[RC_KEY_RESIZE_DOWN].option=strdup("key_resize_down"); 
  rc_values[RC_KEY_RESIZE_DOWN].value=NULL; 
  rc_values[RC_KEY_RESIZE_LEFT].option=strdup("key_resize_left"); 
  rc_values[RC_KEY_RESIZE_LEFT].value=NULL; 
  rc_values[RC_KEY_RESIZE_RIGHT].option=strdup("key_resize_right"); 
  rc_values[RC_KEY_RESIZE_RIGHT].value=NULL; 
  rc_values[RC_KEY_RAISE_LAYER].option=strdup("key_raise_layer"); 
  rc_values[RC_KEY_RAISE_LAYER].value=NULL; 
  rc_values[RC_KEY_LOWER_LAYER].option=strdup("key_lower_layer"); 
  rc_values[RC_KEY_LOWER_LAYER].value=NULL; 
  rc_values[RC_KEY_STICKY].option=strdup("key_sticky"); 
  rc_values[RC_KEY_STICKY].value=NULL; 
  rc_values[RC_KEY_NEXT_WS].option=strdup("key_next_ws"); 
  rc_values[RC_KEY_NEXT_WS].value=NULL; 
  rc_values[RC_KEY_PREV_WS].option=strdup("key_prev_ws"); 
  rc_values[RC_KEY_PREV_WS].value=NULL; 
  rc_values[RC_KEY_ADD_WS].option=strdup("key_add_ws"); 
  rc_values[RC_KEY_ADD_WS].value=NULL; 
  rc_values[RC_KEY_DEL_WS].option=strdup("key_del_ws"); 
  rc_values[RC_KEY_DEL_WS].value=NULL; 
  rc_values[RC_KEY_WS_1].option=strdup("key_ws_1"); 
  rc_values[RC_KEY_WS_1].value=NULL; 
  rc_values[RC_KEY_WS_2].option=strdup("key_ws_2"); 
  rc_values[RC_KEY_WS_2].value=NULL; 
  rc_values[RC_KEY_WS_3].option=strdup("key_ws_3"); 
  rc_values[RC_KEY_WS_3].value=NULL; 
  rc_values[RC_KEY_WS_4].option=strdup("key_ws_4"); 
  rc_values[RC_KEY_WS_4].value=NULL; 
  rc_values[RC_KEY_WS_5].option=strdup("key_ws_5"); 
  rc_values[RC_KEY_WS_5].value=NULL; 
  rc_values[RC_KEY_WS_6].option=strdup("key_ws_6"); 
  rc_values[RC_KEY_WS_6].value=NULL; 
  rc_values[RC_KEY_WS_7].option=strdup("key_ws_7"); 
  rc_values[RC_KEY_WS_7].value=NULL; 
  rc_values[RC_KEY_WS_8].option=strdup("key_ws_8"); 
  rc_values[RC_KEY_WS_8].value=NULL; 
  rc_values[RC_KEY_WS_9].option=strdup("key_ws_9"); 
  rc_values[RC_KEY_WS_9].value=NULL; 
  rc_values[RC_KEY_MOVE_NEXT_WS].option=strdup("key_move_next_ws"); 
  rc_values[RC_KEY_MOVE_NEXT_WS].value=NULL; 
  rc_values[RC_KEY_MOVE_PREV_WS].option=strdup("key_move_prev_ws"); 
  rc_values[RC_KEY_MOVE_PREV_WS].value=NULL; 
  rc_values[RC_KEY_MOVE_WS_1].option=strdup("key_move_ws_1"); 
  rc_values[RC_KEY_MOVE_WS_1].value=NULL; 
  rc_values[RC_KEY_MOVE_WS_2].option=strdup("key_move_ws_2"); 
  rc_values[RC_KEY_MOVE_WS_2].value=NULL; 
  rc_values[RC_KEY_MOVE_WS_3].option=strdup("key_move_ws_3"); 
  rc_values[RC_KEY_MOVE_WS_3].value=NULL; 
  rc_values[RC_KEY_MOVE_WS_4].option=strdup("key_move_ws_4"); 
  rc_values[RC_KEY_MOVE_WS_4].value=NULL; 
  rc_values[RC_KEY_MOVE_WS_5].option=strdup("key_move_ws_5"); 
  rc_values[RC_KEY_MOVE_WS_5].value=NULL; 
  rc_values[RC_KEY_MOVE_WS_6].option=strdup("key_move_ws_6"); 
  rc_values[RC_KEY_MOVE_WS_6].value=NULL; 
  rc_values[RC_KEY_MOVE_WS_7].option=strdup("key_move_ws_7"); 
  rc_values[RC_KEY_MOVE_WS_7].value=NULL; 
  rc_values[RC_KEY_MOVE_WS_8].option=strdup("key_move_ws_8"); 
  rc_values[RC_KEY_MOVE_WS_8].value=NULL; 
  rc_values[RC_KEY_MOVE_WS_9].option=strdup("key_move_ws_9"); 
  rc_values[RC_KEY_MOVE_WS_9].value=NULL; 
  rc_values[RC_CLOSE_X].option=strdup("close_x");
  rc_values[RC_CLOSE_X].value=NULL;
  rc_values[RC_CLOSE_Y].option=strdup("close_y");
  rc_values[RC_CLOSE_Y].value=NULL;
  rc_values[RC_HIDE_X].option=strdup("icon_x");
  rc_values[RC_HIDE_X].value=NULL;
  rc_values[RC_HIDE_Y].option=strdup("icon_y");
  rc_values[RC_HIDE_Y].value=NULL;
  rc_values[RC_ZOOM_X].option=strdup("zoom_x");
  rc_values[RC_ZOOM_X].value=NULL;
  rc_values[RC_ZOOM_Y].option=strdup("zoom_y");
  rc_values[RC_ZOOM_Y].value=NULL;
  rc_values[RC_ACTIVE_TITLEBAR_ALPHA].option=strdup("active_titlebar_alpha");
  rc_values[RC_ACTIVE_TITLEBAR_ALPHA].value=NULL;
  rc_values[RC_INACTIVE_TITLEBAR_ALPHA].option=strdup("inactive_titlebar_alpha");
  rc_values[RC_INACTIVE_TITLEBAR_ALPHA].value=NULL;
  rc_values[RC_LAST].option=NULL; 
 
  config.close.x=0;
  config.close.y=0;
  config.icon.x=0;
  config.icon.y=0;
  config.zoom.x=0;
  config.zoom.y=0;
  
  /* Parse default rc file */ 
  if(!parse_rc(realPath, "Contents/Resources/oroborosxrc", rc_values))
  {
    fprintf(stderr, "%s: File oroborosxrc does not exist.\n", PROGNAME);
    ClearUpSockets();
    exit(1);
  }

  /* Check all values are loaded in */
  for(i=0;i<RC_LAST;i++)
  {
    if(rc_values[i].value==NULL)
    {
      fprintf(stderr, "%s: Option %s not set in file oroborosxrc.\n", PROGNAME, rc_values[i].option);
      ClearUpSockets();
      exit(1);
    }
  }
#ifdef DEBUG
  printf("Done oroborosxrc file.\n");
#endif

  def_font=strdup(rc_values[RC_FONT].value);
 
  /* Parse local rc file */ 
/*  parse_rc(getenv("HOME"), rcfile, rc_values); */

  if(gThemeDir!=NULL)
    reldir=strdup(gThemeDir);
  else {
    reldir=strdup(rc_values[RC_THEME].value);
    gThemeDir=strdup(reldir);
  }
  printf("Theme in oroborosxrc: %s    Using theme: %s\n",rc_values[RC_THEME].value,reldir);
  
//  if(strlen(realPath)>strlen(getenv("HOME")))
//    config.theme_dir=malloc(strlen(realPath)+strlen(reldir)+51);
//  else
    config.theme_dir=malloc(strlen(getenv("HOME"))+strlen(reldir)+65);
    
  if(config.theme_dir==NULL)
  { 
    if(on_cmd_line) {
      fprintf(stderr, "%s: Memory allocation failed in function load_config.\n", PROGNAME);
      ClearUpSockets();
      exit(1);
    } else {
      gQuitFlag=-3;
      return;
    }
  } 

// look for the theme folder inside the user's oroborosx-prefs folder...
  if( gUseShapedTheme==0 )
     sprintf(config.theme_dir,"%s/Library/Preferences/OroborOSX/Themes.localized/%s/non-shaped",getenv("HOME"),reldir);
  else
     sprintf(config.theme_dir,"%s/Library/Preferences/OroborOSX/Themes.localized/%s/shaped",getenv("HOME"),reldir);
     
  printf("Trying theme dir: %s\n",config.theme_dir);
  
//  if(strcmp(rc_values[RC_USE_THEMERC].value, "false"))

  /* Parse theme rc file -now we require it to be read to check that the theme dir exists */
  if(!parse_rc(config.theme_dir, "themerc", rc_values)) {
    // look for the theme folder inside the app bundle's Contents/Resources/Themes/ folder...
    if( gUseShapedTheme==0 )
      sprintf(config.theme_dir,"Contents/Resources/Themes.localized/%s/non-shaped", reldir);
    else
      sprintf(config.theme_dir,"Contents/Resources/Themes.localized/%s/shaped", reldir);
    // check this one worked...
    if(!parse_rc(config.theme_dir, "themerc", rc_values)) {
      if(on_cmd_line) {
        fprintf(stderr, "%s: Failed to find Theme %s.\n", PROGNAME, reldir);
        ClearUpSockets();
        free(reldir);
        exit(33);
      } else {
        free(reldir);
        ShowAlert(BADTHEME_NAME,0,3);
        gQuitFlag=-33;
        return;
      }
    }
  }
     
  free(reldir);

  /* Add theme dir to font path */
  add_dir_to_font_path(config.theme_dir);
 
  /* Set up color replacement table */ 
  color_table[0].name=rc_values[RC_ACTIVE_TEXT_COLOR].option; 
  color_table[0].value=rc_values[RC_ACTIVE_TEXT_COLOR].value; 
  color_table[1].name=rc_values[RC_INACTIVE_TEXT_COLOR].option; 
  color_table[1].value=rc_values[RC_INACTIVE_TEXT_COLOR].value; 
  color_table[2].name=rc_values[RC_ACTIVE_COLOR_1].option; 
  color_table[2].value=rc_values[RC_ACTIVE_COLOR_1].value; 
  color_table[3].name=rc_values[RC_ACTIVE_SHADOW_1].option; 
  color_table[3].value=rc_values[RC_ACTIVE_SHADOW_1].value; 
  color_table[4].name=rc_values[RC_ACTIVE_HILIGHT_1].option; 
  color_table[4].value=rc_values[RC_ACTIVE_HILIGHT_1].value; 
  color_table[5].name=rc_values[RC_INACTIVE_COLOR_1].option; 
  color_table[5].value=rc_values[RC_INACTIVE_COLOR_1].value; 
  color_table[6].name=rc_values[RC_INACTIVE_SHADOW_1].option; 
  color_table[6].value=rc_values[RC_INACTIVE_SHADOW_1].value; 
  color_table[7].name=rc_values[RC_INACTIVE_HILIGHT_1].option; 
  color_table[7].value=rc_values[RC_INACTIVE_HILIGHT_1].value; 
  color_table[8].name=rc_values[RC_ACTIVE_COLOR_2].option; 
  color_table[8].value=rc_values[RC_ACTIVE_COLOR_2].value; 
  color_table[9].name=rc_values[RC_ACTIVE_SHADOW_2].option; 
  color_table[9].value=rc_values[RC_ACTIVE_SHADOW_2].value; 
  color_table[10].name=rc_values[RC_ACTIVE_HILIGHT_2].option; 
  color_table[10].value=rc_values[RC_ACTIVE_HILIGHT_2].value; 
  color_table[11].name=rc_values[RC_INACTIVE_COLOR_2].option; 
  color_table[11].value=rc_values[RC_INACTIVE_COLOR_2].value; 
  color_table[12].name=rc_values[RC_INACTIVE_SHADOW_2].option; 
  color_table[12].value=rc_values[RC_INACTIVE_SHADOW_2].value; 
  color_table[13].name=rc_values[RC_INACTIVE_HILIGHT_2].option; 
  color_table[13].value=rc_values[RC_INACTIVE_HILIGHT_2].value; 

  /* Load pixmaps */
  load_pixmaps(config.theme_dir, color_table, 14); 
/*  load_pixmaps(rc_values[RC_THEME].value, color_table, 14); */
 
  /* Set window placement */ 
  if(!strncasecmp(rc_values[RC_WINDOW_PLACEMENT].value, "center_root", 11)) 
    config.window_placement=CENTER_ROOT; 
  else 
    config.window_placement=CENTER_MOUSE; 
 
  /* Set focus style */ 
  if(!strncasecmp(rc_values[RC_FOCUS_STYLE].value, "sloppy", 6)) 
    config.focus_style=FOCUS_SLOPPY; 
  else 
    config.focus_style=FOCUS_CLICK; 
 
  if(!strncasecmp(rc_values[RC_RAISE_ON_FOCUS].value, "false", 5)) 
    config.raise_on_focus=False; 
  else 
    config.raise_on_focus=True; 
 
  if(!strncasecmp(rc_values[RC_FOCUS_NEW].value, "false", 5)) 
    config.focus_new=False; 
  else 
    config.focus_new=True; 

  /* Set resize mode */
  if(!strncasecmp(rc_values[RC_CONTINUOUS_RESIZE].value, "false", 5))
    config.continuous_resize=False;
  else
    config.continuous_resize=True;
 
  /* Load font */ 
  load_font(rc_values[RC_FONT].value, def_font);
  if(def_font) free(def_font); 
 
  /* Set text alignment */ 
  if(!strncasecmp(rc_values[RC_TEXT_ALIGNMENT].value, "left", 4)) 
    config.text_align=ALIGNMENT_LEFT; 
  else if(!strncasecmp(rc_values[RC_TEXT_ALIGNMENT].value, "right", 5)) 
    config.text_align=ALIGNMENT_RIGHT; 
  else 
    config.text_align=ALIGNMENT_MIDDLE; 
 
  /* Set text colors */ 
  XAllocNamedColor(dpy, DefaultColormap(dpy, screen), rc_values[RC_ACTIVE_TEXT_COLOR].value, &active_col, &dummyc); 
  XAllocNamedColor(dpy, DefaultColormap(dpy, screen), rc_values[RC_INACTIVE_TEXT_COLOR].value, &inactive_col, &dummyc); 
 
  /* Set active font */ 
  gv.function=GXcopy; 
  gv.foreground=active_col.pixel; 
#ifdef I18N_SUPPORT
  if(use_fontset)
    config.font_gc[FRAME_ACTIVE]=XCreateGC(dpy, root, GCFunction|GCForeground, &gv); 
  else
  {
    gv.font=config.font->fid;
    config.font_gc[FRAME_ACTIVE]=XCreateGC(dpy, root, GCFunction|GCForeground|GCFont, &gv);
  }
#else 
  gv.font=config.font->fid; 
  config.font_gc[FRAME_ACTIVE]=XCreateGC(dpy, root, GCFunction|GCForeground|GCFont, &gv); 
#endif /* I18N */   
 
  /* Set inactive font */ 
  gv.foreground=inactive_col.pixel; 
#ifdef I18N_SUPPORT
  if(use_fontset)
    config.font_gc[FRAME_INACTIVE]=XCreateGC(dpy, root, GCFunction|GCForeground, &gv);
  else
    config.font_gc[FRAME_INACTIVE]=XCreateGC(dpy, root, GCFunction|GCForeground|GCFont, &gv);
#else 
  config.font_gc[FRAME_INACTIVE]=XCreateGC(dpy, root, GCFunction|GCForeground|GCFont, &gv);
#endif /* I18N */ 
 
  /* Set margins */ 
  config.top_margin=atoi(rc_values[RC_TOP_MARGIN].value); 
  config.bottom_margin=atoi(rc_values[RC_BOTTOM_MARGIN].value); 
  config.left_margin=atoi(rc_values[RC_LEFT_MARGIN].value); 
  config.right_margin=atoi(rc_values[RC_RIGHT_MARGIN].value); 

  /* new items for widget positions */
  config.close.x=atoi(rc_values[RC_CLOSE_X].value);
  config.close.y=atoi(rc_values[RC_CLOSE_Y].value);
  config.icon.x=atoi(rc_values[RC_HIDE_X].value);
  config.icon.y=atoi(rc_values[RC_HIDE_Y].value);
  config.zoom.x=atoi(rc_values[RC_ZOOM_X].value);
  config.zoom.y=atoi(rc_values[RC_ZOOM_Y].value);
  
  /* default positions for widgets */
  if(config.close.x==0)
    config.close.x=config.left.width+7;
  if(config.close.y==0)
    config.close.y=(config.top.height-config.close.height)/2+1;
  if(config.icon.x==0)
    config.icon.x=config.close.x+config.close.width;
  if(config.icon.y==0)
    config.icon.y=(config.top.height-config.icon.height)/2+1;
  if(config.zoom.x==0)
    config.zoom.x=config.icon.x+config.icon.width;
  if(config.zoom.y==0)
    config.zoom.y=(config.top.height-config.zoom.height)/2+1;

  /* titlebar translucency */
  config.active_titlebar_alpha=atoi(rc_values[RC_ACTIVE_TITLEBAR_ALPHA].value);
  if(config.active_titlebar_alpha==255) config.active_titlebar_alpha=0;
  config.inactive_titlebar_alpha=atoi(rc_values[RC_INACTIVE_TITLEBAR_ALPHA].value);
  if(config.inactive_titlebar_alpha==255) config.inactive_titlebar_alpha=0;
  printf("Got alphas: %d, %d\n",config.active_titlebar_alpha,config.inactive_titlebar_alpha);
  
#ifdef WORKSPACE_SUPPORT 
  /* Set number of workspaces */ 
  config.workspace_count=atoi(rc_values[RC_WORKSPACE_COUNT].value); 
  if(config.workspace_count<=0) config.workspace_count=1; 
#endif 
 
#ifdef KEYBOARD_SUPPORT 
  /* Keyboard bindings */ 
  parse_key(rc_values[RC_KEY_NEXT_WINDOW].value, &config.keys.next_window);
  parse_key(rc_values[RC_KEY_PREV_WINDOW].value, &config.keys.prev_window); 
  parse_key(rc_values[RC_KEY_MAXIMIZE].value, &config.keys.maximize); 
  parse_key(rc_values[RC_KEY_SHADE].value, &config.keys.shade); 
  parse_key(rc_values[RC_KEY_HIDE].value, &config.keys.hide); 
  parse_key(rc_values[RC_KEY_CLOSE].value, &config.keys.close); 
  parse_key(rc_values[RC_KEY_MOVE_UP].value, &config.keys.move_up); 
  parse_key(rc_values[RC_KEY_MOVE_DOWN].value, &config.keys.move_down); 
  parse_key(rc_values[RC_KEY_MOVE_LEFT].value, &config.keys.move_left); 
  parse_key(rc_values[RC_KEY_MOVE_RIGHT].value, &config.keys.move_right); 
  parse_key(rc_values[RC_KEY_RESIZE_UP].value, &config.keys.resize_up); 
  parse_key(rc_values[RC_KEY_RESIZE_DOWN].value, &config.keys.resize_down); 
  parse_key(rc_values[RC_KEY_RESIZE_LEFT].value, &config.keys.resize_left); 
  parse_key(rc_values[RC_KEY_RESIZE_RIGHT].value, &config.keys.resize_right); 
#ifdef LAYER_SUPPORT 
  parse_key(rc_values[RC_KEY_RAISE_LAYER].value, &config.keys.raise_layer); 
  parse_key(rc_values[RC_KEY_LOWER_LAYER].value, &config.keys.lower_layer); 
#endif 
#ifdef WORKSPACE_SUPPORT 
  parse_key(rc_values[RC_KEY_STICKY].value, &config.keys.sticky); 
  parse_key(rc_values[RC_KEY_NEXT_WS].value, &config.keys.next_ws); 
  parse_key(rc_values[RC_KEY_PREV_WS].value, &config.keys.prev_ws); 
  parse_key(rc_values[RC_KEY_ADD_WS].value, &config.keys.add_ws); 
  parse_key(rc_values[RC_KEY_DEL_WS].value, &config.keys.del_ws); 
  parse_key(rc_values[RC_KEY_WS_1].value, &config.keys.ws_1); 
  parse_key(rc_values[RC_KEY_WS_2].value, &config.keys.ws_2); 
  parse_key(rc_values[RC_KEY_WS_3].value, &config.keys.ws_3); 
  parse_key(rc_values[RC_KEY_WS_4].value, &config.keys.ws_4); 
  parse_key(rc_values[RC_KEY_WS_5].value, &config.keys.ws_5); 
  parse_key(rc_values[RC_KEY_WS_6].value, &config.keys.ws_6); 
  parse_key(rc_values[RC_KEY_WS_7].value, &config.keys.ws_7); 
  parse_key(rc_values[RC_KEY_WS_8].value, &config.keys.ws_8); 
  parse_key(rc_values[RC_KEY_WS_9].value, &config.keys.ws_9); 
  parse_key(rc_values[RC_KEY_MOVE_NEXT_WS].value, &config.keys.move_next_ws); 
  parse_key(rc_values[RC_KEY_MOVE_PREV_WS].value, &config.keys.move_prev_ws); 
  parse_key(rc_values[RC_KEY_MOVE_WS_1].value, &config.keys.move_ws_1); 
  parse_key(rc_values[RC_KEY_MOVE_WS_2].value, &config.keys.move_ws_2); 
  parse_key(rc_values[RC_KEY_MOVE_WS_3].value, &config.keys.move_ws_3); 
  parse_key(rc_values[RC_KEY_MOVE_WS_4].value, &config.keys.move_ws_4); 
  parse_key(rc_values[RC_KEY_MOVE_WS_5].value, &config.keys.move_ws_5); 
  parse_key(rc_values[RC_KEY_MOVE_WS_6].value, &config.keys.move_ws_6); 
  parse_key(rc_values[RC_KEY_MOVE_WS_7].value, &config.keys.move_ws_7); 
  parse_key(rc_values[RC_KEY_MOVE_WS_8].value, &config.keys.move_ws_8); 
  parse_key(rc_values[RC_KEY_MOVE_WS_9].value, &config.keys.move_ws_9); 
#endif 
#endif 
 
  /* Free allocated memory */ 
  free_rc(rc_values); 
} 
 
/* 
 * 
 * Free memory used by config structure 
 * 
 */ 

void free_config() 
{ 
#ifdef SHOW_TRACE2 
  printf("TRACE2: free_config\n"); 
#endif 

  rem_dir_from_font_path(config.theme_dir);
  free(config.theme_dir);
  XFreeGC(dpy, config.font_gc[FRAME_ACTIVE]); 
  XFreeGC(dpy, config.font_gc[FRAME_INACTIVE]); 
#ifdef I18N_SUPPORT
  if(use_fontset) 
    XFreeFontSet(dpy, config.fontset); 
  else
    XFreeFont(dpy, config.font);
#else 
  XFreeFont(dpy, config.font); 
#endif /* I18N */ 
  if(config.title_right.width>0 && config.title_right.height>0) {
    XFreePixmap(dpy, config.title_right.pixmap[FRAME_ACTIVE]);
    XFreePixmap(dpy, config.title_right.pixmap[FRAME_INACTIVE]);
    XFreePixmap(dpy, config.title_right.pixmap[FRAME_ACTIVE_SHADED]); 
    XFreePixmap(dpy, config.title_right.pixmap[FRAME_INACTIVE_SHADED]); 
    XFreePixmap(dpy, config.title_right.pixmap[FRAME_ACTIVE_ICONBAR]); 
    XFreePixmap(dpy, config.title_right.pixmap[FRAME_INACTIVE_ICONBAR]); 
    XFreePixmap(dpy, config.title_right.mask[0]);
    XFreePixmap(dpy, config.title_right.mask[1]);
    XFreePixmap(dpy, config.title_right.mask[2]);
  }
  if(config.title_left.width>0 && config.title_left.height>0) {
    XFreePixmap(dpy, config.title_left.pixmap[FRAME_ACTIVE]);
    XFreePixmap(dpy, config.title_left.pixmap[FRAME_INACTIVE]);
    XFreePixmap(dpy, config.title_left.pixmap[FRAME_ACTIVE_SHADED]); 
    XFreePixmap(dpy, config.title_left.pixmap[FRAME_INACTIVE_SHADED]); 
    XFreePixmap(dpy, config.title_left.pixmap[FRAME_ACTIVE_ICONBAR]); 
    XFreePixmap(dpy, config.title_left.pixmap[FRAME_INACTIVE_ICONBAR]); 
    XFreePixmap(dpy, config.title_left.mask[0]);
    XFreePixmap(dpy, config.title_left.mask[1]);
    XFreePixmap(dpy, config.title_left.mask[2]);
  }
//  if(config.title.width>0 && config.title.height>0) {
    XFreePixmap(dpy, config.title.pixmap[FRAME_ACTIVE]);
    XFreePixmap(dpy, config.title.pixmap[FRAME_INACTIVE]);
    XFreePixmap(dpy, config.title.pixmap[FRAME_ACTIVE_SHADED]); 
    XFreePixmap(dpy, config.title.pixmap[FRAME_INACTIVE_SHADED]); 
    XFreePixmap(dpy, config.title.pixmap[FRAME_ACTIVE_ICONBAR]); 
    XFreePixmap(dpy, config.title.pixmap[FRAME_INACTIVE_ICONBAR]); 
    XFreePixmap(dpy, config.title.mask[0]);
    XFreePixmap(dpy, config.title.mask[1]);
    XFreePixmap(dpy, config.title.mask[2]);
//  }
//  if(config.top.width>0 && config.top.height>0) {
    XFreePixmap(dpy, config.top.pixmap[FRAME_ACTIVE]);
    XFreePixmap(dpy, config.top.pixmap[FRAME_INACTIVE]);
    XFreePixmap(dpy, config.top.pixmap[FRAME_ACTIVE_SHADED]); 
    XFreePixmap(dpy, config.top.pixmap[FRAME_INACTIVE_SHADED]); 
    XFreePixmap(dpy, config.top.pixmap[FRAME_ACTIVE_ICONBAR]); 
    XFreePixmap(dpy, config.top.pixmap[FRAME_INACTIVE_ICONBAR]); 
    XFreePixmap(dpy, config.top.mask[0]);
    XFreePixmap(dpy, config.top.mask[1]);
    XFreePixmap(dpy, config.top.mask[2]);
//  }
//  if(config.top_right.width>0 && config.top_right.height>0) {
    XFreePixmap(dpy, config.top_right.pixmap[FRAME_ACTIVE]);
    XFreePixmap(dpy, config.top_right.pixmap[FRAME_INACTIVE]);
    XFreePixmap(dpy, config.top_right.pixmap[FRAME_ACTIVE_SHADED]); 
    XFreePixmap(dpy, config.top_right.pixmap[FRAME_INACTIVE_SHADED]); 
    XFreePixmap(dpy, config.top_right.pixmap[FRAME_ACTIVE_ICONBAR]); 
    XFreePixmap(dpy, config.top_right.pixmap[FRAME_INACTIVE_ICONBAR]); 
    XFreePixmap(dpy, config.top_right.mask[0]); 
    XFreePixmap(dpy, config.top_right.mask[1]);
    XFreePixmap(dpy, config.top_right.mask[2]);
//  }
//  if(config.top_left.width>0 && config.top_left.height>0) {
    XFreePixmap(dpy, config.top_left.pixmap[FRAME_ACTIVE]);
    XFreePixmap(dpy, config.top_left.pixmap[FRAME_INACTIVE]);
    XFreePixmap(dpy, config.top_left.pixmap[FRAME_ACTIVE_SHADED]);
    XFreePixmap(dpy, config.top_left.pixmap[FRAME_INACTIVE_SHADED]);
    XFreePixmap(dpy, config.top_left.pixmap[FRAME_ACTIVE_ICONBAR]); 
    XFreePixmap(dpy, config.top_left.pixmap[FRAME_INACTIVE_ICONBAR]); 
    XFreePixmap(dpy, config.top_left.mask[0]); 
    XFreePixmap(dpy, config.top_left.mask[1]);
    XFreePixmap(dpy, config.top_left.mask[2]);
//  }
//  if(config.left.width>0 && config.left.height>0) {
    XFreePixmap(dpy, config.left.pixmap[FRAME_ACTIVE]); 
    XFreePixmap(dpy, config.left.pixmap[FRAME_INACTIVE]); 
    XFreePixmap(dpy, config.left.mask[0]);
//  }
//  if(config.right.width>0 && config.right.height>0) {
    XFreePixmap(dpy, config.right.pixmap[FRAME_ACTIVE]); 
    XFreePixmap(dpy, config.right.pixmap[FRAME_INACTIVE]); 
    XFreePixmap(dpy, config.right.mask[0]);
//  }
//  if(config.bottom_left.width>0 && config.bottom_left.height>0) {
    XFreePixmap(dpy, config.bottom_left.pixmap[FRAME_ACTIVE]); 
    XFreePixmap(dpy, config.bottom_left.pixmap[FRAME_INACTIVE]); 
    XFreePixmap(dpy, config.bottom_left.mask[0]);
//  }
//  if(config.bottom.width>0 && config.bottom.height>0) {
    XFreePixmap(dpy, config.bottom.pixmap[FRAME_ACTIVE]); 
    XFreePixmap(dpy, config.bottom.pixmap[FRAME_INACTIVE]); 
    XFreePixmap(dpy, config.bottom.mask[0]);
//  }
//  if(config.resize.width>0 && config.resize.height>0) {
    XFreePixmap(dpy, config.resize.pixmap[FRAME_ACTIVE]); 
    XFreePixmap(dpy, config.resize.pixmap[FRAME_INACTIVE]); 
    XFreePixmap(dpy, config.resize.mask[0]);
//  }
  // Lots more to free now, because of the extra window widget states
  XFreePixmap(dpy, config.close.pixmap[FRAME_ACTIVE]); 
  XFreePixmap(dpy, config.close.pixmap[FRAME_INACTIVE]); 
  XFreePixmap(dpy, config.close.pixmap[BUTTON_PRESSED]); 
  XFreePixmap(dpy, config.close.pixmap[BUTTON_OVER]); 
  XFreePixmap(dpy, config.close.pixmap[BUTTON_OVERBACK]); 
  XFreePixmap(dpy, config.close.pixmap[BUTTON_PRESSEDBACK]); 
  XFreePixmap(dpy, config.close.mask[0]); 
  XFreePixmap(dpy, config.zoom.pixmap[FRAME_ACTIVE]); 
  XFreePixmap(dpy, config.zoom.pixmap[FRAME_INACTIVE]); 
  XFreePixmap(dpy, config.zoom.pixmap[BUTTON_PRESSED]); 
  XFreePixmap(dpy, config.zoom.pixmap[BUTTON_OVER]); 
  XFreePixmap(dpy, config.zoom.pixmap[BUTTON_OVERBACK]); 
  XFreePixmap(dpy, config.zoom.pixmap[BUTTON_PRESSEDBACK]); 
  XFreePixmap(dpy, config.zoom.mask[0]); 
  XFreePixmap(dpy, config.icon.pixmap[FRAME_ACTIVE]); 
  XFreePixmap(dpy, config.icon.pixmap[FRAME_INACTIVE]); 
  XFreePixmap(dpy, config.icon.pixmap[BUTTON_PRESSED]); 
  XFreePixmap(dpy, config.icon.pixmap[BUTTON_OVER]); 
  XFreePixmap(dpy, config.icon.pixmap[BUTTON_OVERBACK]); 
  XFreePixmap(dpy, config.icon.pixmap[BUTTON_PRESSEDBACK]); 
  XFreePixmap(dpy, config.icon.mask[0]);
} 
 
/* 
 * 
 * Reload config file 
 * 
 */ 

void reload_config() 
{ 
  XWindowChanges wc; 
  Client *c=NULL; 
  int i, orgicon, was_shaded; 
#ifdef WORKSPACE_SUPPORT 
  int old_ws_count=config.workspace_count; 
#endif 
 
#ifdef SHOW_TRACE1 
  printf("TRACE1: reload_config\n"); 
#endif 

  Allow_HUPs(0);

  /* Remove gravity from all clients */
  c=client_list;
  for(i=0;i<client_count;i++)
  {
    gravitate_client(c, REMOVE_GRAVITY);
    c=c->next;
  }
 
  free_config(); 
  load_config(); 
#ifdef KEYBOARD_SUPPORT 
  ungrab_keys(check_win); 
  grab_keys(check_win); 
#endif 
#ifdef WORKSPACE_SUPPORT 
  if(config.workspace_count!=old_ws_count) 
    config.workspace_count=old_ws_count; 
#endif 
 
  compress_iconbars();
  
  c=client_list; 
  for(i=0;i<client_count;i++) 
  {
    gravitate_client(c, APPLY_GRAVITY);

    // switch off iconbar state so that defines are correct
    orgicon=c->is_icon;
    c->is_icon=NOT_ICONBAR;
    if((get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)==WIN_STATE_SHADED) {
      was_shaded=1;
      set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_SHADED);
    } else
      was_shaded=0;
    
    if(config.top.height==0 || config.top.width==0) {
      if(c->title) XDestroyWindow(dpy,c->title);
      c->title=0;
    } else if(!c->title)
      c->title=create_frame_part(c->frame, TX, TY, TW<1?1:TW, config.top.height, c->has_frame || orgicon!=NOT_ICONBAR);
      
    if(config.top_left.height==0 || config.top_left.width==0) {
      if(c->top_left) XDestroyWindow(dpy,c->top_left);
      c->top_left=0;
    } else if(!c->top_left)
      c->top_left=create_frame_part(c->frame, TLX, TLY, TLW, TLH, c->has_frame || orgicon!=NOT_ICONBAR);
      
    if(config.top_right.height==0 || config.top_right.width==0) {
      if(c->top_right) XDestroyWindow(dpy,c->top_right);
      c->top_right=0;
    } else if(!c->top_right)
      c->top_right=create_frame_part(c->frame, TRX, TRY, TRW, TRH, c->has_frame || orgicon!=NOT_ICONBAR);  
   
//    if(c->has_frame)
//    { 
    /* Side frame parts */
      if(config.left.width==0 || config.left.height==0) {
#ifdef DEBUG
        printf("no left for %#lx ( %d , %d )\n",c->frame,config.left.width,config.left.height);
#endif
        if(c->left) XDestroyWindow(dpy,c->left);
        c->left=0;
      } else if(!c->left) {
#ifdef DEDBUG
        printf("creating left for %#lx ( %d , %d )\n",c->frame,config.left.width,config.left.height);
        printf("LX,LY,LW,LH: %d,%d,%d,%d\n",LX,LY,LW,LH);
#endif
        c->left=create_frame_part(c->frame, LX, LY, LW, LH, was_shaded==0 && c->has_frame);
      }
        
      if(config.right.width==0 || config.right.height==0) {
        if(c->right) XDestroyWindow(dpy,c->right);
        c->right=0;
      } else if(!c->right)
        c->right=create_frame_part(c->frame, RX, RY, RW, RH, was_shaded==0 && c->has_frame);
     
    /* Bottom frame parts */ 
      if(config.bottom.width==0 || config.bottom.height==0) {
        if(c->bottom) XDestroyWindow(dpy,c->bottom);
        c->bottom=0;
      } else if(!c->bottom)
        c->bottom=create_frame_part(c->frame, BX, BY, BW<1?1:BW, config.bottom.height, was_shaded==0 && c->has_frame);

      if(config.bottom_left.width==0 || config.bottom_left.height==0) {
        if(c->bottom_left) XDestroyWindow(dpy,c->bottom_left);
        c->bottom_left=0;
      } else if(!c->bottom_left)
        c->bottom_left=create_frame_part(c->frame, BLX, BLY, BLW, BLH, was_shaded==0 && c->has_frame);

      if(config.resize.width==0 || config.resize.height==0) {
        if(c->resize) XDestroyWindow(dpy,c->resize);
        c->resize=0;
      } else if(!c->resize)
        c->resize=create_frame_part(c->frame, RSX, RSY, RSW, RSH, was_shaded==0 && c->has_frame);
   
    /* Buttons */ 
      if(config.zoom.width==0 || config.zoom.height==0) {
        if(c->zoom) XDestroyWindow(dpy,c->zoom);
        c->zoom=0;
      } else if(!c->zoom)
        c->zoom=create_frame_part(c->frame, HX, HY, HW, HH, orgicon==NOT_ICONBAR && c->has_frame); 

      if(config.icon.width==0 || config.icon.height==0) {
        if(c->icon) XDestroyWindow(dpy,c->icon);
        c->icon=0;
      } else if(!c->icon)
        c->icon=create_frame_part(c->frame, IX, IY, IW, IH, orgicon==NOT_ICONBAR && c->has_frame);

      if(config.close.width==0 || config.close.height==0) {
        if(c->close) XDestroyWindow(dpy,c->close);
        c->close=0;
      } else if(!c->close)
        c->close=create_frame_part(c->frame, CX, CY, CW, CH, orgicon==NOT_ICONBAR && c->has_frame);
//    } 

    if(was_shaded)
      set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_SHADED);
    
    if(orgicon==NOT_ICONBAR) {
      wc.height=c->height; 
      wc.width=c->width; 
      wc.x=c->x; 
      wc.y=c->y;
    } else {
      wc.height=config.top.height;
      wc.width=get_icontab_width();
      if(iconbarpos==ICONBARPOS_TOPLEFT || iconbarpos==ICONBARPOS_BOTLEFT)
        wc.x=c->icon_x; 
      else
        wc.x=c->icon_x+c->icon_w-wc.width; 
      wc.y=c->icon_y;
      // but we must make sure the actual window is moved to the right place when state is iconbar
      XMoveWindow(dpy,c->window, LW, config.top.height);
    }

    // switch back iconbar state
    c->is_icon=orgicon;
    
    configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, c->is_icon!=NOT_ICONBAR, 3);
    redraw_client_frame(c); 
#ifdef KEYBOARD_SUPPORT 
    ungrab_keys(c->window); 
    grab_keys(c->window); 
#endif
    if(c->is_docked)
        c->is_e_m_r |= THEME_CHANGED;
    c=c->next; 
  } 

  Allow_HUPs(1);
} 

