/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 
 
#include "oroborus.h" 
#include "hackcmds.h"
#include "iconbars.h"

void group_new_client(Client *c, int map_state, Window *parentwin)
{
  Window leadwin, tfwin, groupwin, clientwin;
  long dummy=0, readleft, nitems, *propret;
  int head=0, gotTF=0;
  Atom clientleader,clret;
  PropMwmHints *mhints=NULL; 
  XWMHints *hints=NULL; 
  XClassHint xch;
  XID tmpwin;

  // check for WM_TRANSIENTFOR
  hints=XGetTransientForHint(dpy,c->window,(Window*) &tfwin);
  if(hints && tfwin) {
    gotTF=1;
    c->parent=get_client_from_window(tfwin,WINDOW_LE);
    if(gVerboseConsole) printf("TransientFor: %#lx (client: %#lx)\n",tfwin,c->parent);
  } else {
    gotTF=0;
    if(gVerboseConsole) printf("No TransientFor: %d (%#lx)\n",hints,tfwin);
  }
  
  leadwin=0;
  dummy=0;
  clientleader = XInternAtom(dpy,"WM_CLIENT_LEADER",True);
  if(clientleader!=None)
    dummy=XGetWindowProperty(dpy, c->window, clientleader, 0, 1, False, AnyPropertyType,
                                &clret, &head, &nitems, &readleft, (unsigned char **) &propret);
  if(dummy) {
    if(gVerboseConsole) printf("WM_CLIENT_LEADER atom: %#lx (%d %d %d %d %d)\n",
                                            propret,dummy,clret,head,nitems,readleft);
    if(propret) {
        leadwin=(Window) *propret;
        XFree(propret);
        if(gVerboseConsole) printf("  value: %#lx\n",leadwin);
    }
  } else if(gVerboseConsole)
    printf("No WM_CLIENT_LEADER property\n");
    
      // use the group from the client leader, if different from current window, and if exists
      if(leadwin!=c->window && leadwin) {
        head=XGetClassHint(dpy,leadwin,&xch);
        if(xch.res_class) {
          if(c->clientgroup) XFree(c->clientgroup);
          c->clientgroup=strdup(xch.res_class);
          XFree(xch.res_class);
        }
        if(xch.res_name) XFree(xch.res_name);
      }

  // Check for window_group from WM_HINTS
  hints=XGetWMHints(dpy, c->window); 
  if(hints) 
  { 
    if(gVerboseConsole) {
        printf("flags: %#lx\n",hints->flags);
        printf("input: %d\n",hints->input);
        printf("XID  : %#lx\n",hints->window_group);
        printf("state: %#lx\n",hints->initial_state);
    }
    if(hints->flags & WindowGroupHint) {
      tmpwin=hints->window_group;
      groupwin = (Window) tmpwin;
// assuming this what the window_group is normally used for -i.e. it points to parent window ID
      if(c->parent==NULL) {
        c->parent=get_client_from_window(groupwin,WINDOW_LE);
        if(gVerboseConsole) printf("Parent client (from window_group): %#lx\n",c->parent);
      } else if(gVerboseConsole) {
        printf("TransientFor overriding window_group: %#lx\n",
                        get_client_from_window((Window) groupwin,WINDOW_LE));
      }
      if(c->parent==NULL && groupwin) {
        head=XGetClassHint(dpy,groupwin,&xch);
        if(xch.res_class) {
          if(c->clientgroup) XFree(c->clientgroup);
          c->clientgroup=strdup(xch.res_class);
          XFree(xch.res_class);
        }
        if(xch.res_name) XFree(xch.res_name);
      }
    }
    
    /* Sets the initial window state if set by client */ 
    if((hints->flags & StateHint)!=0 && map_state!=IsViewable) {
      if(gVerboseConsole) printf("Used initial state hints (%d)\n",hints->initial_state);
      set_wm_state(c->window, hints->initial_state); 
    }
    
//    if((hints->flags & IconPixmapHint)!=0) {
      //MakeX11Window(&(c->osxwinfo),c->name,c->clientgroup);
      c->osxwinfo.winref=-10;
//    }
    XFree(hints); 
  } else {
// set initial client state
    if(map_state!=IsViewable)
      set_wm_state(c->window, NormalState);
// Need to create a new OS X window, so name appears in dock icon's window list
    //MakeX11Window(&(c->osxwinfo),c->name,c->clientgroup);
// And bring OroborOSX/XDarwin in front so we don't have to go searching for the new window
    c->osxwinfo.winref=-10;
  }

  /* Get MWM hints */ 
  mhints=get_mwm_hints(c->window); 
  if(mhints) 
  { 
    if(gVerboseConsole) {
        printf("---- MWM HINTS...\n");
        printf("flags      : %#lx\n",mhints->flags);
        printf("functions  : %#lx\n",mhints->functions);
        printf("decorations: %#lx\n",mhints->decorations);
        printf("inputMode  : %#lx\n",mhints->inputMode);
        printf("status     : %#lx\n",mhints->status);
        printf("DECOR_BORDER, DECOR_TITLE: %#lx , %#lx\n",MWM_DECOR_BORDER,MWM_DECOR_TITLE); 
    }
    if(mhints->flags==0) {
      if(c->osxwinfo.winref==0)
        c->osxwinfo.winref=-10;
        //MakeX11Window(&(c->osxwinfo),c->name,c->clientgroup);
        //c->osxwinref=MakeWindow(c->name,c->clientclass,0);
    }
    if((mhints->flags&MWM_HINTS_DECORATIONS)!=0 && (mhints->decorations&MWM_DECOR_ALL)==0) 
    { 
      /* Window has no border if *neither* MWM_DECOR_BORDER *nor* MWM_DECOR_TITLE are set */ 
      c->has_frame=((mhints->decorations&(MWM_DECOR_BORDER|MWM_DECOR_TITLE))!=0); 

      /* Set if client can be maximized or not */ 
      c->can_max=((mhints->decorations&MWM_DECOR_MAXIMIZE)!=0); 
    }
    // make a note of modal windows...
    if((mhints->flags&MWM_HINTS_INPUT_MODE)!=0) {
      if((mhints->inputMode&MWM_INPUT_PRIMARY_APPLICATION_MODAL)!=0) {
        c->osxwinfo.winref=-1;
        c->modal_state|=APP_MODAL;
        if(c->parent)
          (c->parent)->modal_state|=HAS_APP_MODAL_CHILD;
      }
      if((mhints->inputMode&MWM_INPUT_SYSTEM_MODAL)!=0) {
        c->osxwinfo.winref=-1;
        c->modal_state|=SYS_MODAL;
        if(c->parent)
          (c->parent)->modal_state|=HAS_SYS_MODAL_CHILD;
      }
/*    } else if(gotTF) {
        c->osxwinfo.winref=-1;
        c->modal_state|=APP_MODAL;
        if(c->parent)
          (c->parent)->modal_state|=HAS_APP_MODAL_CHILD;*/
    } 
    if(c->osxwinfo.winref==0) {
      // create OS X window, and bring OroborOSX/XDarwin in front
      //MakeX11Window(&(c->osxwinfo),c->name,c->clientgroup);
      //c->osxwinref=MakeWindow(c->name,c->clientclass,0);
      c->osxwinfo.winref=-10;
    }
    XFree(mhints); 
  } else {
    if(gotTF) {
        c->osxwinfo.winref=-1;
        c->modal_state|=APP_MODAL;
        if(c->parent)
          (c->parent)->modal_state|=HAS_APP_MODAL_CHILD;
    } 
    if(c->osxwinfo.winref==0)
      c->osxwinfo.winref=-10;
    // create OS X window, and bring OroborOSX/XDarwin in front
    //MakeX11Window(&(c->osxwinfo),c->name,c->clientgroup);
    //c->osxwinref=MakeWindow(c->name,c->clientclass,0);
  }
}

/*
 * original_group - moves the current focus client from its current group back into its original group
 */
void original_group()
{
}

/*
 * mark_group - used by count_and_mark_groups to set the id for all clients in a particular group
 */
void mark_group(long group_winref, int value)
{
    int i;
    Client *c;
    
    if(client_list==NULL) return;
    if(client_count==0) return;
    
    c = client_list;
    for(i=0;i<client_count;i++) {
        if(get_group_for_client(c)==group_winref)
            c->done_icon=value;
        c = c->next;
    }
}

/*
 * count_and_mark_groups - goes through the clients, from top to bottom, and assigns a number to each group
 * This should be used in conjunction with get_marked_group_name to find the names of all the groups.
 */
int count_and_mark_groups()
{
    Window *windows=NULL, w1, w2; 
    int count, i, value=0;
    Client *c, *lc=NULL, *tc=NULL, *fc=NULL;
    XWindowAttributes attr; 
    
    if(client_count==0) return 0;
    if(client_list==NULL) return 0;
    
    init_nearest_iconbar();
    
    XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
    for(i=count-1;i>=0;i--) {
        XGetWindowAttributes(dpy, windows[i], &attr); 
//        if(attr.map_state==IsViewable) 
//        { 
            c=get_client_from_window(windows[i], FRAME);
            if(c!=NULL && c->done_icon==0) {
                if(c->done_icon==0) {
                    value++;
                    mark_group(get_group_for_client(c),value);
                }
            }
//        } 
    }
    
    if(windows) XFree(windows);
    return value;
}

/*
 * get_marked_group_name - returns the name of the group marked with id "which" (set up using count_and_mark_groups)
 * NOTE: *don't free* the returned string -it is a pointer to the name used within the client's structure
 * a copy should be used when, for example, creating the items for the "Change Group" submenu.
 */
char *get_marked_group_name(int which)
{
    Client *c;
    int	i;
    
    if(client_list==NULL) return NULL;
    if(client_count==0) return NULL;
    if(which==0) return NULL;
    
    c = client_list;
    for(i=0;i<client_count;i++) {
        if(c->done_icon==which) return c->clientgroup;
        c = c->next;
    }
    return NULL;
}

/*
 * get_focus_client_grouporg_name - returns the original group name of the current focus client
 * NOTE: *don't free* the returned string! -it is a pointer to the one within the client's structure
 * a copy should be used when placed into the "Change Group" submenu's "Original" item, for example.
 * ALSO: this will return NULL if the current group is still the original group - if the group name is
 * still required on a NULL return, use get_focus_client_group_name.
 */
char *get_focus_client_grouporg_name()
{
    Client *fc;
    
    fc = get_focus_client();
    if(fc==NULL) return NULL;
    if(fc->clientgroup==fc->clientgrouporg) return NULL;
    return fc->clientgrouporg;
}

/*
 * get_focus_client_group_name - returns the current group name of the current focus client
 * NOTE: *don't free* the returned string! -it is a pointer to the one within the client's structure
 * a copy should be used when placed into the "Change Group" submenu's "Original" item, for example.
 */
char *get_focus_client_group_name()
{
    Client *fc;
    
    fc = get_focus_client();
    if(fc==NULL) return NULL;
    return fc->clientgroup;
}

/*
 * add_client_to_group - used when a client is first created, or when it changes group
 */
void add_client_to_group(Client *c)
{
  long i;
  int grouplen;
  Client *ac, *fc=NULL;
  
  sprintf(lastsub,"TRACE2: add_client_to_group  win: %#lx (last: %#lx)\n",C_WINDOW,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: add_client_to_group  win: %#lx (lats: %#lx)\n",C_WINDOW,FC_WINDOW); 
#endif

  if(c==NULL) return;
//  if(client_list==NULL) return;
//  if(client_count==0) return;
//  if(c->osxwinfo.winref==0) return;
  if(c->clientgroup==NULL) return;
  
  grouplen=strlen(c->clientgroup);
// Loop through the clients, looking for matches with group of latest client
  ac=client_list;
  for(i=0;i<client_count;i++) {
    if(ac->osxwinfo.winref!=0)
      if(ac->clientgroup)
        if(strlen(ac->clientgroup)==grouplen)
          if(!strncmp(ac->clientgroup,c->clientgroup,grouplen)) {
            if(fc==NULL) { // check if this is the first match...
                fc = ac;
                if(fc->osxwinfo.group_winref==0) {
                    // only one other client in this group, so create a group proxy window
                    // (also sets up the client's group_winref)
                    MakeGroupWindow(&(c->osxwinfo),c->clientgroup);
                    // update the other client's group proxy window
                    fc->osxwinfo.group_winref = c->osxwinfo.group_winref;
                    // make sure its title changes appropriately
                    CheckOSXWinTitle(&(fc->osxwinfo),fc->name,fc->clientgroup,fc->isoffscreen);
                } else {
                    // set new client's group winref, and make sure group window is reordered
                    c->osxwinfo.group_winref = fc->osxwinfo.group_winref;
                    ReorderGroupWin(fc->osxwinfo.group_winref);
                }
            }
            ReorderOSXWin(&(ac->osxwinfo));
          }
    ac=ac->next;
  }
  
  if(fc==NULL) Add_Group_to_Menu(c->clientgroup);    
}

/*
 * check_delete_group - should be used just before a client is deleted, or just before a client changes group
 */
void check_delete_group(Client *c)
{
    Client *ac, *lc=NULL;
    long i, cnt=0;
    long group_winref;
    
    if(c==NULL) return;
    if(client_list==NULL) return;
    group_winref = get_group_for_client(c);
    if(group_winref==0) return;
    // if this client does not have an OSX proxy window, then it makes no difference to the Window list
    if(c->osxwinfo.winref==0) return;
    
    ac = client_list;
    for(i=0;i<client_count;i++) {
        // only count clients that have an entry in the Window list
        if(ac->osxwinfo.winref!=0)
            if(get_group_for_client(ac)==group_winref) {
                cnt++;
                if(ac!=c && ac->osxwinfo.group_winref!=0) lc=ac;
            }
        ac = ac->next;
    }

        if(cnt==2 && lc!=NULL) { // only one more client left after this is removed, so take away group proxy window
            DestroyGroupWindow(lc->osxwinfo.group_winref);
            lc->osxwinfo.group_winref = 0;
            CheckOSXWinTitle(&(lc->osxwinfo),lc->name,lc->clientgroup,lc->isoffscreen);
        } else if(cnt==1) { // last client in group, so remove group's entry in menu
            Remove_Group_from_Menu(c->clientgroup);
        }
}

/*
 * get_group_for_client - returns the group proxy window reference for the client
 * NOTE: if the client is the only one in a group, the winref returned is actually the one for its own proxy window
 * Also, if the client does not have an entry in the Window list, then it defers to the parent window.
 */
long get_group_for_client(Client *c)
{
    long winref=0;
    
    if(c==NULL) return 0;
    
    while(1) {
        winref = c->osxwinfo.group_winref;
        if(winref==0) winref = c->osxwinfo.winref;
        if(winref==0) {
            if(c->parent==NULL) return 0;
            c = c->parent;
        } else
            return winref;
    }
}

/*
 * raise_group - raises all windows in the group with proxy window reference group_winref
 * They are raised in such a way that the 'within-group' window order remains the same.
 */
void raise_group(long group_winref)
{
    Window *windows=NULL, w1, w2; 
    int count, i;
    Client *c, *lc=NULL, *tc=NULL, *fc=NULL;
    XWindowAttributes attr; 
    
    if(client_count==0) return;
    if(client_list==NULL) return;
    
    if(gVerboseConsole)
        printf("%s info: raising group with winref %#lx\n",PROGNAME,group_winref);
    if(group_winref==0) return;
    
    MyIncreaseXServerGrab();

    XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
    for(i=count-1;i>=0;i--) {
        XGetWindowAttributes(dpy, windows[i], &attr); 
        if(attr.map_state==IsViewable) 
        { 
            c=get_client_from_window(windows[i], FRAME); 
            if(c!=NULL && (get_group_for_client(c)==group_winref)) 
            { 
#ifdef LAYER_SUPPORT 
                if(get_gnome_hint(c->window, WIN_LAYER)>=layer) {
#endif
                    if((c->is_docked==0) && (c->is_icon==NOT_ICONBAR)) {
                        if(lc==NULL) {
                            force_raise_client(c);
                            fc = c;
                        } else
                            stack_client_re_client(c,lc,Below);
                        lc = c;
                    }
                    tc = c;
#ifdef LAYER_SUPPORT 
                }
#endif /* LAYER */ 
            } 
        } 
    }
    
    if(windows) XFree(windows);

    if(fc==NULL) fc = tc;
        
    if(fc!=NULL) {
        if(fc->is_icon==NOT_ICONBAR) {
            raise_client(fc);
            set_focus_client(fc);
        } else
            minmax_client(fc);
    }
    check_ptrwin=1;
    check_top_client=1;
    
    MyDecreaseXServerGrab();
}

/*
 * lower_group - lowers all windows in the group with proxy window group_winref
 * They are lowered in such a way that the 'within-group' window order remains the same.
 */
void lower_group(long group_winref)
{
    Window *windows=NULL, w1, w2; 
    int count, i;
    Client *c, *prevc=NULL;
    XWindowAttributes attr; 
    
    if(client_count<=1) return;
    if(client_list==NULL) return;
    
    if(gVerboseConsole)
        printf("%s info: lowering group with winref %#lx\n",PROGNAME,group_winref);
    if(group_winref==0) return;
        
    MyIncreaseXServerGrab();

    XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
    for(i=0;i<count;i++) {
        XGetWindowAttributes(dpy, windows[i], &attr); 
        if(attr.map_state==IsViewable) 
        { 
            c=get_client_from_window(windows[i], FRAME); 
            if(c!=NULL && (get_group_for_client(c)==group_winref)) 
            { 
#ifdef LAYER_SUPPORT 
                if(get_gnome_hint(c->window, WIN_LAYER)>=layer) {
#endif
                    if((c->is_docked==0) && (c->is_icon==NOT_ICONBAR)) {
                        if(prevc==NULL)
                            lower_client(c);
                        else
                            stack_client_re_client(c,prevc,Above);
                        prevc = c;
                    }
#ifdef LAYER_SUPPORT 
                }
#endif /* LAYER */ 
            } 
        } 
    }
    
    if(windows) XFree(windows);
    check_ptrwin=1;
    check_top_client=1;

    MyDecreaseXServerGrab();
}

/*
 * get_next_group_client - returns the bottom-most client in a *different* group to the current focus client
 */
Client *get_next_group_client() {
    Window *windows=NULL, w1, w2; 
    int count, i;
    Client *c, *bc;
    XWindowAttributes attr; 
    long winref;
    
    if(client_count<=1) return NULL;
    if(client_list==NULL) return NULL;
    
    c = get_focus_client();
    if(c==NULL) return NULL;
    
    winref = get_group_for_client(c);
    if(winref==0) {
        bc = get_bottom_most_client(WIN_LAYER_NORMAL,0);
        if(bc==c) bc = NULL;
        return bc;
    }
    
    XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
    for(i=0;i<count;i++) {
        XGetWindowAttributes(dpy, windows[i], &attr); 
        if(attr.map_state==IsViewable) 
        { 
            c=get_client_from_window(windows[i], FRAME); 
            if(c!=NULL && (c->isoffscreen==0 || gIncludeOffScreen) && get_group_for_client(c)!=winref) 
            { 
#ifdef LAYER_SUPPORT 
                if(get_gnome_hint(c->window, WIN_LAYER)>=layer) {
#endif
                    if((c->is_docked==0) && (c->is_icon==NOT_ICONBAR)) {
                        if(windows) XFree(windows);
                        if(gVerboseConsole)
                            printf("%s info: got next group with winref %#lx (top: %#lx)\n",PROGNAME,c->osxwinfo.group_winref,winref);
                        return c;
                    }
#ifdef LAYER_SUPPORT 
                }
#endif /* LAYER */ 
            } 
        } 
    }
    
    if(windows) XFree(windows);
    if(gVerboseConsole)
    	printf("%s info: got no next group! (top: %#lx)\n",PROGNAME,winref);
    return NULL;
}

/*
 * get_next_group_name - convenient wrapper for get_next_group_client, to return the name of the client's group
 * (used to get the name for the "Next Group" item in the Window menu)
 */
char *get_next_group_name() {
    Client *c;
    char *retval=NULL;
    
    c = get_next_group_client();
    if(c!=NULL) retval = strdup(c->clientgroup);

    return retval;
}

/*
 * raise_next_group - convenient wrapper for get_next_group_client, to raise the client's group
 * (used to raise the "Next Group", when chosen from the Window menu)
 */
void raise_next_group() {
    Client *c;
    
    c = get_next_group_client();
    if(c!=NULL) {
        if(c->osxwinfo.view_state==VIEWSTATE_HIDDEN)
            map_group(get_group_for_client(c));
        else if(get_group_for_client(c)==0)
            raise_client(c);
        else
            raise_group(get_group_for_client(c));
    }
}

/*
 * get_next_client_in_group - returns the bottom-most client in the group with proxy window reference group_winref
 * (used for the "Next Window" item in the window menu)
 */
Client *get_next_client_in_group(long group_winref)
{
    Window *windows=NULL, w1, w2; 
    int count, i;
    Client *c=NULL;
    XWindowAttributes attr;
    
    if(client_count<=1) return NULL;
    if(client_list==NULL) return NULL;
    
    if(group_winref==0) {
        c = get_bottom_most_client(WIN_LAYER_NORMAL,0);
        return c;
    } else {
        XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
        for(i=0;i<count;i++) {
            XGetWindowAttributes(dpy, windows[i], &attr); 
            if(attr.map_state==IsViewable) 
            { 
                c=get_client_from_window(windows[i], FRAME); 
                if(c!=NULL && (c->isoffscreen==0 || gIncludeOffScreen) &&
                        get_group_for_client(c)==group_winref && check_client_modality(c,0)!=get_focus_client()) 
                { 
#ifdef LAYER_SUPPORT 
                    if(get_gnome_hint(c->window, WIN_LAYER)>=layer) {
#endif
                  	 if((c->is_docked==0) && (c->is_icon==NOT_ICONBAR)) {
                            if(gVerboseConsole)
                                printf("%s info: got next in group: %#lx (grpref: %#lx)\n",PROGNAME,c->window,group_winref);
                            if(windows) XFree(windows);
                            return c;
                        }
#ifdef LAYER_SUPPORT 
                    }
#endif /* LAYER */ 
                } 
            } 
        }
    
        if(windows) XFree(windows);
    }
    
    return NULL;
}

/*
 * get_next_title_in_group - convenience wrapper for get_next_client_in_group, to return the title of the
 * client returned by get_next_client_in_group. (Used to put the title into the "Next Window" item.)
 * If "strict" is non-zero, it will only return the bottom-most client in the same group as the focus client.
 * If "strict" is zero, it can return a client from a different group (if the focus client is the only one in
 * its group).
 */
char *get_next_title_in_group(int strict)
{
    Client *fc, *c;
    char *wintitle=NULL, *offset_title=NULL;
    
    fc=get_focus_client();
//#ifdef DEBUG
    if(fc==NULL && gVerboseConsole)
      printf("Got NULL focus client in get_next_title_in_group.\n");
//#endif
    if(fc==NULL) return NULL;
    
    if(strict)
        c=get_next_client_in_group(get_group_for_client(fc));
    else
        c=get_next_client_in_group(fc->osxwinfo.group_winref);

    if(c!=NULL) {  
      if(c->name==NULL && c->clientgroup==NULL) {
        wintitle = malloc(15);
        sprintf(wintitle,"?: Untitled");
      } else if(c->name==NULL) {
        wintitle = malloc(strlen(c->clientgroup)+15);
        sprintf(wintitle,"%s: Untitled",c->clientgroup);
      } else if(c->clientgroup==NULL) {
        wintitle = malloc(strlen(c->name)+5);
        sprintf(wintitle,"?: %s",c->name);
      } else {
        offset_title = c->name;
        if(strlen(c->clientgroup)<strlen(c->name))
            if(!strncmp(c->clientgroup,c->name,strlen(c->clientgroup))) {
                offset_title+=strlen(c->clientgroup);
                // strip out spaces after class name
                while(strlen(offset_title)>1 && !strncmp(offset_title," ",1))
                    offset_title++;
                // strip out colon
                if(strlen(offset_title)>1 && !strncmp(offset_title,":",1))
                    offset_title++;
                // strip out any spaces after colon
                while(strlen(offset_title)>1 && !strncmp(offset_title," ",1))
                    offset_title++;
            }
        wintitle = malloc(strlen(c->clientgroup)+strlen(offset_title)+5);
        sprintf(wintitle,"%s:  %s",c->clientgroup,offset_title);
      }
    }
    
    return wintitle;
}

/*
 * get_next_osxwinref_in_group - convenience wrapper for get_next_client_in_group, to return the winref of the
 * client returned by get_next_client_in_group. (Used to raise the "Next Window", when chosen from Window menu.)
 * If "strict" is non-zero, it will only return the bottom-most client in the same group as the focus client.
 * If "strict" is zero, it can return a client from a different group (if the focus client is the only one in
 * its group).
 */
long get_next_osxwinref_in_group(int raiseit, int strict)
{
    Client *c, *fc;

    fc=get_focus_client();
    
//#ifdef DEBUG
    if(fc==NULL && gVerboseConsole)
      printf("Got NULL focus client in get_next_osxwinref_in_group.\n");
//#endif
    if(fc==NULL) return 0;

    if(strict)
        c=get_next_client_in_group(get_group_for_client(fc));
    else
        c=get_next_client_in_group(fc->osxwinfo.group_winref);

    if(c!=NULL) {
      if(raiseit!=0) {
        if(c->is_docked)
          will_undock_client(c);
        else {
          set_focus_client(c);
          if((c->auto_properties & (AUTO_FOCUS|AUTO_RAISE))==AUTO_FOCUS)
            raise_client(c);
          if(c->al_list!=NULL) XFree(c->al_list);
          c->al_list = NULL;
          c->al_count = 0;
        }
      }
      return c->osxwinfo.winref;
    } else
      return 0;
}

/*
 * get_top_not_this_in_group - returns top client in group with proxy group_winref that is not client "notc".
 */
Client *get_top_not_this_in_group(Client *notc, int layer, int get_icons, long group_winref)
{
    Window *windows=NULL, w1, w2; 
    int count, i;
    Client *c=NULL;
    XWindowAttributes attr;
    
        XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
        for(i=count-1;i>=0;i--) {
            XGetWindowAttributes(dpy, windows[i], &attr); 
            if(attr.map_state==IsViewable) 
            { 
                c=get_client_from_window(windows[i], FRAME);
                if(c!=NULL && c!=notc && (c->isoffscreen==0 || gIncludeOffScreen) &&
                        get_group_for_client(c)==group_winref && check_client_modality(c,0)!=notc) 
                { 
#ifdef LAYER_SUPPORT 
                    if(get_gnome_hint(c->window, WIN_LAYER)>=layer) {
#endif
                        if(((c->is_docked==0) && (c->is_icon==NOT_ICONBAR)) || get_icons) {
                            if(gVerboseConsole)
                                printf("%s info: got top in group: %#lx (grpref: %#lx)\n",PROGNAME,c->window,group_winref);
                            if(windows) XFree(windows);
                            return c;
                        }
#ifdef LAYER_SUPPORT 
                    }
#endif /* LAYER */ 
                } 
            } 
        }
    
        if(windows) XFree(windows);
        
    return NULL;
}

/*
 * map_group - unhides the windows in group proxy group_winref
 */
void map_group(long group_winref)
{
  Window *windows=NULL, w1, w2; 
  int count, i, x, y; 
  Client *c;
  XWindowAttributes attr;

//  if(can_send_hackcmds!=0) return;
  
    if(client_count==0) return;
    if(client_list==NULL) return;
    
    if(gVerboseConsole)
        printf("%s info: map_group with winref %#lx\n",PROGNAME,group_winref);
    if(group_winref==0) return;

  sprintf(lastsub,"TRACE1: map_group %#lx\n",group_winref); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: map_group %#lx\n",group_winref); 
#endif
 
  MyIncreaseXServerGrab(); // grab the server, so nothing happens while we're messing around
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
#ifdef DEBUG
  printf("  children of root: %d\n",count);
#endif

  // Search through the windows that actually exist. (It's possible that some might have
  // disappeared while hidden, but they remain in the client list. This is because the window
  // manager had no notification they were unmapped as they already were unmapped when the
  // user chose to hide OroborOSX.)
  for(i=0;i<count;i++) { 
#ifdef DEBUG
    printf("  child %d (%#lx)\n",i,windows[i]);
#endif
    c=get_client_from_window(windows[i],FRAME_OR_COVER);
    if(c!=NULL && get_group_for_client(c)==group_winref) {
      if(c->cover==windows[i]) {
        printf("moveresize show cover for window %#lx\n",c->window);
        XMoveResizeWindow(dpy,c->cover,c->x,c->y,c->width,c->height);
      } else {
      if(can_send_hackcmds==0) {
        if(c->is_docked==0) {
          XMapWindow(dpy, c->frame); 
          if(c->is_icon!=NOT_ICONBAR)
            c->osxwinfo.view_state=VIEWSTATE_ICONBAR;
          else
            c->osxwinfo.view_state=VIEWSTATE_VISIBLE;
          if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
          if(c->is_icon==NOT_ICONBAR) {
            XMapWindow(dpy, c->window);
            set_wm_state(c->window, NormalState);
          }
          if((c->has_frame) || (c->is_icon!=NOT_ICONBAR)) redraw_client_frame(c);
        } else {
          c->osxwinfo.view_state=VIEWSTATE_DOCKED;
          if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
        }
      } else {
        printf("getattr frame show for window %#lx\n",c->window);
        XGetWindowAttributes(dpy, windows[i], &attr);
//      if(attr.map_state==IsViewable) {
        x = attr.x;
        y = attr.y;
//      }
        XMoveWindow(dpy, windows[i], HACKCMD_STARTX, HACKCMD_STARTY);
        XMoveWindow(dpy, windows[i], HACKCMD_HIDEWINDOW, 0);
        XRaiseWindow(dpy, windows[i]);
        if(c->is_docked!=0) {
          c->osxwinfo.view_state=VIEWSTATE_DOCKED;
        } else if(c->is_icon!=NOT_ICONBAR) {
          c->osxwinfo.view_state=VIEWSTATE_ICONBAR;
        } else {
          set_wm_state(windows[i], NormalState);
          c->osxwinfo.view_state=VIEWSTATE_VISIBLE;
        }
        if(c->isoffscreen==0) {
          if(c->has_frame) redraw_client_frame(c);
          XMoveWindow(dpy,windows[i],FX,FY);
          MyXSync(dpy,0);
        }
        if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
//      } else {
//        set_wm_state(windows[i], NormalState);
//        XMoveWindow(dpy, windows[i], x, y);
//        XFlush(dpy);
      }
      }
    }
  } 

  if(windows) XFree(windows);
  check_ptrwin=1;
  check_top_client=1;

  MyDecreaseXServerGrab(); // release the X server, if it was before

}

/*
 * unmap_group - hides the windows in group proxy group_winref
 */
void unmap_group(long group_winref) {
  Window *windows=NULL, w1, w2; 
  int count, i, x, y; 
  Client *c;
  XWindowAttributes attr;
  
//  focus_before_hide=get_focus_client();
  
//  if(can_send_hackcmds!=0) {
//    SendSimpleHackCommand(HACKCMD_HIDE_ALLXWIN,1);
//    return;
//  }
  
    if(client_count==0) return;
    if(client_list==NULL) return;
    
    if(gVerboseConsole)
        printf("%s info: unmap_group with winref %#lx\n",PROGNAME,group_winref);
    if(group_winref==0) return;

  sprintf(lastsub,"TRACE1: unmap_group %#lx\n",group_winref); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE1: unmap_group %#lx\n",group_winref); 
#endif
 
  MyIncreaseXServerGrab(); // grab the server, so nothing happens while we're messing around
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
#ifdef DEBUG
  printf("  children of root: %d\n",count);
#endif

  for(i=count-1;i>=0;i--) 
  { 
#ifdef DEBUG
    printf("  child %d (%#lx)\n",i,windows[i]);
#endif
    c=get_client_from_window(windows[i],FRAME_OR_COVER);
    if(c!=NULL && get_group_for_client(c)==group_winref) {
      if(c->cover==windows[i]) {
        printf("moveresize hide cover for window %#lx\n",c->window);
        XMoveResizeWindow(dpy,c->cover,-6000,-6000,5,5);
      } else {
        XLowerWindow(dpy,windows[i]);
        if(can_send_hackcmds==0) {
          if(c->is_docked==0) {
            c->ignore_unmap++; 
            XUnmapWindow(dpy, c->frame); 
            c->osxwinfo.view_state=VIEWSTATE_HIDDEN;
            if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
            XUnmapWindow(dpy, c->window); 
            set_wm_state(c->window, WithdrawnState); 
          }
        } else {
          printf("getattr frame hide for window %#lx\n",c->window);
          XGetWindowAttributes(dpy, windows[i], &attr);
          if(attr.map_state==IsViewable) {
            c->osxwinfo.view_state=VIEWSTATE_HIDDEN;
            if(c->osxwinfo.winref) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
            x = attr.x;
            y = attr.y;
            set_wm_state(windows[i], WithdrawnState);
            XMoveWindow(dpy, windows[i], HACKCMD_STARTX, HACKCMD_STARTY);
            XMoveWindow(dpy, windows[i], HACKCMD_HIDEWINDOW, 1);
            /*if(c->isoffscreen==0)
              XMoveWindow(dpy, windows[i], x, y);
            else*/
              XMoveWindow(dpy, windows[i], -5000, -5000);
            MyXSync(dpy,0);
          }
        }
      }
    }
  }

  c = get_top_most_client(WIN_LAYER_NORMAL,0);
  if(c!=NULL) set_focus_client(c);
  
  MyDecreaseXServerGrab(); // release the X server, if it was before
  
  if(windows) XFree(windows);
  check_ptrwin=1;
  check_top_client=1;
  
//  c=client_list;
//  for(i=0;i<client_count;i++) {
//    c=c->next;
//  }
}
