/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 
 
#include "oroborus.h" 
#include "hackcmds.h"
#include "iconbars.h"

// Initial value for number of iconbar steps -use the medium one (really needs to read a pref file)
int iconbar_steps=-1, icontab_width=5, iconbar_width=100; // iconbar_steps now read from prefs
//int iconbar_steps=ICONSTEPS_MED, icontab_width=5, iconbar_width=100;

int get_icontab_width()
{
    int i;

    if(gHiding_Iconbars==0) return iconbar_width;
    
    if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT) {
      i=config.iconbar_left_width;
      if(i<1) i=config.iconbar_right_width;
    } else {
      i=config.iconbar_left_width;
      if(i<1) i=config.iconbar_left_width;
    }
    
    if(i<3)
      i=3;
    else if(i>8)
      i=8;
    
    return i;
}

/*
 * Initialise the flag for checking if we've done a client during a 'nearest iconbar' search
 */
 
void init_nearest_iconbar()
{
  Client *c=NULL;
  int i;
  
  c=client_list;
  for(i=0;i<client_count;i++) {
    c->done_icon=0;
    c=c->next;
  }
}

/*
 * Find the iconbar nearest (vertically only) to the supplied y co-ordinate (ypos)
 */
 
Client *get_nearest_iconbar(int ypos,int state)
{
  int i, diff, mindif;
  Client *c=NULL, *ac;
  
  mindif=10000;
  ac=client_list;
  for(i=0;i<client_count;i++) {
    if(ac->is_icon==state || ac->is_icon==ICONBAR_RAISED || state==NOT_ICONBAR) {
      // only want clients that have already been an iconbar
      if(ac->icon_y!=-10000) {
        diff=ac->icon_y-ypos;
        if(diff<0) diff=-diff;
        if(diff<=mindif) {
          // only take clients that have not already been found in a previous call
          if(ac->done_icon==0) {
            if(state==NOT_ICONBAR) {
              // want a list of iconbar positions for every window that has been 'iconbar-ised'
              c=ac;
              mindif=diff;
            } else if(state==ICONBAR_HIDDEN) {
              // only want iconbars that are hidden
              c=ac;
              mindif=diff;
            } else if(state==ICONBAR_VISIBLE) {
              // only want iconbars that are visible
              c=ac;
              mindif=diff;
            }
          }
        }
      }
    }
    ac=ac->next;
  }
  
  // Note this client as having been found
  if(c!=NULL)
    c->done_icon=1;
    
  return c;
}

/*
 * Hide/unhide a particular iconbar
 */
 
void toggle_slide_iconbar(Client *c)
{
  XWindowChanges wc;

  sprintf(lastsub,"TRACE1: toggle_slide_iconbar  win: %#lx\n",C_WINDOW);
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: toggle_slide_iconbar  win: %#lx\n",C_WINDOW);
#endif

  if(c==NULL) return;
  if(gHiding_Iconbars==0) return;
    
  icontab_width=get_icontab_width();
    
  if(c->is_icon==ICONBAR_HIDDEN) {
#ifdef SHOW_WEIRD
    if(iconbar_state==ICONBAR_VISIBLE)
      printf("%s oddity: Iconbar is hidden, but state is visible!!\n",PROGNAME);
#endif
    if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
      wc.x=c->icon_x+c->icon_w-iconbar_width;
    else
      wc.x=c->icon_x;
    if(c->title) XMapWindow(dpy,c->title);
    if(c->top_left) XMapWindow(dpy,c->top_left);
    if(c->top_right) XMapWindow(dpy,c->top_right);
    wc.y=c->icon_y;
    wc.width=iconbar_width;
    wc.height=c->icon_h;
//    configure_client(c, CWWidth|CWX, &wc, True, False); 
//          printf("doing XMoveResize in toggle_slide_iconbar: %#lx\n",c);
//          XMoveResizeWindow(dpy,c->frame,wc.x,wc.y,wc.width,wc.height);
//    XConfigureWindow(dpy,c->frame,CWWidth|CWX|CWY|CWHeight, &wc);
    configure_client(c, CWWidth|CWX|CWY|CWHeight, &wc, True, 0); 
    c->is_icon=ICONBAR_VISIBLE;
  } else if(c->is_icon==ICONBAR_VISIBLE) {
    if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
      wc.x=c->icon_x+c->icon_w-icontab_width;
    else
      wc.x=c->icon_x;
    wc.y=c->icon_y;
    wc.width=icontab_width;
    wc.height=c->icon_h;
//    configure_client(c, CWWidth|CWX, &wc, True, False);
//          XMoveResizeWindow(dpy,c->frame,wc.x,wc.y,wc.width,wc.height);
    configure_client(c, CWWidth|CWX|CWY|CWHeight, &wc, True, 0);
    c->is_icon=ICONBAR_HIDDEN; 
    wc.stack_mode=Below; 
//    XConfigureWindow(dpy, c->frame, CWStackMode, &wc);
//    redraw_client_frame(c);
    configure_client(c, CWStackMode, &wc, True, 0);
  }
}

/*
 * Hide/unhide all iconbars
 */
 
void slide_all_iconbars(Client *c)
{
  XWindowChanges wc;
  int i, k, neary;
  Client *ac;
  
  if((long) c==-1)
    sprintf(lastsub,"TRACE1: slide_all_iconbars (got -1)\n");
  else
    sprintf(lastsub,"TRACE1: slide_all_iconbars (win: %#lx)\n",C_WINDOW);
  subtrace();
#ifdef SHOW_TRACE1
  if((long) c==-1)
    printf("TRACE1: slide_all_iconbars (got -1)\n");
  else
    printf(,"TRACE1: slide_all_iconbars (win: %#lx)\n",C_WINDOW);
#endif

  lastIconbarOutTicks=0;
  if(gHiding_Iconbars==0) return;
  
  icontab_width=get_icontab_width();

  if(c==NULL) {
    if(iconbar_state==ICONBAR_HIDDEN) {
#ifdef SHOW_WEIRD
      printf("%s oddity: Iconbars already hidden??\n",PROGNAME);
#endif
      return; }
    iconbar_state=ICONBAR_HIDDEN;
    neary=nexticon_y;
    for(k=icontab_width;k>=icontab_width;k-=SLIDESTEP) {
      init_nearest_iconbar();
      ac=get_nearest_iconbar(neary,ICONBAR_VISIBLE);
      while(ac!=NULL) {
        if(ac->is_icon==ICONBAR_VISIBLE || ac->is_icon==ICONBAR_RAISED) {
          if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
            wc.x=ac->icon_x+ac->icon_w-k;
          else
            wc.x=ac->icon_x;
          if(ac->title) XMapWindow(dpy,ac->title);
          if(ac->top_left) XMapWindow(dpy,ac->top_left);
          if(ac->top_right) XMapWindow(dpy,ac->top_right);
          wc.y=ac->icon_y;
          wc.width=k;
          wc.height=ac->icon_h;
//          wc.stack_mode=Below;
//          configure_client(ac, CWWidth|CWX, &wc, True, False);
//          printf("doing XMoveResize in slide_all_iconbars: %#lx\n",ac);
//          XMoveResizeWindow(dpy,ac->frame,wc.x,wc.y,wc.width,wc.height);
//          XConfigureWindow(dpy,ac->frame,CWWidth|CWX|CWY|CWHeight, &wc);
          configure_client(ac, CWWidth|CWX|CWY|CWHeight, &wc, True, 0);
          wc.stack_mode=Below;
//          XConfigureWindow(dpy, ac->frame, CWStackMode, &wc);
//          redraw_client_frame(ac);
          configure_client(ac, CWStackMode, &wc, True, 0);
        }
        ac=get_nearest_iconbar(neary,ICONBAR_VISIBLE);
      }
    }
    ac=client_list;
    for(i=0;i<client_count;i++) {
      if(ac->is_icon==ICONBAR_VISIBLE || ac->is_icon==ICONBAR_RAISED)
        ac->is_icon=ICONBAR_HIDDEN; 
      ac=ac->next;
    }
  } else {
    if(iconbar_state==ICONBAR_VISIBLE) {
//#ifdef SHOW_WEIRD
//      printf("Iconbars already visible\n");
//#endif
      return; }
    iconbar_state=ICONBAR_VISIBLE;
//    wc.stack_mode=Above; 
//    XConfigureWindow(dpy, check_win, CWStackMode, &wc);

    if((long) c!=-1) c->is_icon=ICONBAR_RAISED; // special value which means there is no client to start at
    
    for(k=iconbar_width;k<=iconbar_width;k+=SLIDESTEP) {
    
      if((long) c!=-1) {
        if(c->is_icon==ICONBAR_HIDDEN || c->is_icon==ICONBAR_RAISED) {
          wc.stack_mode=Above; 
//          XConfigureWindow(dpy, c->frame, CWStackMode, &wc);
          configure_client(c, CWStackMode, &wc, True, 0);
          if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
            wc.x=c->icon_x+c->icon_w-k;
          else
            wc.x=c->icon_x;
          wc.y=c->icon_y;
          wc.width=k;
          wc.height=c->icon_h;
//          configure_client(c, CWWidth|CWX, &wc, True, False);
          configure_client(c, CWWidth|CWX|CWY|CWHeight, &wc, True, 0);
//          XMoveResizeWindow(dpy,c->frame,wc.x,wc.y,wc.width,wc.height);
        }
        neary = c->icon_y;
      } else
        neary = nexticon_y;
        
      init_nearest_iconbar();
      ac=get_nearest_iconbar(neary,ICONBAR_HIDDEN);
      
      while(ac!=NULL) {
        if(ac->is_icon==ICONBAR_HIDDEN || ac->is_icon==ICONBAR_RAISED) {
          wc.stack_mode=Above; 
//          XConfigureWindow(dpy, ac->frame, CWStackMode, &wc);
          configure_client(ac, CWStackMode, &wc, True, 0);
          if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
            wc.x=ac->icon_x+ac->icon_w-k;
          else
            wc.x=ac->icon_x;
          wc.y=ac->icon_y;
          wc.width=k;
          wc.height=ac->icon_h;
//          configure_client(ac, CWWidth|CWX, &wc, True, False);
          configure_client(ac, CWWidth|CWX|CWY|CWHeight, &wc, True, 0);
//          XMoveResizeWindow(dpy,ac->frame,wc.x,wc.y,wc.width,wc.height);
        }
        ac=get_nearest_iconbar(neary,ICONBAR_HIDDEN);
      }

    }
    
    ac=client_list;
    for(i=0;i<client_count;i++) {
      if(ac->is_icon==ICONBAR_HIDDEN || ac->is_icon==ICONBAR_RAISED)
        ac->is_icon=ICONBAR_VISIBLE; 
      ac=ac->next;
    }
    
    if((long) c!=-1) c->is_icon=ICONBAR_RAISED; // watch out for the special value!
  }
}

/*
 * Simple routine to unhide all iconbars
 */
void unhide_all_iconbars()
{
    slide_all_iconbars((Client *) -1); // send special "-1" signal, to say there is no client to start with
}

/*
 * 'Glue-like' routine to allow hiding of iconbars without requiring knowledge of Client structure
 */
void hide_all_iconbars()
{
  if(gHiding_Iconbars==0) return;
  slide_all_iconbars(NULL);
}

void change_iconbar_menuspace()
{
  Client *c;
  int neary, newx, newy;
  long i;
  
  sprintf(lastsub,"TRACE1: change_iconbar_menuspace (gMenubarSpace: %d)\n",gMenubarSpace); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: change_iconbar_menuspace (gMenubarSpace: %d)\n",gMenubarSpace); 
#endif 

  if(gMenubarSpace==0)
    gMenubarSpace=21;
  else
    gMenubarSpace=-21;

  if(iconbarpos==ICONBARPOS_BOTLEFT || iconbarpos==ICONBARPOS_BOTRIGHT) {
    if(gMenubarSpace==-21) gMenubarSpace=0;
    return;
  }

  neary=nexticon_y;
  
  init_nearest_iconbar();
  c=get_nearest_iconbar(neary,NOT_ICONBAR);
  while(c!=NULL) {
    newx=c->icon_x;
    newy=c->icon_y+gMenubarSpace;
    if(c->is_icon!=NOT_ICONBAR)
      move_client(c,newx,newy,0);
    else {
      c->icon_x=newx;
      c->icon_y=newy;
    }
    c=get_nearest_iconbar(neary,NOT_ICONBAR);
  }
    
  nexticon_y+=gMenubarSpace;
  
  if(gMenubarSpace==-21) gMenubarSpace=0;
}

void change_iconbar_screen(int newhead)
{
  Client *c;
  int neary, newx, newy, dx, dy;
  
  sprintf(lastsub,"TRACE1: change_iconbar_screen (newhead,current: %d,%d)\n",newhead,iconbar_screen); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: change_iconbar_screen (newhead,current: %d,%d)\n",newhead,iconbar_screen); 
#endif 

  if(iconbar_screen==newhead) {
    compress_iconbars();
    return;
  }
  
  neary=nexticon_y;
    
  dx = screen_x(newhead) - screen_x(iconbar_screen);
  dy = screen_y(newhead) - screen_y(iconbar_screen);
  if(iconbarpos==ICONBARPOS_TOPRIGHT) {
    dx += screen_width(newhead) - screen_width(iconbar_screen);
  } else if(iconbarpos==ICONBARPOS_BOTLEFT) {
    dy += screen_height(newhead) - screen_height(iconbar_screen);
  } else if(iconbarpos==ICONBARPOS_BOTRIGHT) {
    dx += screen_width(newhead) - screen_width(iconbar_screen);
    dy += screen_height(newhead) - screen_height(iconbar_screen);
  }
  
  nexticon_x+=dx;
  nexticon_y+=dy;
  
  iconbar_screen=newhead;
  
  init_nearest_iconbar();
  c=get_nearest_iconbar(neary,NOT_ICONBAR);
  
  while(c!=NULL) {
  
    newx=c->icon_x+dx;
    newy=c->icon_y+dy;
    
    if(c->is_icon!=NOT_ICONBAR)
      move_client(c,newx,newy,0);
    else {
      c->icon_x=newx;
      c->icon_y=newy;
    }
      
    c=get_nearest_iconbar(neary,NOT_ICONBAR);
  }

  compress_iconbars();
}

void change_iconbarpos(int newpos)
{
  Client *c;
  int head=0,swapr2l=0, swapl2r=0, swapt2b=0, swapb2t=0, neary, newx, newy, old_icontab_width;
  
  sprintf(lastsub,"TRACE1: change_iconbarpos (newpos: %d)\n",newpos); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: change_iconbarpos (newpos: %d)\n",newpos); 
#endif 

  if(iconbarpos==newpos) return;
  
  head=iconbar_screen;

//#ifdef XINERAMA_SUPPORT
//    if(xinerama)
//      head=find_head(50, 50);
//#endif /* XINERAMA */

  if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_TOPLEFT) {
    if(newpos==ICONBARPOS_BOTRIGHT || newpos==ICONBARPOS_BOTLEFT)
      swapt2b=1;
  } else if(newpos==ICONBARPOS_TOPRIGHT || newpos==ICONBARPOS_TOPLEFT)
    swapb2t=1;
  
  if(iconbarpos==ICONBARPOS_TOPLEFT || iconbarpos==ICONBARPOS_BOTLEFT) {
    if(newpos==ICONBARPOS_TOPRIGHT || newpos==ICONBARPOS_BOTRIGHT)
      swapl2r=1;
  } else if(newpos==ICONBARPOS_TOPLEFT || newpos==ICONBARPOS_BOTLEFT)
    swapr2l=1;

  iconbarpos=newpos;
  
  neary=nexticon_y;
  if(swapt2b==1) {
    nexticon_y=screen_y(head)+screen_height(head)-neary+screen_y(head);
    if(gMenubarSpace==0) nexticon_y-=22;
  } else if(swapb2t==1) {
    nexticon_y=screen_y(head)+screen_height(head)-neary+screen_y(head);
    if(gMenubarSpace==0) nexticon_y-=22;
  }
  
  if(swapl2r==1) {
    nexticon_x=screen_x(head)+screen_width(head);
  } else if(swapr2l==1) {
    nexticon_x=screen_x(head);
  }
  
  init_nearest_iconbar();
  c=get_nearest_iconbar(neary,NOT_ICONBAR);
  
  old_icontab_width=icontab_width;
  icontab_width=get_icontab_width();
    
  while(c!=NULL) {
  
    newx=c->icon_x;
    newy=c->icon_y;
    
    if(swapl2r==1)
      newx=screen_x(head)+screen_width(head)-c->icon_w;
    else if(swapr2l==1)
      newx=screen_x(head);
    
    if(swapt2b==1) {
      newy=screen_y(head)+screen_height(head)-c->icon_y+screen_y(head);
      if(gMenubarSpace==0) newy-=22;
    } else if(swapb2t==1) {
      newy=screen_y(head)+screen_height(head)-c->icon_y+screen_y(head);
      if(gMenubarSpace==0) newy-=22;
    }

    if(c->is_icon!=NOT_ICONBAR) {
      if(c->icon_w==old_icontab_width) {
        move_resize_client(c,newx,newy,icontab_width,config.top.height,1);
        lower_client(c);
      } else
        move_client(c,newx,newy,1);
    } else {
      c->icon_x=newx;
      c->icon_y=newy;
    }
      
    c=get_nearest_iconbar(neary,NOT_ICONBAR);
  }
  
  compress_iconbars();
  
}

void compress_iconbars()
{
  Client *c;
  int neary, newx, newy, yadd;

  if(iconbar_screen>=xinerama_heads) iconbar_screen=xinerama_heads-1;

  if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_TOPLEFT) {
    newy=screen_y(iconbar_screen);
    yadd=config.top.height;
    if(gMenubarSpace!=0) newy+=22;
  } else {
    newy=screen_y(iconbar_screen)+screen_height(iconbar_screen)-config.top.height;
    yadd=-config.top.height;
  }
  
  neary=newy;
  init_nearest_iconbar();
  c=get_nearest_iconbar(neary,NOT_ICONBAR);

  while(c!=NULL) {
  
    newy+=yadd;
    newx=c->icon_x;
    
    if(c->is_icon!=NOT_ICONBAR) {
      move_client(c,newx,newy,1);
    } else {
      c->icon_x=newx;
      c->icon_y=newy;
    }
    
    c=get_nearest_iconbar(neary,NOT_ICONBAR);
  }

  nexticon_y=newy+yadd;
  
}

/*
 * This has turned into a bit of a monster routine to toggle a client into/from an iconbar!
 */
 
void minmax_client(Client *c) 
{ 
  XWindowChanges wc;
  int i, nowt, nowx, nowy, noww, endt, endx, endy, endw;
  int shaded=0;
  Client *newc=NULL, *ac=NULL;
  Window ptrwin;

  sprintf(lastsub,"TRACE1: minmax_client (win,last: %#lx %#lx)\n",C_WINDOW,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: minmax_client (win,last: %#lx %#lx)\n",C_WINDOW,FC_WINDOW); 
#endif 
 
//    newc=get_focus_client();
    newc=lastfocus_client;

    if( c->is_icon!=NOT_ICONBAR ) {
       if(gX11InFront==0)
         Bring_X_To_Front(1);
       if(newc) {
         set_focus_client(NULL);
//         redraw_client_frame(newc);
         newc=NULL; }
       if(iconbar_state==ICONBAR_HIDDEN && c->is_icon==ICONBAR_HIDDEN)
         toggle_slide_iconbar(c);
       nowx=c->icon_x;
       nowy=c->icon_y;
       endx=c->x-config.left.width;
       endy=c->y;
       nowt=c->icontxt_x;
       noww=c->icon_w;
       if(shrinking_iconbars!=0) {
         endw=c->width+config.left.width+config.right.width;
         endt=c->name_x;
       } else {
         endt=nowt;
         endw=noww; }
       c->is_icon=ICONBAR_RAISED;
       // move iconbar from its location to where the window should be, changing width as necessary
       for(i=iconbar_steps-1;i>=0;i--) {
         wc.x=endx-(endx-nowx)*i/iconbar_steps;
         wc.y=endy-(endy-nowy)*i/iconbar_steps;
         wc.width=endw-(endw-noww)*i/iconbar_steps;
         wc.height=c->icon_h;
         c->icontxt_x=endt-(endt-nowt)*i/iconbar_steps;
         if(i==iconbar_steps-1)
           configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, True, 0);
         else if(shrinking_iconbars!=0)
           configure_client(c, CWWidth|CWX|CWY, &wc, True, 0);
         else
           configure_client(c, CWX|CWY, &wc, True, 0); 
         //sleep(1);  -heh! interesting to watch it do this slowly...!
       }
       wc.x=c->x;
       wc.y=c->y;
       wc.width=c->width;
       wc.height=c->height;
       c->is_icon=NOT_ICONBAR;
       c->osxwinfo.view_state = VIEWSTATE_VISIBLE;
       if(c->osxwinfo.winref!=0) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
       c->icon_x=nowx;
       if(noww!=iconbar_width)
         if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
           c->icon_x=nowx+noww-iconbar_width;
       c->icon_y=nowy;
       c->icon_w=iconbar_width;
       if(shrinking_iconbars!=0)
         c->icontxt_x=nowt;
       XMapWindow(dpy,c->window);
       if(c->has_frame==0) {
         if(c->top_left) XUnmapWindow(dpy,c->top_left);
         if(c->top_right) XUnmapWindow(dpy,c->top_right);
         if(c->title) XUnmapWindow(dpy,c->title);
       } else {
         if(c->close) XMapWindow(dpy,c->close);
         if(c->icon) XMapWindow(dpy,c->icon);
         if(c->zoom) XMapWindow(dpy,c->zoom);
         if(c->left) XMapWindow(dpy,c->left);
         if(c->right) XMapWindow(dpy,c->right);
         if(c->bottom) XMapWindow(dpy,c->bottom);
         if(c->bottom_left) XMapWindow(dpy,c->bottom_left);
         if(c->resize) XMapWindow(dpy,c->resize);
       }
       SendHackTitleToDockCommand(c,0);
    } else {
       // Tell OroborOSX this window is being iconised - doesn't actually do anything, since we still
       // want it to appear in the dock icon's window list, but may be useful at some point...
       c->osxwinfo.view_state = VIEWSTATE_ICONBAR;
       if(c->osxwinfo.winref!=0) ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
       nowx=c->x-config.left.width;
       nowy=c->y;
       noww=c->width+config.left.width+config.right.width;
       nowt=c->name_x;
       if( newc==c ) {
         c->is_icon=ICONBAR_RAISED;
       } else {
         c->is_icon=ICONBAR_VISIBLE; }
       SendHackTitleToDockCommand(c,0);
       // need these sub-windows back, if there was no frame
       if(c->has_frame==0) {
         if(c->top_left) XMapWindow(dpy,c->top_left);
         if(c->top_right) XMapWindow(dpy,c->top_right);
         if(c->title) XMapWindow(dpy,c->title);
       } else {
         if(c->close) XUnmapWindow(dpy,c->close);
         if(c->icon) XUnmapWindow(dpy,c->icon);
         if(c->zoom) XUnmapWindow(dpy,c->zoom);
         if(c->left) XUnmapWindow(dpy,c->left);
         if(c->right) XUnmapWindow(dpy,c->right);
         if(c->bottom) XUnmapWindow(dpy,c->bottom);
         if(c->bottom_left) XUnmapWindow(dpy,c->bottom_left);
         if(c->resize) XUnmapWindow(dpy,c->resize);
       }
       newc=NULL;
       // Check to see if this window has been 'iconbar-ised' before - if not, set new values
       if(c->icon_x==-10000) {
         c->icon_w=iconbar_width;
         if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
           c->icon_x=nexticon_x-c->icon_w;
         else
           c->icon_x=nexticon_x;
         c->icon_y=nexticon_y;
         c->icon_h=config.top.height;
         c->icontxt_x=3+config.iconbar_left_width;
         c->icontxt_y=c->name_y;
         c->icontxt_w=c->name_w;
         c->icontxt_h=c->name_h;
         // Which way are the iconbars going?
         if(iconbarpos==ICONBARPOS_TOPLEFT || iconbarpos==ICONBARPOS_TOPRIGHT)
           nexticon_y+=config.top.height;
         else if(iconbarpos==ICONBARPOS_BOTLEFT || iconbarpos==ICONBARPOS_BOTRIGHT)
           nexticon_y-=config.top.height;
       }
       endx=c->icon_x;
       endy=c->icon_y;
       endw=c->icon_w;
       endt=c->icontxt_x;
       if(shrinking_iconbars==0) {
         noww=endw;
         nowt=endt; }
       // get rid of the actual window content
       c->ignore_unmap++;
       XUnmapWindow(dpy,c->window);
       // move iconbar from the window to where the it's going, changing width as necessary
       for(i=0;i<iconbar_steps;i++) {
         wc.width=noww-(noww-endw)*i/iconbar_steps;
         wc.height=c->icon_h;
         wc.x=nowx-(nowx-endx)*i/iconbar_steps;
         wc.y=nowy-(nowy-endy)*i/iconbar_steps;
         c->icontxt_x=nowt-(nowt-endt)*i/iconbar_steps;
         if(i==0)
           configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, True, 0);
         else if(shrinking_iconbars!=0)
           configure_client(c, CWWidth|CWX|CWY, &wc, True, 0);
         else
           configure_client(c, CWX|CWY, &wc, True, 0); 
         //sleep(1);
       }
       wc.width=endw;
       wc.height=c->icon_h;
       wc.x=endx;
       wc.y=endy;
       if(shrinking_iconbars!=0)
         c->icontxt_x=endt;
       c->is_icon=ICONBAR_VISIBLE;
//       c->osxwinfo.view_state=VIEWSTATE_ICONBAR;
   }

    //shaded=get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED;
    
    // always set the window to shaded -otherwise the window 'flickers' twice when
    // returning from an iconbar, for some reason...
    set_gnome_hint(c->window, WIN_STATE, WIN_STATE_SHADED);
    shaded=WIN_STATE_SHADED;

    if( c->is_icon==NOT_ICONBAR ) {
       overicon=NULL;
       will_hide_all_iconbars();
       raise_client(c);
       set_focus_client(c);
       if(gX11InFront==0)
         Bring_To_Front(0);
    }

    configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, ((shaded==WIN_STATE_SHADED) || (c->is_icon!=NOT_ICONBAR)), 0); 

    if(c->cover && c->height-TH-BH>0 && c->width-LW-RW>0)
      XMoveResizeWindow(dpy,c->cover,c->x+LW,c->y+TH,c->width-LW-RW,c->height-TH-BH);

//    redraw_client_frame(c);

    if(c->is_icon!=NOT_ICONBAR) {
//      lower_client(c);
//      printf("Finished iconbar-ing... %d\n",iconbar_state);
      newc=get_top_most_client(WIN_LAYER_NORMAL,False);
      if( newc==NULL ) {
        set_focus_client(NULL);
      } else if( newc->is_icon!=NOT_ICONBAR ) {
        set_focus_client(NULL);
      } else
        set_focus_client(newc);

      ptrwin=get_pointer_win(root);
      
      if(ptrwin==c->frame) {
//        printf("over new iconbar: %#lx\n",c);
        // We've ended up with the cursor over the new iconbar, so make sure iconbars don't disappear!
        if(iconbar_state==ICONBAR_HIDDEN)
          slide_all_iconbars(c);
        else
          lastIconbarOutTicks=0;
        // ...and that it lights up
        c->is_icon=ICONBAR_RAISED;
        // ...and the width expands to see the whole text, if needed
        if( c->icontxt_w>=TW-6 )
          resize_client(c,c->icontxt_w+6-LW-RW+TLW+TRW,c->icon_h,3);
//        redraw_client_frame(c);
        raise_client(c);
        overicon=c;
        if(can_send_hackcmds) c->is_e_m_r |= IS_ENTERED;
      } else {
        ac=get_client_from_window(ptrwin,FRAME_LE);
//        printf("over client: %#lx , ",ac);
        if(ac!=NULL) {
          if(ac->is_icon==NOT_ICONBAR) {
//            printf("not an iconbar...\n");
            if(iconbar_state==ICONBAR_VISIBLE) {
              c->is_icon=ICONBAR_VISIBLE;
              slide_all_iconbars(NULL);
            } else {
              c->is_icon=ICONBAR_VISIBLE;
              toggle_slide_iconbar(c);
            }
          } else {
//            printf("is an iconbar...\n");
            // we've ended up with the cursor over another iconbar, so make sure iconbars don't disappear
            if(iconbar_state!=ICONBAR_VISIBLE) {
              c->is_icon=ICONBAR_VISIBLE;
              slide_all_iconbars(ac);
            } else
              lastIconbarOutTicks=0;
          }
        } else {
//          printf("iconbar_state: %d\n",iconbar_state);
          if(iconbar_state==ICONBAR_VISIBLE) {
            c->is_icon=ICONBAR_VISIBLE;
            slide_all_iconbars(NULL);
          } else {
            c->is_icon=ICONBAR_VISIBLE;
            toggle_slide_iconbar(c);
          }
        }
      }
      
      newc=get_top_most_client(WIN_LAYER_NORMAL,False);

    } else {
//      printf("Finished un-iconbar-ing...  \n");
      raise_client(c);
      set_focus_client(c);
      if( shaded==WIN_STATE_SHADED )
        toggle_client_shaded(c);
// don't need this code any more...?
//      ptrwin=get_pointer_win(root);
//      printf("ptrwin,frame: %#lx %#lx\n",ptrwin,c->frame);
//      if(ptrwin==c->frame) {
//        printf("iconbar_state: %d\n",iconbar_state);
//        will_hide_all_iconbars();
//      }
    }
} 
 
