/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 
 
#include "oroborus.h"
#include "hackcmds.h"
// FIXME - this is not supposed to be here - see note in force_grab_release
//#include "iconbars.h"

int xserver_grab_level=0;
long check_clients_onscreen=0;
 
/* 
 * 
 * Print out X errors 
 * 
 */ 
int handle_xerror(Display *dpy, XErrorEvent *err) 
{ 
  char msg[1024]; 
 
  sprintf(lastsub,"TRACE1: handle_xerror (code: %d)\n",err->error_code); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: handle_xerror (code: %d)\n",err->error_code); 
#endif 
 
//  XUngrabServer(dpy);
  switch(err->error_code) 
  { 
  case BadAccess:  
    if(err->resourceid==root)
    {
      if(on_cmd_line) {
        fprintf(stderr, "%s: Another window manger is running.\n", PROGNAME);
        ClearUpSockets();
        exit(1);
      } else {
        show_last_ten();
        ShowAlert(OTHERWM_NAME,0,3);
        gQuitFlag=OTHERWM_ERRNO;
        return 0;
      }
    }
    break; 
  case BadWindow: 
    XGetErrorText(dpy, err->error_code, msg, 1024);
#ifdef SHOW_WEIRD
    if(!inhibit_errors) {
      fprintf(stderr, "%s: X BadWindow Warning (%#lx): %s.\n", PROGNAME, err->resourceid, msg);
      show_last_fifteen();
      list_all_clients(); }
#endif
    break;
  case BadDrawable: 
    XGetErrorText(dpy, err->error_code, msg, 1024);
#ifdef SHOW_WEIRD
    fprintf(stderr, "%s: X BadDrawable Warning (%#lx): %s.\n", PROGNAME, err->resourceid, msg);
    show_last_fifteen();
    list_all_clients();
#endif
    break;
  case BadMatch: 
    XGetErrorText(dpy, err->error_code, msg, 1024);
#ifdef SHOW_WEIRD
    fprintf(stderr, "%s: X BadMatch Warning (%#lx): %s.\n", PROGNAME, err->resourceid, msg);
    show_last_fifteen();
    list_all_clients();
#endif
    break; 
  default: 
    XGetErrorText(dpy, err->error_code, msg, 1024);
//    if(on_cmd_line)
      fprintf(stderr, "%s: X Error (number %#lx): %s.\n", PROGNAME, err->resourceid, msg);
//    else
//      return 0;
    show_last_fifteen();
    list_all_clients();
  } 
  return 0; 
} 

int mystrlen(char *str)
{
    if(str==NULL) return -1;
    if(str[0]==0) return 0;
    return strlen(str);
}

void MyXSync(Display *dpy, Bool discard)
{
    if(gVerboseConsole)
        printf("%s info: Calling XSync... (%d)\n",PROGNAME,discard);
    XSync(dpy,discard);
}

/*
 * some routines to deal with nested grabbing
 */
void MyIncreaseXServerGrab()
{
    xserver_grab_level++;
    if( xserver_grab_level < 2 ) {
        if(gVerboseConsole)
            printf("%s info: Grabbing X server\n",PROGNAME);
        XGrabServer(dpy);
        xserver_grab_level=1;
    }
}

void MyDecreaseXServerGrab()
{
    xserver_grab_level--;
    if( xserver_grab_level < 1 ) {
        if(gVerboseConsole)
            printf("%s info: Ungrabbing X server\n",PROGNAME);
        XUngrabServer(dpy);
        xserver_grab_level=0;
    }
}

int MyXServerGrabLevel()
{
    return xserver_grab_level;
}

int check_need_ungrab()
{
  Window win;
  
  if(mouse_mode==MOUSE_RESIZE || mouse_mode==MOUSE_MOVE || mouse_mode==MOUSE_STOPPED || mouse_mode==MOUSE3_TITLE) {
    printf("%s info: ignoring ungrab check (mouse_mode=%d)\n",PROGNAME,mouse_mode);
    return 0;
  }
  
  if(lastfocus_client==NULL)
    win = check_win;
  else
    win = lastfocus_client->frame;
    
  if(pointer_grabbed==0) {
    int retval=XGrabPointer(dpy, win, False, PointerMotionMask|ButtonReleaseMask, GrabModeSync, GrabModeSync, None, normal_curs, CurrentTime);
    if(retval==GrabSuccess) {
        XUngrabPointer(dpy, CurrentTime);
        XAllowEvents(dpy, SyncPointer, CurrentTime);
    } else {
        return retval;
    }
  }
  return 0;
}

void force_grab_release()
{
    XButtonEvent ev;
    int retval, i;
    Window win;
    Client *c=NULL;
    
    if(client_list==NULL || can_send_hackcmds==0) return;
    
    if(mouse_mode==MOUSE_RESIZE || mouse_mode==MOUSE_MOVE || mouse_mode==MOUSE3_TITLE || mouse_mode==MOUSE_STOPPED) {
        printf("%s info: ignoring force ungrab (mouse_mode=%d)\n",PROGNAME,mouse_mode);
        return;
    }

// FIXME - this needs changing so it doesn't mess with check_win, but so that XDarwin itself makes
// sure the click ends up on the root window (by using its inside inhibiting code...)
// Reason is because we should not reference anything to do with the Client struct in this file
// (so we can then remove the #include "iconbars.h" at the start).
    MyIncreaseXServerGrab();
/*            c = client_list;
            win = c->frame;
            XMoveWindow(dpy, check_win, HACKCMD_STARTX, HACKCMD_STARTY);
            XMoveWindow(dpy, check_win, HACKCMD_HIDEWINDOW, 1);
            XSync(dpy,0);
            XMoveResizeWindow(dpy, check_win, 0, 0, XDisplayWidth(dpy,screen),XDisplayHeight(dpy,screen));
            XRaiseWindow(dpy,check_win);*/
            win = check_win;
            XMoveWindow(dpy, win, HACKCMD_STARTX, HACKCMD_STARTY);
            XMoveWindow(dpy, win, HACKCMD_UNGRABPTR, 0);
            XMoveWindow(dpy, win, -1000, -1000);
//            XMoveWindow(dpy, win, FX, FY);
            XSync(dpy,0);
    MyDecreaseXServerGrab();

    retval = -1;
    i = 0;

    did_force_ungrab = GetCarbonTickCount();
    while(retval!=GrabSuccess) {
        retval = XGrabPointer(dpy, check_win, True, PointerMotionMask|ButtonReleaseMask|ButtonPressMask|EnterWindowMask,
                                GrabModeSync, GrabModeSync, None, normal_curs, CurrentTime);
        if(retval==GrabSuccess) {
            XUngrabPointer(dpy,CurrentTime);
            XAllowEvents(dpy, SyncPointer, CurrentTime);
        } else if(i>100) {
            printf("%s warning: Failed to ungrab after 1 second...\n",PROGNAME);
            retval = GrabSuccess;
            check_button_grab = 1;
        } else if(i==1) {
            XSetInputFocus(dpy, check_win, RevertToNone, CurrentTime);
        } else {
            usleep(10000);
            i++;
        }
    }
    
    MyIncreaseXServerGrab();
    XResizeWindow(dpy,check_win,5,5);
    XSync(dpy,0);
    XMoveWindow(dpy, check_win, HACKCMD_STARTX, HACKCMD_STARTY);
    XMoveWindow(dpy, check_win, HACKCMD_HIDEWINDOW, 0);
    XMoveWindow(dpy, check_win, -1000, -1000);
    XLowerWindow(dpy,check_win);
    XSync(dpy,0);
    MyDecreaseXServerGrab();
}

void pointer_grabbing(Window win)
{
  int retval;
  
  return;
  if(win!=0 && win!=pointer_grabbed) {
    if(pointer_grabbed) {
      XUngrabPointer(dpy,CurrentTime);
      XAllowEvents(dpy, SyncPointer, CurrentTime);
    }
    retval = XGrabPointer(dpy, win, True, PointerMotionMask|ButtonReleaseMask|ButtonPressMask|EnterWindowMask,
                              GrabModeAsync, GrabModeAsync, None, normal_curs, CurrentTime);
    if(retval==GrabSuccess) {
      printf("***** grabbed pointer from %#lx to %#lx (%d)\n",pointer_grabbed,win,retval);
      pointer_grabbed = win;
    } else
      printf("***** failed to grab pointer from %#lx to %#lx (%d)\n",pointer_grabbed,win,retval);
  } else if(win==0 && pointer_grabbed!=0) {
    XUngrabPointer(dpy,CurrentTime);
    XAllowEvents(dpy, SyncPointer, CurrentTime);
    printf("***** ungrabbed pointer from %#lx\n",pointer_grabbed);
    pointer_grabbed = 0;
  }
}

#ifdef XINERAMA_SUPPORT
void check_window_pos(int x, int y, int *w, int *h, int *xr, int *yr, int *flag)
{
    int tlh, trh, blh, brh, head=0, badone=0, wo, ho, xo=x, yo=y;
    
      wo = *w;
      ho = *h;

      if(gVerboseConsole)
        printf("check pos: %d %d , %d x %d (%d %d vs %d %d)\n",x,y,wo,ho,x+(wo/2),y+(ho/2),DisplayWidth(dpy,screen)/2,DisplayHeight(dpy,screen)/2);
      if(*flag>=0) // want a specific head
        head=*flag;
      else { // find current head (i.e. containing cursor)
        int mx, my;
        get_mouse_xy(&mx, &my);
        head=find_head(mx, my);
      }
      
      *xr = x;
      *yr = y;
      *flag = head;
            
      {
        int clz=0;
        // check to see if the window cuts across heads
        tlh = strict_find_head(x,y);
        if(tlh<0) clz++;
        trh = strict_find_head(x+wo,y);
        if(trh<0) clz++;
        blh = strict_find_head(x,y+ho);
        if(blh<0) clz++;
        brh = strict_find_head(x+wo,y+ho);
        if(brh<0) clz++;
        if(gVerboseConsole)
          printf("Got heads for tl,tr,bl,br: %d %d %d %d\n",tlh,trh,blh,brh);
        if(clz==1) {
            // three corners are inside screens - give the fourth a screen, based on which others are equal
            if(tlh<0) {
                if(trh==brh) tlh=blh;
                else if(blh==brh) tlh=trh;
                else tlh=find_head(x,y);
            } else if(trh<0) {
                if(tlh==blh) trh=brh;
                else if(blh==brh) trh=tlh;
                else trh=find_head(x+wo,y);
            } else if(blh<0) {
                if(trh==tlh) blh=brh;
                else if(trh==brh) blh=tlh;
                else blh=find_head(x,y+ho);
            } else if(brh<0) {
                if(tlh==blh) brh=trh;
                else if(trh==tlh) brh=blh;
                else brh=find_head(x+wo,y+ho);
            }
        } else if(clz==2) {
            // two corners are inside screens - if they are both the same screen, make all corners the same
            if(tlh==trh && tlh>=0) {
                blh = tlh;
                brh = trh;
            } else if(tlh==blh && tlh>=0) {
                trh = tlh;
                brh = blh;
            } else if(trh==brh && trh>=0) {
                blh = brh;
                tlh = trh;
            } else if(blh==brh && blh>=0) {
                tlh = blh;
                trh = brh;
            } else if(blh==trh && blh==-1) {
                // opposite corners (bl & tr) are off the screens, so find nearest for each
                blh = find_head(x,y+ho);
                trh = find_head(x+wo,y);
            } else if(brh==tlh && brh==-1) {
                // opposite corners (tl & br) are off the screens, so find nearest for each
                tlh = find_head(x,y);
                brh = find_head(x+wo,y+ho);
            } else {
                // two corners not on same screen, so just find nearest edge for other two
                if(tlh<0) tlh = find_head(x,y);
                if(blh<0) blh = find_head(x,y+ho);
                if(trh<0) trh = find_head(x+wo,y);
                if(brh<0) brh = find_head(x+wo,y+ho);
            }
        } else if(tlh==brh && clz==0) {
          // doesn't span heads, so leave it alone...
          head = -1;
          *flag = tlh;
          return;
        }  
      }
// first check if the ratio of the width to height is roughly the same as the global screen
// (some clients, like Uniras, make their windows 'fit' to the screen ratio - Doh!)
        if(DisplayWidth(dpy,screen)*20/wo==DisplayHeight(dpy,screen)*20/ho) {
            *w = wo*screen_width(head)/DisplayWidth(dpy,screen);
            *h = ho*screen_height(head)/DisplayHeight(dpy,screen);
            if(gVerboseConsole)
              printf("Changed width & height from %d x %d to %d x %d\n",wo,ho,*w,*h);
        } else {
// ...or check if the width is too big for the current head
          if(wo>screen_width(head)) {
            *w = wo*screen_width(head)/DisplayWidth(dpy,screen);
            if(gVerboseConsole)
              printf("Too wide for head %d (%d)- changed from %d to %d\n",head,screen_width(head),wo,*w);
          }
// ...and if the height is too big for the current head
          if(ho>screen_height(head)) {
            *h = ho*screen_height(head)/DisplayHeight(dpy,screen);
            if(gVerboseConsole)
              printf("Too tall for head %d (%d)- changed from %d to %d\n",head,screen_height(head),ho,*h);
          }
        }
      
// ...then check if it was trying to centre it horizontally on entire X display (within 15 pixels)
      if(abs((x+(wo/2))-(DisplayWidth(dpy, screen)/2))<15) {
        // if so place it on current head instead
        *xr=screen_x(head)+(screen_width(head)/2)-(*w/2);
        if(gVerboseConsole)
          printf("Got horizontal centre: %d\n",*xr);
      } else if(x<DisplayWidth(dpy,screen)-x-wo) {
        // position it horizontally on current head in same ratio as whole display
        if(x<0) x=0;
        *xr=screen_x(head)+(screen_width(head)*x/DisplayWidth(dpy,screen));
        if(gVerboseConsole)
          printf("Nearest left side: %d %d %d\n",*xr,xo,DisplayWidth(dpy,screen)-xo-wo);
      } else {
        // position it horizontally on current head in same ratio as whole display
        x = DisplayWidth(dpy,screen)-x-wo;
        if(x<0) x=0;
        *xr=screen_x(head)+screen_width(head)-(screen_width(head)*x/DisplayWidth(dpy,screen));
        if(gVerboseConsole)
          printf("Nearest right side: %d %d %d\n",*xr,xo,DisplayWidth(dpy,screen)-xo-wo);
      }
    
// ...then check for vertical centre...
      if(abs((y+(ho/2))-(DisplayHeight(dpy, screen)/2))<15) {
        // put it in the centre of the current head
        *yr=screen_y(head)+(screen_height(head)/2)-(*h/2);
        if(gVerboseConsole)
          printf("Got vertical centre: %d\n",*yr);
      } else if(y<DisplayHeight(dpy,screen)-y-ho) {
        // put it towards the top of the current head in the same ratio as whole display
        if(y<0) y=0;
        *yr=screen_y(head)+(screen_height(head)*y/DisplayHeight(dpy,screen));
        if(gVerboseConsole)
          printf("Nearest top: %d %d %d\n",*yr,yo,DisplayHeight(dpy,screen)-yo-ho);
      } else {
        // put it towards the bottom of the current head in the same ratio as whole display
        y = DisplayHeight(dpy,screen)-y-ho;
        if(y<0) y=0;
        *yr=screen_y(head)+screen_height(head)-*h-(screen_height(head)*y/DisplayHeight(dpy,screen));
        if(gVerboseConsole)
          printf("Nearest bottom: %d %d %d\n",*yr,yo,DisplayHeight(dpy,screen)-yo-ho);
      }
      if(gVerboseConsole)
        printf("XINERAMA: Found head %d... set position to %d,%d\n",head,*xr,*yr);

}
#endif /* XINERAMA */

/*
 *
 * Add directory to font path
 *
 */

void add_dir_to_font_path(char *font_dir)
{
  char **font_path, **new_font_path, *file_name;
  int npaths;
  FILE *file;

  sprintf(lastsub,"TRACE1: add_dir_to_font_path\n");
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: add_dir_to_font_path\n");
#endif

  if((file_name=malloc(strlen(font_dir)+strlen("fonts.dir")+2))==NULL)
  {
    if(on_cmd_line) {
      fprintf(stderr, "%s: Memory allocation failed in function add_dir_to_font_path.\n", PROGNAME);
      ClearUpSockets();
      exit(1);
    } else {
      fprintf(stderr, "%s: Memory allocation failed in function add_dir_to_font_path.\n", PROGNAME);
      ShowAlert(FATALMEM_NAME,0,0);
      gQuitFlag=FATALMEM_ERRNO;
      return;
    }
  }

  if((file=fopen(file_name, "r")))
  {
    fclose(file);
    font_path=XGetFontPath(dpy, &npaths);
    if((new_font_path=malloc(sizeof(char *)*(npaths+1)))==NULL)
    {
      if(on_cmd_line) {
        fprintf(stderr, "%s: Memory allocation failed in function add_dir_to_font_path.\n", PROGNAME);
        ClearUpSockets();
        exit(1);
      } else {
        fprintf(stderr, "%s: Memory allocation failed in function add_dir_to_font_path.\n", PROGNAME);
        ShowAlert(FATALMEM_NAME,0,0);
        gQuitFlag=FATALMEM_ERRNO;
        return;
      }
    }
    memcpy(new_font_path, font_path, sizeof(char*)*npaths);
    new_font_path[npaths]=font_dir;
    XSetFontPath(dpy, new_font_path, npaths+1);
    XFreeFontPath(font_path);
    free(new_font_path);
  }
}

/*
 *
 * Remove directory from font path
 *
 */

void rem_dir_from_font_path(char *font_dir)
{
  char **font_path, **new_font_path;
  int npaths, i, j=0;

  sprintf(lastsub,"TRACE1: rem_dir_from_font_path\n");
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: rem_dir_from_font_path\n");
#endif

  font_path=XGetFontPath(dpy, &npaths);
  if((new_font_path=malloc(sizeof(char *)*npaths))==NULL)
  {
    if(on_cmd_line) {
      fprintf(stderr, "%s: Memory allocation failed in function rem_dir_from_font_path.\n", PROGNAME);
      ClearUpSockets();
      exit(1);
    } else {
      fprintf(stderr, "%s: Memory allocation failed in function rem_dir_from_font_path.\n", PROGNAME);
      ShowAlert(FATALMEM_NAME,0,0);
      gQuitFlag=FATALMEM_ERRNO;
      return;
    }
  }
  for(i=0;i<npaths;i++)
  {
    if(strcmp(font_path[i], font_dir))
      new_font_path[j++]=font_path[i];
  }
  XSetFontPath(dpy, new_font_path, j);
  XFreeFontPath(font_path);
  free(new_font_path);
}

/* 
 * 
 * Get mouse x any y 
 * 
 */ 

void get_mouse_xy(int *x, int *y) 
{ 
  Window w1, w2; 
  int wx, wy, mask; 
 
//  sprintf(lastsub,"TRACE3: get_mouse_xy\n"); 
//  subtrace();
//#ifdef SHOW_TRACE3 
//  printf("TRACE3: get_mouse_xy\n"); 
//#endif 
  if(gX11InFront==0 && mouse_mode==MOUSE_NORMAL && replay_if_drag==0  && can_send_hackcmds==0) {
    // if XDarwin is behind, then it does not track the cursor - so use values set by main
    *x=real_x;
    *y=real_y;
//    XWarpPointer(dpy, None, root, 0, 0, 0, 0, *x, *y);
  } else {
    XQueryPointer(dpy, root, &w1, &w2, x, y, &wx, &wy, &mask);
//    if(mouse_mode==MOUSE_NORMAL && ((w2!=root && w2!=0) || client_list==NULL))
    if(can_send_hackcmds==0)
      if(mouse_mode==MOUSE_NORMAL) {
        Compare_Screens(real_x,real_y,*x,*y,1);	// want to check to see how X11 and 'real' screen compare
        if(*x!=real_x || *y!=real_y) {
          printf("   changing cursor pos in get_mouse: %d,%d  to  %d,%d",*x,*y,real_x,real_y);
          *x=real_x;
          *y=real_y;
        }
      } else if(*x!=real_x || *y!=real_y) {
          printf("   cannot change cursor pos in get_mouse (%d): %d,%d  cf  %d,%d",mouse_mode,*x,*y,real_x,real_y);
      }
  }
} 
 
/* 
 * 
 * Get window under pointer 
 * 
 */ 

Window get_window_container(int x, int y, Window w)
{
  Window *windows=NULL, w1, w2, top=0; 
  int count, i; 
  XWindowAttributes attr; 
 
//  sprintf(lastsub,"TRACE3: get_window_container (w: %#lx  x,y: %d,%d)\n",w,x,y); 
//  subtrace();
//#ifdef SHOW_TRACE3
//  printf("TRACE3: get_window_container (w: %#lx  x,y: %d,%d)\n",w,x,y); 
//#endif
 
  XGetWindowAttributes(dpy, w, &attr); 
  x=x-attr.x;
  y=y-attr.y;

//  sprintf(lastsub,"  doing XQueryTree in get_window_container\n"); 
//  subtrace();
  XQueryTree(dpy, w, &w1, &w2, &windows, &count); 
#ifdef DEBUG
  printf("  attr x,y,w,h & children: %d,%d %d,%d  %d\n",attr.x,attr.y,attr.width,attr.height,count);
#endif

//  sprintf(lastsub,"  done XQueryTree in get_window_container: %d (%#lx)\n",count,windows); 
//  subtrace();
  
  for(i=count-1;i>=0;i--) 
  { 
    XGetWindowAttributes(dpy, windows[i], &attr);
#ifdef DEBUG
    sprintf(lastsub,"  child %d (%#lx): %d,%d %d,%d\n",i,windows[i],attr.x,attr.y,attr.width,attr.height);
    subtrace();
#endif
    if(attr.map_state==IsViewable)
      if(x>=attr.x)
        if(x<attr.x+attr.width)
          if(y>=attr.y)
            if(y<attr.y+attr.height) {
              top=windows[i];
              break; }
  } 

  if(windows) XFree(windows);
  
//  sprintf(lastsub,"TRACE3: leaving get_window_container (top: %#lx)\n",top); 
//  subtrace();
//#ifdef SHOW_TRACE3
//  printf("TRACE3: leaving get_window_container (top: %#lx)\n",top); 
//#endif
 
  return top; 
}

Window get_pointer_win(Window w) 
{ 
  Window w1=0, w2=0, w3=0; 
  int x, y, wx, wy, mask; 
 
//  sprintf(lastsub,"TRACE3: get_pointer_win (w: %#lx)\n",w); 
//  subtrace();
//#ifdef SHOW_TRACE3 
//  printf("TRACE3: get_pointer_win (w: %#lx)\n",w); 
//#endif 

  if(w==0) return w;
  
  if(can_send_hackcmds!=0) {
//    MyIncreaseXServerGrab();
//    XMoveWindow(dpy,cursor_follow_win,-100,-100);
    XQueryPointer(dpy, w, &w1, &w2, &x, &y, &wx, &wy, &mask);
//    XMoveWindow(dpy,cursor_follow_win,x-30,y-30);
//    XSync(dpy,0);
//    MyDecreaseXServerGrab();
  } else if((gX11InFront!=0 && gX11MouseStuck==0) || mouse_mode!=MOUSE_NORMAL || replay_if_drag) {
    if(w==root) {
      w3=get_window_container(real_x,real_y,w);
      if(w3==0) {
//        printf("++ cursor in window: %#lx (%#lx)\n",w3,w3,lastenter);
        if(mouse_mode==MOUSE_NORMAL) {
          XQueryPointer(dpy, w, &w1, &w2, &x, &y, &wx, &wy, &mask);
          // want to check to see how X11 and 'real' screen compare
          if(w2==0)
            Compare_Screens(real_x,real_y,x,y,1);
          else
            Compare_Screens(real_x,real_y,x,y,0);
        }
//  sprintf(lastsub,"TRACE3: leaving get_pointer_win with nothing\n"); 
//  subtrace();
//#ifdef SHOW_TRACE3 
//  printf("TRACE3: leaving get_pointer_win with nothing\n"); 
//#endif 
        return 0;
      }
    }
    XQueryPointer(dpy, w, &w1, &w2, &x, &y, &wx, &wy, &mask);
//    if(w==root && mouse_mode==MOUSE_NORMAL && ((w2!=root && w2!=0) || client_list==NULL))
//    if(w==root) {
//      printf("++ cursor in window: %#lx, %#lx (%#lx)\n",w2,w3,lastenter);
//    }
    if(w==root && mouse_mode==MOUSE_NORMAL) {
      // want to check to see how X11 and 'real' screen compare
      if(w2==w3)
         Compare_Screens(real_x,real_y,x,y,1);
      else
         Compare_Screens(real_x,real_y,x,y,0);
    }
  } else {    // if XDarwin is behind, then it does not track the cursor - so use x,y values set by main
    w2=get_window_container(real_x,real_y,w);
//    if(w==root) {
//      printf("-- cursor in window: %#lx (%#lx, %#lx)\n",w2,lastenter,check_win);
//    }
  }
  
//  sprintf(lastsub,"TRACE3: leaving get_pointer_win (w2: %#lx)\n",w2); 
//  subtrace();
//#ifdef SHOW_TRACE3 
//  printf("TRACE3: leaving get_pointer_win (w2: %#lx)\n",w2); 
//#endif 

  return w2; 
} 
 
/* 
 * 
 * Get window parent 
 * 
 */ 

Window get_window_parent(Window win) 
{ 
  Window parent, dummyw, *windows=NULL; 
  int count; 
 
//  sprintf(lastsub,"TRACE3: get_window_parent (win: %#lx)\n",win); 
//  subtrace();
//#ifdef SHOW_TRACE3 
//  printf("TRACE3: get_window_parent (win: %#lx)\n",win); 
//#endif 
 
  XQueryTree(dpy, win, &dummyw, &parent, &windows, &count); 
  if(windows) XFree(windows); 
  return parent; 
} 
 
/* 
 * 
 * Screen width 
 * 
 */ 
int screen_width(int head) 
{ 
  sprintf(lastsub,"TRACE3: screen_width (head: %d)\n",head); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: screen_width (head: %d)\n",head); 
#endif 
 
#ifdef XINERAMA_SUPPORT 
  if(xinerama) 
    return xinerama_screens[head].width; 
#endif /* XINERAMA */ 
  return XDisplayWidth(dpy, screen); 
} 
 
/* 
 * 
 * Screen height 
 * 
 */ 

int screen_height(int head) 
{ 
  sprintf(lastsub,"TRACE3: screen_height (head: %d)\n",head); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: screen_height (head: %d)\n",head); 
#endif 
 
#ifdef XINERAMA_SUPPORT 
  if(xinerama) 
    return xinerama_screens[head].height; 
#endif /* XINERAMA */ 
  return XDisplayHeight(dpy, screen); 
} 
 
/* 
 * 
 * Screen x 
 * 
 */ 

int screen_x(int head) 
{ 
  sprintf(lastsub,"TRACE3: screen_x (head: %d)\n",head); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: screen_x (head: %d)\n",head); 
#endif 
 
#ifdef XINERAMA_SUPPORT 
  if(xinerama) 
    return xinerama_screens[head].x_org; 
#endif /* XINERAMA */ 
  return 0; 
} 
 
/* 
 * 
 * Screen y 
 * 
 */ 

int screen_y(int head) 
{ 
  sprintf(lastsub,"TRACE3: screen_y (head: %d)\n",head); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: screen_y (head: %d)\n",head); 
#endif 
 
#ifdef XINERAMA_SUPPORT 
  if(xinerama) 
    return xinerama_screens[head].y_org; 
#endif /* XINERAMA */ 
  return 0; 
} 
 
#ifdef XINERAMA_SUPPORT 
/*
 * glue routine for main to get count of xinerama heads
 */

int count_xinerama_heads()
{
  update_xinerama_screens();
  return xinerama_heads;
}
 
/*
 * Glue into Carbon routines to ensure xinerama info is up to date
 */
void update_xinerama_screens()
{
    int count=(xinerama==0?1:Count_Carbon_Screens(1));
    int i=0;
    int offset_x=0, offset_y=0, max_x=-30000, max_y=-30000, min_x=30000, min_y=30000;
    
    if( count < 0 ) return;
    if(xinerama)
        Carbon_To_X11(&offset_x,&offset_y);
    else {
        offset_y = -22;
        return;
    }
    
    check_clients_onscreen=GetCarbonTickCount()+60;
    printf("Updating Xinerama via Carbon... Offset: %d %d (on %d screens)\n",offset_x,offset_y,count);
    
    xinerama_heads = 0;
    while(i<count) {
        int x, y, w, h;
        Get_Carbon_Screen_Bounds(i, &x, &y, &w, &h);
        if( w>0 && h>0 ) {
            int j = 0;
            x += offset_x;
            y += offset_y;
            printf("Bounds for %d:  (%d,%d) [%dx%d]  ",i,x,y,w,h);
            while(j<xinerama_heads) {
                if( x >= xinerama_screens[j].x_org && y >= xinerama_screens[j].y_org &&
                    x+w <= xinerama_screens[j].x_org+xinerama_screens[j].width &&
                    y+h <= xinerama_screens[j].y_org+xinerama_screens[j].height ) {
                    printf("Found container (%d)...\n",j);
                    j = 10000;
                } else if( x <= xinerama_screens[j].x_org && y <= xinerama_screens[j].y_org &&
                    x+w >= xinerama_screens[j].x_org+xinerama_screens[j].width &&
                    y+h >= xinerama_screens[j].y_org+xinerama_screens[j].height ) {
                    printf("Found subscreen (%d)...\n",j);
                    xinerama_screens[j].x_org  = x;
                    xinerama_screens[j].y_org  = y;
                    xinerama_screens[j].width  = w;
                    xinerama_screens[j].height = h;
                    j = 10000;
                } else
                    j++;
            }
            if( j < 10000 ) {
                printf("adding...\n");
                xinerama_screens[xinerama_heads].x_org  = x;
                xinerama_screens[xinerama_heads].y_org  = y;
                xinerama_screens[xinerama_heads].width  = w;
                xinerama_screens[xinerama_heads].height = h;
                xinerama_heads++;
                if( x+w>max_x ) max_x = x + w;
                if( y+h>max_y ) max_y = y + h;
                if( x<min_x ) min_x = x;
                if( y<min_y ) min_y = y;
            }
        }
        i++;
    }
    printf("%s info: Telling XDarwin to update root window (%d,%d) [%dx%d]...\n"
                                    ,PROGNAME,min_x,min_y,max_x-min_x,max_y-min_y);
    MyIncreaseXServerGrab(); // don't want anything else interfering with the X server while this happens
    XMoveWindow(dpy, check_win, HACKCMD_UPDATEROOTW, max_x-min_x);
    XMoveWindow(dpy, check_win, HACKCMD_UPDATEROOTH, max_y-min_y);
//    XMoveWindow(dpy, check_win, HACKCMD_UPDATEROOTX, min_x);
//    XMoveWindow(dpy, check_win, HACKCMD_UPDATEROOTY, min_y);
    XMoveWindow(dpy, check_win, HACKCMD_UPDATEROOTX, 0);
    XMoveWindow(dpy, check_win, HACKCMD_UPDATEROOTY, 0);
    XMoveWindow(dpy, check_win, -1000, -1000);
    XSync(dpy,0);
    MyDecreaseXServerGrab(); // release the X server, if it was before

    (&((_XPrivDisplay)dpy)->screens[screen])->width = max_x-min_x;
    (&((_XPrivDisplay)dpy)->screens[screen])->height = max_y-min_y;
    printf("%s info: XDisplayWidth/Height returns %d x %d\n",
            PROGNAME,XDisplayWidth(dpy,screen),XDisplayHeight(dpy,screen));
}

/* 
 * 
 * Find head 
 * 
 */ 

int generic_find_head(int x, int y, int strict) 
{ 
  int head, min_dist, closest_head; 
 
  sprintf(lastsub,"TRACE3: find_head (x,y: %d,%d)\n",x,y); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: find_head (x,y: %d,%d)\n",x,y); 
#endif 
 
  for(head=0;head<xinerama_heads;head++) 
  { 
    if(x>=xinerama_screens[head].x_org && x<xinerama_screens[head].x_org+xinerama_screens[head].width && y>=xinerama_screens[head].y_org && y<xinerama_screens[head].y_org+xinerama_screens[head].height) 
      return head; 
  } 
  
  // this second check allows the rightmost and bottommost pixel just off the global screen
  for(head=0;head<xinerama_heads;head++) 
  { 
    if(x>=xinerama_screens[head].x_org && x<=xinerama_screens[head].x_org+xinerama_screens[head].width && y>=xinerama_screens[head].y_org && y<=xinerama_screens[head].y_org+xinerama_screens[head].height) 
      return head; 
  } 
  
  /* no head found, so return negative if strict search was required */
  if(strict!=0) return -1;
  
#ifdef DEBUG 
  printf("  no head found, looking for nearest head...\n"); 
#endif 
  
  /* No head found, return head with closest edge */ 
  min_dist=XDisplayWidth(dpy,screen);
  if(XDisplayHeight(dpy,screen)>min_dist) min_dist=XDisplayHeight(dpy,screen); 
  closest_head=0; 
 
  for(head=0;head<xinerama_heads;head++) 
  { 
    int cx = xinerama_screens[head].x_org - x;
    int cy = xinerama_screens[head].y_org - y;
    int dx = xinerama_screens[head].x_org+xinerama_screens[head].width - x;
    int dy = xinerama_screens[head].y_org+xinerama_screens[head].height - y;
    int dist=cx*cx;
    cx = dist;
    cy = cy*cy;
    dx = dx*dx;
    dy = dy*dy;
 
    if(dx<dist) dist=dx;
    if(cy<dist) dist=cy;
    if(dy<dist) dist=dy;
     
    if(dist<min_dist) { 
      closest_head=head; 
      min_dist=dist; 
    } 
  } 
 
  return closest_head; 
} 

int find_head(int x, int y) 
{
    return generic_find_head(x,y,0);
}

int strict_find_head(int x, int y) 
{
    return generic_find_head(x,y,1);
}

int find_rect_head(int x, int y, int w, int h)
{
    int htl, hbr, hcn;
    
    htl=strict_find_head(x,y);
    hbr=strict_find_head(x+w,y+h);
    hcn=strict_find_head(x+w/2,y+h/2);
    if(htl==hbr && htl>=0)
        return htl;
    else if(htl==hcn && htl>=0)
        return htl;
    else if(hbr==hcn && hbr>=0)
        return hbr;
    else if(hcn>=0)
        return hcn;
    else if(htl>=0)
        return htl;
    else if(hbr>=0)
        return hbr;
    else
        return -1;
}

#endif /* XINERAMA */ 
 
#ifdef I18N_SUPPORT 
/* 
 * 
 * some wrapper functions for I18N 
 * 
 */ 

Status XGetWMNameWrapper(Display *dpy, Window window, XTextProperty *text_prop) 
{  
  int status; 
  char **list; 
  int num; 
  
  status = XGetWMName(dpy, window, text_prop); 
  if (!status || !text_prop->value || text_prop->nitems <= 0) 
    return 0; 
  if (text_prop->encoding == XA_STRING) 
    return 1; 
   
  text_prop->nitems = strlen((char *)text_prop->value); 
  status = XmbTextPropertyToTextList(dpy, text_prop, &list, &num); 
   
  if (status >= Success && num > 0 && *list)
  { 
    if(text_prop->value) XFree(text_prop->value); 
    text_prop->value = (unsigned char*)strdup(*list); 
    text_prop->nitems = strlen(*list); 
    if(list) XFreeStringList(list); 
    return 1; 
  } 
   
  return 0; 
} 
 
Status XFetchNameWrapper(Display *dpy, Window window, char **name) 
{  
  XTextProperty text_prop; 
   
  if (XGetWMNameWrapper(dpy, window, &text_prop) != 0)
  { 
    if (text_prop.value && text_prop.nitems > 0)
    { 
      *name = (char *)text_prop.value; 
      return 1; 
    } 
  } 
  *name = NULL; 
  return 0; 
} 
 
XFontSet XLoadQueryFontWrapper(Display *dpy, const char *font_name) 
{ 
  char **miss, *def=NULL; 
  int n_miss; 
  char *base_font_name_list=NULL; 
  int font_name_list_length; 
  XFontSet _fontset; 
 
  font_name_list_length = strlen(font_name) + 3; 
  base_font_name_list = (char *) malloc(sizeof(char) * font_name_list_length); 
  if (!base_font_name_list)
  { 
    if(on_cmd_line) {
      fprintf(stderr, "%s: Memory allocation failed in function XLoadQueryFontWrapper.\n", PROGNAME);
      ClearUpSockets();
      exit(1);
    } else {
      fprintf(stderr, "%s: Memory allocation failed in function XLoadQueryFontWrapper.\n", PROGNAME);
      ShowAlert(FATALMEM_NAME,0,0);
      gQuitFlag=FATALMEM_ERRNO;
      return NULL;
    }
  }
  sprintf(base_font_name_list, "%s,*", font_name); 
  _fontset = XCreateFontSet(dpy, base_font_name_list, &miss, &n_miss, &def); 
  if (!_fontset)
  {
    free(base_font_name_list);
    return NULL;
  }
  if (n_miss)
  { 
    int j; 
    if(on_cmd_line) {
      fprintf(stderr, "%s: XLoadQueryFontWrapper(): missing charset(s)\n", PROGNAME); 
      for (j=0; j<n_miss; j++)  
        fprintf(stderr, "\t%s\n", miss[j]); 
    }
    XFreeStringList(miss); 
  } 
  free(base_font_name_list);
  return _fontset; 
} 
#endif /* I18N */ 
