/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 
 
#include "oroborus.h" 
#include "hackcmds.h"
#include "iconbars.h"
#include "client_properties.h"

/* 
 * 
 * Handle different window gravities 
 * 
 */ 

void gravitate_client(Client *c, int grav) 
{ 
  int dx=0, dy=0, orgicon; 
 
  sprintf(lastsub,"TRACE3: gravitate_client (window: %#lx)\n",C_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: gravitate_client (window: %#lx)\n",C_WINDOW); 
#endif 
 
  // temporarily switch off iconbar flag, so defines are correct
  orgicon=c->is_icon;
  c->is_icon=NOT_ICONBAR;
  
  switch(GRAVITY) 
  { 
  case NorthWestGravity: 
    dx=LW; 
    dy=TH; 
    break; 
  case NorthGravity: 
    dx=(LW+RW)/2; 
    dy=TH; 
    break; 
  case NorthEastGravity: 
    dx=-RW; 
    dy=TH; 
    break; 
  case WestGravity: 
    dx = LW; 
    dy=(TH+BH)/2; 
    break; 
  case EastGravity: 
    dx=-RW; 
    dy=(TH+BH)/2; 
    break;  
  case SouthWestGravity: 
    dx=LW; 
    dy=-BH; 
    break; 
  case SouthGravity: 
    dx=(LW+RW)/2; 
    dy=-BH; 
    break; 
  case SouthEastGravity: 
    dx=-RW; 
    dy=-BH; 
    break; 
  case CenterGravity: 
    dx=(LW+RW)/2; 
    dy=(TH+BH)/2; 
    break; 
  } 
  c->x=c->x+(dx*grav); 
  c->y=c->y+(dy*grav);
  
  // switch back to original iconbar state
  c->is_icon=orgicon;
} 

/* 
 * 
 * Create a new frame part 
 * 
 */ 

Window create_frame_part(Window win, int x, int y, int w, int h, int mapit) 
{ 
  Window new;   
 
  sprintf(lastsub,"TRACE3: create_frame_part %#lx (%#lx)\n",win,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: create_frame_part %#lx (%#lx)\n",win,FC_WINDOW); 
#endif 
  
  if(w<1 || h<1) return 0;
  
  new=XCreateSimpleWindow(dpy, win, x, y, w, h, 0, 0, 0);

  XSelectInput(dpy, new, ButtonPressMask|ButtonReleaseMask); 
  if( mapit ) XMapWindow(dpy, new); 
  return new; 
} 

int still_in_auto_client()
{
  Client *c;
  Window ptrwin;
  
  c = lastfocus_client;
  
  if(lastfocus_client==NULL) return 0;
  if((lastfocus_client->auto_properties & AUTO_FOCUS)==0) return 0;
  
  ptrwin=get_pointer_win(root);
  
  if(lastfocus_client->frame!=ptrwin) return 0;
  
  if(gVerboseConsole)
    printf("Still in autofocus client...\n");
  return 1;
}

int create_new_hackid()
{
    Client *ac;
    int i, maxpos, minid, maxid=0, gotanid=0, tryid=0;
    
    maxpos = XDisplayWidth(dpy, screen)-1;
    minid = maxpos;
    
    ac = client_list;
    if(ac==NULL) return maxpos;
    
    for(i=0;i<client_count;i++) {
        if(ac->hack_id > maxid) maxid = ac->hack_id;
        if(ac->hack_id < minid) minid = ac->hack_id;
        ac=ac->next;
    }
    
    if(minid>1) return minid-1;
    if(maxid<maxpos) return maxid+1;
    
    tryid = minid;
    while(gotanid==0) {
        tryid++;
        if(tryid>maxpos) return 0;
        ac = client_list;
        for(i=0;i<client_count;i++) {
            if(ac->hack_id==tryid) break;
            ac=ac->next;
        }
        if(ac->hack_id!=tryid) gotanid=tryid;
    }
    
    return gotanid;
}

/*
 * Add this client to parent's sibling list
 */

void add_sibling_client(Client *c)
{
  Client *cp, *cnxt;
  
  sprintf(lastsub,"TRACE3: add_sibling_client (window: %#lx)\n",C_WINDOW);
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: add_sibling_client (window: %#lx)\n",C_WINDOW);
#endif

  cp=c->parent;
  if(!cp) return;
  
  cnxt=cp->first_child;
  
  if(cnxt) {
    while(cnxt->next_sibling) {
      cnxt=cnxt->next_sibling;
    }
    cnxt->next_sibling=c;
    c->prev_sibling=cnxt;
    c->next_sibling=NULL;
  } else {
    cp->first_child=c;
    c->next_sibling=NULL;
    c->prev_sibling=NULL;
  }
  cp->child_count++;
}

/* 
 * 
 * Create a new client frame 
 * 
 */ 

void create_new_client_old(Window w, int reposition) 
{ 
  Client *c=NULL; 
  XWindowAttributes attr; 
  XSetWindowAttributes wattr; 
  long dummy; 
  Window parentwin;
  XWindowChanges wc; 
  int x, y, xmax, ymax, xmin, ymin, head=0, showfront=0, retval, new_hack_id=0;
  int inhibit_focus=0, need_ungrab=0, maketrans=0; 
  XClassHint xch;
  char **argv=NULL;
 
  sprintf(lastsub,"TRACE1: create_new_client  win: %#lx (%#lx)\n",w,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: create_new_client  win: %#lx (%#lx)\n",w,FC_WINDOW); 
#endif 
 
  if(gQuitFlag!=0) return;
    
  MyIncreaseXServerGrab();
  /* Get size and position of client */
  retval = XGetWindowAttributes(dpy, w, &attr);
  if(retval!=1) {
    printf("%s warning: XGetWindowAttributes returned status %d - aborting...\n",PROGNAME,retval);
    MyDecreaseXServerGrab();
    return;
  }

  if((c=malloc(sizeof *c))==NULL)
  {
    if(on_cmd_line)
        fprintf(stderr, "%s: Unable to allocate memory for new client.\n", PROGNAME);
    MyDecreaseXServerGrab();
    ClearUpSockets();
    exit(1);
  }
 
  /* Initialize client values */ 
  c->window=w;
  c->cover=0;
  c->al_list=NULL;
  c->al_count=0;
  c->ignore_unmap=0; 
  c->zoom_state=BUTTON_UNPRESSED; 
  c->icon_state=BUTTON_UNPRESSED;
  c->close_state=BUTTON_UNPRESSED; 
  c->can_max=True; 
  c->has_frame=True; 
  c->has_focus=False; 
// assume it isn't an iconbar
  c->is_icon=NOT_ICONBAR;
// start with some "bad" values...
  c->name_x=-1;
  c->name_y=-1;
  c->name_w=-1;
  c->name_h=-1;
  c->icon_x=-10000;
  c->icon_y=-10000;
  c->icon_h=-1;
  c->icon_w=-1;
  c->icontxt_x=-1;
  c->icontxt_y=-1;
  c->icontxt_w=-1;
  c->icontxt_h=-1;
  c->isoffscreen=0;
  c->osx = 0;
  c->osy = 0;
// default to a modeless window
  c->modal_state=MODELESS | DONT_IGNORE_MODAL;
  c->sloppy_family=FOCUS_CLICK;
  c->auto_properties=AUTO_NONE;
  if(config.focus_style==FOCUS_SLOPPY) c->auto_properties |= AUTO_FOCUS;
// set up initial sibling/child list
  c->child_count=0;
  c->parent=NULL;
  c->first_child=NULL;
  c->next_sibling=NULL;
  c->prev_sibling=NULL;
// initialise the OS X window identifier
  c->osxwinfo.winref=0;
  c->osxwinfo.group_winref=0;
  c->osxwinfo.view_state=VIEWSTATE_VISIBLE;
// Start the string info pointing to nothing
  c->clientclass=NULL;
  c->clientname=NULL;
  c->clientcmd=NULL;
  c->clienthost=NULL;
  c->clientgrouporg=NULL;
  c->clientgroup=NULL;
  c->name=NULL;
  c->is_shaped=1;
  c->osxwinfo.translucency=100;
  c->osxwinfo.inactive_translucency=85;
  c->is_dimmed=0;
  c->is_docked=0;
  c->is_e_m_r=0;
  c->dock_timer=0;
  c->undock_timer=0;
  c->osxwinfo.drawalpha=255;
  c->inhibit_level=HACKARG_INHIBITNONE;
  c->real_w=0;
  c->real_h=0;
  c->hack_id=0;
  
  /* Set event mask for client */ 
  XSelectInput(dpy, c->window, PropertyChangeMask|FocusChangeMask); 
 
  new_hack_id=create_new_hackid();

#ifdef I18N_SUPPORT 
  if(use_fontset)
    /* call wrapper function for I18N */ 
    XFetchNameWrapper(dpy, c->window, &c->name); 
  else
    XFetchName(dpy, c->window, &c->name);
#else 
  XFetchName(dpy, c->window, &c->name); 
#endif /* I18N */ 

  if(gQuitFlag!=0) {
    MyDecreaseXServerGrab();
    free(c);
    return;
  }

  c->size=XAllocSizeHints(); 
  XGetWMNormalHints(dpy, w, c->size, &dummy); 
 
  if(gQuitFlag!=0) {
    MyDecreaseXServerGrab();
    free(c);
    return;
  }

  /* Handle clients with border set > 0 and adjust position accordingly */ 
  if(attr.border_width>0) 
  { 
    switch(GRAVITY) 
    { 
    case SouthEastGravity: 
      attr.x=attr.x+attr.border_width*2; 
      attr.y=attr.y+attr.border_width*2; 
      break; 
    case SouthWestGravity: 
    case SouthGravity: 
      attr.y=attr.y+attr.border_width*2; 
      break; 
    case NorthEastGravity: 
    case EastGravity: 
      attr.x=attr.x+attr.border_width*2; 
      break; 
    } 
  } 
  c->x=attr.x; 
  c->y=attr.y; 
  c->width=attr.width; 
  c->height=attr.height; 
 
  if(gVerboseConsole) {
    printf("==== NEW WINDOW: %#lx  (org-geom: %d x %d  @ %d , %d)   hints...\n",
                                            c->window,c->width,c->height,c->x,c->y);
    if(c->name)
      printf("  title: %s\n",c->name);
    else
      printf("  no title!\n");
  }
  head=XGetClassHint(dpy,c->window,&xch);
  if(head!=0) {
    if(gVerboseConsole) {
      printf("  res_name: %s\n",xch.res_name);
      printf("  res_class: %s\n",xch.res_class);
    }
    c->clientname=strdup(xch.res_name);
//    c->clientclass=strdup(xch.res_class);
    c->clientgroup=strdup(xch.res_class);
    c->clientclass=strdup(xch.res_class);
    if(xch.res_name) XFree(xch.res_name);
    if(xch.res_class) XFree(xch.res_class);
  }
  
  if(gVerboseConsole) {
    retval=XGetCommand(dpy,c->window,&argv,&head);
    printf("  ret,argc: %d %d\n",retval,head);
    if(argv) {
      printf("  argv: %s\n",*argv);
      XFreeStringList(argv);
    }
  }
  
  c->parent=NULL;
  
  // Work out which group a client will initially belong to.
  // Note this may be overridden by the user's defaults (in new_client_defaults).
  // Sets up various things in the client's structure, but also returns the window's parent-window
  // (if it has one via TransientFor, via WM_CLIENT_LEADER, or via window_group)
  group_new_client(c, attr.map_state, &parentwin);
  
    // set the group name from its parent, if it has one
    if(c->parent!=NULL) {
      add_sibling_client(c);
      if(c->clientgroup) XFree(c->clientgroup);
      c->clientgroup=strdup(c->parent->clientgroup);
    }

/* All of this lot needs to be done in the default properties, rather than being hardwired here */
  maketrans = 0;
  if(c->clientgroup) {
    // make xeyes start with no frames by default, and xterm start translucent...
    if(strlen(c->clientgroup)==5) {
      if(!strncmp(c->clientgroup,"XEyes",5))
        c->has_frame = False;
      else if(!strncmp(c->clientgroup,"XTerm",5))
        maketrans = 1;
    }
    // hack together a check for openoffice...
    if(strlen(c->clientgroup)==11)
      if(strlen(c->clientname)==11)
        if(!strncmp(c->clientgroup,"VCLSalFrame",11))
          if(!strncmp(c->clientname,"VCLSalFrame",11)) {
            XFree(c->clientgroup);
            c->clientgroup=malloc(16);
            sprintf(c->clientgroup,"OpenOffice.org");
          }
    // hack together a check for xv...
    if(c->clientname) {
      if(strlen(c->clientgroup)>1)
        if(strlen(c->clientname)>1)
          if(!strncmp(c->clientgroup,"XV",2))
            if(!strncmp(c->clientname,"xv",2))
              sprintf(c->clientgroup,"XV");
    }
  }

  // deal with windows that have no class
  if(c->clientgroup==NULL) {
    if(c->name) {
      // hack for ghostscript
      if(strlen(c->name)==11 && !strncmp(c->name,"ghostscript",11)) {
        c->clientgroup = malloc(12);
        sprintf(c->clientgroup,"GhostScript");
      } else if(strlen(c->name)<23) {
        c->clientgroup = strdup(c->name);
      } else {
        // truncate the window title to use for the group name
        c->clientgroup = strdup(c->name);
        c->clientgroup[20]=0x2E;
        c->clientgroup[21]=0x2E;
        c->clientgroup[22]=0x2E;
        c->clientgroup[23]=0;
      }
    } else {
      // no name either...
      c->clientgroup=malloc(15);
      sprintf(c->clientgroup,"Untitled Group");
    }
  }
  c->clientgrouporg = c->clientgroup;
  
  if(gVerboseConsole) printf("Original group: %s\n",c->clientgroup);

//  if(c->osxwinfo.winref==-10) {
    // Now that the client group is sorted out, make the OroborOSX proxy window
    add_client_to_group(c);
    //reorder_dockicon_winlist(c);
    //MakeX11Window(&(c->osxwinfo),c->name,c->clientgroup);
    showfront=1;
//  } else if(c->osxwinfo.winref==-1)
//    c->osxwinfo.winref=0; // This is the case when new window should not be listed in the dock's window list

#ifdef SHAPE_SUPPORT 
  /* Select shape events */ 
  if(shape) 
    XShapeSelectInput(dpy, c->window, ShapeNotifyMask); 
#endif /* SHAPE */ 
 
  /* Create frame window */ 
  wattr.override_redirect=True;
  // Added the LeaveWindowMask to this, so we know when we've left the window
  wattr.event_mask=SubstructureNotifyMask|SubstructureRedirectMask|
                   ButtonPressMask|ButtonReleaseMask|EnterWindowMask|LeaveWindowMask;
  c->frame=XCreateWindow(dpy, root, HACKCMD_STARTX, HACKCMD_STARTY, FW, FH, 0,
                        DefaultDepth(dpy, screen), CopyFromParent, DefaultVisual(dpy, screen),
                        CWOverrideRedirect|CWEventMask, &wattr);
  
/*  if(can_send_hackcmds!=0) {
    c->cover=XCreateSimpleWindow(dpy, root, -2000, -2000, 50, 50, 0, 0, 0);
    XSelectInput(dpy, c->cover, ButtonPressMask|ButtonReleaseMask); 
    XMapWindow(dpy,c->cover);
    XMoveWindow(dpy,c->cover,HACKCMD_STARTX,HACKCMD_STARTY);
    XMoveWindow(dpy,c->cover,HACKCMD_SETID,-999);
  }*/
  
  // Set the cursor we want to see when inside the window -would be nice if this was more like the Mac's
  // normal cursor...
  XDefineCursor(dpy, c->frame, normal_curs); 

  /* This ignores the unmap event when reparenting */ 
  if(attr.map_state==IsViewable) 
    c->ignore_unmap++; 
 
  /* Reparent client window */ 
  XReparentWindow(dpy, c->window, c->frame, LW, TH);
    
  /* Create frame parts if has border and/or title */ 
    // Top frame parts are needed for iconbars, even if no frame
    c->title=create_frame_part(c->frame, TX, TY, TW<1?1:TW, config.top.height, c->has_frame); 
    c->top_left=create_frame_part(c->frame, TLX, TLY, TLW, TLH, c->has_frame); 
    c->top_right=create_frame_part(c->frame, TRX, TRY, TRW, TRH, c->has_frame);  
   
//  if(c->has_frame)
//  { 
    /* Side frame parts */ 
    c->left=create_frame_part(c->frame, LX, LY, LW, LH, c->has_frame); 
    c->right=create_frame_part(c->frame, RX, RY, RW, RH, c->has_frame); 
     
    /* Bottom frame parts */ 
    c->bottom=create_frame_part(c->frame, BX, BY, BW<1?1:BW, config.bottom.height, c->has_frame); 
    c->bottom_left=create_frame_part(c->frame, BLX, BLY, BLW, BLH, c->has_frame); 
    c->resize=create_frame_part(c->frame, RSX, RSY, RSW, RSH, c->has_frame); 
   
    /* Top frame parts 
    c->title=create_frame_part(c->frame, TX, TY, TW<1?1:TW, TH); 
    c->top_left=create_frame_part(c->frame, TLX, TLY, TLW, TLH); 
    c->top_right=create_frame_part(c->frame, TRX, TRY, TRW, TRH);   */
   
    /* Buttons */ 
    c->zoom=create_frame_part(c->frame, HX, HY, HW, HH, c->has_frame); 
    c->icon=create_frame_part(c->frame, IX, IY, IW, IH, c->has_frame);
    c->close=create_frame_part(c->frame, CX, CY, CW, CH, c->has_frame);
//  } 
 
  // Get XDarwin to create the new window - so that the title gets initialised
  XMapWindow(dpy,c->frame);
  // move the window again, so that it does not get 'stuck'
  XMoveWindow(dpy,c->frame,-10000,-10000);
  SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,201,0,0,0,0); // make sure the client is not moved
  redraw_client_frame(c); 

  /* check if the window size and position are sensible */
  if(c->width<2 || c->width>XDisplayWidth(dpy,screen)) {
    printf("%s warning: Got silly width (%d) -setting to 250\n",PROGNAME,c->width);
    c->width=250;
  }
  if(c->height<2 || c->height>XDisplayHeight(dpy,screen)) {
    printf("%s warning: Got silly height (%d) -setting to 200\n",PROGNAME,c->height);
    c->height=200;
  }
  if(c->x<0 || c->x>=XDisplayWidth(dpy,screen)) {
    printf("%s warning: Got silly x co-ord (%d) -setting to 200\n",PROGNAME,c->x);
    c->x=200;
  }
  if(c->y<0 || c->y>=XDisplayHeight(dpy,screen)) {
    printf("%s warning: Got silly y co-ord (%d) -setting to 150\n",PROGNAME,c->y);
    c->y=150;
  }
  
  /* Move the client to position set by hints  -changed this to work more like twm */ 
  if((c->size->flags & USPosition)!=0 ||
        ((c->size->flags & PPosition)!=0 && (c->x>0 || c->y>0)) || !reposition) 
  {
    if(gVerboseConsole)
        printf("Found hints for geometry...   %d x %d  @ %d , %d\n",c->width,c->height,c->x,c->y);
#ifdef XINERAMA_SUPPORT
    if(xinerama)
    {
      if((c->modal_state & STRICT_ANY_MODAL)==0 || c->parent==NULL)
        head=-1;
      else
        head=find_rect_head(c->parent->x,c->parent->y,c->parent->width,c->parent->height);
      check_window_pos(c->x,c->y,&c->width,&c->height,&x,&y,&head);
      /* If window positioned within 15 pixels of center of display, consider it to be centered on the current head */
/*      if(abs((c->x+(c->width/2))-(DisplayWidth(dpy, screen)/2))<15 && abs((c->y+(c->height/2))-(DisplayHeight(dpy, screen)/2))<15)
      {
        get_mouse_xy(&x, &y);
        head=find_head(x, y);
        c->x=screen_x(head)+(screen_width(head)/2)-(c->width/2);
        c->y=screen_y(head)+(screen_height(head)/2)-(c->height/2);
        printf("XINERAMA: Found cursor on head %d... set position to %d,%d\n",head,c->x,c->y);
      }*/
      c->x = x;
      c->y = y;
      //head=find_head(c->x, c->y);
      if(gVerboseConsole) printf("XINERAMA: Window placed on head %d\n",head);
    }
#endif /* XINERAMA */
 
    xmin=screen_x(head)+LW;
    ymin=screen_y(head)+TH;
    if( screen_y(head)==0 && gMenubarSpace!=0 ) ymin+=22;
    xmax=(screen_x(head)+screen_width(head))-c->width-RW; 
    ymax=(screen_y(head)+screen_height(head))-c->height-BH; 
   
    if(gVerboseConsole)
      printf("Got bounds (min/max): %d/%d , %d/%d\n",xmin,xmax,ymin,ymax);
    gravitate_client(c, APPLY_GRAVITY); 
    wc.x=c->x; 
    wc.y=c->y; 
    if(gVerboseConsole)
        printf("Placing via hints: %d x %d @ %d,%d\n",c->width,c->height,c->x,c->y);
 
    /* Keep window fully inside desktop */ 
    if(wc.x>xmax) wc.x=xmax; 
    if(wc.y>ymax) wc.y=ymax; 
    if(wc.x<xmin) wc.x=xmin; 
    if(wc.y<ymin) wc.y=ymin; 
  } 
 
  /* Otherwise follow window placement policy */ 
  else 
  { 
#ifdef XINERAMA_SUPPORT
    if(xinerama)
    {
      head=-1;
      if((c->modal_state & STRICT_ANY_MODAL) && c->parent)
        head=find_rect_head(c->parent->x,c->parent->y,c->parent->width,c->parent->height);
      if(head<0) {
        get_mouse_xy(&x, &y);
        head=find_head(x, y);
        if(gVerboseConsole) printf("XINERAMA: Found cursor (%d,%d) on head %d...\n",x,y,head);
      } else if(gVerboseConsole)
        printf("XINERAMA: Found parent on head %d...\n",head);
    }
#endif /* XINERAMA */

    xmin=screen_x(head)+LW+config.left_margin;
    ymin=screen_y(head)+TH+config.top_margin; 
    if(screen_y(head)==0 && gMenubarSpace!=0) ymin+=22;
    xmax=(screen_x(head)+screen_width(head))-c->width-RW-config.right_margin; 
    ymax=(screen_y(head)+screen_height(head))-c->height-BH-config.bottom_margin; 

    if(gVerboseConsole)
      printf("Got bounds (min/max): %d/%d , %d/%d\n",xmin,xmax,ymin,ymax);

    if(config.window_placement==CENTER_ROOT) 
    { 
      x=screen_x(head)+(screen_width(head)/2); 
      y=screen_y(head)+(screen_height(head)/2);
      if(gVerboseConsole)
        printf("Placing in center of head %d: %d x %d @ %d,%d\n",head,c->width,c->height,x,y);
    } 
    else
    {
      get_mouse_xy(&x, &y); 
      if(gVerboseConsole)
        printf("Placing with cursor: %d x %d @ %d,%d\n",c->width,c->height,x,y);
    }
    wc.x=x-(c->width/2); 
    wc.y=y-(c->height/2); 
 
    /* Keep window fully inside desktop */ 
    if(wc.x>xmax) wc.x=xmax; 
    if(wc.y>ymax) wc.y=ymax; 
    if(wc.x<xmin) wc.x=xmin; 
    if(wc.y<ymin) wc.y=ymin; 
  } 
  wc.height=c->height; 
  wc.width=c->width; 
  wc.border_width=0;
// Add one for the 'jig' to work around the blank windows problem
//  wc.x+=1;

  /* Add client to client list */ 
  if(client_list==NULL) 
  { 
    client_list=c; 
    c->prev=c; 
    c->next=c; 
  } 
  else 
  { 
    c->prev=client_list->prev; 
    c->next=client_list; 
    client_list->prev->next=c; 
    client_list->prev=c; 
  } 
  client_count++; 
  update_gnome_client_list();
  
  need_ungrab = check_need_ungrab();
//  need_ungrab=0;
  
// Make sure the window is updated first time (configure_client must see a change in the
// window dimensions for it to do anything - that's what the '3' does at the end - force a change)
  configure_client(c, CWHeight|CWWidth|CWX|CWY|CWBorderWidth, &wc, False, 3); 
 
  if(get_wm_state(c->window)==IconicState)
    c->dock_timer=-1; // need to make a note to minimise this client
    
  
  if(need_ungrab && last_replay_client==lastfocus_client && gX11InFront!=0) {
    XMoveWindow(dpy,c->frame,-FW-10,-FH-10);
    printf("%s info: found grab - not raising new client\n",PROGNAME);
    show_client(c);
    //stack_client_re_client(c,lastfocus_client,Below);
    XMoveWindow(dpy,c->frame,FX,FY);
    inhibit_focus = 1;
    showfront = 0;
    raise_client(c);
    //XRaiseWindow(dpy,c->frame);
  } else {
    show_client(c);
    if(gVerboseConsole)
      printf("%s info: no grab (%d) - raising new client\n",PROGNAME,need_ungrab);
    inhibit_focus = 0;
    XRaiseWindow(dpy,c->frame);
  //raise_client(c);
  }
  
  c->hack_id = new_hack_id;
  SendHackCommand(c,0,HACKCMD_SETID,0,0,0,0,0); // uses the id stored for client, so args ignored

  if(c->auto_properties & (AUTO_FOCUS|NONAUTO_FIRSTCLICK))
    SendHackCommand(c,1,HACKCMD_SETFIRSTCLICK,1,0,0,0,0);
  
  //SendHackCommand(c,1,HACKCMD_INHIBITFULL,0,0,0,0,0);
  SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,HACKARG_INHIBITNONE,0,0,0,0);
  SendHackCommand(c,1,HACKCMD_DOSHADOW,-1,0,0,0,0); // inhibit shadow changes (tell XDarwin when myself)
    
  /* Button grab for click-to-focus */ 
  XGrabButton(dpy, AnyButton, AnyModifier, c->frame, False, ButtonPressMask|ButtonReleaseMask, GrabModeSync, GrabModeAsync, None, None); 
 
#ifdef KEYBOARD_SUPPORT 
  grab_keys(c->window); 
#endif /* KEYBOARD */ 
 
#ifdef WORKSPACE_SUPPORT 
  set_gnome_hint(c->window, WIN_WORKSPACE, workspace); 
#endif /* WORKSPACE */ 
 
//  SendSimpleHackCommand(HACKCMD_INHIBITINFO,0);
  
  MyDecreaseXServerGrab();

  MakeX11Window(&(c->osxwinfo),c->name,c->clientgroup);

  if(maketrans) {
    if(gVerboseConsole)
      printf("Making %#lx window translucent...\n",C_WINDOW);
    c->osxwinfo.translucency=ACTIVE_TRANSLUCENT1;
    c->osxwinfo.inactive_translucency=INACTIVE_TRANSLUCENT1;
    SendHackCommand(c,1,HACKCMD_TRANSLUCENT,ACTIVE_TRANSLUCENT1,0,0,0,0);
    //Setup_Properties_Menu_Translucency(&c->osxwinfo);
  }
  
  /* Check if we want to focus new clients */ 
  if(c->is_docked!=0 || c->is_icon!=NOT_ICONBAR || !config.focus_new || inhibit_focus ||
    (get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS) || still_in_auto_client()!=0) {
    if((gBGWinSetting & BGWIN_TRANSLUCENT)==BGWIN_TRANSLUCENT)
      SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.inactive_translucency,0,0,0,0);
    if((gBGWinSetting & BGWIN_DIMMED)==BGWIN_DIMMED)
      SendHackCommand(c,1,HACKCMD_DIMWIN,1,0,0,0,0);
    SendHackCommand(c,1,HACKCMD_SETSHADOW,0,0,0,0,0);
    if(lastfocus_client!=NULL)
      if(lastfocus_client->osxwinfo.winref!=0)
         FocusOSXWin(&(lastfocus_client->osxwinfo));
    redraw_client_frame(c);
  } else
    set_focus_client(c);

  SendHackTitleToDockCommand(c,0);

  /* Set client to correct state */   
  if(c->dock_timer==-1) {
    c->dock_timer=0;
    // turn new client into an iconbar/docked-icon
    toggle_minimise_client(c,0);
    lower_client(c);
  } else if(showfront!=0 && gX11InFront==0 && c->is_docked==0 && c->is_icon==NOT_ICONBAR)
    Bring_To_Front(0);
    
  
// 'Jig' it by one pixel (to make sure content appears on secondary screens)...
/*  wc.x=c->x+1; 
  wc.y=c->y; 
  wc.height=c->height; 
  wc.width=c->width; 
  configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, False); 
 
  wc.x=c->x-1; 
  wc.y=c->y; 
  wc.height=c->height; 
  wc.width=c->width; 
  configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, False); 
*/

  sprintf(lastsub,"TRACE1: exit create_new_client\n");
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: exit create_new_client\n");
#endif
} 

void create_new_client(Window w, int reposition) 
{ 
  Client *c=NULL; 
  XWindowAttributes attr; 
  XSetWindowAttributes wattr; 
  long dummy; 
  Window parentwin;
  XWindowChanges wc; 
  int x, y, xmax, ymax, xmin, ymin, head=0, showfront=0, retval, new_hack_id=0;
  int inhibit_focus=0, need_ungrab=0, maketrans=0; 
  XClassHint xch;
  char **argv=NULL;
 
  sprintf(lastsub,"TRACE1: create_new_client  win: %#lx (%#lx)\n",w,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: create_new_client  win: %#lx (%#lx)\n",w,FC_WINDOW); 
#endif 
 
  if(gQuitFlag!=0) return;
    
  MyIncreaseXServerGrab();
  /* Get size and position of client */
  retval = XGetWindowAttributes(dpy, w, &attr);
  if(retval!=1) {
    printf("%s warning: XGetWindowAttributes returned status %d - aborting...\n",PROGNAME,retval);
    MyDecreaseXServerGrab();
    return;
  }

  if((c=malloc(sizeof *c))==NULL)
  {
    if(on_cmd_line)
        fprintf(stderr, "%s: Unable to allocate memory for new client.\n", PROGNAME);
    MyDecreaseXServerGrab();
    ClearUpSockets();
    exit(1);
  }
 
  /* Initialize client values */ 
  c->window=w;
  c->cover=0;
  c->al_list=NULL;
  c->al_count=0;
  c->ignore_unmap=0; 
  c->zoom_state=BUTTON_UNPRESSED; 
  c->icon_state=BUTTON_UNPRESSED;
  c->close_state=BUTTON_UNPRESSED; 
  c->can_max=True; 
  c->has_frame=True; 
  c->has_focus=False; 
// assume it isn't an iconbar
  c->is_icon=NOT_ICONBAR;
// start with some "bad" values...
  c->name_x=-1;
  c->name_y=-1;
  c->name_w=-1;
  c->name_h=-1;
  c->icon_x=-10000;
  c->icon_y=-10000;
  c->icon_h=-1;
  c->icon_w=-1;
  c->icontxt_x=-1;
  c->icontxt_y=-1;
  c->icontxt_w=-1;
  c->icontxt_h=-1;
  c->isoffscreen=0;
  c->osx = 0;
  c->osy = 0;
// default to a modeless window
  c->modal_state=MODELESS | DONT_IGNORE_MODAL;
  c->sloppy_family=FOCUS_CLICK;
  c->auto_properties=AUTO_NONE;
  if(config.focus_style==FOCUS_SLOPPY) c->auto_properties |= AUTO_FOCUS;
// set up initial sibling/child list
  c->child_count=0;
  c->parent=NULL;
  c->first_child=NULL;
  c->next_sibling=NULL;
  c->prev_sibling=NULL;
// initialise the OS X window identifier
  c->osxwinfo.winref=0;
  c->osxwinfo.group_winref=0;
  c->osxwinfo.view_state=VIEWSTATE_VISIBLE;
// Start the string info pointing to nothing
  c->clientclass=NULL;
  c->clientname=NULL;
  c->clientcmd=NULL;
  c->clienthost=NULL;
  c->clientgrouporg=NULL;
  c->clientgroup=NULL;
  c->name=NULL;
  c->is_shaped=1;
  c->osxwinfo.translucency=100;
  c->osxwinfo.inactive_translucency=85;
  c->is_dimmed=0;
  c->is_docked=0;
  c->is_e_m_r=0;
  c->dock_timer=0;
  c->undock_timer=0;
  c->osxwinfo.drawalpha=255;
  c->inhibit_level=HACKARG_INHIBITNONE;
  c->real_w=0;
  c->real_h=0;
  c->hack_id=0;
  
  /* Set event mask for client */ 
  XSelectInput(dpy, c->window, PropertyChangeMask|FocusChangeMask); 
 
  new_hack_id=create_new_hackid();

  if(gQuitFlag!=0) {
    MyDecreaseXServerGrab();
    free(c);
    return;
  }

  c->size=XAllocSizeHints(); 
  XGetWMNormalHints(dpy, w, c->size, &dummy); 
 
  if(gQuitFlag!=0) {
    MyDecreaseXServerGrab();
    free(c);
    return;
  }

  /* Handle clients with border set > 0 and adjust position accordingly */ 
  if(attr.border_width>0) 
  { 
    switch(GRAVITY) 
    { 
    case SouthEastGravity: 
      attr.x=attr.x+attr.border_width*2; 
      attr.y=attr.y+attr.border_width*2; 
      break; 
    case SouthWestGravity: 
    case SouthGravity: 
      attr.y=attr.y+attr.border_width*2; 
      break; 
    case NorthEastGravity: 
    case EastGravity: 
      attr.x=attr.x+attr.border_width*2; 
      break; 
    } 
  } 
  c->x=attr.x; 
  c->y=attr.y; 
  c->width=attr.width; 
  c->height=attr.height; 
 
 
  // go find all the properties, based on the default properties rules
  new_client_properties(c,gMainRules,attr.map_state,reposition);
 
    // set up the sibling list
    if(c->parent!=NULL) {
      add_sibling_client(c);
      // group name inhertance now happens in new_client_properties
      //if(c->clientgroup) XFree(c->clientgroup);
      //c->clientgroup=strdup(c->parent->clientgroup);
    }

  // deal with windows that have no class
  if(c->clientgroup==NULL) {
    if(c->name) {
      // use the title as the group name
      if(strlen(c->name)<23) {
        c->clientgroup = strdup(c->name);
      } else {
        // too long, so truncate it
        c->clientgroup = strdup(c->name);
        c->clientgroup[20]=0x2E;
        c->clientgroup[21]=0x2E;
        c->clientgroup[22]=0x2E;
        c->clientgroup[23]=0;
      }
    } else {
      // no name either...
      c->clientgroup=malloc(15);
      sprintf(c->clientgroup,"Untitled Group");
    }
  }
  c->clientgrouporg = c->clientgroup;
  
  //if(gVerboseConsole)
    printf("Original group: '%s' (%d)\n",c->clientgroup,c->osxwinfo.winref);

    add_client_to_group(c);
    //reorder_dockicon_winlist(c);
//  if(c->osxwinfo.winref==-10) {
    if( c->auto_properties & NO_GRAB_FOCUS ) {
        inhibit_focus = 1;
        showfront = 0;
        printf("*** NO_GRAB_FOCUS is set (%d)\n",c->auto_properties);
    } else {
        printf("*** NO_GRAB_FOCUS is *NOT* set (%d)\n",c->auto_properties);
        inhibit_focus = 0;
        showfront = 1;
    }
    
    if(lastfocus_client)
        if( lastfocus_client->auto_properties & RETAIN_FOCUS ) inhibit_focus=1;
//  } else if(c->osxwinfo.winref==-1)
//    c->osxwinfo.winref=0; // This is the case when new window should not be listed in the dock's window list

#ifdef SHAPE_SUPPORT 
  /* Select shape events */ 
  if(shape) 
    XShapeSelectInput(dpy, c->window, ShapeNotifyMask); 
#endif /* SHAPE */ 
 
  /* Create frame window */ 
  wattr.override_redirect=True;
  // Added the LeaveWindowMask to this, so we know when we've left the window
  wattr.event_mask=SubstructureNotifyMask|SubstructureRedirectMask|
                   ButtonPressMask|ButtonReleaseMask|EnterWindowMask|LeaveWindowMask;
  c->frame=XCreateWindow(dpy, root, HACKCMD_STARTX, HACKCMD_STARTY, FW, FH, 0,
                        DefaultDepth(dpy, screen), CopyFromParent, DefaultVisual(dpy, screen),
                        CWOverrideRedirect|CWEventMask, &wattr);
  
/*  if(can_send_hackcmds!=0) {
    c->cover=XCreateSimpleWindow(dpy, root, -2000, -2000, 50, 50, 0, 0, 0);
    XSelectInput(dpy, c->cover, ButtonPressMask|ButtonReleaseMask); 
    XMapWindow(dpy,c->cover);
    XMoveWindow(dpy,c->cover,HACKCMD_STARTX,HACKCMD_STARTY);
    XMoveWindow(dpy,c->cover,HACKCMD_SETID,-999);
  }*/
  
  // Set the cursor we want to see when inside the window -would be nice if this was more like the Mac's
  // normal cursor...
  XDefineCursor(dpy, c->frame, normal_curs); 

  /* This ignores the unmap event when reparenting */ 
  if(attr.map_state==IsViewable) 
    c->ignore_unmap++; 
 
  /* Reparent client window */ 
  XReparentWindow(dpy, c->window, c->frame, LW, TH);
    
  /* Create frame parts if has border and/or title */ 
    // Top frame parts are needed for iconbars, even if no frame
    c->title=create_frame_part(c->frame, TX, TY, TW<1?1:TW, config.top.height, c->has_frame); 
    c->top_left=create_frame_part(c->frame, TLX, TLY, TLW, TLH, c->has_frame); 
    c->top_right=create_frame_part(c->frame, TRX, TRY, TRW, TRH, c->has_frame);  
   
//  if(c->has_frame)
//  { 
    /* Side frame parts */ 
    c->left=create_frame_part(c->frame, LX, LY, LW, LH, c->has_frame); 
    c->right=create_frame_part(c->frame, RX, RY, RW, RH, c->has_frame); 
     
    /* Bottom frame parts */ 
    c->bottom=create_frame_part(c->frame, BX, BY, BW<1?1:BW, config.bottom.height, c->has_frame); 
    c->bottom_left=create_frame_part(c->frame, BLX, BLY, BLW, BLH, c->has_frame); 
    c->resize=create_frame_part(c->frame, RSX, RSY, RSW, RSH, c->has_frame); 
   
    /* Top frame parts 
    c->title=create_frame_part(c->frame, TX, TY, TW<1?1:TW, TH); 
    c->top_left=create_frame_part(c->frame, TLX, TLY, TLW, TLH); 
    c->top_right=create_frame_part(c->frame, TRX, TRY, TRW, TRH);   */
   
    /* Buttons */ 
    c->zoom=create_frame_part(c->frame, HX, HY, HW, HH, c->has_frame); 
    c->icon=create_frame_part(c->frame, IX, IY, IW, IH, c->has_frame);
    c->close=create_frame_part(c->frame, CX, CY, CW, CH, c->has_frame);
//  } 
 
  // Get XDarwin to create the new window - so that the title gets initialised
  XMapWindow(dpy,c->frame);
  // move the window again, so that it does not get 'stuck'
  XMoveWindow(dpy,c->frame,-10000,-10000);
  //SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,201,0,0,0,0); // make sure the client is not moved
  redraw_client_frame(c); 

  /* Add client to client list */ 
  if(client_list==NULL) 
  { 
    client_list=c; 
    c->prev=c; 
    c->next=c; 
  } 
  else 
  { 
    c->prev=client_list->prev; 
    c->next=client_list; 
    client_list->prev->next=c; 
    client_list->prev=c; 
  } 
  client_count++; 
  update_gnome_client_list();
  
  need_ungrab = check_need_ungrab();
//  need_ungrab=0;
  
// Make sure the window is updated first time (configure_client must see a change in the
// window dimensions for it to do anything - that's what the '3' does at the end - force a change)
/*  x=c->x;
  y=c->y;
  wc.x=-c->width-10;
  wc.y=-c->height-20;*/
  wc.x=c->x;
  wc.y=c->y;
  wc.height=c->height; 
  wc.width=c->width; 
  wc.border_width=0;
// Add one for the 'jig' to work around the blank windows problem
//  wc.x+=1;
  configure_client(c, CWX|CWY|CWHeight|CWWidth|CWBorderWidth, &wc, False, 3); 
 
//  if(get_wm_state(c->window)==IconicState)
//    c->dock_timer=-1; // need to make a note to minimise this client
    
  
  if(need_ungrab && last_replay_client==lastfocus_client && gX11InFront!=0) {
    XMoveWindow(dpy,c->frame,-FW-10,-FH-10);
    printf("%s info: found grab - not raising new client\n",PROGNAME);
    show_client(c);
    //stack_client_re_client(c,lastfocus_client,Below);
    XMoveWindow(dpy,c->frame,FX,FY);
    inhibit_focus = 1;
    showfront = 0;
    raise_client(c);
    //XRaiseWindow(dpy,c->frame);
  } else {
    show_client(c);
    if(gVerboseConsole)
      printf("%s info: no grab (%d) - raising new client\n",PROGNAME,need_ungrab);
    //inhibit_focus = 0;
    XRaiseWindow(dpy,c->frame);
  //raise_client(c);
  }
  
  c->hack_id = new_hack_id;
  SendHackCommand(c,0,HACKCMD_SETID,0,0,0,0,0); // uses the id stored for client, so args ignored

  if(c->auto_properties & (AUTO_FOCUS|NONAUTO_FIRSTCLICK))
    SendHackCommand(c,1,HACKCMD_SETFIRSTCLICK,1,0,0,0,0);
  
  //SendHackCommand(c,1,HACKCMD_INHIBITFULL,0,0,0,0,0);
  SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,HACKARG_INHIBITNONE,0,0,0,0);
  SendHackCommand(c,1,HACKCMD_DOSHADOW,-1,0,0,0,0); // inhibit shadow changes (tell XDarwin when myself)
    
  /* Button grab for click-to-focus */ 
  XGrabButton(dpy, AnyButton, AnyModifier, c->frame, False, ButtonPressMask|ButtonReleaseMask, GrabModeSync, GrabModeAsync, None, None); 
 
#ifdef KEYBOARD_SUPPORT 
  grab_keys(c->window); 
#endif /* KEYBOARD */ 
 
#ifdef WORKSPACE_SUPPORT 
  set_gnome_hint(c->window, WIN_WORKSPACE, workspace); 
#endif /* WORKSPACE */ 
 
//  SendSimpleHackCommand(HACKCMD_INHIBITINFO,0);
  
  MyDecreaseXServerGrab();

    // Now that the client group is sorted out, make the OroborOSX proxy window
    if( c->dock_timer & 2 ) {
        MakeX11Window(&(c->osxwinfo),c->name,c->clientgroup);
        c->dock_timer ^= 2;
    }

  if(c->osxwinfo.translucency!=ACTIVE_TRANSLUCENT0) maketrans=1;
      
  /* Check if we want to focus new clients */ 
  if(c->is_docked!=0 || c->is_icon!=NOT_ICONBAR /*|| !config.focus_new*/ || inhibit_focus ||
    (get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS) || still_in_auto_client()!=0) {
    if((gBGWinSetting & BGWIN_TRANSLUCENT)==BGWIN_TRANSLUCENT || maketrans)
      SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.inactive_translucency,0,0,0,0);
    if((gBGWinSetting & BGWIN_DIMMED)==BGWIN_DIMMED)
      SendHackCommand(c,1,HACKCMD_DIMWIN,1,0,0,0,0);
    SendHackCommand(c,1,HACKCMD_SETSHADOW,0,0,0,0,0);
    if(lastfocus_client!=NULL)
      if(lastfocus_client->osxwinfo.winref!=0)
         FocusOSXWin(&(lastfocus_client->osxwinfo));
    redraw_client_frame(c);
  } else {
    if(maketrans) {
      if(gVerboseConsole)
        printf("Making %#lx window translucent...\n",C_WINDOW);
      SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.translucency,0,0,0,0);
      Setup_Properties_Menu_Translucency(&c->osxwinfo);
    }
    set_focus_client(c);
  }
  
  SendHackTitleToDockCommand(c,0);

  /* Set client to correct state */   
  if(c->dock_timer & 1) {
    c->dock_timer ^= 1;
    // turn new client into an iconbar/docked-icon
    toggle_minimise_client(c,0);
    lower_client(c);
  } else if(showfront!=0 && gX11InFront==0 && c->is_docked==0 && c->is_icon==NOT_ICONBAR) {
    if( Is_Whole_Screen_Mode()==1 ) {
      Wait_OOSX_To_Front();
      gX11InFront = 1;
      come_to_front();
    } else if(c->dock_timer & 4) {
      Wait_OOSX_To_Front();
      Start_Whole_Screen_Mode();
      gX11InFront = 1;
      come_to_front();
    } else
      Bring_To_Front(0);
  }    
  // reinitialise this timer, now that it's use as a temporary flag is finished
  c->dock_timer = 0;
  
// 'Jig' it by one pixel (to make sure content appears on secondary screens)...
/*  wc.x=c->x+1; 
  wc.y=c->y; 
  wc.height=c->height; 
  wc.width=c->width; 
  configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, False); 
 
  wc.x=c->x-1; 
  wc.y=c->y; 
  wc.height=c->height; 
  wc.width=c->width; 
  configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, False); 
*/

  sprintf(lastsub,"TRACE1: exit create_new_client\n");
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: exit create_new_client\n");
#endif
} 
