/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 
 
#include "oroborus.h"
#include "communication.h"
#include "hackcmds.h"
#include "iconbars.h"
#include "client_properties.h"

// initial positions for iconbars
int nexticon_x=0, nexticon_y=22, grabstat_before_hide=0, gWaitingForSIGCHLD=0;
Display *dpy; 
Window root, check_win, lastenter;
Window lastfocus=0, pointer_grabbed=0; 
int hacking_focus=False, inhibit_errors=False;
Cursor resize_curs, move_curs, normal_curs; 
int screen, iconbar_state=ICONBAR_HIDDEN; 
Client *overicon=NULL, *current_resize_client=NULL;
Client *client_list=NULL, *lastfocus_client=NULL, *last_client_entered_hack=NULL; 
long client_count; 
Config config; 
Atom icccm[ICCCM_ATOM_COUNT], motif[MOTIF_ATOM_COUNT], gnome[GNOME_ATOM_COUNT]; 
char *rcfile=RCFILE;
RulesList *gMainRules=NULL;

#ifdef SHAPE_SUPPORT 
int shape, shape_event; 
#endif 
 
#ifdef WORKSPACE_SUPPORT 
int workspace=0; 
#endif 
 
#ifdef XINERAMA_SUPPORT 
int xinerama; 
int xinerama_heads;
#ifndef EMULATE_XINERAMA
XineramaScreenInfo *xinerama_screens; 
#else
XineramaScreenInfo xinerama_screens[EMULATED_SCREENS]=
{
  {0, 0, 0, 600, 1003},
  {0, 600, 100, 400, 903},
  {0, 1000, 200, 280, 803} 
};
#endif
#endif 

#ifdef I18N_SUPPORT
int use_fontset=False;
#endif
 
/*  
 * 
 * Quit nicely 
 * 
 */ 

void quit() 
{ 
  char script[50];
  
  sprintf(lastsub,"TRACE1: quit\n"); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: quit\n"); 
#endif 
 
  Restart_XDarwin();
  
  // make sure we don't try to send anything else to XDarwin and it doesn't send anything back
  // (otherwise we will get a crash on quit...)
  /* Remove all frames */ 
  while(client_list)  
    delete_client(client_list, 1); 
 
  /* Free cursors */ 
  XFreeCursor(dpy, normal_curs);
  XFreeCursor(dpy, resize_curs); 
  XFreeCursor(dpy, move_curs); 
 
  /* Free theme data */ 
  free_config(); 

  XSetInputFocus(dpy, root, RevertToPointerRoot, CurrentTime); 

  if( can_send_hackcmds!=0 ) {
    SendSimpleHackCommand(HACKCMD_INHIBITINFO,1);
    // Very last thing we do before closing display is to tell XDarwin about the quit
    if(launched_xdarwin==0)	// shouldn't get this, but maybe one day...
        XMoveWindow(dpy, check_win, -23451, -100);	// tell XDarwin we'll quit, but let it continue
    else
        XMoveWindow(dpy, check_win, -23451, -101);	// tell XDarwin to quit
  }
  
  XCloseDisplay(dpy);
  printf("%s info: successfully closed X display\n",PROGNAME);
  if(can_send_hackcmds!=0) ClearUpSockets();
  can_send_hackcmds=0;
  
  if(gQuitFlag==0)
    gQuitFlag=1; 

  if(Display_Name) free(Display_Name);
} 

/* 
 * 
 * Handle signals 
 * 
 */ 
void signal_handler(int signal) 
{
  int dummy;
  
  sprintf(lastsub,"TRACE1: signal_handler: %d\n",signal); 
  subtrace();
//#ifdef SHOW_TRACE1 
  if(signal!=SIGCHLD)
    printf("TRACE1: signal_handler: %d\n",signal);
  else if(gWaitingForSIGCHLD) {
    printf("Got SIGCHLD signal -clearing gWaitingForSIGCHLD\n");
    gWaitingForSIGCHLD = 0;
  }
//#endif 
 
  switch(signal) 
  { 
  case SIGINT: 
  case SIGTERM: 
    quit(); 
    break;
  case SIGHUP:
//    Restart_XDarwin();
//    reload_config();
    show_last_fifteen();
    list_all_windows();
    dummy=sleep(1);
    //system("/bin/sleep 1");
//    printf("quitting...\n");
//    quit();
    break;
  case SIGINFO:
    show_last_fifteen();
    list_all_clients();
    dummy=sleep(1);
    break;
  case SIGCHLD: 
//    wait(NULL); 
    break; 
  } 
} 
 
/* 
 * 
 * Create a frame for each window that already exists 
 * 
 */ 

void scan_for_windows() 
{ 
  unsigned int count, i; 
  Window w1, w2, *windows=NULL; 
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE1: scan_for_windows\n"); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: scan_for_windows\n"); 
#endif 
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
  if(gQuitFlag==0)
    for(i=0; i<count; i++) 
    { 
      XGetWindowAttributes(dpy, windows[i], &attr);
      if(gQuitFlag!=0) break;
      if(!attr.override_redirect && attr.map_state==IsViewable) 
        create_new_client(windows[i], False); 
    } 
  if(windows) XFree(windows); 
} 
 
/* 
 * 
 * Initialize window manager 
 * 
 */ 

void init_wm() 
{ 
  int dummy;
  int i, head=0;
  int val;
  long w2;
 
  sprintf(lastsub,"TRACE1: init_wm\n"); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: init_wm\n"); 
#endif 
 
  /* ICCCM stuff */ 
  icccm[WM_STATE]=XInternAtom(dpy, "WM_STATE", False); 
  icccm[WM_PROTOCOLS]=XInternAtom(dpy, "WM_PROTOCOLS", False); 
  icccm[WM_DELETE]=XInternAtom(dpy, "WM_DELETE_WINDOW", False); 
  icccm[WM_CHANGE_STATE]=XInternAtom(dpy, "WM_CHANGE_STATE", False); 
 
  /* Mwm hints */ 
  motif[MOTIF_WM_HINTS]=XInternAtom(dpy, "_MOTIF_WM_HINTS", False); 
 
  /* GNOME hints */ 
  gnome[WIN_CLIENT_LIST]=XInternAtom(dpy, "_WIN_CLIENT_LIST", False); 
  gnome[WIN_STATE]=XInternAtom(dpy, "_WIN_STATE", False); 
  gnome[WIN_HINTS]=XInternAtom(dpy, "_WIN_HINTS", False); 
  gnome[WIN_LAYER]=XInternAtom(dpy, "_WIN_LAYER", False); 
  gnome[WIN_SUPPORTING_WM_CHECK]=XInternAtom(dpy, "_WIN_SUPPORTING_WM_CHECK", False); 
  gnome[WIN_PROTOCOLS]=XInternAtom(dpy, "_WIN_PROTOCOLS", False); 
  gnome[WIN_DESKTOP_BUTTON_PROXY]=XInternAtom(dpy, "_WIN_DESKTOP_BUTTON_PROXY", False); 
#ifdef WORKSPACE_SUPPORT 
  gnome[WIN_WORKSPACE]=XInternAtom(dpy, "_WIN_WORKSPACE", False); 
  gnome[WIN_WORKSPACE_COUNT]=XInternAtom(dpy, "_WIN_WORKSPACE_COUNT", False); 
#endif 
 
  /* Create GNOME check window */ 
  check_win=XCreateSimpleWindow(dpy, root, -5000, -5000, 5, 5, 0, 0, 0); 
   
  /* Set up GNOME properties */ 
  set_gnome_hint(check_win, WIN_SUPPORTING_WM_CHECK, check_win); 
  set_gnome_hint(root, WIN_SUPPORTING_WM_CHECK, check_win); 
  set_gnome_hint(check_win, WIN_DESKTOP_BUTTON_PROXY, check_win); 
  set_gnome_hint(root, WIN_DESKTOP_BUTTON_PROXY, check_win); 
  val = check_win;
  XChangeProperty(dpy, root, gnome[WIN_SUPPORTING_WM_CHECK], XA_CARDINAL, 32, PropModeReplace, (unsigned char *) &w2, 1);
  XChangeProperty(dpy, check_win, gnome[WIN_SUPPORTING_WM_CHECK], XA_CARDINAL, 32, PropModeReplace, (unsigned char *) &w2, 1);
#ifdef WORKSPACE_SUPPORT 
  set_gnome_hint(root, WIN_WORKSPACE_COUNT, config.workspace_count); 
  set_gnome_hint(root, WIN_WORKSPACE, 0); 
#endif 

//  cursor_follow_win=XCreateSimpleWindow(dpy, root, -2000, -2000, 50, 50, 0, 0, 0); 

#ifdef SHAPE_SUPPORT 
  /* Check for shape support */ 
  shape=XShapeQueryExtension(dpy, &shape_event, &dummy); 
#endif 
 
  head=0;
#ifdef XINERAMA_SUPPORT
#ifndef EMULATE_XINERAMA 
  /* Check xinerama extension */ 
  xinerama=XineramaQueryExtension(dpy, &dummy, &dummy); 
  if(xinerama) {
    xinerama_screens=XineramaQueryScreens(dpy, &xinerama_heads);
    if(xinerama_screens==NULL) {
      printf("%s warning: Got NULL from QueryScreens, but QueryExtension gave %d !!\n",PROGNAME,xinerama);
      xinerama=0;
    } else
      head=find_head(50,50);
  }
#else
  xinerama=True;
  xinerama_heads=EMULATED_SCREENS;
#endif
  if(xinerama) {
    printf("%s info: Real Xinerama screens found: %d\n",PROGNAME,xinerama_heads);
    for(i=0;i<xinerama_heads;i++)
      printf("    head %d (x,y w,h):  %d,%d %d,%d\n",i,screen_x(i),screen_y(i),screen_width(i),screen_height(i));
    update_xinerama_screens();
  } else
    printf("%s info: Xinerama not found...\n",PROGNAME);
#endif 
 
  /* Set event mask for root window */ 
  XSelectInput(dpy, root, SubstructureRedirectMask|SubstructureNotifyMask|ButtonPressMask|ButtonReleaseMask|FocusChangeMask|PropertyChangeMask); 
 
  normal_curs=XCreateFontCursor(dpy, NORMAL_CURS);
  resize_curs=XCreateFontCursor(dpy, RESIZE_CURS); 
  move_curs=XCreateFontCursor(dpy, MOVE_CURS); 

// set up the correct iconbar positions, depending on iconbar location option
  if(iconbarpos==ICONBARPOS_TOPLEFT || iconbarpos==ICONBARPOS_TOPRIGHT)
    nexticon_y=screen_y(head)+22;
  else
    nexticon_y=screen_y(head)+screen_height(head)-config.top.height;

  if(iconbarpos==ICONBARPOS_TOPLEFT || iconbarpos==ICONBARPOS_BOTLEFT)
    nexticon_x=screen_x(head);
  else
    nexticon_x=screen_x(head)+screen_width(head)-1;
} 
 
/* 
 * 
 * Set up signal handler 
 * 
 */ 

void init_signals() 
{ 
  struct sigaction act; 
 
  sprintf(lastsub,"TRACE1: init_signals\n"); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: init_signals\n"); 
#endif 
  
  act.sa_handler=signal_handler; 
  act.sa_flags=0; 
  sigaction(SIGTERM, &act, NULL); 
  sigaction(SIGINT, &act, NULL); 
  sigaction(SIGHUP, &act, NULL); 
  sigaction(SIGCHLD, &act, NULL); 
  sigaction(SIGINFO, &act, NULL); 
} 

/*
 * Sets up the main rules list by parsing the rules files
 */
void init_rules_list()
{
    char *filename=NULL;
    int parse_err;
    
    if(gMainRules!=NULL) 
        clear_rules_list(gMainRules);
    else
        gMainRules = new_rules_list();

    if(gMainRules==NULL) {
        fprintf(stderr,"%s error: failed to create rules list!\n",PROGNAME);
        return;
    }
    
    filename = malloc(50);
    sprintf(filename,"/tmp/oroborosx-rules.%s",randomID);
    
    parse_err = parse_rules_file(filename,gMainRules);
    printf("parse_rules_file returned with error: %d\n",parse_err);
    sprintf(filename,"rm -f /tmp/oroborosx-rules.%s",randomID);
    system(filename);
    free(filename);
}

/* 
 * 
 * Main routine for 'oroborus part' - now exits back to main, rather than calling the event loop
 * 
 */ 

int orob_main(int argc, char *argv[], char *dpy_str) 
{
  char str[100];
  int i, sret;
  char def_dpy[10]=":0.0";
  char *auth_str=NULL;
  
#ifdef I18N_SUPPORT 
  char *locale=NULL; 
#endif /* I18N */ 

  sprintf(lastsub,"TRACE1: orob_main (argc: %d)\n",argc); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: orob_main (argc: %d)\n",argc); 
#endif 

  auth_str=getenv("XAUTHORITY");
  if(auth_str) {
    if(strlen(auth_str)<=1) {
      free(auth_str);
      auth_str=NULL;
    }
  }
  if(auth_str==NULL) {
    char *tmp_str=getenv("HOME");
    if(tmp_str) {
      auth_str=malloc(strlen(tmp_str)+12);
      sprintf(auth_str,"%s/.Xauthority",tmp_str);
//      free(tmp_str);
      setenv("XAUTHORITY",auth_str,1);
      printf("%s info: setting XAUTHORITY to %s\n",PROGNAME,auth_str);
    } else {
      fprintf(stderr,"%s: HOME environment variable not set\n",PROGNAME);
      return -156;
    }
  } else {
    printf("%s info: XAUTHORITY is already set to %s\n",PROGNAME,auth_str);
  }

  if(auth_str) {
    free(auth_str);
    auth_str=NULL;
  }
    
  on_cmd_line=1;
  for(i=0;i<argc;i++)
  {
#ifdef DEBUG
    printf("arg %d :  %s\n",i,argv[i]);
#endif
    if(!strncmp(argv[i], "-psn", 4))
      on_cmd_line=0;
    if(!strncmp(argv[i], "-config", 7) && i+1<argc)
      rcfile=argv[++i];
    else if(!strncmp(argv[i], "-display", 8) && i+1<argc)
      dpy_str=argv[++i];
  }
 
#ifdef I18N_SUPPORT 
  locale=setlocale(LC_ALL, "");
  if(locale!=NULL)
  {
#ifdef DEBUG
    printf("locale=%s\n", locale);
#endif
    if(strncmp(locale, "C", 1) && strncmp(locale, "POSIX", 5) && XSupportsLocale())
      use_fontset=True;
  }
#ifdef DEBUG
  if(use_fontset)
    printf("setting use_fontset=True\n");
  else
    printf("setting use_fontset=False\n");
#endif
#endif /* I18N */ 
   
  /* Initialize X */ 
  if( dpy_str==NULL ) {
    if(!(dpy=XOpenDisplay(dpy_str)))
      dpy_str=def_dpy;
    printf("%s info: Display string: %s %#lx\n",PROGNAME,dpy_str,dpy);
  }
  
  if(!(dpy=XOpenDisplay(dpy_str))) {
    if(on_cmd_line)
        fprintf(stderr, "%s: Can't open display, X may not be running.\n", PROGNAME);
    return -1;
  }
 
  init_signals(); 

// Get the display name to send to the X11 app launcher
  if(dpy_str==NULL) {
    dpy_str=getenv("DISPLAY");
    printf("%s info: Display string: %s %#lx\n",PROGNAME,dpy_str,dpy);
  }
  
  if(dpy_str!=NULL)
    Display_Name=strdup(dpy_str);
  else {
    fprintf(stderr,"%s: Could not find name of display\n",PROGNAME);
    return -2;
  }
  
  XSetErrorHandler(handle_xerror); 
   
  root=XDefaultRootWindow(dpy); 
  screen=XDefaultScreen(dpy); 
 
  load_config(); 
  init_wm();
  
#ifdef KEYBOARD_SUPPORT
  init_keyboard();
#endif

  sret = iconbarpos;
  if(iconbarpos==ICONBARPOS_TOPRIGHT)
    change_iconbarpos(ICONBARPOS_BOTLEFT);
  else if(iconbarpos==ICONBARPOS_TOPLEFT)
    change_iconbarpos(ICONBARPOS_BOTRIGHT);
  else if(iconbarpos==ICONBARPOS_BOTLEFT)
    change_iconbarpos(ICONBARPOS_TOPRIGHT);
  else if(iconbarpos==ICONBARPOS_BOTRIGHT)
    change_iconbarpos(ICONBARPOS_TOPLEFT);
  change_iconbarpos(sret);

//  compress_iconbars();
  scan_for_windows(); 

  if(gQuitFlag!=0)	// got an error (probably a wm is already running)
    return gQuitFlag;
  
  XMapWindow(dpy, check_win);

        xdarwin_socket = 0;
        
        if(can_send_hackcmds!=0 && main_socket!=0) {
        
            SendSimpleHackCommand(HACKCMD_COMMUNICATE,1); // tell XDarwin to connect to socket
            
            i=0;
            sret=0;
            while(i<15 && sret==0) {
                sret=check_listen_accept_name(main_socket,str);
                if(sret==0) {
                    i++;
                    if(i==15) {
                        fprintf(stderr,"%s: XDarwin did not register. Aborting.\n",PROGNAME);
                        return -3;
                    }
                    printf("%s Info: Waiting for XDarwin to register...\n",PROGNAME);
                    sleep(1);
                } else if(sret>0) {
                    printf("%s info: Server heard from socket client: %s\n",PROGNAME,str);
                    if(strlen(str)==7) {
                        if(strncasecmp(str,"XDarwin",7)==0) {
                            xdarwin_socket=sret;
                            printf("%s info: Set xdarwin_socket=%d\n",PROGNAME,xdarwin_socket);
                            SendSimpleHackCommand(HACKCMD_INHIBITINFO,0);
                        } else
                            sret = 0;
                    } else
                        sret = 0;
                }
            }
		
            if( sret < 0 ) {
                fprintf(stderr,"%s: Got an error from check_listen_accept_name: %d\n",PROGNAME,sret);
                ClearUpSockets();
                exit(1);
            }
        }
        
#ifdef KEYBOARD_SUPPORT
  grab_keys(check_win); 
#endif 

    if(can_send_hackcmds!=0) {
        // don't want anything else interfering with the X server while this happens
        MyIncreaseXServerGrab();
        MyXSync(dpy,0);
        XMoveWindow(dpy, check_win, HACKCMD_STARTX, HACKCMD_STARTY); // start the hack command sequence
        XMoveWindow(dpy, check_win, HACKCMD_SETID, -999);	// special window for relaying commands
        XMoveWindow(dpy, check_win, -5000, -5000);
        MyXSync(dpy,0);
        MyDecreaseXServerGrab();

/*	XMapWindow(dpy, cursor_follow_win);
        XMoveWindow(dpy, cursor_follow_win, HACKCMD_STARTX, HACKCMD_STARTY);
        XMoveWindow(dpy, cursor_follow_win, HACKCMD_SETID, -999);*/
    }
  
  set_focus_client(get_focus_client());
  XDefineCursor(dpy, root, normal_curs);

  return 0; 
}
