/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 
 
#include "oroborus.h" 
 
/* 
 * 
 * Load pixmap file 
 * 
 */ 

int load_pixmap_file(char *dir, char *file_name, Pixmap *pm, Pixmap *mask, int *w, int *h, XpmColorSymbol *colsym, int numsym) 
{ 
  int err;
  char *full_file_name=NULL; 
  char *data[]={ "1 1 2 1", 
                 "  c None", 
                 ". c #000000", 
                 "." }; 
  XpmAttributes attr; 
 
  if(gQuitFlag!=0) return 0;
  
  sprintf(lastsub,"TRACE2: load_pixmap_file %s\n",file_name); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: load_pixmap_file %s\n",file_name); 
#endif 
 
  attr.valuemask=XpmSize|XpmColorSymbols; 
  attr.colorsymbols=colsym; 
  attr.numsymbols=numsym; 
 
  /* Build full file name */ 
  if((full_file_name=malloc(strlen(dir)+strlen(file_name)+3))==NULL)
  { 
    if(on_cmd_line) {
      fprintf(stderr, "%s: Memory allocation failed in function load_pixmap_file.\n", PROGNAME);
      ClearUpSockets();
      exit(1);
    } else {
      fprintf(stderr, "%s: Memory allocation failed in function load_pixmap_file.\n", PROGNAME);
      gQuitFlag=-2;
      return 0;
    }
  }
 
  sprintf(full_file_name, "%s/%s", dir, file_name);  
 
  /* Read pixmap and get size */ 
  err=XpmReadFileToPixmap(dpy, root, full_file_name, pm, mask, &attr);
  /* Free allocated memory */ 
  free(full_file_name);
  
  if(err!=0)
  {
    if(err==-1) {	// No such file: may not be needed for this theme...
      printf("%s warning: Can't find pixmap %s  -not used\n", PROGNAME, file_name);
    } else {		// hmmm... real errors... never mind, let's carry on and let it blow up...
      printf("%s warning: Can't load pixmap %s (got error %d in XpmReadFileToPixmap)\n", PROGNAME, file_name, err);
    }
// just create a dummy pixmap
    XpmCreatePixmapFromData(dpy, root, data, pm, mask, &attr);
// and register the height and width as zero
    if(w) *w=0; 
    if(h) *h=0; 
  } else {
//    printf("   pixmap %s loaded OK\n",file_name);
    if(w) *w=attr.width; 
    if(h) *h=attr.height; 
  }
  
  return err;
} 
 
/* 
 * 
 * Load pixmaps -note extra ones for new widget states
 * 
 */ 

void load_pixmaps(char *pixmap_dir, XpmColorSymbol *colsym, int numsyms) 
{ 
  int got_title_iconbar=0;
  
  sprintf(lastsub,"TRACE1: load_pixmaps\n"); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: load_pixmaps\n"); 
#endif 
 
 
  /* Load resize pixmaps */ 
  load_pixmap_file(pixmap_dir, "resize-active.xpm", &config.resize.pixmap[FRAME_ACTIVE], &config.resize.mask[0], &config.resize.width, &config.resize.height, colsym, numsyms); 
  load_pixmap_file(pixmap_dir, "resize-inactive.xpm", &config.resize.pixmap[FRAME_INACTIVE], NULL, NULL, NULL, colsym, numsyms); 
 
 
  /* Load zoom button pixmaps */ 
  load_pixmap_file(pixmap_dir, "zoom-active.xpm", &config.zoom.pixmap[FRAME_ACTIVE], &config.zoom.mask[0], &config.zoom.width, &config.zoom.height, colsym, numsyms); 
  load_pixmap_file(pixmap_dir, "zoom-inactive.xpm", &config.zoom.pixmap[FRAME_INACTIVE], NULL, NULL, NULL, colsym, numsyms);
  
  if(load_pixmap_file(pixmap_dir, "zoom-active-pressed.xpm", &config.zoom.pixmap[BUTTON_PRESSED], NULL, NULL, NULL, colsym, numsyms)==-1) {
    printf("%s info: trying alpha 4 theme compatibility...\n",PROGNAME);
    // backward compatibility with old (confusing!) names
    load_pixmap_file(pixmap_dir, "zoom-activeoff.xpm", &config.zoom.pixmap[BUTTON_ACTIVEOFF], NULL, NULL, NULL, colsym, numsyms);
    load_pixmap_file(pixmap_dir, "zoom-inactiveoff.xpm", &config.zoom.pixmap[BUTTON_INACTIVEOFF], NULL, NULL, NULL, colsym, numsyms);
    load_pixmap_file(pixmap_dir, "zoom-pressed.xpm", &config.zoom.pixmap[BUTTON_PRESSED], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "zoom-over.xpm", &config.zoom.pixmap[BUTTON_OVER], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "zoom-overback.xpm", &config.zoom.pixmap[BUTTON_OVERBACK], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "zoom-pressedback.xpm", &config.zoom.pixmap[BUTTON_PRESSEDBACK], NULL, NULL, NULL, colsym, numsyms);
  } else {
    // new names...
    load_pixmap_file(pixmap_dir, "zoom-active-over.xpm", &config.zoom.pixmap[BUTTON_OVER], NULL, NULL, NULL, colsym, numsyms);
    load_pixmap_file(pixmap_dir, "zoom-active-off.xpm", &config.zoom.pixmap[BUTTON_ACTIVEOFF], NULL, NULL, NULL, colsym, numsyms);
    load_pixmap_file(pixmap_dir, "zoom-inactive-off.xpm", &config.zoom.pixmap[BUTTON_INACTIVEOFF], NULL, NULL, NULL, colsym, numsyms);
    load_pixmap_file(pixmap_dir, "zoom-inactive-over.xpm", &config.zoom.pixmap[BUTTON_OVERBACK], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "zoom-inactive-pressed.xpm", &config.zoom.pixmap[BUTTON_PRESSEDBACK], NULL, NULL, NULL, colsym, numsyms);
  }
 
 
  /* Load 'iconise' button pixmaps */
  if(load_pixmap_file(pixmap_dir, "icon-inactive.xpm", &config.icon.pixmap[FRAME_INACTIVE], NULL, NULL, NULL, colsym, numsyms)==-1) {
    // backward compatibility with old (confusing!) names
    printf("%s info: trying alpha 4 theme compatibility...\n",PROGNAME);
    load_pixmap_file(pixmap_dir, "hide-active.xpm", &config.icon.pixmap[FRAME_ACTIVE], &config.icon.mask[0], &config.icon.width, &config.icon.height, colsym, numsyms);
    load_pixmap_file(pixmap_dir, "hide-inactive.xpm", &config.icon.pixmap[FRAME_INACTIVE], NULL, NULL, NULL, colsym, numsyms);
    load_pixmap_file(pixmap_dir, "hide-pressed.xpm", &config.icon.pixmap[BUTTON_PRESSED], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "hide-over.xpm", &config.icon.pixmap[BUTTON_OVER], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "hide-overback.xpm", &config.icon.pixmap[BUTTON_OVERBACK], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "hide-pressedback.xpm", &config.icon.pixmap[BUTTON_PRESSEDBACK], NULL, NULL, NULL, colsym, numsyms);
  } else {
    // new names...
    load_pixmap_file(pixmap_dir, "icon-active.xpm", &config.icon.pixmap[FRAME_ACTIVE], &config.icon.mask[0], &config.icon.width, &config.icon.height, colsym, numsyms);
    load_pixmap_file(pixmap_dir, "icon-active-pressed.xpm", &config.icon.pixmap[BUTTON_PRESSED], NULL, NULL, NULL, colsym, numsyms);
    load_pixmap_file(pixmap_dir, "icon-active-over.xpm", &config.icon.pixmap[BUTTON_OVER], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "icon-inactive-over.xpm", &config.icon.pixmap[BUTTON_OVERBACK], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "icon-inactive-pressed.xpm", &config.icon.pixmap[BUTTON_PRESSEDBACK], NULL, NULL, NULL, colsym, numsyms);
  }
  
 
  /* Load close button pixmaps */ 
  load_pixmap_file(pixmap_dir, "close-active.xpm", &config.close.pixmap[FRAME_ACTIVE], &config.close.mask[0], &config.close.width, &config.close.height, colsym, numsyms); 
  load_pixmap_file(pixmap_dir, "close-inactive.xpm", &config.close.pixmap[FRAME_INACTIVE], NULL, NULL, NULL, colsym, numsyms);
  
  if(load_pixmap_file(pixmap_dir, "close-active-pressed.xpm", &config.close.pixmap[BUTTON_PRESSED], NULL, NULL, NULL, colsym, numsyms)==-1) {
    printf("%s info: trying alpha 4 theme compatibility...\n",PROGNAME);
    // backward compatibility with old (confusing!) names
    load_pixmap_file(pixmap_dir, "close-pressed.xpm", &config.close.pixmap[BUTTON_PRESSED], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "close-over.xpm", &config.close.pixmap[BUTTON_OVER], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "close-overback.xpm", &config.close.pixmap[BUTTON_OVERBACK], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "close-pressedback.xpm", &config.close.pixmap[BUTTON_PRESSEDBACK], NULL, NULL, NULL, colsym, numsyms);
  } else {
    // new names...
    load_pixmap_file(pixmap_dir, "close-active-over.xpm", &config.close.pixmap[BUTTON_OVER], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "close-inactive-over.xpm", &config.close.pixmap[BUTTON_OVERBACK], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "close-inactive-pressed.xpm", &config.close.pixmap[BUTTON_PRESSEDBACK], NULL, NULL, NULL, colsym, numsyms);
  }
  
 
  /* Load title pixmaps */ 
  load_pixmap_file(pixmap_dir, "title-active.xpm", &config.title.pixmap[FRAME_ACTIVE], &config.title.mask[0], &config.title.width, &config.title.height, colsym, numsyms); 
  load_pixmap_file(pixmap_dir, "title-inactive.xpm", &config.title.pixmap[FRAME_INACTIVE], NULL, NULL, NULL, colsym, numsyms);
  
  if(load_pixmap_file(pixmap_dir, "title-active-iconbar.xpm", &config.title.pixmap[FRAME_ACTIVE_ICONBAR], &config.title.mask[2], NULL, NULL, colsym, numsyms)==-1) {
    printf("%s info: trying alpha 4 theme compatibility...\n",PROGNAME);
    // backward compatibility, with title-(in)active-shaded as iconbar and title-(in)active as shaded pixmaps
    load_pixmap_file(pixmap_dir, "title-active-shaded.xpm", &config.title.pixmap[FRAME_ACTIVE_ICONBAR], &config.title.mask[2], NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "title-inactive-shaded.xpm", &config.title.pixmap[FRAME_INACTIVE_ICONBAR], NULL, NULL, NULL, colsym, numsyms);
    load_pixmap_file(pixmap_dir, "title-active.xpm", &config.title.pixmap[FRAME_ACTIVE_SHADED], &config.title.mask[1], NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "title-inactive.xpm", &config.title.pixmap[FRAME_INACTIVE_SHADED], NULL, NULL, NULL, colsym, numsyms);
  } else {
    got_title_iconbar=1;
    // new theme structure (titlebars, iconbars, and shaded titlebars all have different pixmaps)
    load_pixmap_file(pixmap_dir, "title-inactive-iconbar.xpm", &config.title.pixmap[FRAME_INACTIVE_ICONBAR], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "title-active-shaded.xpm", &config.title.pixmap[FRAME_ACTIVE_SHADED], &config.title.mask[1], NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "title-inactive-shaded.xpm", &config.title.pixmap[FRAME_INACTIVE_SHADED], NULL, NULL, NULL, colsym, numsyms);
  }


  config.title_left.width=0;
  config.title_left.height=0;
  config.title_right.width=0;
  config.title_right.height=0;
  /* Load top pixmaps */ 
  if(load_pixmap_file(pixmap_dir, "top-active.xpm", &config.top.pixmap[FRAME_ACTIVE], &config.top.mask[0], &config.top.width, &config.top.height, colsym, numsyms)==-1) {
    // backwards compatibility, with top the same as title pixmaps
    printf("%s info: assuming theme does not use separate title background pixmaps...\n",PROGNAME);
    if(got_title_iconbar) {
      load_pixmap_file(pixmap_dir, "title-active-shaded.xpm", &config.top.pixmap[FRAME_ACTIVE_SHADED], &config.top.mask[1], NULL, NULL, colsym, numsyms); 
      load_pixmap_file(pixmap_dir, "title-inactive-shaded.xpm", &config.top.pixmap[FRAME_INACTIVE_SHADED], NULL, NULL, NULL, colsym, numsyms);
      load_pixmap_file(pixmap_dir, "title-active-iconbar.xpm", &config.top.pixmap[FRAME_ACTIVE_ICONBAR], &config.top.mask[2], NULL, NULL, colsym, numsyms); 
      load_pixmap_file(pixmap_dir, "title-inactive-iconbar.xpm", &config.top.pixmap[FRAME_INACTIVE_ICONBAR], NULL, NULL, NULL, colsym, numsyms);
    } else {
      load_pixmap_file(pixmap_dir, "title-active.xpm", &config.top.pixmap[FRAME_ACTIVE_SHADED], &config.top.mask[1], NULL, NULL, colsym, numsyms); 
      load_pixmap_file(pixmap_dir, "title-inactive.xpm", &config.top.pixmap[FRAME_INACTIVE_SHADED], NULL, NULL, NULL, colsym, numsyms);
      load_pixmap_file(pixmap_dir, "title-active-shaded.xpm", &config.top.pixmap[FRAME_ACTIVE_ICONBAR], &config.top.mask[2], NULL, NULL, colsym, numsyms); 
      load_pixmap_file(pixmap_dir, "title-inactive-shaded.xpm", &config.top.pixmap[FRAME_INACTIVE_ICONBAR], NULL, NULL, NULL, colsym, numsyms);
    }
    load_pixmap_file(pixmap_dir, "title-active.xpm", &config.top.pixmap[FRAME_ACTIVE], &config.top.mask[0], NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "title-inactive.xpm", &config.top.pixmap[FRAME_INACTIVE], NULL, NULL, NULL, colsym, numsyms);
    config.top.width=config.title.width;
    config.top.height=config.title.height;
    config.title.width=0;
    config.title.height=0;
  } else {
    // top is now the main part of the titlebar, with title only being used behind the title text itself
    load_pixmap_file(pixmap_dir, "top-inactive.xpm", &config.top.pixmap[FRAME_INACTIVE], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "top-active-shaded.xpm", &config.top.pixmap[FRAME_ACTIVE_SHADED], &config.top.mask[1], NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "top-inactive-shaded.xpm", &config.top.pixmap[FRAME_INACTIVE_SHADED], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "top-active-iconbar.xpm", &config.top.pixmap[FRAME_ACTIVE_ICONBAR], &config.top.mask[2], NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "top-inactive-iconbar.xpm", &config.top.pixmap[FRAME_INACTIVE_ICONBAR], NULL, NULL, NULL, colsym, numsyms);
    // also need the patterns that go left and right of the title text
    if(load_pixmap_file(pixmap_dir, "title-left-inactive.xpm", &config.title_left.pixmap[FRAME_INACTIVE], NULL, NULL, NULL, colsym, numsyms)==-1) {
      // no left/right patterns for the title text (two pixels padding of title pixmap will be used instead)
    } else {
      // we do have the patterns that go left and right of the title text...
      load_pixmap_file(pixmap_dir, "title-left-active.xpm", &config.title_left.pixmap[FRAME_ACTIVE], &config.title_left.mask[0], &config.title_left.width, &config.title_left.height, colsym, numsyms);
      load_pixmap_file(pixmap_dir, "title-left-active-shaded.xpm", &config.title_left.pixmap[FRAME_ACTIVE_SHADED], &config.title_left.mask[1], NULL, NULL, colsym, numsyms); 
      load_pixmap_file(pixmap_dir, "title-left-inactive-shaded.xpm", &config.title_left.pixmap[FRAME_INACTIVE_SHADED], NULL, NULL, NULL, colsym, numsyms);
      load_pixmap_file(pixmap_dir, "title-left-active-iconbar.xpm", &config.title_left.pixmap[FRAME_ACTIVE_ICONBAR], &config.title_left.mask[2], NULL, NULL, colsym, numsyms); 
      load_pixmap_file(pixmap_dir, "title-left-inactive-iconbar.xpm", &config.title_left.pixmap[FRAME_INACTIVE_ICONBAR], NULL, NULL, NULL, colsym, numsyms);
      load_pixmap_file(pixmap_dir, "title-right-active.xpm", &config.title_right.pixmap[FRAME_ACTIVE], &config.title_right.mask[0], &config.title_right.width, &config.title_right.height, colsym, numsyms);
      load_pixmap_file(pixmap_dir, "title-right-inactive.xpm", &config.title_right.pixmap[FRAME_INACTIVE], NULL, NULL, NULL, colsym, numsyms);
      load_pixmap_file(pixmap_dir, "title-right-active-shaded.xpm", &config.title_right.pixmap[FRAME_ACTIVE_SHADED], &config.title_right.mask[1], NULL, NULL, colsym, numsyms); 
      load_pixmap_file(pixmap_dir, "title-right-inactive-shaded.xpm", &config.title_right.pixmap[FRAME_INACTIVE_SHADED], NULL, NULL, NULL, colsym, numsyms);
      load_pixmap_file(pixmap_dir, "title-right-active-iconbar.xpm", &config.title_right.pixmap[FRAME_ACTIVE_ICONBAR], &config.title_right.mask[2], NULL, NULL, colsym, numsyms); 
      if(load_pixmap_file(pixmap_dir, "title-right-inactive-iconbar.xpm", &config.title_right.pixmap[FRAME_INACTIVE_ICONBAR], NULL, NULL, NULL, colsym, numsyms)==-1) {
         printf("Checking for badly expanded file with #0 suffix... ");
         if(load_pixmap_file(pixmap_dir, "title-right-inactive-iconbar.#0", &config.title_right.pixmap[FRAME_INACTIVE_ICONBAR], NULL, NULL, NULL, colsym, numsyms)==0)
           printf("Found it (blame Stuffit Expander!)\n");
         else
           printf("still failed. Assume not used...\n");
      }
    }
  }
 
 
  /* Load top-left pixmaps */ 
  load_pixmap_file(pixmap_dir, "top-left-active.xpm", &config.top_left.pixmap[FRAME_ACTIVE], &config.top_left.mask[0], &config.top_left.width, &config.top_left.height, colsym, numsyms); 
  load_pixmap_file(pixmap_dir, "top-left-inactive.xpm", &config.top_left.pixmap[FRAME_INACTIVE], NULL, NULL, NULL, colsym, numsyms); 
  
  if(load_pixmap_file(pixmap_dir, "top-left-inactive-iconbar.xpm", &config.top_left.pixmap[FRAME_INACTIVE_ICONBAR], NULL, NULL, NULL, colsym, numsyms)==-1) {
    printf("%s info: trying alpha 4 theme compatibility...\n",PROGNAME);
    // backward compatibility, with top-left-(in)active-shaded as iconbar and top-left-(in)active as shaded
    load_pixmap_file(pixmap_dir, "top-left-active-shaded.xpm", &config.top_left.pixmap[FRAME_ACTIVE_ICONBAR], &config.top_left.mask[2], NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "top-left-inactive-shaded.xpm", &config.top_left.pixmap[FRAME_INACTIVE_ICONBAR], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "top-left-active.xpm", &config.top_left.pixmap[FRAME_ACTIVE_SHADED], &config.top_left.mask[1], NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "top-left-inactive.xpm", &config.top_left.pixmap[FRAME_INACTIVE_SHADED], NULL, NULL, NULL, colsym, numsyms); 
    config.iconbar_left_width=config.top_left.width;
    config.iconbar_left_height=config.top_left.height;
  } else {
    // new theme structure (titlebars, iconbars, and shaded titlebars all have different pixmaps)
    load_pixmap_file(pixmap_dir, "top-left-active-shaded.xpm", &config.top_left.pixmap[FRAME_ACTIVE_SHADED], &config.top_left.mask[1], NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "top-left-inactive-shaded.xpm", &config.top_left.pixmap[FRAME_INACTIVE_SHADED], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "top-left-active-iconbar.xpm", &config.top_left.pixmap[FRAME_ACTIVE_ICONBAR], &config.top_left.mask[2], &config.iconbar_left_width, &config.iconbar_left_height, colsym, numsyms); 
  }
  
  
  /* Load top-right pixmaps */ 
  load_pixmap_file(pixmap_dir, "top-right-active.xpm", &config.top_right.pixmap[FRAME_ACTIVE], &config.top_right.mask[0], &config.top_right.width, &config.top_right.height, colsym, numsyms); 
  load_pixmap_file(pixmap_dir, "top-right-inactive.xpm", &config.top_right.pixmap[FRAME_INACTIVE], NULL, NULL, NULL, colsym, numsyms);
  
  if(load_pixmap_file(pixmap_dir, "top-right-inactive-iconbar.xpm", &config.top_right.pixmap[FRAME_INACTIVE_ICONBAR], NULL, NULL, NULL, colsym, numsyms)==-1) {
    printf("%s info: trying alpha 4 theme compatibility...\n",PROGNAME);
    // backward compatibility, with top-right-(in)active-shaded as iconbar and top-right-(in)active as shaded
    load_pixmap_file(pixmap_dir, "top-right-active-shaded.xpm", &config.top_right.pixmap[FRAME_ACTIVE_ICONBAR], &config.top_right.mask[2], NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "top-right-inactive-shaded.xpm", &config.top_right.pixmap[FRAME_INACTIVE_ICONBAR], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "top-right-active.xpm", &config.top_right.pixmap[FRAME_ACTIVE_SHADED], &config.top_right.mask[1], NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "top-right-inactive.xpm", &config.top_right.pixmap[FRAME_INACTIVE_SHADED], NULL, NULL, NULL, colsym, numsyms);
    config.iconbar_right_width=config.top_right.width;
    config.iconbar_right_height=config.top_right.height;
  } else {
    // new theme structure (titlebars, iconbars, and shaded titlebars all have different pixmaps)
    load_pixmap_file(pixmap_dir, "top-right-active-shaded.xpm", &config.top_right.pixmap[FRAME_ACTIVE_SHADED], &config.top_right.mask[1], NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "top-right-inactive-shaded.xpm", &config.top_right.pixmap[FRAME_INACTIVE_SHADED], NULL, NULL, NULL, colsym, numsyms); 
    load_pixmap_file(pixmap_dir, "top-right-active-iconbar.xpm", &config.top_right.pixmap[FRAME_ACTIVE_ICONBAR], &config.top_right.mask[2], &config.iconbar_right_width, &config.iconbar_right_height, colsym, numsyms); 
  }

 
  /* Load left pixmaps */ 
  load_pixmap_file(pixmap_dir, "left-active.xpm", &config.left.pixmap[FRAME_ACTIVE], &config.left.mask[0], &config.left.width, &config.left.height, colsym, numsyms); 
  load_pixmap_file(pixmap_dir, "left-inactive.xpm", &config.left.pixmap[FRAME_INACTIVE], NULL, NULL, NULL, colsym, numsyms); 
 
 
  /* Load right pixmaps */ 
  load_pixmap_file(pixmap_dir, "right-active.xpm", &config.right.pixmap[FRAME_ACTIVE], &config.right.mask[0], &config.right.width, &config.right.height, colsym, numsyms); 
  load_pixmap_file(pixmap_dir, "right-inactive.xpm", &config.right.pixmap[FRAME_INACTIVE], NULL, NULL, NULL, colsym, numsyms); 
 
 
  /* Load bottom-left pixmaps */ 
  load_pixmap_file(pixmap_dir, "bottom-left-active.xpm", &config.bottom_left.pixmap[FRAME_ACTIVE], &config.bottom_left.mask[0], &config.bottom_left.width, &config.bottom_left.height, colsym, numsyms); 
  load_pixmap_file(pixmap_dir, "bottom-left-inactive.xpm", &config.bottom_left.pixmap[FRAME_INACTIVE], NULL, NULL, NULL, colsym, numsyms); 
 
 
  /* Load bottom pixmaps */ 
  load_pixmap_file(pixmap_dir, "bottom-active.xpm", &config.bottom.pixmap[FRAME_ACTIVE], &config.bottom.mask[0], &config.bottom.width, &config.bottom.height, colsym, numsyms); 
  load_pixmap_file(pixmap_dir, "bottom-inactive.xpm", &config.bottom.pixmap[FRAME_INACTIVE], NULL, NULL, NULL, colsym, numsyms); 
} 
 

