/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 

#include "hackcmds.h"
#include "oroborus.h"
#include "communication.h"
#include "iconbars.h"

int mouse_x, mouse_y, lastx=-100, lasty=-100, lastcnt=0, lastbadenter=0, got_button=False;
int check_button_grab=0, check_ptrwin=0, check_top_client=0, delayed_can_focus=0;
double lastmv=0.0;
Client *lastc=NULL, *last_button1_client=NULL;
int replay_x, replay_y, replay_if_drag=0, orgmove_x, orgmove_y, replay_motion=0;//, got_replay_pointer=0;
XEvent replay_button_event;
XEvent replay_motion_event;
Window replay_win, last_pointer_grabbed=0;
Time last_button1_time, replay_time, last_motion_time;
long last_gotbutton_time=0, next_get_cursorpos_time=0, last_get_cursorpos_time=0;
long no_grabbing_until=0, did_force_ungrab=0, delayed_enter_ticks=0;
Client *last_replay_client=NULL, *delayed_enter_client=NULL;
#ifdef KEYBOARD_SUPPORT
int tab_client_state;
Time last_tab_time;
#endif

#ifdef KEYBOARD_SUPPORT

/*
 * 
 * Handle key press event
 *
 */

void event_key_press(XKeyEvent *ev)
{
  Client *c=NULL;

  sprintf(lastsub,"TRACE1: event_handle_key_press (win: %#lx)\n",ev->window);
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_handle_key_press (win: %#lx)\n",ev->window);
#else
#ifdef SHOW_EVENTS
  printf("%s X11 event: key press for win %#lx\n",PROGNAME,ev->window);
#else
  if(gVerboseConsole)
    printf("%s X11 event: key press for win %#lx\n",PROGNAME,ev->window);
#endif
#endif

  c=get_client_from_window(ev->window, WINDOW);

  ev->state=ev->state & (Mod1Mask|ControlMask|ShiftMask);
  
  /* If we have an active client */
  if(c!=NULL)
  {
    /* Maximize client */
    if(ev->keycode==config.keys.maximize.keycode && ev->state==config.keys.maximize.modifier)
      toggle_client_maximized(c,0);
      
    /* Shade client */
    if(ev->keycode==config.keys.shade.keycode && ev->state==config.keys.shade.modifier)
      toggle_client_shaded(c);

    /* Hide client */
    if(ev->keycode==config.keys.hide.keycode && ev->state==config.keys.hide.modifier)
      minmax_client(c);
     
    /* Close client */ 
    if(ev->keycode==config.keys.close.keycode && ev->state==config.keys.close.modifier) 
      close_client(c); 
       
    /* Move client up */ 
    if(ev->keycode==config.keys.move_up.keycode && ev->state==config.keys.move_up.modifier)
    {
      mouse_mode=MOUSE_STOPPED; 
      move_client(c, c->x, c->y-10, 0); 
    }
       
    /* Move client down */ 
    if(ev->keycode==config.keys.move_down.keycode && ev->state==config.keys.move_down.modifier)
    {
      mouse_mode=MOUSE_STOPPED; 
      move_client(c, c->x, c->y+10, 0); 
    }
       
    /* Move client left */ 
    if(ev->keycode==config.keys.move_left.keycode && ev->state==config.keys.move_left.modifier)
    {
      mouse_mode=MOUSE_STOPPED; 
      move_client(c, c->x-10, c->y, 0); 
    }
       
    /* Move client right */ 
    if(ev->keycode==config.keys.move_right.keycode && ev->state==config.keys.move_right.modifier) 
    {
      mouse_mode=MOUSE_STOPPED;
      move_client(c, c->x+10, c->y, 0); 
    }
 
    if(!(get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_MAXIMIZED) && !(get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)) 
    { 
      if(ev->keycode==config.keys.resize_up.keycode && ev->state==config.keys.resize_up.modifier) 
      {
        mouse_mode=MOUSE_STOPPED;
        resize_client(c, c->width, c->height-(c->size->flags & (PResizeInc?c->size->height_inc:10)), 0); 
      }
      if(ev->keycode==config.keys.resize_down.keycode && ev->state==config.keys.resize_down.modifier) 
      {
        mouse_mode=MOUSE_STOPPED;
        resize_client(c, c->width, c->height+(c->size->flags & (PResizeInc?c->size->height_inc:10)), 0);
      } 
      if(ev->keycode==config.keys.resize_left.keycode && ev->state==config.keys.resize_left.modifier) 
      {
        mouse_mode=MOUSE_STOPPED;
        resize_client(c, c->width-(c->size->flags & (PResizeInc?c->size->width_inc:10)), c->height, 0);
      } 
      if(ev->keycode==config.keys.resize_right.keycode && ev->state==config.keys.resize_right.modifier) 
      {
        mouse_mode=MOUSE_STOPPED;
        resize_client(c, c->width+(c->size->flags & (PResizeInc?c->size->width_inc:10)), c->height, 0); 
      }
    } 
     
#ifdef LAYER_SUPPORT 
    /* Raise client layer */ 
    if(ev->keycode==config.keys.raise_layer.keycode && ev->state==config.keys.raise_layer.modifier) 
      set_client_layer(c, get_client_layer(c)+1); 
       
    /* Lower client layer */ 
    if(ev->keycode==config.keys.lower_layer.keycode && ev->state==config.keys.lower_layer.modifier) 
      set_client_layer(c, get_client_layer(c)-1); 
#endif /* LAYER */ 
 
#ifdef WORKSPACE_SUPPORT 
    /* Toggle sticky */ 
    if(ev->keycode==config.keys.sticky.keycode && ev->state==config.keys.sticky.modifier) 
      toggle_client_sticky(c); 
       
    /* Move window to next workspace */ 
    if(ev->keycode==config.keys.move_next_ws.keycode && ev->state==config.keys.move_next_ws.modifier) 
      move_client_to_workspace(c, workspace+1); 
 
    /* Move window to prev workspace */ 
    if(ev->keycode==config.keys.move_prev_ws.keycode && ev->state==config.keys.move_prev_ws.modifier) 
      move_client_to_workspace(c, workspace-1); 
       
    /* Move window to specific workspace */ 
    if(ev->keycode==config.keys.move_ws_1.keycode && ev->state==config.keys.move_ws_1.modifier) 
      move_client_to_workspace(c, 0);      
    if(ev->keycode==config.keys.move_ws_2.keycode && ev->state==config.keys.move_ws_2.modifier) 
      move_client_to_workspace(c, 1); 
    if(ev->keycode==config.keys.move_ws_3.keycode && ev->state==config.keys.move_ws_3.modifier) 
      move_client_to_workspace(c, 2); 
    if(ev->keycode==config.keys.move_ws_4.keycode && ev->state==config.keys.move_ws_4.modifier) 
      move_client_to_workspace(c, 3); 
    if(ev->keycode==config.keys.move_ws_5.keycode && ev->state==config.keys.move_ws_5.modifier) 
      move_client_to_workspace(c, 4); 
    if(ev->keycode==config.keys.move_ws_6.keycode && ev->state==config.keys.move_ws_6.modifier) 
      move_client_to_workspace(c, 5); 
    if(ev->keycode==config.keys.move_ws_7.keycode && ev->state==config.keys.move_ws_7.modifier) 
      move_client_to_workspace(c, 6); 
    if(ev->keycode==config.keys.move_ws_8.keycode && ev->state==config.keys.move_ws_8.modifier) 
      move_client_to_workspace(c, 7); 
    if(ev->keycode==config.keys.move_ws_9.keycode && ev->state==config.keys.move_ws_9.modifier) 
      move_client_to_workspace(c, 8); 
#endif /* WORKSPACE */ 
  } 
 
  /* Next window */ 
  if(ev->keycode==config.keys.next_window.keycode && ev->state==config.keys.next_window.modifier) 
  { 
    Client *c2=NULL; 
 
    /* Check time of last tab */ 
    if(ev->time-last_tab_time<TAB_RAISE_TIMEOUT) 
    { 
     /* If window was hidden then hide it again 
     if(tab_client_state==IconicState && c!=NULL) */
     if((tab_client_state!=0) && (c!=NULL) )
        minmax_client(c); 
    } 
    last_tab_time=ev->time; 
    if(c)
      c2=get_next_client(c);
    else if(client_list)
      c2=get_next_client(client_list->prev); 
    if(c2!=NULL) 
    { 
/*      tab_client_state=get_wm_state(c2->window); */
      tab_client_state=c2->is_icon;
      if(tab_client_state!=0)
        minmax_client(c2);
      raise_client(c2); 
      set_focus_client(c2); 
    } 
  }

  /* Previous window */ 
  if(ev->keycode==config.keys.prev_window.keycode && ev->state==config.keys.prev_window.modifier) 
  { 
    Client *c2=NULL; 
 
    /* Check time of last tab */ 
    if(ev->time-last_tab_time<TAB_RAISE_TIMEOUT) 
    { 
     /* If window was hidden then hide it again 
     if(tab_client_state==IconicState && c!=NULL)  */
     if(tab_client_state!=0 && c!=NULL)
        minmax_client(c); 
    } 
    last_tab_time=ev->time; 
    if(c)
      c2=get_prev_client(c);
    else if(client_list)
      c2=get_prev_client(client_list->next); 
    if(c2!=NULL) 
    { 
/*      tab_client_state=get_wm_state(c2->window); */
      tab_client_state=c2->is_icon;
      if(tab_client_state!=0)
        minmax_client(c2);
      raise_client(c2);
      set_focus_client(c2); 
    }
  } 
 
#ifdef WORKSPACE_SUPPORT   
  /* Switch to next workspace */ 
  if(ev->keycode==config.keys.next_ws.keycode && ev->state==config.keys.next_ws.modifier) 
    change_to_workspace(workspace+1); 
 
  /* Switch to prev workspace */ 
  if(ev->keycode==config.keys.prev_ws.keycode && ev->state==config.keys.prev_ws.modifier) 
    change_to_workspace(workspace-1); 
 
  /* Add new workspace */ 
  if(ev->keycode==config.keys.add_ws.keycode && ev->state==config.keys.add_ws.modifier) 
    set_gnome_hint(root, WIN_WORKSPACE_COUNT, config.workspace_count+1); 
 
  /* Delete last workspace */ 
  if(ev->keycode==config.keys.del_ws.keycode && ev->state==config.keys.del_ws.modifier) 
    set_gnome_hint(root, WIN_WORKSPACE_COUNT, config.workspace_count-1); 
 
  /* Switch to specific workspace */ 
  if(ev->keycode==config.keys.ws_1.keycode && ev->state==config.keys.ws_1.modifier) 
    change_to_workspace(0); 
  if(ev->keycode==config.keys.ws_2.keycode && ev->state==config.keys.ws_2.modifier) 
    change_to_workspace(1); 
  if(ev->keycode==config.keys.ws_3.keycode && ev->state==config.keys.ws_3.modifier) 
    change_to_workspace(2); 
  if(ev->keycode==config.keys.ws_4.keycode && ev->state==config.keys.ws_4.modifier) 
    change_to_workspace(3); 
  if(ev->keycode==config.keys.ws_5.keycode && ev->state==config.keys.ws_5.modifier) 
    change_to_workspace(4); 
  if(ev->keycode==config.keys.ws_6.keycode && ev->state==config.keys.ws_6.modifier) 
    change_to_workspace(5); 
  if(ev->keycode==config.keys.ws_7.keycode && ev->state==config.keys.ws_7.modifier) 
    change_to_workspace(6); 
  if(ev->keycode==config.keys.ws_8.keycode && ev->state==config.keys.ws_8.modifier) 
    change_to_workspace(7); 
  if(ev->keycode==config.keys.ws_9.keycode && ev->state==config.keys.ws_9.modifier) 
    change_to_workspace(8); 
#endif /* WORKSPACE */ 
} 
#endif /* KEYBOARD */ 
 
/* 
 * 
 * Handle button press events 
 * 
 */ 

void event_button_press(XButtonEvent *ev) 
{ 
  Client *c=NULL; 
  Client *win=NULL;
  Window ptr_win; 
  int replay=False, focus=True, raise=True, xinfront=True, noshift_state;
//  unsigned int stateorg; 
//  XWindowChanges wc; 

  sprintf(lastsub,"TRACE1: event_button_press (win,last: %#lx,%#lx,%d,%d)\n",ev->window,FC_WINDOW,ev->button,ev->state); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_button_press (win,last: %#lx,%#lx,%d,%d)\n",ev->window,FC_WINDOW,ev->button,ev->state); 
#else
#ifdef SHOW_EVENTS
  printf("%s X11 event: button down (win,last: %#lx,%#lx,%d,%d)\n",PROGNAME,ev->window,FC_WINDOW,ev->button,ev->state); 
#else
    if(gVerboseConsole)
        printf("%s X11 event: button down (win,last: %#lx,%#lx,%d,%d)\n",PROGNAME,ev->window,FC_WINDOW,ev->button,ev->state); 
#endif
#endif
 
  /*got_button=True;*/

/*
This "replay_if_drag" stuff is totally inoperative, so forget it for now...

  if(replay_if_drag==2) {
#ifdef SHOW_EVENTS
    printf("   got event again: replaying...\n");
#endif
    // got click event for the second time, so try to replay it to the window
    // does not work for some reason...?
    XAllowEvents(dpy, ReplayPointer, CurrentTime);
    replay_if_drag=0;
    return;
  } else if(replay_if_drag==1) {
#ifdef SHOW_EVENTS
    printf("   allowing events (sync) on button down...\n");
#endif
    replay_if_drag=0;
    XAllowEvents(dpy, SyncPointer, CurrentTime);
  }
*/

  if( gotCommandDown ) {
    if(gVerboseConsole)
        printf("%s info: got butdown event before command key up - cancelling...\n",PROGNAME);
    gotCommandDown = 0;
  }
//  replay_button_event=*(XEvent *)ev;

//  win=get_focus_client();
  win=lastfocus_client;
  if(gX11InFront==0)
    win=NULL;
  else
    xinfront=False;
  
  if(ev->window!=check_win) {
    last_replay_client = NULL;
    did_force_ungrab = 0;
  }
/*
  if( keepInFrontUntil==-1 ) {
    keepInFrontUntil=0;
    printf("%s info: Leaving command key shortcut mode...\n",PROGNAME);
  }
*/
  last_pointer_grabbed = pointer_grabbed;
  pointer_grabbing(0);
  c=get_client_from_window(ev->window, FRAME_OR_COVER); 

  if(c==NULL) {
    ptr_win = get_window_parent(ev->window);
    c=get_client_from_window(ptr_win, FRAME_OR_COVER);
    if(c!=NULL) ev->window=ptr_win;
  }
  
  ev->state=ev->state & (Mod1Mask|ControlMask|ShiftMask|Mod2Mask);
  noshift_state = ev->state & (Mod1Mask|ControlMask|Mod2Mask);

  current_resize_client=NULL;
 
  check_button_grab = 1;
  
  // always send scroll wheel events in root to the current focus window
  if(c==NULL && (ev->button==4 || ev->button==5)) {
    if(lastfocus_client==NULL) return;
    c=lastfocus_client;
    ev->window = c->window;
  }
  
  /* Clicked on root */ 
  if(c==NULL) 
  { 
//#ifdef DEBUG
    if(gVerboseConsole)
      printf("Button pressed in non-client %d %d %d\n",ev->x_root,ev->y_root,ev->state);
//#endif
/*    if(ev->button==1)
        c=check_for_undock(ev->x_root,ev->y_root);
        if(c!=NULL) {
            XAllowEvents(dpy, SyncPointer, CurrentTime);
            c->dock_timer=0;
            c->undock_timer=0;
            undock_client(c);
            return;
        }*/
    /* Proxy root click */ 
//    c=get_focus_client();
    // don't want any window to have focus...
    gone_to_back();
    XSendEvent(dpy, check_win, False, SubstructureNotifyMask, (XEvent *)ev);
  } 
  /* Clicked somewhere on a client */ 
  else 
  { 
//#ifdef DEBUG
    if(gVerboseConsole)
      printf("Button pressed in client win %#lx, %d %d\n",C_WINDOW,c->is_docked,ev->state);
//#endif
/*    if(c->is_docked && ev->button==1) {
        XAllowEvents(dpy, SyncPointer, CurrentTime);
        c->dock_timer=0;
        c->undock_timer=0;
        undock_client(c);
        return;
    }*/
    
    ptr_win=get_pointer_win(c->frame);
    // Don't want clicks to go through to the client if it ignores clicks
    if( c->auto_properties&NONAUTO_IGNORECLICKS )
        if( ptr_win == c->window ) ptr_win = c->frame;
    // This can set back above, unfortunately, so there's another test later which resets
    // the replay flags to prevent clicks getting through to a client which ignores clicks
    if((ptr_win==c->frame || ptr_win==0) && (c->auto_properties&NONAUTO_DRAGFIRSTCLICK)==0)
        ptr_win = c->window;
    
    if( c->is_icon!=NOT_ICONBAR )
    {
        // clicked on an iconbar -don't wait for mouse up, just go ahead and restore it
      minmax_client(c);
    }
    
    /* If we want to resize */ 
    else if((ptr_win==c->resize && ((ev->button==Button1 && (noshift_state==0 || ev->state==Mod2Mask)) || (ev->button==Button2 && noshift_state==0))) || (ev->button==Button2 && ev->state==(Mod1Mask|ControlMask))) 
    { 
//      printf("Unmapping window in buttonpress\n");
//      c->ignore_unmap++; 
//      XUnmapWindow(dpy, c->window); 
//      XUngrabPointer(dpy,CurrentTime);
      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, resize_curs, CurrentTime); 
         
      /* Set mouse_x and mouse_y to position of cursor relative to southeast corner of window */ 
      mouse_x=FW-(ev->x_root-FX); 
      mouse_y=FH-(ev->y_root-FY); 
      mouse_mode=MOUSE_RESIZE;
      current_resize_client=c;
// start the timer for motion events
      if(gX11InFront==0)
        last_motion_time=ev->time+BG_FIRST_MOTION_GAP;
      else
        last_motion_time=ev->time+FIRST_MOTION_GAP;
        
      if((ev->button==Button1 && ev->state==Mod2Mask) || (ev->button==Button2 && noshift_state==0)) {
        raise=False;
        focus=False;
        xinfront=False;
      }
    } 
 
    /* If we want to maximise */ 
    else if(ptr_win==c->zoom && (ev->button==Button1 || ev->button==Button2 || ev->button==Button3) && noshift_state==0) 
    { 
      raise=False;
      focus=False;
      xinfront=False;
// This next bit of code makes it behave like OS X windows -i.e. it comes straight in front as soon
// as maximise is pressed, rather than waiting for the user to 'let go' of the button -but I don't like it!!
//      raise_client(c);
//      set_focus_client(c);
//      win=c;
//      raise=True;
//      focus=True;
//      xinfront=true;
      if(c==win)
         c->zoom_state=BUTTON_PRESSED;
      if(c!=win)
         c->zoom_state=BUTTON_PRESSEDBACK;
      mouse_mode=MOUSE_ZOOM; 
//      XUngrabPointer(dpy,CurrentTime);
      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, None, CurrentTime); 
// start the timer for motion events (don't need to bother with such a large gap at the start)
      last_motion_time=ev->time; 
      redraw_client_buttons(c); 
    } 
 
    /* If we want to 'iconbar-ise' */ 
    else if(ptr_win==c->icon && ev->button==Button1 && noshift_state==0) 
    { 
      if(c==win)
         c->icon_state=BUTTON_PRESSED; 
      if(c!=win)
         c->icon_state=BUTTON_PRESSEDBACK;
      mouse_mode=MOUSE_ICON; 
//      XUngrabPointer(dpy,CurrentTime);
      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, None, CurrentTime); 
// start the timer for motion events (don't need to bother with such a large gap at the start)
      last_motion_time=ev->time; 
      raise=False;
      focus=False;
      xinfront=False;
      redraw_client_buttons(c); 
    } 
 
    /* If we want to close */ 
    else if(ptr_win==c->close && ev->button==Button1 && noshift_state==0) 
    { 
      if(c==win)
         c->close_state=BUTTON_PRESSED;
      if(c!=win)
         c->close_state=BUTTON_PRESSEDBACK;
      mouse_mode=MOUSE_CLOSE; 
//      XUngrabPointer(dpy,CurrentTime);
      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, None, CurrentTime); 
      raise=False;
      focus=False;
      xinfront=False;
// start the timer for motion events (don't need to bother with such a large gap at the start)
      last_motion_time=ev->time; 
      redraw_client_buttons(c); 
    } 
 
    /* If we want to windowshade */ 
    else if(ptr_win!=c->window && ev->time-last_button1_time<DBL_CLICK && c==last_button1_client && ((ev->button==Button1 && (noshift_state==0 || ev->state==Mod2Mask)) || (ev->button==Button2 && noshift_state==0))) 
    { 
        // got a double-click on the title-bar
#ifdef DEBUG
        printf("Got button down again in titlebar...\n");
#endif
        if(c->has_frame!=0) toggle_client_shaded(c);
        last_button1_time=0; 
        last_button1_client=NULL;
        raise=False;
        focus=False;
        xinfront=False;
    }
     
    /* If we want to move */ 
    else if((ptr_win!=c->window && ((ev->button==Button1 && (noshift_state==0 || ev->state==Mod2Mask)) || (ev->button==Button2 && noshift_state==0))) || (ev->button==Button1 && ev->state==(Mod1Mask|ControlMask))) 
    { 
#ifdef DEBUG
        printf("Got button down in titlebar...\n");
#endif
 //       XUngrabPointer(dpy,CurrentTime);
        //printf("Got button down in window %#lx\n",ptr_win);
        //printf("frame,win,title,bottom: %#lx %#lx %#lx %#lx\n",c->frame,c->window,c->title,c->bottom);
        XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, move_curs, CurrentTime); 
 
        /* Set mouse_x and mouse_y to position of cursor relative to root window */ 
        mouse_x=ev->x_root; 
        mouse_y=ev->y_root;
        orgmove_x=mouse_x;
        orgmove_y=mouse_y;
        mouse_mode=MOUSE_MOVE; 
        if((ev->button==Button1 || ev->button==Button2) && (noshift_state==0 || ev->state==Mod2Mask) ) {
          if(c->has_frame!=0) {
            last_button1_time=ev->time;
            last_button1_client=c;
          }
          if(ev->state==Mod2Mask || ev->button==Button2) {
            raise=False;
            focus=False;
            xinfront=False;
          }
        } else {
          last_button1_time=0;
          last_button1_client=NULL;
          raise=False;
          focus=False;
          xinfront=False;
        }
// start the timer for motion events
        if(gX11InFront==0)
          last_motion_time=ev->time+BG_FIRST_MOTION_GAP;
        else
          last_motion_time=ev->time+FIRST_MOTION_GAP;
    } 
 
    /* If we want to 'iconbar-ise' */ 
    else if(ptr_win!=c->window && ev->button==Button3 && noshift_state==0) 
    {
      mouse_mode=MOUSE3_TITLE; 
//      XUngrabPointer(dpy,CurrentTime);
      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, None, CurrentTime); 
//      minmax_client(c); 
      raise=False;
      focus=False;
      xinfront=False;
    }

    /* If we want to send to back */ 
    else if((ptr_win!=c->window && ev->button==Button3 && ev->state==ShiftMask) || (ev->button==Button3 && ev->state==(Mod1Mask|ControlMask))) 
    { 
      lower_client(c); 
      raise=False; 
      focus=False; 
      xinfront=False;
    } 
 
#ifdef WORKSPACE_SUPPORT 
    /* If we want to make window sticky */ 
    else if(ptr_win!=c->window && ev->button==Button2 && ev->state==Mod1Mask) 
      toggle_client_sticky(c); 
#endif /* WORKSPACE */ 
 
#ifdef LAYER_SUPPORT 
    /* If we want to move to a higher layer */ 
    else if(ptr_win!=c->window && ev->button==Button1 && ev->state==Mod1Mask) 
      set_client_layer(c, get_client_layer(c)+1); 
 
    /* If we want to send to a lower layer */ 
    else if(ptr_win!=c->window && ev->button==Button3 && ev->state==Mod1Mask) 
      set_client_layer(c, get_client_layer(c)-1); 
#endif /* LAYER */ 
    else if( gX11InFront==0 ) {
      // left-click in a window should just bring XDarwin/OroborOSX in front, since it isn't already
      if(ev->button==Button1 && ev->state==0) {
//This "replay_if_drag" stuff is totally inoperative, so forget it for now...
//        replay_if_drag=1;
#ifdef DEBUG
        printf("Got button in window -is but1\n");
#endif
        raise=True;
        focus=True;
        xinfront=True;
      } else {
        replay=True;
        raise=False;
        focus=False;
        xinfront=False;
#ifdef DEBUG
        printf("Got button in window -not but1\n");
#endif
      }
//    } else if( get_focus_client()!=c ) {
    } else if( lastfocus_client!=c ) {
      // left-click in a window should just bring it in front, since it isn't already
      if(ev->button==Button1 && ev->state==0) {
        if(gVerboseConsole)
          printf("But1 click in non-focus win...\n");
//This "replay_if_drag" stuff is totally inoperative, so forget it for now...
//        replay_if_drag=1;
        raise=True;
        focus=True;
//        if(lastfocus_client==c) focus=False;
//        if(get_top_most_client(WIN_LAYER_NORMAL,False)==c) raise=False;
        // if this window is set to accept the first click, then allow it go through
        if((c->auto_properties & NONAUTO_FIRSTCLICK)==NONAUTO_FIRSTCLICK ||
                                client_modal_firstclick(c)==lastfocus_client)
            replay=True;
      } else {
        if(gVerboseConsole)
          printf("Click (not but1) in non-focus win...\n");
        replay=True;
        raise=False;
        focus=False;
      }
      xinfront=False;
    } else {
      if(gVerboseConsole)
        printf("Click in focus win...\n");
      // click in focus window, so allow it through
      replay=True;
      raise=False;
      focus=False;
      xinfront=False;
      // if it is an autofocus window set to raise on click, then raise it
      if((c->auto_properties & AUTO_CLICKRAISE)==AUTO_CLICKRAISE)
        raise=True;
    }
 
    if( c->auto_properties & NONAUTO_IGNORECLICKS ) {
        replay = False;
        replay_if_drag = 0;
    }
    
    if((replay || xinfront) && lastfocus_client==c &&
        !focus && c->is_icon==NOT_ICONBAR) quick_set_focus_client(c);
    // get this done a.s.a.p. before XDarwin resets the modifier if it's emulated middle/right-click
    if(replay) {
      last_replay_client = c;
      XAllowEvents(dpy, ReplayPointer, CurrentTime);
      MyXSync(dpy,0);
      // Need to allow time for the client to do its pointer grab before event_idle does its own (for checking
      // purposes) - so we set a timer which prevents event_idle from doing it for the next second (or until
      // some other button event is received)
      last_gotbutton_time = GetCarbonTickCount();
      no_grabbing_until = last_gotbutton_time+60;
      if(gVerboseConsole)
        printf("%s info: Waiting until TickCount>%d before checking grab\n",PROGNAME,no_grabbing_until);
    } else
      no_grabbing_until = 0;
    
    if(c->is_icon==NOT_ICONBAR) {
      Client *ac = NULL;
      if(can_send_hackcmds && (focus || xinfront)) MyIncreaseXServerGrab();
      ac = check_client_modality(c,0);
      win =ac;
      if(gVerboseConsole)
        printf("here with win/focus/raise/xinfront=%#lx, %d, %d, %d\n",AC_WINDOW,focus,raise,xinfront);
      if(can_send_hackcmds==0) {
        if(xinfront) {
          int old_mm=mouse_mode; // we want to make sure the X11 window really does
          mouse_mode=-1;         // come in front (in case of keepInFrontUntil==-1)
          if(raise) force_raise_client(c);
          // need to bring XDarwin in front before it notices and brings itself in front
          Bring_X_To_Front(1);
          if(focus) set_focus_client(c);
//          if(raise) raise_client(c); 
          Bring_To_Front(0);
          mouse_mode=old_mm;
        }
        if(focus) set_focus_client(c);
        //else if( lastfocus_client==c ) quick_set_focus_client(c);
        if(raise) raise_client(c); 
      } else {
        // the modified XDarwin does not bring itself in front
        if(xinfront) {
          int old_mm=mouse_mode; // we want to make sure the X11 window really does
          mouse_mode=-1;         // come in front (in case of keepInFrontUntil==-1)
//          if(raise) force_raise_client(c);
          if(raise) XRaiseWindow(dpy,check_win);
          if(focus) set_focus_client(NULL);
//          if(raise) force_raise_client(c);
          Bring_To_Front(1);
          gX11InFront=1;
          mouse_mode=old_mm;
        }
        if(raise) raise_client(c);
        if(focus && lastfocus_client!=win) set_focus_client(win);
        //if(xinfront) come_to_front();
/*        if(focus) {
          if(raise) {
            if(!xinfront) {
              if(get_top_most_client(WIN_LAYER_NORMAL,False)!=c)
                raise_client(c);
            } else if(get_top_most_client(WIN_LAYER_NORMAL,False)!=win)
              raise_client(c);
          }
          if(lastfocus_client!=win)
            set_focus_client(win);
        } else if(raise) {
          if(get_top_most_client(WIN_LAYER_NORMAL,False)!=c || xinfront)
            raise_client(c);
        }*/
//        if(lastfocus_client==c && !focus) quick_set_focus_client(c);
//        if(gInterleaving_Windows==0) redraw_client_frame(c);
        if(focus || xinfront) MyDecreaseXServerGrab();
        if( xinfront && Is_Whole_Screen_Mode()==1 )
            Wait_OOSX_To_Front();
      }
    }

    if(replay)  {
//      XAllowEvents(dpy, ReplayPointer, CurrentTime);
#ifdef SHOW_EVENTS
      printf("  allowed events for replay on button down... %d,%d,%d,%d \n",(ev->state&ShiftMask),(ev->state&ControlMask),(ev->state&Mod1Mask),(ev->state&Mod2Mask));
#else
      if(gVerboseConsole)
        printf("  allowed events for replay on button down... %d,%d,%d,%d \n",(ev->state&ShiftMask),(ev->state&ControlMask),(ev->state&Mod1Mask),(ev->state&Mod2Mask));
#endif
      // Event can be sent through to window
//      XAllowEvents(dpy, SyncPointer, CurrentTime);
//      XSendEvent(dpy, c->window, False, ButtonPressMask, (XEvent *) &replay_button_event);
      ev->window = c->window;
//      printf("Sending lastmods to XDarwin after replay: %#lx\n",lastsentmods);
//      SendModsToXDarwin(lastsentmods,lastsentmods);
//      got_replay_pointer=2;
    } else if(replay_if_drag==1) {
//      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, None, CurrentTime); 
 
    /* Set mouse_x and mouse_y to position of cursor relative to root window */ 
      replay_x=ev->x_root; 
      replay_y=ev->y_root; 
#ifdef SHOW_EVENTS
      printf("  Starting drag check: %d,%d\n",replay_x,replay_y);
#else
      if(gVerboseConsole)
        printf("  Starting drag check: %d,%d\n",replay_x,replay_y);
#endif
      replay_time=ev->time;
      replay_win=c->frame;
      XAllowEvents(dpy, SyncPointer, CurrentTime);
// try to get around the cursor bug...
      if(xinfront) {
        XUndefineCursor(dpy, c->frame);
        XDefineCursor(dpy, c->frame, normal_curs);
      }
// This is not working, so forget it for now...
      replay_if_drag=0;
    } else {
#ifdef SHOW_EVENTS
      printf("  allowing events (sync) on button down...\n");
#else
      if(gVerboseConsole)
        printf("  allowing events (sync) on button down...\n");
#endif
      XAllowEvents(dpy, SyncPointer, CurrentTime);
// try to get around the cursor bug...
      if(xinfront) {
        XUndefineCursor(dpy, c->frame);
        XDefineCursor(dpy, c->frame, normal_curs);
      }
    }
  } 
} 

Client *check_for_hackid(int hackid)
{
    Client *c;
    int i;
    
    c=client_list;
    if(c==NULL) return NULL;
    if(hackid==0) return NULL;
    
    for(i=0;i<client_count;i++) {
        if(hackid==c->hack_id)
            return c;
        c=c->next;
    }
    return NULL;
}

/* 
 * 
 * Handle button release events 
 * 
 */ 

void event_button_release(XButtonEvent *ev) 
{ 
  Client *c=NULL; 
  int dx, dy;
 
  sprintf(lastsub,"TRACE1: event_button_release (win,last: %#lx,%#lx,%d,%d)\n",ev->window,lastfocus,ev->button,ev->state);
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_button_release (win,last: %#lx,%#lx,%d,%d)\n",ev->window,lastfocus,ev->button,ev->state);
#else
#ifdef SHOW_EVENTS
  if(ev->button!=-5)
    printf("%s X11 event: button up (win,last: %#lx,%#lx,%d,%d)\n",PROGNAME,ev->window,lastfocus,ev->button,ev->state);
#else
    if(gVerboseConsole && ev->button!=-5)
      printf("%s X11 event: button up (win,last: %#lx,%#lx,%d,%d)\n",PROGNAME,ev->window,lastfocus,ev->button,ev->state);
#endif
#endif 
 
/*
This "replay_if_drag" stuff is totally inoperative, so forget it for now...

  if(replay_if_drag!=0) {
    replay_if_drag=0;
#ifdef SHOW_EVENTS
    printf("   allowing events (sync) on button release...\n");
#endif
  }
*/

 
    if(ev->button==-5) {
#ifdef DEBUG
        printf("-------- Got button -5 at %d %d\n",ev->x_root,ev->y_root);
#endif
        check_ptrwin = 1;
        if(ev->y_root==100) {
            c=check_for_undock(ev->x_root);
            if(c!=NULL) {
//                XAllowEvents(dpy, SyncPointer, CurrentTime);
                c->dock_timer=0;
                c->undock_timer=0;
                undock_client(c);
//                XUngrabPointer(dpy, CurrentTime); 
                return;
            }
        } else if(ev->y_root==101) {
            /*if(mouse_mode!=MOUSE_NORMAL) {
#ifdef SHOW_EVENTS
                printf("-------- Ignoring entered at %d %d\n",ev->x_root,ev->y_root);
#endif
                return;
            }*/
            c=check_for_hackid(ev->x_root);
            if(c!=NULL) {
#ifdef SHOW_EVENTS
                printf("%s from XDarwin: entered win %#lx,%#lx,%d\n",PROGNAME,c->window,last_client_entered_hack,c->is_e_m_r);
#else
                if(gVerboseConsole)
                  printf("%s from XDarwin: entered win %#lx,%#lx,%d\n",PROGNAME,c->window,last_client_entered_hack,c->is_e_m_r);
#endif
//                XAllowEvents(dpy, SyncPointer, CurrentTime);
//                XUngrabPointer(dpy, CurrentTime);
                c->is_e_m_r |= IS_ENTERED;
                if(c->is_docked && GetCarbonTickCount()>c->dock_timer) {
                    c->dock_timer=0;
                    c->undock_timer=0;
                    undock_client(c);
                }
/*                c=get_lowest_entered_client(1);
                if(c!=last_client_entered_hack) {
                    //entered_client(c,1);
                    last_client_entered_hack=c;
                    if(c==NULL)
                        lastenter=0;
                    else
                        lastenter=c->window;
                }*/
                return;
            }
        } else if(ev->y_root==102) {
            /*if(mouse_mode!=MOUSE_NORMAL) {
#ifdef SHOW_EVENTS
                printf("-------- Ignoring exited at %d %d\n",ev->x_root,ev->y_root);
#endif
                return;
            }*/
            c=check_for_hackid(ev->x_root);
            if(c!=NULL) {
                if((c->is_e_m_r & LOCK_ENTERED)==LOCK_ENTERED) {
                    c->is_e_m_r ^= LOCK_ENTERED;
                    if(c->is_icon!=NOT_ICONBAR) {
#ifdef SHOW_EVENTS
                        printf("%s info: ignoring iconbar exit (from XDarwin) %#lx,%#lx,%d\n",PROGNAME,c->window,last_client_entered_hack,c->is_e_m_r);
#else
                        if(gVerboseConsole)
                            printf("%s info: ignoring iconbar exit (from XDarwin) %#lx,%#lx,%d\n",PROGNAME,c->window,last_client_entered_hack,c->is_e_m_r);
#endif
                        return;
                    }
                }
#ifdef SHOW_EVENTS
                printf("%s from XDarwin: exited client %#lx,%#lx,%d\n",PROGNAME,c->window,last_client_entered_hack,c->is_e_m_r);
#else
                if(gVerboseConsole)
                    printf("%s from XDarwin: exited client %#lx,%#lx,%d\n",PROGNAME,c->window,last_client_entered_hack,c->is_e_m_r);
#endif
//                XAllowEvents(dpy, SyncPointer, CurrentTime);
//                XUngrabPointer(dpy, CurrentTime); 
                if(c->is_docked && GetCarbonTickCount()>c->dock_timer) {
                    c->dock_timer=0;
                    c->undock_timer=0;
                    undock_client(c);
                }
                c->is_e_m_r &= NOT_ENTERED;
/*                c=get_lowest_entered_client(1);
                if(c!=last_client_entered_hack) {
                    //entered_client(c,1);
                    last_client_entered_hack=c;
                    if(c==NULL)
                        lastenter=0;
                    else
                        lastenter=c->window;
                }*/
                return;
            }
        }
        return;
    }
        
  if( gotCommandDown ) {
    if(gVerboseConsole)
        printf("%s info: got butup event before command key up - cancelling...\n",PROGNAME);
    gotCommandDown = 0;
  }
  if(ev->window!=check_win) {
    last_replay_client = NULL;
    did_force_ungrab = 0;
  }
  current_resize_client=NULL;
  c=get_client_from_window(ev->window, FRAME_OR_COVER);
  no_grabbing_until = 0;
  
  /* Release was on root window */ 
  if(c==NULL) 
  { 
#ifdef DEBUG
    printf("Button released in root %d %d\n",ev->x_root,ev->y_root);
#endif
    /* Proxy root release */ 
    XSendEvent(dpy, check_win, False, SubstructureNotifyMask, (XEvent *)ev); 
  } 
  /* Clicked somewhere on a client */ 
  else 
  { 
#ifdef DEBUG
    printf("Button released in win %#lx, %d\n",c->window,c->is_docked);
#endif
/*    if(c->is_docked && ev->button==1) {
        XAllowEvents(dpy, SyncPointer, CurrentTime);
        c->dock_timer=0;
        c->undock_timer=0;
        undock_client(c);
        return;
    }*/

    if(c->is_icon!=NOT_ICONBAR)
    {
      /* minmax_client(c); */
    }
    else
    {
      c->zoom_state=BUTTON_UNPRESSED; 
      c->icon_state=BUTTON_UNPRESSED;
      c->close_state=BUTTON_UNPRESSED; 
       
      /* Release of button 3 occured on titlebar */
      if(mouse_mode==MOUSE3_TITLE && ev->button==3) {
        if(get_pointer_win(c->frame)!=c->window) {
          //pointer_grabbing(last_pointer_grabbed);
          XUngrabPointer(dpy, CurrentTime); 
//          XGrabPointer(dpy, c->frame, True, ButtonPressMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, resize_curs, CurrentTime);
          XAllowEvents(dpy, SyncPointer, CurrentTime);
          mouse_mode=MOUSE_NORMAL;
          got_button=False;
          toggle_minimise_client(c,1); // perform the alternative minimisation method
          return;
        }
      /* Release occured on maximise button */
      } else if(mouse_mode==MOUSE_ZOOM || mouse_mode==MOUSE_ZOOM_OUT) {
        redraw_client_buttons(c); 
        if(get_pointer_win(c->frame)==c->zoom) {
          int wasFront = gX11InFront;
          raise_client(c);
          gX11InFront = 1;
          if( lastfocus_client!=c ) set_focus_client(c);
          if(wasFront==0 && Is_Whole_Screen_Mode()==0 ) {
            gX11InFront = 0;
            Bring_To_Front(0);
          }
          toggle_client_maximized(c,ev->button-1);
          if(wasFront==0 && Is_Whole_Screen_Mode()==1 )
            Wait_OOSX_To_Front();
        }
      }
      /* Release occured on iconise button */ 
      else if(mouse_mode==MOUSE_ICON || mouse_mode==MOUSE_ICON_OUT) {
        redraw_client_buttons(c); 
        if(get_pointer_win(c->frame)==c->icon)
          toggle_minimise_client(c,0);
      }
      /* Release occured on close button */ 
      else if(mouse_mode==MOUSE_CLOSE || mouse_mode==MOUSE_CLOSE_OUT) {
        redraw_client_buttons(c); 
        if(get_pointer_win(c->frame)==c->close) 
          close_client(c);
      }
      /* If we are resizing */
      else if(mouse_mode==MOUSE_RESIZE) 
      { 
        /* Only allow resizing if not shaded */ 
        if((get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)==0) {
          dx=(ev->x_root-FX-LW-RW)+mouse_x; 
          dy=(ev->y_root-FY-TH-BH)+mouse_y; 
//#ifdef DEBUG
//          printf("Mapping window & Setting w,h in release: %d,%d\n",dx,dy);
//#endif
          // make sure configure event is sent
//          XMapWindow(dpy, c->window); 
          mouse_mode=MOUSE_NORMAL;
          c->is_e_m_r &= NOT_RESIZING;
          //if( dx!=0 || dy!=0 )
          resize_client(c, dx, dy, 2);
          if(c->has_frame)
            redraw_client_title(c);	// make sure title is redrawn (so it doesn't have size on it any more)
        }
      }
      /* If moving window */ 
      else if(mouse_mode==MOUSE_MOVE) 
      { 
        dx=orgmove_x-ev->x_root;
        dy=orgmove_y-ev->y_root;
        if((ev->time-last_button1_time<DBL_CLICK) && (last_button1_client==c) && dx<6 && dy<6 && dx>-6 && dy>-6) {
          last_button1_time=ev->time;
          last_button1_client=c;
#ifdef DEBUG
          printf("Got fast button up for MOUSE_MOVE...\n");
#endif
        } else {
#ifdef DEBUG
          printf("Got slow button up for MOUSE_MOVE...\n");
#endif
          dx=ev->x_root-mouse_x; 
          dy=ev->y_root-mouse_y; 
          mouse_x=ev->x_root; 
          mouse_y=ev->y_root;
          mouse_mode=MOUSE_NORMAL;
          c->is_e_m_r &= NOT_MOVING;
          move_client(c, c->x + dx, c->y + dy, 0);
//          redraw_client_frame(c);	// make sure it is redrawn translucent
        }
      }
      else if(ev->time-last_button1_time<DBL_CLICK && (last_button1_client==c))
      {
        if(get_pointer_win(c->frame)==c->title) {
#ifdef DEBUG
          printf("Got fast button up in titlebar (not MOUSE_MOVE)...\n");
#endif
          toggle_client_shaded(c);
          //if(c->has_frame) redraw_client_frame(c);	// make sure it is redrawn translucent
        } else
#ifdef DEBUG
          printf("Got fast button up (not MOUSE_MOVE, nor title)...\n");
#endif
        last_button1_time=0; 
        last_button1_client=NULL;
      }
//      if(mouse_mode==MOUSE_RESIZE)
//        resize_client(c, c->width, c->height, 0); 
    }

  } 

  //pointer_grabbing(last_pointer_grabbed);
  XUngrabPointer(dpy, CurrentTime); 
//  XAllowEvents(dpy, SyncPointer, CurrentTime);
  mouse_mode=MOUSE_NORMAL;
//  if(lastfocus_client==NULL)
//    XGrabPointer(dpy, check_win, True, ButtonPressMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, resize_curs, CurrentTime);
//  else
//    XGrabPointer(dpy, lastfocus_client->frame, True, ButtonPressMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, resize_curs, CurrentTime);
  XAllowEvents(dpy, SyncPointer, CurrentTime);
  
  got_button=False;
} 

/*
 * Attempt to sort out a problem that sometimes appears, where the X11 cursor looks like
 * it has frozen (i.e. been grabbed) after a menu item has been chosen that did not work?
 * Don't know why it does this - but this attempt does not fix it either.
 * UPDATE: turns out it's OSX's mouse tracking that's problematic, not XDarwin...
 */
 
void force_ungrab()
{
  XUngrabPointer(dpy, CurrentTime); 
  mouse_mode=MOUSE_NORMAL;
//  if(lastfocus_client==NULL)
//    XGrabPointer(dpy, check_win, True, ButtonPressMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, resize_curs, CurrentTime);
//  else
//    XGrabPointer(dpy, lastfocus_client->frame, True, ButtonPressMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, resize_curs, CurrentTime);
  XAllowEvents(dpy, SyncPointer, CurrentTime);
  
  got_button=False;
}

/* 
 * 
 * Handle motion notify events 
 * 
 */ 

void event_motion_notify(XMotionEvent *ev) 
{ 
  Client *c=NULL; 
  Client *win=NULL;
  int height, width, dx=0, dy=0;
  Window w1, w2; 
  int wx, wy, mask, x, y; 
//  int now_x, now_y; 
  XEvent e, *ep=NULL;
  XMotionEvent ee;

//  sprintf(lastsub,"TRACE1: event_motion_notify (win,time,last: %#lx,%#lx,%#lx)\n",ev->window,ev->time,lastfocus); 
//  subtrace();
//#ifdef SHOW_TRACE1
//  printf("TRACE1: event_motion_notify (win,time,last: %#lx,%#lx,%#lx)\n",ev->window,ev->time,lastfocus); 
//#else
#ifdef SHOW_EVENTS
  printf("%s X11 event: motion (win,time,last: %#lx,%#lx,%#lx)\n",PROGNAME,ev->window,ev->time,lastfocus); 
#else
  if(gVerboseConsole)
    printf("%s X11 event: motion (win,time,last: %#lx,%#lx,%#lx)\n",PROGNAME,ev->window,ev->time,lastfocus); 
#endif
//#endif 
 
  check_ptrwin=1;
  if(replay_motion!=0) return;
  
  no_grabbing_until = 0;

  // hack an XEvent pointer to the XMotionEvent
  ep=(XEvent *) &ee;
  
  /* If there are other motion events on this window, use the latest one! (well, maybe not...) */
  if(XCheckTypedWindowEvent(dpy, ev->window, MotionNotify,&e)) 
  { 
// Why not actually use the latest motion event!?
/*    if(e.time>ev->time)
      *ev=e;
#ifdef DEBUG 
    printf("Got another motion event: %d\n",e.time);
#endif
    while(XCheckTypedWindowEvent(dpy, ev->window, MotionNotify,(XEvent *) &e)) {
// Why not actually use the latest motion event!?
      if(e.time>ev->time)
        *ev=e;
//      dx++;
#ifdef DEBUG 
      printf("Got more motion events: %d\n",e.time);
#endif
    }
hmmmm... probably because it doesn't work, for some strange reason...??? */
    *ep=e;
    while(XCheckTypedWindowEvent(dpy, ev->window, MotionNotify,&e)) 
      *ep=e;
//    XPutBackEvent(dpy,ep);
    replay_motion=0;
    *ev=ee;
//    return; 
  }
  
/*
  if(!got_button) return;
*/
  if(gX11InFront==0)
    win=NULL;
  else
    win=lastfocus_client;
//    win=get_focus_client();

  c=get_client_from_window(ev->window, FRAME_OR_COVER); 
  if(c!=NULL) 
  { 
    /* If there is a release event on this window, ignore this motion */ 
    if(XCheckTypedWindowEvent(dpy, ev->window, ButtonRelease,&e)) 
    { 
#ifdef DEBUG
      printf("Got button release in motion (%d)\n",mouse_mode);
#endif
      XPutBackEvent(dpy,&e); 
      replay_motion=0;
      return; 
    }
    /* If resizing */ 
    if(mouse_mode==MOUSE_RESIZE) 
    { 
      /* Don't allow resizing if maximized or shaded 
      if(get_gnome_hint(c->window, WIN_STATE) & (WIN_STATE_MAXIMIZED|WIN_STATE_SHADED)) */ 
      if(get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)
        return;
 
      // don't want to get swamped by these events, so only allow so many per second
      // (this also helps spurious motion events which occur when really only clicking...)
//      if(ev->time-last_motion_time<RESIZE_MOTION_GAP) {
//#ifdef DEBUG
//        printf("Not time yet for motion (MOUSE_RESIZE) -put it back (%d,%d)\n",ev->x_root,ev->y_root);
//#endif
//        replay_motion_event=*(XEvent *)ev;
//        last_motion_time-=20; 
//        replay_motion=1;
//        return; }
      replay_motion=0;

      width=(ev->x_root-FX-LW-RW)+mouse_x; 
      height=(ev->y_root-FY-TH-BH)+mouse_y;
      if(c->width!=width || c->height!=height) {
#ifdef DEBUG
        printf("Setting w,h in motion: %d,%d  (%d,%d)\n",width,height,c->width,c->height);
#endif
        c->is_e_m_r |= IS_RESIZING;
//        resize_client(c, width, height, !config.continuous_resize); 
        resize_client(c, width, height, 0); 
//        c->is_e_m_r &= NOT_RESIZING;
        // reset the timer for motion events
        last_motion_time=ev->time; 
      }
    } 
 
    /* If moving window */ 
    else if(mouse_mode==MOUSE_MOVE) 
    { 
      // don't want to get swamped by these events, so only allow so many per second
      // (this also helps spurious motion events which occur when really only clicking...)
//      if(ev->time-last_motion_time<MOVE_MOTION_GAP) {
//#ifdef DEBUG
//        printf("Not time yet for motion (MOUSE_MOVE) -put it back (%d,%d)\n",ev->x_root,ev->y_root);
//#endif
//        replay_motion_event=*(XEvent *)ev;
//        replay_motion=1;
//        last_motion_time-=20; 
//        return; }

      dx=ev->x_root-mouse_x; 
      dy=ev->y_root-mouse_y; 
      mouse_x=ev->x_root; 
      mouse_y=ev->y_root; 
      if(dx!=0 || dy!=0) {
        // reset the timer for motion events
        last_motion_time=ev->time; 
        c->is_e_m_r |= IS_MOVING;
        move_client(c, c->x + dx, c->y + dy, 0);
        c->is_e_m_r &= NOT_MOVING;
      }
      replay_motion=0;
    } 
    else if(mouse_mode==MOUSE_ZOOM)
    // must check for enter/leave of window widgets
    {
      if(get_pointer_win(c->frame)!=c->zoom) {
        mouse_mode=MOUSE_ZOOM_OUT;
        if(c==win)
          c->zoom_state=BUTTON_OVER;
        if(c!=win)
          c->zoom_state=BUTTON_OVERBACK;  
        redraw_client_buttons(c); }
    }   
    else if(mouse_mode==MOUSE_ZOOM_OUT)
    {     
       if(get_pointer_win(c->frame)==c->zoom) {
         mouse_mode=MOUSE_ZOOM;
         if(c==win)
           c->zoom_state=BUTTON_PRESSED;
         if(c!=win)
           c->zoom_state=BUTTON_PRESSEDBACK;
         redraw_client_buttons(c); }
    } 
    else if(mouse_mode==MOUSE_ICON)
    {
      if(get_pointer_win(c->frame)!=c->icon) {
        mouse_mode=MOUSE_ICON_OUT;
        if(c==win)
          c->icon_state=BUTTON_OVER;
        if(c!=win)
          c->icon_state=BUTTON_OVERBACK;
        redraw_client_buttons(c); }
    }
    else if(mouse_mode==MOUSE_ICON_OUT)
    {
       if(get_pointer_win(c->frame)==c->icon) {
         mouse_mode=MOUSE_ICON;
         if(c==win)
           c->icon_state=BUTTON_PRESSED;
         if(c!=win)
           c->icon_state=BUTTON_PRESSEDBACK;
         redraw_client_buttons(c); }
    }
    else if(mouse_mode==MOUSE_CLOSE)
    {
      if(get_pointer_win(c->frame)!=c->close) {
        mouse_mode=MOUSE_CLOSE_OUT;
        if(c==win)
          c->close_state=BUTTON_OVER;
        if(c!=win)
          c->close_state=BUTTON_OVERBACK;
        redraw_client_buttons(c); } 
    } 
    else if(mouse_mode==MOUSE_CLOSE_OUT)
    {
       if(get_pointer_win(c->frame)==c->close) {
         mouse_mode=MOUSE_CLOSE;
         if(c==win)
           c->close_state=BUTTON_PRESSED;
         if(c!=win)
           c->close_state=BUTTON_PRESSEDBACK;
         redraw_client_buttons(c); }
    }
    XQueryPointer(dpy, root, &w1, &w2, &x, &y, &wx, &wy, &mask);
    if(can_send_hackcmds==0) Compare_Screens(real_x,real_y,x,y,2);
  } 
} 
 
/* 
 * 
 * Handle idle time -looks out for cursor over window widgets
 * 
 */ 

void event_idle(int moved)
{
  Client *c=NULL;
  Client *win=NULL;
  Client *top=NULL;
  Window ptrwin=0;
  int retval;
  long nowticks;
  
//  int now_x, now_y;

#ifdef DEBUG
#ifndef SHOW_IDLE
// Showing all of the debugging junk is a bit much during idle time (background mouse tracking),
// so this makes sure we don't see it unless we specifically ask for it by defining SHOW_IDLE
  return;
#endif
#endif

  c=last_client_entered_hack;
//  sprintf(lastsub,"TRACE1: event_idle (check_ptrwin: %d  last: %#lx %#lx)\n",check_ptrwin,FC_WINDOW,C_WINDOW);
//  subtrace();
//#ifdef SHOW_TRACE1
//  printf("TRACE1: event_idle (check_ptrwin: %d  last: %#lx %#lx)\n",check_ptrwin,FC_WINDOW,C_WINDOW);
//#endif

/*
This "replay_if_drag" stuff is totally inoperative, so forget it for now...

  // replaying for drag in background window does not work (for some reason the XAllowEvents with
  // replayPointer does not actually send the event back to the window...)
  if(replay_if_drag==1) {
    // Hmmm... CurrentTime is not actually the value of the current time,
    // so this timer is rubbish anyway...
    if(CurrentTime-replay_time>DBL_CLICK) {
      get_mouse_xy(&now_x,&now_y);
#ifdef SHOW_EVENTS
      printf("  check for drag: %d,%d  (%d,%d)\n",now_x,now_y,replay_x,replay_y);
      printf("  currenttime/replaytime: %#lx,%#lx\n",CurrentTime,replay_time);
#endif
      if(now_x>replay_x+2)
         replay_if_drag=2;
      else if(now_x<replay_x-2)
         replay_if_drag=2;
      else if(now_y>replay_y+2)
         replay_if_drag=2;
      else if(now_y<replay_y-2)
         replay_if_drag=2;
      if(replay_if_drag==2) {
#ifdef SHOW_EVENTS
        printf("  re-sending event for drag...\n");
#endif
        XPutBackEvent(dpy,&replay_button_event);
      }
    } else
#ifdef SHOW_EVENTS
      printf("  waiting for DBL_CLICK time...\n");
#endif
    return;
  }
*/

  //if(mouse_mode!=MOUSE_NORMAL) printf("GETTING OUT OF event_idle BECAUSE mouse_mode=%d\n",mouse_mode);
  if(mouse_mode!=MOUSE_NORMAL)
    return;

  nowticks=GetCarbonTickCount();
/*  XGetInputFocus(dpy, &focus_win, &dummy); */

/*  if(lastfocus_client!=NULL && last_gotbutton_time!=0) {
    retval=XGrabPointer(dpy, lastfocus_client->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, normal_curs, CurrentTime);
    if(retval==GrabSuccess) {
      XUngrabPointer(dpy, CurrentTime);
      if(GetCarbonTickCount()>last_gotbutton_time+10) last_gotbutton_time=0;
      if(inhibiting_info!=0) {
        printf("Got %d from XGrabPointer -uninhibiting info. %d %d %d\n",retval,GrabSuccess,AlreadyGrabbed,GrabFrozen);
        SendSimpleHackCommand(HACKCMD_INHIBITINFO,0);
      }
    } else if(inhibiting_info==0) {
      printf("Got %d from XGrabPointer -inhibiting info. %d %d %d\n",retval,GrabSuccess,AlreadyGrabbed,GrabFrozen);
      SendSimpleHackCommand(HACKCMD_INHIBITINFO,1);
    }
  }
  
  // don't want to do anything odd if the cursor has been grabbed by a client
  if(inhibiting_info!=0) return;*/

  if(delayed_enter_ticks!=0 && nowticks>delayed_enter_ticks+2)
    delayed_entered_client(delayed_enter_client,delayed_can_focus);

  if(no_grabbing_until!=0) {
    if(nowticks>no_grabbing_until) {
      if(gVerboseConsole)
        printf("TickCount reached %d - checking grab\n",no_grabbing_until);
      check_button_grab = 1;
      no_grabbing_until = 0;
    } else {
      if(gVerboseConsole)
        printf("Not yet reached %d (%d %d)\n",no_grabbing_until,nowticks,MyXServerGrabLevel());
      return;
    }
#ifdef DEBUG
  } else {
      printf("Checks: %d %d %d %d\n",MyXServerGrabLevel(),hacking_focus,check_ptrwin,pointer_grabbed);
#endif
  }
    
  grabstat_before_hide=1;
  //if(lastfocus_client!=NULL && check_button_grab) {
  if(check_button_grab && pointer_grabbed==0 && no_grabbing_until==0) {
//    MyIncreaseXServerGrab();
    if(lastfocus_client==NULL)
        ptrwin = check_win;
    else
        ptrwin = lastfocus_client->frame;
    retval=XGrabPointer(dpy, ptrwin, False, ButtonPressMask|ButtonReleaseMask, GrabModeSync, GrabModeSync, None, None, CurrentTime);
    if(gVerboseConsole)
      printf("%s info: Got %d from XGrabPointer (%d %d)\n",PROGNAME,retval,check_button_grab,hacking_focus);
    if(retval!=GrabSuccess) {
      //XUngrabPointer(dpy, CurrentTime);
      // try again, just in case it was me that had grabbed it!
      //retval=XGrabPointer(dpy, ptrwin, False, ButtonPressMask|ButtonReleaseMask, GrabModeSync, GrabModeSync, None, None, CurrentTime);
      //printf("Got %d from second XGrabPointer\n",retval);
      //if(retval!=GrabSuccess) {
//        MyDecreaseXServerGrab();
        check_button_grab=2;
        return;
      //}
    }
    ptrwin = 0;
    XUngrabPointer(dpy, CurrentTime);
    XAllowEvents(dpy, SyncPointer, CurrentTime);
    check_button_grab = 0;
    if(lastfocus_client) quick_set_focus_client(lastfocus_client);
//    MyDecreaseXServerGrab();
  }
  grabstat_before_hide=0;
/*    
  if(can_send_hackcmds==0) {

    if(check_ptrwin) {
      //check_ptrwin=0;
      ptrwin=get_pointer_win(root);
    } else
      ptrwin = lastenter;
      
//    MyDecreaseXServerGrab();

    if( ptrwin!=lastenter ) {
      if( ptrwin!=lastbadenter || lastenter!=0 ) {
#ifdef SHOW_IDLE
        printf("  discrepancy: %#lx %#lx\n",ptrwin,lastenter );
#endif
        lastenter=ptrwin;
        if(lastenter!=0) {
          c=get_client_from_window(lastenter, FRAME_OR_COVER_LE);
          if(c==NULL) {
#ifdef SHOW_IDLE
            printf("  failed to find client: %#lx %#lx\n",lastenter, root );
#endif
            lastenter=0;
            lastbadenter=ptrwin;
          }
        }
        if(c!=NULL && c->is_docked)
          return;
        else
          entered_client(c,True);
      }
    } else if(lastenter!=0)
      c=get_client_from_window(lastenter, FRAME_OR_COVER_LE);

  } else if(check_ptrwin==0) {
//    MyDecreaseXServerGrab();
    c=last_client_entered_hack;
    if(c==NULL) {
        if(lastenter!=0) {
            printf("%s warning: non-zero lastenter with NULL last_client_entered_hack: %#lx\n",PROGNAME,lastenter);
            lastenter = 0;
        }
    } else if(c->frame!=lastenter) {
        printf("%s warning: lastenter does not match with last_client_entered_hack: %#lx,%#lx\n",PROGNAME,lastenter,c->frame);
        lastenter = c->frame;
    }
    ptrwin=lastenter;
  } else {
    //check_ptrwin=0;
    ptrwin=get_pointer_win(root);
    if(gX11InFront!=0) {
        c=get_top_most_client(WIN_LAYER_NORMAL,False);
        if(c!=NULL)
            if(c->frame==ptrwin && c!=last_client_entered_hack)
                c->is_e_m_r |= IS_ENTERED;
    }
//    MyDecreaseXServerGrab();
    c=last_client_entered_hack;
    if(c==NULL) {
        lastenter=0;
        if(ptrwin!=0) {
            c=get_client_from_window(ptrwin,FRAME_OR_COVER);
            if(c!=NULL) {
                win=get_lowest_entered_client(1);
                if(win==NULL) {
                    c=NULL;
                } else if(win==c) {
                    entered_client(c,1);
                    lastenter=c->frame;
                    last_client_entered_hack=c;
                } else if((c->is_e_m_r & IS_ENTERED)==0) {
                    c=NULL;
                } else {
                    win=c;
                    entered_client(c,1);
                    lastenter=c->frame;
                    last_client_entered_hack=c;
                }
            }
        }
    } else {
        lastenter=c->frame;
        if((c->is_e_m_r & IS_ENTERED)==0 || ptrwin!=lastenter) {
            c=get_client_from_window(ptrwin,FRAME_OR_COVER);
            if(c==NULL) {
                entered_client(NULL,1);
                lastenter=0;
                last_client_entered_hack=NULL;
            } else {
                win=get_lowest_entered_client(1);
                if(win==NULL) {
                    c=NULL;
                    entered_client(NULL,1);
                    lastenter=0;
                    last_client_entered_hack=NULL;
                } else if(win==c) {
                    entered_client(c,1);
                    lastenter=c->frame;
                    last_client_entered_hack=c;
                } else if((c->is_e_m_r & IS_ENTERED)==0) {
                    c=NULL;
                    entered_client(NULL,1);
                    lastenter=0;
                    last_client_entered_hack=NULL;
                } else {
                    win=c;
                    entered_client(c,1);
                    lastenter=c->frame;
                    last_client_entered_hack=c;
                }
            }
        }
    }
  }*/
  
  ptrwin=0;

  if( hacking_focus ) return;
  
  if(gX11InFront==0)
    win=NULL;
  else
    win=lastfocus_client;
//    win=get_focus_client();
  
  if(win!=NULL && check_top_client) {
    top=get_top_most_client(WIN_LAYER_NORMAL,False);
    if(top!=win) {
      if((win->auto_properties & AUTO_FOCUS)==0) {
        if(gVerboseConsole)
          printf("%s info: Noticed that top and focus clients are different...\n",PROGNAME);
        set_focus_client(top);
        win=lastfocus_client;
//        win=get_focus_client();
      }
    }
  }


  if(moved || check_top_client || check_ptrwin) {
  check_top_client = 0;
  check_ptrwin = 0;
  
  if(c!=NULL)
  {
    // if over different client, make sure window widgets of last client are reset
    if(lastc!=c) {
      if(lastc!=NULL) {
        if(lastc->is_icon==NOT_ICONBAR) {
#ifdef SHOW_IDLE
          printf("Resetting buttons of last win %#lx (%#lx)\n",lastc->window,C_WINDOW);
#endif
          lastc->zoom_state=BUTTON_UNPRESSED;
          lastc->icon_state=BUTTON_UNPRESSED;
          lastc->close_state=BUTTON_UNPRESSED;
          redraw_client_buttons(lastc);
        }
      }
    }
    lastc=c;
    ptrwin=get_pointer_win(c->frame);
    // check if we're over any of the window widgets
    if(ptrwin==c->zoom || ptrwin==c->icon || ptrwin==c->close) {
#ifdef SHOW_IDLE
       printf("Over buttons of current win %#lx (%#lx)\n",c->window,lastc->window);
#endif
       if(c->zoom_state==BUTTON_UNPRESSED || c->icon_state==BUTTON_UNPRESSED || c->close_state==BUTTON_UNPRESSED) {
         // set "x - +" glyphs in all of them appropriately
         if( win==c ) {
           c->zoom_state=BUTTON_OVER;
           c->icon_state=BUTTON_OVER;
           c->close_state=BUTTON_OVER;
         } else {
           c->zoom_state=BUTTON_OVERBACK;
           c->icon_state=BUTTON_OVERBACK;
           c->close_state=BUTTON_OVERBACK;
         }
#ifdef SHOW_IDLE
         printf("Lighting up buttons of current win %#lx (%#lx)\n",c->window,lastc->window);
#endif
         redraw_client_buttons(c);
       }
    } else if(c->zoom_state!=BUTTON_UNPRESSED || c->icon_state!=BUTTON_UNPRESSED || c->close_state!=BUTTON_UNPRESSED) {
       // one of the widgets had a glyph, so make sure none of them have now
       c->zoom_state=BUTTON_UNPRESSED;
       c->icon_state=BUTTON_UNPRESSED;
       c->close_state=BUTTON_UNPRESSED;
#ifdef SHOW_IDLE
       printf("Switching off buttons of current win %#lx (%#lx)\n",c->window,lastc->window);
#endif
       redraw_client_buttons(c);
       lastc=NULL;
    }
  }
  else if(lastc!=NULL)
  {
    // not over a window any more, so make sure last one has no glyphs in window widgets
    if(lastc->zoom_state!=BUTTON_UNPRESSED || lastc->icon_state!=BUTTON_UNPRESSED || lastc->close_state!=BUTTON_UNPRESSED) {
      lastc->zoom_state=BUTTON_UNPRESSED;
      lastc->icon_state=BUTTON_UNPRESSED;
      lastc->close_state=BUTTON_UNPRESSED;
#ifdef SHOW_IDLE
       printf("Switching off buttons of last client %#lx (%#lx)\n",lastc->window,C_WINDOW);
#endif
      redraw_client_buttons(lastc);
      lastc=NULL;
    }
  }
  
  }
}


/* 
 * 
 * Handle enter notify events - and also leave notify events
 * 
 */ 

void event_enter_notify(XCrossingEvent *ev, int entering) 
{ 
  Window ptrwin=0;
  Client *c=NULL; 
  XEvent e, ee; 

  check_ptrwin=1;
  check_top_client=1;
/*  if(can_send_hackcmds!=0) {
    if(XCheckTypedEvent(dpy, EnterNotify, &e)) {
      while(XCheckTypedEvent(dpy, EnterNotify, &e))
        ;
      entering=1;
      ev->window = ((XCrossingEvent *) &e)->window;
    }
    if(entering && ev->window) {
      if(gVerboseConsole)
        printf("%s X11 event: Entered window %#lx\n",PROGNAME,ev->window);
      c=get_client_from_window(ev->window,FRAME_OR_COVER);
      if(c!=NULL) {
        if(gX11InFront!=0 && (c==lastfocus_client || client_modal_firstclick(c)==lastfocus_client
                || (c->auto_properties & (AUTO_FOCUS|NONAUTO_FIRSTCLICK)))) {
          pointer_grabbing(0);
        } else {
          pointer_grabbing(c->frame);
        }
      }
/*    } else if(entering==0 && ev->window) {
      if(lastfocus_client)
        pointer_grabbing(lastfocus_client->frame);
      else
        pointer_grabbing(check_win);* /
    }
    return;
  }*/
    
  sprintf(lastsub,"TRACE1: event_enter_notify (win,lastenter,entering,lastfocus: %#lx,%#lx,%d,%#lx)\n",ev->window,lastenter,entering,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_enter_notify (win,lastenter,entering,lastfocus: %#lx,%#lx,%d,%#lx)\n",ev->window,lastenter,entering,lastfocus); 
#else
#ifdef SHOW_EVENTS
  printf("%s X11 event: enter/leave (win,lastenter,entering,lastfocus: %#lx,%#lx,%d,%#lx)\n",PROGNAME,ev->window,lastenter,entering,lastfocus); 
#else
  if(gVerboseConsole)
    printf("%s X11 event: enter/leave (win,lastenter,entering,lastfocus: %#lx,%#lx,%d,%#lx)\n",PROGNAME,ev->window,lastenter,entering,lastfocus); 
#endif
#endif
 
  MyIncreaseXServerGrab();
  /* If there are other enter/leave notify events, ignore this one */ 
  if(XCheckTypedEvent(dpy, EnterNotify|LeaveNotify, &e))
  {
    ee=e;
    while(XCheckTypedEvent(dpy, EnterNotify|LeaveNotify, &e))
      ee=e;
    XPutBackEvent(dpy, &ee);
#ifdef DEBUG
    printf("Found other events -ignoring enter\n");
#endif
    MyDecreaseXServerGrab();
    return;
  }
  
  if( entering && ev->window ) {
//    ptrwin=get_pointer_win(root);
//    if(ptrwin!=ev->window) {
//      printf("Got enter %#lx, but in window %#lx %#lx\n",ev->window,ptrwin,lastenter);
//      MyDecreaseXServerGrab();
//      return; }
    c=get_client_from_window(ev->window,FRAME_OR_COVER);
#ifdef DEBUG
    if(c==NULL)
       printf("  got NULL for frame - not changing lastenter\n");
#endif
    if(c!=NULL) {
        if(c->is_docked) {
            MyDecreaseXServerGrab();
            return;
        }
//       SendHackCommand(c,2,HACKCMD_HITCHECK,real_x+diffscrn_x,real_y+diffscrn_y,0,0,0);
       if(lastenter!=ev->window)
         entered_client(c,mouse_mode!=MOUSE_STOPPED);
#ifdef DEBUG
       else
         printf("  entered lastenter?!\n");
#endif
       lastenter=ev->window;
       last_client_entered_hack=c;
    }
  } else {
    if(ev->window==lastenter && lastenter) {
      c=get_client_from_window(ev->window,FRAME_OR_COVER);
      if(c!=NULL) {
        if(c->is_docked) {
            MyDecreaseXServerGrab();
            return;
        }
        ptrwin=get_pointer_win(c->frame);
        if(ptrwin) {
#ifdef DEBUG
          printf("  ignoring leave event: [%#lx], %#lx\n",lastenter,ptrwin);
#endif
          MyDecreaseXServerGrab();
          return;
        }
        if(c->is_icon!=NOT_ICONBAR) {
          ptrwin=get_pointer_win(root);
          if(ptrwin) {
            c=get_client_from_window(ptrwin,FRAME_OR_COVER);
            if(c!=NULL) {
              if(c->is_icon!=NOT_ICONBAR) {
                MyDecreaseXServerGrab();
                return; }
            }
          }
        }
      }
    }
    ptrwin=get_pointer_win(root);
    if(!ptrwin) {
      lastenter=0;
      entered_client(NULL,mouse_mode!=MOUSE_STOPPED);
      last_client_entered_hack=NULL;
#ifdef DEBUG
    } else {
      printf("Got leave, but cursor in window %#lx %#lx\n",ptrwin,lastenter);
#endif
    }
  }
  MyDecreaseXServerGrab();
#ifdef DEBUG
  printf("  set lastenter to [%#lx] %#lx, [%#lx] %#lx\n",lastenter,ev->window,c==NULL?0:c->frame,C_WINDOW);
#endif

} 
 
/* 
 * 
 * Handle focus in events 
 * 
 */ 

void event_focus_in(XFocusChangeEvent *ev) 
{ 
  Client *c=NULL; 
  XEvent e, ee; 
 
  sprintf(lastsub,"TRACE1: event_focus_in (win,last: %#lx,%#lx)\n",ev->window,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_focus_in (win,last: %#lx,%#lx)\n",ev->window,FC_WINDOW); 
#else
#ifdef SHOW_EVENTS
  printf("%s X11 event: focus in (win,last: %#lx,%#lx)\n",PROGNAME,ev->window,FC_WINDOW); 
#else
  if(gVerboseConsole)
    printf("%s X11 event: focus in (win,last: %#lx,%#lx)\n",PROGNAME,ev->window,FC_WINDOW); 
#endif
#endif
  
  check_ptrwin=1;
  return;
  
  /* If there are other focus in events, ignore this one */ 
  MyIncreaseXServerGrab();
/*  if(XCheckTypedEvent(dpy, FocusIn, &e)) 
  { 
#ifdef DEBUG
    printf("  ignored...\n");
#endif
    ee=e;
    while(XCheckTypedEvent(dpy, FocusIn, &e)) 
      ee=e;
    XPutBackEvent(dpy, &ee); 
  }
  else*/
  {
    c=get_client_from_window(ev->window, WINDOW); 
    if(c!=NULL) 
    { 
      if(c->is_icon==NOT_ICONBAR)
        if(c==lastfocus_client)
            quick_set_focus_client(c);
        else
            set_focus_client(c);
/*      if(c->is_icon==NOT_ICONBAR && ev->window!=lastfocus) {
        if(config.raise_on_focus || config.focus_style==FOCUS_CLICK) 
          raise_client(c); 
        lastfocus=ev->window;
        lastfocus_client=c;
        redraw_client_frame(c);
      }*/
    } 
    else {
      hacking_focus = 1;
//      lastfocus=0;
//      set_focus_client(get_top_most_client(WIN_LAYER_NORMAL,False)); 
    }
  }
  MyDecreaseXServerGrab();
} 
 
/* 
 * 
 * Handle focus out events 
 * 
 */ 

void event_focus_out(XFocusChangeEvent *ev) 
{ 
  Client *c=NULL; 
 
  sprintf(lastsub,"TRACE1: event_focus_out (win,last: %#lx,%#lx)\n",ev->window,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_focus_out (win,last: %#lx,%#lx)\n",ev->window,FC_WINDOW); 
#else
#ifdef SHOW_EVENTS
  printf("%s X11 event: focus out (win,last: %#lx,%#lx)\n",PROGNAME,ev->window,FC_WINDOW); 
#else
  if(gVerboseConsole)
    printf("%s X11 event: focus out (win,last: %#lx,%#lx)\n",PROGNAME,ev->window,FC_WINDOW); 
#endif
#endif 
 
  check_ptrwin=1;
//  c=get_client_from_window(ev->window, WINDOW); 
//  if(c!=NULL && c==lastfocus_client) {
//    hacking_focus = 1;
//    set_focus_client(NULL);
//    lastfocus_client=NULL;
//    redraw_client_frame(c);
//  }
} 
 
/* 
 * 
 * Handle destroy notify events 
 * 
 */  

void event_destroy_notify(XDestroyWindowEvent *ev) 
{ 
  Client *c; 
   
  sprintf(lastsub,"TRACE1: event_destroy_notify (win,last: %#lx,%#lx)\n",ev->window,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_destroy_notify (win,last: %#lx,%#lx)\n",ev->window,FC_WINDOW); 
#else
#ifdef SHOW_EVENTS
  printf("%s X11 event: destroy notify (win,last: %#lx,%#lx)\n",PROGNAME,ev->window,FC_WINDOW); 
#else
  if(gVerboseConsole)
    printf("%s X11 event: destroy notify (win,last: %#lx,%#lx)\n",PROGNAME,ev->window,FC_WINDOW); 
#endif
#endif 
 
//  if(ev->window==check_win) {
//    printf("check_win destroyed! Quitting...\n");
//    exit(0);
//  }
  
  check_ptrwin=1;
  check_top_client=1;
  if(ev->window==lastfocus)
    lastfocus=0;

  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) { 
    delete_client(c, 0); 
  } else if(lastenter!=0) {
    c=get_client_from_window(ev->window, FRAME);
    if(c!=NULL) {
#ifdef DEBUG
      printf("  checking lastenter: [%#lx]",c->frame);
#endif
      if(c->frame==lastenter) {
#ifdef DEBUG
         printf("  got lastenter in destroy: [%#lx]",lastenter);
#endif
         lastenter=0;
      }
    }
  }
} 
 
/* 
 * 
 * Handle unmap notify events 
 * 
 */ 

void event_unmap_notify(XUnmapEvent *ev) 
{ 
  Client *c=NULL; 
  XEvent e; 

  sprintf(lastsub,"TRACE1: event_unmap_notify (win,last: %#lx,%#lx)\n",ev->window,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_unmap_notify (win,last: %#lx,%#lx)\n",ev->window,FC_WINDOW); 
#else
#ifdef SHOW_EVENTS
  printf("%s X11 event: unmap notify (win,last: %#lx,%#lx)\n",PROGNAME,ev->window,FC_WINDOW); 
#else
  if(gVerboseConsole)
    printf("%s X11 event: unmap notify (win,last: %#lx,%#lx)\n",PROGNAME,ev->window,FC_WINDOW); 
#endif
#endif
 
  check_ptrwin=1;
  check_top_client=1;
  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) 
  { 
    if(c->ignore_unmap) 
      c->ignore_unmap--; 
    else 
    { 
      delete_client(c, 0);
        // check to see if there are any more unmaps to come...
      while (XCheckTypedEvent(dpy, UnmapNotify, &e)) 
      { 
        c=get_client_from_window(((XUnmapEvent *) &e)->window, WINDOW );
        // if windows have already vanished, the following get_top_most_client might cause BadWindow
        // errors, so go back and deal with this next unmapped window
        if(c!=NULL) {
          if(c->ignore_unmap)
            c->ignore_unmap--;
          else
            delete_client(c, 0);
        }
      }
      if(still_in_auto_client()==0)
        set_focus_client(get_top_most_client(WIN_LAYER_NORMAL,False)); 
    } 
  } 
} 
 
/* 
 * 
 * Handle map request events 
 * 
 */ 
void event_map_request(XMapRequestEvent *ev) 
{ 
  Client *c=NULL; 
 
//  MyIncreaseXServerGrab();
  sprintf(lastsub,"TRACE1: event_map_request (win,last: %#lx,%#lx)\n",ev->window,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_map_request (win,last: %#lx,%#lx)\n",ev->window,FC_WINDOW); 
#else
#ifdef SHOW_EVENTS
  printf("%s X11 event: map request (win,last: %#lx,%#lx)\n",PROGNAME,ev->window,FC_WINDOW); 
#else
  if(gVerboseConsole)
    printf("%s X11 event: map request (win,last: %#lx,%#lx)\n",PROGNAME,ev->window,FC_WINDOW); 
#endif
#endif 
 
  check_ptrwin=1;
  check_top_client=1;
  c=get_client_from_window(ev->window, WINDOW); 
  if(c==NULL) {
    create_new_client(ev->window, True); 
  } else { 
    raise_client(c);
    if(still_in_auto_client()==0)
      set_focus_client(c); 
  } 
//  MyDecreaseXServerGrab();
} 
 
/* 
 * 
 * Handle configure request events 
 * 
 */ 

void event_configure_request(XConfigureRequestEvent *ev) 
{ 
  Client *c=NULL; 
  XWindowChanges wc; 
  XEvent e, ee;
  XWindowAttributes wattr;
  int no_grab_focus=0;
 
  sprintf(lastsub,"TRACE1: event_configure_request (win,last: %#lx,%#lx)\n",ev->window,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_configure_request (win,last: %#lx,%#lx)\n",ev->window,FC_WINDOW); 
#else
#ifdef SHOW_EVENTS
  printf("%s X11 event: configure request (win,last: %#lx,%#lx)\n",PROGNAME,ev->window,FC_WINDOW);
//#else
//  if(gVerboseConsole)
//    printf("%s X11 event: configure request (win,last: %#lx,%#lx)\n",PROGNAME,ev->window,FC_WINDOW);
#endif
#endif 
 
  check_ptrwin=1;
  check_top_client=1;
  MyIncreaseXServerGrab();
  // If there are other configure requests on this window, ignore this one 
/*  if(XCheckTypedWindowEvent(dpy, ev->window, ConfigureRequest, &e)) 
  { 
    ee=e;
    while(XCheckTypedWindowEvent(dpy, ev->window, ConfigureRequest, &e)) {
      ee=e;
    }
    XPutBackEvent(dpy, &ee);
    MyDecreaseXServerGrab();
    return; 
  }*/ 
  // Actually, can't really do that because they may not just be changes to the size of the
  // window. They might also be changes to the position or to the stacking order.
  // We can't really afford to miss any of them!
  // It may be possible to check ahead for configures that match this one (i.e. look ahead for
  // another resize/move/stacking and ignore this one if one is found), but then we would still
  // need to make sure the *order* of requests is kept the same, so it would mean we would have
  // to put all events, except the current one and all matching ones *before* the last matching,
  // back onto the event queue in the same order we found them...
  
  wc.x=ev->x; 
  wc.y=ev->y; 
  wc.width=ev->width; 
  wc.height=ev->height; 
  wc.sibling=ev->above; 
  wc.stack_mode=ev->detail; 
  wc.border_width=0; 
 
//  XGrabServer(dpy);
  c=get_client_from_window(ev->window, WINDOW);
  if(c!=NULL && gVerboseConsole) {
    printf("%s info: got configure request: %#lx %d (",PROGNAME,ev->window,ev->value_mask);
    if(ev->value_mask & (CWX|CWY))
      printf(" @ %d , %d ",ev->x,ev->y);
    if(ev->value_mask & (CWWidth|CWHeight))
      printf(" %d x %d ",ev->width,ev->height);
    if(ev->value_mask & CWStackMode) {
      if(ev->detail==Above) printf(" Above ");
      if(ev->detail==Below) printf(" Below ");
      if(ev->value_mask & CWSibling) printf(" %#lx ",ev->above);
    }
    printf(")\n");
  }
  
// Check for cases when focus should not be grabbed from current focus window...

// In this case the client has tried to bring itself in front when it has just lost a grab
// (OpenOffice does this for some reason) - we just ignore the request
  if(c!=NULL && c==last_replay_client && did_force_ungrab!=0 && ev->value_mask==CWStackMode && ev->detail==Above) {
    if(GetCarbonTickCount()<did_force_ungrab+600) {
      printf("%s info: ignoring configure request (it noticed it lost the grab...)\n",PROGNAME);
      MyDecreaseXServerGrab();
      did_force_ungrab = 0;
      return;
    }
  }
// In this case, raise to just below current focus window.
  if(lastfocus_client!=NULL && wc.stack_mode==Above && gX11InFront!=0) {
    if((lastfocus_client->auto_properties & (RETAIN_FOCUS|AUTO_FOCUS))==RETAIN_FOCUS ||
        (c!=NULL && (c->auto_properties & NO_GRAB_FOCUS)==NO_GRAB_FOCUS)) {
      if(wc.sibling==0 || wc.sibling==lastfocus_client->frame || (ev->value_mask & CWSibling)==0) {
        wc.stack_mode=Below;
        wc.sibling=lastfocus_client->frame;
        ev->value_mask |= CWSibling;
      }
    }
  }
  if(c==NULL)
    XConfigureWindow(dpy, ev->window, ev->value_mask, &wc);
  else if(c->is_docked==0) { // do "will_undock_client" first if docked and coming in front
// thought I could get away without the need for the client to receive a configurenotify event
// here, but it appears some clients expect it, even when they change the window themselves.
// This is a bit of a hack to make sure windows that are being dragged (like with AbiWord toolbars)
// don't get fiddled in their positions as they are dragged - i.e. don't fiddle if button is down...
    if( Carbon_Button() )
      configure_client(c, ev->value_mask, &wc, c->is_icon!=NOT_ICONBAR, 1);
    else
      configure_client(c, ev->value_mask, &wc, c->is_icon!=NOT_ICONBAR, 100);
  }
  
// If a client is trying to come in front (not advised!) then bring X11 in front
  if((ev->value_mask & CWStackMode) && (ev->detail==Above)) {
    if(c!=NULL && c->is_docked!=0) {
      will_undock_client(c);
      configure_client(c, ev->value_mask, &wc, c->is_icon!=NOT_ICONBAR, 100); 
    } else if (gX11InFront==0) {
      if(c==NULL) {
        XGetWindowAttributes(dpy,ev->window,&wattr);
        if(wattr.map_state==IsViewable) {
          Bring_To_Front(0);
        } else if(gVerboseConsole)
          printf("%s info: didn't come to front since window not mapped...\n",PROGNAME);
      } else {
        if((c->auto_properties & NO_GRAB_FOCUS)==0) Bring_To_Front(0);
      }
    }
  }

  MyDecreaseXServerGrab();
} 
 
/* 
 *  
 * Handle property notify events 
 * 
 */ 

void event_property_notify(XPropertyEvent *ev) 
{ 
  Client *c=NULL; 
  long dummy;
 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_property_notify (win,last,atom: %#lx,%#lx,%d)\n",ev->window,FC_WINDOW,ev->atom);
#else
#ifdef SHOW_EVENTS
  printf("%s X11 event: property notify (win,last,atom: %#lx,%#lx,%d)\n",PROGNAME,ev->window,FC_WINDOW,ev->atom);
#else
  if(gVerboseConsole)
    printf("%s X11 event: property notify (win,last,atom: %#lx,%#lx,%d)\n",PROGNAME,ev->window,FC_WINDOW,ev->atom);
#endif
#endif 
 
  check_ptrwin=1;
  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) 
  { 
    switch(ev->atom) 
    { 
    case XA_WM_NAME: 
      if(c->name) XFree(c->name); 
      XFetchName(dpy, c->window, &c->name); 
      XClearWindow(dpy, c->title); 
      redraw_client_title(c); 
      SendHackTitleToDockCommand(c,0);
      break; 
    case XA_WM_NORMAL_HINTS: 
      XGetWMNormalHints(dpy, c->window, c->size, &dummy); 
      break; 
    } 
  } 
#ifdef WORKSPACE_SUPPORT 
  else 
  { 
    if(ev->atom==gnome[WIN_WORKSPACE_COUNT]) 
      update_workspace_count(); 
  } 
#endif /* WORKSPACE */ 

  sprintf(lastsub,"TRACE1: leaving event_property_notify...\n"); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: leaving event_property_notify...\n"); 
#endif 
 
} 
 
/* 
 * 
 * Handle client message events 
 * 
 */ 

void event_client_message(XClientMessageEvent *ev) 
{ 
  Client *c=NULL; 
 
  sprintf(lastsub,"TRACE1: event_client_message (win,last: %#lx,%#lx)\n",ev->window,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_client_message (win,last: %#lx,%#lx)\n",ev->window,FC_WINDOW); 
#else
#ifdef SHOW_EVENTS
  printf("%s X11 event: client message (win,last: %#lx,%#lx)\n",PROGNAME,ev->window,FC_WINDOW); 
#else
  if(gVerboseConsole)
    printf("%s X11 event: client message (win,last: %#lx,%#lx)\n",PROGNAME,ev->window,FC_WINDOW); 
#endif
#endif 
 
  check_ptrwin=1;
  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) 
  { 
    /* Client wants to minimise */ 
    if(ev->message_type==icccm[WM_CHANGE_STATE] && ev->format==32 && ev->data.l[0]==IconicState) 
      toggle_minimise_client(c,0);
//      minmax_client(c); 
 
    /* Client wants to toggle shaded state */ 
    if(ev->message_type==gnome[WIN_STATE] && ev->format==32 && (ev->data.l[0] & WIN_STATE_SHADED)) 
      toggle_client_shaded(c); 
       
#ifdef WORKSPACE_SUPPORT 
    /* Client wants to toggle sticky state */ 
    if(ev->message_type==gnome[WIN_STATE] && ev->format==32 && (ev->data.l[0] & WIN_STATE_STICKY)) 
      toggle_client_sticky(c); 
#endif /* WORKSPACE */ 
 
#ifdef LAYER_SUPPORT 
    /* Client wants to change layer */ 
    if(ev->message_type==gnome[WIN_LAYER] && ev->format==32) 
      set_client_layer(c, ev->data.l[0]); 
#endif /* LAYER */ 
  } 
  else if(ev->window==root)  
  { 
#ifdef WORKSPACE_SUPPORT 
    /* Change to new workspace */ 
    if(ev->message_type==gnome[WIN_WORKSPACE] && ev->format==32) 
      change_to_workspace(ev->data.l[0]); 
#endif /* WORKSPACE */ 
  } 
} 
 
#ifdef SHAPE_SUPPORT 
/* 
 * 
 * Handle shape events 
 * 
 */ 

void event_shape(XShapeEvent *ev) 
{ 
  Client *c=NULL; 
  XEvent e, ee; 
 
  sprintf(lastsub,"TRACE1: event_shape (win,last: %#lx,%#lx)\n",ev->window,FC_WINDOW); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_shape (win,last: %#lx,%#lx)\n",ev->window,FC_WINDOW); 
#else
#ifdef SHOW_EVENTS
  printf("%s X11 event: shape (win,last: %#lx,%#lx)\n",PROGNAME,ev->window,FC_WINDOW); 
#else
//  if(gVerboseConsole)
    printf("%s X11 event: shape (win,last: %#lx,%#lx)\n",PROGNAME,ev->window,FC_WINDOW); 
#endif
#endif 
 
  check_ptrwin=1;
  check_top_client=1;
  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) 
  { 
    c->is_shaped=1;
    /* If there are other shape events on this window, ignore this one */
//    XGrabServer(dpy);
    if(XCheckTypedWindowEvent(dpy, ev->window, shape_event, &e)) 
    {
      ee=e;
      while(XCheckTypedWindowEvent(dpy, ev->window, shape_event, &e))
        ee=e;
      XPutBackEvent(dpy, &ee); 
//      XUngrabServer(dpy);
      return; 
    } 
//    XUngrabServer(dpy);

//    SendHackCommand(c,1,HACKCMD_INHIBITFULL,1,0,0,0,0);
//    SendHackCommand(c,1,HACKCMD_INHIBITREFRESH,0,0,0,0,0);
    SendHackCommand(c,1,HACKCMD_INHIBITLEVEL,HACKARG_INHIBITDRAW,0,0,0,0);
    SendHackCommand(c,1,HACKCMD_DOSHADOW,-999,0,0,0,0);
    set_client_shape(c,0);
    SendHackCommand(c,1,HACKCMD_DOSHADOW,-998,0,0,0,0);
    // make sure it is redrawn translucent
    if((c->has_frame) || (c->is_icon!=NOT_ICONBAR)) redraw_client_frame(c);
//    XSync(dpy,False);
    if((c->is_e_m_r & IS_RESIZING)==0 || mouse_mode==MOUSE_NORMAL)
        SendHackCommand(c,2,HACKCMD_INHIBITLEVEL,HACKARG_INHIBITNONE,1,0,0,0);
    else
        SendHackCommand(c,2,HACKCMD_INHIBITLEVEL,HACKARG_INHIBITNONE,1,0,0,0);
/*    if((c->is_e_m_r & IS_RESIZING)==0 || mouse_mode==MOUSE_NORMAL)
        SendHackCommand(c,1,HACKCMD_INHIBITFULL,2,0,0,0,0);
    else
        SendHackCommand(c,1,HACKCMD_INHIBITREFRESH,3,0,0,0,0);*/
//    if(c==get_top_most_client(WIN_LAYER_NORMAL,True)) {
//      printf("Raising after shape\n");
//      raise_client(c); }
  } 
} 
#endif /* SHAPE */ 
 
/* 
 * 
 * Main event loop 
 * 
 */ 

void do_event_loop(int sleepus,int trackMouse) 
{ 
int gotev=0, sleeporg=0, cursorpos_gap=4, rcvlen;
char buf[100];
int *v1, *v2, *v3, *v4;
long last_get_cursorpos_time=0;
XEvent ev;
Client *c;

//  sprintf(lastsub,"TRACE1: do_event_loop (sleep,moved: %d,%d)\n",sleepus,trackMouse); 
//  subtrace();
//#ifdef SHOW_TRACE1 
//  printf("TRACE1: do_event_loop (sleep,moved: %d,%d)\n",sleepus,trackMouse); 
//#endif 
    sleeporg=sleepus;

//    got_replay_pointer=1;
    
    Start_Time_Check(60);	// only want to stay here for 1 second at most
    
    Restart_XDarwin();
    
/*    if( trackMouse==-999 ) {
      gotev = XPending(dpy);
      Stop_XDarwin();
      usleep(sleepus);
      return;
    }*/
    
/*XGetInputFocus(dpy, &lastenter, &dummy);*/
 
    if(gX11InFront==0)
      cursorpos_gap = 8;
    else
      cursorpos_gap = 4;
        
    if(can_send_hackcmds!=0) {
      if(xdarwin_socket) {
        int gotsomething=0;
        while(1) {
        rcvlen = check_listen_get_bytes(xdarwin_socket,buf);
        if(rcvlen==0) {
            printf("Got zero from XDarwin...\n");
            ClearUpSockets();
            exit(1);
        } else if(rcvlen==-2) {
            printf("XDarwin ended session.\n");
            ClearUpSockets();
            exit(1);
        } else if(rcvlen==-1) {
            //if(gotsomething==0)
            //    return;	// nothing in receive buffer
            //SendSimpleHackCommand(HACKCMD_INHIBITINFO,0);
            break;
        } else if(rcvlen<0) {
            printf("Got an error from check_listen_get_bytes: %d\n",rcvlen);
            exit(1);
        } else if(rcvlen==16) {
            gotsomething=1;
            v1 = (int *) &buf[0];
            v2 = (int *) &buf[4];
            v3 = (int *) &buf[8];
            v4 = (int *) &buf[12];
            //printf("%s info: GOT INFO FROM XDARWIN: %d %d %d %d\n",PROGNAME,*v1,*v2,*v3,*v4);
            if(*v1==-999) {
                // Got check_win's hackID - probably means screens have changed, so check it out
                c = NULL;
                update_xinerama_screens();
            } else
                c = get_client_from_hackid(*v1);
            if(c!=NULL) {
                switch(*v2) {
                    case 30000:	// window moved through external agent
                        Cocoa_To_Carbon(v3,v4);
                        //if(gVerboseConsole)
                            printf("After Cocoa_To_Carbon: %d %d",*v3,*v4);
                        Carbon_To_X11(v3,v4);
                        //if(gVerboseConsole)
                            printf("  After Carbon_To_X11: %d %d\n",*v3,*v4);
                        *v4 += TH;
                        *v3 += LW;
                        if(c->isoffscreen==0) {
                            int orgx=c->icon_x;
                            int orgy=c->icon_y;
                            if(c->is_icon==NOT_ICONBAR) {
                                orgx = c->x;
                                orgy = c->y;
                            }
                            if( *v3!=orgx || *v4!=orgy ) {
                                if(mouse_mode!=MOUSE_NORMAL)
                                    move_client(c,orgx,orgy,1);	// correct the position
                                else if(Carbon_Button())
                                    move_client(c,*v3,*v4,1);	// update the X server
                                else
                                    move_client(c,*v3,*v4,100);	// update the X server
                                if(c->is_icon!=NOT_ICONBAR)	// but we must move iconbars back into place
                                    move_client(c,orgx,orgy,1);
                                else if(c->osx==-10000 && c->osy==-10000)
                                    move_client(c,orgx,orgy,1);
                            }
                        } else if(c->is_icon==NOT_ICONBAR) {
                            if(gVerboseConsole)
                                printf("HackID %d : Hiding pos changed from %d %d to %d %d\n",
                                                                c->hack_id,c->osx,c->osy,*v3,*v4);
                            c->osx = *v3;
                            c->osy = *v4;
                        } else if( *v3!=c->icon_x || *v4!=c->icon_y )
                            move_client(c,c->icon_x,c->icon_y,1);
                        break;
                    case 30001:	// window resized through external agent
                        if(c->is_icon==NOT_ICONBAR) {
                            *v3-=LW+RW;
                            *v4-=BH+TH;
                            if( c->width!=*v3 || c->height!=*v4 ) resize_client(c,c->width,c->height,0);
                        } else if( *v3!=c->icon_w || *v4!=c->icon_h )
                            resize_client(c,c->icon_w,c->icon_h,0);
                        break;
                    case 30002:	// window changed order (only understand raised and lowered at the moment)
                        if(*v3==1 && *v4==0) {
                            if(c->isoffscreen==1) {
                                check_ptrwin=1;
                                check_top_client=1;
                                c->isoffscreen=0;
                                //c->ignore_unmap=0;
                                //XMapWindow(dpy,c->frame);
                                move_client(c,c->osx,c->osy,1);
                                if(c->osxwinfo.winref)
                                    ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
                            }
                            //if(c->is_icon==NOT_ICONBAR) {
                                if(gVerboseConsole)
                                    printf("%s info: bringing window %#lx (hackid: %d) in front\n",
                                                                        PROGNAME,c->window,c->hack_id);
                                if(c->osx==-10000 && c->osy==-10000) c->osx=-20000;
                                if(c->is_icon!=NOT_ICONBAR) {
                                    minmax_client(c);
                                    c->osx=-10000;
                                    c->osy=-10000;
                                } else if(c->osxwinfo.winref)
                                    change_focus(c->osxwinfo.winref);
                                else
                                    raise_client(c);
                            //}
                        } else if( *v3==-9999 && *v4==-9999 ) {
                            if(c->is_icon==NOT_ICONBAR) {
                                if(gVerboseConsole)
                                    printf("%s info: sending window %#lx (hackid: %d) to back\n",
                                                        PROGNAME,c->window,c->hack_id);
                                lower_client(c);
                            }
                        } else
                            printf("%s warning (HackID %d): UNKNOWN WINDOW ORDER INFO: %d %d %d %d\n",
                                                    PROGNAME,c->hack_id,*v1,*v2,*v3,*v4);
                        break;
                    case 30003:	// window appeared/disappeared
                        if(*v3==0 && c->isoffscreen==0) {
                            check_ptrwin=1;
                            check_top_client=1;
                            if(gGotCVDHidden==0) {
                                gGotCVDHidden=1;
                                Setup_Options_Menu(0);
                            }
                            if(c->is_icon==NOT_ICONBAR) {
                                c->osx = c->x;
                                c->osy = c->y;
                            } else {
                                c->osx = c->icon_x;
                                c->osy = c->icon_y;
                            }
                            //if(gVerboseConsole)
                                printf("HackID %d : Storing pos before hiding: %d %d\n",
                                                                c->hack_id,c->osx,c->osy);
                            move_client(c,-FW-10,-FH-10,0);
                            c->isoffscreen=1;
                            if(c->osxwinfo.winref)
                                ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
                            //c->ignore_unmap++;
                            //XUnmapWindow(dpy,c->frame);
                        } else if(*v3==1 && c->isoffscreen==1) {
                            check_ptrwin=1;
                            check_top_client=1;
                            c->isoffscreen=0;
                            //if(gVerboseConsole)
                                printf("HackID %d : Restoring hiding pos: %d %d\n",
                                                            c->hack_id,c->osx,c->osy);
                            //c->ignore_unmap=0;
                            //XMapWindow(dpy,c->frame);
                            // For some reason, this shading hack is needed to get the windows to
                            // reappear when VirtualDesktop brings them back on-screen
                            //toggle_client_shaded(c);
                            move_client(c,c->osx,c->osy,1);
                            //toggle_client_shaded(c);
                            //set_client_shape(c,0);
                            if(c->osxwinfo.winref)
                                ChangedOSXWinState(&(c->osxwinfo),c->name,c->clientgroup,c->isoffscreen);
                        } else
                            printf("HackID %d : Got odd combination: %d %d %d\n",
                                                    c->hack_id,c->isoffscreen,*v3,*v4);
                        break;
                    default:
                        printf("%s warning: GOT UNKNOWN INFO FROM XDARWIN: %d %d %d %d\n",
                                                                    PROGNAME,*v1,*v2,*v3,*v4);
                        break;
                }
            }
        } else if(rcvlen!=8) {
            gotsomething=1;
            printf("Got strange length from XDarwin: %d (%s)\n",rcvlen,buf);
        } else {
            gotsomething=1;
            v1 = (int *) &buf[0];
            v2 = (int *) &buf[4];
            //printf("OroborOSX info: Received from XDarwin: %d %d\n",*v1,*v2);
            if(*v2>=30000) {
                switch(*v2) {
                    case 99999:
                        if(*v1==99999) {
                            // XDarwin wants to quit, so OroborOSX asks instead...
#ifdef SHOW_EVENTS
                            printf("%s info: got quit message from XDarwin\n",PROGNAME);
#else
                            if(gVerboseConsole)
                                printf("%s info: got quit message from XDarwin\n",PROGNAME);
#endif
                            if( Ask_Quit()==0 ) {
                                XMoveWindow(dpy, check_win, -23451, -99);	// tell XDarwin not to quit
                            } else {
                                quit();	// this will tell XDarwin to quit, so we don't need to tell it here
                                exit(0);
                            }
                        } else if(*v1==1) {
                            ;
//                    		printf("%s warning: got message from XDarwin to check events\n",PROGNAME);
                        } else {
                            printf("%s warning: got unknown message from XDarwin: %d\n",PROGNAME,*v1);
                        }
                        break;
                    case 30000:
                        if(*v1==1) {
                            keepInFrontUntil = -1;
                            if(gVerboseConsole)
                                printf("%s info: Entering command key shortcut mode...\n",PROGNAME);
                            SetCmdKeyCursor(1);
                        }
                        break;
                    default:
                        printf("***** GOT INFO FROM XDARWIN: %d %d\n",*v1,*v2);
                        break;
                }
            } else {
                // for now, send it through to the horrific hack that was used for v0.8 preview...
                ev.xbutton.button=-5;
                ev.xbutton.x_root=*v1;
                ev.xbutton.y_root=*v2;
                event_button_release(&ev.xbutton);
            }
          }
        }
      }
      if(client_list!=NULL) {	// no point tracking if there are no clients!
        long current_ticks = GetCarbonTickCount();
        if(check_clients_onscreen!=0 && current_ticks>check_clients_onscreen) {
          Client *c=NULL;
          int i=0;
          check_clients_onscreen = 0;
          c = client_list;
          // go through clients, making sure they are on screen
          for(i=0;i<client_count;i++) {
            if(gVerboseConsole) printf("%s info: checking client %#lx after screen change\n",PROGNAME,c);
            check_client_ok(c);
            c = c->next;
          }
          // may need to change iconbar positions - just 'compress' them into the appropriate place
          compress_iconbars();
        }
        if(current_ticks>next_get_cursorpos_time) {
          if(real_x!=lastx || real_y!=lasty) {
            XMoveWindow(dpy,check_win,HACKCMD_MOUSEMOVED_X,HACKCMD_MOUSEMOVED_Y);
            XMoveWindow(dpy,check_win,-1000,-1000);
            last_get_cursorpos_time=GetCarbonTickCount();
            next_get_cursorpos_time=last_get_cursorpos_time+cursorpos_gap;
            lastx = real_x;
            lasty = real_y;
            trackMouse = 1;
            //printf("Asked for cursor pos: %d %d\n",last_get_cursorpos_time,next_get_cursorpos_time);
          } else {
            next_get_cursorpos_time=GetCarbonTickCount()+4;
          }
        }
      }
    }
        
//    while(got_replay_pointer!=0) {
    while(XPending(dpy)!=0) 
//    while(Is_Time_Up()==0 && XPending(dpy)!=0 && got_replay_pointer==1) 
//    while(Is_Time_Up()==0 && XPending(dpy)!=0) 
//    while(Is_Time_Up()==0 && XCheckMaskEvent(dpy,0xffffffff,&ev)!=0) 
//    while(XCheckMaskEvent(dpy,0xffffffff,&ev)!=0) 
    { 
      
      XNextEvent(dpy, &ev); 

      gotev = 1;

      switch(ev.type) 
      { 
#ifdef KEYBOARD_SUPPORT 
      case KeyPress:
#ifdef SHOW_EVENTS
        printf("%s X11 event: Got a key press event\n",PROGNAME);
#else
        if(gVerboseConsole)
          printf("%s X11 event: Got a key press event\n",PROGNAME);
#endif
        event_key_press(&ev.xkey); 
        break; 
#endif /* KEYBOARD */ 
      case ButtonPress:
#ifdef SHOW_EVENTS
        printf("%s X11 event: Got a button down event\n",PROGNAME);
#endif
        event_button_press(&ev.xbutton); 
        cursorpos_gap=5;
        next_get_cursorpos_time=last_get_cursorpos_time+5;
        break; 
      case ButtonRelease: 
#ifdef SHOW_EVENTS
        printf("%s X11 event: Got a button up event\n",PROGNAME);
#endif
        event_button_release(&ev.xbutton); 
        break; 
      case MotionNotify: 
#ifdef SHOW_EVENTS
        printf("%s X11 event: Got a motion event\n",PROGNAME);
#endif
        event_motion_notify(&ev.xmotion); 
        next_get_cursorpos_time=last_get_cursorpos_time+2;
        cursorpos_gap=2;
        trackMouse=1;
        break; 
      case LeaveNotify:
#ifdef SHOW_EVENTS
        printf("%s X11 event: Got a leave event\n",PROGNAME);
#endif
        event_enter_notify(&ev.xcrossing,False);
        next_get_cursorpos_time=last_get_cursorpos_time+2;
        trackMouse=1;
        break;
      case EnterNotify: 
#ifdef SHOW_EVENTS
        printf("%s X11 event: Got an enter event\n",PROGNAME);
#endif
        event_enter_notify(&ev.xcrossing,True);
        next_get_cursorpos_time=last_get_cursorpos_time+2;
        trackMouse=1;
        break; 
      case FocusIn: 
#ifdef SHOW_EVENTS
        printf("%s X11 event: Got a focus in event\n",PROGNAME);
#endif
        event_focus_in(&ev.xfocus); 
        break; 
      case FocusOut: 
#ifdef SHOW_EVENTS
        printf("%s X11 event: Got a focus out event\n",PROGNAME);
#endif
        event_focus_out(&ev.xfocus); 
        break; 
      case DestroyNotify: 
#ifdef SHOW_EVENTS
        printf("%s X11 event: Got a destroy event\n",PROGNAME);
#endif
        event_destroy_notify(&ev.xdestroywindow); 
        break; 
      case UnmapNotify: 
#ifdef SHOW_EVENTS
        printf("%s X11 event: Got an unmap event\n",PROGNAME);
#endif
        event_unmap_notify(&ev.xunmap); 
        break; 
      case MapRequest: 
#ifdef SHOW_EVENTS
        printf("%s X11 event: Got a map request event\n",PROGNAME);
#endif
        event_map_request(&ev.xmaprequest); 
        break; 
      case ConfigureRequest: 
#ifdef SHOW_EVENTS
        printf("%s X11 event: Got a configure event\n",PROGNAME);
#endif
        event_configure_request(&ev.xconfigurerequest); 
        break; 
      case PropertyNotify: 
        event_property_notify(&ev.xproperty); 
        break; 
      case ClientMessage: 
#ifdef SHOW_EVENTS
        printf("%s X11 event: Got a client message event\n",PROGNAME);
#endif
        event_client_message(&ev.xclient); 
        break; 
      default:
        gotev = 0;
#ifdef SHAPE_SUPPORT 
        if(ev.type==shape_event) {
          if(shape) {
#ifdef SHOW_EVENTS
            printf("%s X11 event: Got a shape event\n",PROGNAME);
#else
            if(gVerboseConsole)
              printf("%s X11 event: Got a shape event\n",PROGNAME);
#endif // SHOW_EVENTS
            event_shape((XShapeEvent *)&ev); 
            gotev = 1;
          } else {
#ifdef SHOW_EVENTS
            printf("%s X11 event: Got a shape event -but not used...\n",PROGNAME);
#else
            if(gVerboseConsole)
              printf("%s X11 event: Got a shape event -but not used...\n",PROGNAME);
#endif // SHOW_EVENTS 
          }
        } else {
#ifdef SHOW_EVENTS
          if(ev.type!=22) printf("%s X11 event: Got an unknown event, type: %d\n",PROGNAME,ev.type);
#else
          if(gVerboseConsole && ev.type!=22) printf("%s X11 event: Got an unknown event, type: %d\n",PROGNAME,ev.type);
#endif // SHOW_EVENTS 
        }
#else // NO SHAPE 
#ifdef SHOW_EVENTS
        if(ev.type!=22) printf("%s X11 event: Got an unknown event, type: %d\n",PROGNAME,ev.type);
#else
        if(gVerboseConsole && ev.type!=22) printf("%s X11 event: Got an unknown event, type: %d\n",PROGNAME,ev.type);
#endif // SHOW_EVENTS 
#endif // SHAPE 
        break; 
      }
      
      if(overicon!=NULL)
        next_get_cursorpos_time=last_get_cursorpos_time+2;
//    if(gX11InFront==0)
      if(can_send_hackcmds!=0 && client_list!=NULL)	// no point tracking if there are no clients!
        if(GetCarbonTickCount()>next_get_cursorpos_time) {
          if(real_x!=lastx || real_y!=lasty) {
            XMoveWindow(dpy,check_win,HACKCMD_MOUSEMOVED_X,HACKCMD_MOUSEMOVED_Y);
            XMoveWindow(dpy,check_win,-1000,-1000);
            last_get_cursorpos_time=GetCarbonTickCount();
            next_get_cursorpos_time=last_get_cursorpos_time+cursorpos_gap;
            lastx = real_x;
            lasty = real_y;
            trackMouse = 1;
            //printf("Asked for cursor pos: %d %d\n",last_get_cursorpos_time,next_get_cursorpos_time);
          } else {
            next_get_cursorpos_time=GetCarbonTickCount()+4;
          }
        }
    }
    
    if(replay_motion!=0)
      XPutBackEvent(dpy, (XEvent *) &replay_motion_event); 
    replay_motion=0;

    /*if(current_resize_client!=NULL) {
        c=current_resize_client;
        if(c->resize_timer>0) {
            if(c->resize_timer>GetCarbonTickCount()) {
                unsigned int mask = (CWHeight|CWWidth|CWX|CWY);
                XWindowChanges wc;
                XConfigureEvent ce; 
                c->resize_timer=0;
                wc.x=LW;
                wc.y=TH;
                wc.width=c->width;
                wc.height=c->height;
                XConfigureWindow(dpy, c->window, mask, &wc);
                ce.type=ConfigureNotify;
                ce.display=dpy;
                ce.event=c->window; 
                ce.window=c->window; 
                ce.x=c->x; 
                ce.y=c->y; 
                ce.width=c->width; 
                ce.height=c->height;
                ce.border_width=0; 
//		  ce.above=c->frame; 
                ce.above=0; 
                ce.override_redirect=0; 
                XSendEvent(dpy, c->window, False, StructureNotifyMask, (XEvent *)&ce);
                //set_client_shape(c,0);
            }
        }
    }*/
    // only need to check mouse-over stuff if the mouse has moved
    if(trackMouse!=0 || replay_if_drag!=0)
      if(cursorpos_gap>2) event_idle(trackMouse);
//      if(got_replay_pointer<2) event_idle();
      
    gotev = 0;

    if(mouse_mode==MOUSE_STOPPED) mouse_mode=MOUSE_NORMAL;
    
  Stop_XDarwin();
    
#ifdef SHOW_WEIRD
  if(sleepus>100000 || sleepus<10) {
    printf("%s oddity: Got very strange value for sleepus: %d  (was %d)\n",PROGNAME,sleepus,sleeporg);
    show_last_fifteen(); }
#endif 
  if(sleepus>100000 || sleepus<10) sleepus=100000;

// Come back very quickly if our time is up, or if the gap till next cursor request is small,
// or if we are dragging (MOUSE_DRAG), resizing (MOUSE_RESIZE) or a window widget is pressed (etc.)
//  if(Is_Time_Up()!=0 || cursorpos_gap<3 || mouse_mode!=MOUSE_NORMAL)
  if(cursorpos_gap<3 || mouse_mode!=MOUSE_NORMAL)
    sleepus = 3000;

//  sprintf(lastsub,"TRACE1: sleeping in do_event_loop (sleep: %d, mouse: %d)\n",sleepus,mouse_mode); 
//  subtrace();
//#ifdef SHOW_TRACE1 
//  printf("TRACE1: sleeping in do_event_loop (sleep: %d, mouse: %d)\n",sleepus,mouse_mode); 
//#endif

//  if(got_replay_pointer==2) {
//    got_replay_pointer=3;
////    printf("Staying for two seconds...\n");
//    Start_Time_Check(120);	// only want to stay here for 10 ticks at most
//  } else if(got_replay_pointer==3) {
//    if(Is_Time_Up()!=0) {
//      got_replay_pointer=0;
////      printf("Getting out...\n");
//    }
//  } else if(got_replay_pointer==1) {
//    got_replay_pointer=0;
//  }
  
    // oroborus slept for 5 microseconds! -caused far too much CPU usage...
    // This can happily go to several thousand microseconds without noticeable problems
    // However, we do get out a.s.a.p. if there was already considerable time spent in here (10 ticks)
//  if(got_replay_pointer==2)
//    usleep(10000);
//  else if(Is_Time_Up()==0)

    usleep(sleepus);

  
//  }
  
//  sprintf(lastsub,"TRACE1: leaving do_event_loop...\n"); 
//  subtrace();
//#ifdef SHOW_TRACE1 
//  printf("TRACE1: leaving do_event_loop...\n"); 
//#endif 
 
//  Restart_XDarwin();
} 
