/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 

#include "oroborus.h" 
#include "hackcmds.h"
#include "iconbars.h"

#define USE_CHECK_WIN

void SendHackTitleToDockCommand(Client *c, int todock)
{
    unsigned char sent_title[120], *title=NULL;
    int i, l, p1, title_cmd, gotnotitle=0;
    Window win=0;
    
    if(c==NULL) return;
    if(can_send_hackcmds==0) return;

    if(c->is_icon!=NOT_ICONBAR) {
        MyIncreaseXServerGrab();
        XMoveWindow(dpy, check_win, HACKCMD_STARTX, HACKCMD_STARTY);	// start the hack command sequence
        XMoveWindow(dpy, check_win, HACKCMD_SETCURWIN, c->hack_id);	// Tell XDarwin which window we want
        XMoveWindow(dpy, check_win, HACKCMD_MINIWIN_TITLE, -2);		// Tell XDarwin to make title string nil
        XMoveWindow(dpy, check_win, HACKCMD_SETCURWIN,0);		// switch off the window of interest
        MyDecreaseXServerGrab();
        return;
    }
    
    if(c->name)
        title = c->name;
    else if(c->clientcmd)
        title = c->clientcmd;
    else if(c->clientgroup)
        title = c->clientgroup;
    else {
        title = malloc(5);
        sprintf(title,"?");
        gotnotitle=1;
        printf("**** Started off with no title! %d\n",c->hack_id);
    }
    
    if(title) {
        l = strlen(title);
#ifdef DEBUG
        printf("Got '%s' as title...\n",title);
#endif
        if(l>99) {
            for(i=0;i<70;i++)
                sent_title[i]=title[i];
            title[70] = 33; // dot character
            title[71] = 33;
            title[72] = 33;
            for(i=73;i<101;i++)
                sent_title[i]=title[i+l-100];
        } else {
            for(i=0;i<=l;i++)
                sent_title[i]=title[i];
        }

#ifdef DEBUG
        printf("Sending '%s' as title...\n",sent_title);
#endif
    
        MyIncreaseXServerGrab();
        XMoveWindow(dpy, check_win, HACKCMD_STARTX, HACKCMD_STARTY); // start the hack command sequence

        // tell XDarwin which window we're interested in
        if(todock==0)
            XMoveWindow(dpy, check_win, HACKCMD_SETCURWIN, c->hack_id);
        else // want the X11 window out of harm's way if it is to be docked
            XMoveWindow(dpy, c->frame, HACKCMD_SETCURWIN, c->hack_id);
        
        win = check_win;

        XMoveWindow(dpy, win, HACKCMD_MINIWIN_TITLE, -1); // Tell XDarwin to start a new title string
        
        // Send the title string
        title_cmd=HACKCMD_MINIWIN_TITLE1;
        for(i=0;i<strlen(sent_title);i++) {
            p1 = sent_title[i];
            XMoveWindow(dpy, win, title_cmd, p1);
            // alternate position, since X server will not send through command if values are the same
            title_cmd = HACKCMD_MINIWIN_TITLE-(title_cmd-HACKCMD_MINIWIN_TITLE);
        }
    
        XMoveWindow(dpy, win, HACKCMD_MINIWIN_TITLE, 0); // Commit the title string
    
    } else {
        MyIncreaseXServerGrab();
        XMoveWindow(dpy, check_win, HACKCMD_STARTX, HACKCMD_STARTY); // start the hack command sequence

        // tell XDarwin which window we're interested in
        if(todock==0)
            XMoveWindow(dpy, check_win, HACKCMD_SETCURWIN, c->hack_id);
        else // want the X11 window out of harm's way if it is to be docked
            XMoveWindow(dpy, c->frame, HACKCMD_SETCURWIN, c->hack_id);
        win = check_win;
        if(gotnotitle) free(title);
    }
    
    if(todock==0) {
        XMoveWindow(dpy,check_win,HACKCMD_SETCURWIN,0);	// switch off the window of interest
//        XMoveWindow(dpy, c->frame, FX, FY); // The X server thinks the window has moved, so put it back
    } else {
        c->is_docked=1;
        XMoveWindow(dpy, win, HACKCMD_DOCKWIN, 1); // Send it to the dock
        XMoveWindow(dpy,check_win,HACKCMD_SETCURWIN,0);	// switch off the window of interest
    }
    
    MyDecreaseXServerGrab();
}

void SendKeycodeModsToXDarwin(int *modlist)
{
    int key=0, side=0, keycode=0, i;
    if(can_send_hackcmds==0) return;
    
    MyIncreaseXServerGrab(); // don't want anything else interfering with the X server while this happens
    XMoveWindow(dpy, check_win, HACKCMD_STARTX, HACKCMD_STARTY);
    for(i=0;i<12;i++) {
        keycode = *modlist++;
        if( keycode>7 && keycode < 256 ) {
            int param = key + keycode*8;
            if( side ) param = -param;
            printf("sending mod info: %d %d %d (%d)\n",key,side,keycode,param);
            XMoveWindow(dpy, check_win, HACKCMD_CHANGEMODS, param);
        }
        if(side) {
            key++;
            side = 0;
        } else
            side = 1;
    }
    XMoveWindow(dpy, check_win, -1000, -1000);
    XSync(dpy,0);
    MyDecreaseXServerGrab(); // don't want anything else interfering with the X server while this happens
}

void SendModsToXDarwin(long new, long old)
{
    SendSimpleHackCommand(HACKCMD_SENDMODS,(int) new);
}

void SendKeyToXDarwin(int key, int down)
{
    if( down )
        SendSimpleHackCommand(HACKCMD_SENDKEYDOWN,key);
    else
        SendSimpleHackCommand(HACKCMD_SENDKEYUP,key);
}

void SendSimpleHackCommand(int cmd, int param)	// uses check_win to send a command not related to a client
{
    if(can_send_hackcmds==0) return;

    if(gVerboseConsole)
        if(cmd==HACKCMD_UNGRABPTR) printf("%s info: Asking for fake button events...\n",PROGNAME);
    MyIncreaseXServerGrab(); // don't want anything else interfering with the X server while this happens
    XMoveWindow(dpy, check_win, HACKCMD_STARTX, HACKCMD_STARTY);
    XMoveWindow(dpy, check_win, cmd, param);
    XMoveWindow(dpy, check_win, -1000, -1000);
    XSync(dpy,0);
    MyDecreaseXServerGrab(); // release the X server, if it was before
    
#ifdef DEBUG
    printf("Sent simple hack command: %d %d\n",cmd,param);
#endif
    if(cmd==HACKCMD_INHIBITINFO)
        inhibiting_info = param;
}

void SendHackCommand(Client *c, int nparam, int orgcmd, int p1, int p2, int p3, int p4, int p5)
{
    int params[5], i, cmd;
    Window	win=0;

    if(c==NULL) return;
    if(can_send_hackcmds==0) return;
    
    params[0]=p1;
    params[1]=p2;
    params[2]=p3;
    params[3]=p4;
    params[4]=p5;
    
    if(c->is_docked!=0 && orgcmd!=HACKCMD_DOCKWIN && orgcmd!=HACKCMD_TRANSLUCENT && orgcmd!=HACKCMD_DIMWIN && orgcmd!=HACKCMD_SETID && orgcmd!=HACKCMD_INHIBITLEVEL && orgcmd!=HACKCMD_UPDATEALL && orgcmd!=HACKCMD_SETDRAWALPHA) {
        printf("Entered SendHackCommand for docked client: %d %d\n",orgcmd,p1);
        return;
    }
    
    if(orgcmd<-23499 || orgcmd>-23400) {
        printf("%s warning: out of range hack-command (%d) for window %#lx (client %#lx)\n",
                                                                    PROGNAME,orgcmd,C_WINDOW,c);
        show_last_fifteen();
        return;
    }
    
#ifdef DEBUG
    printf("Sending to XDarwin for id %d: %d %d %d %d %d %d\n",c->hack_id,orgcmd,p1,p2,p3,p4,p5);
#endif
    if(orgcmd==HACKCMD_SETDRAWALPHA) {
        if(c->osxwinfo.drawalpha==p1)
            return;
        else
            c->osxwinfo.drawalpha=p1;
    } else if(orgcmd==HACKCMD_SETID) {
        if(c->hack_id==0) {
            printf("%s warning: got zero hack_id for client %#lx !!\n",PROGNAME,c);
            return;
        }
        params[0] = c->hack_id;
        nparam = 1;
        printf("Sending hack_id = %d to XDarwin...\n",params[0]);
    } else if(orgcmd==HACKCMD_INHIBITLEVEL && p1<200)
        c->inhibit_level=p1;
    
    win = check_win;
//    win = c->frame;
    
    MyIncreaseXServerGrab(); // don't want anything else interfering with the X server while this happens

    //XFlush(dpy);
    XSync(dpy,0);

    XMoveWindow(dpy, win, HACKCMD_STARTX, HACKCMD_STARTY); // start the hack command sequence

#ifdef USE_CHECK_WIN
//    if(c->hack_id>0 && orgcmd!=HACKCMD_SETDRAWALPHA && orgcmd!=HACKCMD_DIMWIN && orgcmd!=HACKCMD_SETID)
    if(c->hack_id>0 && orgcmd!=HACKCMD_SETID) {
//        printf("Told XDarwin to use window %#lx %d\n",c->window,c->hack_id);
        XMoveWindow(dpy, win, HACKCMD_SETCURWIN, c->hack_id);	// tell XDarwin which window we want
    } else
#endif
        win = c->frame;
    
    cmd = orgcmd;
//    if(orgcmd==-23456) win = check_win;
    
    // Send the command and the arguments
    for(i=0;i<nparam;i++) {
        XMoveWindow(dpy, win, cmd, params[i]);
        cmd--;
//        if(win==check_win && orgcmd==-23456) win=c->frame;
    }
    
#ifdef USE_CHECK_WIN
//    if(c->hack_id<=0 || orgcmd==HACKCMD_SETDRAWALPHA || orgcmd==HACKCMD_DIMWIN || orgcmd==HACKCMD_SETID) {
    if(c->hack_id<=0 || orgcmd==HACKCMD_SETID) {
#endif
        // Don't bother putting back if it has been docked, or if it is being dragged
        if((orgcmd!=HACKCMD_DOCKWIN || p1==0) && c->is_docked==0 && orgcmd!=HACKCMD_SETWINPOS)
            XMoveWindow(dpy, c->frame, FX, FY); // The X server thinks the window has moved, so put it back
        XMoveWindow(dpy,check_win,-1000,-1000);
#ifdef USE_CHECK_WIN
    } else {
//        if((orgcmd!=HACKCMD_DOCKWIN || p1==0) && c->is_docked==0 && orgcmd!=HACKCMD_SETWINPOS) {
//            XMoveWindow(dpy,c->frame,HACKCMD_SETCURWIN,0);	// switch off the window of interest
//            XMoveWindow(dpy,c->frame,FX,FY);
//        } else
//            printf("Finished using window %#lx %d\n",c->window,c->hack_id);
            XMoveWindow(dpy,check_win,HACKCMD_SETCURWIN,0);	// switch off the window of interest
//            if((orgcmd!=HACKCMD_DOCKWIN || p1==0) && c->is_docked==0 && orgcmd!=HACKCMD_SETWINPOS)
//                XMoveWindow(dpy,c->frame,FX,FY);
//        XMoveWindow(dpy,check_win,-1000,-1000);
    }
#endif
    XSync(dpy,0);
    //XFlush(dpy);
    
    MyDecreaseXServerGrab(); // release the X server, if it was before

    if(orgcmd==HACKCMD_INHIBITINFO) {
        inhibiting_info = p1;
    } else if(orgcmd==HACKCMD_DOCKWIN && p1<2) {
        c->is_docked = p1;
//        if(p1==0) {
//            XMapWindow(dpy,c->window);
//            set_client_shape(c,0); // need to force a complete redraw of the window if it's undocked
//        } else {
//            c->ignore_unmap++;
//            XUnmapWindow(dpy,c->window); // try to avoid getting any other events from this window
//        }
    } else if(orgcmd==HACKCMD_DIMWIN && p1<2 && p1!=c->is_dimmed) {
        // p1>=2 allows for temp drawing of frame without being dim
        c->is_dimmed = p1;
        //set_client_shape(c,0); // need to force a complete redraw of the window if it's dimmed/undimmed
        // also need to redraw the frame so it is translucent
        //if((c->has_frame) || (c->is_icon!=NOT_ICONBAR)) redraw_client_frame(c);
    }

}
