//
//  XServer.m
//
//  This class handles the interaction between the Cocoa front-end
//  and the Darwin X server thread.
//
//  Created by Andreas Monitzer on January 6, 2001.
//
/*
 * Copyright (c) 2001 Andreas Monitzer. All Rights Reserved.
 * Copyright (c) 2002-2003 Torrey T. Lyons. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE ABOVE LISTED COPYRIGHT HOLDER(S) BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name(s) of the above copyright
 * holders shall not be used in advertising or otherwise to promote the
 * sale, use or other dealings in this Software without prior written
 * authorization.
 */
/* $XFree86: xc/programs/Xserver/hw/darwin/quartz/XServer.m,v 1.9 2003/04/30 23:15:39 torrey Exp $ */

#include "communication.h"
#include "privatewinlist.h"
#include "quartzCommon.h"
#include "aquaCommon.h"
#include "insidecheck.h"

#define BOOL xBOOL
#include "X.h"
#include "Xproto.h"
#include "os.h"
#include "darwin.h"
#undef BOOL

#import "XServer.h"
#import "Preferences.h"

#include <unistd.h>
#include <stdio.h>
#include <sys/syslimits.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <pwd.h>
#include <signal.h>
#include <fcntl.h>

// For power management notifications
#import <mach/mach_port.h>
#import <mach/mach_interface.h>
#import <mach/mach_init.h>
#import <IOKit/pwr_mgt/IOPMLib.h>
#import <IOKit/IOMessage.h>
#include <IOKit/hidsystem/IOLLEvent.h>

#define ENQUEUE(xe)                                                         \
{                                                                           \
    char byte = 0;                                                          \
    DarwinEQEnqueue(xe);                                                    \
    /* signal there is an event ready to handle */                          \
    write(eventWriteFD, &byte, 1);                                          \
}

// Types of shells
enum {
    shell_Unknown,
    shell_Bourne,
    shell_C
};

typedef struct {
    char *name;
    int type;
} shellList_t;

static shellList_t const shellList[] = {
    { "csh",    shell_C },		// standard C shell
    { "tcsh",   shell_C },		// ... needs no introduction
    { "sh",     shell_Bourne },		// standard Bourne shell
    { "zsh",    shell_Bourne },		// Z shell
    { "bash",   shell_Bourne },		// GNU Bourne again shell
    { NULL,	shell_Unknown }
};

extern int argcGlobal;
extern char **argvGlobal;
extern char **envpGlobal;
extern int main(int argc, char *argv[], char *envp[]);
extern void HideMenuBar(void);
extern void ShowMenuBar(void);
extern void QuartzReallySetCursor();
static void childDone(int sig);
static void powerDidChange(void *x, io_service_t y, natural_t messageType,
                           void *messageArgument);

static NSPort *signalPort;
static NSPort *returnPort;
static NSPortMessage *signalMessage;
static pid_t clientPID;
static XServer *oneXServer;
static NSRect aquaMenuBarBox;
static io_connect_t root_port;
static unsigned int old_flags=0;
static unsigned int allmods=NX_ALPHASHIFTMASK|NX_SHIFTMASK|NX_CONTROLMASK|NX_ALTERNATEMASK|NX_COMMANDMASK|NX_SECONDARYFNMASK;
static unsigned int allbutfnmods=NX_ALPHASHIFTMASK|NX_SHIFTMASK|NX_CONTROLMASK|NX_ALTERNATEMASK|NX_COMMANDMASK;
int showing_prefs=0, mainThreadBusy=0;
int xserver_curpos_x=-10000, xserver_curpos_y=-10000;


@implementation XServer

- (id)init
{
    self = [super init];
    oneXServer = self;

    serverState = server_NotStarted;
    serverLock = [[NSRecursiveLock alloc] init];
    clientPID = 0;
    sendServerEvents = NO;
    serverVisible = NO;
    rootlessMenuBarVisible = YES;
    queueShowServer = YES;
    quartzServerQuitting = NO;
    mouseState = 0;
    eventWriteFD = quartzEventWriteFD;

    // set up a port to safely send messages to main thread from server thread
    signalPort = [[NSPort port] retain];
    returnPort = [[NSPort port] retain];
    signalMessage = [[NSPortMessage alloc] initWithSendPort:signalPort
                    receivePort:returnPort components:nil];

    // set up receiving end
    [signalPort setDelegate:self];
    [[NSRunLoop currentRunLoop] addPort:signalPort
                                forMode:NSDefaultRunLoopMode];
    [[NSRunLoop currentRunLoop] addPort:signalPort
                                forMode:NSModalPanelRunLoopMode];

    [[NSNotificationCenter defaultCenter] addObserver:self
    selector:@selector(CTSWindowDidHide:)
    name:@"CTSWindowDidHide" object:nil];
    
    [[NSNotificationCenter defaultCenter] addObserver:self
    selector:@selector(CTSWindowDidShow:)
    name:@"CTSWindowDidShow" object:nil];

    return self;
}

void *GetWinByNumber(int number)
{
    PrivateListRec *aRec;
    
    aRec = private_list_start;
    
    while(aRec) {
        if(aRec->rw) {
            AquaWindowRec *winRec = AQUA_WINREC(aRec->rw);
            if([winRec->window windowNumber]==number) {
                //if(verboseConsole)
                    //ErrorF("===== Got window with Number %d ====\n",number);
                return aRec->rw;
            }
            aRec = aRec->next;
        }
    }
    
    NSLog(@"Failed to get window with Number %d !!!!!!!",number);
    return NULL;
}

// Glue routine to get at it from rootlessAquaImp
void startFlushTimer()
{
    NSLog(@"Starting flush timer for the first time...");
    [oneXServer startTimer];
}

- (void)CTSWindowDidHide:(NSNotification *)notification
{
    int number=[(NSNumber *)[notification object] intValue];
    void *rw=GetWinByNumber(number);
    //NSLog(@"In CTSWindowDidHide with window number %d",number);
    
    if(rw!=NULL) {
        AquaWindowRec *winRec = AQUA_WINREC(rw);
        //if([winRec->window isKindOfClass:windowClass]) {
            if(verboseConsole)
                NSLog(@"Got WindowDidHide for HackID %d",winRec->hackID);
            [winRec->view sendSpecialEvent:-9999 which:30003 dx:0 dy:0];
        //} else
        //    NSLog(@"****** Got WindowDidHide for non-X11 window (%d)",number);
    } else
        NSLog(@"****** GOT BAD WINDOWDIDHIDE!!!! %d",number);
    
}

- (void)CTSWindowDidShow:(NSNotification *)notification
{
    int number=[(NSNumber *)[notification object] intValue];
    void *rw=GetWinByNumber(number);
    //NSLog(@"In CTSWindowDidShow with window number %d",number);
    
    if(rw!=NULL) {
        AquaWindowRec *winRec = AQUA_WINREC(rw);
        //if([winRec->window isKindOfClass:windowClass]) {
            if(verboseConsole)
                NSLog(@"Got WindowDidShow for HackID %d",winRec->hackID);
            [winRec->view sendSpecialEvent:-9999 which:30003 dx:1 dy:1];
        //} else
        //    NSLog(@"****** Got WindowDidShow for non-X11 window (%d)",number);
    } else
        NSLog(@"****** GOT BAD WINDOWDIDSHOW!!!! %d",number);
}

- (NSApplicationTerminateReply)
        applicationShouldTerminate:(NSApplication *)sender
{
    // Quit if the X server is not running
    if ([serverLock tryLock]) {
        quartzServerQuitting = YES;
        serverState = server_Done;
        if (clientPID != 0)
            kill(clientPID, SIGINT);
        return NSTerminateNow;
    }

    // Hide the X server and stop sending it events
    [self showServer:NO];
    sendServerEvents = NO;

    if (clientPID != 0 || !quartzStartClients) {
        int but;

        but = NSRunAlertPanel(NSLocalizedString(@"Quit X server?",@""),
                              NSLocalizedString(@"Quitting the X server will terminate any running X Window System programs.",@""),
                              NSLocalizedString(@"Quit",@""),
                              NSLocalizedString(@"Cancel",@""),
                              nil);

        switch (but) {
            case NSAlertDefaultReturn:		// quit
                break;
            case NSAlertAlternateReturn:	// cancel
                if (serverState == server_Running)
                    sendServerEvents = YES;
                return NSTerminateCancel;
        }
    }

    quartzServerQuitting = YES;
    if (clientPID != 0)
        kill(clientPID, SIGINT);

    // At this point the X server is either running or starting.
    if (serverState == server_Starting) {
        // Quit will be queued later when server is running
        return NSTerminateLater;
    } else if (serverState == server_Running) {
        [self quitServer];
    }

    return NSTerminateNow;
}

// Ensure that everything has quit cleanly
- (void)applicationWillTerminate:(NSNotification *)aNotification
{
    // Make sure the client process has finished
    if (clientPID != 0) {
        NSLog(@"Waiting on client process...");
        sleep(2);

        // If the client process hasn't finished yet, kill it off
        if (clientPID != 0) {
            int clientStatus;
            NSLog(@"Killing client process...");
            killpg(clientPID, SIGKILL);
            waitpid(clientPID, &clientStatus, 0);
        }
    }

    // Wait until the X server thread quits
    [serverLock lock];
}

// Tell the window manager that something has happened to the window
- (void)sendSpecialEvent:(int)hackid which:(int)which
{
    int buf[2], send_err;
    NSPoint hitCheckPt, aPt;
    
//    if(oroborosx_info_inhibit) return;
    if(oroborosx_socket==0) return;
    
    buf[0] = hackid;
    buf[1] = which;
    send_err = send_bytes_to_socket(oroborosx_socket,8,(char *) &buf[0]);
    
    if( send_err < 0 )
        NSLog(@"send_bytes_to_socket err: %d",send_err);
    else if(verboseConsole)
        NSLog(@"Sent: %d %d",buf[0],buf[1]);
}

- (void)check_modifiers:(unsigned int)new old:(unsigned int)old event:(xEvent *)xe
                                                            fromEvent:(NSEvent *)anEvent
{
    [self getMousePosition:xe fromEvent:anEvent];
    if( old != new ) {
        //if(verboseConsole)
            NSLog(@"Forcing modifiers change: %#lx %#lx",new,old);
        xe->u.u.type = kXDarwinUpdateModifiers;
        /*if(new==0x10000) {
            xe->u.clientMessage.u.l.longs0 = new + 0x20000;
        } else*/
            xe->u.clientMessage.u.l.longs0 = new;
        [self sendXEvent:xe];
    }
}

// returns YES when event was handled
- (BOOL)translateEvent:(NSEvent *)anEvent
{
    xEvent xe;
    static BOOL mouse1Pressed = NO;
    static BOOL gotCommandDown = NO;
    NSEventType type;
    unsigned int flags;

    if (!sendServerEvents) {
        return NO;
    }

    type  = [anEvent type];
    // mask out aeverything but the modifiers we are interested in
    if(type==NSFlagsChanged) 
        flags = [anEvent modifierFlags] & allmods;
    else
        flags = ([anEvent modifierFlags] & allbutfnmods)
              | (old_flags & NX_SECONDARYFNMASK);

    if (!quartzRootless) {
        // Check for switch keypress
        if ((type == NSKeyDown) && (![anEvent isARepeat]) &&
            ([anEvent keyCode] == [Preferences keyCode]))
        {
            unsigned int switchFlags = [Preferences modifiers];

            // Switch if all the switch modifiers are pressed, while none are
            // pressed that should not be, except for caps lock.
            if (((flags & switchFlags) == switchFlags) &&
                ((flags & ~(switchFlags | NSAlphaShiftKeyMask)) == 0))
            {
                [self toggle];
                return YES;
            }
        }

        if (!serverVisible)
            return NO;
    }

    memset(&xe, 0, sizeof(xe));

    switch (type) {
        case NSLeftMouseUp:
            if (quartzRootless && !mouse1Pressed) {
                // MouseUp after MouseDown in menu - ignore
                return NO;
            }
            [self check_modifiers:flags old:old_flags event:&xe fromEvent:anEvent];
            mouse1Pressed = NO;
            xe.u.u.type = ButtonRelease;
            xe.u.u.detail = 1;
            break;
        case NSLeftMouseDown:
            if (quartzRootless) {
                NSWindow *theWindow = [anEvent window];
                if ((quartzUsesNSWindow && !theWindow) ||
                    (!quartzUsesNSWindow && theWindow)) {
                    // Click in non X window - ignore
                    return NO;
                }
                if(quartzUsesNSWindow && theWindow) {
                    // Hack to make sure the view's mInside flag is on
                    //NSLog(@"Button 1 down: Making sure inside window...");
                    [[theWindow contentView] mouseEntered:anEvent];
                    /*xe.u.keyButtonPointer.rootX--;
                    xe.u.keyButtonPointer.rootY--;
                    xe.u.u.type = MotionNotify;
                    [self sendXEvent:&xe];
                    xe.u.keyButtonPointer.rootX++;
                    xe.u.keyButtonPointer.rootY++;*/
                }
            }
            //if( (flags & 0x10000) && (old_flags & 0x10000)==0 ) flags ^= 0x10000;
            [self check_modifiers:flags old:old_flags event:&xe fromEvent:anEvent];
            mouse1Pressed = YES;
            xe.u.u.type = ButtonPress;
            xe.u.u.detail = 1;
            break;
        case NSMouseMoved:
        case NSLeftMouseDragged:
        case NSRightMouseDragged:
        case NSOtherMouseDragged:
            [self check_modifiers:flags old:old_flags event:&xe fromEvent:anEvent];
            xe.u.u.type = MotionNotify;
            //xe.u.u.type = kXDarwinFakeMoveMouse;
            break;
        case NSSystemDefined:
        {
            long hwButtons = [anEvent data2];

            if (![anEvent subtype]==7)
                return NO; // we only use multibutton mouse events
            if (mouseState == hwButtons)
                return NO; // ignore double events
            mouseState = hwButtons;

            //if( (flags & 0x10000) && (old_flags & 0x10000)==0 ) flags ^= 0x10000;
            [self check_modifiers:flags old:old_flags event:&xe fromEvent:anEvent];
            xe.u.u.type = kXDarwinUpdateButtons;
            xe.u.clientMessage.u.l.longs0 = [anEvent data1];
            xe.u.clientMessage.u.l.longs1 =[anEvent data2];
            break;
        }
        case NSScrollWheel:
            [self check_modifiers:flags old:old_flags event:&xe fromEvent:anEvent];
            xe.u.u.type = kXDarwinScrollWheel;
            xe.u.clientMessage.u.s.shorts0 = [anEvent deltaY];
            break;
        case NSKeyDown:
        case NSKeyUp:
            // If the mouse is not on the valid X display area,
            // we don't send the X server key events.
            if (![self getMousePosition:&xe fromEvent:nil])
                return NO;
            // If we have the prefs window showing,
            // don't send key events to the X server
            if(showing_prefs && quartzRootless)
                return NO;
            [self check_modifiers:flags old:old_flags event:&xe fromEvent:nil];
            if (type == NSKeyDown)
                xe.u.u.type = KeyPress;
            else
                xe.u.u.type = KeyRelease;
            xe.u.u.detail = [anEvent keyCode];
            if(verboseConsole) NSLog(@"Got keycode %#lx",xe.u.u.detail);
            break;
        case NSFlagsChanged:
            if( flags != old_flags ) {
                if(verboseConsole)
                    NSLog(@"Noticed modifiers change: %#lx %#lx",flags,old_flags);
                [self getMousePosition:&xe fromEvent:nil];
                xe.u.u.type = kXDarwinUpdateModifiers;
                xe.u.clientMessage.u.l.longs0 = flags;
                if( gotCommandDown ) {
                // check for changes except caps lock or command
                    if( ((flags^old_flags) | (NX_ALPHASHIFTMASK|NX_COMMANDMASK)) !=
                                            (NX_ALPHASHIFTMASK|NX_COMMANDMASK) ) {
                        gotCommandDown = NO;
                        if(verboseConsole)
                            NSLog(@"Got another modifier before command-key up - cancelling");
                    }
                }
                //old_flags = flags;
            }
            break;
        case NSRightMouseDown:
            //if( (flags & 0x10000) && (old_flags & 0x10000)==0 ) flags ^= 0x10000;
            [self check_modifiers:flags old:old_flags event:&xe fromEvent:anEvent];
            xe.u.u.type = ButtonPress;
            xe.u.u.detail = 3;
            break;
        case NSRightMouseUp:
            //if( (flags & 0x10000) && (old_flags & 0x10000)==0 ) flags ^= 0x10000;
            [self check_modifiers:flags old:old_flags event:&xe fromEvent:anEvent];
            xe.u.u.type = ButtonRelease;
            xe.u.u.detail = 3;
            break;
        case NSOtherMouseDown:          // undocumented MouseDown
            //if( (flags & 0x10000) && (old_flags & 0x10000)==0 ) flags ^= 0x10000;
            [self check_modifiers:flags old:old_flags event:&xe fromEvent:anEvent];
            xe.u.u.type = ButtonPress;
            xe.u.u.detail = 2;
            break;
        case NSOtherMouseUp:            // undocumented MouseUp
            //if( (flags & 0x10000) && (old_flags & 0x10000)==0 ) flags ^= 0x10000;
            [self check_modifiers:flags old:old_flags event:&xe fromEvent:anEvent];
            xe.u.u.type = ButtonRelease;
            xe.u.u.detail = 2;
            break;
        default:
            if(verboseConsole) NSLog(@"Received unhandled event: %d",type);
            return NO;
            break;
    }
    
    if( gotCommandDown ) {
        if( type != NSFlagsChanged && type != NSMouseMoved ) {
            gotCommandDown = NO;
            if(verboseConsole)
                NSLog(@"Got another event before command-key up - cancelling");
        } else if( (old_flags & NX_COMMANDMASK) && (flags & NX_COMMANDMASK)==0 ) {
            gotCommandDown = NO;
            if(verboseConsole)
                NSLog(@"---- Switching into command-key-shortcut mode");
            [self sendSpecialEvent:1 which:30000];
        }
    } else if( (flags & NX_COMMANDMASK) && (old_flags & NX_COMMANDMASK)==0 ) {
        gotCommandDown = YES;
        if(verboseConsole) NSLog(@"Noticed command key go down...");
    }
    old_flags = flags;

    [self sendXEvent:&xe];

    // Rootless: Send first NSLeftMouseDown to Cocoa windows and views so
    // window ordering can be suppressed.
    // Don't pass further events - they (incorrectly?) bring the window
    // forward no matter what.
    if (quartzRootless  &&
        (type == NSLeftMouseDown || type == NSLeftMouseUp) &&
        [anEvent clickCount] == 1 && [anEvent window])
    {
        return NO;
    }

    return YES;
}

// Return mouse coordinates, inverting y coordinate.
// The coordinates are extracted from an event or the current mouse position.
// For rootless mode, the menu bar is treated as not part of the usable
// X display area and the cursor position is adjusted accordingly.
// Returns YES if the cursor is not in the menu bar.
- (BOOL)getMousePosition:(xEvent *)xe fromEvent:(NSEvent *)anEvent
{
    NSPoint pt;

    if (anEvent) {
        NSWindow *eventWindow = [anEvent window];

        if (eventWindow) {
            pt = [anEvent locationInWindow];
            pt.x += [eventWindow frame].origin.x;
            pt.y += [eventWindow frame].origin.y;
        } else {
            pt = [NSEvent mouseLocation];
        }
    } else {
        pt = [NSEvent mouseLocation];
    }

    xserver_curpos_x = (int)(pt.x);
    xserver_curpos_y = (int)(pt.y);
    xe->u.keyButtonPointer.rootX = xserver_curpos_x;

    /*if (quartzRootless && NSMouseInRect(pt, aquaMenuBarBox, NO)) {
        // mouse in menu bar - tell X11 that it's just below instead
        xe->u.keyButtonPointer.rootY = aquaMenuBarHeight;
        return NO;
    } else*/ {
        NSRect frame=[[[NSScreen screens] objectAtIndex:0] frame];
        xe->u.keyButtonPointer.rootY =// (int)frame.origin.y +
         (int)NSHeight(frame) - xserver_curpos_y;
        //NSLog(@"Cursor pos: %d %d (adjust by %d+%d: %d)",(int)pt.x,(int)pt.y,
        // (int)frame.origin.y,(int)NSHeight(frame), xe->u.keyButtonPointer.rootY);
        return YES;
    }
}

// Append a string to the given enviroment variable
+ (void)append:(NSString*)value toEnv:(NSString*)name
{
    setenv([name cString],
        [[[NSString stringWithCString:getenv([name cString])]
            stringByAppendingString:value] cString],1);
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
    // Block SIGPIPE
    // SIGPIPE repeatably killed the (rootless) server when closing a
    // dozen xterms in rapid succession. Those SIGPIPEs should have been
    // sent to the X server thread, which ignores them, but somehow they
    // ended up in this thread instead.
    {
        sigset_t set;
        sigemptyset(&set);
        sigaddset(&set, SIGPIPE);
        // pthread_sigmask not implemented yet
        // pthread_sigmask(SIG_BLOCK, &set, NULL);
        sigprocmask(SIG_BLOCK, &set, NULL);
    }

    if (quartzRootless == -1) {
        // The display mode was not set from the command line.
        // Show mode pick panel?
        if ([Preferences modeWindow]) {
            if ([Preferences rootless])
                [startRootlessButton setKeyEquivalent:@"\r"];
            else
                [startFullScreenButton setKeyEquivalent:@"\r"];
            [modeWindow makeKeyAndOrderFront:nil];
        } else {
            // Otherwise use default mode
            quartzRootless = [Preferences rootless];
            [self startX];
        }
    } else {
        [self startX];
    }
}

// Start the X server thread and the client process
- (void)startX
{
    NSDictionary *appDictionary;
    NSString *appVersion;

    [modeWindow close];

    // Calculate the height of the menu bar so rootless mode can avoid it
    if (quartzRootless) {
        aquaMenuBarHeight = NSHeight([[[NSScreen screens] objectAtIndex:0] frame]) -
                            NSMaxY([[[NSScreen screens] objectAtIndex:0] visibleFrame]) - 1;
        aquaMenuBarBox =
            NSMakeRect(0, NSMaxY([[[NSScreen screens] objectAtIndex:0] visibleFrame]) + 1,
                       NSWidth([[[NSScreen screens] objectAtIndex:0] frame]),
                       aquaMenuBarHeight);
    }

    // Write the XDarwin version to the console log
    appDictionary = [[NSBundle mainBundle] infoDictionary];
    appVersion = [appDictionary objectForKey:@"CFBundleShortVersionString"];
    if (appVersion)
        NSLog(@"\n%@", appVersion);
    else
        NSLog(@"No version");

    // Start the X server thread
    serverState = server_Starting;
    [NSThread detachNewThreadSelector:@selector(run) toTarget:self
              withObject:nil];

    [self startTimer];
    // Start the X clients if started from GUI
    if (quartzStartClients) {
        [self startXClients];
    }

    if (quartzRootless) {
        // There is no help window for rootless; just start
        [helpWindow close];
        helpWindow = nil;
    } else {
        IONotificationPortRef notify;
        io_object_t anIterator;

        // Register for system power notifications
        root_port = IORegisterForSystemPower(0, &notify, powerDidChange,
                                             &anIterator);
        if (root_port) {
            CFRunLoopAddSource([[NSRunLoop currentRunLoop] getCFRunLoop],
                               IONotificationPortGetRunLoopSource(notify),
                               kCFRunLoopDefaultMode);
        } else {
            NSLog(@"Failed to register for system power notifications.");
        }
        
        // Show the X switch window if not using dock icon switching
        if (![Preferences dockSwitch])
            [switchWindow orderFront:nil];

        if ([Preferences startupHelp]) {
            // display the full screen mode help
            [helpWindow makeKeyAndOrderFront:nil];
            queueShowServer = NO;
        } else {
            // start running full screen and make sure X is visible
            ShowMenuBar();
            [self closeHelpAndShow:nil];
        }
    }
}

// Finish starting the X server thread
// This includes anything that must be done after the X server is
// ready to process events.
- (void)finishStartX
{
    sendServerEvents = YES;
    serverState = server_Running;

    if (quartzRootless) {
        [self forceShowServer:[NSApp isActive]];
    } else {
        [self forceShowServer:queueShowServer];
    }

    if (quartzServerQuitting) {
        [self quitServer];
        [NSApp replyToApplicationShouldTerminate:YES];
    }
}

// Start the first X clients in a separate process
- (BOOL)startXClients
{
    struct passwd *passwdUser;
    NSString *shellPath, *NOdashShellName, *commandStr, *startXPath;
    NSMutableString *safeStartXPath;
    NSRange aRange;
    NSBundle *thisBundle;
    const char *shellPathStr, *newargv[3], *shellNameStr;
    int fd[2], outFD, length, shellType, i;
    char *mystr=NULL;
    
    // Register to catch the signal when the client processs finishes
    signal(SIGCHLD, childDone);

    // Get user's password database entry
    passwdUser = getpwuid(getuid());

    // Find the shell to use
    if ([Preferences useDefaultShell])
        shellPath = [NSString stringWithCString:passwdUser->pw_shell];
    else
        shellPath = [Preferences shellString];

//    dashShellName = [NSString stringWithFormat:@"-%@",
//                            [shellPath lastPathComponent]];
    NOdashShellName = [NSString stringWithFormat:@"%@",
                            [shellPath lastPathComponent]];
    shellPathStr = [shellPath cString];
    shellNameStr = [[shellPath lastPathComponent] cString];

    if (access(shellPathStr, X_OK)) {
        NSLog(@"Shell %s is not valid!", shellPathStr);
        return NO;
    }

    // Find the type of shell
    for (i = 0; shellList[i].name; i++) {
        if (!strcmp(shellNameStr, shellList[i].name))
            break;
    }
    shellType = shellList[i].type;

    newargv[0] = [NOdashShellName cString];
/*    if (shellType == shell_Bourne) {
        // Bourne shells need to be told they are interactive to make
        // sure they read all their initialization files.
        newargv[1] = "-i";
        newargv[2] = NULL;
    } else {
        newargv[1] = NULL;
    }*/
    // since the startXClients process does not do anything with OroborOSX,
    // there's no need to make this a login shell, nor interactive, nor to
    // run any of the user's init scripts...
    newargv[1] = "-f";
    newargv[2] = NULL;

    // Create a pipe to communicate with the X client process
    NSAssert(pipe(fd) == 0, @"Could not create new pipe.");

    // Open a file descriptor for writing to stdout and stderr
    outFD = open("/dev/console", O_WRONLY, 0);
    if (outFD == -1) {
        outFD = open("/dev/null", O_WRONLY, 0);
        NSAssert(outFD != -1, @"Could not open shell output.");
    }

    // Fork process to start X clients in user's default shell
    // Sadly we can't use NSTask because we need to start a login shell.
    // Login shells are started by passing "-" as the first character of
    // argument 0. NSTask forces argument 0 to be the shell's name.
    clientPID = vfork();
    if (clientPID == 0) {

        // Inside the new process:
        if (fd[0] != STDIN_FILENO) {
            dup2(fd[0], STDIN_FILENO);	// Take stdin from pipe
            close(fd[0]);
        }
        close(fd[1]);			// Close write end of pipe
        if (outFD == STDOUT_FILENO) {	// Setup stdout and stderr
            dup2(outFD, STDERR_FILENO);
        } else if (outFD == STDERR_FILENO) {
            dup2(outFD, STDOUT_FILENO);
        } else {
            dup2(outFD, STDERR_FILENO);
            dup2(outFD, STDOUT_FILENO);
            close(outFD);
        }

        // Setup environment
        setenv("HOME", passwdUser->pw_dir, 1);
        setenv("SHELL", shellPathStr, 1);
        setenv("LOGNAME", passwdUser->pw_name, 1);
        setenv("USER", passwdUser->pw_name, 1);
        setenv("TERM", "unknown", 1);
        if (chdir(passwdUser->pw_dir))	// Change to user's home dir
            NSLog(@"Could not change to user's home directory.");

        execv(shellPathStr, (char * const *)newargv);	// Start user's shell

        NSLog(@"Could not start X client process with errno = %i.", errno);
        _exit(127);
    }

    // In parent process:
    close(fd[0]);	// Close read end of pipe
    close(outFD);	// Close output file descriptor

    thisBundle = [NSBundle bundleForClass:[self class]];
    startXPath = [thisBundle pathForResource:@"startXClients" ofType:nil];
    if (!startXPath) {
        NSLog(@"Could not find startXClients in application bundle!");
        return NO;
    }

    // We will run the startXClients script with the path in single quotes
    // in case there are problematic characters in the path. We still have
    // to worry about there being single quotes in the path. So, replace
    // all instances of the ' character in startXPath with '\''.
    safeStartXPath = [NSMutableString stringWithString:startXPath];
    aRange = NSMakeRange(0, [safeStartXPath length]);
    while (aRange.length) {
        aRange = [safeStartXPath rangeOfString:@"'" options:0 range:aRange];
        if (!aRange.length)
            break;
        [safeStartXPath replaceCharactersInRange:aRange
                        withString:@"\'\\'\'"];
        aRange.location += 4;
        aRange.length = [safeStartXPath length] - aRange.location;
    }

    if ([Preferences addToPath]) {
        commandStr = [NSString stringWithFormat:@"'%@' :%d %@\n",
                        safeStartXPath, [Preferences display],
                        [Preferences addToPathString]];
        mystr = malloc(50+strlen([[Preferences addToPathString] cString]));
        sprintf(mystr,"/tmp/oroborosx-%s.dir/startXClients :%d %s\n", getenv("OROBOROSX_RANDOMID"), [Preferences display],[[Preferences addToPathString] cString]);
    } else {
        commandStr = [NSString stringWithFormat:@"'%@' :%d\n",
                        safeStartXPath, [Preferences display]];
        mystr = malloc(50);
        sprintf(mystr,"/tmp/oroborosx-%s.dir/startXClients :%d\n", getenv("OROBOROSX_RANDOMID"), [Preferences display]);
    }

//    length = [commandStr cStringLength];
//    if (write(fd[1], [commandStr cString], length) != length) {
    NSLog(@"Command: %s",mystr);
    length = strlen(mystr);
    if (write(fd[1], mystr, length) != length) {
        NSLog(@"Write to X client process failed.");
        free(mystr);
        return NO;
    }

    // Close the pipe so that shell will terminate when xinit quits
    close(fd[1]);
    free(mystr);

    return YES;
}

// Run the X server thread
- (void)run
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

    [serverLock lock];
    main(argcGlobal, argvGlobal, envpGlobal);
    serverVisible = NO;
    [pool release];
    [serverLock unlock];
    QuartzMessageMainThread(kQuartzServerDied, nil, 0);
}

// Full screen mode was picked in the mode pick panel
- (IBAction)startFullScreen:(id)sender
{
    [Preferences setModeWindow:[startupModeButton intValue]];
    [Preferences saveToDisk];
    quartzRootless = FALSE;
    [self startX];
}

- (void)flushWindows:(NSTimer *)aTimer
{
//    NSLog(@"Flushing windows (flushWindows)");
//    PrivateFlushWindows();
//    [self startTimer];
    if( need_window_flush > 5 ) {
        xEvent xe;
        xe.u.u.type = kXDarwinNeedWindowFlush;
        if(verboseConsole)
            NSLog(@"Sending kXDarwinNeedWindowFlush to X Server (%d)",need_window_flush);
        [self sendXEvent:&xe];
    } else if( need_window_flush > 0 )
        need_window_flush++;
}

- (void)startTimer
{
    NSTimer *aTimer;
    
    aTimer = [NSTimer scheduledTimerWithTimeInterval:0.02 target:self
            selector:@selector(flushWindows:) userInfo:nil repeats:YES];
    NSLog(@"Started new flush timer %#lx",aTimer);
//        [[NSRunLoop currentRunLoop] addTimer: aTimer forMode: NSEventTrackingRunLoopMode];
}

// Rootless mode was picked in the mode pick panel
- (IBAction)startRootless:(id)sender
{
    [Preferences setModeWindow:[startupModeButton intValue]];
    [Preferences saveToDisk];
    quartzRootless = TRUE;
    [self startX];
}

// Close the help splash screen and show the X server
- (IBAction)closeHelpAndShow:(id)sender
{
    if (sender) {
        int helpVal = [startupHelpButton intValue];
        [Preferences setStartupHelp:helpVal];
        [Preferences saveToDisk];
    }
    [helpWindow close];
    helpWindow = nil;

    [self forceShowServer:YES];
    [NSApp activateIgnoringOtherApps:YES];
}

// Show the X server when sent message from GUI
- (IBAction)showAction:(id)sender
{
    [self forceShowServer:YES];
}

// Show or hide the X server or menu bar in rootless mode
- (void)toggle
{
    if (quartzRootless) {
#if 0
        // FIXME: Remove or add option to not dodge menubar
        if (rootlessMenuBarVisible)
            HideMenuBar();
        else
            ShowMenuBar();
        rootlessMenuBarVisible = !rootlessMenuBarVisible;
#endif
    } else {
        [self showServer:!serverVisible];
    }
}

// Show or hide the X server on screen
- (void)showServer:(BOOL)show
{
    // Do not show or hide multiple times in a row
    if (serverVisible == show)
        return;

    if (sendServerEvents) {
        [self sendShowHide:show];
    } else if (serverState == server_Starting) {
        queueShowServer = show;
    }
}

// Show or hide the X server irregardless of the current state
- (void)forceShowServer:(BOOL)show
{
    serverVisible = !show;
    [self showServer:show];
}

// Tell the X server to show or hide itself.
// This ignores the current X server visible state.
//
// In full screen mode, the order we do things is important and must be
// preserved between the threads. X drawing operations have to be performed
// in the X server thread. It appears that we have the additional
// constraint that we must hide and show the menu bar in the main thread.
//
// To show the X server:
//   1. Capture the displays. (Main thread)
//   2. Hide the menu bar. (Must be in main thread)
//   3. Send event to X server thread to redraw X screen.
//   4. Redraw the X screen. (Must be in X server thread)
//
// To hide the X server:
//   1. Send event to X server thread to stop drawing.
//   2. Stop drawing to the X screen. (Must be in X server thread)
//   3. Message main thread that drawing is stopped.
//   4. If main thread still wants X server hidden:
//     a. Release the displays. (Main thread)
//     b. Unhide the menu bar. (Must be in main thread)
//   Otherwise we have already queued an event to start drawing again.
//
- (void)sendShowHide:(BOOL)show
{
    xEvent xe;

    [self getMousePosition:&xe fromEvent:nil];

    if (show) {
        if (!quartzRootless) {
            QuartzFSCapture();
            HideMenuBar();
        }
        xe.u.u.type = kXDarwinShow;
        [self sendXEvent:&xe];

        // the mouse location will have moved; track it
        xe.u.u.type = MotionNotify;
        [self sendXEvent:&xe];

        // inform the X server of the current modifier state
        //xe.u.u.type = kXDarwinUpdateModifiers;
        //xe.u.clientMessage.u.l.longs0 = [[NSApp currentEvent] modifierFlags];
        //[self sendXEvent:&xe];

        // put the pasteboard into the X cut buffer
        [self readPasteboard];
    } else {
        // put the X cut buffer on the pasteboard
        [self writePasteboard];

        xe.u.u.type = kXDarwinHide;
        [self sendXEvent:&xe];
    }

    serverVisible = show;
}

// Enable or disable rendering to the X screen
- (void)setRootClip:(BOOL)enable
{
    xEvent xe;

    xe.u.u.type = kXDarwinSetRootClip;
    xe.u.clientMessage.u.l.longs0 = enable;
    [self sendXEvent:&xe];
}

// Tell the X server to read from the pasteboard into the X cut buffer
- (void)readPasteboard
{
    xEvent xe;

    xe.u.u.type = kXDarwinReadPasteboard;
    [self sendXEvent:&xe];
}

// Tell the X server to write the X cut buffer into the pasteboard
- (void)writePasteboard
{
    xEvent xe;

    xe.u.u.type = kXDarwinWritePasteboard;
    [self sendXEvent:&xe];
}

- (void)quitServer
{
    xEvent xe;

    xe.u.u.type = kXDarwinQuit;
    [self sendXEvent:&xe];

    // Revert to the Mac OS X arrow cursor. The main thread sets the cursor
    // and it won't be responding to future requests to change it.
    [[NSCursor arrowCursor] set];

    serverState = server_Quitting;
}

- (void)sendXEvent:(xEvent *)xe
{
    // This field should be filled in for every event
    xe->u.keyButtonPointer.time = GetTimeInMillis();

#if 0
    // FIXME: Really?
    if (quartzRootless  &&
        (ev->type == NSLeftMouseDown  ||  ev->type == NSLeftMouseUp  ||
        (ev->type == NSSystemDefined && ev->data.compound.subType == 7)))
    {
        // mouse button event - send mouseMoved to this position too
        // X gets confused if it gets a click that isn't at the last
        // reported mouse position.
        xEvent moveEvent = *ev;
        xe.u.u.type = NSMouseMoved;
        [self sendXEvent:&moveEvent];
    }
#endif

    ENQUEUE(xe);
}

- (void)sendFakeButtonEvent:(int)which
{
    xEvent xe;
    
    [self getMousePosition:&xe fromEvent:nil];
    xe.u.u.type = MotionNotify;
    [self sendXEvent:&xe];

    if(verboseConsole)
        NSLog(@"telling X server to send fake button %d press and release",which); 

    // Only fake button 1 is meant to be on root window
    if( which == 1 ) simulateClickRoot = 1;
    xe.u.u.type = ButtonPress;
    xe.u.u.detail = which;
    [self sendXEvent:&xe];
    xe.u.u.type = ButtonRelease;
    xe.u.u.detail = which;
    [self sendXEvent:&xe];
}

// Handle messages from the X server thread
- (void)handlePortMessage:(NSPortMessage *)portMessage
{
    unsigned msg = [portMessage msgid];
    
    // obviously can't be busy any more...
    if( mainThreadBusy && msg!=kQuartzWinToResize ) {
        NSLog(@"Got mainThreadBusy in handlePortMessage!");
        mainThreadBusy = 0;
    }

    switch(msg) {
        case kQuartzServerHidden:
            // Make sure the X server wasn't queued to be shown again while
            // the hide was pending.
            if (!quartzRootless && !serverVisible) {
                QuartzFSRelease();
                ShowMenuBar();
            }
            break;

        case kQuartzServerStarted:
            [self finishStartX];
            break;

        case kQuartzServerDied:
            sendServerEvents = NO;
            serverState = server_Done;
            if (!quartzServerQuitting) {
                [NSApp terminate:nil];	// quit if we aren't already
            }
            break;

        case kQuartzCursorUpdate:
            // not used with the NSCursor implementation
            //QuartzReallySetCursor();
            break;

        case kQuartzPostEvent:
        {
            const xEvent *xe = [[[portMessage components] lastObject] bytes];
            ENQUEUE(xe);
        }
        break;

        case kQuartzWinToResize:
        {
            const int *privateid = [[[portMessage components] lastObject] bytes];
            if(verboseConsole)
                NSLog(@"Got kQuartzWinToResize %d message from server",*privateid); 
            QuartzHandleResize(*privateid);
        }
        break;
        
        case kQuartzWinToDock:
        {
            const int *hackid = [[[portMessage components] lastObject] bytes];
            if(verboseConsole)
                NSLog(@"Got kQuartzWinToDock %d message from server",*hackid); 
            QuartzHandleDocking(*hackid,1);
        }
        break;

        case kQuartzWinToUndock:
        {
            const int *hackid = [[[portMessage components] lastObject] bytes];
            if(verboseConsole)
                NSLog(@"Got kQuartzWinToUndock %d message from server",*hackid); 
            QuartzHandleDocking(*hackid,0);
        }
        break;
        
        case kQuartzUndockedWin:
        {
            const int *hackid = [[[portMessage components] lastObject] bytes];
            if(verboseConsole)
                NSLog(@"Got kQuartzUndockedWin %d message from server",*hackid); 
            [self sendSpecialEvent:*hackid which:100];
        }
        break;
        
        case kQuartzEnteredWin:
        {
            const int *hackid = [[[portMessage components] lastObject] bytes];
            if(verboseConsole)
                NSLog(@"Got kQuartzEnteredWin %d message from server",*hackid); 
            [self sendSpecialEvent:*hackid which:101];
        }
        break;
        
        case kQuartzExitedWin:
        {
            const int *hackid = [[[portMessage components] lastObject] bytes];
            if(verboseConsole)
                NSLog(@"Got kQuartzExitedWin %d message from server",*hackid); 
            [self sendSpecialEvent:*hackid which:102];
        }
        break;
        
        case kQuartzSendFakeButton1:
            if(verboseConsole)
                NSLog(@"Got kQuartzSendFakeButton1 message from server");
            [self sendFakeButtonEvent:1];
            break;
            
        case kQuartzSendFakeButton2:
            if(verboseConsole)
                NSLog(@"Got kQuartzSendFakeButton2 message from server");
            [self sendFakeButtonEvent:2];
            break;
            
        case kQuartzSendFakeButton3:
            if(verboseConsole)
                NSLog(@"Got kQuartzSendFakeButton3 message from server");
            [self sendFakeButtonEvent:3];
            break;
            
        case kQuartzWantMouseMoved:
        {
            xEvent xe;
            [self getMousePosition:&xe fromEvent:nil];
            xe.u.u.type = MotionNotify;
//            NSLog(@"Got kQuartzWantMouseMoved message from server");
            [self sendXEvent:&xe];
        }
        break;
        
        case kQuartzSendKeyPressEvent:
        {
            const int *keynum = [[[portMessage components] lastObject] bytes];
            xEvent xe;
            if(showing_prefs && quartzRootless)
                break;
            if(verboseConsole)
                NSLog(@"Got key down from OroborOSX: %d",*keynum);
            [self getMousePosition:&xe fromEvent:nil];
            xe.u.u.type = KeyPress;
            xe.u.u.detail = *keynum;
            [self sendXEvent:&xe];
            break;
        }
        
        case kQuartzSendKeyReleaseEvent:
        {
            const int *keynum = [[[portMessage components] lastObject] bytes];
            xEvent xe;
            if(showing_prefs && quartzRootless)
                break;
            if(verboseConsole)
                NSLog(@"Got key up from OroborOSX: %d",*keynum);
            [self getMousePosition:&xe fromEvent:nil];
            xe.u.u.type = KeyRelease;
            xe.u.u.detail = *keynum;
            [self sendXEvent:&xe];
            break;
        }
        
        case kQuartzCheckMods:
        {
            const int *flags = [[[portMessage components] lastObject] bytes];
            long orobmods = 0;
            xEvent xe;
            if(verboseConsole)
                NSLog(@"Got mods from OroborOSX: %d",*flags);
            if(*flags & 0x4) orobmods = 0x10000;	// caps lock
            if(*flags & 0x2) orobmods += 0x20000;	// shift
            if(*flags & 0x10) orobmods += 0x40000;	// control
            if(*flags & 0x8) orobmods += 0x80000;	// option
            if(*flags & 0x1) orobmods += 0x100000;	// command
            //if(*flags & 0x20) orobmods += ???; -- need to find what fn key is...
            [self getMousePosition:&xe fromEvent:nil];
            xe.u.u.type = kXDarwinUpdateModifiers;
            xe.u.clientMessage.u.l.longs0 = orobmods;
            old_flags = orobmods;
            [self sendXEvent:&xe];
            break;
        }
        
        default:
            NSLog(@"Unknown message from server thread.");
    }
}

// Quit the X server when the X client process finishes
- (void)clientProcessDone:(int)clientStatus
{
    if (WIFEXITED(clientStatus)) {
        int exitStatus = WEXITSTATUS(clientStatus);
        if (exitStatus != 0)
            NSLog(@"X client process terminated with status %i.", exitStatus);
    } else {
        NSLog(@"X client process terminated abnormally.");
    }

    if (!quartzServerQuitting) {
        [NSApp terminate:nil];	// quit if we aren't already
    }
}

// Called when the user clicks the application icon,
// but not when Cmd-Tab is used.
// Rootless: Don't switch until applicationWillBecomeActive.
- (BOOL)applicationShouldHandleReopen:(NSApplication *)theApplication
            hasVisibleWindows:(BOOL)flag
{
    if ([Preferences dockSwitch] && !quartzRootless) {
        [self showServer:YES];
    }
    return NO;
}

- (void)applicationWillResignActive:(NSNotification *)aNotification
{
    [self showServer:NO];
}

- (void)applicationWillBecomeActive:(NSNotification *)aNotification
{
    if (quartzRootless)
        [self showServer:YES];
}

@end


// Send a message to the main thread, which calls handlePortMessage in
// response. Must only be called from the X server thread because
// NSPort is not thread safe.
void QuartzMessageMainThread(unsigned msg, void *data, unsigned length)
{
    //if( need_window_flush>4 ) PrivateFlushWindows();
//    if (msg == kQuartzPostEvent) {
    if( mainThreadBusy && verboseConsole && msg!=kQuartzWinToResize )
        NSLog(@"WARNING: main thread busy - may hang... (%d %#lx %d)",
                                                            msg,data,length);
    /*if(verboseConsole)
        NSLog(@"Sending message %d (%#lx %d) to main thread",msg,data,length);*/
    if (data && length>0) {
        NSData *eventData = [NSData dataWithBytes:data length:length];
        NSArray *eventArray = [NSArray arrayWithObject:eventData];
        NSPortMessage *newMessage =
                [[NSPortMessage alloc]
                        initWithSendPort:signalPort
                        receivePort:returnPort components:eventArray];
        [newMessage setMsgid:msg];
        [newMessage sendBeforeDate:[NSDate distantPast]];
        [newMessage release];
    } else {
        [signalMessage setMsgid:msg];
        [signalMessage sendBeforeDate:[NSDate distantPast]];
    }
}

// Handle SIGCHLD signals
static void childDone(int sig)
{
    int clientStatus;

    if (clientPID == 0)
        return;

    // Make sure it was the client task that finished
    if (waitpid(clientPID, &clientStatus, WNOHANG) == clientPID) {
        if (WIFSTOPPED(clientStatus))
            return;
        clientPID = 0;
        [oneXServer clientProcessDone:clientStatus];
    }
}

static void powerDidChange(
    void *x,
    io_service_t y,
    natural_t messageType,
    void *messageArgument)
{
    switch (messageType) {
        case kIOMessageSystemWillSleep:
            if (!quartzRootless) {
                [oneXServer setRootClip:FALSE];
            }
            IOAllowPowerChange(root_port, (long)messageArgument);
            break;
        case kIOMessageCanSystemSleep:
            IOAllowPowerChange(root_port, (long)messageArgument);
            break;
        case kIOMessageSystemHasPoweredOn:
            if (!quartzRootless) {
                [oneXServer setRootClip:TRUE];
            }
            break;
    }
    
}
