/*
 * NSView subclass for Mac OS X rootless X server
 *
 * Copyright (c) 2001 Greg Parker. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE ABOVE LISTED COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name(s) of the above copyright
 * holders shall not be used in advertising or otherwise to promote the sale,
 * use or other dealings in this Software without prior written authorization.
 */
/* $XFree86: xc/programs/Xserver/hw/darwin/quartz/XView.h,v 1.2 2002/07/15 19:58:31 torrey Exp $ */

#include "fakeBoxRec.h"

typedef struct _AlphaListRec AlphaListRec;  
struct _AlphaListRec {
    fakeBoxRec *rects;
    int count;
    unsigned int alpha;
    int bpr;
    int bpp;
    AlphaListRec *next;
};

#import <Cocoa/Cocoa.h>

@interface XView : NSQuickDrawView
{
    NSRect mMyFrame;
    int mHackID;
    int mPrivateID;
    int mLockFocusCount;
    BOOL mWantFlipped;
    BOOL mInDrawRect;
    BOOL mWaitForShape;
    BOOL mFirstClick;
    BOOL mInside;
    BOOL mMyMiniature;
    BOOL mResizing;
    BOOL mOneOfMine;
    BOOL mIsMyLockFocus;
    BOOL mLockedPortBits;
    NSRect mNewResizeFrame;
    char *mPixels;
    int mBPR;
    int mBPP;
    char *mOldPixels;
    int mOldWidth;
    int mOldHeight;
    int mOldBPP;
    int mOldBPR;
    NSTrackingRectTag mRectTag;
    int mStoredCount;
    fakeBoxRec *mStoredRects;
    AlphaListRec *mAlphaRectsList;
    AlphaListRec *mAlphaRectsLast;
}

- (void)setFrameOrigin:(NSPoint)pt;
- (void)setFrameSize:(NSSize)size;
- (void)setFrame:(NSRect)frame;
- (NSRect)frame;
- (void)setWantFlipped:(BOOL)val;
- (void)setWaitForShape:(BOOL)val;
- (void)setInDrawRect:(BOOL)val;
- (BOOL)isInDrawRect;
- (void)waitLockPortBits;
- (void)myUnlockPortBits;
- (void)initLocks:(int)count bits:(BOOL)locked;
- (void)lockFocus;
- (void)unlockFocus;
- (void)myLockFocus;
- (void)myUnlockFocus;
- (int)portIsLocked;

- (void)initAlphaRects;
- (void)storeShapeRects:(fakeBoxRec *)rects count:(int)count;
- (void)addAlphaRects:(fakeBoxRec *)rects count:(int)count
        alpha:(unsigned int)alpha BPR:(int)bpr BPP:(int)bpp;
- (void)freeAlphaRects;
- (void)copyFromSrc:(char *)srcBits srcBPR:(int)srcBPR srcBPP:(int)srcBPP
                alpha:(unsigned int)alpha
                rects:(fakeBoxRec *)rectList count:(int)count;
- (void)drawRect:(NSRect)aRect;
- (void)displayIfNeeded;
- (void)display;

- (BOOL)isFlipped;
- (BOOL)isOpaque;
- (BOOL)acceptsFirstResponder;
- (BOOL)acceptsFirstMouse:(NSEvent *)theEvent;
- (BOOL)shouldDelayWindowOrderingForEvent:(NSEvent *)theEvent;

- (void)mouseDown:(NSEvent *)anEvent;
- (void)rightMouseDown:(NSEvent *)anEvent;
- (void)otherMouseDown:(NSEvent *)anEvent;

- (void)setAllIDs:(int)privateid hackID:(int)hackid;
- (int)getMyHackID;
- (int)getMyPrivateID;
- (void)createTrackingRect;
- (void)deleteTrackingRect;
- (NSTrackingRectTag)getMyTrackingRect;
- (void)setMyTrackingRect:(NSTrackingRectTag)tag;
- (void)mouseEntered:(NSEvent *)anEvent;
- (void)mouseExited:(NSEvent *)anEvent;
- (void)sendSpecialEvent:(int)info which:(int)which dx:(int)dx dy:(int)dy;
- (void)setFirstClick:(BOOL)val;
- (BOOL)getFirstClick;
- (BOOL)getInside;
- (void)setMyMiniature:(BOOL)val;
- (BOOL)getMyMiniature;
- (NSRect)getNewResizeFrame;
- (void)setWaitResizeFlag:(NSRect)newFrame oldPix:(void *)oldPix;
- (void)unsetWaitResizeFlag;
- (BOOL)getWaitResizeFlag;
- (void)windowDidDeminiaturize:(NSNotification *)notification;

@end
