/**************************************************************
 *
 * Support for using the Quartz Window Manager cursor
 * NSCursor version
 *
 **************************************************************/
/*
 * Copyright (c) 2001-2003 Torrey T. Lyons and Greg Parker.
 *                 All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE ABOVE LISTED COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name(s) of the above copyright
 * holders shall not be used in advertising or otherwise to promote the sale,
 * use or other dealings in this Software without prior written authorization.
 */
/* $XFree86: xc/programs/Xserver/hw/darwin/quartz/quartzCursor.c,v 1.5 2003/01/23 00:34:26 torrey Exp $ */

// hack around conflicting BOOL definition
#define BOOL NSBOOL
#import <AppKit/NSCursor.h>
#import <AppKit/NSImage.h>
#import <AppKit/NSBitmapImageRep.h>
#import <Foundation/NSAutoreleasePool.h>
#undef BOOL

// hack
#define __EMX__

#include "insidecheck.h"
#include "communication.h"
#include "quartzCommon.h"
#include "quartzCursor.h"
#include "darwin.h"

#include "mi.h"
#include "scrnintstr.h"
#include "cursorstr.h"
#include "mipointrst.h"
#include "globals.h"
#include "servermd.h"

// hack
#undef __EMX__

#include "myXPlugin.h"

// Maximum NSCursor size
#define CURSORWIDTH 32
#define CURSORHEIGHT 32

// Cursor debugging
#define CURSOR_DEBUG ErrorF

typedef struct {
    int                     qdCursorMode;
    int                     qdCursorVisible;
    int                     useQDCursor;
    QueryBestSizeProcPtr    QueryBestSize;
    miPointerSpriteFuncPtr  spriteFuncs;
} QuartzCursorScreenRec, *QuartzCursorScreenPtr;

static int darwinCursorScreenIndex = -1;
static unsigned long darwinCursorGeneration = 0;
static  CursorPtr quartzLatentCursor = NULL;

int quartz_no_cursor_warp=1;

#define CURSOR_PRIV(pScreen) \
    ((QuartzCursorScreenPtr)pScreen->devPrivates[darwinCursorScreenIndex].ptr)

#define HIDE_QD_CURSOR(pScreen, visible)                                \
    if (visible) {                                                      \
        int ix;                                                         \
        for (ix = 0; ix < QUARTZ_PRIV(pScreen)->displayCount; ix++) {   \
            CGDisplayHideCursor(QUARTZ_PRIV(pScreen)->displayIDs[ix]);  \
        }                                                               \
        visible = FALSE;                                                \
    } ((void)0)

#define SHOW_QD_CURSOR(pScreen, visible)                                \
    {                                                                   \
        int ix;                                                         \
        for (ix = 0; ix < QUARTZ_PRIV(pScreen)->displayCount; ix++) {   \
            CGDisplayShowCursor(QUARTZ_PRIV(pScreen)->displayIDs[ix]);  \
        }                                                               \
        visible = TRUE;                                                 \
    } ((void)0)


/* swiped from Apple's X Server code :O */
static Bool
load_cursor (CursorPtr src, int screen)
{
    uint32_t *data;
    uint32_t rowbytes;
    int width, height;
    int hot_x, hot_y;

    uint32_t fg_color, bg_color;
    uint8_t *srow, *sptr;
    uint8_t *mrow, *mptr;
    uint32_t *drow, *dptr;
    unsigned xcount, ycount;

    myxp_error err;

    width = src->bits->width;
    height = src->bits->height;
    hot_x = src->bits->xhot;
    hot_y = src->bits->yhot;

#ifdef ARGB_CURSOR
    if (src->bits->argb != NULL)
    {
	rowbytes = src->bits->width * sizeof (CARD32);
	data = (uint32_t *) src->bits->argb;
    }
    else
#endif
    {
	fg_color = 0xFF00 | (src->foreRed >> 8);
	fg_color <<= 16;
	fg_color |= src->foreGreen & 0xFF00;
	fg_color |= src->foreBlue >> 8;

	bg_color = 0xFF00 | (src->backRed >> 8);
	bg_color <<= 16;
	bg_color |= src->backGreen & 0xFF00;
	bg_color |= src->backBlue >> 8;

	fg_color = htonl (fg_color);
	bg_color = htonl (bg_color);

	/* round up to 8 pixel boundary so we can convert whole bytes */
	rowbytes = ((src->bits->width * 4) + 31) & ~31;
	data = alloca (rowbytes * src->bits->height);

	if (!src->bits->emptyMask)
	{
	    ycount = src->bits->height;
	    srow = src->bits->source; mrow = src->bits->mask;
	    drow = data;

	    while (ycount-- > 0)
	    {
		xcount = (src->bits->width + 7) / 8;
		sptr = srow; mptr = mrow;
		dptr = drow;

		while (xcount-- > 0)
		{
		    uint8_t s, m;
		    int i;

		    s = *sptr++; m = *mptr++;
		    for (i = 0; i < 8; i++)
		    {
#if BITMAP_BIT_ORDER == MSBFirst
			if (m & 128)
			    *dptr++ = (s & 128) ? fg_color : bg_color;
			else
			    *dptr++ = 0;
			s <<= 1; m <<= 1;
#else
			if (m & 1)
			    *dptr++ = (s & 1) ? fg_color : bg_color;
			else
			    *dptr++ = 0;
			s >>= 1; m >>= 1;
#endif
		    }
		}

		srow += BitmapBytePad (src->bits->width);
		mrow += BitmapBytePad (src->bits->width);
		drow = (uint32_t *) ((char *) drow + rowbytes);
	    }
	}
	else
	{
	    memset (data, 0, src->bits->height * rowbytes);
	}
    }

    err = myxp_set_cursor (width, height, hot_x, hot_y, data, rowbytes);
    return err == Success;
}

/*
 * MakeNSCursor: construct an NSCursor from an X11 CursorPtr
 * pCursor must be no larger than 16x16.
 *
 * X11 cursor is 1-bit source and 1-bit mask and fore and back colors.
 * The mapping from X11 cursor bits to X11 colors to NSImage colors is:
 * (source,mask | X11 color   | NSImage color)
 *          00  | transparent | transparent
 *          01  | back color  | back color
 *          10  | undefined   | transparent
 *          11  | fore color  | fore color
 */
static NSCursor *MakeNSCursor(CursorPtr pCursor)
{
    int w, h;
    int width = pCursor->bits->width;
    int height = pCursor->bits->height;
    NSPoint hotSpot = NSMakePoint(pCursor->bits->xhot, pCursor->bits->yhot);

    int factor = 1;
    int nswidth = 16; // width and height of NSImage
    int nsheight = 16; // NSCursor requires exactly 16x16 pixels

    // Image bits are 32-bit RGBA
    // X11 color components are 16-bit - shift to get 8-bit
    unsigned int transparent = 0x00000000;
    unsigned int foreColor = ((((pCursor->foreRed   >> 8) & 0xff) << 24) |
                              (((pCursor->foreGreen >> 8) & 0xff) << 16) |
                              (((pCursor->foreBlue  >> 8) & 0xff) << 8)  |
                              0xff);
    unsigned int backColor = ((((pCursor->backRed   >> 8) & 0xff) << 24) |
                              (((pCursor->backGreen >> 8) & 0xff) << 16) |
                              (((pCursor->backBlue  >> 8) & 0xff) << 8)  |
                              0xff);
    
    NSBitmapImageRep *imageRep;
    unsigned int *imageBits;
    NSImage *image;
    NSCursor *nsCursor = nil;
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

    // deal with cursors > 16x16
    if(width>16 || height>16) {
       hotSpot = NSMakePoint(pCursor->bits->xhot/2, pCursor->bits->yhot/2);
       factor = 2;
    }
    
    // Create a cursor bitmap
    imageRep = [[[NSBitmapImageRep alloc] 
                    initWithBitmapDataPlanes:NULL
                    pixelsWide:nswidth    pixelsHigh:nsheight
                    bitsPerSample:8       samplesPerPixel:4
                    hasAlpha:YES          isPlanar:NO
                    colorSpaceName:NSCalibratedRGBColorSpace
                    bytesPerRow:nswidth*4 bitsPerPixel:32] autorelease];
    if (!imageRep) goto done;
    imageBits = (unsigned int *)[imageRep bitmapData];
    bzero(imageBits, nswidth * nsheight * 4);

    if(factor==2) {
    // Draw the cursor in the bitmap.
        if(verboseConsole) {
            CURSOR_DEBUG("Cursor (%d x %d):\n", width, height);
            CURSOR_DEBUG("foreground 0x%x ('#'), background 0x%x ('+')\n", 
                        foreColor, backColor);
        }
        
        for (h = 0; h < height; h++) {
        
            unsigned long sourceRow = 
                (pCursor->bits->source[h*4]   << 24) | (pCursor->bits->source[h*4+1] << 16) |
                (pCursor->bits->source[h*4+2] << 8)  | pCursor->bits->source[h*4+3];
            unsigned long maskRow = 
                (pCursor->bits->mask[h*4]   << 24) | (pCursor->bits->mask[h*4+1] << 16 ) |
                (pCursor->bits->mask[h*4+2] << 8)  | pCursor->bits->mask[h*4+3];
            unsigned long bitmask = 0x80000000;
            
            for (w = 0; w < width; w++, bitmask >>= 1) {
                if ((maskRow & bitmask) == 0) CURSOR_DEBUG(" ");
                else if ((sourceRow & bitmask) == 0) CURSOR_DEBUG("+");
                else CURSOR_DEBUG("#");
            }
            if(verboseConsole) CURSOR_DEBUG("|\n");
        }
    }
    
    for (h = 0; h < height; h+=factor) {
        // 16-bit row. High width bits are the bits to use.
        unsigned long sourceRow = 
            (pCursor->bits->source[h*4] << 8) | pCursor->bits->source[h*4+1];
        unsigned long maskRow = 
            (pCursor->bits->mask[h*4] << 8) | pCursor->bits->mask[h*4+1];
        unsigned long bitmask = 0x8000;

        if(factor==2) {

            unsigned long sourceRow2 = 
                (pCursor->bits->source[h*4+4] << 24) | (pCursor->bits->source[h*4+5] << 16) |
                (pCursor->bits->source[h*4+6] << 8)  | pCursor->bits->source[h*4+7];
            unsigned long maskRow2 = 
                (pCursor->bits->mask[h*4+4] << 24) | (pCursor->bits->mask[h*4+5] << 16 ) |
                (pCursor->bits->mask[h*4+6] << 8)  | pCursor->bits->mask[h*4+7];
        
            sourceRow = (sourceRow << 16) | (pCursor->bits->source[h*4+2] << 8) | pCursor->bits->source[h*4+3];
            maskRow   = (maskRow << 16)   | (pCursor->bits->mask[h*4+2] << 8)   | pCursor->bits->mask[h*4+3];
            bitmask   = bitmask << 16;
            
            for (w = 0; w < width; w+=2, bitmask >>= 2) {
                unsigned int color;
                unsigned long bitmask2 = bitmask >> 1;
                
//                if((maskRow & bitmask)  == 0 && (maskRow2 & bitmask)  == 0 &
//                   (maskRow & bitmask2) == 0 && (maskRow2 & bitmask2) == 0)
//                   color = transparent;
//                else {
                    if((maskRow & bitmask) == 0)
                        color = transparent;
                    else if((sourceRow & bitmask) == 0)
                        color = (backColor & 0xFCFCFCFC) >> 2;
                    else
                        color = (foreColor & 0xFCFCFCFC) >> 2;
                        
                    if((maskRow2 & bitmask) == 0)
                        ;
                    else if((sourceRow2 & bitmask) == 0)
                        color = color + ((backColor & 0xFCFCFCFC) >> 2);
                    else
                        color = color + ((foreColor & 0xFCFCFCFC) >> 2);
                        
                    if((maskRow & bitmask2) == 0)
                        ;
                    else if((sourceRow & bitmask2) == 0)
                        color = color + ((backColor & 0xFCFCFCFC) >> 2);
                    else
                        color = color + ((foreColor & 0xFCFCFCFC) >> 2);
                        
                    if((maskRow2 & bitmask2) == 0)
                        ;
                    else if((sourceRow2 & bitmask2) == 0)
                        color = color + ((backColor & 0xFCFCFCFC) >> 2);
                    else
                        color = color + ((foreColor & 0xFCFCFCFC) >> 2);

                    if((color & 0xFF) > 0) color |= 0xFF;
                    if((color & 0xFF00) > 0xFB00) color |= 0xFF00;
                    if((color & 0xFF0000) > 0xFB0000) color |= 0xFF0000;
                    if((color & 0xFF000000) > 0xFB000000) color |= 0xFF000000;
//                }
            
                imageBits[(h*nswidth + w)/2] = color;

                if(verboseConsole) {
                    if (color == transparent) CURSOR_DEBUG(" ");
                    else if (color == backColor) CURSOR_DEBUG("+");
                    else if (color == foreColor) CURSOR_DEBUG("#");
                    else CURSOR_DEBUG("x");
                }
            }
            if(verboseConsole) CURSOR_DEBUG("|\n");

        } else {
        
            for (w = 0; w < width; w++, bitmask >>= 1) {
                unsigned int color;
                if ((maskRow & bitmask) == 0) color = transparent;
                else if ((sourceRow & bitmask) == 0) color = backColor;
                else color = foreColor;
            
                imageBits[h*nswidth + w] = color;

//                if (color == transparent) CURSOR_DEBUG(" ");
//                else if (color == backColor) CURSOR_DEBUG("+");
//                else CURSOR_DEBUG("#");
            }
//            CURSOR_DEBUG("|\n");

        }
    }
    if(verboseConsole) ErrorF("XDarwin info: created a new NSCursor\n");
    // Create an NSImage from the cursor bitmap.
    image = [[[NSImage alloc] initWithSize:[imageRep size]] autorelease];
    if (!image) goto done;
    [image addRepresentation:imageRep];

    // Create an NSCursor from the image.
    nsCursor = [[NSCursor alloc] initWithImage:image hotSpot:hotSpot];

    done:
    [pool release];

    return nsCursor;
}


/*
===========================================================================

 Pointer sprite functions

===========================================================================
*/


/*
 * A realized cursor's devPrivate is always one of the following:
 * - a constructed NSCursor 
 * - a retained copy of [NSCursor arrowCursor]
 * - nil (iff using X11 software cursor in full-screen mode)
 * In particular:
 * - a nil devPrivate always means X11 software cursor is in use
 * - a non-nil devPrivate is always a retained NSCursor.
 */

/*
 * QuartzRealizeCursor
 * Convert the X cursor representation to QuickDraw format if possible.
 */
Bool
QuartzRealizeCursor(
    ScreenPtr pScreen,
    CursorPtr pCursor )
{
    Bool result;
    NSCursor *nsCursor;
    QuartzCursorScreenPtr ScreenPriv = CURSOR_PRIV(pScreen);

    if(!pCursor || !pCursor->bits)
        return FALSE;
 
    if( myxp_use_xp_cursor ) return TRUE;

    if (pCursor->bits->height <= CURSORHEIGHT  &&  
        pCursor->bits->width  <= CURSORWIDTH  &&  
        ScreenPriv->useQDCursor) 
    {
        // Cursor is small enough for NSCursor
        nsCursor = MakeNSCursor(pCursor);
        result = (nsCursor != nil);
    } else {
        // Cursor is too big for NSCursor. 
        if (quartzRootless) {
            // Rootless mode can't use a software cursor - use an arrow instead
            nsCursor = [[NSCursor arrowCursor] retain];
            result = TRUE;
        } else {
            // Full-screen mode can use an X11 software cursor
            nsCursor = nil;
            result = (*ScreenPriv->spriteFuncs->RealizeCursor)
                (pScreen, pCursor);
        }
    }

    // save the NSCursor, if any
    pCursor->devPriv[pScreen->myNum] = (pointer) nsCursor;

    return result;
}


/*
 * QuartzUnrealizeCursor
 * Free the storage space associated with a realized cursor.
 */
Bool
QuartzUnrealizeCursor(
    ScreenPtr pScreen,
    CursorPtr pCursor )
{
    Bool result;
    QuartzCursorScreenPtr ScreenPriv = CURSOR_PRIV(pScreen);
    NSCursor *nsCursor = (NSCursor *)pCursor->devPriv[pScreen->myNum];

    if( myxp_use_xp_cursor ) return TRUE;

    if (nsCursor) {
        // We constructed an NSCursor or retained a copy of the arrow. 
        [nsCursor release];
        result = TRUE;
    } else if (quartzRootless) {
        // This should never happen.
        ErrorF("QuartzUnrealizeCursor - nil cursor in rootless mode!\n");
        result = FALSE;
    } else {
        // We used a software cursor
        result = (*ScreenPriv->spriteFuncs->UnrealizeCursor)
            (pScreen, pCursor);
    }

    pCursor->devPriv[pScreen->myNum] = NULL;

    return result;
}


/*
 * QuartzSetCursor
 * Set the cursor sprite and position.
 * Use QuickDraw cursor if possible.
 */
static void
QuartzSetCursor(
    ScreenPtr       pScreen,
    CursorPtr       pCursor,
    int             x,
    int             y)
{
    QuartzCursorScreenPtr ScreenPriv = CURSOR_PRIV(pScreen);
    NSCursor *nsCursor = pCursor ? (NSCursor *)pCursor->devPriv[pScreen->myNum]:nil;

    if(verboseConsole) ErrorF("XDarwin info: into QuartzSetCursor (%#lx)\n",pCursor);
    quartzLatentCursor = pCursor;

    // Don't touch Mac OS cursor if X is hidden!
    if (!quartzServerVisible)
        return;

    if( myxp_use_xp_cursor ) {
        if (!pCursor) {
            if (ScreenPriv->qdCursorVisible)
            {
                //ErrorF("Hiding xp_cursor...\n");
                myxp_hide_cursor();
                ScreenPriv->qdCursorVisible = FALSE;
            }
        }
        else
        {
            load_cursor (pCursor, pScreen->myNum);

            if (!ScreenPriv->qdCursorVisible)
            {
                ErrorF("---- Using myxp_show_cursor...\n");
                myxp_show_cursor();
                ScreenPriv->qdCursorVisible = TRUE;
            }
        }
        return;
    }

    if (!pCursor) {
        if(verboseConsole) ErrorF("XDarwin info: removing cursor (pCursor=nil)\n");
        // Remove the cursor completely.
        HIDE_QD_CURSOR(pScreen, ScreenPriv->qdCursorVisible);
        if (! ScreenPriv->qdCursorMode)
            (*ScreenPriv->spriteFuncs->SetCursor)(pScreen, 0, x, y);
    }
    else if (nsCursor != nil) {
        // An NSCursor is available. Use it.
        if (! ScreenPriv->qdCursorMode) {
            // remove the X cursor
            (*ScreenPriv->spriteFuncs->SetCursor)(pScreen, 0, x, y);
            ScreenPriv->qdCursorMode = TRUE;
        }
        if(verboseConsole) ErrorF("XDarwin info: changing NSCursor\n");
        [nsCursor set];
        SHOW_QD_CURSOR(pScreen, ScreenPriv->qdCursorVisible);
    }
    else if (quartzRootless) {
        // This should never happen.
        ErrorF("QuartzSetCursor: nil cursor in rootless mode!");
    }
    else {
        // No NSCursor available. Use X11 software cursor instead.
        HIDE_QD_CURSOR(pScreen, ScreenPriv->qdCursorVisible);
        ScreenPriv->qdCursorMode = FALSE;
        (*ScreenPriv->spriteFuncs->SetCursor)(pScreen, pCursor, x, y);
    }
}


/*
 * QuartzMoveCursor
 * Move the cursor. This is a noop for QuickDraw.
 */
static void
QuartzMoveCursor(
    ScreenPtr   pScreen,
    int         x,
    int         y)
{
    QuartzCursorScreenPtr ScreenPriv = CURSOR_PRIV(pScreen);

    // only the X11 software cursor needs to be explicitly moved
    if (!ScreenPriv->qdCursorMode)
        (*ScreenPriv->spriteFuncs->MoveCursor)(pScreen, x, y);
}


static miPointerSpriteFuncRec quartzSpriteFuncsRec = {
    QuartzRealizeCursor,
    QuartzUnrealizeCursor,
    QuartzSetCursor,
    QuartzMoveCursor
};

/*
===========================================================================

 Pointer screen functions

===========================================================================
*/

/*
 * QuartzCursorOffScreen
 */
static Bool QuartzCursorOffScreen(ScreenPtr *pScreen, int *x, int *y)
{
    return FALSE;
}


/*
 * QuartzCrossScreen
 */
static void QuartzCrossScreen(ScreenPtr pScreen, Bool entering)
{
    return;
}


/*
 * QuartzWarpCursor
 *  Change the cursor position without generating an event or motion history.
 *  The input coordinates (x,y) are in pScreen-local X11 coordinates.
 *
 */
static void
QuartzWarpCursor(
    ScreenPtr               pScreen,
    int                     x,
    int                     y)
{
    static int              neverMoved = TRUE;

    if (quartz_no_cursor_warp) {
        // Don't move the cursor the first time. This is the jump-to-center
        // initialization, and it's annoying because we may still be in MacOS.
        quartz_no_cursor_warp = 0;
        return;
    }

    if (quartzServerVisible) {
        CGDisplayErr        cgErr;
        CGPoint             cgPoint;
        // Only need to do this for one display. Any display will do.
        CGDirectDisplayID   cgID = QUARTZ_PRIV(pScreen)->displayIDs[0];
        CGRect              cgRect = CGDisplayBounds(cgID);

        // Convert (x,y) to CoreGraphics screen-local CG coordinates.
        // This is necessary because the X11 screen and CG screen may not
        // coincide. (e.g. X11 screen may be moved to dodge the menu bar)

        // Make point in X11 global coordinates
        cgPoint = CGPointMake(x + dixScreenOrigins[pScreen->myNum].x,
                              y + dixScreenOrigins[pScreen->myNum].y);
        // Shift to CoreGraphics global screen coordinates
        cgPoint.x += darwinMainScreenX;
        cgPoint.y += darwinMainScreenY;
        // Shift to CoreGraphics screen-local coordinates
        cgPoint.x -= cgRect.origin.x;
        cgPoint.y -= cgRect.origin.y;

        cgErr = CGDisplayMoveCursorToPoint(cgID, cgPoint);
        if (cgErr != CGDisplayNoErr) {
            ErrorF("Could not set cursor position with error code 0x%x.\n",
                    cgErr);
        }
    }

    miPointerWarpCursor(pScreen, x, y);
    miPointerUpdate();
}


static miPointerScreenFuncRec quartzScreenFuncsRec = {
    QuartzCursorOffScreen,
    QuartzCrossScreen,
    QuartzWarpCursor,
    DarwinEQPointerPost,
    DarwinEQSwitchScreen
};

/*
===========================================================================

 Other screen functions

===========================================================================
*/

/*
 * QuartzCursorQueryBestSize
 * Handle queries for best cursor size
 */
static void
QuartzCursorQueryBestSize(
   int              class,
   unsigned short   *width,
   unsigned short   *height,
   ScreenPtr        pScreen)
{
    QuartzCursorScreenPtr ScreenPriv = CURSOR_PRIV(pScreen);

    if (class == CursorShape) {
        *width = CURSORWIDTH;
        *height = CURSORHEIGHT;
    } else {
        (*ScreenPriv->QueryBestSize)(class, width, height, pScreen);
    }
}


/*
 * QuartzInitCursor
 * Initialize cursor support
 */
Bool
QuartzInitCursor(
    ScreenPtr   pScreen )
{
    QuartzCursorScreenPtr   ScreenPriv;
    miPointerScreenPtr      PointPriv;
    DarwinFramebufferPtr    dfb = SCREEN_PRIV(pScreen);

    // initialize software cursor handling (always needed as backup)
    if (!miDCInitialize(pScreen, &quartzScreenFuncsRec)) {
        return FALSE;
    }

    // allocate private storage for this screen's QuickDraw cursor info
    if (darwinCursorGeneration != serverGeneration) {
        if ((darwinCursorScreenIndex = AllocateScreenPrivateIndex()) < 0)
            return FALSE;
        darwinCursorGeneration = serverGeneration;
    }

    ScreenPriv = xcalloc( 1, sizeof(QuartzCursorScreenRec) );
    if (!ScreenPriv) return FALSE;

    CURSOR_PRIV(pScreen) = ScreenPriv;

    // override some screen procedures
    ScreenPriv->QueryBestSize = pScreen->QueryBestSize;
    pScreen->QueryBestSize = QuartzCursorQueryBestSize;

    // initialize QuickDraw cursor handling
    PointPriv = (miPointerScreenPtr)
                    pScreen->devPrivates[miPointerScreenIndex].ptr;

    ScreenPriv->spriteFuncs = PointPriv->spriteFuncs;
    PointPriv->spriteFuncs = &quartzSpriteFuncsRec;

    if (!quartzRootless)
        ScreenPriv->useQDCursor = QuartzFSUseQDCursor(dfb->colorBitsPerPixel);
    else
        ScreenPriv->useQDCursor = TRUE;
    ScreenPriv->qdCursorMode = TRUE;
    ScreenPriv->qdCursorVisible = TRUE;

    if(myxp_use_xp_cursor) myxp_init(0);

    return TRUE;
}


// X server is hiding. Restore the Aqua cursor.
void QuartzSuspendXCursor(
    ScreenPtr pScreen )
{
    QuartzCursorScreenPtr ScreenPriv = CURSOR_PRIV(pScreen);

    [[NSCursor arrowCursor] set];
    SHOW_QD_CURSOR(pScreen, ScreenPriv->qdCursorVisible);
}


// X server is showing. Restore the X cursor.
void QuartzResumeXCursor(
    ScreenPtr pScreen,
    int x,
    int y )
{
    QuartzSetCursor(pScreen, quartzLatentCursor, x, y);
}
