/*
 * Generic rootless to Aqua specific glue code
 *
 * This code acts as a glue between the generic rootless X server code
 * and the Aqua specific implementation, which includes definitions that
 * conflict with stardard X types.
 *
 * Greg Parker     gparker@cs.stanford.edu
 */
/* $XFree86: xc/programs/Xserver/hw/darwin/quartz/rootlessAquaGlue.c,v 1.6 2003/04/30 23:15:39 torrey Exp $ */

#include "communication.h"
#include "privatewinlist.h"
#include "quartzCommon.h"
#include "darwin.h"
#include "rootlessAqua.h"
#include "rootlessAquaImp.h"
#include "rootless.h"
#include "aqua.h"

#include "regionstr.h"
#include "scrnintstr.h"
#include "picturestr.h"
#include "globals.h" // dixScreenOrigins[]

int in_command_hack=0;

extern char *gResizeDeathBits;

/////////////////////////////////////////
// Rootless mode callback glue

static void
AquaGlueIsInsideFrame(RootlessFramePtr pFrame)
{
    AquaIsInsideWindow(pFrame->devPrivate);
}


static void
AquaGlueCreateFrame(ScreenPtr pScreen, RootlessFramePtr pFrame,
                    RootlessFramePtr pUpper)
{
    int sx = dixScreenOrigins[pScreen->myNum].x + darwinMainScreenX;
    int sy = dixScreenOrigins[pScreen->myNum].y + darwinMainScreenY;

    if( pFrame->x==-23451 && pFrame->y==-6789 ) {
        sx = 0;
        sy = 0;
    }

    pFrame->devPrivate = AquaNewWindow(pUpper ? pUpper->devPrivate : NULL,
                                       pFrame->x + sx, pFrame->y + sy,
                                       pFrame->w, pFrame->h,
                                       pFrame->isRoot);
}


static void
AquaGlueDestroyFrame(ScreenPtr pScreen, RootlessFramePtr pFrame)
{
    AquaDestroyWindow(pFrame->devPrivate);
}


static void
AquaGlueMoveFrame(ScreenPtr pScreen, RootlessFramePtr pFrame,
                  int oldX, int oldY)
{
    static int	lastxpos;
    static void	*lastwinrec;

    int sx = dixScreenOrigins[pScreen->myNum].x + darwinMainScreenX;
    int sy = dixScreenOrigins[pScreen->myNum].y + darwinMainScreenY;

    if(pFrame->x==-23451 && pFrame->y==-6789) {
        in_command_hack=1;
        lastwinrec = NULL;	// start of new sequence, so assume lastwinrec should be NULL
    } else if(pFrame->x==-23451 && pFrame->y==-9876) {
        QuartzMessageMainThread(kQuartzWantMouseMoved,nil,0);
        in_command_hack=0;
        return;
/*    } else if(pFrame->x==-23451 && (pFrame->y==-100 || pFrame->y==-101 || pFrame->y==-99)) {
        if(lastwinrec==NULL)
            AquaHackCommand(pFrame->devPrivate, -23495, pFrame->y+101);
        else
            AquaHackCommand(lastwinrec, -23495, pFrame->y+101);
//        in_command_hack=0;
        return;
*/    } else if(in_command_hack==1) {
        if(pFrame->x==-23455) {
            if(pFrame->y<=0) {
//           	 ErrorF("---- clearing lastwinrec %#lx\n",lastwinrec);
                lastwinrec = NULL;
            } else {
                lastwinrec = GetWinByHackID(pFrame->y);
//          	  ErrorF("---- got lastwinrec %#lx\n",lastwinrec);
            }
            return;
        } else if(pFrame->x==-23456) {
            lastxpos = pFrame->y;
            return;
        } else if(pFrame->x==-23457) {
            pFrame->x = lastxpos;
//        	ErrorF("---- fake move of window to %d %d %#lx\n",lastxpos,pFrame->y,lastwinrec);
            if(lastwinrec==NULL)
                AquaMoveWindow(pFrame->devPrivate, pFrame->x+sx, pFrame->y+sy);
            else
                AquaMoveWindow(lastwinrec, pFrame->x+sx, pFrame->y+sy);
//       	 in_command_hack=0;
            return;
        } else if(pFrame->x>-23500 && pFrame->x<=-23400) {
            if(lastwinrec==NULL)
                AquaHackCommand(pFrame->devPrivate, pFrame->x, pFrame->y);
            else {
                AquaHackCommand(lastwinrec, pFrame->x, pFrame->y);
            }
        } else
            in_command_hack=0;
    }
    
    if(in_command_hack==0)
        AquaMoveWindow(pFrame->devPrivate, pFrame->x + sx, pFrame->y + sy);
}


static void
AquaGlueStartResizeFrame(ScreenPtr pScreen, RootlessFramePtr pFrame,
                         int oldX, int oldY,
                         unsigned int oldW, unsigned int oldH, void *oldPixels)
{
    int sx = dixScreenOrigins[pScreen->myNum].x + darwinMainScreenX;
    int sy = dixScreenOrigins[pScreen->myNum].y + darwinMainScreenY;

    AquaStartResizeWindow(pFrame->devPrivate,
                          pFrame->x + sx, pFrame->y + sy, pFrame->w, pFrame->h, oldPixels);
}

static void
AquaGlueFinishResizeFrame(ScreenPtr pScreen, RootlessFramePtr pFrame,
                          int oldX, int oldY,
                          unsigned int oldW, unsigned int oldH)
{
    int sx = dixScreenOrigins[pScreen->myNum].x + darwinMainScreenX;
    int sy = dixScreenOrigins[pScreen->myNum].y + darwinMainScreenY;

    AquaFinishResizeWindow(pFrame->devPrivate,
                           pFrame->x + sx, pFrame->y + sy,
                           pFrame->w, pFrame->h);
}


static void
AquaGlueRestackFrame(ScreenPtr pScreen, RootlessFramePtr pFrame,
                     RootlessFramePtr pOldPrev,
                     RootlessFramePtr pNewPrev)
{
    AquaRestackWindow(pFrame->devPrivate,
                      pNewPrev ? pNewPrev->devPrivate : NULL);
}


static void
AquaGlueReshapeFrame(ScreenPtr pScreen, RootlessFramePtr pFrame,
                     RegionPtr pNewShape)
{
    BoxRec shapeBox = {0, 0, pFrame->w, pFrame->h};

    // Don't correct for dixScreenOrigins here.
    // pNewShape is in window-local coordinates.

    if (pFrame->isRoot) return; // shouldn't happen; mi or dix covers this

    // Hack to make sure pointer is updated for temporary translucent drawing buffer
    pFrame->pixelData = gResizeDeathBits;
    
    REGION_INVERSE(pScreen, pNewShape, pNewShape, &shapeBox);
    AquaReshapeWindow(pFrame->devPrivate,
                      (fakeBoxRec *) REGION_RECTS(pNewShape),
                      REGION_NUM_RECTS(pNewShape));
}


static void
AquaGlueUpdateRegion(ScreenPtr pScreen, RootlessFramePtr pFrame,
                     RegionPtr pDamage)
{
    AquaUpdateRects(pFrame->devPrivate,
                    (fakeBoxRec *) REGION_RECTS(pDamage),
                    REGION_NUM_RECTS(pDamage));
}


static void
AquaGlueStartDrawing(ScreenPtr pScreen, RootlessFramePtr pFrame)
{
    AquaStartDrawing(pFrame->devPrivate, &pFrame->pixelData,
                     &pFrame->bytesPerRow, &pFrame->depth,
                     &pFrame->bitsPerPixel);
}

static void
AquaGlueStopDrawing(ScreenPtr pScreen, RootlessFramePtr pFrame)
{
    AquaStopDrawing(pFrame->devPrivate);
}


static RootlessFrameProcs aquaRootlessProcs = {
    AquaGlueCreateFrame,
    AquaGlueDestroyFrame,
    AquaGlueMoveFrame,
    AquaGlueStartResizeFrame,
    AquaGlueFinishResizeFrame,
    AquaGlueRestackFrame,
    AquaGlueReshapeFrame,
    AquaGlueUpdateRegion,
    AquaGlueStartDrawing,
    AquaGlueStopDrawing,
    AquaGlueIsInsideFrame
};


///////////////////////////////////////
// Rootless mode initialization.
// Exported by rootlessAqua.h

/*
 * AquaDisplayInit
 *  Find all Aqua screens.
 */
void
AquaDisplayInit(void)
{
    darwinScreensFound = AquaDisplayCount();
}


/*
 * AquaAddScreen
 *  Init the framebuffer and record pixmap parameters for the screen.
 */
Bool
AquaAddScreen(int index, ScreenPtr pScreen)
{
    DarwinFramebufferPtr dfb = SCREEN_PRIV(pScreen);
    QuartzScreenPtr displayInfo = QUARTZ_PRIV(pScreen);
    CGRect cgRect;
    CGDisplayCount numDisplays;
    CGDisplayCount allocatedDisplays = 0;
    CGDirectDisplayID *displays = NULL;
    CGDisplayErr cgErr;
    int componentCount;

    dfb->colorType = TrueColor;
    AquaScreenInit(index, &dfb->x, &dfb->y, &dfb->width, &dfb->height,
                    &dfb->pitch, &dfb->bitsPerComponent,
                    &componentCount, &dfb->bitsPerPixel);
    dfb->colorBitsPerPixel = dfb->bitsPerComponent * componentCount;

    // No frame buffer - it's all in window pixmaps.
    dfb->framebuffer = NULL; // malloc(dfb.pitch * dfb.height);

    // Get all CoreGraphics displays covered by this X11 display.
    cgRect = CGRectMake(dfb->x, dfb->y, dfb->width, dfb->height);
    do {
        cgErr = CGGetDisplaysWithRect(cgRect, 0, NULL, &numDisplays);
        if (cgErr) break;
        allocatedDisplays = numDisplays;
        displays = xrealloc(displays,
                            numDisplays * sizeof(CGDirectDisplayID));
        cgErr = CGGetDisplaysWithRect(cgRect, allocatedDisplays, displays,
                                      &numDisplays);
        if (cgErr != CGDisplayNoErr) break;
    } while (numDisplays > allocatedDisplays);

    if (cgErr != CGDisplayNoErr || numDisplays == 0) {
        ErrorF("Could not find CGDirectDisplayID(s) for X11 screen %d: %dx%d @ %d,%d.\n",
               index, dfb->width, dfb->height, dfb->x, dfb->y);
        return FALSE;
    }

    // This X11 screen covers all CoreGraphics displays we just found.
    // If there's more than one CG display, then video mirroring is on
    // or PseudoramiX is on.
    displayInfo->displayCount = allocatedDisplays;
    displayInfo->displayIDs = displays;

    return TRUE;
}


/*
 * AquaSetupScreen
 *  Setup the screen for rootless access.
 */
Bool
AquaSetupScreen(int index, ScreenPtr pScreen)
{
    // Add Aqua specific replacements for fb screen functions
    pScreen->PaintWindowBackground = AquaPaintWindow;
    pScreen->PaintWindowBorder = AquaPaintWindow;
/*
#ifdef RENDER
    {
        PictureScreenPtr ps = GetPictureScreen(pScreen);
        ps->Composite = AquaComposite;
    }
#endif / * RENDER */

    // Initialize generic rootless code
    return RootlessInit(pScreen, &aquaRootlessProcs);
}


/*
 * AquaInitInput
 */
void
AquaInitInput(int argc, char **argv)
{
    // Nothing here
}
