/*
 * Rootless implementation for Mac OS X Aqua environment
 */
/*
 * Copyright (c) 2001 Greg Parker. All Rights Reserved.
 * Copyright (c) 2002 Torrey T. Lyons. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE ABOVE LISTED COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name(s) of the above copyright
 * holders shall not be used in advertising or otherwise to promote the sale,
 * use or other dealings in this Software without prior written authorization.
 */
/* $XFree86: xc/programs/Xserver/hw/darwin/quartz/rootlessAquaImp.m,v 1.7 2003/04/30 23:15:40 torrey Exp $ */

#include "insidecheck.h"
#include "communication.h"
#include "rootlessAquaImp.h"
#include "fakeBoxRec.h"
#include "quartzCommon.h"
#include "aquaCommon.h"
#include "privatewinlist.h"
#include "pseudoramiX.h"
#import <Cocoa/Cocoa.h>
#include <ApplicationServices/ApplicationServices.h>
#import "XWindow.h"

// Only defined on OSX10.2, so make sure we set it ourselves if compiling on pre-OSX10.2
#ifndef NSAppKitVersionNumber10_1
#define NSAppKitVersionNumber10_1 620
#endif

extern XWindow *currentKeyWindow;
extern char *gResizeDeathBits;
extern int gLastDeathSize, gResizeDeathBPR;
static int waitForResize=0, waitOneSec=0;
int simulateClickRoot=0, gotOtherClick=0;

extern void ErrorF(const char *, ...);

XWindow *currentKeyWindow=nil;

void QuartzHandleResize(int privateid)
{
    void *rw = GetWinByPrivateID(privateid);
    
    if(!rw) {
        NSLog(@"QuartzHandleResize: failed to get window for PrivateID %d !!",
                                                                        privateid);
    } else {
        AquaWindowRec *winRec = AQUA_WINREC(rw);
//        [winRec->view setInDrawRect:YES];
        [winRec->view setWantFlipped:NO];
//        [winRec->window setAutodisplay:NO];
        [winRec->window setOneOfMine:YES];
        if (floor(NSAppKitVersionNumber) > NSAppKitVersionNumber10_0)
            [winRec->window setOpaque:YES];
        if(verboseConsole) NSLog(@"before mysetframe...");
//        [winRec->window setContentSize:[winRec->view getNewResizeFrame].size];
        [winRec->window mySetFrame:[winRec->view getNewResizeFrame] display:YES];
        if(verboseConsole) NSLog(@"after mysetframe...");
        [winRec->window setOneOfMine:NO];
//        [winRec->window setAutodisplay:NO];
//        [winRec->view setWantFlipped:NO];
        [winRec->view setInDrawRect:NO];
        /*{
            NSRect frame=[winRec->view frame];
            fakeBoxRec bounds = {0, 0, NSWidth(frame), NSHeight(frame)};
            AquaUpdateRects(rw, &bounds, 0);
        }*/
        //[winRec->view unsetWaitResizeFlag];
        waitForResize=0;
        mainThreadBusy=0;
    }
}

void QuartzHandleDocking(int hackid, int what)
{
    void *rw = GetWinByHackID(hackid);
    
    if(!rw) {
        NSLog(@"QuartzHandleDocking (%d): failed to get window for HackID %d !!",
                                                                        what,hackid);
    } else {
        AquaWindowRec *winRec = AQUA_WINREC(rw);
        if(what==0) {
            [winRec->window deminiaturize:nil];
            [winRec->view setMyMiniature:NO];
        } else {
            [winRec->view setMyMiniature:YES];
            [winRec->window miniaturize:nil];
        }
    }
}

/*
 * AquaDisplayCount
 *  Return the number of displays.
 *  Multihead note: When rootless mode uses PseudoramiX, the
 *  X server only sees one screen; only PseudoramiX itself knows
 *  about all of the screens.
 */
int AquaDisplayCount()
{
    int aquaNumScreens = [[NSScreen screens] count];

    if (noPseudoramiXExtension) {
        return aquaNumScreens;
    } else {
        return 1; // only PseudoramiX knows about the rest
    }
}


void AquaScreenInit(int index, int *x, int *y, int *width, int *height,
                    int *rowBytes, int *bps, int *spp, int *bpp)
{
    *spp = 3;
    *bps = CGDisplayBitsPerSample(kCGDirectMainDisplay);
    *bpp = CGDisplayBitsPerPixel(kCGDirectMainDisplay);

    if (noPseudoramiXExtension) {
        NSScreen *screen = [[NSScreen screens] objectAtIndex:index];
        NSRect frame = [screen frame];

        // set x, y so (0,0) is top left of main screen
        *x = NSMinX(frame);
        *y = NSHeight([[[NSScreen screens] objectAtIndex:0] frame]) - NSHeight(frame) -
            NSMinY(frame);

        *width =  NSWidth(frame);
        *height = NSHeight(frame);
        *rowBytes = (*width) * (*bpp) / 8;

        // Shift the usable part of main screen down to avoid the menu bar.
        if (NSEqualRects(frame, [[[NSScreen screens] objectAtIndex:0] frame])) {
            *y      += aquaMenuBarHeight;
            *height -= aquaMenuBarHeight;
        }

    } else {
        int i;
        NSRect unionRect = NSMakeRect(0, 0, 0, 0);
        NSArray *screens = [NSScreen screens];

        // Get the union of all screens (minus the menu bar on main screen)
        for (i = 0; i < [screens count]; i++) {
            NSScreen *screen = [screens objectAtIndex:i];
            NSRect frame = [screen frame];
            ErrorF("Found screen: %dx%d @ %d,%d  ",(int)frame.size.width,
            (int)frame.size.height,(int)frame.origin.x,(int)frame.origin.y);
            frame.origin.y = [[[NSScreen screens] objectAtIndex:0] frame].size.height -
                             frame.size.height - frame.origin.y;
            if (NSEqualRects([screen frame], [[[NSScreen screens] objectAtIndex:0] frame])) {
                frame.origin.y    += aquaMenuBarHeight;
                frame.size.height -= aquaMenuBarHeight;
            }
            ErrorF("- After adjust: %d\n",(int)frame.origin.y);
            unionRect = NSUnionRect(unionRect, frame);
        }

        // Use unionRect as the screen size for the X server.
        *x = unionRect.origin.x;
        *y = unionRect.origin.y;
        *width = unionRect.size.width;
        *height = unionRect.size.height;
        *rowBytes = (*width) * (*bpp) / 8;
        ErrorF("Global bounds: %dx%d @ %d,%d\n",*width,*height,*x,*y);

        // Tell PseudoramiX about the real screens.
        // InitOutput() will move the big screen to (0,0),
        // so compensate for that here.
        for (i = 0; i < [screens count]; i++) {
            NSScreen *screen = [screens objectAtIndex:i];
            NSRect frame = [screen frame];
            int j;

            // Skip this screen if it's a mirrored copy of an earlier screen.
            for (j = 0; j < i; j++) {
                if (NSEqualRects(frame, [[screens objectAtIndex:j] frame])) {
                    ErrorF("PseudoramiX screen %d is a mirror of screen %d.\n",
                           i, j);
                    break;
                }
            }
            if (j < i) continue; // this screen is a mirrored copy

            frame.origin.y = [[[NSScreen screens] objectAtIndex:0] frame].size.height -
                             frame.size.height - frame.origin.y;

            if (NSEqualRects([screen frame], [[[NSScreen screens] objectAtIndex:0] frame])) {
                frame.origin.y    += aquaMenuBarHeight;
                frame.size.height -= aquaMenuBarHeight;
            }

            ErrorF("PseudoramiX screen %d added: %dx%d @ (%d,%d).\n", i,
                   (int)frame.size.width, (int)frame.size.height,
                   (int)frame.origin.x, (int)frame.origin.y);

            frame.origin.x -= unionRect.origin.x;
            frame.origin.y -= unionRect.origin.y;

            ErrorF("PseudoramiX screen %d placed at X11 coordinate (%d,%d).\n",
                   i, (int)frame.origin.x, (int)frame.origin.y);

            PseudoramiXAddScreen(frame.origin.x, frame.origin.y,
                                 frame.size.width, frame.size.height);
        }
    }
}


void AquaIsInsideWindow(void *rw)
{
    AquaWindowRec *winRec = AQUA_WINREC(rw);

    got_inside_x11_window = 0;
    got_inside_firstclick = 0;

    if( !rw || simulateClickRoot!=0 ) return;
    // the mainThreadBusy check prevents a problem with cursor changes while the main thread
    // is waiting for drawRect to finish
    //if(!rw || mainThreadBusy ) return;

    if( winRec->hackID == -999 ) return;

    if( [winRec->view getInside] ) {
        got_inside_x11_window = 1;
        if( gotOtherClick!=0 || (winRec->window_state & OROB_WINDOW_FIRSTCLICK)!=0 )
            got_inside_firstclick = 1;
    }
    //NSLog(@"PrivateID %d : %d %d",winRec->privateID,got_inside_x11_window,got_stclick);
}

/*
 * AquaNewWindow
 *  Create a new on-screen window.
 *  Rootless windows must not autodisplay! Autodisplay can cause a deadlock.
 *    Event thread - autodisplay: locks view hierarchy, then window
 *    X Server thread - window resize: locks window, then view hierarchy
 *  Deadlock occurs if each thread gets one lock and waits for the other.
*/
void *AquaNewWindow(void *upperw, int x, int y, int w, int h, int isRoot)
{
    NSRect mainframe=[[[NSScreen screens] objectAtIndex:0] frame];
    AquaWindowRec *winRec = (AquaWindowRec *)malloc(sizeof(AquaWindowRec));
    NSRect frame = NSMakeRect(x, NSHeight(mainframe) -y -h, w, h);
                              //   +(int)mainframe.origin.y - y - h, w, h);
    XWindow *theWindow;
    XView *theView;
    unsigned int styleMask=NSMiniaturizableWindowMask;

    winRec->dirtyRgn = NULL;
    if(x!=-23451 || y!=-6789) {
        // default for non-'managed' windows is to accept first click
        winRec->window_state = OROB_WINDOW_NOTDRAWN | OROB_WINDOW_FIRSTCLICK;
        winRec->undrawn_x = x;
        winRec->undrawn_y = y;
        frame.origin.x = -10000.0;
        frame.origin.y = -10000.0;
    } /*else {
        //styleMask |= (NSClosableWindowMask|NSResizableWindowMask);
        styleMask |= NSTitledWindowMask;
    }*/

    // Create an NSWindow for the new X11 window
    theWindow = [[XWindow alloc] initWithContentRect:frame
                                  styleMask:styleMask
                                  backing:NSBackingStoreBuffered
                                  defer:YES];
    if (!theWindow) return NULL;

    if(x==-23451 && y==-6789) {
        char tmpOrobTitle[12];
        sprintf(tmpOrobTitle,"OroborOSX");
        if(verboseConsole)
            NSLog(@"initialising window title: %s",tmpOrobTitle);
        [theWindow setTitle:[NSString stringWithCString:tmpOrobTitle]];
    }

    {
        BOOL gotMatch;
        [theWindow setDepthLimit:NSBestDepth(NSDeviceWhiteColorSpace,
                                                5,16,NO,&gotMatch)];
//        if(!gotMatch) {
            NSLog(@"BPP %d, BPS: %d",
                NSBitsPerPixelFromDepth([theWindow depthLimit]),
                NSBitsPerSampleFromDepth([theWindow depthLimit]));
//        }
    }
    //[theWindow setBackgroundColor:[NSColor clearColor]];  // erase transparent
    [theWindow setBackgroundColor:[NSColor blackColor]];  // erase transparent
    [theWindow setAlphaValue:1.0];       // draw opaque
    [theWindow setOpaque:YES];           // changed when window is shaped or made translucent

    [theWindow useOptimizedDrawing:YES]; // Has no overlapping sub-views
    [theWindow setAutodisplay:NO];       // See comment above
    [theWindow disableFlushWindow];      // We do all the flushing manually
    [theWindow setHasShadow:!isRoot];    // All windows have shadows except root
//    [theWindow setHasShadow:NO];
    [theWindow setReleasedWhenClosed:YES]; // Default, but we want to be sure

    theView = [[XView alloc] initWithFrame:frame];
    [theWindow mySetContentView:theView];
    [theWindow setShowsResizeIndicator:NO];
    [theWindow setInitialFirstResponder:theView];
    [theWindow setDelegate:theWindow];
    [theView setWantFlipped:NO];
    [theWindow setOneOfMine:YES];
    [theView unsetWaitResizeFlag];
    [theView initLocks:0 bits:NO];
    [theView storeShapeRects:NULL count:-9999];  // -9999 inits to NULL

    [theWindow setAcceptsMouseMovedEvents:YES];

    winRec->window = theWindow;
    winRec->view = theView;
    winRec->hackID = 0;
    winRec->alpha_mask = 0xFF000000;
    winRec->transBG_mask = 0xFF000000;
    
    if(x==-23451 && y==-6789) {
        // default for 'managed' windows is to not accept first click
        // -will be informed through hackcmds if this should be changed
        winRec->window_state = 0;
        [theView setFirstClick:NO];
        [theWindow mySetStyleMask:NSTitledWindowMask];
    } else {
        // default for non-'managed' windows is to accept first click
        winRec->window_state = OROB_WINDOW_NOTDRAWN | OROB_WINDOW_FIRSTCLICK;
        winRec->undrawn_x = x;
        winRec->undrawn_y = y;
        [theView setFirstClick:YES];
        [theWindow mySetStyleMask:0];
    }
    
    if (upperw) {
        AquaWindowRec *upperRec = AQUA_WINREC(upperw);
        int uppernum = [upperRec->window windowNumber];
        [theWindow myOrderWindow:NSWindowBelow relativeTo:uppernum];
    } else {
        if (!isRoot) {
            [theWindow myOrderFront:nil];
            winRec->port = NULL;
        }
    }

    if (!isRoot) {
        winRec->rootGWorld = NULL;
        [theView myLockFocus];
        // Fill the window with white to make sure alpha channel is set
        NSEraseRect(frame);
        winRec->port = [theView qdPort];
        // CreateCGContextForPort(winRec->port, &winRec->context);
        winRec->context = [[NSGraphicsContext currentContext] graphicsPort];
//        SyncCGContextOriginWithPort( winRec->context, winRec->port );
//        CGContextTranslateCTM( winRec->context, 0, frame.size.height );

        //  Flip the y axis so that positive Y points down
        //  Note that this will cause text drawn with Core Graphics
        //  to draw upside down
//        CGContextScaleCTM( winRec->context, 1.0, -1.0 );
//        CGContextSetAlpha( winRec->context, 0.5 );
        [theView myUnlockFocus];
        winRec->privateID = (int)winRec->port;
    } else {
        // Allocate the offscreen graphics world for root window drawing
        GWorldPtr rootGWorld;
        Rect globalRect;

        SetRect(&globalRect, x, y, x+w, y+h);
        if (NewGWorld(&rootGWorld, 0, &globalRect, NULL, NULL, 0))
            return NULL;
        winRec->rootGWorld = rootGWorld;
        winRec->privateID = (int)rootGWorld;
        
        //verboseConsole = 1;
        if(!gResizeDeathBits) {
            gResizeDeathBPR = NSWidth([theView frame])*4;
            gLastDeathSize=NSHeight([theView frame])*gResizeDeathBPR;
            gResizeDeathBits=malloc(gLastDeathSize);
            NSLog(@"(Root) new deathbits: %#lx, size: %#lx, bpr: %d",
                                    gResizeDeathBits,gLastDeathSize,gResizeDeathBPR);
        } else if(gLastDeathSize<NSWidth([theView frame])*NSHeight([theView frame])*4) {
            free(gResizeDeathBits);
            gResizeDeathBPR = NSWidth([theView frame])*4;
            gLastDeathSize=NSHeight([theView frame])*gResizeDeathBPR;
            gResizeDeathBits=malloc(gLastDeathSize);
            NSLog(@"(Root) increased deathbits: %#lx, size: %#lx, bpr: %d",
                                    gResizeDeathBits,gLastDeathSize,gResizeDeathBPR);
        }
    }

            NSLog(@"BPP %d, BPS: %d",
                NSBitsPerPixelFromDepth([theWindow depthLimit]),
                NSBitsPerSampleFromDepth([theWindow depthLimit]));

    [theView setMyTrackingRect:0];
    [theView setAllIDs:winRec->privateID hackID:winRec->hackID];
    
    if(!isRoot)
        AddToPrivateWindowList(winRec);
//    else
//        startFlushTimer();
    
    [theWindow setOneOfMine:NO];

    return winRec;
}


void AquaDestroyWindow(void *rw)
{
    AquaWindowRec *winRec = AQUA_WINREC(rw);

    winRec->window_state = 0;

    if( (winRec->window_state & OROB_WINDOW_DOCKED)!=0 ) {
        if(verboseConsole)
            NSLog(@"PrivateID %d : in DestroyWindow for docked window",winRec->privateID);
    }

    RemoveFromPrivateWindowList(winRec);

    [winRec->view deleteTrackingRect];
    [winRec->window myOrderOut:nil];
    [winRec->window close];
    [winRec->view release];
    if (winRec->rootGWorld) {
        DisposeGWorld(winRec->rootGWorld);
    }
    free(rw);
}


void AquaHackCommand(void *rw, int x, int y)
{
    AquaWindowRec *winRec = AQUA_WINREC(rw);
    float z;
/*    int pos;
*/
//    NSLog(@"Got hack command (for %#lx): %d %d",rw,x,y);
    if(x==-23400) {
        z=y;
        z=z*0.01;
        [winRec->window setOpaque:NO];
        [winRec->window setAlphaValue:z];
        if(verboseConsole)
            NSLog(@"Set window translucency to %d",y);
    } else if(x==-23440) {
        if(verboseConsole)
            NSLog(@"Got HACKCMD_COMEINFRONT request from OroborOSX (%d)",y);
        if( y > 0 && ![NSApp isActive] )
            [NSApp activateIgnoringOtherApps:YES];
    } else if(x==-23434) {
        if(verboseConsole)
            NSLog(@"Setting whole screen mode: %d",y);
        if(y==0)
            End_Whole_Screen_Mode();
        else
            Start_Whole_Screen_Mode();
    } else if(x==-23401) {
        if(y==1) {
            if(winRec->hackID<=0) {
                NSLog(@"PrivateID %d : invalid window to Dock! (%d)",
                                        winRec->privateID,winRec->hackID);
            } else if((winRec->window_state & OROB_WINDOW_DOCKED)==0) {
                if(verboseConsole) NSLog(@"PrivateID %d : docking window (%d)",
                                        winRec->privateID,winRec->hackID);
                QuartzMessageMainThread(kQuartzWinToDock,&winRec->hackID,4);
                winRec->window_state |= OROB_WINDOW_DOCKED;
            } else
                NSLog(@"PrivateID %d : already docked!",winRec->privateID);
        } else if(y==0) {
            if(winRec->window_state & OROB_WINDOW_DOCKED) {
                winRec->window_state ^= OROB_WINDOW_DOCKED;
                [winRec->view setMyMiniature:NO];
                if(verboseConsole)
                    NSLog(@"PrivateID %d : window was undocked",winRec->privateID);
            }
        } else if(winRec->window_state & OROB_WINDOW_DOCKED) {
            if(verboseConsole) NSLog(@"PrivateID %d : undocking window (%d)",
                                        winRec->privateID,winRec->hackID);
            QuartzMessageMainThread(kQuartzWinToUndock,&winRec->hackID,4);
            winRec->window_state ^= OROB_WINDOW_DOCKED;
        } else
            NSLog(@"PrivateID %d : window not docked!",winRec->privateID);
    } else if(x==-23402) {
        if(y==-998) {
            winRec->window_state |= OROB_WINDOW_SHAPEINHIBIT;
            if(verboseConsole) NSLog(@"Inhibited reshape...");
        } else if(y==-999) {
            if(winRec->window_state & OROB_WINDOW_SHAPEINHIBIT)
                winRec->window_state ^= OROB_WINDOW_SHAPEINHIBIT;
            if(verboseConsole) NSLog(@"Allowed reshape...");
        } else if(y==-198) {
            if(verboseConsole) NSLog(@"switching on waitForShape flag...");
            [winRec->view setWaitForShape:YES];
            winRec->window_state |= OROB_WINDOW_NOFLUSH;
        } else if(y==-199) {
            if([winRec->view isInDrawRect]) {
                int cnt=6;
                if(verboseConsole) NSLog(@"switching off waitForShape flag...");
                [winRec->view setWaitForShape:NO];
                if(winRec->window_state & OROB_WINDOW_NOFLUSH)
                    winRec->window_state ^= OROB_WINDOW_NOFLUSH;
                usleep(10000);
                while([winRec->view isInDrawRect] && cnt>0) {
                    usleep(20000);
                    cnt--;
                }
                while([winRec->view isInDrawRect]) {
                    NSLog(@"Still waiting for drawRect to notice...");
                    usleep(20000);
                }
                //NSLog(@"drawRect has noticed...");
            } else {
                // This tends to happen with certain windows that have 'step' resizing
                // when the user moves the cursor but the window does not actually get
                // larger - it does not get into drawRect, but OroborOSX does not know,
                // so it still tells XDarwin to unset the flag...
                NSRect frame = [winRec->window frame];
                fakeBoxRec bounds = {0, 0, NSWidth(frame), NSHeight(frame)};
                //if(verboseConsole)
                //    NSLog(@"Got waitForShape unset, but not in drawRect???");
                [winRec->view setWaitForShape:NO];
                if(winRec->window_state & OROB_WINDOW_NOFLUSH)
                    winRec->window_state ^= OROB_WINDOW_NOFLUSH;
                AquaUpdateRects(rw, &bounds, 0);
                if (floor(NSAppKitVersionNumber) <= NSAppKitVersionNumber10_1) {
                    [winRec->window setHasShadow:NO];
                    [winRec->window setHasShadow:YES];
                } else
                    [winRec->window invalidateShadow];
            }
        }
/*        if(y==0)
            [winRec->window setDim:NO update:YES];
        else if(y==1)
            [winRec->window setDim:YES update:YES];
        else if(y==2)
            [winRec->window setDim:NO update:NO];
        else if(y==3)
            [winRec->window setDim:YES update:NO];
        else if(y==-1)
            [winRec->window setInhibitShadowChange:YES];
        else if(y==-2)
            [winRec->window setInhibitShadowChange:NO];
        else if(y==-3) {*/
        /*if(y==-3) {
            if(verboseConsole)
                NSLog(@"PrivateID %d : Redoing shadow (%d)",
                                    winRec->privateID,winRec->window_state);
            if(![winRec->window hasShadow])
                [winRec->window setHasShadow:YES];
            else if (floor(NSAppKitVersionNumber) <= NSAppKitVersionNumber10_1) {
                [winRec->window setHasShadow:NO];
                [winRec->window setHasShadow:YES];
            } else
                [winRec->window invalidateShadow];
            if((winRec->window_state & OROB_WINDOW_SHADOWINHIBIT)!=0)
                winRec->window_state -= OROB_WINDOW_SHADOWINHIBIT;
        } else if(y==-4) {
            winRec->window_state |= OROB_WINDOW_SHADOWINHIBIT;
            if(verboseConsole)
                NSLog(@"PrivateID %d : Inhibiting shadow change (%d)",
                                winRec->privateID,winRec->window_state);
        }*//* else if(y==-5) {
            if(![winRec->window hasShadow]) {
                if(verboseConsole)
                    NSLog(@"adding shadow for HackID %d",[winRec->window getMyHackID]);
                //[winRec->window enableFlushWindow];
                //[winRec->window setAutodisplay:YES];
                //[[winRec->window contentView] lockFocus];
                [winRec->window setHasShadow:YES];
                //[[NSGraphicsContext currentContext] flushGraphics];
                //[[winRec->window contentView] unlockFocus];
                //[winRec->window setAutodisplay:NO];
                //[winRec->window disableFlushWindow];
            }
        }
        // want to prevent annoying flicker when shadow changes (hmmm... doesn't work...)
        // since OroborOSX will immediately call set_client_shape right after this hack command,
        // and since the shape will not change, there's no need to change the shadow.
*/    } else if(x==-23403) {
        if(verboseConsole)
            NSLog(@"Bringing all windows in front");
        if (! [NSApp isActive])
            [NSApp activateIgnoringOtherApps:YES]; // make sure XDarwin is frontmost app
        [NSApp arrangeInFront:nil]; // make sure all windows come in front
/*    } else if(x==-23404) {
        if(y==0)
            [winRec->window mySetHasShadow:NO];
        else
            [winRec->window mySetHasShadow:YES];
*/    } else if(x==-23405) {
            if(y>=0 && y<256) {
                winRec->alpha_mask = y<<24;
                if(verboseConsole)
                    NSLog(@"PrivateID %d : Set alpha value to %d",winRec->privateID,y);
            } else
                NSLog(@"Can't set alpha value to %d!!!",y);
/*        [winRec->window setMyDrawAlpha:y];
    } else if(x==-23406) {
        [winRec->window setMyTranslucentAlpha:y];
    } else if(x==-23407) {
        [winRec->window setMyTranslucentRed:y];
    } else if(x==-23408) {
        [winRec->window setMyTranslucentGreen:y];
    } else if(x==-23409) {
        [winRec->window setMyTranslucentBlue:y];
*/    } else if(x==-23410) {
        [NSApplication hide:nil];
/*    } else if(x==-23411) {
        hitcheckptx = y;
        ErrorF("======== got X: %d ",hitcheckptx);
    } else if(x==-23412) {
        NSPoint hitCheckPt = NSMakePoint(hitcheckptx,NSHeight([[[NSScreen screens] objectAtIndex:0] frame]) - y);
        NSPoint aPt = [winRec->window convertScreenToBase:hitCheckPt];
        ErrorF(", got Y: %d \n",y);
        if([winRec->window willHitWindow:aPt])
            ErrorF("+++++++++++++++++++++ %d,%d will hit window\n",(int)(aPt.x),(int)(aPt.y));
        else
            ErrorF("--------------------- %d,%d will not hit window\n",(int)(aPt.x),(int)(aPt.y));*/
    } else if(x==-23413) {
        [winRec->window setMyTitlePart:y];
    } else if(x==-23414) {
        [winRec->window setMyTitlePart:y];
    } else if(x==-23415) {
        [winRec->window setMyTitlePart:y];
/*    } else if(x==-23416) {
        oroborosx_info_inhibit=y;*/
    } else if(x==-23417) {
        if(verboseConsole) NSLog(@"Setting can_come_in_front to %d",y);
        can_come_in_front=y;
    } else if(x==-23418) {
        [NSApp deactivate];
    } else if(x==-23419) {
        QuartzMessageMainThread(kQuartzSendFakeButton1,nil,0);
    } else if(x==-23426) {
        //NSLog(@"Sending kQuartzSendKeyPressEvent to main thread: %d",y);
        QuartzMessageMainThread(kQuartzSendKeyPressEvent,&y,sizeof(int));
    } else if(x==-23427) {
        //NSLog(@"Sending kQuartzSendKeyReleaseEvent to main thread: %d",y);
        QuartzMessageMainThread(kQuartzSendKeyReleaseEvent,&y,sizeof(int));
    } else if(x==-23424) {
        int key=0, side=0, keycode=0;
        if( y < 0 ) {
            y = -y;
            side = 1;
        }
        keycode = y/8;
        key = y - keycode*8;
        DarwinChangeModifierNXKeycode( key, side, keycode );
    } else if(x==-23425) {
        if(y<100) {
//            if(verboseConsole)
//                NSLog(@"------- Sending check mods message to main thread (%d)...",y);
            QuartzMessageMainThread(kQuartzCheckMods,&y,sizeof(int));
        }
/*    } else if(x==-23450) {
        if(y==0)
            [[winRec->window contentView] setInhibitFull:NO];
        else if(y==2) {
            [[winRec->window contentView] setInhibitFull:NO];
            mySpecialResize(winRec);
        } else if(y==3) {
            [[winRec->window contentView] setInhibitFull:YES];
            mySpecialResize(winRec);
        } else
            [[winRec->window contentView] setInhibitFull:YES];
    } else if(x==-23449) {
        if(y==0)
            [[winRec->window contentView] setInhibitRefresh:NO];
        else if(y==2) {
            [[winRec->window contentView] setInhibitRefresh:NO];
            mySpecialResize(winRec);
        } else if(y==3) {
            [[winRec->window contentView] setInhibitRefresh:YES];
            mySpecialResize(winRec);
        } else
            [[winRec->window contentView] setInhibitRefresh:YES];
*/    } else if(x==-23433) {
            if(y>0) {
                if((winRec->window_state & OROB_WINDOW_FIRSTCLICK)==0) {
                    if(verboseConsole)
                        NSLog(@"PrivateID %d : enabling firstclick",winRec->privateID);
                    winRec->window_state |= OROB_WINDOW_FIRSTCLICK;
                    [winRec->view setFirstClick:YES];
                } else if(verboseConsole)
                    NSLog(@"PrivateID %d : firstclick already enabled",winRec->privateID);
            } else if(winRec->window_state & OROB_WINDOW_FIRSTCLICK) {
                if(verboseConsole)
                    NSLog(@"PrivateID %d : disabling firstclick",winRec->privateID);
                winRec->window_state ^= OROB_WINDOW_FIRSTCLICK;
                [winRec->view setFirstClick:NO];
            }
    } else if(x==-23432) {
            if(y>0) {
                // must make sure window is offscreen when it's hidden
                AquaMoveWindow(rw, -10001, -10002);
                if(verboseConsole)
                    NSLog(@"Moved window to -10001,-10002...");
                winRec->window_state |= OROB_WINDOW_HIDDEN;
            } else if( winRec->window_state & OROB_WINDOW_HIDDEN ) {
                // we rely on window manager to put window back in correct place
                // when window is unhidden
                winRec->window_state ^= OROB_WINDOW_HIDDEN;
            }
            if(verboseConsole)
                NSLog(@"HACKID %d : set hidden state to %d",winRec->hackID);
    } else if(x==-23460) {
/*        if(y==0) {
            [winRec->view setWaitForShape:NO];
            NSLog(@"Private ID %d : Unset waitForShape flag",winRec->privateID);
        }*/
            if(y==200) {
                if( winRec->window_state & OROB_WINDOW_MOVEINHIBIT ) {
                    winRec->window_state ^= OROB_WINDOW_MOVEINHIBIT;
                    if(verboseConsole)
                        NSLog(@"Private ID %d : move inhibit off",winRec->privateID);
                } else
                    NSLog(@"Private ID %d : move inhibit already off",winRec->privateID);
            } else if(y==201) {
                if( winRec->window_state & OROB_WINDOW_MOVEINHIBIT )
                    NSLog(@"Private ID %d : move inhibit already on",winRec->privateID);
                else {
                    winRec->window_state |= OROB_WINDOW_MOVEINHIBIT;
                    if(verboseConsole)
                        NSLog(@"Private ID %d : move inhibit on",winRec->privateID);
                }
            } /*else
                [[winRec->window contentView] setInhibitLevel:y];
    } else if(x==-23461) {
            if(y>0) mySpecialResize(winRec);
*/    } else if(x==-23459) {
            verboseConsole=y;
            NSLog(@"Set verbose console mode to %d",y);
    } else if(x==-23490) {
        int dummy=0;
        if(y==0) {
            if(oroborosx_socket) {
                NSLog(@"Closing OroborOSX socket: %d",oroborosx_socket);
                dummy = close(oroborosx_socket);
                oroborosx_socket=0;
            }
        } else {
            char *socket_path=NULL;
            if(oroborosx_socket!=0) {
                NSLog(@"Closing OroborOSX socket and reconnecting... (%d)",oroborosx_socket);
                dummy = close(oroborosx_socket);
                oroborosx_socket=0;
            }
            socket_path = malloc(strlen(MAIN_SOCKET_PATH)+15);
            sprintf(socket_path,"%s.%s",MAIN_SOCKET_PATH,getenv("OROBOROSX_RANDOMID"));
            oroborosx_socket = init_socket_connect(socket_path,"XDarwin");
            free(socket_path);
            if( oroborosx_socket<=0 ) {
                NSLog(@"Failed to connect to OroborOSX socket: %d",oroborosx_socket);
                oroborosx_socket=0;
            } else
                NSLog(@"Created OroborOSX socket: %d",oroborosx_socket);
        }
/*    } else if(x==-23495) {
        if(gotQuitResult==10) {
            if(y==2) {
                NSLog(@"OroborOSX told me not to quit...");
                gotQuitResult=-1;
            } else {
                NSLog(@"OroborOSX told me it's OK to quit...");
                gotQuitResult=1;
            }
        } else if(y==0) {
            oroborosx_info_inhibit=1;
            can_come_in_front=-1;
            NSLog(@"OroborOSX told me to quit...");
            QuartzMessageMainThread(kQuartzOroborOSXQuitting);
        } else {
            NSLog(@"OroborOSX told me it's quitting...");
            oroborosx_info_inhibit=1;
            can_come_in_front=3;
        }
*/    } else if(x==-23498) {
        showing_prefs = y;
    } else if(x==-23499) {
        winRec->hackID=y;
        [winRec->view setAllIDs:0 hackID:y];
        if(verboseConsole)
            NSLog(@"PrivateID %d : set HackID to %d",winRec->privateID,y);
    }

}

void AquaMoveWindow(void *rw, int x, int y)
{
    AquaWindowRec *winRec = AQUA_WINREC(rw);
    NSRect mainframe=[[[NSScreen screens] objectAtIndex:0] frame];
    NSPoint topLeft = NSMakePoint(x, NSHeight(mainframe)-y);//+(int)mainframe.origin.y);
    NSRect frame = [winRec->window frame];
    
    if( winRec->hackID == -999 ) return;

    if( frame.origin.x == topLeft.x && frame.origin.y+frame.size.height == topLeft.y ) {
        if( verboseConsole )
            NSLog(@"PrivateID %d : Don't need to move window - already there",
                                                                    winRec->privateID);
        return;
    }
    
    if( (winRec->window_state & OROB_WINDOW_NOTDRAWN)!=0 ) {
        winRec->undrawn_x = x;
        winRec->undrawn_y = y;
        if(verboseConsole)
            NSLog(@"PrivateID %d : not moving undrawn window",winRec->privateID);
        return;
    }
    if( (winRec->window_state & OROB_WINDOW_DOCKED)!=0 ) {
        if(verboseConsole)
            NSLog(@"PrivateID %d : not moving docked window",winRec->privateID);
        return;
    }
    if( (winRec->window_state & OROB_WINDOW_HIDDEN)!=0 ) {
        if(verboseConsole)
            NSLog(@"PrivateID %d : not moving hidden window",winRec->privateID);
    	return;
    }
    if( (winRec->window_state & OROB_WINDOW_MOVEINHIBIT)!=0 ) {
        if(verboseConsole)
            NSLog(@"PrivateID %d : moving inhibited",winRec->privateID);
    	return;
    }
    if([winRec->view isInDrawRect]) {
       NSLog(@"In AquaMoveWindow while in drawRect!!!");
       usleep(15000);
       while([winRec->view isInDrawRect]) {
          NSLog(@"Waiting for drawRect in AquaMoveWindow");
          usleep(20000);
       }
    }
    //if(winRec->hackID>0) topLeft.y+=22;
    [winRec->window setOneOfMine:YES];
    [winRec->window mySetFrameTopLeftPoint:topLeft];
    [winRec->window setOneOfMine:NO];
}


// C routine to flush window from privatewinlist
void AquaFlushWindow(void *wr)
{
    if(wr) {
        AquaWindowRec *winRec = AQUA_WINREC(wr);
        
        if(winRec->window_state & OROB_WINDOW_NEEDSFLUSH) {
            winRec->window_state ^= OROB_WINDOW_NEEDSFLUSH;
            /*if(winRec->window_state & OROB_WINDOW_LOCKDIRTY)
                winRec->window_state ^= OROB_WINDOW_LOCKDIRTY;
            else {*/
            if (! winRec->rootGWorld) {
                [winRec->view myLockFocus];
                winRec->port = [winRec->view qdPort];
                if([winRec->view portIsLocked]) {
                    //NSLog(@"flushing window %#lx (port locked)",winRec->privateID);
                    QDFlushPortBuffer(winRec->port, winRec->dirtyRgn);
                } else {
                    //NSLog(@"flushing window %#lx (locking port)",winRec->privateID);
                    [winRec->view waitLockPortBits];
                    QDFlushPortBuffer(winRec->port, winRec->dirtyRgn);
                    [winRec->view myUnlockPortBits];
                }
                [winRec->view myUnlockFocus];
                SetEmptyRgn(winRec->dirtyRgn);
            }
            //}
        }
    }
}

/*
 * AquaStartResizeWindow
 *  Resize the on screen window.
 */
void AquaStartResizeWindow(void *rw, int x, int y, int w, int h, void *orgPixels)
{
    AquaWindowRec *winRec = AQUA_WINREC(rw);
    
    if(!rw) return;
    
    if( (winRec->window_state & OROB_WINDOW_DOCKED)!=0 ) {
        if(verboseConsole)
            NSLog(@"PrivateID %d : not starting resize of docked window",winRec->privateID);
        return;
    }

    
    if( (winRec->window_state & OROB_WINDOW_HIDDEN)!=0 ) {
//        if(verboseConsole)
//            NSLog(@"not moving hidden window %d",winRec->hackID);
        x=-w-10;
        y=-h-10;
    }
    
    {
    int *privateID;
    NSRect mainframe=[[[NSScreen screens] objectAtIndex:0] frame];
    NSRect frame = NSMakeRect(x, NSHeight(mainframe) -y -h, w, h);
                               // +(int)mainframe.origin.y - y - h, w, h);

    if( winRec->hackID == -999 ) {
        [winRec->window setOneOfMine:YES];
        [winRec->window mySetFrame:frame display:NO];
        [winRec->window setOneOfMine:NO];
        return;
    }
    
    privateID=malloc(sizeof(int));
//    winRec->window_state |= OROB_WINDOW_NOFLUSH;
    [winRec->view deleteTrackingRect];
    [winRec->view setWaitResizeFlag:frame oldPix:orgPixels];

    if( (winRec->window_state & OROB_WINDOW_NOTDRAWN)!=0 )
        winRec->window_state ^= OROB_WINDOW_NOTDRAWN;
    
    *privateID=winRec->privateID;
    
    waitForResize=1;
    QuartzMessageMainThread(kQuartzWinToResize,privateID,sizeof(int));
    mainThreadBusy=1;
    if([winRec->view getWaitResizeFlag]) {
        int cnt=10;
        while([winRec->view getWaitResizeFlag] && cnt>0) {
            usleep(10000);
            cnt--;
        }
        while([winRec->view getWaitResizeFlag]) {
            NSLog(@"Still waiting in X server thread for waitResizeFlag...");
            usleep(20000);
        }
    }
    if(verboseConsole) NSLog(@"Finished StartResizeWindow...");
    
    free(privateID);
    }
}


void AquaFinishResizeWindow(void *rw, int x, int y, int w, int h)
{
    AquaWindowRec *winRec = AQUA_WINREC(rw);

    if( winRec->hackID == -999 ) return;

    if( (winRec->window_state & OROB_WINDOW_DOCKED)!=0 ) {
        if(verboseConsole)
            NSLog(@"PrivateID %d : not finishing resize of docked window",
                                                        winRec->privateID);
        return;
    }

    // refresh everything? fixme yes for testing -not any more (will be done when reshaped...)
    [winRec->view createTrackingRect];
    if(verboseConsole) NSLog(@"Done FinishResizeWindow...");
}


/*
 * AquaUpdateRects
 *  Flush rectangular regions from a window's backing buffer
 *  (or PixMap for the root window) to the screen.
 */
void AquaUpdateRects(void *rw, fakeBoxRec *fakeRects, int count)
{
    AquaWindowRec *winRec = AQUA_WINREC(rw);
    fakeBoxRec *rects, *end;
//    static RgnHandle rgn = NULL;
    static RgnHandle box = NULL;

    if( winRec->hackID == -999 ) return;

    //return;

        if( [winRec->view isInDrawRect] ) {
            if( (winRec->alpha_mask & 0xFF000000)!=0xFF000000 ) {
                PixMapHandle pix;
                int rowBytes, depth, bpp;
                
                [winRec->view waitLockPortBits];
                pix=GetPortPixMap([winRec->view qdPort]);
                rowBytes = GetPixRowBytes(pix) & 0x3fff; // fixme is mask needed?
                depth = (**pix).cmpCount * (**pix).cmpSize; // fixme use GetPixDepth(pix)?
                bpp = (**pix).pixelSize;
                [winRec->view myUnlockPortBits];
                
                [winRec->view addAlphaRects:fakeRects count:count
                            alpha:winRec->alpha_mask BPR:rowBytes BPP:bpp];
            }
            if(verboseConsole) NSLog(@"Not flushing while in drawRect");
            return;
        }

    if( (winRec->window_state & OROB_WINDOW_DOCKED)!=0 ) {
        if(verboseConsole)
            NSLog(@"PrivateID %d : not flushing docked window",winRec->privateID);
        return;
    }

//    if (!rgn) rgn = NewRgn();
    if (!box) box = NewRgn();
    if( !winRec->dirtyRgn ) winRec->dirtyRgn = NewRgn();

    if (winRec->rootGWorld) {
        // FIXME: Draw from the root PixMap to the normally
        // invisible root window.
    } else {
        
        if(count==0) {
            Rect qdrect;
            qdrect.left = fakeRects->x1;
            qdrect.top = fakeRects->y1;
            qdrect.right = fakeRects->x2;
            qdrect.bottom = fakeRects->y2;

            if(winRec->window_state & OROB_WINDOW_NEEDSFLUSH)
                winRec->window_state ^= OROB_WINDOW_NEEDSFLUSH; 

            //winRec->window_state |= OROB_WINDOW_LOCKDIRTY;

            count=1;
            if(winRec->hackID<=0)
                NSLog(@"PrivateID %d : Updating all of non-managed window",winRec->privateID);
            else if(verboseConsole)
                NSLog(@"PrivateID %d : Updating all of managed window (%d)",
                                                        winRec->privateID,winRec->hackID);
            
//            if( !box ) box = NewRgn();
            
            RectRgn(box, &qdrect);
            QDFlushPortBuffer(winRec->port, box);
            
            SetEmptyRgn(box);
            SetEmptyRgn(winRec->dirtyRgn);
            
            if(winRec->window_state & OROB_WINDOW_NOTDRAWN) {
                winRec->window_state ^= OROB_WINDOW_NOTDRAWN; 
                AquaMoveWindow(rw, winRec->undrawn_x, winRec->undrawn_y);
            }

            //if(winRec->window_state & OROB_WINDOW_LOCKDIRTY) {
                //if(need_window_flush==0) need_window_flush = 1;
                //winRec->window_state ^= OROB_WINDOW_LOCKDIRTY;
            /*} else { // state must have changed during flush, but we can ignore it
                ;
            }*/

            //if( need_window_flush>1 ) PrivateFlushWindows();
            return;
        }

        if( (winRec->alpha_mask & 0xFF000000)!=0xFF000000 ) {
            PixMapHandle pix;
            int rowBytes, depth, bpp;

            [winRec->view myLockFocus];

            [winRec->view waitLockPortBits];
            pix=GetPortPixMap([winRec->view qdPort]);
            rowBytes = GetPixRowBytes(pix) & 0x3fff; // fixme is mask needed?
            depth = (**pix).cmpCount * (**pix).cmpSize; // fixme use GetPixDepth(pix)?
            bpp = (**pix).pixelSize;
            [winRec->view myUnlockPortBits];

            [winRec->view copyFromSrc:gResizeDeathBits srcBPR:rowBytes srcBPP:bpp
                            alpha:winRec->alpha_mask
                            rects:fakeRects count:count];

            [winRec->view myUnlockFocus];
        }

        if((winRec->window_state & OROB_WINDOW_NOFLUSH)==0) {
    
            //winRec->window_state |= OROB_WINDOW_LOCKDIRTY;
            winRec->window_state |= OROB_WINDOW_NEEDSFLUSH;

            for (rects = fakeRects, end = fakeRects+count; rects < end; rects++) {
                Rect qdrect;
                qdrect.left = rects->x1;
                qdrect.top = rects->y1;
                qdrect.right = rects->x2;
                qdrect.bottom = rects->y2;

                RectRgn(box, &qdrect);
                UnionRgn(winRec->dirtyRgn, box, winRec->dirtyRgn);
                //RectRgn(winRec->dirtyRgn, &qdrect);
            }

            //QDFlushPortBuffer(winRec->port, winRec->dirtyRgn);
            //SetEmptyRgn(winRec->dirtyRgn);

            /*if(winRec->window_state & OROB_WINDOW_LOCKDIRTY)
                winRec->window_state ^= OROB_WINDOW_LOCKDIRTY;*/
                if(need_window_flush==0) need_window_flush = 1;
            /*else { // state must have been changed during a flush attempt, so we do the flush
                QDFlushPortBuffer(winRec->port, winRec->dirtyRgn);
                SetEmptyRgn(winRec->dirtyRgn);
            }*/

        }
    }

    if( need_window_flush>1 ) PrivateFlushWindows();
        
//    SetEmptyRgn(rgn);
    SetEmptyRgn(box);
}


/*
 * AquaRestackWindow
 *  Change the window order. Put the window behind upperw or on top if
 *  upperw is NULL.
 */
void AquaRestackWindow(void *rw, void *upperw)
{
    AquaWindowRec *winRec = AQUA_WINREC(rw);

    if( (winRec->window_state & OROB_WINDOW_DOCKED)!=0 ) {
        if(verboseConsole)
            NSLog(@"PrivateID %d : not restacking docked window",winRec->privateID);
        return;
    }

    [winRec->window setOneOfMine:YES];
    if (upperw) {
        AquaWindowRec *upperRec = AQUA_WINREC(upperw);
        int uppernum = [upperRec->window windowNumber];
        [winRec->window myOrderWindow:NSWindowBelow relativeTo:uppernum];
    } else {
        [winRec->window myMakeKeyAndOrderFront:nil];
    }
    [winRec->window setOneOfMine:NO];
}


/*
 * AquaReshapeWindow
 *  Set the shape of a window. The rectangles are the areas that are
 *  not part of the new shape.
 */
void AquaReshapeWindow(void *rw, fakeBoxRec *fakeRects, int count)
{
    AquaWindowRec *winRec = AQUA_WINREC(rw);
    NSRect frame = [winRec->view frame];
    int winHeight = NSHeight(frame);

    if( winRec->hackID == -999 ) return;

/*    if( (winRec->window_state & OROB_WINDOW_SHAPEINHIBIT)!=0 ) {
        if(verboseConsole)
            NSLog(@"PrivateID %d : can't reshape - inhibited",winRec->privateID);
//        return;
    }*/

    if( (winRec->window_state & OROB_WINDOW_DOCKED)!=0 ) {
        if(verboseConsole)
            NSLog(@"PrivateID %d : not reshaping docked window",winRec->privateID);
        return;
    }


    if([winRec->view isInDrawRect]) {
        [winRec->view storeShapeRects:fakeRects count:count];
        //[winRec->view freeAlphaRects];
        return;
    }
	 
    if(verboseConsole) NSLog(@"starting AquaReshapeWindow...");

    [winRec->view myLockFocus];

    // If window is currently shaped we need to undo the previous shape
    if (![winRec->window isOpaque]) {
        [[NSColor whiteColor] set];
        NSRectFillUsingOperation(frame, NSCompositeDestinationAtop);
    }
    if(verboseConsole) NSLog(@"Finished undoing shape...");

    if (count > 0) {
        fakeBoxRec *rects, *end;
        fakeBoxRec bounds = {0, 0, NSWidth(frame), winHeight};

        // Make transparent if window is now shaped.
        [winRec->window setOpaque:NO];

        // Clear the areas outside the window shape
        [[NSColor clearColor] set];
        for (rects = fakeRects, end = fakeRects+count; rects < end; rects++) {
            int rectHeight = rects->y2 - rects->y1;
            NSRectFill( NSMakeRect(rects->x1, //rects->y1,
                                   winHeight - rects->y1 - rectHeight,
                                   rects->x2 - rects->x1, rectHeight) );
        }


        if((winRec->window_state & OROB_WINDOW_NOFLUSH)==0) {
            [[NSGraphicsContext currentContext] flushGraphics];

        // force update of window shadow (unless inhibited)
            if (floor(NSAppKitVersionNumber) <= NSAppKitVersionNumber10_1) {
                [winRec->window setHasShadow:NO];
                [winRec->window setHasShadow:YES];
            } else
                [winRec->window invalidateShadow];	// new for OSX10.2
        }
        
    } else {
        fakeBoxRec bounds = {0, 0, NSWidth(frame), winHeight};

        //[winRec->window setOpaque:NO];
        if(![winRec->view isInDrawRect])
           AquaUpdateRects(rw, &bounds, 0);
    }

//    [winRec->view setWaitForShape:NO];
    [winRec->view myUnlockFocus];
    
    if(verboseConsole) NSLog(@"Finished AquaReshapeWindow...");
}


/* AquaStartDrawing
 *  When a window's buffer is not being drawn to, the CoreGraphics
 *  window server may compress or move it. Call this routine
 *  to lock down the buffer during direct drawing. It returns
 *  a pointer to the backing buffer and its depth, etc.
 */
void AquaStartDrawing(void *rw, char **bits,
                      int *rowBytes, int *depth, int *bpp)
{
    AquaWindowRec *winRec = AQUA_WINREC(rw);
    PixMapHandle pix;

    if( (winRec->window_state & OROB_WINDOW_DOCKED)!=0 ) {
        if(verboseConsole)
            NSLog(@"PrivateID %d : in StartDrawing for docked window",winRec->privateID);
    }

    //if( winRec->hackID == -999 )
    //    NSLog(@"In StartDrawing for check_win");

    winRec->window_state |= OROB_WINDOW_DRAWING;

    if (! winRec->rootGWorld) {
        [winRec->view myLockFocus];
        winRec->port = [winRec->view qdPort];
        [winRec->view waitLockPortBits];
        //LockPortBits(winRec->port);
        [winRec->view myUnlockFocus];
        pix = GetPortPixMap(winRec->port);
    } else {
        pix = GetGWorldPixMap(winRec->rootGWorld);
        LockPixels(pix);
    }

    *rowBytes = GetPixRowBytes(pix) & 0x3fff; // fixme is mask needed?
    *depth = (**pix).cmpCount * (**pix).cmpSize; // fixme use GetPixDepth(pix)?
    *bpp = (**pix).pixelSize;

    if ( (winRec->alpha_mask & 0xFF000000)!=0xFF000000) {
        if(!gResizeDeathBits) {
            gResizeDeathBPR=*rowBytes;
            gLastDeathSize=gResizeDeathBPR*NSHeight([winRec->view frame]);
            gResizeDeathBits=malloc(gLastDeathSize);
            NSLog(@"New deathbits: %#lx, size: %#lx (bpr: %d)",
                        gResizeDeathBits,gLastDeathSize,gResizeDeathBPR);
        } else if(gLastDeathSize<*rowBytes*NSHeight([winRec->view frame])) {
            free(gResizeDeathBits);
            gResizeDeathBPR=*rowBytes;
            gLastDeathSize=*rowBytes*NSHeight([winRec->view frame]);
            gResizeDeathBits=malloc(gLastDeathSize);
            NSLog(@"Increased deathbits: %#lx, size: %#lx (bpr: %d)",
                        gResizeDeathBits,gLastDeathSize,gResizeDeathBPR);
        } else if(verboseConsole)
            NSLog(@"In AquaStartDrawing for alpha (src: %#lx, bpr: %d)",
                                                    gResizeDeathBits,*rowBytes);
        *bits = gResizeDeathBits;
        //*rowBytes = gResizeDeathBPR;
        //if([winRec->view isInDrawRect])
        //    *bits = GetPixBaseAddr(pix);
    } else {
        int offset = 0;
        //if( winRec->hackID > 0 ) offset = *rowBytes*22;
        *bits = GetPixBaseAddr(pix) + offset;
    }
}


/*
 * AquaStopDrawing
 *  When direct access to a window's buffer is no longer needed, this
 *  routine should be called to allow CoreGraphics to compress or
 *  move it.
 */
void AquaStopDrawing(void *rw)
{
    AquaWindowRec *winRec = AQUA_WINREC(rw);

    if(winRec->window_state & OROB_WINDOW_DRAWING)
        winRec->window_state ^= OROB_WINDOW_DRAWING;

    if( (winRec->window_state & OROB_WINDOW_DOCKED)!=0 ) {
        if(verboseConsole)
            NSLog(@"PrivateID %d : in StopDrawing docked window",winRec->privateID);
    }

    //if( winRec->hackID == -999 )
    //    NSLog(@"In StopDrawing for check_win");

    if (! winRec->rootGWorld) {
        if(verboseConsole && [winRec->view isInDrawRect])
            NSLog(@"In AquaStopDrawing while in drawRect");      
        //UnlockPortBits(winRec->port);
        [winRec->view myUnlockPortBits];
    } else {
        PixMapHandle pix = GetGWorldPixMap(winRec->rootGWorld);
        UnlockPixels(pix);
    }
}
